var PgBootstrapHelpers = function(options, f, version) {
    if (typeof options.reqVersion != "function") {
        options.reqVersion = function () { return true; }
    }

    // More options
    options.framework_prefix = 'bs' + version;
    options.container_selector = '.container,.container-fluid';
    options.grid_container_selector = 'row';
    options.style = '';
    options.skipVersion = options.skipVersion || function() { return true; };

    var pgFrameworkHelpers = new PgFrameworkHelpers('Bootstrap', options, f, version, PgBSMenuPointColumnSize);

    pgFrameworkHelpers.getCodeForLarge = function () {
        var code_large;
        switch(this.version) {
            case 5:
            case 4:
                code_large = '<pgdiv class="col-1 pinegrow-ui-helper"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
            case 3:
                code_large = '<pgdiv class="col-xs-1 pinegrow-ui-helper"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
        }

        return code_large;
    }

    pgFrameworkHelpers.getCodeForSmall = function () {
        var code_small;
        switch(this.version) {
            case 5:
            case 4:
                code_small = '<pgdiv class="col-1 pinegrow-ui-helper d-block d-sm-none" style="padding-left:0;padding-right:0;"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
            case 3:
                code_small = '<pgdiv class="col-xs-1 pinegrow-ui-helper visible-xs visible-xs-block" style="padding-left:0;padding-right:0;"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
        }

        return code_small;
    }

    pgFrameworkHelpers.getClassForSize = function(size, prefix, add) {
        if(add === null || add === undefined) {
            add = '';
        }  else {
            add = add + '';
        }
        return prefix + (size == options.size_for_all ? '' : '-' + size) + (add.length ? ('-' + add) : '');
    }

    pgFrameworkHelpers.getColumnSizes = function(pgel_or_node) {
        var re = new RegExp('^col(\\-(' + options.sizes.join('|') + '))?(\\-([0-9]+|auto))?');

        var map = {};

        var list = pgel_or_node instanceof pgParserNode ? pgel_or_node.getClasses() : pgel_or_node.classList;

        list.forEach(function(cls) {
            var m = cls.match(re);
            if(m) {
                var size = m[2] || options.size_for_all;
                var cols = m[4] || '';
                map[size] = cols;
            }
        })
        return map;
    }

    var delayed_tasks = new PgDelayedTasks();

    pgFrameworkHelpers.executeScriptInPageDelayed = function(page, code) {
        delayed_tasks.run('runCode' + code + page.uid, function() {
            pinegrow.executeScriptInPage(page, code);
        }, 1000)
    }

    pgFrameworkHelpers.makeFieldLabelForSize = function(label, size) {
        return '<span class="sc-label">' + label + '</span><span class="sc-device-size">' + size + '</span>';
    }

    pgFrameworkHelpers.getDevicesControlFactory = function(control_id, title, func) {

        var gh = pgFrameworkHelpers;
        //Returns function that creates the control
        var forSize = function(f) {
            options.sizes.forEach(function(size) {
                f(size, size == options.size_for_all ? '' : '-' + size, options.prefix + control_id + '.' + size);
            })
        }

        return function() {
            //will keep helper functions and vars private

            //Create control, needs a unique id
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //Register subfields.
            c.onDefine = function () {

                //These fields will not be shown when just registered. This is neccessary to get values.
                forSize(function(size, size_token, field_key) {
                    var field = func(size, size_token);
                    c.registerInputField(field_key, field);
                })
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $("<div/>", {class: 'sc-device-control'});
                $('<div class="sc-device-control-title">' + title + '</div>').appendTo($div);
                forSize(function(size, size_token, field_key) {
                    var field_def = func(size, size_token);
                    var field = c.showInputField($div, field_key, field_def);
                    field.$field.attr('data-device-size', size);
                })
                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function(values) {
                //return;
                var none_or_all = true;

                forSize(function(size, size_token, field_key) {
                    var $field = c.get$Field(field_key);
                    if(c.values[field_key]) {
                        $field.addClass('sc-has-value');
                        if(size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('sc-has-value');
                    }
                })
                if(none_or_all) {
                    c.$element.addClass('sc-none-or-all');
                } else {
                    c.$element.removeClass('sc-none-or-all');
                }
            }

            //Return the control
            return c;
        };
    }

    var open_device_sections_map = {};

    pgFrameworkHelpers.onSectionWithDevicesControlsShown = function(secdef, uiSection) {

        var gh = pgFrameworkHelpers;

        //var setActive = function($li, )
        var active = {}
        var no_active = true;

        options.sizes.forEach(function(size) {
            if(open_device_sections_map[secdef.section_key] && open_device_sections_map[secdef.section_key][size]) {
                active[size] = true;
            }
        })

        var fields = uiSection.$content.get(0).querySelectorAll('[data-device-size]');

        fields.forEach(function(field) {
            var size = field.getAttribute('data-device-size');
            if(!active[size]) {
                var has = field.classList.contains('sc-has-value');
                active[size] = has;
                no_active = no_active && !has;
            }
        })

        if(no_active) {
            active[options.size_for_all] = true;
        }

        fields.forEach(function(field) {
            var size = field.getAttribute('data-device-size');
            var display = field.getAttribute('data-device-display') || 'flex';

            if(active[size]) {
                pgShow(field, display);
            } else {
                pgHide(field);
            }
        })

        var html = '<ul class="sc-device-selector">';
        options.sizes.forEach(function(size) {
            html += '<li data-size="' + size + '" class="' + (active[size] ? 'sc-active' : '') + '">' + size + '</li>';
        })
        html += '</ul>';
        var $c = $(html);
        uiSection.$title.append($c);

        $c.on('click', 'li', function(e) {
            e.preventDefault();
            e.stopPropagation();

            var $li = $(this);
            var size = $li.attr('data-size');
            var a = $li.hasClass('sc-active');
            a = !a;

            active[size] = a;

            if(!a) {
                if($c.find('li.sc-active').length == 1) {
                    pinegrow.showQuickError('At least one device size must be active.');
                    return;
                }
            }

            if(!open_device_sections_map[secdef.section_key]) {
                open_device_sections_map[secdef.section_key] = {};
            }

            open_device_sections_map[secdef.section_key][size] = a;

            var $fields = uiSection.$content.find('[data-device-size="' + size + '"]');
            $fields.each(function(i, field) {
                var display = field.getAttribute('data-device-display') || 'flex';

                if (a) {
                    pg$Show($(field), display);
                    $li.addClass('sc-active');
                } else {
                    pg$Hide($(field));
                    $li.removeClass('sc-active');
                }
            });
            updateLabels();
        })

        var updateVisibility = function() {
            var page = pinegrow.getSelectedPage();
            fields.forEach(function(field) {
                var size = field.getAttribute('data-device-size');
                if(gh.isSizeVisible(page, size)) {
                    field.classList.remove('sc-size-not-visible');
                } else {
                    field.classList.add('sc-size-not-visible');
                }
            })

            $c.children().each(function(i, li) {
                var size = li.getAttribute('data-size');
                if(gh.isSizeVisible(page, size)) {
                    li.classList.remove('sc-size-not-visible');
                } else {
                    li.classList.add('sc-size-not-visible');
                }
            })
        }

        var updateLabels = function() {
            var num_shown = 0;
            var all_shown = false;

            options.sizes.forEach(function(size) {
                if(active[size]) {
                    num_shown++;
                    if(size == options.size_for_all) {
                        all_shown = true;
                    }
                }
            })

            var $controls = uiSection.$content.find('.sc-device-control');

            if(num_shown > 1) {
                $controls.removeClass('sc-single-size');
            } else {
                $controls.addClass('sc-single-size');
            }
            if(all_shown) {
                $controls.addClass('sc-for-all-shown');
            } else {
                $controls.removeClass('sc-for-all-shown');
            }
        }

        updateVisibility();
        updateLabels();

    }

    var directions = ['', 'x', 'y', 't', 'r', 'b', 'l'];
    var spacingVal = ['auto', 0, 1, 2, 3, 4, 5];
    var spacing = 5;


    var marginPaddingObjs = {};
    var marginPaddingActiveBtn = {};

    var rowGapMainObj = {
        directions: ['', 'x', 'y'],
        valuesCount: 5
    }
    var rowGapObjs = {};

    var positionObjs = {
        directions: ['top', 'start', 'end', 'bottom'],
        fields: {},
        activeBtn: null,
        skipCheckingChanges: false
    };

    var getDirectionName = function(direction) {
        if (options.reqVersion(5)) {
            if (direction == 'l') {
                direction = 's';
            } else if (direction == 'r') {
                direction = 'e';
            }
        }

        return direction;
    }

    var getNumFromMarginPaddingValue = function (value) {
        if (!value) return null;

        var arr = value.split('-');
        if (arr.length > 1) return arr[arr.length - 1];

        return '';
    }

    var getNewMarginPaddingValueFor = function (type, direction, size, field_key, values, valDirection, negative) {
        var value = values[field_key + valDirection];
        if (!value) return null;

        return type + direction + (size != 'xs' ? '-' + size : '') + '-' + (negative ? 'n' : '') + getNumFromMarginPaddingValue(value);
    }

    var updateAllMarginPaddingFields = function (pgel, type, size, field_key, values, skipDirection, negative) {
        for (var i = 3; i < directions.length; i++) {
            var direction = getDirectionName(directions[i]);

            pgel.removeClass(values[field_key + direction]);

            if (direction != skipDirection) {
                var field = marginPaddingObjs[size][type + direction];

                var currValue = getNewMarginPaddingValueFor(type, direction, size, field_key, values, '', negative);
                values[field_key + direction] = currValue;
                field.setValue(currValue);
            }
        }
    }


    var getClassNameForMarginAndPadding = function (type, direction, size, num, negative) {
        if (num == undefined || num == null) return null;
        return type + direction + (size != 'xs' ? '-' + size : '') + '-' + (negative ? 'n' : '') + num;
    }

    var setAllMarginPaddingValue = function (pgel, type, size, control_id, values, value) {
        var field_key = control_id + '.' + size + '.' + type;

        var col = pinegrow.getCollectionForElement(pgel);

        pinegrow.makeChanges(col, 'Set spacing', function() {

            col.forEachElement(function(pgel) {

                var allMarginPaddingUpdated = false;

                for (var i = 1; i < directions.length; i++) {
                    var direction = getDirectionName(directions[i]);

                    for (var j = 0; j < spacingVal.length; j++) {
                        var found = false, negative = false;
                        currClass = getClassNameForMarginAndPadding(type, direction, size, spacingVal[j], negative);
                        if (pgel.hasClass(currClass)) {
                            found = true;
                        }
                        else {
                            negative = true;
                            currClass = getClassNameForMarginAndPadding(type, direction, size, spacingVal[j], negative);
                            if (pgel.hasClass(currClass)) {
                                found = true;
                            }
                        }

                        if (found) {
                            pgel.removeClass(currClass);

                            var newClass = type + (size != 'xs' ? '-' + size : '') + '-' + (negative ? 'n' : '') + spacingVal[j];
                            pgel.addClass(newClass);
                            values[field_key] = newClass;

                            updateAllMarginPaddingFields(pgel, type, size, field_key, values, direction);

                            allMarginPaddingUpdated = true;
                            break;
                        }
                    }

                    if (allMarginPaddingUpdated) break;
                }
            })
        })
    }

    var isAllValuesEqual = function (pgel, type, size, control_id, direction, value) {
        var numValue = getNumFromMarginPaddingValue(value);

        var directionClassFound = true;
        for (var i = 3; i < directions.length; i++) {
            var dir = getDirectionName(directions[i]);

            if (dir != direction) {
                var directionClass = type + dir + (size != 'xs' ? '-' + size : '') + '-' + numValue;
                if (!pgel.hasClass(directionClass)) {
                    directionClassFound = false;
                    break;
                }
            }
        }

        return directionClassFound;
    }

    var toggleIconActivation = function (e, type, control) {
        var $target = $(e.target);
        var $icon = $target.closest('.change-all-icon');
        $icon.toggleClass('active');

        var isActive = $icon.hasClass('active');

        if (isActive) {
            var size = $icon.data('size');
            setAllMarginPaddingValue(control.obj, type, size, control.control_id, control.values);
        }
    }

    pgFrameworkHelpers.getDevicesMarginAndPaddingControlFactory = function(control_id, func) {
        //Returns function that creates the control
        var types = ['m', 'p'];

        var forSize = function(f) {
            options.sizes.forEach(function(size) {
                f(size, size == options.size_for_all ? '' : '-' + size, options.prefix + control_id + '.' + size);
            })
        }

        return function() {
            //will keep helper functions and vars private
            var updateFieldsValues = function (size, size_token, field_key) {
                var types = ['p', 'm'];
                var directions = ['t', 'r', 'l', 'b'];

                var getPureValue = function (type) {
                    var type_field_key = field_key + '.' + type;
                    var value_directions = ['', 'x', 'y'];
                    var values = [];

                    for (var i = 0; i < value_directions.length; i++) {
                        var dir = value_directions[i]
                        var value = c.values[type_field_key + dir];
                        if (value) {
                            var value_arr = value.split('-');

                            values.push({
                                direction: dir,
                                value: value_arr[value_arr.length - 1]
                            });
                        }
                    }

                    if (values.length > 0) {
                        return values;
                    }
                    return null;
                }

                var getNewValue = function (type, direction, value) {
                    return type + direction + (size_token ? '-' + size_token : '') + '-' + value;
                }

                var isAllowedToOverwrite = function (value_direction, type_direction) {
                    if (type_direction == '') {
                        return true;
                    }
                    else if (type_direction == 'x') {
                        if ([
                            getDirectionName('r'), getDirectionName('l')
                        ].indexOf(value_direction) > -1) return true;
                    }
                    else if (type_direction == 'y') {
                        if ([
                            getDirectionName('t'), getDirectionName('b')
                        ].indexOf(value_direction) > -1) return true;
                    }

                    return false;
                }

                for (var i = 0; i < types.length; i++) {
                    var type = types[i];
                    var values = getPureValue(type);
                    var $activeBtn = marginPaddingActiveBtn[size][type];

                    if (values && values.length > 0) {
                        for (var k = 0; k < values.length; k++) {
                            var value_obj = values[k];
                            var pure_value = value_obj.value;
                            var type_direction = value_obj.direction;

                            for (var j = 0; j < directions.length; j++) {
                                var direction = getDirectionName(directions[j]);
                                if (!isAllowedToOverwrite(direction, type_direction)) continue;

                                var type_dir_field_key = field_key + '.' + type + direction;

                                if (!c.values[type_dir_field_key]) {
                                    var pgField = marginPaddingObjs[size][type + direction];
                                    var newValue = getNewValue(type, direction, pure_value);

                                    c.values[type_dir_field_key] = newValue;
                                    pgField.setValue(newValue);
                                }
                            }

                            if (value_obj.direction == '') {
                                $activeBtn.addClass('active');
                                break;
                            }
                        }
                    }
                    else {
                        $activeBtn.removeClass('active');
                    }
                }
            }

            var hasValue = function (field_key) {
                var types = ['m', 'p'];
                var directions = ['t', 'l', 'r', 'b'];

                for (let i = 0; i < types.length; i++) {
                    var type = types[i];

                    for (let j = 0; j < directions.length; j++) {
                        var direction = getDirectionName(directions[j]);

                        var full_field_key = field_key + '.' + type + direction;
                        if (c.values[full_field_key]) return true;
                    }
                }

                return false;
            }

            //Create control, needs a unique id
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //Register subfields.
            c.onDefine = function () {
                forSize(function(size, size_token, field_key) {
                    for (var i = 0; i < types.length; i++) {
                        for (var j = 0; j < directions.length; j++) {
                            var direction = getDirectionName(directions[j]);

                            var field = func(size, types[i], direction, size_token);
                            var full_field_key = field_key + '.' + types[i] + direction;
                            c.registerInputField(full_field_key, field);
                        }
                    }
                });
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $('<div class="crsa-css-gui bs4-margin-and-padding"></div>');
                var _this = this;
                var type = '', direction = '';
                var field_def, field, full_field_key;

                forSize(function(size, size_token, field_key) {
                    var $container = $('<div class="margin-and-padding-container text-center" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="sc-device-size">' + size + '</span>').appendTo($container);

                    marginPaddingObjs[size] = {};
                    marginPaddingActiveBtn[size] = {};

                    // Margin top
                    type = 'm'; direction = getDirectionName('t');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-top crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Margin left
                    direction = getDirectionName('l');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-left crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    var $marginBorder = $('<div class="margin-border"></div>').appendTo($container);

                    var $marginIcon = $('<div class="change-all-margin-icon change-all-icon">\
                                                <i class="icon-same"></i>\
                                            </div>').appendTo($marginBorder);

                    marginPaddingActiveBtn[size]['m'] = $marginIcon;
                    $marginIcon.data('size', size);
                    $marginIcon.on('click', function (e) {
                        toggleIconActivation(e, 'm', _this);
                    });

                    var $paddingContainer = $('<div class="padding-container"></div>').appendTo($marginBorder);

                    // Padding top
                    type = 'p'; direction = getDirectionName('t');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-top crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Padding left
                    direction = getDirectionName('l');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-left crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    var $paddingIcon = $('<div class="padding-border">\
                            <div class="change-all-padding-icon change-all-icon"><i class="icon-same"></i></div>\
                        </div>').appendTo($paddingContainer).find('> .change-all-icon');

                    marginPaddingActiveBtn[size]['p'] = $paddingIcon;
                    $paddingIcon.data('size', size);
                    $paddingIcon.on('click', function (e) {
                        toggleIconActivation(e, 'p', _this);
                    });

                    // Padding right
                    direction = getDirectionName('r');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-right crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Padding bottom
                    direction = getDirectionName('b');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-bottom crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Margin right
                    type = 'm'; direction = getDirectionName('r');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-right crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;


                    // Margin bottom
                    direction = getDirectionName('b');
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-bottom crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;
                });

                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function(values) {
                //return;
                var none_or_all = true;

                forSize(function(size, size_token, field_key) {
                    // Check padding and margin
                    updateFieldsValues(size, size_token, field_key);

                    var $field = c.$element.find('> [data-device-size="' + size + '"]');
                    if (hasValue(field_key)) {
                        $field.addClass('sc-has-value');
                        if (size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('sc-has-value');
                    }
                })
                if(none_or_all) {
                    c.$element.addClass('sc-none-or-all');
                } else {
                    c.$element.removeClass('sc-none-or-all');
                }
            }

            //Return the control
            return c;
        };
    }

    pgFrameworkHelpers.getSpacingControl = function() {
        return pgFrameworkHelpers.getDevicesMarginAndPaddingControlFactory('margin.padding', function(size, type, direction_token, size_token) {
            var getAllOptions = function () {
                var base = type + direction_token + size_token + '-';
                var optionsValue = [];

                if (options.reqVersion(4.2) /* && options.skipVersion(5) */) {
                    if (base.startsWith('m')) {
                        for (var i = 1; i <= spacing; i++) {
                            var ni = -1 * i;
                            optionsValue.push({ 'key' : base + 'n' + i, 'name' :  ni, num: ni });
                        }
                    }
                }

                optionsValue.push({ 'key' : base + 'auto', 'name' : 'Auto' });

                for (var i = 0; i <= spacing; i++) {
                    optionsValue.push({ 'key' : base + i, 'name' : i, num: i });
                }

                return optionsValue;
            }

            var field = {
                type : 'select',
                name : null,
                placeholder : direction_token,
                show_empty : true,
                options : getAllOptions,
                action : 'custom',
                can_make_change_action: 'apply_class',
                draggable_list : true,
                async_options_function: true,
                fill_list_on_open: true,
                preview_class: true,
                on_preview_class: function($el, fn, value, values) {
                    var old_value = field.get_value($el, fn, values, field);
                    field.set_value($el, value, values, old_value, 'input', field);
                },
                more_prop: {
                    size: size,
                    type: type,
                    direction: direction_token
                },
                get_value: function(pgel, fn, values, fv) {
                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (pgel.hasClass(optionsValue[i].key)) return optionsValue[i].key;
                    }

                    var size      = fv.more_prop.size,
                        type      = fv.more_prop.type,
                        direction = fv.more_prop.direction;

                    if (direction && direction != 'x' && direction != 'y') {
                        var field_key = options.prefix + 'margin.padding.' + size + '.' + type;
                        if (values[field_key]) {
                            var currValue = getNewMarginPaddingValueFor(type, direction, size, field_key, values, '');
                            values[field_key + direction] = currValue;
                            return currValue
                        }

                        if (direction == getDirectionName('r') || direction == getDirectionName('l')) {
                            if (values[field_key + 'x']) {
                                var currValue = getNewMarginPaddingValueFor(type, direction, size, field_key, values, 'x');
                                values[field_key + direction] = currValue;
                                return currValue;
                            }
                        }
                        else if (direction == getDirectionName('t') || direction == getDirectionName('b')) {
                            if (values[field_key + 'y']) {
                                var currValue = getNewMarginPaddingValueFor(type, direction, size, field_key, values, 'y');
                                values[field_key + direction] = currValue;
                                return currValue;
                            }
                        }
                    }

                    return '';
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var optionsValue = fdef.options();

                    $.each(optionsValue, function (i, opt) {
                        if (opt.key == value) return true;
                        if (pgel.hasClass(opt.key)) {
                            pgel.removeClass(opt.key);
                        }
                    });

                    //pgel.removeClass(oldValue);

                    var size       = fdef.more_prop.size,
                        type       = fdef.more_prop.type,
                        direction  = fdef.more_prop.direction;

                    var control_id = options.prefix + 'margin.padding.';

                    var $icon = marginPaddingActiveBtn[size][type];
                    var isActive = $icon.hasClass('active');
                    if (isActive || isAllValuesEqual(pgel, type, size, control_id, direction, value)) {
                        var field_key = control_id + size + '.' + type;
                        var currClass = getClassNameForMarginAndPadding(type, '', size, getNumFromMarginPaddingValue(oldValue));
                        var newClass  = getClassNameForMarginAndPadding(type, '', size, getNumFromMarginPaddingValue(value));
                        pgel.removeClass(currClass);

                        for (var i = 3; i < directions.length; i++) {
                            var dir = getDirectionName(directions[i]);

                            var colNum;
                            if (dir == direction) {
                                colNum = getNumFromMarginPaddingValue(oldValue);
                            }
                            else {
                                colNum = getNumFromMarginPaddingValue(value);
                            }

                            var directionClass = type + dir + (size != 'xs' ? '-' + size : '') + '-' + colNum;
                            pgel.removeClass(directionClass);
                        }

                        pgel.addClass(newClass);
                        values[field_key] = newClass;

                        updateAllMarginPaddingFields(pgel, type, size, field_key, values, direction);
                    }
                    else {
                        var field_key = control_id + size + '.' + type
                        var oldClass = values[field_key];

                        values[field_key] = '';
                        pgel.removeClass(oldClass);

                        for (var i = 3; i < directions.length; i++) {
                            var dir = getDirectionName(directions[i]);

                            if (dir != direction) {
                                pgel.addClass(values[field_key + dir]);
                            }
                        }
                        pgel.addClass(value);
                    }

                    return value;
                }
            }

            return field;
        })
    }

    var getRowGapClassName = function(direction, size, value) {
        return 'g' + direction + size + '-' + value;
    }

    var getRowGapValueFor = function(className, direction, size) {
        if (className) {
            return className.replace('g' + direction + size + '-', '');
        }

        return null;
    }

    pgFrameworkHelpers.getDevicesRowGapControlFactory = function(control_id, func) {
        //Returns function that creates the control
        var directions = rowGapMainObj.directions;
        var rowGapValue = rowGapMainObj.valuesCount;

        var forSize = function(f) {
            options.sizes.forEach(function(size) {
                f(size, size == options.size_for_all ? '' : '-' + size, options.prefix + control_id + '.' + size);
            })
        }

        var hasValue = function(pgel, size_token) {
            for (var i = 0; i < directions.length; i++) {
                var dir = directions[i];

                for (var j = 0; j <= rowGapValue; j++) {
                    if ( pgel.hasClass( getRowGapClassName(dir, size_token, j) )) {
                        return true;
                    }
                }
            }

            return false;
        }

        return function() {
            //Create control, needs a unique id
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //Register subfields.
            c.onDefine = function () {
                forSize(function(size, size_token, field_key) {
                    for (var j = 0; j < directions.length; j++) {
                        var field = func(size, directions[j], size_token);
                        var full_field_key = field_key + '.' + directions[j];
                        c.registerInputField(full_field_key, field);
                    }
                });
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $('<div class="bs5-row-gap sc-device-control sc-single-size"></div>');
                var $label = $('<div class="sc-device-control-title">Gutter</div>').appendTo($div);
                var direction = '';
                var field_def, field, full_field_key;

                forSize(function(size, size_token, field_key) {
                    var $container = $('<div class="row-gap-container" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="sc-label">Gutter</span><span class="sc-device-size">' + size + '</span>').appendTo($container);


                    var $row = $('<div class="main-row-gap-container"></div>').appendTo($container);
                    rowGapObjs[size] = {};

                    var $fieldsContainer = $('<div class="row-gap-fields-container"></div>').appendTo($row);

                    // x direction
                    var $xCol = $('<div></div>').appendTo($fieldsContainer);

                    direction = 'x';
                    field_def = func(size, direction, size_token);
                    full_field_key = field_key + '.' + direction;
                    field = c.showInputField($xCol, full_field_key, field_def);
                    field.$field.addClass('');
                    rowGapObjs[size][direction] = field;

                    // y direction
                    var $yCol = $('<div></div>').appendTo($fieldsContainer);

                    direction = 'y';
                    field_def = func(size, direction, size_token);
                    full_field_key = field_key + '.' + direction;
                    field = c.showInputField($yCol, full_field_key, field_def);
                    field.$field.addClass('');
                    rowGapObjs[size][direction] = field;
                });

                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function(values) {
                //return;
                var none_or_all = true;

                forSize(function(size, size_token, field_key) {
                    var $field = c.$element.find('> [data-device-size="' + size + '"]');
                    if (hasValue(c.obj, size_token)) {
                        $field.addClass('sc-has-value');
                        if (size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('sc-has-value');
                    }
                })
                if(none_or_all) {
                    c.$element.addClass('sc-none-or-all');
                } else {
                    c.$element.removeClass('sc-none-or-all');
                }
            }

            //Return the control
            return c;
        };
    }

    var getPositionClassName = function(direction, value) {
        return direction + '-' + value;
    }

    var getPositionValueFor = function(direction, value) {
        if (value) {
            return value.replace(direction + '-', '');
        }

        return null;
    }

    pgFrameworkHelpers.getGappingControl = function() {
        return pgFrameworkHelpers.getDevicesRowGapControlFactory('row.gutter', function(size, direction_token, size_token){
            var valuesCount = rowGapMainObj.valuesCount;

            var getAllOptions = function () {
                var optionsValue = [];

                for (var i = 0; i <= valuesCount; i++) {
                    optionsValue.push({
                        'key': getRowGapClassName(direction_token, size_token, i),
                        'name': i,
                        'num': i
                    });
                }

                return optionsValue;
            }

            var field = {
                type : 'select',
                name : null,
                placeholder : direction_token,
                show_empty : true,
                options : getAllOptions,
                action : 'custom',
                async_options_function: true,
                more_prop: {
                    size: size,
                    direction: direction_token
                },
                get_value: function(pgel, fn, values, fv) {
                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (pgel.hasClass(optionsValue[i].key)) return optionsValue[i].key;
                    }

                    var size      = fv.more_prop.size,
                        direction = fv.more_prop.direction;

                    var size_token = size == options.size_for_all ? '' : '-' + size;

                    for (var i = 0; i <= valuesCount; i++) {
                        var className = getRowGapClassName('', size_token, i);
                        if (pgel.hasClass(className)) {
                            return getRowGapClassName(direction, size_token, i);
                            break;
                        }
                    }

                    return null;
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var optionsValue = fdef.options();

                    $.each(optionsValue, function (i, opt) {
                        if (opt.key == value) return true;
                        if (pgel.hasClass(opt.key)) {
                            pgel.removeClass(opt.key);
                        }
                    });

                    var size       = fdef.more_prop.size,
                        direction  = fdef.more_prop.direction,
                        opDirection = direction === 'x' ? 'y' : 'x';

                    var size_token = size == options.size_for_all ? '' : '-' + size;

                    var notAdded = true;

                    // if two directions set
                    for (var i = 0; i <= valuesCount; i++) {
                        var mainClassName = getRowGapClassName('', size_token, i);
                        if (pgel.hasClass(mainClassName)) {
                            pgel.removeClass(mainClassName);

                            // remove value from opposite side
                            pgel.addClass( getRowGapClassName(opDirection, size_token, i) );
                        }
                    }

                    // Check if the opposite direction has same value
                    var currentValue = getRowGapValueFor(value, direction, size_token);
                    var opDirClassName = getRowGapClassName(opDirection, size_token, currentValue);

                    if (pgel.hasClass(opDirClassName)) {
                        pgel.removeClass(opDirClassName);
                        pgel.addClass(getRowGapClassName('', size_token, currentValue));
                        notAdded = false
                    }

                    if (notAdded) {
                        pgel.addClass(value);
                    }

                    return value;
                }
            }

            return field;
        });
    }

    pgFrameworkHelpers.getArrangeControlFactory = function(control_id, fun) {
        return function() {

            var directions = positionObjs.directions;
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            c.onDefine = function () {
                for (var i = 0; i < directions.length; i++) {
                    var direction = directions[i];
                    var field = options.prefix + '.' + direction;
                    this.registerInputField(field, fun(direction));
                }
            }

            c.onShow = function () {
                var $container = $('<div class="crsa-css-gui bs5-position"></div>');

                var $positionCont = $('<div class="position-container text-center"></div>').appendTo($container);

                for (var i = 0; i < directions.length; i++) {
                    var direction = directions[i];
                    var field_key = options.prefix + '.' + direction;
                    var field = this.showInputField($positionCont, field_key, fun(direction));
                    positionObjs.fields[direction] = field;

                    // Set dir
                    var dir = direction;
                    if (dir == 'start') {
                        dir = 'left';
                    } else if (dir == 'end') {
                        dir = 'right';
                    }
                    field.$field.addClass('crsa-field-' + dir + ' crsa-small-input crsa-align-center-input crsa-inline-container');
                }

                var $positionOverallBorder = $('<div class="position-overall-border"></div>').appendTo($positionCont);

                var $positoinPaddingTop = $('<div class="position-padding-top"></div>').appendTo($positionOverallBorder);
                var $positoinPaddingTop = $('<div class="position-padding-left"></div>').appendTo($positionOverallBorder);
                var $positionBorder = $('<div class="position-border"></div>').appendTo($positionOverallBorder);
                var $positionPaddingRight = $('<div class="position-padding-right"></div>').appendTo($positionOverallBorder);
                var $positionPaddingRight = $('<div class="position-padding-bottom"></div>').appendTo($positionOverallBorder);

                var $equalIcon = positionObjs.activeBtn = $('<div class="change-all-icon"><i class="icon-same"></i></div>').appendTo($positionBorder);
                $equalIcon.on('click', function(e) {
                    var target = e.target;
                    var $changeIcon = $(target).closest('.change-all-icon');
                    $changeIcon.toggleClass('active');
                    var isActive = $changeIcon.hasClass('active');

                    if (isActive) {
                        for (var i = 0; i < directions.length; i++) {
                            var dir = directions[i];
                            var value = positionObjs.fields[dir].getValue();
                            var valueNum = getPositionValueFor(dir, value);

                            if (value) {
                                positionObjs.skipCheckingChanges = true;
                                for (var j = 0; j < directions.length; j++) {
                                    var currDir = directions[j];

                                    var currValue = positionObjs.fields[dir].getValue();
                                    var newValue = getPositionClassName(currDir, valueNum);
                                    if (currValue != newValue) {
                                        positionObjs.fields[currDir].setValue(newValue, true);
                                    }
                                }

                                positionObjs.skipCheckingChanges = false;
                                break;
                            }
                        }
                    }
                });

                return $container;
            }

            c.onSetValues = function () {
            }

            return c
        }
    }

    pgFrameworkHelpers.getArrangeControl = function() {
        return pgFrameworkHelpers.getArrangeControlFactory('arrange', function(direction) {
            var base = direction + '-';
            var field = {
                type: 'select',
                name: null,
                action: 'apply_class',
                show_empty: true,
                preview_class: true,
                options: [
                    { key: base + '0',   name: '0' },
                    { key: base + '50',  name: '50' },
                    { key: base + '100', name: '100' },
                ],
                more_prop: {
                    direction: direction,
                },
                on_changed: function(pgel, prop, value, oldValue, fieldDef) {
                    if (positionObjs.activeBtn.hasClass('active') && !positionObjs.skipCheckingChanges) {
                        positionObjs.skipCheckingChanges = true;
                        var directions = positionObjs.directions;
                        var direction = fieldDef.more_prop.direction;

                        for (var i = 0; i < directions.length; i++) {
                            var dir = directions[i];
                            var valueNum = getPositionValueFor(direction, value);

                            var currValue = positionObjs.fields[dir].getValue();
                            var newValue = getPositionClassName(dir, valueNum);
                            if (currValue != newValue) {
                                positionObjs.fields[dir].setValue(newValue, true);
                            }
                        }
                        positionObjs.skipCheckingChanges = false;
                    }
                }
            }

            return field;
        })
    }

    pgFrameworkHelpers.init();

    var getAllColorsFromPgel = function(pgel, what_list, all_base_colors) {
        var list = [];

        var has = {};
        var all_base_colors = ['white', 'light', 'muted', 'dark', 'primary', 'secondary', 'success', 'info', 'warning', 'danger'];

        pgel.walkSelfAndChildren(function(node) {
            if(node.isElement) {
                var classes = node.getClasses().slice(0);
                classes.forEach(function(cls) {
                    for(var n = 0; n < what_list.length; n++) {
                        var what = what_list[n];
                        for (var i = 0; i < all_base_colors.length; i++) {
                            if (cls.indexOf(what + '-' + all_base_colors[i]) >= 0 && list.indexOf(all_base_colors[i]) < 0 && !has[all_base_colors[i]]) {
                                list.push(all_base_colors[i]);
                                has[all_base_colors[i]] = true;
                                break;
                            }
                        }
                    }
                })
            }
            return true;
        })
        return list;
    }

    var replaceColor = function(col, what_list, from_color, to_color) {
        /*
        what: all,text,background,border
        from: *, red, blue...
        to: red, blue
         */


        pinegrow.makeChanges(col, 'Swap colors', function() {
            col.forEachElement(function(pgel) {
                what_list.forEach(function(what) {
                    var search = what + '-' + from_color;
                    var replace = what + '-' + to_color;

                    pgel.walkSelfAndChildren(function(node) {
                        if(node.isElement) {
                            var classes = node.getClasses();
                            classes.forEach(function(cls) {

                                if(cls.indexOf(search) >= 0) {
                                    node.removeClass(cls);
                                    cls = cls.replace(search, replace);
                                    node.addClass(cls);
                                }
                            })
                        }
                        return true;
                    })
                })
            })
        })

        pinegrow.selectedElements.reselect();
    }

    var getMenuActionsForColorChange = function(pgel) {
        var target_col = new PgCollection(pgel);

        var getSubmenu = function(what, what_list) {

            if(!what_list) what_list = [what];

            var getColorHtml = function(c, col) {
                /*
                var col = c;
                if(simple_colors.indexOf(c) < 0) {
                    col += '-500';
                }*/
                var bck = col;//getTailwindColor(col);

                var tc = chroma.contrast(bck, '#ffffff') < 4.5 ? tc = '#000' : '#fff';
                return `<span class="tw-dropdown-menu-color" style="background:${bck};color:${tc};">${c}</span>`;
            }

            var submenu = [
                {
                    type: 'header',
                    label: 'Swap color...'
                }
            ];

            var base_colors = [];
            var base_colors_values = {};

            var color_options = f.bsGetBackgroundColorOptions('');

            var page = pinegrow.getSelectedPage();

            var var_prefix = version === 5 ? '--bs-' : '--';

            color_options.forEach(function(i) {
                var a = i.key.replace('bg-', '');
                base_colors.push(a);
                base_colors_values[a] = page.getCSSVariableValue(var_prefix + a) || f.bsGetBootstrapColor(a);
            })

            var list = getAllColorsFromPgel(pgel, what_list, base_colors);

            list.forEach(function (from_c) {
                var sub = [
                    {
                        type: 'header',
                        label: '...with color'
                    }
                ]

                base_colors.forEach(function (color) {
                    sub.push({
                        label: getColorHtml(color, base_colors_values[color]),
                        action: function () {
                            replaceColor(target_col, what_list, from_c, color);
                        }
                    })
                })

                submenu.push({
                    label: getColorHtml(from_c, base_colors_values[from_c]),
                    submenu: sub
                })
            })
            return submenu;
        }


        var a = [
            {
                type: 'header',
                label: 'Swap colors'
            },
            {
                label: 'Text',
                submenu: getSubmenu('text', ['text', 'link'])
            },
            {
                label: 'Background',
                submenu: getSubmenu('bg', ['bg', 'btn', 'alert', 'table', 'btn-outline', 'badge', 'list-group-item'])
            },
            {
                label: 'Borders',
                submenu: getSubmenu('border', ['border'])
            },
            {
                label: 'Background & Borders',
                submenu: getSubmenu('bg', ['bg', 'border', 'btn', 'alert', 'table', 'btn-outline', 'badge', 'list-group-item'])
            },
            {
                label: 'All',
                submenu: getSubmenu('all', ['text', 'link', 'bg', 'border', 'btn', 'alert', 'table', 'btn-outline', 'badge', 'list-group-item'])
            }
        ]
        return a;
    }

    f.on_get_menu_commands = function(page, pgel, commands, $el) {

        commands.push({
            html: '<span class="tw-color-swap"><span class="span1"></span><span class="span2"></span></span>',
            tooltip: 'Swap colors',
            action: function (pgel, $menu_el) {

                var a = getMenuActionsForColorChange(pgel);

                var menu = new CrsaContextMenu(a, $menu_el);
                menu.showForElement($menu_el, null, 0);
            }
        })
    }

    return pgFrameworkHelpers;
}
