var PgFrameworkHelpers = function(fname, options, f, version, PgMenuPointColumnSizeClass) {

    var _this = this;

    options.showGridInMenu = (options.showGridInMenu === undefined || options.showGridInMenu === null) ? true : options.showGridInMenu;
    options.showColumnInMenu = (options.showColumnInMenu === undefined || options.showColumnInMenu === null) ? true : options.showColumnInMenu;

    this.options = options;
    this.framework = f;
    this.version = version;

    var use_display_helpers = options.createDisplayHelpers === undefined ? true : options.createDisplayHelpers;

    //Add page header buttons

    this.onFrameworkActivatedOnPage = function(page) {
        page.getPageViews().forEach(function(pv) {
            pv.updatePgStyles();
            _this.show(pv);
            addHeaderToolbarItems(pv);
        })
    }

    pinegrow.addEventHandler('on_show_display_helpers_changed', function() {
        if(!use_display_helpers) return;

        pinegrow.getAllPages().forEach(function(page) {
            if(page.hasFramework(f)) {
                page.getPageViews().forEach(function(pv) {
                    onShowColumnsChangedForPageView(pv);
                    onShowGridChangedForPageView(pv);
                })
            }
        })
    })

    pinegrow.addEventHandler('on_page_view_header', function(page, pv) {
        if(page.hasFramework(f)) {
            addHeaderToolbarItems(pv);
        }
    })

    var addHeaderToolbarItems = function(pv) {
        if(!use_display_helpers) return;

        if(pv.data.bsShowGridToolbarItem) return;

        var item = new PgButtonToolbarItem('<span>Grid</span>', 'Toggle grid helper overlays in ' + fname + ' containers.');
        pv.addHeaderToolbarItem(options.prefix, item);
        item.setActive(showGrid);

        item.onClick = function() {
            pv.data.bsShowGrid = !this.getActive();
            this.setActive(pv.data.bsShowGrid);
            onShowGridChangedForPageView(pv);

            pinegrow.showMode('Grid overlays in ' + fname + ' containers', pv.data.bsShowGrid)
            if(pv.data.bsShowGrid && !pinegrow.getShowDisplayHelpers()) {
                pinegrow.showQuickMessage('Note, Display helpers are OFF.');
            }
        }

        pv.data.bsShowGridToolbarItem = item;

        item = new PgButtonToolbarItem('<span>Cols</span>', 'Toggle ' + fname + ' column helper overlays.');
        pv.addHeaderToolbarItem(options.prefix, item);
        item.setActive(showColumns);

        item.onClick = function() {
            pv.data.bsShowColumns = !this.getActive();
            this.setActive(pv.data.bsShowColumns);
            onShowColumnsChangedForPageView(pv);

            pinegrow.showMode(fname + ' column overlays', pv.data.bsShowColumns)

            if(pv.data.bsShowColumns && !pinegrow.getShowDisplayHelpers()) {
                pinegrow.showQuickMessage('Note, Display helpers are OFF.');
            }
        }

        pv.data.bsShowColumnsToolbarItem = item;
    }

    ///End header buttons


    //Get device sizes
    this.getDeviceSizesForPageViewMenu = function(page, list) {
        var sizes = options.sizes;
        var sizes_breakpoints = options.sizes_breakpoints;
        var sizes_view_sizes = options.sizes_view_sizes;
        var sizes_icons = options.sizes_icons;
        var sizes_names = options.sizes_names;

        list.splice(0, list.length); //remove all default values
        for(var i = 0; i < sizes.length; i++) {
            var next_w = i + 1 < sizes.length ? sizes_breakpoints[i+1] : sizes_breakpoints[i] + 1;

            list.push({
                name: sizes[i].toUpperCase() + '<span style="opacity:0.7;"> / ' + sizes_names[i] + ' [' + sizes_breakpoints[i] + 'px' + ((i != sizes.length - 1) ? ('...' + (next_w - 1) + 'px') : (' and up')) + ']</span>',
                name_short: sizes[i].toUpperCase(),
                w: sizes_view_sizes[i],
                w_range: [sizes_breakpoints[i], (i + 1 < sizes.length) ? next_w - 1 : 9999],
                icon: sizes_icons[i],
                size: sizes[i],
                slug: sizes[i] == options.size_for_all ? '' : '-' + sizes[i]
            })
        }
    }

    //Update helpers on change

    this.updateHelpersAfterChange = function(page, info, page_col) {
        if(!use_display_helpers) return;

        if(page && page_col) {
            if(true || (info && info.source && info.source instanceof PgCodeEditView)) {
                //element code edit, lets update whole area

                //add els to list
                if(!page.data.bsHelpersUpdateCol) {
                    page.data.bsHelpersUpdateCol = new PgCollection();
                }
                page.data.bsHelpersUpdateCol.add(page_col);

                page.delayedTasks.run(options.prefix + 'updateHelpers', function() {

                    var p = new CrsaProfile();

                    if(page.data.bsHelpersUpdateCol) {
                        page.data.bsHelpersUpdateCol.normalize();

                        page.data.bsHelpersUpdateCol.forEachElement(function (pgel) {
                            if (!pgel.isDeleted) {
                                _this.updateHelpers(pgel);
                            }
                        })

                        page.data.bsHelpersUpdateCol = new PgCollection();
                    }

                    p.show('BS4 on_page_changes_done delayed');
                }, 1000)
            } else {
                /*
                disabled, didn't work well

                page_col.forEachElement(function (pgel) {
                    if (isColumn(pgel)) {
                        gh.updateColumnOverlay(pgel)
                    } else if(true || (pgel.parent && pgel.parent.hasClass('row'))) {
                        gh.removeColumnOverlay(pgel)
                    }
                    if(isContainer(pgel)) {
                        gh.updateGridHelper(pgel)
                    } else if(pgel.getData('bshelperHasGridHelper')) {
                        gh.removeGridHelper(pgel)
                    }
                })*/
            }
        }
    }
    //End update helpers on change


    if (options.showGridInMenu) {
        //Grid helper
        var gridHelperKey = options.prefix + 'grid-helper-' + this.version;

        var showGrid = pinegrow.getSetting(gridHelperKey, '0') == '1';

        var gridHelper = new PgPageViewHelper(fname + ' ' + this.version + ' grid', gridHelperKey, 'pg-internal-page-view-bootstrap-grid', null, null, '1', f);

        gridHelper.onGetCurrentValue = function() {
            return showGrid;
        }

        gridHelper.onChange = function(v) {
            showGrid = v;

            pinegrow.getAllPages().forEach(function(page) {
                if(page.hasFramework(f)) {
                    page.getPageViews().forEach(function(pv) {

                        if(pv.data.bsShowGridToolbarItem) {
                            pv.data.bsShowGridToolbarItem.setActive(v);
                            delete pv.data.bsShowGrid;
                        }

                        onShowGridChangedForPageView(pv);
                    })
                }
            })
        }
    }

    var onShowGridChangedForPageView = function(pv) {
        if(!use_display_helpers) return;

        var v = getShowGridForPageView(pv);
        if(v) {
            _this.updateGridHelperIn$El(pv.get$Body(), v)
        } else {
            _this.removeGridHelperIn$El(pv.get$Body());
        }
    }

    var getShowGridForPageView = function(pv) {
        if(!use_display_helpers) return;

        var on = pinegrow.getShowDisplayHelpers();
        if('bsShowGrid' in pv.data) return pv.data.bsShowGrid && on;
        return showGrid && on;
    }



    this.removeGridHelperIn$El = function($el) {
        $el.find('.pghelper-' + options.framework_prefix + '-grid').remove();
    }

    this.updateGridSizes = function(pv) {
        if(!use_display_helpers) return;

        if(getShowGridForPageView(pv)) {
            var small_max_size = this.options.num_columns * 40;
            pv.get$Body().find('.pghelper-' + options.framework_prefix + '-grid').each(function(i, g) {
                var $g = $(g);
                if($g.parent().innerWidth() < small_max_size) {
                    $g.addClass('pghelper-' + options.framework_prefix + '-grid-narrow');
                } else {
                    $g.removeClass('pghelper-' + options.framework_prefix + '-grid-narrow');
                }
            })
        }
    }

    this.updateGridHelperIn$El = function($el, val) {

        var val = val === undefined ? showGrid : val;

        if(!val || !$el) return;

        if($el.closest('[contenteditable="true"]').length) return;

        var conts = [];

        var cls = [];

        var code_large = this.getCodeForLarge();
        var code_small = this.getCodeForSmall();

        var small_max_size = this.options.num_columns * 40;

        //this.removeGridHelperIn$El($el);

        $el.find('.pghelper-' + options.framework_prefix + '-grid').addClass('pghelper-' + options.framework_prefix + '-grid-old')

        selfAndFind($el, options.container_selector).each(function(i,c) {
            var $cont = $(c);
            if ($cont.is('pgdiv')) return;

            var $grid = $cont.find('>.pghelper-' + options.framework_prefix + '-grid');
            if($grid.length) {
                $grid.removeClass('pghelper-' + options.framework_prefix + '-grid-old');
            } else {
                var add = true;
                var p = $cont.css('position');
                if (p == 'static') {
                    var has_positioned_children = false;
                    $cont.children().each(function(i, el) {
                        if($(el).css('position') == 'absolute') {
                            has_positioned_children = true;
                            return false;
                        }
                    })
                    if(has_positioned_children) {
                        //adding relative to $cont will change the children positions
                        add = false;
                    } else {
                        $cont.addClass('pghelper-' + options.framework_prefix + '-grid-parent-relative');
                    }
                }

                if(add) {
                    var cont_w = $cont.innerWidth();

                    var html = '<pgdiv class="' + options.grid_container_selector + ' pinegrow-ui-helper pghelper-' + options.framework_prefix + '-grid';
                    if(cont_w < small_max_size) {
                        html += ' pghelper-' + options.framework_prefix + '-grid-narrow';
                    }

                    html += '">';

                    for (var i = 0; i < options.num_columns; i++) {
                        html += code_large;
                    }
                    //make a special version with no padding for xs
                    /*
                    for (var i = 0; i < options.num_columns; i++) {
                        html += code_small;
                    }*/
                    html += '</pgdiv>';
                    var $grid = $(html);
                    $cont.prepend($grid);
                }

                var pgel = getElementPgNode($cont);
                if (pgel) {
                    pgel.setData('bshelperHasGridHelper', true);
                }
            }
        })
        $el.find('.pghelper-' + options.framework_prefix + '-grid-old').remove();
    }

    this.updateGridHelper = function(pgel) {
        pgel.get$DOMElements().forEach(function($col) {
            _this.updateGridHelperIn$El($col)
        })
    }

    this.removeGridHelper = function(pgel) {
        pgel.setData('bshelperHasGridHelper', false);
        pgel.get$DOMElements().forEach(function($col) {
            _this.removeGridHelperIn$El($col)
        })
    }

    ///End Grid helper

    if (options.showColumnInMenu) {
        var columnHelperKey = options.prefix + 'column-helper-' + this.version;

        var showColumns = pinegrow.getSetting(columnHelperKey, '0') == '1';

        var columnsHelper = new PgPageViewHelper(fname + ' ' + this.version + ' columns', columnHelperKey, 'pg-internal-page-view-bootstrap-columns', null, null, '1', f);

        columnsHelper.onGetCurrentValue = function() {
            return showColumns;
        }

        columnsHelper.onChange = function(v) {
            showColumns = v;

            pinegrow.getAllPages().forEach(function(page) {
                if(page.hasFramework(f)) {
                    page.getPageViews().forEach(function(pv) {

                        if(pv.data.bsShowColumnsToolbarItem) {
                            pv.data.bsShowColumnsToolbarItem.setActive(v);
                            delete pv.data.bsShowColumns;
                        }

                        onShowColumnsChangedForPageView(pv);
                    })
                }
            })
        }
    }


    var onShowColumnsChangedForPageView = function(pv) {
        var v = getShowColumnsForPageView(pv);
        if(v) {
            _this.updateColumnHelperIn$El(pv.get$Body(), v)
        } else {
            _this.removeColumnHelperIn$El(pv.get$Body())
        }
    }

    var getShowColumnsForPageView = function(pv) {
        var on = pinegrow.getShowDisplayHelpers();
        if('bsShowColumns' in pv.data) return pv.data.bsShowColumns && on;
        return showColumns && on;
    }

    this.removeColumnHelperIn$El = function($el) {
        $el.find('.pghelper-' + options.framework_prefix + '-column').remove();
    }

    this.updateColumnHelperIn$El = function($el, val) {

        var val = val === undefined ? showColumns : val;

        if(!val || !$el) return;

        if($el.closest('[contenteditable="true"]').length) return;

        this.removeColumnHelperIn$El($el);

        selfAndFind($el, all_col_classes_selector).not('.pinegrow-ui-helper').each(function(i, col) {
            var $col = $(col);

            var map = _this.getColumnSizes(col);
            var html = '';
            options.sizes.forEach(function(size) {
                if(size in map) {
                    html += '<pgspan>' + size + _this.getColumnSizeAsShortString(map[size], ' ') + '</pgspan>';
                }
            })
            addColumnOverlay($col, html);
        })
    }

    this.updateColumnOverlay = function(pgel) {
        if(!use_display_helpers) return;

        var html = null;

        var page = pgel.getPage();
        if(page) {
            page.getPageViews().forEach(function(pv) {
                if(getShowColumnsForPageView(pv)) {
                    if(html === null) {
                        var map = _this.getColumnSizes(pgel);
                        html = '';
                        options.sizes.forEach(function(size) {
                            if(size in map) {
                                html += '<pgspan>' + size + _this.getColumnSizeAsShortString(map[size], ' ') + '</pgspan>';
                            }
                        })
                    }
                    var $col = pgel.get$DOMElement(pv.get$Html());
                    if($col) {
                        if(html.length) {
                            //is column
                            var $overlay = $col.find('>.pghelper-' + options.framework_prefix + '-column');
                            if($overlay.length === 0) {
                                $overlay = addColumnOverlay($col, html, pgel);
                            } else {
                                $overlay.find('>.pghelper-' + options.framework_prefix + '-column-inner').html(html);
                            }
                        } else {
                            $col.find('>.pghelper-' + options.framework_prefix + '-column').remove();
                        }
                    }
                }
            })
        }
    }

    this.removeColumnOverlay = function(pgel) {
        pgel.get$DOMElements().forEach(function($col) {
            $col.find('>.pghelper-' + options.framework_prefix + '-column').remove();
        })
        //pgel.setData('bshelperHasColumnOverlay', false);
    }

    var addColumnOverlay = function($col, sizeshtml, pgel) {
        var html = '<pgdiv class="pinegrow-ui-helper pghelper-' + options.framework_prefix + '-column"><pgdiv class="pghelper-' + options.framework_prefix + '-column-inner">' + sizeshtml + '</pgdiv></pgdiv>';
        var $o = $(html);

        var add = true;
        var p = $col.css('position');
        if (p == 'static') {
            var has_positioned_children = false;
            $col.children().each(function(i, el) {
                if($(el).css('position') == 'absolute') {
                    has_positioned_children = true;
                    return false;
                }
            })
            if(has_positioned_children) {
                //adding relative to $cont will change the children positions
                add = false;
            } else {
                $col.addClass('pghelper-' + options.framework_prefix + '-grid-parent-relative');
            }
        }

        if(add) {
            $col.append($o);

            if(!pgel) {
                pgel = getElementPgNode($col);
                if(pgel) {
                    //pgel.setData('bshelperHasColumnOverlay', true);
                }
            }

            return $o;
        }

        return null;
    }


    ///Column helper

    //ENd column helper

    this.updateHelpersIn$El = function($el, pv) {
        if(!use_display_helpers) return;

        this.updateGridHelperIn$El($el, pv ? getShowGridForPageView(pv) : undefined);
        this.updateColumnHelperIn$El($el, pv ? getShowColumnsForPageView(pv) : undefined);
    }

    this.show = this.updateHelpersOnPageView = function(pv) {
        if(!use_display_helpers) return;

        this.updateHelpersIn$El(pv.get$Body(), pv);
    }

    this.updateHelpers = function(pgel) {
        var page = pgel.getPage();
        if(page) {
            this.updateHelpersOnPage(page);
        }
    }

    this.updateHelpersOnPage = function(page) {
        page.getPageViews().forEach(function (pv) {
            _this.updateHelpersOnPageView(pv);
        })
    }



    // this.getClassForSize = function(size, prefix, add) {
    //     if(add === null || add === undefined) {
    //         add = '';
    //     }  else {
    //         add = add + '';
    //     }
    //     return prefix + (size == options.size_for_all ? '' : '-' + size) + (add.length ? ('-' + add) : '');
    // }

    // this.getColumnSizes = function(pgel_or_node) {
    //     var re = new RegExp('^col(\\-(' + options.sizes.join('|') + '))?(\\-([0-9]+|auto))?');

    //     var map = {};

    //     var list = pgel_or_node instanceof pgParserNode ? pgel_or_node.getClasses() : pgel_or_node.classList;

    //     list.forEach(function(cls) {
    //         var m = cls.match(re);
    //         if(m) {
    //             var size = m[2] || options.size_for_all;
    //             var cols = m[4] || '';
    //             map[size] = cols;
    //         }
    //     })
    //     return map;
    // }

    this.getColumnSizeAsShortString = function(num, prefix) {
        if(prefix === undefined) prefix = '';
        var sptxt = '';
        switch(num) {
            case '':
                break;
            case 'auto':
                sptxt = ' A';
                break;
            default:
                sptxt = ' ' + num;
                break;
        }
        if(sptxt.length) {
            sptxt = prefix + sptxt;
        }
        return sptxt;
    }

    this.getViewCss = function () {
        if(!use_display_helpers) return '';

        return '\
        pgdiv { display:block;}\
        .pghelper-' + options.framework_prefix + '-grid {\
            position:absolute;\
            top:0;\
            width: 100%;\
            height: 100%;\
            z-index:1;\
            pointer-events:none;\
            outline: none !important;\
            overflow: hidden;\
        }\
        .pghelper-' + options.framework_prefix + '-grid > pgdiv {\
            height:100%;\
            outline: none !important;\
            border-left: 1px solid rgba(0,0,0,0.07);\
        }\
        .pghelper-' + options.framework_prefix + '-grid > pgdiv:first-child {\
            border-left:none;\
        }\
        .pghelper-' + options.framework_prefix + '-grid.pghelper-' + options.framework_prefix + '-grid-narrow > pgdiv {\
            padding:0;\
        }\
        .pghelper-' + options.framework_prefix + '-grid > pgdiv > pgdiv {\
            background:rgba(0,0,0,0.03);\
            height:100%;\
            outline: none !important;\
        }\
        .pghelper-' + options.framework_prefix + '-grid-parent-relative {\
            position:relative;\
        }\
        .pghelper-' + options.framework_prefix + '-column {\
            position:absolute;\
            left:0;\
            right:0;\
            top:0;\
            bottom:0;\
            background:rgba(51, 173, 214, 0.06);\
            padding: 5px 5px;\
            pointer-events:none;\
            border: 1px solid #b5d9e4;\
        }\
        .pghelper-' + options.framework_prefix + '-column-inner {\
            background: rgba(51, 173, 214, 0.12);\
            height: 100%;\
            pointer-events:none;\
            text-align: right;\
            padding: 0px 6px;\
        }\
        .pghelper-' + options.framework_prefix + '-column-inner pgspan {\
            font-size:13px;\
            text-transform:uppercase;\
            font-family: sans-serif;\
            color: #319ec3;\
            display: inline-block;\
            margin-left: 10px;\
        }\
        ' + options.style;
    }

    this.onPageViewResized = function(page, pv) {
        this.updateMenuPointsAfterViewResize(page, pv);
        this.updateGridSizes(pv);
    }

    this.addPrefixToSectionsAndFields = function(sections) {
        var r = {};
        $.each(sections, function (skey, sdef) {
            if (sdef.fields) {
                var fields = {};
                $.each(sdef.fields, function (fkey, fdef) {
                    fields[options.prefix + fkey] = fdef;
                })
                sdef.fields = fields;
            }
            r[options.prefix + skey] = sdef;
        })
        return r;
    }

    ///Menu points
    this.getMenuPoints = function(page, pgel, menu, points) {
        if(this.options.isColumn(pgel)) {
            //Will add 2 points for column size: one for the current visible setting and one for the current device size
            var col_point_auto = PgMenuPointColumnSizeClass('right', menu, false, false, this);
            points.push(col_point_auto);

            var col_point_current_size = PgMenuPointColumnSizeClass('right', menu, this.getCurrentPageViewSize(menu.pageView), col_point_auto, this);
            points.push(col_point_current_size);

            col_point_auto.hidePointInDrag = col_point_current_size;
            col_point_current_size.hidePointInDrag = col_point_auto;
        }
    }

    this.updateMenuPointsAfterViewResize = function(page, pv) {
        var size = this.getCurrentPageViewSize(pv);
        var points = [];
        pinegrow.selectedElements.getMenuPoints(pv).forEach(function(p) {
            if(p instanceof PgMenuPointColumnSizeClass) {
                if(p.forceSize) {
                    p.forceSize = size;
                }
                p.update();
                points.push(p);
            }
        })
        //do it in 2 passes
        points.forEach(function(p) {d
            p.update();
        })
    }

    ///End menu points


    //Page libs
    var pagelib_url = f.getBaseUrl() + '/pagelib';
   if(isApp()) {
       var pagelib_dir = path.join(crsaMakeFileFromUrl(f.getBaseUrl()), 'pagelib');

   }

    this.getPageLibs = function(list) {
        var r = [];
        list.forEach(function(item) {
            r.push({
                name: item.name,
                url: pagelib_url + '/' + item.url,
                selector: '.pg-lib-item',
                framework: f,
                removeClasses: ['pg-lib-item','pg-lib-item-skip-replace-images']
            })
        })
        return r;
    }

    this.addBuildPageLibsFeature = function(base_url, sources) {

        var buildPageLibs = function() {

            var path = require('path');
            var dir = pagelib_dir;

            var pb = new PgPluginBuilder(f);

            var master = pb.readFile('pagelib/master.html');

            master.withoutEvents(function() {

                //Combine all in one
                var all = master.clone(true);
                var main = all.findOne('main');

                var idx = -1;

                var addHelp = function(source, section) {
                    if(source.help) {
                        var pgel = pgCreate('<div class="pg-lib-help"><a href="' + source.help.link + '" class="external">' + source.help.label + '</a></div>');
                        var content = section.findOne('.pg-lib-section-content');
                        pgel.insertBefore(content)
                    }
                }

                var doSource = function() {
                    idx++;
                    if(sources.length > idx) {

                        var source = sources[idx];

                        pinegrow.showQuickMessage('Doing ' + source.url + '...');

                        if(source.import) {
                            var node = pb.readFile( crsaMakeUrlFromFile(path.join(dir, source.import)));
                            node = node.findOne('.pg-lib-section')
                            main.append(node);
                            addHelp(source, node);
                            doSource();
                        } else {

                            pb.getSourceNode(base_url + source.url, function (node) {

                                var h1 = node.findOne('.bd-title');
                                var lead = node.findOne('.bd-lead');

                                var section = pgCreate('<div class="pg-lib-section"><div class="pg-lib-section-info"><h2 class="pg-lib-section-title">' + h1.html() + '</h2><div class="pg-lib-section-desc">' + lead.html() + '</div></div><div class="pg-lib-section-content"></div></div>');

                                main.append(section);

                                var section_content = section.findOne('.pg-lib-section-content');

                                node.find('.bd-example').forEach(function (example) {

                                    if (source.filter && !source.filter(example)) return;

                                    var wrap = pgCreate('<div class="pg-lib-group"></div>');
                                    section_content.append(wrap);

                                    example.children.forEach(function (pgel) {
                                        if (pgel.isElement) {
                                            //add to our lib

                                            pgel.removeAttribute('style'); //dont want that
                                            pgel.addClass('pg-lib-item');

                                            if(!source.skip_replace_images) {
                                                pb.replaceHolderImages(pgel);
                                            } else {
                                                pgel.addClass('pg-lib-item-skip-replace-images');
                                            }

                                            wrap.append(pgel);

                                            if (source.after) source.after(pgel, example, node, main, section_content)
                                        }
                                    })

                                    if(source.after_group) source.after_group(wrap);
                                })

                                addHelp(source, section);

                                doSource();
                            })
                        }

                    } else {
                        //lets write it out
                        var file = path.join(dir, 'index.html');
                        var fs = require('fs');
                        fs.writeFileSync(file, all.toStringOriginal(true), {encoding: 'utf8'});
                        pinegrow.showQuickMessage('Done!');
                    }
                }
                doSource();
            })
        }

        var $menu = $('<li class="dropdown"><a href="#" class="dropdown-toggle" data-toggle="dropdown"><span>BS Dev</span></a>\
                <ul class="dropdown-menu with-checkboxes">\
                    <li><a href="#" class="bs-build-pagelibs">Build Page Libs</a></li>\
                </ul>\
            </li>');

        pinegrow.addPluginControlToTopbar(f, $menu);

        $menu.find('.bs-build-pagelibs').on('click', function(e) {
            buildPageLibs();
        })
    }

    //End build page libs

    this.isSizeVisible = function(page_or_pv, size) {
        var pv = page_or_pv instanceof PgPageView ? page_or_pv : page_or_pv.activeView;
        return pv.deviceWidth >= options.sizes_breakpoints_map[size];
    }

    var current_pv = null;
    var current_pv_size = null;

    this.getCurrentPageViewSizeCached = function(pv) {

    }

    this.getCurrentPageViewSize = function(pv) {
        var size = options.size_for_all;
        var sizes = options.sizes;

        for(var i = 0; i < sizes.length; i++) {
            if(this.isSizeVisible(pv, sizes[i])) {
                size = sizes[i];
            } else {
                break;
            }
        }
        return size;
    }

    //Init stuff

    var selfAndFind = function($el, sel) {
        var $r = $el.find(sel);
        if($el.is(sel)) $r = $r.add($el);
        return $r;
    }


    var all_col_classes_selector = [];

    this.init = function () {
        if (options.col_prefix_class) {
            all_col_classes_selector.push('.' + options.col_prefix_class);
        }

        options.sizes.forEach(function(s) {
            for(var i = 1; i <= options.num_columns; i++) {
                all_col_classes_selector.push('.' + _this.getClassForSize(s, options.col_prefix_class, i));
            }
        })

        all_col_classes_selector = all_col_classes_selector.join(',');
    }

    //Resources

    var include_remote = true;

    var onAddResourceFilter = function(res) {
        var remote = res.url.startsWith('https://');
        return include_remote === remote;
    }

    f.resources_name = options.version_name || f.name;
    f.customResourceNamespace = '';

    if(!options.no_include_local_question) {
        f.on_add_resources = function (is_project, done) {
            pinegrow.showAlert(`<p>${f.name} can be included to the page in two different ways:</p><ul><li>From  <b>remote CDN server</b>. This is the simplest option but it requires internet connection.</li><li>With <b>local files in your project</b>. All required ${f.name} files will be copied to your project.</li></ul><p>How would you like to include ${f.name}?</p>`, 'Use CDN or local files?', 'Cancel', 'Add with CDN', null, function () {
                include_remote = true;
                done(true);
            }, 'Add with local files', function () {
                include_remote = false;
                done(true);
            })
        }
    }

    f.resources.description = `Include <b>${f.resources_name}</b> CSS and JS files. Updating resources will add (or update to) the latest version of ${f.name} CSS and JS to the page.`;

    this.addResource = function(url, detect, always, force_type, force_footer) {
        var relative_url = null;
        var source = null;

        if(crsaIsAbsoluteUrl(url)) {

        } else {
            relative_url = url;
            url = f.getResourceUrl('template/' + url);
            source = url;
        }
        res = new PgComponentTypeResource(url, null, force_type);
        res.relative_url = relative_url;
        res.source = source;
        res.detect = detect;
        res.replace = true;
        res.footer = force_footer !== undefined ? force_footer : (res.type === 'application/javascript');
        if(!always) {
            res.on_add_resource_filter = onAddResourceFilter;
        }
        f.resources.add(res);
        return res;
    }
}