/**
 * Created by Matjaz on 2/3/14.
 */

$(function() {

    $('body').one('pinegrow-ready', function(e, pinegrow) {
        var f = new PgFramework('pg.html', 'Plain HTML');

        f.name_short = 'HTML';

        f.description = 'Plain HTML starting pages and components.';
        f.author = 'Pinegrow';
        f.author_link = 'http://pinegrow.com';

        f.setScriptFileByScriptTagId('plugin-html'); //get url if script is included directly into edit.html
        pinegrow.addFramework(f, 100); //show at the bottom

        //let every page use this framework
        f.default = true;
        f.show_in_manager = false;

        f.lock_css_files = [/(^|\/)normalize\.css/i];

        f.ignore_css_files = [/(^|\/)theme\.css/i];

        var simple = false;

        if(pgf.html_plugin) {
            simple = pgf.html_plugin.simple || false;
        }

        var tag_to_def_map = {}
        var default_def = null;

        f.getType = function($elnull, pgel, skip_actions) {
            return tag_to_def_map[pgel.tagName] || default_def;
        }


        var just_essential = pgf.kids || false;

        if(just_essential && false) {
            var list_containers = [];
            var list_content = [];
            var list_components = [];

            var addComponent = function(type, name, sel, code, on_can_insert_into, options) {
                var container = new PgComponentType(prefix + type, name);
                container.selector = sel;
                container.code = code;
                container.empty_placeholder = true;
                container.tags = 'major';

                container.on_can_insert_into = on_can_insert_into;

                if(options) {
                    $.each(options, function(key, val) {
                        container[key] = val;
                    })
                }
                //Add it to our framework
                f.addComponentType(container);
                return container;
            }

            //Containers

            var parentsAre = function(pgel, sel) {
                var p = pgel;
                while(p && !p.rootNode) {
                    if(!p.isSelector(sel)) return false;
                    p = p.parent;
                }
                return true;
            }

            var sections_selector = 'header,section,footer';
            var blocks_selector = sections_selector + ',div';

            var insertInBlocksOnly = function(pgel) {
                if(parentsAre(pgel, 'div,body,html,' + sections_selector)) return true;
                return false;
            }

            var insertInBlocksAndLinksOnly = function(pgel) {
                if(parentsAre(pgel, 'a,div,body,html,' + sections_selector)) return true;
                return false;
            }

            var isSection = function(pgel) {
                return pgel.isSelector(sections_selector);
            }

            var isBlock = function(pgel) {
                return pgel.isSelector(blocks_selector);
            }

            var canInsertSection = function(pgel, def) {
                //can insert into pgel?

                if(pgel.closest(sections_selector)) {
                    return false;
                }
                if(pgel.tagName === 'body') return true;

                if(parentsAre(pgel, 'div,body,html')) return true;

                return false;
            }

            var canInsertContent = function(pgel, def) {
                //must be in block
            }

            var insertInSectionsOnly = function(pgel, def) {
                return pgel.parent.isSelector(sections_selector);
            }

            list_containers.push(addComponent('header', 'Header', 'header', `<header></header>`, canInsertSection));

            list_containers.push(addComponent('section', 'Section', 'section', `<section></section>`, canInsertSection));

            list_containers.push(addComponent('footer', 'Footer', 'footer', `<footer></footer>`, canInsertSection));

            list_containers.push(addComponent('div', 'Div', 'div', `<div></div>`, insertInBlocksOnly));

            list_containers.push(addComponent('columns.centered', 'Columns - Centered', '.container', `<div class="container"><div class="row"><div class="col-md-4 pg-empty-placeholder"></div><div class="col-md-4 pg-empty-placeholder"></div><div class="col-md-4 pg-empty-placeholder"></div></div></div>`, function(pgel, def) {
                return isSection(pgel);
            }));

            list_containers.push(addComponent('columns', 'Columns', '.container-fluid', `<div class="container-fluid"><div class="row"><div class="col-md-4"></div><div class="col-md-4"></div><div class="col-md-4"></div></div></div>`));

            // Content

            var h_sel = 'h1,h2,h3,h4,h5,h6';
            var text_sel = 'p,span,button,.btn,b,em,i'

            for(var h = 1; h <= 6; h++) {
                list_content.push(addComponent('h' + h, 'H' + h, 'h' + h, `<h${h}>Heading ${h}</h${h}>`, function(pgel) {
                    //can insert into
                    return isBlock(pgel);

                    //if(pgel.closest(h_sel)) return false;
                    //if(insertInBlocksAndLinksOnly(pgel)) return true;
                }));
            }

            list_content.push(addComponent('p', 'Paragraph', 'p', `<p>This is some nice text.</p>`, function(pgel) {
                return isBlock(pgel);
            }));

            list_content.push(addComponent('img', 'Image', 'img', `<img src="${pinegrow.getPlaceholderImage()}">`, function(pgel, def) {
                return isBlock(pgel) || pgel.isSelector('p');
            }));

            /*
             list_content.push(addComponent('link', 'Link', 'a', `<a href="">Link</a>`, function(pgel, def) {
             return isBlock(pgel) || pgel.isSelector('p');
             }));
             */

            list_content.push(addComponent('button', 'Button', '.btn', `<button class="btn btn-primary">Click me</button>`, function(pgel, def) {
                return isBlock(pgel) || pgel.isSelector('p');
            }));

            var section = new PgFrameworkLibSection('pg.html.containers', 'Layout');
            //Pass components in array
            section.setComponentTypes(list_containers);
            f.addLibSection(section);

            section = new PgFrameworkLibSection('pg.html.content', 'Content');
            //Pass components in array
            section.setComponentTypes(list_content);
            f.addLibSection(section);

        }

        //will generate definitions from data
        var html5 = [
            {'name' : 'Sections', key: 'html-sections', els: [
                    {
                        tag: 'section/p',
                        name: 'section',
                        major: true,
                        preview_description: `Use a SECTION element to create a section (chapter) of the webpage. Each section usually contains a heading and content.`
                    },
                    {
                        tag:'nav/p',
                        name: 'nav',
                        major: true,
                        preview_description: `Use a NAV element to create a navigation menu for the website.`
                    },'article/p','aside/p',
                    {
                        tag: 'h1',
                        name: 'h1',
                        code: '<h1>Heading 1</h1>',
                        preview_description: `Use H1 (heading 1) for the main headline on the page.`
                    },
                    {
                        tag: 'h2',
                        name: 'h2',
                        code: '<h2>Heading 2</h2>',
                        preview_description: `Use H2 (heading 2) for section headings.`
                    },
                    {
                        tag: 'h3',
                        name: 'h3',
                        code: '<h3>Heading 3</h3>',
                        preview_description: `Use H3 (heading 3) for section sub-headings.`
                    },
                    {
                        tag: 'h4',
                        name: 'h4',
                        code: '<h4>Heading 4</h4>',
                        preview_description: `Use H4 (heading 4) for section sub-sub-headings.`
                    },
                    'h5:Heading 5:h5','h6:Heading 6:h6',
                    {
                        tag: 'header/p',
                        name: 'header',
                        major: true,
                        preview_description: `Use a HEADER element to create the most important part of the webpage, usually at the top of the page.`
                    },
                    {
                        tag: 'footer/p',
                        name: 'footer',
                        major: true,
                        preview_description: `Use a FOOTER element for displaying the information about the website, usually at the bottom of the page.`
                    },'address','main/p']},
            {'name' : 'Grouping elements', key: 'html-grouping', els: [
                    {
                        tag: 'p',
                        name: 'p',
                        code: '<p>This is a paragraph, one or more lines of text. Each paragraph is displayed in a new line.</p>',
                        preview_description: `Use a P (paragraph) element to display text.`
                    },
                    {
                        tag: 'hr',
                        name: 'hr',
                        code: '<hr>',
                        preview_description: `Use a HR (horizontal line) element to display a horizontal divider line between two elements.`
                    },
                    'pre:Preformated text','blockquote',
                    {tag: 'ol/p', props: {start: {name: 'Start'}}},'ul/p',
                    {
                        tag: 'li:List item',
                        parent_selector: 'ul,ol'
                    },'dl','dt:Term','dd:Definition','figure','figcaption',
                    {
                        tag: 'div',
                        name: 'div',
                        empty_placeholder: true,
                        code: '<div></div>',
                        preview_description: `Use a DIV element to group one or more HTML elements.`
                    }]},
            {'name' : 'Text', key: 'html-text', els: [
                    {
                        tag:'a',
                        name: 'Anchor',
                        selector: function() { return false;},
                        drag_helper: '<div>Anchor</div>',
                        code: '<a name="anchor"></a>',
                        props: {
                            'name' : {name: 'Name'}
                        }
                    },
                    {
                        tag:'a',
                        name: 'a',
                        code: '<a href="">Link</a>',
                    },
                    'em','strong','small','s','cite','q','dfn','abbr','data','time','code','var','samp','kbd','sub','sup','i','b','u','mark','ruby','rt','rp','bdi','bdo',
                    {
                        tag: 'span',
                        name: 'span',
                        code: '<span>This is a span.</span>',
                        preview_description: `Use a SPAN element to group one or more words, so that you can style them with CSS. SPAN is displayed inline - it will not break the line of text.`
                    },
                    'br','wbr']},
            {'name' : 'Edits', key: 'html-edits', els: ['ins','del']},
            {'name' : 'External content', key: 'html-embed', els: [
                    {
                        tag: 'img/e',
                        name: 'img',
                        preview_description: `Use an IMG element to display an image.`,
                        code : function() {
                            return '<img src="' + pinegrow.getPlaceholderImage() + '">';
                        },
                        props: {
                            'src' : {name: 'Source', type: 'image', autocomplete_same_tag: true},
                            'alt' : {name: 'Alt text'},
                            'width' : {
                                name: 'Width',
                                type: 'slider',
                                slider_def_unit: '',
                                slider_min: 0
                            },
                            'height' : {
                                name: 'Height',
                                type: 'slider',
                                slider_def_unit: '',
                                slider_min: 0
                            },
                            'loading': {
                                name: 'Lazy loading',
                                type: 'checkbox',
                                value: 'lazy'
                            },
                            'getmorebutton' : {
                                show_if: function() {
                                    return !pgf.kids;
                                },
                                name: 'Get more images',
                                type: 'button',
                                func: function(obj, $el) {
                                    pgShowQuickImageBrowser($el);
                                }
                            },
                            'srcset': {
                                name: 'Srcset',
                                type: 'text',
                                textarea: true,
                                start_section: 'Responsive images'
                            },
                            'sizes': {
                                name: 'Sizes',
                                type: 'text',
                                textarea: true
                            },

                            'calcsizesbutton' : {
                                show_if: function() {
                                    return !pgf.kids;
                                },
                                name: 'Calculate sizes',
                                type: 'button',
                                func: function(obj, $el) {
                                    var r = new PgResponsiveImages();
                                    r.calculateResponsiveSizes(obj, function(sizes) {
                                        if(sizes.length) {
                                            pinegrow.makeChanges(obj, 'Calculate sizes', function () {
                                                obj.setAttr('sizes', sizes.join(',\n'));
                                            })
                                            pinegrow.selectedElements.reselect();
                                            pinegrow.showQuickMessage('Done.');
                                        } else {
                                            pinegrow.showQuickError('Failed to calculate image sizes.')
                                        }
                                    })
                                }
                            },

                        }
                    },
                    {
                        tag: 'picture',
                        name: 'picture',
                        preview_description: `Use a PICTURE element to display an image with different sources and formats.`,
                        action_menu: {
                            add: ['html-source']
                        },
                        code : function() {
                            return '<picture><img src="' + pinegrow.getPlaceholderImage() + '"></picture>';
                        },
                        props: {
                            picturenotice: {
                                name: 'Select &lt;img&gt; and &lt;source&gt; child elements to setup the picture.',
                                type: 'displayhtml',
                                action: 'custom'
                            },
                            pictureaddsource : {
                                show_if: function() {
                                    return !pgf.kids;
                                },
                                name: 'Add source',
                                type: 'button',
                                func: function(obj, $el) {
                                    var api = new PgApi();
                                    api.insertElements(`<source type="image/webp">`, null, null, {
                                        append_before_selector: 'img',
                                        select: true
                                    })
                                }
                            },
                        }
                    },
                    {
                        tag: 'link/e',
                        code : function() {
                            return '<link>';
                        },
                        props: {
                            'href' : {name: 'Href', file_picker: true, file_picker_no_proxy: true},
                            'rel' : {name: 'Rel'}
                        },
                        on_inserted : function (pgel, cp) {
                            if (pgel.getAttr('rel') == "stylesheet") {
                                cp.updateStylesheetsList();
                            }
                        },
                        on_deleted : function (pgel, cp) {
                            if (pgel.getAttr('rel') == "stylesheet") {
                                cp.updateStylesheetsList();
                            }
                        }
                    },
                    {
                        tag: 'iframe/e',
                        name: pgf.kids ? 'iframe' : 'iframe',
                        drag_helper: '<div>Embedded media</div>',
                        preview: '<div>YouTube video or similar media.</div>',
                        preview_description: 'Use an IFRAME element to include a media or content from another website, for example a YouTube video.',
                        props: {
                            'iframesrc' : {
                                name: 'Source url',
                                attribute: 'src',
                                action: 'element_attribute',
                                file_picker: true, file_picker_no_proxy: true, autocomplete_same_tag: true
                            },
                            'iframewidth' : {
                                name: 'Width',
                                type: 'slider',
                                attribute: 'width',
                                action: 'element_attribute',
                                slider_def_unit: '',
                                slider_min: 0
                            },
                            'iframeheight' : {
                                name: 'Height',
                                type: 'slider',
                                attribute: 'height',
                                action: 'element_attribute',
                                slider_def_unit: '',
                                slider_min: 0
                            },
                            'iframeloading': {
                                name: 'Lazy loading',
                                type: 'checkbox',
                                value: 'lazy',
                                attribute: 'loading',
                                action: 'element_attribute',
                            },
                        },
                        on_inserted : function (pgel, cp) {
                            if(!pgel.getAttr('src')) {
                                //ask about the embed
                                pinegrow.showQuickDialog('Insert iframe', 'Enter the HTML code', '<p>Paste the HTML &lt;iframe&gt; code from YouTube or similar service into the code box and click OK</p><p>Note that you can\'t select &lt;iframe&gt; by clicking on them on the page. Use the Tree structure panel to select them.</p>', {
                                    code: {
                                        type: 'text',
                                        textarea: true,
                                        name: 'HTML code',
                                        value: '',
                                        validate: function(obj, prop, value) {
                                            value = (value || '').trim();
                                            var msg = `The field should contain code that looks like: <code>&lt;iframe src="..."&gt;&lt;/iframe&gt;`;
                                            if(!value) return msg;
                                            try {
                                                var pgel = pgCreate(value, false);
                                                if(pgel.tagName !== 'iframe' || pgel.validateTree().length) {
                                                    return msg;
                                                }
                                            } catch(err) {
                                                return msg;
                                            }
                                        }
                                    }
                                }, 'Cancel', 'OK', null, function(values) {

                                    pinegrow.makeChanges(pgel, 'Embed an iframe', function() {
                                        var new_pgel = pgCreate(values.code.trim());
                                        pgel.replaceWith(new_pgel);
                                    })
                                })
                            }
                        },
                    },
                    {
                        tag: 'embed/e',
                        props: {
                            'src' : {name: 'Source url', file_picker: true, file_picker_no_proxy: true},
                            'width' : {name: 'Width'},
                            'height' : {name: 'Height'}
                        }
                    },
                    {
                        tag: 'object/e',
                        props: {
                            'data' : {name: 'Data'},
                            'type' : {name: 'Mime type'},
                            'width' : {name: 'Width'},
                            'height' : {name: 'Height'}
                        }
                    },
                    'param/e',
                    {
                        tag: 'video',
                        name: 'video',
                        code: '<video>Your browser doesn\'t support HTML5 video tag.</video>',
                        action_menu: {
                            add: ['html-source']
                        },
                        props: {
                            'src' : {name: 'Src', file_picker: true, file_picker_no_proxy: true},
                            /*
                            'videonotice': {
                                name: 'Note, playback of H.264 videos (usually, .mp4 files) is not supported in Pinegrow. All videos will play ok in the browser.',
                                type: 'displayhtml',
                                action: 'custom'
                            },*/
                            'poster' : {name: 'Poster', type: 'image', file_picker: true, file_picker_no_proxy: true},
                            width: {
                                name: 'Width',
                                type: 'slider',
                                slider_def_unit: '',
                                action: 'element_attribute',
                                attribute: 'width',
                                placeholder: '200'
                            },
                            height: {
                                name: 'Height',
                                type: 'slider',
                                slider_def_unit: '',
                                action: 'element_attribute',
                                attribute: 'height',
                                placeholder: '200'
                            },
                            'autoplay': {
                                name: 'Autoplay',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'autoplay'
                            },
                            'controls': {
                                name: 'Controls',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'controls',
                                empty_attribute: true
                            },
                            'loop': {
                                name: 'Loop',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'loop'
                            },
                            'muted': {
                                name: 'Muted',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'muted'
                            },
                            'preload': {
                                name: 'Preload',
                                type: 'select',
                                action: 'element_attribute',
                                show_empty: true,
                                attribute: 'preload',
                                options: [
                                    {key: 'none', name: 'none'},
                                    {key: 'metadata', name: 'metadata'},
                                    {key: 'auto', name: 'auto'}
                                ]
                            },
                            'crossorigin': {
                                name: 'Cross-origin',
                                type: 'select',
                                action: 'element_attribute',
                                show_empty: true,
                                attribute: 'crossorigin',
                                options: [
                                    {key: 'anonymous', name: 'anonymous'},
                                    {key: 'use-credentials', name: 'use-credentials'}
                                ]
                            }
                        }
                    },
                    {
                        tag: 'audio',
                        name: 'audio',
                        code: '<audio>Your browser doesn\'t support HTML5 audio tag.</audio>',
                        action_menu: {
                            add: ['html-source']
                        },
                        props: {
                            'src' : {name: 'Src', file_picker: true, file_picker_no_proxy: true},
                            'volume': {
                                name: 'Volume',
                                type: 'slider',
                                action: 'element_attribute',
                                slider_def_unit: '',
                                slider_step: 0.1,
                                slider_min: 0,
                                slider_max: 1.0,
                                attribute: 'volume'
                            },
                            'autoplay': {
                                name: 'Autoplay',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'autoplay'
                            },
                            'controls': {
                                name: 'Controls',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'controls',
                                empty_attribute: true
                            },
                            'loop': {
                                name: 'Loop',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'loop'
                            },
                            'muted': {
                                name: 'Muted',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'muted'
                            },
                            'preload': {
                                name: 'Preload',
                                type: 'select',
                                action: 'element_attribute',
                                show_empty: true,
                                attribute: 'preload',
                                options: [
                                    {key: 'none', name: 'none'},
                                    {key: 'metadata', name: 'metadata'},
                                    {key: 'auto', name: 'auto'}
                                ]
                            },
                            'crossorigin': {
                                name: 'Cross-origin',
                                type: 'select',
                                action: 'element_attribute',
                                show_empty: true,
                                attribute: 'crossorigin',
                                options: [
                                    {key: 'anonymous', name: 'anonymous'},
                                    {key: 'use-credentials', name: 'use-credentials'}
                                ]
                            }
                        }
                    },
                    {
                        tag: 'source',
                        name: 'Source',
                        code: '<source>',
                        props: {
                            'sourcesrc' : {
                                name: 'Src',
                                attribute: 'src',
                                file_picker: true,
                                file_picker_no_proxy: true,
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName !== 'picture'
                                }
                            },
                            'sourcesrcset' : {
                                name: 'Srcset',
                                textarea: true,
                                attribute: 'srcset',
                                type: 'text',
                                file_picker: true,
                                file_picker_no_proxy: true,
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName === 'picture'
                                },
                                on_changed: function(pgel, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
                                    if(eventType === 'change' && value) {
                                        var mime = pinegrow.getMimeType(value);
                                        if(mime) {
                                            $field.closest('.pg-ui-section-content').find('.crsa-field-sourcetype input').val(mime).trigger('change');
                                        }
                                    }
                                }
                            },
                            'sourcetype': {
                                name: 'Type',
                                type: 'text',
                                action: 'element_attribute',
                                show_empty: true,
                                attribute: 'type',
                                open_list_on_focus: true,
                                options: [
                                    {key: 'image/jpeg', name: 'image/jpeg'},
                                    {key: 'image/png', name: 'image/png'},
                                    {key: 'image/webp', name: 'image/webp'},
                                    {key: 'image/gif', name: 'image/gif'},
                                    {key: 'image/svg+xml', name: 'image/svg+xml'},
                                    {key: 'image/apng', name: 'image/apng'},
                                    {key: 'image/avif', name: 'image/avif'},
                                    {key: 'video/mp4', name: 'video/mp4'},
                                    {key: 'video/webm', name: 'video/webm'},
                                    {key: 'video/ogg', name: 'video/ogg'},
                                    {key: 'video/quicktime', name: 'video/quicktime'},
                                    {key: 'audio/mpeg', name: 'audio/mpeg'},
                                    {key: 'audio/ogg', name: 'audio/ogg'},
                                    {key: 'audio/wav', name: 'audio/wav'}
                                ]
                            },
                            'sourcesizes' : {
                                name: 'Sizes',
                                textarea: true,
                                type: 'text',
                                attribute: 'sizes',
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName === 'picture'
                                }
                            },
                            'sourcemedia' : {
                                name: 'Media',
                                textarea: false,
                                attribute: 'media',
                                type: 'text',
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName === 'picture'
                                }
                            },
                            'sourcewidth' : {
                                name: 'Width',
                                type: 'slider',
                                attribute: 'width',
                                slider_def_unit: '',
                                slider_min: 0,
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName === 'picture'
                                }
                            },
                            'sourceheight' : {
                                name: 'Height',
                                type: 'slider',
                                attribute: 'height',
                                slider_def_unit: '',
                                slider_min: 0,
                                show_if: function(values, pgel, $field) {
                                    return pgel.parent.tagName === 'picture'
                                }
                            },
                        }
                    },'track/e','canvas/e','map/e','area/e',
                    {
                        tag: 'svg',
                        selector: 'svg,path,polygon,circle,line,rect,g',
                        name: 'Svg',
                        code: '<svg height="100" width="100">\
                    <circle cx="50" cy="50" r="40" stroke="#9e9e9e" stroke-width="3" fill="#cccccc" />\
                    Sorry, your browser does not support inline SVG.\
                </svg>',
                        props: {
                            svgwidth: {
                                name: 'Width',
                                type: 'slider',
                                slider_def_unit: '',
                                action: 'element_attribute',
                                attribute: 'width',
                                placeholder: '100',
                                show_if: function(values, pgel) { return pgel.tagName === 'svg'; }
                            },
                            svgheight: {
                                name: 'Height',
                                type: 'slider',
                                slider_def_unit: '',
                                action: 'element_attribute',
                                attribute: 'height',
                                placeholder: '100',
                                show_if: function(values, pgel) { return pgel.tagName === 'svg'; }
                            },
                            /*
                            svgx: {
                                name: 'X',
                                type: 'slider',
                                slider_def_unit: 'px',
                                action: 'element_attribute',
                                attribute: 'x'
                            },
                            svgy: {
                                name: 'Y',
                                type: 'slider',
                                slider_def_unit: 'px',
                                action: 'element_attribute',
                                attribute: 'y'
                            },*/
                            svgviewBox: {
                                name: 'viewBox',
                                type: 'slider',
                                slider_def_unit: '',
                                action: 'element_attribute',
                                attribute: 'viewBox',
                                placeholder: '0 0 100 100',
                                show_if: function(values, pgel) { return pgel.tagName === 'svg'; }
                            },
                            'preserveAspectRatio': {
                                name: 'Preserve aspect ratio',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'true',
                                attribute: 'preserveAspectRatio',
                                empty_attribute: true,
                                show_if: function(values, pgel) { return pgel.tagName === 'svg'; }
                            },
                            svgFillWithCurrent: {
                                name: 'Fill with current color',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'currentColor',
                                attribute: 'fill'
                            },
                            svgStrokeWithCurrent: {
                                name: 'Stroke with current color',
                                type: 'checkbox',
                                action: 'element_attribute',
                                value: 'currentColor',
                                attribute: 'stroke'
                            },
                        }
                    },
                    {
                        tag: 'math',
                        code: '<math><mrow><mi>a</mi><mo>⁢</mo><msup><mi>x</mi><mn>2</mn></msup><mo>+</mo><mi>b</mi><mo>⁢</mo><mi>x</mi><mo>+</mo><mi>c</mi></mrow></math>',
                        drag_helper: '<b>a + b</b>'
                    }
                ]},
            {'name' : 'Tables', key: 'html-tables', els: [
                {
                    tag:'table/p'
                },'caption','colgroup/p','col/p','tbody/p','thead/p','tfoot/p','tr/p',
                {
                    tag: 'td',
                    name: 'Td',
                    code: '<td>Data</td>',
                    props: {
                        colspan: {
                            name: 'Colspan',
                            type: 'slider',
                            slider_def_unit: '',
                            action: 'element_attribute',
                            attribute: 'colspan'
                        },
                        rowspan: {
                            name: 'Rowspan',
                            type: 'slider',
                            slider_def_unit: '',
                            action: 'element_attribute',
                            attribute: 'rowspan'
                        }
                    }
                },'th']},
            {'name' : 'Forms', key: 'html-forms', els: [
                {
                    tag: 'form/p',
                    name: 'Form',
                    props: {
                        'action' : {name: 'Action', file_picker: true, file_picker_no_proxy: true},
                        'method' : {name: 'Method', type: 'select', options: [
                            {key: 'post', name: 'post'},
                            {key: 'get', name: 'get'}
                        ]},
                        'name' : {name: 'Name'},
                        'target' : {name: 'Target'}
                    }
                },'fieldset/p','legend',{
                    tag: 'label',
                    props: {
                        'for' : {
                            name: 'For',
                            placeholder: 'input id'
                        }
                    }},
                {
                    tag: 'input/e',
                    name: 'Input',
                    props: {
                        'name' : {name: 'Name'},
                        'type' : {name: 'Type', type: 'select',
                            'options' : [
                                {key: 'text', name: 'Text'},
                                {key: 'password', name: 'Password'},
                                {key: 'checkbox', name: 'Checkbox'},
                                {key: 'radio', name: 'Radio'},
                                {key: 'number', name: 'Number'},
                                {key: 'email', name: 'Email'},
                                {key: 'file', name: 'File'},
                                {key: 'url', name: 'Url'},
                                {key: 'search', name: 'Search'},
                                {key: 'tel', name: 'Tel'},
                                {key: 'color', name: 'Color'},
                                {key: 'datetime', name: 'Datetime'},
                                {key: 'datetime-local', name: 'Datetime local'},
                                {key: 'date', name: 'Date'},
                                {key: 'month', name: 'Month'},
                                {key: 'time', name: 'Time'},
                                {key: 'week', name: 'Week'},
                                {key: 'range', name: 'range'},
                                {key: 'submit', name: 'submit'},
                                {key: 'hidden', name: 'Hidden'}
                            ]
                        },
                        checked: {
                            type: 'checkbox',
                            name: 'Checked',
                            action: 'element_attribute',
                            attribute: 'checked',
                            value: 'checked'
                        },
                        value: {
                            type: 'text',
                            name: 'Value',
                            action: 'custom',
                            get_value: function(pgel) {
                                return pgel.attr('value');
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                pgel.attr('value', value);
                                return value;
                            }
                        },
                        placeholder: {
                            type: 'text',
                            name: 'Placeholder',
                            action: 'element_attribute',
                            attribute: 'placeholder'
                        },
                        'value' : {name: 'Value'},
                        'size' : {name: 'Size'},
                        'maxlength' : {name: 'Max length'},
                        'required' : {
                            type: 'checkbox',
                            action: 'element_attribute',
                            attribute: 'required',
                            name: 'Required',
                            value: 'true'
                        },
                    }
                },
                {
                    tag: 'textarea/e',
                    name: 'Textarea',
                    props: {
                        'name' : {name: 'Name'},
                        placeholder: {
                            type: 'text',
                            name: 'Placeholder',
                            action: 'element_attribute',
                            attribute: 'placeholder'
                        },
                        'required' : {
                            type: 'checkbox',
                            action: 'element_attribute',
                            attribute: 'required',
                            name: 'Required',
                            value: 'true'
                        },
                    }
                },
                {
                    tag:'button',
                    name: 'Button',
                    code: '<button>Click me!</button>',
                    props: {
                        'name' : {name: 'Name'},
                        'value' : {name: 'Value'},
                        'buttontype' : {name: 'Type', type: 'select', attribute: 'type', show_empty: true, options: [
                            {key: 'submit', name: 'Submit'},
                            {key: 'button', name: 'Button'},
                            {key: 'reset', name: 'Reset'}
                        ]}
                    }
                },
                {
                    tag:'select/e',
                    name:'Select',
                    props: {
                        'name' : {name: 'Name'},
                        'value' : {name: 'Value'},
                        multiple: {
                            type: 'checkbox',
                            name: 'Multiple',
                            action: 'element_attribute',
                            attribute: 'multiple',
                            value: 'multiple'
                        }
                    }
                },
                {
                    tag:'option/e',
                    props: {
                        'text' : {name: 'Text', action: 'element_html'},
                        'value' : {name: 'Value'}
                    }
                },
                'datalist/e','optgroup/e','keygen/e','output/e','progress/e','meter/e']},
            {'name' : 'Interactive elements', key: 'html-interactive', els: ['details','summary','menuitem','menu/p']},
            {'name' : 'Head elements', key: 'html-head', els: ['title',
                {tag: 'meta-author', 'name' : 'Meta author', code: '<meta name="author" content="name">', selector: 'meta[name="author"]'},
                {tag: 'meta-description', 'name' : 'Meta description', code: '<meta name="description" content="name">', selector: 'meta[name="description"]'},
                {tag: 'meta-keywords', 'name' : 'Meta keywords', code: '<meta name="keywords" content="name">', selector: 'meta[name="keywords"]'}
            ]},
            {'name' : 'Extras', skip_lib: true, key: 'html-extras', els:
                ['body',
                    {tag: 'tag', 'name' : 'Tag', selector_new: function(pgel) {return true}, display_name: 'tag', priority: 2001}]}
        ];
        var single = ['area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'];


        var bckimage = {
            type : 'image',
            name: 'Bck. image',
            action : 'custom',
            file_picker_no_proxy: true,
            can_make_change_action: 'element_bckimage',
            get_value: function(pgel) {
                return pgel.getInlineStylePropertyValue('background-image', true);

            },
            set_value: function(pgel, value, values, oldValue, eventType) {
                pgel.setInlineStylePropertyValue('background-image', value, true);

                return value;
            }
        }

        var capitalize = function(s) {
            return s;
            return s.charAt(0).toUpperCase() + s.slice(1)
        }

        var html_sections = null;//crsaAddStandardSections({});

        //add some sections that are most useful
        var section_layout = new PgFrameworkLibSection('html.top.layout', 'Layout');
        f.addLibSection(section_layout);

        var section_content = new PgFrameworkLibSection('html.top.content', 'Content');
        f.addLibSection(section_content);

        var section_media = new PgFrameworkLibSection('html.top.media', 'Media');
        f.addLibSection(section_media);

        var getLinkElement = function(pgel) {
            return pgel.getLinkElement();
        }

        var createLink = function(pgel, $button, values, $fc, force_loc) {
            var grid_section = $fc.data('propertyField').grid_section;
            link_sec_def.closed = false;
            grid_section.openClose(true, true);

            var $panel = $button.closest('.pg-grid');
            crsaFuncs.addLinkAroundElement(pinegrow.getCollectionForElement(pgel), force_loc );
            pinegrow.selectedElements.reselect();
            setTimeout(function() {
                $panel.find('.crsa-field-html-link-href input').get(0).focus();
            }, 10)
        }

        var link_sec_def = {
            name: '&lt;a&gt; Link',
            position: 0,
            default_closed: pgf.kids || false,
            show_if: function(pgel) {
                return pgel.closest('head') === null && no_link_for.indexOf(pgel.tagName) < 0;
            },
            aaaforce_open: function(pgel) {
                if(getLinkElement(pgel)) {
                    return null;
                } else {
                    return 'closed';
                }
            },
            fields: {

                'html.link.create': {
                    type: 'button',
                    aaashow_button_in_section_title: function(pgel) {
                        return getLinkElement(pgel) === null;
                    },
                    name: 'Create link',
                    func: function(pgel, $button, values, $fc, event) {
                        event.stopPropagation();
                        createLink(pgel, $button, values, $fc);
                    },
                    helptext: 'Create the link around or inside (for headings and paragraphs) the element. Right-click for options.',
                    on_button_context_menu: function(pgel, $button, values, $fc) {
                        return [
                            {
                                type: 'header',
                                label: 'Create link'
                            },
                            {
                                label: 'Around the element',
                                action: function() {
                                    createLink(pgel, $button, values, $fc, 'outside');
                                }
                            },
                            {
                                label: 'Inside the element',
                                action: function() {
                                    createLink(pgel, $button, values, $fc, 'inside');
                                }
                            }
                        ]
                    },
                    show_if: function(values, pgel, $field) {
                        return getLinkElement(pgel) === null;
                    }
                },
                'html.link.href' : {
                    name: 'Href',
                    type: 'text',
                    file_picker: isApp(),
                    file_picker_no_proxy: true,
                    action: 'custom',
                    can_make_change_action: 'element_attribute',
                    attribute: 'href',
                    can_make_change_element: getLinkElement,
                    set_value: function(pgel, value) {
                        if(value !== null && value.length) {
                            var a = getLinkElement(pgel);
                            if (!a && value.length) {
                                crsaFuncs.addLinkAroundElement(pgel, true /* keep this one selected */);
                                a = getLinkElement(pgel);
                            }
                            if (a) {
                                a.setAttr('href', value);
                            }
                        }
                        pinegrow.dispatchEvent('on_element_link_changed', pgel.getPage());
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.getAttr('href');
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return getLinkElement(pgel) !== null;
                    },
                    on_changed: function(pgel, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
                        if(eventType === 'change') {
                            //pinegrow.selectedElements.reselect();
                        }
                    }
                },
                'html.link.target' : {
                    name: 'Open in new tab',
                    type: 'checkbox',
                    value: '1',
                    action: 'custom',
                    can_make_change_action: 'element_attribute',
                    attribute: 'target',
                    can_make_change_element: getLinkElement,
                    set_value: function(pgel, value) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            if(value === '1') {
                                a.setAttr('target', '_blank');
                            } else {
                                a.removeAttribute('target');
                            }
                        }
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.getAttr('target') === '_blank' ? '1' : null;
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return pgf.kids && getLinkElement(pgel) !== null;
                    }
                },

                'html.link.target2' : {
                    name: 'Target',
                    type: 'text',
                    action: 'custom',
                    options: [
                        { key: '_blank', name: '_blank'},
                        { key: '_parent', name: '_parent'},
                        { key: '_top', name: '_top'},
                    ],
                    open_list_on_focus: true,
                    can_make_change_action: 'element_attribute',
                    attribute: 'target',
                    can_make_change_element: getLinkElement,
                    set_value: function(pgel, value) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            if(value) {
                                a.setAttr('target', value);
                            } else {
                                a.removeAttribute('target');
                            }
                        }
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.getAttr('target');
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return !pgf.kids && getLinkElement(pgel) !== null;
                    }
                },
                'html.link.rel' : {
                    name: 'Rel',
                    type: 'text',
                    action: 'custom',
                    can_make_change_action: 'element_attribute',
                    attribute: 'rel',
                    can_make_change_element: getLinkElement,
                    set_value: function(pgel, value) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            if(value) {
                                a.setAttr('rel', value);
                            } else {
                                a.removeAttribute('rel');
                            }
                        }
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.getAttr('rel');
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return !pgf.kids && getLinkElement(pgel) !== null;
                    }
                },
                'html.link.name' : {
                    name: 'Name',
                    type: 'text',
                    action: 'custom',
                    open_list_on_focus: true,
                    can_make_change_action: 'element_attribute',
                    can_make_change_element: getLinkElement,
                    attribute: 'name',
                    set_value: function(pgel, value) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            if(value) {
                                a.setAttr('name', value);
                            } else {
                                a.removeAttribute('name');
                            }
                        }
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.getAttr('name');
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return !pgf.kids && getLinkElement(pgel) !== null && !values['html.link.href']
                    }
                },
                'html.link.label' : {
                    name: 'Inner HTML',
                    type: 'text',
                    action: 'custom',
                    open_list_on_focus: true,
                    can_make_change_action: 'element_html',
                    can_make_change_element: getLinkElement,
                    set_value: function(pgel, value, values, oldValue, eventType, fieldDef, $field) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            if(a !== pgel && a.contains(pgel)) {
                                //the a contains the selected element, that will be destroyed with the call to a.html()
                                setTimeout(function() {
                                    pinegrow.selectElement(a);
                                    pinegrow.showQuickMessage('Selected the parent &lt;a&gt; element.');
                                    setTimeout(function() {
                                            $field.find('input').get(0).focus();
                                        }, 10)

                                },0)
                            }
                            a.html(value || '');
                        }
                        return value;
                    },
                    get_value: function(pgel) {
                        var a = getLinkElement(pgel);
                        if(a) {
                            return a.html();
                        }
                        return null;
                    },
                    show_if: function(values, pgel, $field) {
                        return !pgf.kids && getLinkElement(pgel) !== null;
                    }
                },
                'html.link.remove': {
                    type: 'button',
                    name: 'Remove the link',
                    func: function(pgel) {
                        var col = pinegrow.getCollectionForElement(pgel);
                        var a_col = new PgCollection();
                        var select_col = new PgCollection();

                        col.forEachElement(function(pgel) {
                            var a = getLinkElement(pgel);
                            if(a) {
                                a_col.add(a);
                                if(a !== pgel) {
                                    select_col.add(pgel)
                                } else {
                                    var children = a.getChildren();
                                    if(children.length === 1) {
                                        select_col.add(children[0])
                                    } else {
                                        select_col.add(a.parent);
                                    }
                                }
                            }
                        })

                        crsaFuncs.detagElement(a_col);
                        pinegrow.selectedElements.selectCollection(select_col);
                    },
                    show_if: function(values, pgel, $field) {
                        return getLinkElement(pgel) !== null;
                    }
                }
            }
        }

        var no_link_for = ['html', 'head', 'body', 'meta', 'base', 'link', 'br', 'hr', 'script', 'input', 'select', 'textarea', 'option', 'optgroup', 'table', 'tbody', 'thead', 'tfoot', 'tr', 'colgroup', 'ul', 'ol', 'source'];

        $.each(html5, function(i, sec) {
            var section = null;

            if(!just_essential) {
                section = new PgFrameworkLibSection(sec.key + '.2', sec.name);
                section.closed = true; //closed by default
            }

            if(!sec.skip_lib) {
                if(!simple && section) {
                    f.addLibSection(section);
                }
            }

            $.each(sec.els, function(j, tag) {

                var opts = {};
                if(typeof tag == 'object') {
                    opts = tag;
                    tag = opts.tag;
                }

                var a = tag.split(':');
                tag = a[0];
                opts.inner = a.length > 1 ? a[1] : null;

                if(a.length > 2 && !opts.name) {
                    opts.name = a[2];
                }

                var is_plc = tag.indexOf('/p') >= 0;
                var is_emp = tag.indexOf('/e') >= 0;
                if(is_plc) is_emp = true;
                tag = tag.replace('/e', '').replace('/p', '');
                var is_single = single.indexOf(tag) >= 0;

                var preview_code = is_single ? '<' + tag + '>' : '<' + tag + '>' + tag + '</' + tag + '>';

                var cls = is_plc ? ' class="pg-empty-placeholder"' : '';

                if(pgf.empty_ph_attribute && is_plc) {
                    cls = ' data-empty-placeholder';
                }

                var inner = is_emp ? '' : (opts.inner ? opts.inner : tag);

                var code = is_single ? '<' + tag + cls + '>' : '<' + tag + cls + '>' + inner + '</' + tag + '>';

                var def = new PgComponentType('html-' + tag, opts.name ? opts.name.toLowerCase() : capitalize(tag));

                def.selector = opts.selector ? opts.selector : tag;
                def.selector_new = opts.selector_new ? opts.selector_new : null;
                def.code = opts.code ? opts.code : code;
                def.preview = preview_code;
                def.empty_placeholder = opts.empty_placeholder || is_plc;
                def.priority = opts.property ? opts.property : 2000;
                def.parent_selector = opts.parent_selector ? opts.parent_selector : null;

                if(opts.drag_helper) def.drag_helper = opts.drag_helper;
                if(opts.action_menu) def.action_menu = opts.action_menu;
                if(opts.preview_description) def.preview_description = opts.preview_description;

                def.display_name = opts.display_name ? opts.display_name : null;

                if(opts.on_inserted) def.on_inserted = opts.on_inserted;
                if(opts.on_deleted) def.on_deleted = opts.on_deleted;

                if(opts.file_picker) def.file_picker = true;

                if(tag === 'tag') {
                    default_def = def;
                } else {
                    tag_to_def_map[tag] = def;
                }

                def.sections = {
                    'pg-classes' : {
                        position: 1,
                        default_closed: pgf.kids || false,
                        name: 'Classes',
                        fields: {
                            'rules' : {
                                'type' : 'rules',
                                'name' : null,
                                'action' : 'rules'
                            }
                            /*,

                             'html.bck.image' : bckimage,
                             'style' : {
                             'type' : 'text',
                             'name' : 'Style',
                             'action' : 'element_attribute',
                             attribute: 'style'
                             }*/

                            /*,
                             'classes': {
                             type: 'select',
                             name: 'Classes',
                             action: 'custom',
                             multiple: true,
                             can_add_items: true,
                             options: function() {
                             return pinegrow.insight.getClasses();
                             },
                             get_value: function(obj) {
                             var $el = obj.data;
                             var pgel = new pgQuery($el);
                             return (pgel.attr('class') || '').split(' ').join(',');
                             },
                             set_value: function(obj, value, values, oldValue, eventType) {
                             var $el = obj.data;
                             var pgel = new pgQuery($el);
                             if(value) {
                             pgel.attr('class', value.replace(/\,/g, ' '));
                             } else {
                             pgel.removeAttr('class');
                             }
                             return value;
                             }
                             } */
                        }
                    },
                    'pg-info' : {
                        position: 1,
                        default_closed: pgf.kids || false,
                        name: 'Info',
                        fields: {
                            'element_id': {
                                'type': 'text',
                                'name': 'Id',
                                action: 'custom',
                                options: function () {
                                    return pinegrow.insight.getIds();
                                },
                                can_make_change_action: 'element_id',
                                live_update: false,
                                get_value: function (pgel) {
                                    return pgel.attr('id');
                                },
                                set_value: function (pgel, value, values, oldValue, eventType) {
                                    var same = false;
                                    var oldsame = false;
                                    if (!value) {
                                        pgel.removeAttr('id');
                                        // $.fn.crsa('setNeedsUpdate', false, $el);todo30
                                        return value;
                                    }
                                    if (value) {
                                        var ex = pgel.closest('body').find('#' + value);
                                        if (ex.length > 0 && ex != pgel) {
                                            crsaQuickMessage("Another element already has the id <b>" + value + "</b>. Search the tree to find it.", 4000, false);
                                            same = true;
                                        }
                                    }
                                    if (oldValue) {
                                        var ex = pgel.closest('body').find('#' + oldValue);
                                        if (ex.length > 0 && ex != pgel) {
                                            //crsaQuickMessage("Another element already has the id <b>" + value + "</b>. Search the tree to find it.", 4000, false);
                                            oldsame = true;
                                        }
                                    }
                                    pgel.attr('id', value);
                                    if (pgel.isSelector('input,select,textarea')) {
                                        if (!oldValue) oldValue = "";
                                        var lab = pgel.closest('label[for="' + oldValue + '"]');
                                        if (lab.length == 0) {
                                            lab = pgel.parent.find('label[for="' + oldValue + '"]');
                                        }
                                        if (lab.length == 0 && !same) {
                                            lab = pgel.closest('body').find('label[for="' + oldValue + '"]');
                                        }
                                        if (lab.length == 1 && !same && !oldsame) {
                                            lab.attr('for', value);
                                            crsaQuickMessage("Label's <b>&quot;for&quot;</b> attribute updated to <b>" + value + "</b>", 4000);
                                        } else if (lab.length > 1) {
                                            crsaQuickMessage("Multiple labels for <b>" + oldValue + "</b> exist. Please update <b>&quot;for&quot;</b> attributes manually.", 4000);
                                        }
                                    }
                                    //$.fn.crsa('setNeedsUpdate', false, $el);todo30
                                    return value;
                                }
                            },
                            'pinegrow_name': {
                                'type': 'text',
                                'name': 'Name',
                                'action': 'element_attribute',
                                attribute: 'data-pg-name',
                                placeholder: 'name shown in the tree',
                                update_tree: true
                            },
                            'title': {
                                'type': 'text',
                                'name': 'Title',
                                'action': 'element_attribute',
                                attribute: 'title'
                            },
                            /*
                            'empty_placeholder': {
                                'type': 'checkbox',
                                'name': 'Empty placeholder',
                                'action': 'apply_class',
                                value: 'pg-empty-placeholder'
                            }*/
                        }
                    }

                };

                if(opts.major) {
                    def.tags = 'major';
                }

                if(pgf.kids) {
                    delete def.sections['pg-info'].fields['pinegrow_name'];
                    delete def.sections['pg-info'].fields['empty_placeholder'];
                }

                //Can have link?
                // if(no_link_for.indexOf(tag) < 0) {
                def.sections['pg.html.link'] = link_sec_def;
                //  }

                if(opts.props) {
                    def.sections['attributes'] = {
                            name: '&lt;' + tag + '&gt;' + (def.name != tag ? (' ' + def.name) : ''),
                            position: 0,
                            fields : {}
                        }

                    $.each(opts.props, function(pkey, pdef) {
                        var attr_key = pkey;
                        if(!pdef.type) pdef.type = 'text';
                        if(!pdef.action) {
                            pdef.action = 'element_attribute';
                            if(!pdef.attribute) {
                                pdef.attribute = pkey;
                                attr_key = tag + '.' + pkey;
                            } else {

                            }
                        }
                        def.sections.attributes.fields[attr_key] = pdef;
                    });
                }

                /*def.sections['Info'] = {

                };*/

                def.sections['pg-attributes'] = {
                    name: 'Attribute editor',
                    default_closed: true,
                    position: 2,
                    fields: {
                        'pg_attributes_editor' : {
                            type: 'custom',
                            name: 'attributes_editor',
                            action: 'none',
                            control: pgCreateAttributesEditorControl
                        }
                    }
                };

                f.addComponentType(def);

                if(section) {
                    section.addComponentType(def);
                }
            });
        });

        //body/root

        var prefix = 'html-';

        var body = {
            'type' : 'body',
            'selector' : 'body',
            'name' : 'Body',
            action_menu : {
                add : [prefix + 'header', prefix + 'main', prefix + 'section', prefix + 'footer', prefix + 'navbar'],
                on_add : function(pgel, pgnew, newdef, prepend) {
                    if (newdef.type == prefix + 'header' || newdef.type == prefix + 'navbar') {
                        pgel.prepend(pgnew);
                    }
                    else if (newdef.type == prefix + 'footer') {
                        pgel.append(pgnew);
                    } else {
                        if (prepend) {
                            pgel.prepend(pgnew);
                        } else {
                            pgel.append(pgnew);
                        }
                    }
                }
            }
        }
        f.addComponentType(body);
        tag_to_def_map['body'] = body;

        //Add common most useful elements

        var getDefList = function(l) {
            var r = [];
            l.forEach(function(d) {
                if(tag_to_def_map[d]) {
                    r.push(tag_to_def_map[d])
                } else {
                    console.error(`HTML: Definition for ${d} not found.`);
                }
            })
            return r;
        }

        section_layout.setComponentTypes(getDefList(['div', 'section', 'main', 'header', 'footer', 'nav']))
        section_content.setComponentTypes(getDefList(['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'img', 'picture', 'a', 'button', 'span', 'ul', 'ol', 'li']));
        section_media.setComponentTypes(getDefList(['iframe', 'video', 'audio']));

        var head = {
            'type' : 'head',
            'selector' : 'head',
            'name' : 'Head',
        }
        f.addComponentType(head);
        tag_to_def_map['head'] = head;

        var getPgQueryForHeadElement = function(head, sel, code) {
            if (head.is('html')) {
                var h = head.findOne('head');
                if(!h) {
                    h = pgCreate('<head></head>');
                    head.prepend(h);
                }
                head = h;
            }
            var e = head.find(sel);
            if(!e.length) {
                if(!code) return null;
                var pgel = pgCreateNodeFromHtml(code);
                head.append(pgel);
                return pgel;
            }
            return e[0];
        }

        var get$Head = function($el) {
            return $el.find('head');
        }

        var getMetaValue = function(pgel, key) {
            var pgel = getPgQueryForHeadElement(pgel, 'meta[name="' + key + '"]');
            return pgel ? pgel.attr('content') : null;
        }

        var setMetaValue = function(pgel, key, value) {
            var head = pgel;
            if (pgel.is('html')) head = pgel.findOne('head');

            var pgel = getPgQueryForHeadElement(head, 'meta[name="' + key + '"]', '<meta name="' + key + '">');
            if(value === null) {
                pgel && pgel.remove();
            } else {
                pgel.attr('content', value);
            }
        }

        var getFaviconElement = function(pgel, create) {
            var head = pgel.findOne('head');
            if(head) {
                var fs = head.find('link[rel]');
                var f = null;
                for(var i = 0; i < fs.length; i++) {
                    if((fs[i].getAttribute('rel') || '').split(' ').indexOf('icon') >= 0) {
                        return fs[i];
                        break;
                    }
                }
                if (create) {
                    f = pgCreate(`<link rel="icon">`);
                    f.appendTo(head);
                    return f;
                }
            }
            return null;
        }

        var html = {
            'type' : 'html',
            'selector' : 'html',
            'name' : 'Page',
            'sections' : {
                'html.page' : {
                    position: 0, //at the top of props
                    name: 'Page',
                    fields: {
                        'page.title' : {
                            'type' : 'text',
                            'name' : 'Title',
                            'action' : 'custom',
                            can_make_change_action: 'element_html',
                            can_make_change_element: function(pgel) {
                                return pgel.findOne('title');
                            },
                            options : function() {
                                return pinegrow.insight.getContentsForTag('title');
                            },
                            get_value: function(pgel) {
                                var pgel = getPgQueryForHeadElement(pgel, 'title');
                                return pgel ? pgel.html() : null;
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                var pgel = getPgQueryForHeadElement(pgel, 'title', '<title></title>');
                                if(value === null) {
                                    pgel && pgel.remove();
                                } else {
                                    pgel.html(value);
                                }
                                //if(eventType == 'change') pinegrow.updateTree(get$Head(obj.data));todo30
                                return value;
                            }
                        },
                        'page.description' : {
                            'type' : 'text',
                            'name' : 'Description',
                            'action' : 'custom',
                            can_make_change_action: 'element_attribute',
                            attribute: 'content',
                            can_make_change_element: function(pgel) {
                                return pgel.findOneWithAttrValue('name', 'description', 'meta');
                            },
                            options : function() {
                                return pinegrow.insight.getMetaValuesForName('description');
                            },
                            get_value: function(pgel) {
                                return getMetaValue(pgel, 'description');
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                setMetaValue(pgel, 'description', value);
                                //if(eventType == 'change') pinegrow.updateTree(get$Head(obj.data));todo30
                                return value;
                            }
                        },
                        'page.keywords' : {
                            'type' : 'text',
                            //'multiple' : true,
                            //can_add_items: true,
                            'name' : 'Keywords',
                            'action' : 'custom',
                            can_make_change_action: 'element_attribute',
                            attribute: 'content',
                            can_make_change_element: function(pgel) {
                                return pgel.findOneWithAttrValue('name', 'keywords', 'meta');
                            },
                            'options' : function() {
                                return pinegrow.insight.getKeywords();
                            },
                            get_value: function(pgel) {
                                return getMetaValue(pgel, 'keywords');
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                setMetaValue(pgel, 'keywords', value);
                                //if(eventType == 'change') pinegrow.updateTree(get$Head(obj.data));todo30
                                return value;
                            }
                        },
                        'page.author' : {
                            'type' : 'text',
                            'name' : 'Author',
                            'action' : 'custom',
                            can_make_change_action: 'element_attribute',
                            attribute: 'content',
                            can_make_change_element: function(pgel) {
                                return pgel.findOneWithAttrValue('name', 'author', 'meta');
                            },
                            options : function() {
                                return pinegrow.insight.getMetaValuesForName('author');
                            },
                            get_value: function(pgel) {
                                return getMetaValue(pgel, 'author');
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                setMetaValue(pgel, 'author', value);
                                //if(eventType == 'change') pinegrow.updateTree(get$Head(obj.data));todo30
                                return value;
                            }
                        },
                        'page.favicon': {
                            'type' : 'text',
                            'name' : 'Favicon',
                            'action' : 'custom',
                            file_picker: true,
                            file_picker_no_proxy: true,
                            can_make_change_action: 'element_attribute',
                            attribute: 'content',
                            can_make_change_element: function(pgel) {
                                return pgel.findOneWithAttrValue('name', 'author', 'meta');
                            },
                            options : function() {
                                return pinegrow.insight.getMetaValuesForName('author');
                            },
                            get_value: function(pgel) {
                                var f = getFaviconElement(pgel);
                                if(f) {
                                    return f.getAttribute('href')
                                } else {
                                    return null;
                                }
                            },
                            set_value: function(pgel, value, values, oldValue, eventType) {
                                if(value) {
                                    var f = getFaviconElement(pgel, true);
                                    f.setAttribute('href', value);
                                    var mt = pinegrow.getMimeType(value);
                                    f.setAttribute('type', mt || '');
                                } else {
                                    var f = getFaviconElement(pgel);
                                    f && f.remove();
                                }
                                return value;
                            },
                            validate: function(pgel, prop, value) {
                                if(value) {
                                    var mt = pinegrow.getMimeType(value);
                                    if(!mt || !mt.startsWith('image/')) {
                                        return 'Favicon should be an image.'
                                    }
                                }
                            }
                        },
                    }
                }
            }
        }
        f.addComponentType(html);
        tag_to_def_map['html'] = html;

        //Register starting page template
        f.addTemplateProjectFromResourceFolder('template', null, 0, function (node) {
            node.required = node.name !== 'normalize.css' && !node.name.endsWith('.html');
        });

        var design_provider = new PgDmDesignProviderForPlainHTML(f);
        pgDmStore.registerDesignProvider('plainhtml', design_provider);

        f.on_dm_detect_design_providers = function(page, list) {
            list.push('plainhtml');
        }
    });

});
