class PgDmDesignProvider {
    constructor(key, name) {
        this.key = key || 'generic';
        this.name = name || 'Generic design provider';
        this.skills = [];
        this.engine = null;
        this.colors = null;
        this.settings = {};
        this.settings_key = 'design-settings';

        this.setDefaultSettings();
    }

    setDefaultSettings() {
        this.settings = {}
    }

    setSettings(settings, skip_update) {
        //this.setDefaultSettings();
        var _this = this;
        $.each(settings, function(key, val) {
            _this.settings[key] = val;
        })
        if(!skip_update) {
            pgDmStore.update();
        }
        var pi = pinegrow.getCurrentProjectInfo();
        pi.setSetting(this.settings_key, this.settings);
        pi.save();
    }

    firstActivateInProject(done) {
        done && done();
    }

    activate() {
        if(this.engine) {
            this.engine.destroy();
            this.engine = null;
        }

        this.setDefaultSettings();

        var pi = pinegrow.getCurrentProjectInfo();
        var settings = pi.getSetting(this.settings_key, null);
        if(settings) {
            this.settings = $.extend(this.settings, settings);
        }
        //this.getEngine(); //initialize engine
    }

    onProjectLoaded(project) {

    }

    onColorListChanged() {}

    getEngine(done) {
        console.error('PgDmDesignProvider.getEngine should be implemented by provider.')
    }

    addSkill(skill) {
        this.skills.push(skill);
    }

    getSkills() {
        return this.skills;
    }

    getColorsForPicker() {
        return this.colors.getColorsForPicker();
    }

    canEditColors() {
        return false;
    }

    supportsColorShades() {
        return false;
    }

    getActiveFonts() {
        return [];
    }
}

var PgDmSkill = PgDmAddOn = function(key, name, init) {

    var _this = this;

    this.key = key;
    this.name = name;
    this.description = null;
    this.controls = {};
    this.params = {};

    this.value_sources = {};

    var section = {
        key: key,
        name: name,
        fields: this.controls
    }

    pgDmStore.registerSkill(this);

    this.isSet = function(params) {
        var has = false;
        $.each(params, function(key, val) {
            if(val !== null) {
                has = true;
                return false;
            }
        })
        /*
        $.each(this.controls, function(key, def) {
            if(def.subkey in params && params[def.subkey] !== null) {
                has = true;
            }
        })*/
        return has;
    }

    this.getParam = function(subkey, params) {
        return params[subkey] || null;
    }

    this.setParam = function(subkey, value) {
        var key = this.key + '.' + subkey;
        this.params[key] = value;
    }

    this.addControl = function(subkey, def) {
        def.key = this.key + '.' + subkey;
        this.controls[def.key] = def;

        def.subkey = subkey;

        def.action = 'custom';
        def.set_value = this.controlSetValue.bind(this);
        def.get_value = this.controlGetValue.bind(this);

        def.on_field_updated = function(pgel, $field, fdef, control) {
            var info;
            var field = $field.get(0);
            var subfield = false;
            var first_field = true;

            if(field.parentNode.classList.contains('crsa-field')) {
                subfield = true;
                first_field = !field.previousSibling || field.previousSibling.nodeName === 'LABEL';
            }

            var label = field.querySelector('label');
            var parent_label = null;

            if(field.classList.contains('inherited-value')) {
                field.classList.remove('inherited-value');
                if(label) label.removeAttribute('title');
            }

            if(subfield && first_field) {
                field.parentNode.classList.remove('inherited-value');
                parent_label = field.parentNode.querySelector('label');
                if(parent_label) parent_label.removeAttribute('title');
            }

            if(info = _this.value_sources[fdef.subkey]) {
                if(info.type === 'size') {
                    var label_txt = `Set for ${info.size.name_short} and up.`;

                    field.classList.add('inherited-value');
                    if(label) label.setAttribute('title', label_txt);

                    if(subfield) {
                        field.parentNode.classList.add('inherited-value');
                        if(parent_label) parent_label.setAttribute('title', label_txt);
                    }
                }
            } else {

            }
        }

        def.on_changed = function(obj, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
            var field = $field.get(0);
            var subfield = false;
            var first_field = true;

            if(field.parentNode.classList.contains('crsa-field')) {
                subfield = true;
            }

            if(field.classList.contains('inherited-value')) {
                field.classList.remove('inherited-value');
            }
            if(subfield) {
                if(field.parentNode.querySelector('.inherited-value')) {
                    field.parentNode.classList.add('inherited-value');
                } else {
                    field.parentNode.classList.remove('inherited-value');
                }
            }
        }

        if(def.type == 'color') {
            def.palette = function() {
                return pgDmStore.design.getColorPalette();
            }
        }
    }

    this.getControl = function(subkey) {
        return this.controls[this.key + '.' + subkey] || null;
    }

    this.getPropSection = function(pgel) {
        return section;
    }

    this.controlSetValue = function(pgel, value, values, oldValue, eventType, fdef) {
        var _this = this;
        pgDmStore.setSkillParam(pgel, _this, fdef, value);
        return value;
    }

    this.controlGetValue = function(pgel, fn, values, fdef, skip_prev_value) {
        var _this = this;
        this.value_sources[fdef.subkey] = null;

        var val = pgDmStore.getSkillParam(pgel, _this, fdef);
        if(val !== null) {
            this.value_sources[fdef.subkey] = {type: 'own'};
            return val;
        }
        if(!skip_prev_value) {
            //get value for same control at lower device size
            var sizes = pgDmStore.design.getDeviceSizes();
            var ignore = true;
            var current_size = pgDmStore.getCurrentSize();
            for(var i = sizes.length - 1; i >= 0; i--) {
                if(sizes[i].size === current_size) {
                    ignore = false;
                } else if(!ignore) {
                    val = pgDmStore.getSkillParam(pgel, _this, fdef, sizes[i].size);
                    if(val !== null) {
                        this.value_sources[fdef.subkey] = {type: 'size', size: sizes[i]};
                        return val;
                    }
                }
            }
            //get value for some control in preceeding style
        }
        return null;
    }

    this.getCSS = function(params, style) {
        return this.onGetCSS(params, style);
    }

    this.getConfigObject = function(params, config) {
        this.onGetConfigObject && this.onGetConfigObject(params, config);
    }

    this.getHTMLForStyleDisplay = function(params) {
        return '';
    }

    if(init) {
        init.call(this);
    }
    //Add controls to DM

    //Add CSS

    //Add Properties to panel
}

class PgDmDesignSkill {

    constructor(key, name, init)
    {
        //this.node = pgCreate(`<dmdesignskill skill="${key}"></dmdesignskill>`);

        this.node = null;
        this.key = key;
        this.name = name;
        this.description = null;
        this.controls = {};
        this.params = {};

        this.locked = {}
        this.$locks = {}
        this.fields = {}

        this.value_sources = {};

        this.section = {
            key: key,
            name: name,
            fields: this.controls
        }

        //pgDmStore.registerDesignSkill(this);

        if(init) {
            init.call(this);
        }
    }


    load(node) {
        var _this = this;

        this.node = node;

        //this.populateDefaultParams();

        //get locks
        $.each(this.controls, function(key, def) {
            var locked = pgDmStore.isParamLocked(_this, def.subkey, def.dm_locked || false);
            _this.updateLockedStatusForControl(def.subkey, locked);
        })

        if(this.onLoad) {
            this.onLoad(node);
        } else {
            $.each(this.controls, function(key, def) {
                var val = node.getAttribute(def.subkey);
                if(val === null && ('default_value' in def)) {
                    val = def.default_value;
                }
                _this.params[def.subkey] = val;
                pgDmStore.design.setDesignParameter(_this, def.subkey, val);
            })
        }
    }

    populateDefaultParams() {

    }

    getParam(subkey, def) {
        var val = this.params[subkey] || def || null;

        this.onGetParam && (val = this.onGetParam(subkey, val));

        if(val === null) {
            var key = this.key + '.' + subkey;
            if (this.controls[key]) {
                if (this.controls[key].default_value) {
                    return this.controls[key].default_value;
                }
            }
        }
        return val;
    }

    setParam(subkey, value, skip_node_update) {
        var _this = this;

        this.params[subkey] = value;
        pgDmStore.design.setDesignParameter(this, subkey, value);

        if(skip_node_update) return;

        pgDmStore.db.makeChanges(null, 'Set design parameter', function() {
            if(_this.onSetParam) {
                _this.onSetParam(subkey, value);
            } else {
                _this.node.setAttribute(subkey, value);
            }
        }, 'design_panel');
    }

    surpriseMeForRangeControls(done) {
        var _this = this;

        $.each(this.controls, function(key, def) {
            if(!def.dm_no_lock && def.dm_no_lock !== 'placeholder') {
                if (def.type === 'range' && !_this.isLocked(def.subkey)) {
                    _this.setParam(def.subkey, getRand(
                        ('magic_range_min' in def) ? def.magic_range_min : def.range_min,
                        ('magic_range_max' in def) ? def.magic_range_max : def.range_max
                    ))
                }
            }
        })
        done && done();
    }

    surpriseMe(done) {
        return this.onSurpriseMe ? this.onSurpriseMe(done) : (done && done());
    }

    isLocked(subkey) {
        return this.locked[subkey] || false;
    }

    updateLockedStatusForControl(subkey, val) {
        this.locked[subkey] = val;
        if(this.$locks[subkey]) {
            if(val) {
                this.$locks[subkey].addClass('like');
            } else {
                this.$locks[subkey].removeClass('like');
            }
        }
    }

    addControl(subkey, def) {
        var _this = this;

        def.key = this.key + '.' + subkey;
        this.controls[def.key] = def;

        def.subkey = subkey;

        def.action = 'custom';
        def.set_value = this.controlSetValue.bind(this);
        def.get_value = this.controlGetValue.bind(this);

        def.skip_did_change = true;

        if(def.type === 'checkbox') {
            //def.live_update = false;
        }

        if(def.dm_locked) {
            this.locked[subkey] = true;
        }

        var addLock = function($fc) {
            var name = name || 'Lock';
            _this.$locks[subkey] = $('<div class="pg-dm-like pg-dm-lock-' + subkey + (_this.locked[subkey] ? ' like' : '') + '"><i class="closed icon icon-Lock_open"></i><i class="open icon icon-Lock_closed"></i></div>').appendTo($fc)
                .on('click', function(e) {
                    e.preventDefault();
                    var like = !(_this.locked[def.subkey] || false);
                    _this.locked[def.subkey] = like;
                    pgDmStore.setParamLocked(_this, def.subkey, like);
                    if(like) {
                        $(this).addClass('like');
                    } else {
                        $(this).removeClass('like');
                    }
                })
            addTooltip(_this.$locks[subkey], 'Lock to keep Surprise Me from changing this design option.')
        }

        var addEmptyLock = function($fc) {
            $('<div class="pg-dm-like pg-dm-lock-' + subkey + '"><i class="open icon icon-Lock_closed" style="visibility:hidden;"></i></div>').appendTo($fc);
        }

        def.on_field_updated = function(pgel, $field, fdef, control) {
        }


        def.on_fields_created = function(obj, name, val, def, $field, values, recycled, propField) {
            if(!recycled && (def.type !== 'button' || (('dm_no_lock' in def) && def.dm_no_lock === false))) {
                if(!def.dm_no_lock) {
                    addLock($field);
                } else if(def.dm_no_lock === 'placeholder') {
                    addEmptyLock($field);
                }
            }
            _this.fields[subkey] = propField;
        }

        if(def.default_value) {

        }
    }

    getControl(subkey) {
        return this.controls[this.key + '.' + subkey] || null;
    }

    getField(subkey) {
        return this.fields[subkey] || null;
    }

    getSection() {
        return this.section;
    }

    controlSetValue(pgel, value, values, oldValue, eventType, fdef) {
        var _this = this;
        this.setParam(fdef.subkey, value);
        pgDmStore.update();
        return value;
    }

    controlGetValue(pgel, fn, values, fdef, skip_prev_value) {
        var _this = this;

        var val = pgDmStore.design.getDesignParameter(this, fdef.subkey);
        if(val === null && fdef.default_value) val = fdef.default_value;
        return val;
    }

    setVar(name, value) {
        return `$${name}: ${value};
`;
    }

    getCSS(params, style) {
        return this.onGetCSS(params, style);
    }

    getHTMLForStyleDisplay(params) {
        return '';
    }

    onGetCSS(part) {
        return '';
    }

    onGetConfigObject(part, config) {

    }
}

var PgDmStyle = function(name_or_node) {

    this.items = {};
    this.items_size_map = {};
    this.store = pgDmStore;
    this.current_state = null;

    if(name_or_node instanceof pgParserNode) {
        this.node = name_or_node;
        this.name = this.node.getAttribute('name');
        this.selector = this.node.getAttribute('selector');
        this.node_items = this.node.findOne('dmitems');

        var _this = this;

        this.node_items.children.forEach(function(item_node) {
            if(item_node.tagName == 'dmitem') {
                var item = new PgDmSkillItem(item_node);
                _this.items[item.getKey()] = item;
            }
        })
        this.updateSizeMap();
    } else {
        this.name = name_or_node;
        this.selector = '.' + pgDmStore.class_prefix + this.name;

        this.node = pgCreate(`<dmstyle name="${this.name}" selector="${this.selector}"><dmitems></dmitems></dmstyle>`);
        this.node_items = this.node.findOne('dmitems');
    }
}

PgDmStyle.prototype.addSkillItem = function(skill) {
    var item = new PgDmSkillItem(skill);
    item.setSize(this.store.getCurrentSize());
    item.setState(this.store.getCurrentState());

    var key = item.getKey();
    this.items[key] = item;
    this.node_items.append(item.node);
    this.updateSizeMap();
    return item;
}

PgDmStyle.prototype.removeSkillItem = function(item) {
   ////
}

PgDmStyle.prototype.updateSizeMap = function() {

    var _this = this;
    this.items_size_map = {};

    pgDmStore.design.getDeviceSizes().forEach(function(def) {
        _this.items_size_map[def.size] = [];
    })

    var map = this.items_size_map;

    $.each(this.items, function(key, item) {
        var size = item.size;
        if(item.size && map[size]) {
            map[size].push(item);
        }
    })
}

PgDmStyle.prototype.getItemsForSize = function(size) {
    return this.items_size_map[size] || [];
}

PgDmStyle.prototype.setSkillParam = function(skill, control, value) {
    var skillItem = this.getSkillItem(skill) || this.addSkillItem(skill);
    skillItem.setParam(control, value);
}

PgDmStyle.prototype.getSkillParam = function(skill, control, force_size) {
    var skillItem = this.getSkillItem(skill, force_size);
    return skillItem ? skillItem.getParam(control) : null;
}

PgDmStyle.prototype.getSkillItem = function(skill, force_size) {
    var key = this.store.getSkillItemKey(skill, force_size);
    if(!this.items[key]) {
        return false;
        //this.addSkillItem(skill);
    }
    return this.items[key];
}

PgDmStyle.prototype.getStates = function() {
    var r = {};
    $.each(this.items, function(key, item) {
        r[item.state] = true;
    })
    return r;
}

PgDmStyle.prototype.getCSS = function() {
    var _this = this;

    var css = '';
    var def_size = pgDmStore.default_size;

    var states = ['none', 'hover', 'active', 'focus']

    $.each(this.items_size_map, function(size, list) {

        var states_map = {
            none: '',
            hover: '',
            active: '',
            focus: ''
        }

        if(list.length) {
            list.forEach(function(item) {
                states_map[item.state] += item.getCSS(_this);
            })
        }

        states.forEach(function(state) {
            if(states_map[state].length) {

                var selector = _this.selector;
                if(state !== 'none') {
                    selector += ':' + state;
                }
                if(size == def_size) {
                    css += `${selector} {\n${states_map[state]} \n}`;
                } else {
                    css += `@include media-breakpoint-up(${size}) {\n ${selector} {\n${states_map[state]} \n} \n}`;
                }
            }
        })
    })

    return css;
}


PgDmStyle.prototype.getHTMLForStyleDisplay = function() {
    var _this = this;

    var html = '';
    var def_size = pgDmStore.default_size;

    var states = ['none', 'hover', 'active', 'focus']

    $.each(this.items_size_map, function(size, list) {

        var states_map = {
            none: '',
            hover: '',
            active: '',
            focus: ''
        }

        var has = false;

        if(list.length) {
            list.forEach(function(item) {
                var item_html = item.getHTMLForStyleDisplay();
                if(item_html.length) {
                    states_map[item.state] += item_html;
                    has = true;
                }
            })
        }

        if(has) {
            html += `<div class="pg-dm-style-list-item-preview-size"><label>${size}</label>`;
            states.forEach(function (state) {
                if (states_map[state].length) {

                    if (state !== 'none') {
                    }

                    html += `<div class="pg-dm-style-list-item-preview-state"><label>${state}</label>`;
                    html += states_map[state];
                    if (size == def_size) {
                    } else {
                    }
                    html += '</div>';
                }
            })
            html += '</div>';
        }
    })

    return html;
}

/*

Style

 - xs
 - md
 - lg

 */


var PgDmSkillItem = function(skill_or_node) {

    this.params = {};
    this.size = pgDmStore.default_size;
    this.state = pgDmStore.default_state;

    if(skill_or_node instanceof pgParserNode) {
        this.node = skill_or_node;
        var key = this.node.getAttribute('skill');
        this.skill = pgDmStore.getSkillByKey(key);

        var _this = this;

        this.node.getAttrList().forEach(function(a) {
            if((a.name !== 'skill' && a.name !== 'size' && a.name !== 'state') && a.value !== null) {
                _this.params[a.name] = a.value;
            }
        })
        this.size = this.node.getAttribute('size') || pgDmStore.default_size;
        this.state = this.node.getAttribute('state') || pgDmStore.default_state;
        if(this.state == 'normal') this.state = 'none';
    } else {
        this.skill = skill_or_node;
        this.node = pgCreate(`<dmitem skill="${this.skill.key}"></dmitem>`);
        this.node.setAttribute('size', this.size);
        this.node.setAttribute('state', this.state);
    }
}

PgDmSkillItem.prototype.setSize = function(size) {
    this.size = size;
    this.node.setAttribute('size', this.size);
}

PgDmSkillItem.prototype.setState = function(state) {
    this.state = state;
    this.node.setAttribute('state', this.state);
}

PgDmSkillItem.prototype.getKey = function() {
    return pgDmStore.getSkillItemKey(this.skill, this.size, this.state);
}

PgDmSkillItem.prototype.isSet = function() {
    return this.skill.isSet(this.params);
}

PgDmSkillItem.prototype.getParam = function(control) {
    return this.params[control.subkey] || null;
}

PgDmSkillItem.prototype.setParam = function(control, value) {
    this.params[control.subkey] = value;
    if(value !== null) {
        this.node.setAttribute(control.subkey, value);
    } else {
        this.node.removeAttribute(control.subkey);
    }
}

PgDmSkillItem.prototype.getSkill = function() {

}

PgDmSkillItem.prototype.getDeviceSize = function() {

}

PgDmSkillItem.prototype.getState = function() {

}

PgDmSkillItem.prototype.getCSS = function(style) {
    if(this.skill.isSet(this.params)) {
        return this.skill.getCSS(this.params, style);
    } else {
        return '';
    }
}

PgDmSkillItem.prototype.getHTMLForStyleDisplay = function() {
    return this.skill.getHTMLForStyleDisplay(this.params);
}




var PgDmFont = function() {

    this.family = null;
    this.weight = null;
    this.italic = false;
    this.key = null;

    this.getInitCSS = function() {
        return '';
    }

    this.getRuleCSS = function() {
        return `@include m-font-${this.key};\n`;
    }
}

var PgDmFonts = function() {

    var _this = this;

    this.fonts = {}

    this.clear = function() {
        this.fonts = {};
    }

    this.setFont = function(font) {
        this.fonts[font.key] = font;
    }

    this.getFont = function(key) {
        return this.fonts[key];
    }

    this.getInitCSS = function() {
        var css = '';
        $.each(this.fonts, function(key, f) {
            css += f.getInitCSS();
        })
        return css;
    }

    this.getGoogleFontsUrl = function() {
        // https://fonts.googleapis.com/css?family=M+PLUS+1p:300,400|Roboto+Condensed:300,300i,400
        var map = {};
        var r = [];

        $.each(this.fonts, function(key, font) {
            var lib_font = pinegrow.fontLibrary.getFontByFamily(font.family);
            if(lib_font && lib_font.type === 'google') {

                r.push(lib_font.getGoogleFontUrlToken());
                /*
                if (!map[font.family]) {
                    map[font.family] = [];
                }
                var w = font.weight ? font.weight.toString() : null;
                if (font.italic) {
                    if (!w) w = '400';
                    w += 'i';
                }
                if (w && map[font.family].indexOf(w) < 0) {
                    map[font.family].push(w);
                }
                if (false) {
                    //add bold
                    var b = (parseInt(font.weight || '400') + 200).toString();
                    if (map[font.family].indexOf(b) < 0) {
                        map[font.family].push(b);
                    }
                }*/
            }
        })

        $.each(map, function(family, vlist) {
            //r.push(family.replace(/\s/g, '+') + ':' + vlist.join(','));
        })
        return r.length ? 'https://fonts.googleapis.com/css?family=' + r.join('|') + '&display=swap' : null;
    }


    this.getRemoteFontsUrls = function() {

        var r = [];
        $.each(this.fonts, function(key, font) {
            var lib_font = pinegrow.fontLibrary.getFontByFamily(font.family);
            if(lib_font && (lib_font.type === 'remote' || lib_font.type === 'adobe')) {
                if(r.indexOf(lib_font.url) < 0) r.push(lib_font.url);
            } else if(lib_font && lib_font.type === 'local') {
                var url = lib_font.getUrl();
                if(r.indexOf(lib_font.url) < 0) r.push(lib_font.url);
            }
        })
        return r;
    }

    this.getUsedFonts = function() {
        var r = [];
        $.each(this.fonts, function(key, font) {
            var lib_font = pinegrow.fontLibrary.getFontByFamily(font.family);
            if(lib_font && r.indexOf(lib_font) < 0) r.push(lib_font);
        })
        return r;
    }

    this.getFontUrls = function() {
        var r = [];
        var gf_url = this.getGoogleFontsUrl();
        if(gf_url) {
            r.push(gf_url);
        }

        this.getRemoteFontsUrls().forEach(function(url) {
            r.push(url)
        })
        return r;
    }

    this.getImportStatements = function(cs_url) {
        var p = pinegrow.getCurrentProject();
        return this.getFontUrls().map(function(url) {
            if(cs_url) {
                //try to make url relative
                if(p.isUrlInProject(url)) {
                    url = crsaMakeLinkRelativeTo(url, cs_url);
                }
            }
            return `@import url('${url}');`;
        })
    }

    this.getFontFontFaceStatements = function() {
        var r = [];
        $.each(this.fonts, function(key, font) {
            var lib_font = pinegrow.fontLibrary.getFontByFamily(font.family);
            if(lib_font) {
                var code = lib_font.getCode();
                if(code.css) {
                    if (r.indexOf(code.css) < 0) r.push(code.css);
                }
            }
        })
        return r;
    }

}




var PgDmDesign = function() {

    this.node = pgCreate(`<dmdesign active></dmdesign>`);
    this.fonts_node = this.node.findOne('dmfonts');
    this.colors_node = this.node.findOne('dmcolors');

    var params = {}
    var colors = []

    this.palettes = new PgDmPalettes();

    var hasUnitOrVar = function(val) {
        return typeof val === 'string' && val.match(/[a-z\%\$]/);
    }

    var isColor = function(val) {
        return typeof val === 'string' && !val.match(/\$/);
    }

    this.options = {
        default_font_types: ['heading', 'lead', 'text'],
        font_sizes: {
            min: -2,
            max: 6
        },
        max_fonts: 10,
        max_colors: 10,
        reserved_colors: [
            {name: 'light', color: '#fff'},
            {name: 'gray', color: '#888'},
            {name: 'dark', color: '#000'}
        ],
        palette: null,
        device_sizes: [
            {size: 'xs', name_short: 'XS', name: 'Extra Small', breakpoint: 0, w: 320, icon: 'icon-Xs'},
            {size: 'sm', name_short: 'SM', name: 'Small', breakpoint: 576, w: 640, icon: 'icon-IPAD'},
            {size: 'md', name_short: 'MD', name: 'Medium', breakpoint: 768, w: 768, icon: 'icon-tableta-lezeca'},
            {size: 'lg', name_short: 'LG', name: 'Large', breakpoint: 992, w: 1024, icon: 'icon-ikone100--copy_laptop'},
            {size: 'xl', name_short: 'XL', name: 'Extra Large', breakpoint: 1200, w: 1280, icon: 'icon-Lg'},
        ]
    }



    for(var i = 0; i < this.options.device_sizes.length; i++) {
        if(i === this.options.device_sizes.length - 1) {
            this.options.device_sizes[i].w_range = [this.options.device_sizes[i].breakpoint, 99999];
        } else {
            this.options.device_sizes[i].w_range = [this.options.device_sizes[i].breakpoint, this.options.device_sizes[i+1].breakpoint - 1];
        }
    }

    this.getDeviceSizes = function() {
        return this.options.device_sizes;
    }

    this.getDeviceSizeDef = function(size) {
        for(var i = 0; i < this.options.device_sizes.length; i++) {
            if(this.options.device_sizes[i].size === size) return this.options.device_sizes[i];
        }
    }

    this.fonts = new PgDmFonts();

    this.getFont = function(type) {
        return this.fonts.getFont(type);

        for(var i = 0; i < this.options.fonts.length; i++) {
            if(this.options.fonts[i].type == type) return this.options.fonts[i];
        }
        return null;
    }

    this.getFontDeclarationForCSS = function(name) {
        return `    @include m-font-${name};\n`;
    }

    this.getFontTypes = function() {
        var r = [];
        for(var i = 1; i <= this.options.max_fonts; i++) {
            if(i <= this.options.default_font_types.length) {
                r.push(this.options.default_font_types[i-1])
            } else {
                var name = params[`fonts.font_${i}_name`] || null;
                if(name) {
                    r.push(name);
                }
            }
        }
        return r;
    }

    this.getFontSizeForCSS = function(size) {
        if(hasUnitOrVar(size)) return size;
        var val = parseFloat(size);
        if(val >= this.options.font_sizes.min && val <= this.options.font_sizes.max) {
            return `map-get($m-sizes, ${val})`;
        } else {
            return size;
        }
    }

    this.getDesignParameter = function(skill, subkey) {
        var _this = this;
        var skill_key = typeof skill === 'object' ? skill.key : skill;

        var key = `${skill_key}.${subkey}`;
        if(key in params) return params[key];
        return null;
    }

    this.setDesignParameter = function(skill, subkey, value) {
        var _this = this;

        var key = `${skill.key}.${subkey}`;
        params[key] = value;

        return;

        if(name == 'fontcombo') {
            value.toFonts(this.fonts);

        } else if(name == 'palette') {
            this.options.palette = value;
            _this.colors_node.removeAllChildren();

            value.getColorsInfo().forEach(function(c) {
                pgCreate(`<dmcolor name="${c.name}" value="${c.color.css()}"></dmcolor>`).appendTo(_this.colors_node);
            })
        } else {
            if(value !== null) {
                this.node.setAttribute(name, value);
            } else {
                this.node.removeAttribute(name);
            }
        }
    }

    this.getColorForCSS = function(val) {

        var a = val.split('-');
        if(a.length === 2) {
            return `m-color(${a[0]}, ${a[1]})`
        }
        return null;
    }

    this.findColorByName = function(name) {
        for(var i = 0; i < colors.length; i++) {
            if(colors[i].name === name) {
                return colors[i];
            }
        }
    }

    this.setColor = function(idx, name, color) {
        if(colors.length <= idx) {
            colors.push({
                name: name,
                color: chroma(color),
                idx: idx
            })
        } else {
            colors[idx].color = chroma(color);
            colors[idx].name = name;
        }
    }

    for(var i = 0; i < this.options.max_colors; i++) {
        colors.push({
            name: '',
            color: null,
            idx: i
        })
    }

    this.options.max_custom_colors = this.options.max_colors - this.options.reserved_colors.length;

    var idx = this.options.max_custom_colors;

    for(var i = 0; i < this.options.reserved_colors.length; i++) {
        var c = this.options.reserved_colors[i];
        this.setColor(idx, c.name, c.color);
        this.setDesignParameter('colors', `color${idx}`, c.color);
        idx++;
    }


    this.renameColor = function(name, new_name) {
        var c = this.findColorByName(name);
        if(c) {
            c.name = new_name;
        }
    }

    this.getColors = function() {
        var r = [];
        for(var i = 0; i < colors.length; i++) {
            if(colors[i].name) {
                r.push(colors[i])
            }
        }
        return r;
    }

    this.getColorPalette = function() {
        return this.options.palette;
    }

    this.setColorPalette = function(pal) {
        this.options.palette = pal;
    }

    var color_shade_step = 0.25;
    var color_shade_steps = 2;

    this.setColorInterval = function(percent) {
        color_shade_step = percent / 100.0;
    }

    var color_shade_desc = {
        '-2': 'darker',
        '-1': 'dark',
        '0': 'normal',
        '1': 'light',
        '2': 'lighter'
    }

    this.getColorShade = function(color, step) {
        var f = color_shade_step * step;
        if(step == 0) {
            return color;
        } else if(step < 0) {
            return color.darken(-f);
        } else {
            return color.brighten(f);
        }
    }

    this.getColorShadeSASSCode = function(color, step) {
        var f = parseInt(color_shade_step * step * 100);
        if(step == 0) {
            return color;
        } else if(step < 0) {
            return `scale-color(${color}, $lightness: ${f}%)`;
        } else {
            return `scale-color(${color}, $lightness: ${f}%)`;
        }
    }

    this.getColorShadeDescriptions = function() {
        return color_shade_desc;
    }
}


var PgDmElementStateManager = function() {

    var hasStateSet = false;

    var clearStates = function() {
        if(hasStateSet) {
            pinegrow.getAllPageViews().forEach(function (pv) {
                pv.getDocument().querySelectorAll('.pg-state-hover,.pg-state-active,.pg-state-focus').forEach(function (node) {
                    node.classList.remove('pg-state-hover', 'pg-state-active', 'pg-state-focus');
                })
            })
        }
        hasStateSet = false;
    }

    this.setState = function(state) {
        clearStates();
        if(state != 'none') {
            pinegrow.selectedElements.forEachElement(function (pgel) {
                pgel.get$DOMElements().forEach(function ($el) {
                    $el.addClass('pg-state-' + state);
                })
            })
            hasStateSet = true;
        }
    }
}


var PgDmStore = function() {

    /*

    class

     */

    var _this = this;

    var data = {}
    var skills = {}
    var styles = []
    var styles_map = {}
    var design_skills = {}

    this.class_prefix = 'm-';

    this.db = null;

    //set from load
    this.designs_node = null;
    this.locked_node = null;

    //this.node = pgCreate(`<dmproject><dmdesigns></dmdesigns><dmstyles></dmstyles></dmproject>`);
    //this.designs_node = this.node.findOne('dmdesigns');
    //this.styles_node = this.node.findOne('dmstyles');

    this.default_size = 'xs';
    this.default_state = 'none';

    var current_style = null;
    var current_size = this.default_size;
    var current_state = this.default_state;

    var engine = null;

    this.design = new PgDmDesign();

    this.stateManager = new PgDmElementStateManager();

    this.fontLibrary = new PgDmFontLibrary();

    this.localImages = new PgRandomLocalImage();

    //this.designs_node.append(this.design.node);

    var design_providers = {}
    var active_design_provider = null;

    this.registerDesignProvider = function(key, provider) {
        design_providers[key] = provider;
    }

    this.activateDesignProvider = function(key, skip_save) {
        var current = active_design_provider;
        active_design_provider = design_providers[key] || null;

        if(active_design_provider && current !== active_design_provider) {
            active_design_provider.activate();

            design_skills = {};

            active_design_provider.getSkills().forEach(function(skill) {
                _this.registerDesignSkill(skill);
            });
        }

        if(!skip_save) {
            var pi = pinegrow.getCurrentProjectInfo();
            pi.setSetting('active-design-provider', key);
            pi.save();
        }
    }

    this.getActiveDesignProvider = function() {
        return active_design_provider;
    }

    this.registerSkill = function(skill) {
        skills[skill.key] = skill;
    }

    this.getSkillByKey = function(key) {
        return skills[key] || null;
    }

    this.registerDesignSkill = function(skill) {
        design_skills[skill.key] = skill;

//        this.design.node.append(skill.node);
    }

    this.getDesignSkill = function(key) {
        return design_skills[key] || false;
    }

    this.forEachDesignSkill = function(func) {
        $.each(design_skills, function(key, skill) {
            func(skill)
        })
    }

    this.save = function() {
        var project = pinegrow.getCurrentProject();
        var path = require('path');
        var fs = require('fs');

        //todo save to projectdb
        project.requireDB(function(db) {



        }, false /* create */)

        return;

        var file_name = path.join(project.getDir(), 'pinegrow.magic.xml');
        var str = this.node.toStringOriginal(true);

        crsaWriteFileWithBackup(fs, file_name, str, 'utf8');
    }

    this.addNewDesign = function(source_node) {

        pgDmStore.db.makeChanges(null, source_node ? 'Duplicate design' : 'Add design', function() {
            var new_design = (source_node || pgDmStore.design.node).clone();

            if(!source_node) {
                _this.designs_node.append(new_design);
            } else {
                new_design.insertAfter(source_node);
            }

            _this.forEachDesignSkill(function(skill) {
                var sn = new_design.findOne(`dmdesignskill[skill="${skill.key}"]`);
                if(!sn) {
                    sn = pgCreate(`<dmdesignskill skill="${skill.key}"></dmdesignskill>`);
                    sn.appendTo(new_design);
                    skill.load(sn);
                    skill.populateDefaultParams();
                } else {
                    skill.load(sn);
                }
            })

            _this.setActiveDesign(new_design);
        }, 'design_panel');
    }

    this.deleteDesign = function(node) {

        if(_this.designs_node.find('dmdesign').length === 1) {
            pinegrow.showQuickError('Can not delete the only design. A project needs at least one design.')
        } else {
            pgDmStore.db.makeChanges(null, 'Delete design', function () {
                var next = null;
                if(node.hasAttribute('active')) {
                    next = node.next() || node.prev();
                }
                node.remove();

                next && _this.setActiveDesign(next);
            }, 'design_panel');
        }
    }


    this.getActiveDesignNode = function() {
        var a = this.designs_node.findOne('dmdesign[active]');
        if(a) return a;
        return this.designs_node.findOne('dmdesign')
    }

    this.setActiveDesign = function(new_design) {

        var _this = this;

        this.designs_node.withoutEvents(function() {
            _this.designs_node.find('dmdesign[active]').forEach(function(node) {
                node.removeAttribute('active');
            })

            new_design.setAttribute('active', null);
        })

        this.forEachDesignSkill(function(skill) {
            var skill_node = new_design.findOne(`dmdesignskill[skill="${skill.key}"]`);
            if(skill_node) {
                skill.load(skill_node);
            } else {
                skill_node = pgCreate(`<dmdesignskill skill="${skill.key}"></dmdesignskill>`);
                new_design.append(skill_node);
                skill.load(skill_node);
                skill.populateDefaultParams();
            }
        })



        this.design.node = new_design;

        _this.update();
        _this.designPanel.updatePanelDisplay();
    }

    this.isDesignEnabled = function() {
        return _this.designs_node !== null && active_design_provider !== null;
    }

    this.load = function(force_create) {
        var project = pinegrow.getCurrentProject();

        //page = page || pinegrow.getSelectedPage();
        var path = require('path');
        var fs = require('fs');

       // try {

        project.requireDB(function(db) {
            _this.db = db;

            if(force_create) {
                _this.designs_node = db.getOrCreateNode('dmdesigns');
            } else {
                _this.designs_node = db.findOne('dmdesigns');
            }

            if(_this.designs_node) {

                _this.locked_node = db.getOrCreateNode('dmlocked');

                var design_node = _this.designs_node.findOne('dmdesign[active]') || _this.designs_node.findOne('dmdesign');

                if (design_node) {

                } else {
                    design_node = pgCreate(`<dmdesign active></dmdesign>`);
                    _this.designs_node.append(design_node);
                    //_this.update();
                }

                _this.designPanel.show();
                _this.setActiveDesign(design_node);
                pinegrow.dispatchEvent('on_dm_project_loaded', null);
            } else {
                _this.designPanel.show();
            }
            /*
            _this.designPanel.updatePanelDisplay();
            _this.designPanel.updateKeepList();

             */


        }, !force_create /* skip create */)
    }

    /*
    var onPageSaved = function() {
        _this.save();
    }

     */

    this.addClass = function(cls) {

    }

    this.setParamLocked = function(skill, key, val) {
        var k = (skill.key + '-' + key).replace(/\./g, '-');
        if(this.locked_node) {
            val ? this.locked_node.setAttribute(k, 'true') : this.locked_node.setAttribute(k, 'false');
            this.db.is_changed = true; //just set the flag, no need to dispatch events
        }
    }

    this.isParamLocked = function(skill, key, default_val) {
        if(default_val === undefined) default_val = false;
        var k = (skill.key + '-' + key).replace(/\./g, '-');
        if(this.locked_node && this.locked_node.hasAttribute(k)) {
            return this.locked_node.getAttribute(k) === 'true';
        }
        return default_val;
    }

    this.surpriseMe = function(skip_skills) {
        this.forEachDesignSkill(function(skill) {
            if(!skip_skills || skip_skills.indexOf(skill) < 0) {
                skill.surpriseMe();
            }
        })
        this.update();
        this.designPanel.updatePanelDisplay();
    }

    this.makeDesignFromPhoto = function(src, with_surprise_me) {

        var colors_skill = pgDmStore.getDesignSkill('colors');

        var harmony = colors_skill.getParam('harmony');

        colors_skill.getPaletteFromImage(src, function() {

            //the palette is set
            var bck_skill = pgDmStore.getDesignSkill('background');
            bck_skill.setParam('image', src);

            if(with_surprise_me) {
                pgDmStore.surpriseMe([colors_skill, bck_skill]);
            } else {
                this.update();
                this.designPanel.updatePanelDisplay();
            }
        }, null, harmony);
    }

    this.assignStyle = function(pgel, style) {
        var cls = style.selector.replace('.', '');

        if(!pgel.hasClass(cls)) {
            pgel.addClass(cls);
        }
    }

    this.assignStyleToSelectedElements = function(style) {
        var _this = this;
        pinegrow.selectedElements.forEachElement(function(pgel) {
            _this.assignStyle(pgel, style);
        })
    }

    this.rememberSelectedStyleForSelectedElements = function(style) {
        pinegrow.selectedElements.forEachElement(function(pgel) {
            pgel.setData('dm-selected-style', style);
        })
    }

    this.getPropSections = function(pgel) {
        var map = {};
        $.each(skills, function(key, skill) {
            var s = skill.getPropSection(pgel);
            if(s) {
                map[s.key] = s;
            }
        })
        return map;
    }

    this.getDesignSections = function() {
        var map = {};
        $.each(design_skills, function(key, skill) {
            var s = skill.getSection();
            if(s) {
                map[s.key] = s;
            }
        })
        return map;
    }

    this.addStyle = function(style) {
        styles.push(style);
        styles_map[style.selector] = style;
        this.styles_node.append(style.node);
        pinegrow.dispatchEvent('on_dm_style_added', null, style);
    }

    this.getCurrentStyle = function() {
        return current_style;
    }

    this.setCurrentStyle = function(style, just_in_time) {
        current_style = style;
        setCurrentStateForCurrentStyle();
        this.stateManager.setState(current_state);
        pinegrow.dispatchEvent('on_dm_current_style_changed', null, style, just_in_time);
    }

    this.triggerCurrentStyleChanged = function() {
        this.setCurrentStyle(current_style);
    }

    this.getSkillItemKey = function(skill, size, state) {
        return skill.key + '-' + (size || current_size) + '-' + (state || current_state);
    }

    this.getCurrentSize = function() {
        return current_size;
    }

    this.setCurrentSize = function(s) {
        current_size = s;
        this.setCurrentStyle(current_style); //trigger update
    }

    this.getCurrentState = function() {
        return current_state;
    }

    this.setCurrentState = function(s) {
        current_state = s;
        if(current_style) {
            current_style.current_state = s;
        }
        this.setCurrentStyle(current_style); //trigger update
    }

    this.findStyleByName = function(name) {
        for(var i = 0; i < styles.length; i++) {
            if(styles[i].name == name) {
                return styles[i];
            }
        }
        return null;
    }

    this.getStylesForPgel = function(pgel) {
        var r = [];
        pgel.getClasses().forEach(function(cls) {

            if(cls.startsWith(_this.class_prefix)) {
                //var style_name = cls.replace(_this.class_prefix, '');

                var style = styles_map['.' + cls] || null;
                if (style) {
                    style.idx = styles.indexOf(style);
                    r.push(style);
                }
            }
        })
        r.sort(function(a, b) {
            return a.idx - b.idx;
        })
        return r;
    }

    this.setSkillParam = function(pgel, skill, control, value) {
        var style = this.getCurrentStyle();
        var changed = true;

        if(!style) {
            style = new PgDmStyle('bubi' + getRand(1, 10000));
            this.addStyle(style);
            this.assignStyleToSelectedElements(style);
            pgDmStore.rememberSelectedStyleForSelectedElements(style);
            this.setCurrentStyle(style, true /* just_in_time */);
            changed = false;
        }

        if(typeof control === 'string') {
            control = skill.getControl(control);
        }
        style.setSkillParam(skill, control, value);

        //this.assignStyle(pgel, style);
        if(changed) {
            pinegrow.dispatchEvent('on_dm_style_changed', null, style);
        }

        this.update();
    }

    this.getSkillParam = function(pgel, skill, control, force_size) {
        var style = this.getCurrentStyle();

        if(!style) return null;

        if(typeof control === 'string') {
            control = skill.getControl(control);
        }

        return style.getSkillParam(skill, control, force_size);
    }

    this.getStyles = function() {
        return styles;
    }

    var forEachStyle = function(func) {
        styles.forEach(func);
    }

    this.setSASSEngine = function(e) {
        engine = e;

        engine.onGetCSS = function(part) {
            var css = '';

            if(part == 'init' || part == 'import') {
                var o = {
                    css: ''
                }
                pinegrow.dispatchEvent('on_dm_get_css', null, part, o);
                css += o.css;
            }

            if(part == 'rules') {
                var o = {
                    css: ''
                }
                pinegrow.dispatchEvent('on_dm_get_css', null, part, o);
                css += o.css;

                forEachStyle(function(style) {
                    css += style.getCSS();
                })
            }

            return css;
        }
    }

    this.update = function(done) {
        if(active_design_provider) {
            active_design_provider.getEngine(function(engine) {
                engine.update(done);
            })
        }
    }

    //current_style = new PgDmStyle('bubi');
    //this.addStyle(current_style);

    var onProjectLoaded = function(project) {
        var pi = pinegrow.getCurrentProjectInfo();
        var key = pi.getSetting('active-design-provider');
        key && _this.activateDesignProvider(key, true);
        pgDmStore.load();

        //activate on any pages that are open from before
        pinegrow.getAllPages().forEach(function(page) {
            onPageLoaded(page);
        })

        _this.designPanel.updatePanelDisplay();
    }

    var onProjectClosed = function() {
        if(active_design_provider) {
            active_design_provider.destroyEngine();
            active_design_provider = null;
        }
        _this.designs_node = null;
        _this.design.node = null;
        _this.locked_node = null;
        _this.designPanel.updatePanelDisplay();
    }

    var onPageLoaded = function(page) {
        if(pgDmStore.isDesignEnabled()) {
            active_design_provider.getEngine(function(engine) {
                engine.activateOnPage(page);
            })
        }
    }

    var onGetMenuPoints = function(page, pgel, menu, points) {

        var insertMenu = function(point, p) {

            var im = new PgInsertLibElementMenu(pgel, null, true, p, true /* recommended */);
            var actions = im.getActions();
            var m = new CrsaContextMenu(actions, point.$element);
            m.showForElement(point.$element);
            return false;
        }

        var insert_after = new PgElementPageViewMenuPoint('bottom-outside', menu, true, '+');

        insert_after.onClick = function() {
            insertMenu(this, 'insertAfter');
        }
        points.push(insert_after);

        if(pgel.canHaveChildren()) {
            var append = new PgElementPageViewMenuPoint(function () {
                return pgel.isEmpty() ? 'center' : 'bottom-inside'
            }, menu, true, '+');

            append.onClick = function () {
                insertMenu(this, 'append');
            }
            points.push(append);
        }

        /*

        var insert_before = new PgElementPageViewMenuPoint('left-outside', menu, true, '+');

        insert_before.onClick = function() {
            insertMenu(this, 'insertBefore');
        }
        points.push(insert_before);

        var prepend = new PgElementPageViewMenuPoint('left-inside', menu, true, '+');

        prepend.onClick = function() {
            insertMenu(this, 'prepend');
        }
        points.push(prepend);

        */
    }

    var setCurrentStateForCurrentStyle = function() {
        current_state = (current_style ? current_style.current_state : null) || 'none';
    }

    var onElementSelected = function() {
        var pgel = pinegrow.selectedElements.getLastSelected();
        if(pgel) {
            var styles = pgDmStore.getStylesForPgel(pgel);
            if(styles.length) {
                var sel_style = pgel.getData('dm-selected-style');
                if(sel_style && styles.indexOf(sel_style) >= 0) {
                    current_style = sel_style;
                } else {
                    current_style = styles[0];
                }
            } else {
                current_style = null;
            }
        } else {
            current_style = null;
        }

        setCurrentStateForCurrentStyle();

        _this.stateManager.setState(current_state);

        pinegrow.dispatchEvent('on_dm_current_style_changed', null, current_style, false, true /* on sel element */);
    }



    var onPageSaved = function() {
        if(pgDmStore.isDesignEnabled()) {
            active_design_provider.getEngine(function(engine) {
                engine.saveCSS();
            })
            //pinegrow.refreshPageChangedStatusForAllPages()
        }
    }

    this.destroy = function() {
        pinegrow.removeEventHandler('on_page_saved', onPageSaved)
        pinegrow.removeEventHandler('on_page_loaded', onPageLoaded)
        //pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected)
        pinegrow.removeEventHandler('on_project_loaded', onProjectLoaded)
        pinegrow.removeEventHandler('on_project_closed', onProjectClosed)
        //pinegrow.removeEventHandler('on_get_menu_points', onGetMenuPoints)
    }

    $('body').one('pinegrow-ready', function(e, pinegrow) {
        pinegrow.addEventHandler('on_page_saved', onPageSaved);
        pinegrow.addEventHandler('on_page_loaded', onPageLoaded);
        //pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
        pinegrow.addEventHandler('on_project_loaded', onProjectLoaded)
        pinegrow.addEventHandler('on_project_closed', onProjectClosed)
        //pinegrow.addEventHandler('on_get_menu_points', onGetMenuPoints)
    })
}

//if(pgIsMagic()) {
    window.pgDmStore = new PgDmStore();
//}

/*

var bubi = new PgDmStyle('bubi');

var s1 = new PgDmSkillHeadingLine();

bubi.addSkillItem = new PgDmSkillItem(s1, 'MD', 'hover');

store.addStyle(bubi);

store.assignStyle(pgel, bubi);

store.update();


Reserved style names = BS classes

Style Misko:
    heading.space
    .text-white

for each .misko remove all reserved classes and re-add them
 */



var PgDmStylePill = function(style) {

    this.$element = $(`<div class="pg-dm-style-pill" data-id="${style.name}">${style.name}</div>`);
}

PgDmStylePill.prototype.setActive = function(a) {
    if(a) {
        this.$element.addClass('active');
    } else {
        this.$element.removeClass('active');
    }
}



