var PgDmStyleListItem = function(style) {

    this.style = style;
    this.$element = $(`<div class="pg-dm-style-list-item"></div>`);
    this.$element.data('item', this);
}

PgDmStyleListItem.prototype.update = function() {
    var html = '';
    html += `<name>${this.style.name}</name>`;

    var preview_html = this.style.getHTMLForStyleDisplay();

    if(preview_html.length) {
        html += `<div class="pg-sm-style-list-item-preview">${preview_html}</div>`;
    }
    this.$element.get(0).innerHTML = html;
}

PgDmStyleListItem.prototype.setActive = function(val) {
    if(val) {
        this.$element.addClass('active');
    } else {
        this.$element.removeClass('active');
    }
}




var PgDmStyleListView = function() {

    var _this = this;

    this.$element = $('<div class="pg-dm-style-list"></div>');

    var $items = $('<div class="pg-dm-style-list-items"></div>').appendTo(this.$element);

    var view = this.view = new PgUIView(this.$element, 'pgdmstylelist');

    var style_items = new WeakMap();
    var style_items_list = [];

    var active_item = null;

    $items.on('click', '.pg-dm-style-list-item', function(e) {
        e.preventDefault();
        var item = $(this).data('item');
        pgDmStore.assignStyleToSelectedElements(item.style);
        pgDmStore.setCurrentStyle(item.style);
        //pgDmStore.triggerCurrentStyleChanged();
    })

    this.show = function() {
        view.whenVisible(function() {

            style_items_list.forEach(function(item) {
                //item.$element.detach();
            })

            pgDmStore.getStyles().forEach(function(style) {

                var item;
                if(style_items.has(style)) {
                    item = style_items.get(style);
                } else {
                    item = new PgDmStyleListItem(style);
                    style_items.set(style, item);
                    style_items_list.push(item);
                    item.update();
                    $items.append(item.$element);
                }

            })
        })
    }

    this.show();

    var onStyleAdded = function(pagenull, style) {
        _this.show();
    }

    var onRepaintList = function() {
        var style_items = new WeakMap();
        var style_items_list = [];
        $items.empty();
        _this.show();
    }

    var onStyleChanged = function(pagenull, style) {
        view.whenVisible(function() {
            var item = style_items.get(style);
            if(item) {
                item.update();
            }
        }, 'updateStyle' + style.name)
    }

    var onStyleSelected = function(pagenull, style, just_in_time) {
        view.whenVisible(function() {
            if(active_item) {
                active_item.setActive(false);
                active_item = null;
            }
            var item = style_items.get(style);
            if(item) {
                item.setActive(true);
                active_item = item;
            }
        }, 'setActiveStyle', 1)
    }

    pinegrow.addEventHandler('on_dm_style_added', onStyleAdded);
    pinegrow.addEventHandler('on_dm_project_loaded', onRepaintList);
    pinegrow.addEventHandler('on_dm_style_changed', onStyleChanged);
    pinegrow.addEventHandler('on_dm_current_style_changed', onStyleSelected);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_dm_style_added', onStyleAdded);
        pinegrow.removeEventHandler('on_dm_project_loaded', onRepaintList);
        pinegrow.removeEventHandler('on_dm_style_changed', onStyleChanged);
        pinegrow.removeEventHandler('on_dm_current_style_changed', onStyleSelected);
    }
}