var PgTutorialSubtitles = function() {

    var map = {}

    this.set = function(key, val) {
        map[key] = val;
    }

    this.get = function(key) {
        if(!map[key]) {
            //console.error(`Subtitles for ${key} missing!`)
        }
        return map[key] || '';
    }
}

var pgTutorialSubtitles = new PgTutorialSubtitles();



$(function() {

    $('body').one('pinegrow-ready', function(e, pinegrow) {

        var PgTutorial = function() {

            var _this = this;

            this.name = '';
            this.allowRandomAccess = false;
            this.warnOnRandomAccess = true;

            var lessons = [];

            this.registerThisTutorial = pgf.kids || false;
            this.showOnStartScreen = true;

            pinegrow.addEventHandler('on_register_tutorials', function(page, manager) {
                _this.registerThisTutorial && manager.addTutorial(_this)
            })

            this.addLesson = function(lesson) {
                if(!lesson.chapter) {
                    lesson.chapter = this.name;
                }
                lesson.tutorial = this;
                lessons.push(lesson);
            }

            this.initDone = function() {
                pinegrow.recentFiles.populate();
                if(pgf.featured_projects) {
                    pinegrow.backend.getPublicProjects();
                }
            }

            this.getLessons = function() {
                return lessons;
            }

            var findLesson = this.findLesson = function(key) {
                for(var i = 0; i < lessons.length; i++) {
                    if(lessons[i].key === key) return lessons[i];
                }
                return null;
            }



            this.openLesson = function(key, done, force_save) {

                var lesson = _this.findLesson(key);

                var open = function() {

                    var missing = []

                    if(lesson.requireProducts.length && !pinegrow.isTrial()) {
                         for(var i = 0; i < lesson.requireProducts.length; i++) {
                             if(!pinegrow.hasActivatedProduct(lesson.requireProducts[i])) {
                                 var addon = pgAddOnManager.findAddon(lesson.requireProducts[i]);
                                 missing.push(addon ? addon.name : lesson.requireProducts[i]);
                             }
                         }
                    }

                    if(missing.length) {
                        pinegrow.showAlert(`<p>This ${pgf.lesson_name} requires the following add-ons that are not activated for your ${pgf.product_name_short} license:</p>
<ul><li>${missing.join('</li><li>')}</li></ul>`, 'One or more add-ons are required', 'Cancel', 'Upgrade your license', null, function() {
                            pinegrow.showBuyProductScreen();
                        })
                        return;
                    }

                    window.parent && window.parent.postMessage({
                        action: 'pgStartedLoadingProject'
                    }, '*');

                    if(isApp()) {
                        var gui = require('nw.gui');
                        var path = require('path');
                        var data_dir = path.join(gui.App.dataPath, 'pg_tutorials', key);
                        var fs = require('fs');
                        var open_async = false;

                        var openProject = function() {
                            var project = pinegrow.openProject(data_dir, null, null, {
                                lesson: lesson,
                                name: lesson.name
                            });
                        }

                        if(fs.existsSync(data_dir)) {
                        } else {
                            var done = false;

                            if(lesson.main_lesson) {
                                var prev_lesson = _this.getPreviousLesson(lesson);
                                if(prev_lesson) {
                                    var prev_data_dir = path.join(gui.App.dataPath, 'pg_tutorials', prev_lesson.key);
                                    if(fs.existsSync(prev_data_dir)) {
                                        //copy this
                                        crsaCopyFolder(prev_data_dir, data_dir, function() {
                                            openProject()
                                        }, true, true);
                                        done = true;
                                        open_async = true;
                                    }
                                }
                            }
                            if(!done) {
                                open_async = true;
                                crsaCreateDirs(data_dir);
                                lesson.copyFilesToFolder(data_dir, function() {
                                    openProject()
                                });
                            }
                        }
                        if(!open_async) {
                            openProject()
                        }

                    } else {
                        var url = 'https://tutorial_' + key;
                        var p = new CrsaProject();

                        p.fromServer(url, function (p) {
                            pinegrow.setCurrentProject(p);
                            done && done();
                        })
                    }
                }

                if(lesson.external_url) {
                    pinegrow.openExternalUrl(lesson.external_url);
                    pinegrow.backend.setLessonStatus(lesson, 'done');
                    lesson.setStatus('done');
                    return;
                }
                var cp = pinegrow.getCurrentProject();
                if(cp) {
                    pinegrow.backend.saveAndCloseCurrentProject(function() {
                        open();
                    }, force_save)
                } else {
                    open();
                }
            }



            this.openNextLesson = function(done) {
                //pinegrow.backend.
                var p = pinegrow.getCurrentProject();
                var next = this.getNextLesson(p.lesson);
                if(next) {
                    if(pinegrow.currentUser.isLoggedInAndActive() || next.demo) {
                        this.openLesson(next.key, done, true);
                    } else {
                        pinegrow.currentUser.onlyForMembersNotice('do all 58 adventures', 'You completed all demo adventures', `<p>Congratulations, you completed the first ten adventures.</p>`);
                        pinegrow.backend.saveAndCloseCurrentProject();
                    }
                } else {
                    pinegrow.showAlert(this.all_lessons_done || pgf.all_lessons_done, 'Well done!', null, this.all_lessons_done_button || pgf.all_lessons_done_button);
                    pinegrow.backend.saveAndCloseCurrentProject();
                }
            }

            this.getNextLesson = function(lesson) {
                var idx = lessons.indexOf(lesson);
                if(idx + 1 < lessons.length && !lessons[idx + 1].external_url) {
                    return lessons[idx + 1];
                }
            }

            this.getPreviousLesson = function(lesson) {
                var idx = lessons.indexOf(lesson);
                if(idx - 1 >= 0) {
                    return lessons[idx - 1];
                }
            }

            this.getAllLessonParts = function(lesson) {
                var main = lesson.main_lesson || lesson;
                var r = [];
                lessons.forEach(function(l) {
                    if(l === lesson || l === main || l.main_lesson === main) {
                        r.push(l);
                    }
                })
                return r;
            }

        }

        window.PgTutorial = PgTutorial;

        window.PgTutorialLesson = function(key, icon, chapter, img) {

            this.key = key;
            this.name = null;
            this.description = null;
            this.short_description = null;
            this.video = null;
            this.status = 'locked';
            this.icon = icon;
            this.chapter = chapter;
            this.image = img;
            this.demo = false;
            this.tutorial = null;
            this.requireProducts = [];

            if(!pgf.kids) {
                this.adventure_label = 'Steps';
            }

            this.subtitles = pgTutorialSubtitles.get(key);

            var files = []
            var resources = []
            var hidden_files = {}

            this.resources = resources;

            this.getProjectId = function() {
                return 'https://tutorial_' + this.key;
            }

            this.setStatus = function(s) {
                if(s === 'none') s = 'locked';
                this.status = s;
            }

            this.addPage = function(name, content) {
                files.push({
                    name: name,
                    content: content
                })
            }

            this.addStylesheet = function(name, content) {
                files.push({
                    name: name,
                    content: content
                })
            }

            this.addHiddenFile = function(name, content) {
                hidden_files[name] = content;
            }

            this.getHiddenFileContent = function(name) {
                return hidden_files[name] || null;
            }

            this.getFileContent = function(name) {
                for(var i = 0; i < files.length; i++) {
                    if(files[i].name === name) {
                        return files[i].content;
                    }
                }
                return null;
            }

            this.addResource = function(url, mapped_url) {
                resources.push({
                    source_url: url,
                    dom_url: mapped_url
                })
            }

            this.addBootstrap = function() {
                this.addResource('assets/js/popper.min.js','frameworks/bootstrap5/template/assets/js/popper.min.js');

                this.addResource('bootstrap/js/bootstrap.min.js','frameworks/bootstrap5/template/bootstrap/js/bootstrap.min.js');

                this.addResource('bootstrap/css/bootstrap.min.css', 'frameworks/bootstrap5/template/bootstrap/css/bootstrap.min.css');
            }

            this.addPGIA = function() {
                this.addResource('pgia/lib/pgia.js','frameworks/pg.animations/resources/pgia/lib/pgia.js');
            }

            this.addFilesToProject = function(project) {
                files.forEach(function(f) {
                    project.addFile(f.name, f.content);
                })

                resources.forEach(function(r) {
                    project.addSourceUrlToDOMUrlMap(r.source_url, r.dom_url);
                })

                if(this.inject_css) {
                    project.inject_css = this.inject_css;
                }
            }

            this.copyFilesToFolder = function(dir, done) {
                var path = require('path');
                var fs = require('fs');
                files.forEach(function(f) {
                    var file = path.join(dir, f.name);
                    var paths = crsaGetFileDir(file);

                    if(!fs.existsSync(paths)) {
                        crsaCreateDirs(paths, fs);
                    }
                    fs.writeFileSync(file, f.content, {encoding: 'utf8'});
                })

                var remote_count = 0;

                resources.forEach(function(r) {
                    var dest_file = path.join(dir, r.source_url);
                    var paths = crsaGetFileDir(dest_file);

                    if(!fs.existsSync(paths)) {
                        crsaCreateDirs(paths, fs);
                    }

                    if(crsaIsRemoteUrl(r.dom_url)) {
                        remote_count++;
                        pgCopyRemoteUrlToLocalPath(r.dom_url, dest_file, function() {
                            setTimeout(function() {
                                remote_count--;
                                if(remote_count <= 0) {
                                    done && done();
                                }
                            },1);
                        })
                    } else {
                        var source_file = pgGetAppFilePath('/' + r.dom_url);
                        crsaCopyFileSync(fs, source_file, dest_file);
                    }
                })

                if(remote_count <= 0) {
                    done && done();
                }
            }

            this.setLibrary = function(d) {

            }

            this.setStyle = function(d) {

            }

            this.getThumbUrl = function() {
                return 'help/examples/pinepoems/thumbnail.png';
            }

            this.getVideoButton = function(vid, title) {
                return `<div class="pg-tutorial-video-button" data-vid="${vid}"><i class="icon icon-play-round"></i> ${title || 'Watch the video'}</div>`;
            }

            this.makeDesc = function(vid, about_element, adventure, lib) {

                var html = this.getVideoButton(vid) + `<div class="pg-info-about-element">${about_element}</div><div class="pg-info-adventure">`;
                if(adventure) {

                    adventure = this.addInteractiveElementsToHTML(adventure);

                    var rep = `<li>Click on <b>${tooltip('Check', 'check')}</b> when you are done.`;

                    if(isApp()) {
                        rep += ` Use <b>${tooltip('Restart', 'restart_lesson')}</b> to restart the tutorial.`;
                    }
                    rep += `</li></ol>`;


                    adventure = adventure.replace('</ol>', rep);

                    html += `<h2>${this.adventure_label || 'The Adventure'}</h2>${adventure}</div>`;
                }

                if(lib) {
                    html += `<div class="pg-info-lib">${lib}</div>`
                }
                return html;
            }

            this.getHtmlForLibButton = function(name, code, add_class, tip, target_sel, place_error_msg) {
                return `<div title="${tip || 'Drag me to the page!'}" class="pg-info-panel-lib-button-code ${add_class} aaacrsa-factory-element aaacrsa-factory-element-text" data-code="${pgEncodeAttribute(code)}" draggable="true" data-target-selector="${target_sel || ''}" data-place-error="${place_error_msg || ''}"><name>${name}</name></div>`;
            }

            this.check = function() {

                var oks = [
                    'You crushed it.',
                    'Great work.',
                    'Good job.',
                    'That was brilliant.',
                    'Excellent work.'
                ];

                var oops = [

                ]

                var status = {
                    result: 'ok'
                }

                var page = pinegrow.getSelectedPage();

                var all_pages = {};

                pinegrow.getAllPages().forEach(function(p) {
                    all_pages[p.name] = p;
                })

                if(this.onCheck) {
                    this.onCheck(page, status, all_pages);
                }

                if(!status.title) {
                    status.title = status.result === 'ok' ? 'Congratulation!' : 'Oops';
                }
                if(!status.message) {
                    status.message = status.result === 'ok' ? 'You crushed it.' : 'Take a closer look, something is not right.';
                }

                if(status.result === 'ok') {
                    pinegrow.backend.setLessonStatus(this, 'done');
                    this.setStatus('done');

                    var next = tutorial.getNextLesson(this);
                    if(next) {
                        if(next.status === 'locked') {
                            next.setStatus('active');
                        }
                    }
                    pinegrow.tutorialsManager.updateStartScreen();
                } else if(this.status !== 'active') {
                    pinegrow.backend.setLessonStatus(this, 'active');
                    this.setStatus('active');
                    pinegrow.tutorialsManager.updateStartScreen();
                }

                return status;
            }

            this.addInteractiveElementsToHTML = function(adventure) {
                return pinegrow.tutorialsManager.addInteractiveElementsToHTML(adventure)
            }
        }



        var showElement = function(tag) {
            return '<span class="html-tag">' + escapeHtmlCode(`<${tag}>`) + '</span>';
        }

        PgTutorial.prototype.showElement = showElement;

        var select = function(text, sel, page) {
            if(!sel) {
                sel = text;
                text = showElement(text);
            } else {
                text = escapeHtmlCode(text);
            }
            return `<a data-select data-html="true" title="Click here to select the element." href="${sel}" data-page="${page || ''}">${text}</a>`;
        }

        PgTutorial.prototype.select = select;

        var showRule = function(sel, done) {

            var view = pgUIViewManager.getView('style');
            var $vel = view.get$Element();

            if(!pinegrow.selectedElements.getLastSelected()) {
                pinegrow.showQuickError('Select the element on the page first.');
            } else {

                if (!$vel.is(':visible')) {
                    pinegrow.showQuickError('First, open the Style panel.')
                } else {
                    sel = sel.toLowerCase();
                    var $rules = $('.rules-list');
                    var has = false;
                    $rules.find('.rule-selector').each(function(i, e) {
                        var $e = $(e);
                        if($e.text().toLowerCase() === sel) {
                            has = true;
                            pgScrollToItemInContainer($e, $rules, false, true, function () {
                                pgRevealElement($e, done);
                                pinegrow.showQuickMessage('Click on the CSS rule to select it.');
                            }, -50)
                        }
                    })
                    if(!has) {
                        pinegrow.showQuickError(`The CSS rule ${sel} is not shown for the selected element.`)
                    }
                }
            }
        }
        PgTutorial.prototype.showRule = showRule;

        var field = function(text, field) {
            return tooltip(text, 'Click to show the field.', 'field', field)
        }
        PgTutorial.prototype.field = field;

        var rule = function(sel) {
            return tooltip(`<code>${sel}</code> CSS rule`, 'Click to show the CSS rule.', 'rule', sel)
        }
        PgTutorial.prototype.rule = rule;

        PgTutorial.prototype.showLibElement = function(text, sel) {
            return tooltip(text, 'Click to show the element in the Insert panel.', 'lib', sel);
        }

        PgTutorial.prototype.showGridElement = function(text, sel) {
            return tooltip(text, 'Click to show the element in the CSS Grid editor.', 'grid', sel);
        }

        PgTutorial.prototype.showWPAction = function(text, sel, assigned) {
            sel = (assigned ? ('.pg-actions-assigned ') : ('.pg-actions-list ')) + sel;
            sel = sel.replace(/"/g, "'");
            return tooltip(text, 'Click to show the action in the WordPress panel.', 'field', 'cms:' + sel);
        }

        PgTutorial.prototype.showWPActionName = function(text, sel, assigned) {
            return this.showWPAction(text, sel + ' name span', assigned)
        }

        PgTutorial.prototype.showWPActionField = function(text, action, field) {
            var sel = `[data-action="${action}"] .crsa-field-${action}`;
            if(field) sel += '-' + field;
            return this.showWPAction(text, sel, true)
        }

        PgTutorial.prototype.showWPActionAddItem = function(text, action) {
            var sel = `[data-action="${action}"] .pg-action-repeat-item-add`;
            return this.showWPAction(text, sel, true)
        }

        PgTutorial.prototype.showInteraction = function(text, action, index, edit) {
            var map = {
                'interaction': 'animation-action',
                'scroll': 'scroll-scene',
                'hide': 'animation-hide',
                'show': 'animation-show',
                'apply': 'animation-apply'
            }

            return `<b><span data-pginfo-animation="${map[action] || action}" data-pginfo-animation-index="${index || ''}" data-pginfo-animation-edit="${edit  || ''}">${text}</span></b>`;
        }

        PgTutorial.prototype.showTimeline = function(text, key, param, extra) {
            return `<b><span data-pginfo-animation-timeline="${key}" data-pginfo-animation-timeline-param="${param || ''}" data-pginfo-animation-timeline-extra="${extra || ''}">${text}</span></b>`;
        }

        PgTutorial.prototype.showProjectFile = function(text, file) {
            file = file || text;
            var sel = `[data-file-name='${file}']`;
            return tooltip(text, 'Click to show the file in the Project panel.', 'field', 'project:' + sel);
        }

        var page = function(name) {
            return tooltip(name + '.html', 'Click to show the page.', 'page', name)
        }
        PgTutorial.prototype.page = page;

        var showField = function(field, show_panel) {
            if(show_panel === undefined && !pgf.kids) show_panel = true;
            var b = field.split('|');
            var border_fields = null;
            if(b.length === 2) {
                border_fields = b[1];
                field = b[0];
            }

            var done = function() {
                if(border_fields) {
                    $el.find(border_fields).each(function(i, e) {
                        pg$Flash($(e), 'orange_border_flash_slow', 3200);
                    })
                }
            }

            var names = {
                prop: 'Attributes panel',
                style: 'Style panel',
                cms: 'WordPress panel',
                project: 'Project panel'
            }

            var a = field.split(':');

            if(a.length === 1) {
                pgRevealElement($(a[0]));
                return;
            }
            var panel = a[0];

            var needs_select = panel !== 'project';

            var view = pgUIViewManager.getView(panel);
            var $el = view.get$Element();

            if(needs_select && !pinegrow.selectedElements.getLastSelected()) {
                pinegrow.showQuickError('Select the element on the page first.');
            } else {

                if (!$el.is(':visible')) {
                    if(!show_panel) {
                        pinegrow.showQuickError('First, open the ' + names[panel] + '.')
                    } else {
                        pinegrow.showUIPanel(panel, false, false, false);
                        setTimeout(function() {
                            showField(field, false);
                        }, 100)
                    }
                } else {
                    var raw_sel = (panel !== 'style' && panel !== 'prop');
                    var $r = $el.find(raw_sel ? a[1] : ('.crsa-field-' + a[1]));
                    if($r.length > 0 && $r.is(':visible')) {

                        var $grid = $r.closest('.pg-grid,.rules-list');
                        if($grid.length) {
                            pgScrollToItemInContainer($r, $grid, false, true, function () {
                                pgRevealElement($r, done);
                            }, -50)
                        } else {
                            pgRevealElement($r, done);
                        }

                        //pgRevealElement($r);
                    }
                }
            }
        }

        PgTutorial.prototype.showField = showField;

        var tag = function(t, content, attributes) {
            var h = `<span class="tag">${escapeHtmlCode(`<${t}`)}`;
            if(attributes) {
                $.each(attributes, function(a, val) {
                    h += ` <span class="attr">${a}="${val}"</span>`;
                })
            }
            h += '&gt;</span>';
            if(content) {
                h += content;
            }
            if(['img'].indexOf(t) < 0) {
                h += `<span class="tag">${escapeHtmlCode(`</${t}>`)}</span>`;
            }
            return h;
        }
        PgTutorial.prototype.tag = tag;




        var revealElementMenuIcon = function(sel, not_found_msg) {
            not_found_msg = not_found_msg || 'The element is not found.';

            !pinegrow.getShowDisplayHelpers() && pinegrow.setShowDisplayHelpers(true, false);
            var pgel = pinegrow.selectedElements.getLastSelected();
            if(!pgel) {
                pinegrow.showQuickError('First, select the element on the page.')
            } else {
                var page = pinegrow.getSelectedPage();
                if(pinegrow.inlineEditManager.isInEdit()) {
                    pinegrow.showQuickError('Text edit mode is already active!');
                } else {
                    var $el = page.$page.find('.crsa-inline-menu:visible ' + sel);
                    if($el.length === 0 || !$el.is(':visible')) {
                        pinegrow.showQuickError(not_found_msg);
                    } else {
                        pgRevealElement($el);
                    }
                }
            }
        }

        PgTutorial.prototype.revealElementMenuIcon = revealElementMenuIcon;

        var revealTextEditButton = function(sel) {
            !pinegrow.getShowDisplayHelpers() && pinegrow.setShowDisplayHelpers(true, false);
            var pgel = pinegrow.selectedElements.getLastSelected();
            if(!pgel) {
                pinegrow.showQuickError('First, select the element on the page.')
            } else {
                var page = pinegrow.getSelectedPage();
                if(!pinegrow.inlineEditManager.isInEdit()) {
                    pinegrow.showQuickError('Text edit mode is not active!');
                } else {
                    pgRevealElement($('.pg-inline-edit-toolbar ' + sel));
                }
            }
        }

        PgTutorial.prototype.revealTextEditButton = revealTextEditButton;

        var tooltips = {
            code_editor: {
                text:`Click on the <i class='icon icon-KODA'></i> icon in the top toolbar to open/close the code editor. Click here to show the icon.`,
                icon: 'icon-KODA',
                click: function() {
                    pgRevealElement($('.commands-toolbar .icon-KODA'));
                }
            },
            page_view: {
                text: `The area where the page is displayed.`,
                click: function() {
                    pgRevealElement($('.page-tab-canvas'));
                }
            },
            check: {
                text: `Press the Check button in the top toolbar to complete the ${pgf.lesson_name}. Click here to show the button.`,
                icon: 'icon-check',
                click: function() {
                    pgRevealElement($('.pg-lesson-check-button'));
                }
            },
            restart_lesson: {
                text: `Click the <i class='icon icon-refresh'></i> in in the ${pgf.lesson_name} heading.`,
                icon: 'icon-refresh',
                click: function() {
                    var $r = $('.pg-tutorial-restart-button');
                    pgScrollToItemInContainer($r, $r.closest('.pg-info-viewer-content'), false, true, function () {
                        pgRevealElement($r);
                    }, -50)
                }
            },
            css_grid: {
                text:`The <i class='icon icon-grid'></i> icon in the selected element menu to open the CSS Grid editor or press ${pgShowKbdCombo('CMD+G')}. Click here to show the icon.`,
                icon: 'icon-grid',
                click: function() {
                    revealElementMenuIcon('.icon-grid')
                }
            },
            edit_text: {
                text:`The <i class='icon icon-text_edit'></i> icon in the selected element menu. Click here to show the icon.`,
                icon: 'icon-text_edit',
                click: function() {
                    revealElementMenuIcon('.icon-text_edit')
                }
            },
            edit_text_done: {
                text: 'The last button in the blue Edit text menu on the selected element. Only visible in Edit text mode. Click to show the button.',
                click: function() {
                    revealTextEditButton('.pg-text-edit-menu-done');
                }
            },
            insert_char: {
                text: 'The &bigstar; button in the blue Edit text menu on the selected element. Only visible in Edit text mode. Click to show the button.',
                click: function() {
                    revealTextEditButton('.pg-insert-char-menu-link');
                }
            },
            edit_text_link: {
                text: 'The Link button in the blue Edit text menu on the selected element. Only visible in Edit text mode. Click to show the button.',
                click: function() {
                    revealTextEditButton('.pg-text-edit-menu-link');
                }
            },
            insert_menu: {
                text: 'Click on the + Insert menu to add elements to the page.',
                icon: 'icon-plus',
                click: function() {
                    pgRevealElement($('.menu-open-library'));
                }
            },
            edit_text_span: {
                text: 'The Span button in the blue Edit text menu on the selected element. Only visible in Edit text mode. Click to show the button.',
                click: function() {
                    revealTextEditButton('.medium-editor-action-phraseSpan');
                }
            },
            duplicate: {
                text:`Click on the <i class='icon icon-duplicate'></i> icon in the selected element menu to duplicate the element.`,
                icon: 'icon-duplicate',
                click: function() {
                    revealElementMenuIcon('.icon-duplicate');
                }
            },
            delete: {
                text:`Click on the <i class='icon icon-bin'></i> icon in the selected element menu to delete the element.`,
                icon: 'icon-duplicate',
                click: function() {
                    revealElementMenuIcon('.icon-bin');
                }
            },
            more_actions: {
                text:`Click on the <i class='icon icon-down'></i> icon in the selected element menu.`,
                icon: 'icon-down',
                click: function() {
                    revealElementMenuIcon('.actions .icon-down');
                }
            },
            attributes_panel: {
                text:`Click on the <i class='icon icon-settings'></i> tab in the panel on the right side of the page.`,
                icon: 'icon-settings',
                click: function() {
                    pinegrow.showUIPanel('prop', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-settings'));
                }
            },
            click_icon: {
                text:`Click on the <i class='icon icon-touch'></i> icon in the selected element menu to test the link.`,
                icon: 'icon-touch',
                click: function() {
                    revealElementMenuIcon('.icon-touch', 'The selected element doesn\'t have a link.');
                }
            },
            style: {
                text:`Click on the <i class='icon icon-Pin15'></i> tab to show the Style panel.`,
                icon: 'icon-Pin15',
                click: function() {
                    pinegrow.showUIPanel('style', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin15'));
                }
            },
            design: {
                text:`Click on the <i class='icon icon-color'></i> tab to show the Design panel.`,
                icon: 'icon-color',
                click: function() {
                    pinegrow.showUIPanel('pgdm', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin15'));
                }
            },
            classstyles: {
                text:`Click on the <i class='icon classes'></i> tab to show the Class Styles panel.`,
                icon: 'icon-classes',
                click: function() {
                    pinegrow.showUIPanel('classstyles', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin15'));
                }
            },
            structure: {
                text:`Click on the <i class='icon icon-Pin11'></i> tab in the panel on the right side of the page.`,
                icon: 'icon-Pin11',
                click: function() {
                    pinegrow.showUIPanel('tree', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin11'));
                }
            },
            tree_panel: {
                text:`<i class='icon icon-Pin11'></i> The Tree panel shows the structure of the HTML document.`,
                icon: 'icon-Pin11',
                click: function() {
                    pinegrow.showUIPanel('tree', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin11'));
                }
            },
            library_panel: {
                text:`<i class='icon icon-import'></i> Use the Library panel to add HTML elements and components to the page.`,
                icon: 'icon-import',
                click: function() {
                    pinegrow.showUIPanel('library', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin11'));
                }
            },
            interactions_panel: {
                text:`<i class='icon icon-113'></i> The Interactions panel is used to create web animations.`,
                icon: 'icon-113',
                click: function() {
                    pinegrow.showUIPanel('animations', false, false, true);
                    //pgRevealElement($('.tab-icon.icon-Pin11'));
                }
            },
            menu_toggle: {
                text: `Click on the <i class='icon icon-Pin13'></i> icon in the top toolbar to switch the element menus on and off.`,
                icon: 'icon-Pin13',
                click: function() {
                    pgRevealElement($('.commands-toolbar .icon-Pin13'));
                }
            },
            test_clicks: {
                text: `Click on the <i class='icon icon-touch'></i> icon in the top toolbar to toggle test clicks.`,
                icon: 'icon-touch',
                click: function() {
                    pgRevealElement($('.commands-toolbar .icon-touch'));
                }
            },
            refresh_page: {
                text: `Click on the <i class='icon icon-refresh'></i> icon in the top toolbar to refresh the page view.`,
                icon: 'icon-refresh',
                click: function() {
                    pgRevealElement($('.commands-toolbar .icon-refresh'));
                }
            },
            empty_placeholder: {
                text:`Click on the <i class='icon icon-Pin21'></i> icon in the selected element menu to toggle the empty placeholder.`,
                icon: 'icon-Pin21',
                click: function() {
                    revealElementMenuIcon('.icon-Pin21', '');
                }
            },
            insert_character: {
                text:`Click on the ★ icon in the Edit text menu to insert a special character.`,
            },
            save_css: {
                text:`Click on the <i class='icon icon-plus-square'></i> icon on the Style attribute.`,
                icon: 'icon-plus-square'
            },
            preview: {
                text: isApp() ? `Click on the <i class='icon icon-27'></i> icon in the toolbar to open the current page in a browser.` : `Click on the <i class='icon icon-27'></i> icon in the toolbar to open the current page in a new browser tab.`,
                icon: 'icon-27',
                click: function() {
                    pgRevealElement($('.preview-link .icon-27, #crsa-topbar .icon-27'));
                }
            },
            project_menu: {
                text:`Click on the Menu button in the top toolbar to edit project settings.`,
                click: function() {
                    pgRevealElement($('.menu-project-menu'));
                }
            },
            wp_menu: {
                text: `Click on the WordPress menu in the top toolbar to export the theme or plugin.`,
                click: function() {
                    pgRevealElement($('.wp-menu-button'));
                }
            },
            wp_panel: {
                text:`<i class='icon icon-WP'></i> Use the WordPress panel to add WordPress actions to elements.`,
                icon: 'icon-WP',
                click: function() {
                    pinegrow.showUIPanel('cms', false, false, true);
                }
            },
            project_panel: {
                text:`<i class='icon icon-Pin11'></i> Project panel displays the project files and folders.`,
                icon: 'icon-Pin11',
                click: function() {
                    pinegrow.showUIPanel('project', false, false, true);
                }
            },
        }

        window.pgInfoPanelgetTooltipInfo = function(key) {
            return tooltips[key] || null;
        }


        var tooltip = function(text, tip, action, param) {
            var tip_text = tip;
            if(tooltips[tip]) tip_text = tooltips[tip].text;
            var s = `<span class="has-tooltip" data-tooltip-key="${tip}" data-html="true" title="${tip_text}" data-tooltip-action="${action || ''}" data-tooltip-action-param="${param || ''}">${text}`;
            if(tooltips[tip] && tooltips[tip].icon) s += ` <i class="icon ${tooltips[tip].icon}"></i>`;
            s += `</span>`;
            return s;
        }

        PgTutorial.prototype.tooltip = tooltip;

        var codeEditor = function(text, file) {
            return tooltip(text,'Click on the code editor tab with this name.', 'code_editor_tab', file || text);
        }

        PgTutorial.prototype.codeEditor = codeEditor;

        PgTutorial.prototype.showStyleSizeTab = function(size, text) {
            text = text || `${size} tab`;
            var sel = `style:niga,.pg-dm-sizes-control [data-size=${size}]`;
            return `<b>${this.field(text, sel)}</b>`
        }

        PgTutorial.prototype.showStyleSizeTabCreateRule = function(text) {
            text = text || `Create the rule`;
            var sel = `style:niga,.pg-css-editor-no-rule-overlay button`;
            return `<b>${this.field(text, sel)}</b>`
        }

        PgTutorial.prototype.showBSField = function( text, sel) {
            var asel = `prop:niga,.crsa-field-bs5-bs${sel}`;
            if(sel.indexOf('.') >= 0) {
                asel = 'prop:niga,' + sel;
            }
            return `<b>${this.field(text, asel)}</b>`
        }

        PgTutorial.prototype.showTWField = function( text, sel) {
            var asel = `prop:niga,.crsa-field-tw-${sel}`;
            if(sel.indexOf('.') >= 0) {
                asel = 'prop:niga,' + sel;
            }
            return `<b>${this.field(text, asel)}</b>`
        }

        pinegrow.addEventHandler('on_info_panel_tooltip_clicked', function(pagenull, node, event) {
            var tip = node.getAttribute('data-tooltip-key');
            if(tip && tooltips[tip] && tooltips[tip].click) {
                event.preventDefault();
                tooltips[tip].click();
            }

            var action = node.getAttribute('data-tooltip-action');
            var param = node.getAttribute('data-tooltip-action-param');

            switch(action) {
                case 'rule':
                    showRule(param);
                    break;
                case 'field':
                    showField(param);
                    break;
                case 'page':
                    pinegrow.openOrShowPage(pinegrow.getCurrentProject().getAbsoluteUrl(param + '.html'), null, true);
                    break;
                case 'code_editor_tab':
                    $('.lm_stack:not(.pg-page-area) .lm_title').each(function(i, e) {
                        if(e.innerHTML === param) {
                            pgRevealElement($(e));
                            return false;
                        }
                    })
                    break;
                case 'lib':
                    var win = pinegrow.getQuickLibWindow();
                    if (!win || !win.win.visible) {
                        pinegrow.showQuickError('The Library panel is not open.')
                    } else {
                        var $r = win.lib.view.get$Element().closest('.pg-quick-win').find(param);
                        if($r.length > 0 && $r.is(':visible')) {
                            var $grid = $r.closest('.pg-grid');
                            if($grid.length) {
                                pgScrollToItemInContainer($r, $grid, false, true, function () {
                                    pgRevealElement($r);
                                }, -50)
                            } else {
                                pgRevealElement($r);
                            }
                        }
                    }
                    break;
                case 'grid':
                    var $grid = $('.pg-dm-grid-view:visible');
                    if($grid.length === 0) {
                        pinegrow.showQuickError('The CSS Grid editor is not open.')
                    } else {
                        pgRevealElement($grid.find(param));
                    }
                    break;
            }
        });


        pinegrow.addEventHandler('on_show_nice_project_view', function(pagenull, project, options, done) {
            if(project.lesson && pgf.kids) {
                options.title = 'The Adventure menu';
                options.pages = false;

                var makeButton = function(label, icon, desc) {
                    return `<i class="icon ${icon}"></i><span>${label}</span><p>${desc}</p>`;
                }
                var my_classes = 'pg-rich-button';

                options.buttons.close.label = makeButton('Go to the Start screen', 'icon-276', 'Pause the adventure and go to the home screen. Everything you did is saved.');
                options.buttons.close.classes = my_classes;
                options.buttons.close.tooltip = null;

                options.buttons.extras = [];

                options.buttons.extras.push({
                    label: makeButton('Restart', 'icon-redo', 'Restart the adventure. Your current work will be lost. Save it as a project first, if you want to keep it.'),
                    classes: my_classes,
                    action: function() {
                        done();
                        var h = new PgProjectMenuHelper(project);
                        h.restartLesson();
                    }
                })

                options.buttons.extras.push({
                    label: makeButton('Save as project', 'icon-project', 'Save the current adventure as a project so that you can keep editing it after you finish the adventure.'),
                    classes: my_classes,
                    action: function() {
                        var p = project;

                        if(pinegrow.backend.canCreateProject()) {
                            var ask = function (val) {
                                pinegrow.showPrompt((val ? `<p><b>A project with this name already exists.</b> Please give it a unique name.</p>` : '') + `<p>Enter the name of the new project:</p>`, 'Save the adventure as a project', val || p.lesson.name, null, null, function (value) {
                                    var url = 'project_' + pgMakeSlug(value);
                                    p.name = value;
                                    pinegrow.backend.saveAsNewProject(p, url, function () {
                                        pinegrow.showQuickMessage('Adventure was saved as a new project ' + value + '.')
                                    })
                                })
                            }

                            ask();
                        }
                    }
                })
            }
        })


        var tutorial = new PgTutorial();

        pinegrow.kidsTutorial = tutorial;

    });
});



