var PgTailwindHelpers = function(options, f, version) {
    if (typeof options.reqVersion != "function") {
        options.reqVersion = function () { return true; }
    }

    var rc = options.rc;

    // More options
    options.framework_prefix = 'tw' + version;
    options.container_selector = '.container';
    options.grid_container_selector = 'grid';
    options.style = '';
    options.no_include_local_question = true;
    options.showGridInMenu = false;
    options.showColumnInMenu = false;
    options.createDisplayHelpers = false;

    var makeResponsiveClass = options.makeResponsiveClass;
    var registerClasses = options.registerClasses;

    var mainPoints = options.spacingPoints;

    var pgFrameworkHelpers = new PgFrameworkHelpers('Tailwind', options, f, version, PgBSMenuPointColumnSize);
    var bm = new PgToggleButtonMaker();

    pgFrameworkHelpers.getCodeForLarge = function () {
        var code_large;
        switch(this.version) {
            case 4:
                code_large = '<pgdiv class="col-1 pinegrow-ui-helper"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
            case 3:
                code_large = '<pgdiv class="col-xs-1 pinegrow-ui-helper"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
        }

        return code_large;
    }

    pgFrameworkHelpers.getCodeForSmall = function () {
        var code_small;
        switch(this.version) {
            case 4:
                code_small = '<pgdiv class="col-1 pinegrow-ui-helper d-block d-sm-none" style="padding-left:0;padding-right:0;"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
            case 3:
                code_small = '<pgdiv class="col-xs-1 pinegrow-ui-helper visible-xs visible-xs-block" style="padding-left:0;padding-right:0;"><pgdiv class="pinegrow-ui-helper"></pgdiv></pgdiv>';
                break;
        }

        return code_small;
    }

    pgFrameworkHelpers.getClassForSize = function(size, prefix, add) {
        if(add === null || add === undefined) {
            add = '';
        }  else {
            add = add + '';
        }
        return prefix + (size == options.size_for_all ? '' : '-' + size) + (add.length ? ('-' + add) : '');
    }

    pgFrameworkHelpers.getColumnSizes = function(pgel_or_node) {
        var re = new RegExp('^col(\\-(' + options.sizes.join('|') + '))?(\\-([0-9]+|auto))?');

        var map = {};

        var list = pgel_or_node instanceof pgParserNode ? pgel_or_node.getClasses() : pgel_or_node.classList;

        list.forEach(function(cls) {
            var m = cls.match(re);
            if(m) {
                var size = m[2] || options.size_for_all;
                var cols = m[4] || '';
                map[size] = cols;
            }
        })
        return map;
    }

    var delayed_tasks = new PgDelayedTasks();

    pgFrameworkHelpers.executeScriptInPageDelayed = function(page, code) {
        delayed_tasks.run('runCode' + code + page.uid, function() {
            pinegrow.executeScriptInPage(page, code);
        }, 1000)
    }

    pgFrameworkHelpers.makeFieldLabelForSize = function(label, size) {
        return '<span class="tw-label">' + label + '</span><span class="tw-device-size">' + size + '</span>';
    }

    pgFrameworkHelpers.getDevicesControlFactory = function(control_id, title, func, add_class, parent_section_id) {

        var gh = pgFrameworkHelpers;
        //Returns function that creates the control

        //We'll just do one size because selecting the device size is done with the device size control in prop panel
        var forSize = function(f) {
            /* options.sizes */['all'].forEach(function(size) {
                f(size, size == options.size_for_all ? '' : '-' + size, options.prefix + control_id + '.' + size);
            })
        }

        return function() {
            //will keep helper functions and vars private

            //Create control, needs a unique id
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //Register subfields.
            c.onDefine = function () {

                //These fields will not be shown when just registered. This is neccessary to get values.
                forSize(function(size, size_token, field_key) {
                    var field = func(size, size_token);
                    field.parent_section_id = parent_section_id;
                    c.registerInputField(field_key, field);
                })
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $("<div/>", {class: 'tw-device-control ' + add_class});
                $('<div class="tw-device-control-title">' + title + '</div>').appendTo($div);
                forSize(function(size, size_token, field_key) {
                    var field_def = func(size, size_token);
                    var field = c.showInputField($div, field_key, field_def);
                    field.$field.attr('data-device-size', size);
                })
                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function(values) {
                //return;
                var none_or_all = true;

                forSize(function(size, size_token, field_key) {
                    var $field = c.get$Field(field_key);
                    if(c.values[field_key]) {
                        $field.addClass('tw-has-value');
                        if(size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('tw-has-value');
                    }
                })
                if(none_or_all) {
                    c.$element.addClass('tw-none-or-all');
                } else {
                    c.$element.removeClass('tw-none-or-all');
                }
            }

            //Return the control
            return c;
        };
    }

    var open_device_sections_map = {};

    pgFrameworkHelpers.onSectionWithDevicesControlsShown = function(secdef, uiSection) {

        var gh = pgFrameworkHelpers;

        //var setActive = function($li, )
        var active = {}
        var no_active = true;

        var sizes = ['all'];//options.sizes;

        active['all'] = true;

        var fields = uiSection.$content.get(0).querySelectorAll('[data-device-size]');


        sizes.forEach(function(size) {
            if(open_device_sections_map[secdef.section_key] && open_device_sections_map[secdef.section_key][size]) {
                active[size] = true;
            }
        })



        fields.forEach(function(field) {
            var size = field.getAttribute('data-device-size');
            if(!active[size]) {
                var has = field.classList.contains('tw-has-value');
                active[size] = has;
                no_active = no_active && !has;
            }
        })

        if(no_active) {
            active[options.size_for_all] = true;
        }


        fields.forEach(function(field) {
            var size = field.getAttribute('data-device-size');
            var display = field.getAttribute('data-device-display') || 'flex';

            if(active[size]) {
                pgShow(field, display);
            } else {
                pgHide(field);
            }
        })



/*
        var html = '<ul class="tw-device-selector">';
        sizes.forEach(function(size) {
            html += '<li data-size="' + size + '" class="' + (active[size] ? 'tw-active' : '') + '">' + size + '</li>';
        })
        html += '</ul>';
        var $c = $(html);
        uiSection.$title.append($c);


        $c.on('click', 'li', function(e) {
            e.preventDefault();
            e.stopPropagation();

            var $li = $(this);
            var size = $li.attr('data-size');
            var a = $li.hasClass('tw-active');
            a = !a;

            active[size] = a;

            if(!a) {
                if($c.find('li.tw-active').length == 1) {
                    pinegrow.showQuickError('At least one device size must be active.');
                    return;
                }
            }

            if(!open_device_sections_map[secdef.section_key]) {
                open_device_sections_map[secdef.section_key] = {};
            }

            open_device_sections_map[secdef.section_key][size] = a;

            var $fields = uiSection.$content.find('[data-device-size="' + size + '"]');
            $fields.each(function(i, field) {
                var display = field.getAttribute('data-device-display') || 'flex';

                if (a) {
                    pg$Show($(field), display);
                    $li.addClass('tw-active');
                } else {
                    pg$Hide($(field));
                    $li.removeClass('tw-active');
                }
            });
            updateLabels();
        })
        */


        var updateVisibility = function() {
            var page = pinegrow.getSelectedPage();
            fields.forEach(function(field) {
                var size = field.getAttribute('data-device-size');
                if(gh.isSizeVisible(page, size)) {
                    field.classList.remove('tw-size-not-visible');
                } else {
                    field.classList.add('tw-size-not-visible');
                }
            })

            /*

            $c.children().each(function(i, li) {
                var size = li.getAttribute('data-size');
                if(gh.isSizeVisible(page, size)) {
                    li.classList.remove('tw-size-not-visible');
                } else {
                    li.classList.add('tw-size-not-visible');
                }
            })
            */

        }

        var updateLabels = function() {
            var num_shown = 0;
            var all_shown = false;

            sizes.forEach(function(size) {
                if(active[size]) {
                    num_shown++;
                    if(size == options.size_for_all) {
                        all_shown = true;
                    }
                }
            })

            var $controls = uiSection.$content.find('.tw-device-control');

            if(num_shown > 1) {
                $controls.removeClass('tw-single-size');
            } else {
                $controls.addClass('tw-single-size');
            }
            if(all_shown) {
                $controls.addClass('tw-for-all-shown');
            } else {
                $controls.removeClass('tw-for-all-shown');
            }
        }

        updateVisibility();
        updateLabels();

    }

    var marginPaddingObjs = {};
    var marginPaddingActiveBtn = {};
    var spacingTypes = {
        m: {
            directions: ['', 'x', 'y', 't', 'l', 'r', 'b'],
            default_points: options.getSpacingPoints().concat(['px', 'auto']),
            points: []
        },
        p: {
            directions: ['', 'x', 'y', 't', 'l', 'r', 'b'],
            default_points: options.getSpacingPoints().concat(['px']),
            points: []
        }
    }

    var positionObjs = {};
    var positionActiveBtn = {};
    var posDirections = ['', 'x', 'y', 'top', 'left', 'right', 'bottom'];
    var default_posPoints = ['0', 'auto'];

    var posPoints = [];

    var getPositionPoints = function() {
        if(options.getTWVersion() < 2) {
            return default_posPoints;
        } else {
            var r = options.getSpacingPoints();
            r.push('1/2');
            r.push('1/3');
            r.push('2/3');
            r.push('1/4');
            r.push('2/4');
            r.push('3/4');

            var m = ['auto', 'full'];

            m = m.concat(r);
            for(var i = 0; i < r.length; i++) {
                m.push('-' + r[i])
            }
            return m;
        }
    }
    var transformObjs = {};
    var transformTypes = {
        translate: {
            directions: ['x', 'y'],
            default_points: options.getSpacingPoints().concat(["px", "full", "1/2"]),
            points: [],
        },
        scale: {
            directions: ['x', 'y'],
            default_points: ["0", "50", "75", "90", "95", "100", "105", "110", "125", "150"],
            points: []
        },
        skew: {
            directions: ['x', 'y'],
            default_points: ["0", "1", "2", "3", "6", "12"],
            points: []
        },
        rotate: {
            directions: [''],
            default_points: ["0", "1", "2", "3", "6", "12", "45", "90", "180"],
            points: []
        }
    }

    var borderObjs = {};
    var borderTypes = {
        "border": {
            directions: ['', 't', 'r', 'b', 'l'],
            default_points: ["", "0", "2", "4", "8"],
            points: []
        },
        "rounded": {
            directions: ['', 't', 'r', 'b', 'l', "tl", "tr", "br", "bl"],
            default_points: ["none", "sm", "", "md", "lg", "xl", "2xl", "3xl", "full"],
            points: []
        }
    }
    var borderBtn = {};

    pgFrameworkHelpers.onTailwindUpdated = function () {
        for (var type in spacingTypes) {
            spacingTypes[type].points = [];
        }

        posPoints = [];

        for (var type in transformTypes) {
            transformTypes[type].points = [];
        }

        for (var type in borderTypes) {
            borderTypes[type].points = [];
        }
    }

    var getNumValue = function (base, value) {
        if (!value) return null;

        var minusValue = false;
        var currValue = value;
        if (currValue.startsWith('-')) {
            minusValue = true;
            currValue = currValue.substr(1, currValue.length);
        }

        if (currValue.startsWith(base)) {
            currValue = currValue.replace(base, '');
            if (currValue.length > 0) {
                currValue = currValue.substr(1, currValue.length);
            }
        }
        else {
            return null;
        }

        return minusValue ? `-${currValue}` : currValue;
    }

    var getNumFromMarginPaddingValue = function (base, value) {
        return getNumValue(base, value);
    }

    var getNewMarginPaddingValueFor = function (base, type, direction, size, field_key, values, valDirection) {
        var num = getNumFromMarginPaddingValue(base, values[field_key + valDirection]) + '';

        var prefix = '';
        if (num.startsWith('-')) {
            prefix = '-';
            num = num.substr(1, num.length);
        }

        return (size != options.size_for_all ? `${size}:` : '') + prefix + type + direction + '-' + num;
    }

    var getMPDirectionsFor = function (direction) {
        switch (direction) {
            case '':
                return ['t', 'l', 'r', 'b'];
            case 'x':
                return ['l', 'r'];
            case 'y':
                return ['t', 'b'];
            default:
                return [direction];
        }
    }

    var getClassNameForPosition = function (direction, size, value) {
        if (!value) return null;

        var currDir = direction;
        if (direction == '') {
            currDir = 'inset';
        }
        else if (['x', 'y'].indexOf(direction) > -1) {
            currDir = 'inset-' + direction;
        }

        var prefix = '';
        if (value.startsWith('-')) {
            prefix = '-';
            value = value.substr(1, value.length);
        }

        return (size != options.size_for_all ? `${size}:` : '')  + prefix + currDir + '-' + value;
    }

    var getPosDirectionsFor = function (direction) {
        switch (direction) {
            case '':
                return ['top', 'left', 'right', 'bottom'];
            case 'x':
                return ['left', 'right'];
            case 'y':
                return ['top', 'bottom'];
            default:
                return [direction];
        }
    }

    var getClassNameForMarginAndPadding = function (type, direction, size, num) {
        if (!num) return null;
        num += ''; // Change it to string

        var prefix = '';
        if (num.startsWith('-')) {
            prefix = '-';
            num = num.substr(1, num.length);
        }
        return (size != options.size_for_all ? `${size}:` : '') + prefix + type + direction + '-' + num;
    }

    var setMPValueOnFields = function (type, direction, size, value) {
        var dirs = getMPDirectionsFor(direction);

        for (var i = 0; i < dirs.length; i++) {
            var dir = dirs[i];
            var pgField = marginPaddingObjs[size][type + dir];
            var newClass = getClassNameForMarginAndPadding(type, dir, size, value);
            pgField && pgField.setValue( newClass );
        }
    }

    var getNumFromPositionValue = function (base, value) {
        return getNumValue(base, value);
    }

    var setPosValueOnFields = function (direction, size, value) {
        var dirs = getPosDirectionsFor(direction);

        for (var i = 0; i < dirs.length; i++) {
            var dir = dirs[i];
            var pgField = positionObjs[size][dir];
            var newClass = getClassNameForPosition(dir, size, value);
            pgField && pgField.setValue( newClass );
        }
    }

    var getNumFromBorderValue = function (base, value, points) {
        var num = getNumValue(base, value);
        //if(num === '')
        return points.indexOf(num) > -1 ? num : null;
    }

    var setBorderValueOnFields = function (type, direction, size, value) {
        var dirs = getBorderDirectionsFor(type, direction);

        for (var i = 0; i < dirs.length; i++) {
            var dir = dirs[i];
            var pgField = borderObjs[size][type][dir];
            var newClass = getClassNameForBorder(type, dir, size, value);
            pgField && pgField.setValue( newClass );
        }
    }

    var getNumFromScaleValue = function (base, value) {
        return getNumValue(base, value);
    }

    var getScaleDirectionsFor = function (direction) {
        switch (direction) {
            case '':
                return ['x', 'y'];
            default:
                return [direction];
        }
    }

    var getClassNameForScale = function(type, dir, size, value, clsPrefix) {
        if (!value) return null;

        var size_t = (size == options.size_for_all ? '' : size + ":")
        var currDir = (dir == '' ? '' : dir + '-');


        return size_t + clsPrefix + type + '-' + currDir + value;
    }

    var getClassNameForBorder = function (type, direction, size, value) {
        if (!value && value != "") return null;

        var size_t = (size == options.size_for_all ? '' : `${size}:`);
        var mainBase = size_t + type;
        var currDirBase = (direction ? '-' + direction : '');

        if(type === 'rounded' && value === 'normal') {
            //value = '';
        }
        if(type === 'border' && value === 'px') {
            //value = '';
        }
        var valueBase = (value ? '-' + value : '');

        return mainBase + currDirBase + valueBase;
    }

    var getBorderDirectionsFor = function (type, direction) {
        if (type == 'border') {
            switch (direction) {
                case '':
                    return ['t', 'l', 'r', 'b'];
                default:
                    return [direction];
            }
        }
        else if (type == 'rounded') {
            switch (direction) {
                case '':
                    return ['tl', 'tr', 'bl', 'br'];
                case 't':
                    return ['tr', 'tl'];
                case 'l':
                    return ['tl', 'bl'];
                case 'r':
                    return ['tr', 'br'];
                case 'b':
                    return ['bl', 'br'];
                default:
                    return [direction];
            }
        }
    }

    var toggleIconActivation = function (e) {
        var $target = $(e.target);
        var $icon = $target.closest('.change-all-icon,.change-all-border-raduis');
        $icon.toggleClass('active');
        var isActive = $icon.hasClass('active');

        if (isActive) {
            var getKey = function (key, type, direction) {
                return key.replace('{type}', type)
                        .replace('{direction}', direction)
            }


            var valuesObj = $icon.data('valueObject')[options.size_for_all];
            var directions = $icon.data('directions');
            var type = $icon.data('type');

            for (var i = 0; i < directions.length; i++) {
                var dir = directions[i];
                var frKey = $icon.data('frKey');
                var sdKey = $icon.data('sdKey');

                var pgField = valuesObj;
                if (frKey) {
                    var key = getKey(frKey, type, dir);
                    pgField = pgField[key];
                }

                if (sdKey) {
                    var key = getKey(sdKey, type, dir);
                    pgField = pgField[key];
                }

                var value = pgField.getValue();
                if (value) pgField.setValue(value, true);
            }
        }
    }

    var isAllowedToOverwriteMP = function (value_direction, type_direction) {
        if (type_direction == '') {
            return true;
        } else if (type_direction == 'x') {
            if (['r', 'l'].indexOf(value_direction) > -1) return true;
        } else if (type_direction == 'y') {
            if (['t', 'b'].indexOf(value_direction) > -1) return true;
        }

        return false;
    }

    var isAllowedToOverwritePos = function (value, dir) {
        if (dir == '') {
            return true;
        } else if (dir == 'x') {
            if (['right', 'left'].indexOf(value) > -1) return true;
        } else if (dir == 'y') {
            if (['top', 'bottom'].indexOf(value) > -1) return true;
        }

        return false;
    }

    var isAllowedToOverwriteBor = function (value, dir) {
        if (dir == '') {
            return true;
        } else if (dir == 't') {
            if (['tl', 'tr'].indexOf(value) > -1) return true;
        } else if (dir == 'r') {
            if (['tr', 'br'].indexOf(value) > -1) return true;
        } else if (dir == 'b') {
            if (['bl', 'br'].indexOf(value) > -1) return true;
        } else if (dir == 'l') {
            if (['tl', 'bl'].indexOf(value) > -1) return true;
        }

        return false;
    }

    var getBorderDirectionProp = function (type, direction_token) {
        if (type == 'border') {
            switch (direction_token) {
                case 't':
                    return 'top';
                case 'l':
                    return 'left';
                case 'r':
                    return 'right';
                case 'b':
                    return 'bottom';
                default:
                    return null;
            }
        }
        else if (type == 'rounded') {
            switch (direction_token) {
                case 'tl':
                case 'l':
                case 't':
                    return 'top-left';
                case 'tr':
                case 'r':
                    return 'top-right';
                case 'bl':
                    return 'bottom-left';
                case 'br':
                case 'b':
                    return 'bottom-right';
                default:
                    return null;
            }
        }
    }

    var getOptionsSpacing = function (base, type) {
        if(options.twc.has_data()) {
            return options.twc.getOptions(
                base, // BASE
                type === 'p' ? 'padding' : 'margin', // PROP
                type === 'p' ? /padding/ : /margin/, // PROP RE
                type === 'm', // MINUS
                function (options) { // FILTER FUN
                    var updatePoints = false;
                    if (spacingTypes[type].points.length == 0) updatePoints = true;

                    var oldValue = "-";
                    for (var i = 0; i < options.length; i++) {
                        var { name } = options[i];

                        if (oldValue.startsWith('-') && !name.startsWith('-') && name !== "") {
                            options.splice( i, 0, { key: '', name: '' });
                        }

                        if (updatePoints) {
                            spacingTypes[type].points.push(name);
                        }
                        oldValue = name;
                    }
                }
            );
        }

        spacingTypes[type].points = spacingTypes[type].default_points;
        var points = spacingTypes[type].points.slice(0);

        var optionsValue = [];
        var restOptions = [];

        var fn = "unshift";
        if (type == 'p') {
            var pointPx = points[points.length - 1];
            points.splice(points.length - 1, 1);
            optionsValue.push({ 'key': base + pointPx, 'name' : pointPx });
            optionsValue.push({ 'key' : '', 'name' : '' });
        }
        else {
            restOptions = points.splice(points.length - 2, 2);
        }

        for (var i = 0; i < points.length; i++) {
            var ni = points[i];
            optionsValue[fn]({ 'key': base + points[i], 'name': ni, num: ni });
        }

        if (type == 'm') {
            for (var i = 0; i < restOptions.length; i++) {
                optionsValue.push({ 'key': base + restOptions[i], 'name' : restOptions[i] });
            }

            optionsValue.push({ 'key' : '', 'name' : '' });
            for (var i = 1; i < points.length; i++) {
                var ni = -1 * points[i];
                optionsValue.push({ 'key': '-' + base + points[i], 'name': ni, num: ni });
            }
        }

        return optionsValue;
    }
    rc.add(function() { registerClasses( 'margin.padding.all.m', getOptionsSpacing('m-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.mx', getOptionsSpacing('mx-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.my', getOptionsSpacing('my-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.mt', getOptionsSpacing('mt-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.ml', getOptionsSpacing('ml-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.mr', getOptionsSpacing('mr-', 'm'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.mb', getOptionsSpacing('mb-', 'm'), 'tw.spacing')});

    rc.add(function() { registerClasses( 'margin.padding.all.p', getOptionsSpacing('p-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.px', getOptionsSpacing('px-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.py', getOptionsSpacing('py-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.pt', getOptionsSpacing('pt-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.pl', getOptionsSpacing('pl-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.pr', getOptionsSpacing('pr-', 'p'), 'tw.spacing')});
    rc.add(function() { registerClasses( 'margin.padding.all.pb', getOptionsSpacing('pb-', 'p'), 'tw.spacing')});

    var getOptionsPosition = function (base, direction, skip_fulling_points) {
        if(options.twc.has_data()) {
            return options.twc.getOptions(
                base, // BASE
                direction, // PROP
                null, // PROP RE
                true, // MINUS
                function (options) { // FILTER FUN
                    var updatePoints = false;
                    if (posPoints.length == 0 && !skip_fulling_points) updatePoints = true;

                    var oldValue = "-";
                    for (var i = 0; i < options.length; i++) {
                        var { name } = options[i];

                        if (oldValue.startsWith('-') && !name.startsWith('-') && name !== "") {
                            options.splice( i, 0, { key: '', name: '' });
                        }

                        if (updatePoints) {
                            posPoints.push(name);
                        }
                        oldValue = name;
                    }
                }
            );
        }
        var optionsValue = [];
        posPoints = getPositionPoints();

        optionsValue.push({ 'key' : '', 'name': '' });
        for (var i = 0; i < posPoints.length; i++) {
            var point = posPoints[i];

            optionsValue.push({ 'key' : base + point, 'name': point });
        }

        return optionsValue;
    }

    //We're passing section ids here because these classes are not auto-detected from scaning field definitions
    rc.add(function() { registerClasses( 'position.all', getOptionsPosition('inset-', 'top', true), 'tw.position')});
    rc.add(function() { registerClasses( 'position.all.x', getOptionsPosition('inset-x-', 'left'), 'tw.position')});
    rc.add(function() { registerClasses( 'position.all.y', getOptionsPosition('inset-y-', 'top'), 'tw.position')});
    rc.add(function() { registerClasses( 'position.dir.all.top', getOptionsPosition('top-', 'top'), 'tw.position')});
    rc.add(function() { registerClasses( 'position.dir.all.left', getOptionsPosition('left-', 'left'), 'tw.position')});
    rc.add(function() { registerClasses( 'position.dir.all.right', getOptionsPosition('right-', 'right'), 'tw.position')});
    rc.add(function() { registerClasses( 'position.dir.all.bottom', getOptionsPosition('bottom-', 'bottom'), 'tw.position')});

    var getOptionsTransform = pgFrameworkHelpers.getOptionsTransform = function (base, type, direction, skip_fulling_points) {
        if(options.twc.has_data()) {
            var prop;
            if(direction === 'both') {
                prop = [`--tw-${type}-x`, `--tw-${type}-y`];
            } else {
                prop = `--tw-${type}${direction ? '-' + direction : ''}`;
            }
            return options.twc.getOptions(
                base, // BASE
                prop, // PROP
                null, // PROP RE
                type != 'scale', // MINUS
                function (options) { // FILTER FUN
                    var updatePoints = false;
                    if (transformTypes[type].points.length == 0 && !skip_fulling_points) updatePoints = true;

                    var oldValue = "-";
                    for (var i = 0; i < options.length; i++) {
                        var { name } = options[i];

                        if (oldValue.startsWith('-') && !name.startsWith('-') && name !== "") {
                            options.splice( i, 0, { key: '', name: '' });
                        }

                        if (updatePoints) {
                            transformTypes[type].points.push(name);
                        }
                        oldValue = name;
                    }
                }
            );
        }
        var optionsValue = [];

        transformTypes[type].points = transformTypes[type].default_points
        var points = transformTypes[type].points;

        var fn = "unshift";
        if (type == "scale") {
            fn = "push";
            optionsValue.push({ 'key': '', 'name': '' });
        }

        for (var i = 0; i < points.length; i++) {
            var point = crsaCamelize(points[i]);
            var ni = parseInt(points[i]);
            optionsValue[fn]({ 'key' : base + points[i], 'name': point, num: ni });
        }


        if (type != "scale") {
            optionsValue.push({ 'key': '', 'name': '' });

            for (var i = 1; i < points.length; i++) {
                var option = {
                    'key': '-' + base + points[i]
                }
                if (isNaN(parseInt(points[i])) || points[i].indexOf('/') > -1) {
                    option['name'] = '-' + crsaCamelize(points[i]);
                }
                else {
                    var ni = -1 * points[i];
                    option['name'] = ni;
                    option['num'] = ni;
                }
                optionsValue.push(option);
            }
        }

        return optionsValue;
    }


    rc.add(function() { registerClasses( 'transform.translate.x.all', getOptionsTransform('translate-x-', 'translate', 'x'), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.translate.y.all', getOptionsTransform('translate-y-', 'translate', 'y'), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.scale.all', getOptionsTransform('scale-', 'scale', 'both', true), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.scale.x.all', getOptionsTransform('scale-x-', 'scale', 'x'), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.scale.y.all', getOptionsTransform('scale-y-', 'scale', 'y'), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.rotate.all', getOptionsTransform('rotate-', 'rotate', ''), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.skew.x.all', getOptionsTransform('skew-x-', 'skew', 'x'), 'tw.transform')});
    rc.add(function() { registerClasses( 'transform.skew.y.all', getOptionsTransform('skew-y-', 'skew', 'y'), 'tw.transform')});

    var getOptionsBorder = function (base, type, dir) {
        if(options.twc.has_data()) {
            var prop = '';
            if (type == 'border') {
                prop = "border" + (dir ? '-' + dir : '') + '-width';
            }
            else if (type == 'rounded') {
                prop = "border" + (dir ? '-' + dir : '') + '-radius'
            }
            return options.twc.getOptions(
                base, // BASE
                prop, // PROP
                null, // PROP RE
                false,
                function (options) { // FILTER FUN
                    var updatePoints = false;
                    if (borderTypes[type].points.length == 0) updatePoints = true;

                    for (var i = 0; i < options.length; i++) {
                        var option = options[i];
                        var { name, key } = option;
                        var point = name;

                        if (name.startsWith('-')) {
                            name = name.substring(1, name.length);
                            point = name;
                        }
                        else if (name == "" && key != "") {
                            if (type == 'border') {
                                name = '1px';
                            }
                            else if (type == 'rounded') {
                                name = 'normal';
                            }
                        }
                        if(point === 'normal' || point === '1px') {
                            point = '';
                        }

                        options[i] = { key, name };

                        if (updatePoints) {
                            borderTypes[type].points.push(point);
                        }
                    }
                }
            );
        }
        var optionsValue = [];

        borderTypes[type].points = borderTypes[type].default_points;
        var points = borderTypes[type].points;

        for (var i = 0; i < points.length; i++) {
            var point = points[i];
            if (point.length == 0) {
                if (type == "border") {
                    optionsValue.push({ key: base, name: "1px" });
                }
                else if (type == "rounded") {
                    optionsValue.push({ key: base, name: "Normal" });
                }
            }
            else {
                var option = { key: base + "-" + point, name: crsaCamelize(point) };
                var pointNum = parseInt(point);
                if (!isNaN(pointNum)) option['num'] = pointNum;
                optionsValue.push(option);
            }
        }

        return optionsValue;
    }
    rc.add(function() { registerClasses( 'border.border.all', getOptionsBorder('border', 'border', ''), 'tw.border')});
    rc.add(function() { registerClasses( 'border.border.t.all', getOptionsBorder('border-t', 'border', 'top'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.border.l.all', getOptionsBorder('border-l', 'border', 'left'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.border.r.all', getOptionsBorder('border-r', 'border', 'right'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.border.b.all', getOptionsBorder('border-b', 'border', 'bottom'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded', getOptionsBorder('rounded', 'rounded', ''), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.t.all', getOptionsBorder('rounded-t', 'rounded', 'top-left'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.l.all', getOptionsBorder('rounded-l', 'rounded', 'top-left'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.r.all', getOptionsBorder('rounded-r', 'rounded', 'top-right'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.b.all', getOptionsBorder('rounded-b', 'rounded', 'bottom-right'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.tl.all', getOptionsBorder('rounded-tl', 'rounded', 'top-left'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.tr.all', getOptionsBorder('rounded-tr', 'rounded', 'top-right'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.bl.all', getOptionsBorder('rounded-bl', 'rounded', 'bottom-left'), 'tw.border')});
    rc.add(function() { registerClasses( 'border.rounded.br.all', getOptionsBorder('rounded-br', 'rounded', 'bottom-right'), 'tw.border')});

    pgFrameworkHelpers.getDevicesMarginAndPaddingControlFactory = function(control_id, func) {
        //Returns function that creates the control

        var forSize = function(f) {
            /* options.sizes */ ['all'].forEach(function(size) {
                f(size, size == options.size_for_all ? '' : '-' + size, options.prefix + control_id + '.' + size);
            })
        }

        return function() {
            //will keep helper functions and vars private
            var updateFieldsValues = function (size, size_token, field_key) {
                var getPureValue = function (type) {
                    var type_field_key = field_key + '.' + type;
                    var directions = spacingTypes[type].directions;
                    var values = [];

                    for (var i = 0; i < 3; i++) {
                        var dir = directions[i]
                        var value = c.values[type_field_key + dir];
                        if (value) {
                            var value_arr = value.split('-');

                            values.push({
                                direction: dir,
                                value: value_arr[value_arr.length - 1]
                            });
                        }
                    }

                    if (values.length > 0) {
                        return values;
                    }
                    return null;
                }

                var getNewValue = function (type, direction, value) {
                    return type + direction + (size_token ? '-' + size_token : '') + '-' + value;
                }

                for (var type in spacingTypes) {
                    var values = getPureValue(type);
                    var $activeBtn = marginPaddingActiveBtn[size][type];

                    $activeBtn.removeClass('active');

                    if (values && values.length > 0) {
                        for (var k = 0; k < values.length; k++) {
                            var value_obj = values[k];
                            var pure_value = value_obj.value;
                            var type_direction = value_obj.direction;

                            var directions = spacingTypes[type].directions;
                            for (var j = 3; j < directions.length; j++) {
                                var direction = directions[j];
                                if (!isAllowedToOverwriteMP(direction, type_direction)) continue;

                                var type_dir_field_key = field_key + '.' + type + direction;

                                if (!c.values[type_dir_field_key]) {
                                    var pgField = marginPaddingObjs[size][type + direction];
                                    var newValue = getNewValue(type, direction, pure_value);

                                    c.values[type_dir_field_key] = newValue;
                                    pgField.setValue(newValue);
                                }
                            }

                            if (value_obj.direction == '') {
                                $activeBtn.addClass('active');
                                break;
                            }
                        }
                    }
                }
            }

            var hasValue = function (field_key) {
                for (var type in spacingTypes) {
                    var directions = spacingTypes[type].directions;

                    for (var j = 3; j < directions.length; j++) {
                        var direction = directions[j];

                        var full_field_key = field_key + '.' + type + direction;
                        if (c.values[full_field_key]) return true;
                    }
                }

                return false;
            }

            //Create control, needs a unique id
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //Register subfields.
            c.onDefine = function () {
                forSize(function(size, size_token, field_key) {
                    for (var type in spacingTypes) {
                        var directions = spacingTypes[type].directions;

                        for (var j = 0; j < directions.length; j++) {
                            var field = func(size, type, directions[j], size_token);
                            var full_field_key = field_key + '.' + type + directions[j];
                            c.registerInputField(full_field_key, field);
                        }
                    }
                });
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $('<div class="crsa-css-gui bs4-margin-and-padding"></div>');
                var type = '', direction = '';
                var field_def, field, full_field_key;

                forSize(function(size, size_token, field_key) {
                    var $container = $('<div class="margin-and-padding-container text-center" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="tw-device-size">' + size + '</span>').appendTo($container);

                    marginPaddingObjs[size] = {};
                    marginPaddingActiveBtn[size] = {};

                    // Margin top
                    type = 'm'; direction = 't';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-top crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Margin left
                    direction = 'l';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-left crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    var $marginBorder = $('<div class="margin-border"></div>').appendTo($container);

                    var $marginIcon = $('<div class="change-all-margin-icon change-all-icon">\
                        <i class="icon-same"></i>\
                    </div>').appendTo($marginBorder);

                    var directions = spacingTypes[type].directions;
                    marginPaddingActiveBtn[size]['m'] = $marginIcon;
                    $marginIcon.data('frKey', "{type}{direction}");
                    $marginIcon.data('type', type);
                    $marginIcon.data('directions', directions.slice(3, directions.length));
                    $marginIcon.data('valueObject', marginPaddingObjs);
                    $marginIcon.on('click', toggleIconActivation);

                    var $paddingContainer = $('<div class="padding-container"></div>').appendTo($marginBorder);

                    // Padding top
                    type = 'p'; direction = 't';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-top crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Padding left
                    direction = 'l';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-left crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    var $paddingIcon = $('<div class="padding-border">\
                        <div class="change-all-padding-icon change-all-icon"><i class="icon-same"></i></div>\
                    </div>').appendTo($paddingContainer).find('> .change-all-icon');

                    directions = spacingTypes[type].directions;
                    marginPaddingActiveBtn[size]['p'] = $paddingIcon;
                    $paddingIcon.data('frKey', "{type}{direction}");
                    $paddingIcon.data('type', type);
                    $paddingIcon.data('directions', directions.slice(3, directions.length));
                    $paddingIcon.data('valueObject', marginPaddingObjs);
                    $paddingIcon.on('click', toggleIconActivation);

                    // Padding right
                    direction = 'r';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-right crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Padding bottom
                    direction = 'b';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($paddingContainer, full_field_key, field_def);
                    field.$field.addClass('crsa-field-padding-bottom crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;

                    // Margin right
                    type = 'm'; direction = 'r';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-right crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;


                    // Margin bottom
                    direction = 'b';
                    field_def = func(size, type, direction, size_token);
                    full_field_key = field_key + '.' + type + direction;
                    field = c.showInputField($container, full_field_key, field_def);
                    field.$field.addClass('crsa-field-margin-bottom crsa-small-input crsa-align-center-input crsa-inline-container');
                    marginPaddingObjs[size][type + direction] = field;
                });

                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function(values) {
                //return;
                var none_or_all = true;

                forSize(function(size, size_token, field_key) {
                    // Check padding and margin
                    updateFieldsValues(size, size_token, field_key);

                    var $field = c.$element.find('> [data-device-size="' + size + '"]');
                    if (hasValue(field_key)) {
                        $field.addClass('tw-has-value');
                        if (size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('tw-has-value');
                    }
                })
                if(none_or_all) {
                    c.$element.addClass('tw-none-or-all');
                } else {
                    c.$element.removeClass('tw-none-or-all');
                }
            }

            //Return the control
            return c;
        };
    }

    pgFrameworkHelpers.getSpacingControl = function (parent_section_id) {
        return pgFrameworkHelpers.getDevicesMarginAndPaddingControlFactory('margin.padding', function(size, type, direction_token, size_token) {

            var getAllOptions = function () {
                // var size_t = (size_token ? `${size}:` : '');
                var base =  type + direction_token + '-';
                return getOptionsSpacing(base, type);
            }

            var field = {
                type : 'select',
                name : null,
                //placeholder : direction_token,
                // show_empty : false,//true,
                default_value: '',
                drag_starting_index: spacingTypes[type].points.length,
                options : getAllOptions,
                action : 'custom',
                can_make_change_action: 'apply_class',
                preview_class: true,
                preview_class_from_value: function(value, fn, fdef) {
                    return value;
                },
                /*
                preview_class_get_values: function(fn, fdef) {

                },

                 */
                draggable_list : true,
                async_options_function: true,
                fill_list_on_open: true,
                parent_section_id : parent_section_id,
                //on_field_updated: options.onFieldUpdated,
                more_prop: {
                    size: size,
                    type: type,
                    direction: direction_token
                },
                preview_class: true,
                responsive_class: function(val, pgel) {
                    //dummy, just return
                    //a hack to get prefix for class preview
                    if(val === '___dummy___') {
                        return makeResponsiveClass(val, pgel);
                    }
                    return val;
                },
                get_value: function(pgel, fn, values, fv) {
                    var class_setter = pgel.getPage().classSetter;

                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (class_setter.hasClass(pgel, makeResponsiveClass(optionsValue[i].key))) return optionsValue[i].key;
                    }

                    var size      = fv.more_prop.size,
                        type      = fv.more_prop.type,
                        direction = fv.more_prop.direction;

                    if (direction && direction != 'x' && direction != 'y') {
                        var field_key = options.prefix + 'margin.padding.' + size + '.' + type;
                        if (values[field_key]) {
                            var base = type;
                            var currValue = getNewMarginPaddingValueFor(base, type, direction, size, field_key, values, '');
                            values[field_key + direction] = currValue;
                            return currValue
                        }

                        if (direction == 'r' || direction == 'l') {
                            if (values[field_key + 'x']) {
                                var base = type + 'x';
                                var currValue = getNewMarginPaddingValueFor(base, type, direction, size, field_key, values, 'x');
                                values[field_key + direction] = currValue;
                                return currValue;
                            }
                        }
                        else if (direction == 't' || direction == 'b') {
                            if (values[field_key + 'y']) {
                                var base = type + 'y';
                                var currValue = getNewMarginPaddingValueFor(base, type, direction, size, field_key, values, 'y');
                                values[field_key + direction] = currValue;
                                return currValue;
                            }
                        }
                    }

                    return '';
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var class_setter = pgel.getPage().classSetter;

                    class_setter.removeClass(pgel, makeResponsiveClass(oldValue));

                    var size = fdef.more_prop.size,
                        type = fdef.more_prop.type,
                        direction = fdef.more_prop.direction;

                    var num = getNumFromMarginPaddingValue(type + direction, value);

                    var directions = spacingTypes[type].directions;
                    var points = spacingTypes[type].points;

                    var $icon = marginPaddingActiveBtn[size][type];
                    var isActive = $icon.hasClass('active');

                    var valuesNum = [];
                    for (var i = 0; i < directions.length; i++) {
                        var dir = directions[i];

                        for (var j = 0; j < points.length; j++) {
                            var point = points[j];
                            var classFound = false;
                            var className = makeResponsiveClass(getClassNameForMarginAndPadding(type, dir, size, point));
                            if (class_setter.hasClass(pgel, className)) {
                                classFound = true;
                            }
                            else if (type == 'm') {
                                className = makeResponsiveClass(getClassNameForMarginAndPadding(type, dir, size, '-' + point));
                                if (class_setter.hasClass(pgel, className)) {
                                    point = '-' + point;
                                    classFound = true;
                                }
                            }

                            if (classFound) {
                                valuesNum[dir] = point;

                                var dirs = getMPDirectionsFor(dir);
                                for (var k = 0; k < dirs.length; k++) {
                                    valuesNum[dirs[k]] = point;
                                }
                            }
                        }
                    }

                    var top    = direction == 't' ? num : valuesNum['t'];
                    var left   = direction == 'l' ? num : valuesNum['l'];
                    var right  = direction == 'r' ? num : valuesNum['r'];
                    var bottom = direction == 'b' ? num : valuesNum['b'];

                    var removeDirections = {
                        '': valuesNum[''],
                        'x': valuesNum['x'],
                        'y': valuesNum['y'],
                        't': valuesNum['t'],
                        'l': valuesNum['l'],
                        'r': valuesNum['r'],
                        'b': valuesNum['b'],
                    };
                    var addDirections = {};
                    if (isActive || (top && top == left && left == right && right == bottom)) {
                        addDirections[''] = isActive ? num : top;
                    }
                    else {
                        if (top && top == bottom) {
                            addDirections['y'] = top;
                        }
                        else {
                            addDirections['t'] = top;
                            addDirections['b'] = bottom;
                        }
                        if (left && left == right) {
                            addDirections['x'] = left;
                        }
                        else {
                            addDirections['l'] = left;
                            addDirections['r'] = right;
                        }
                    }

                    for (var dir in removeDirections) {
                        var dirValue = removeDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForMarginAndPadding(type, dir, size, dirValue);
                            class_setter.removeClass(pgel, makeResponsiveClass(newClass));
                        }
                    }

                    for (var dir in addDirections) {
                        var dirValue = addDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForMarginAndPadding(type, dir, size, dirValue);
                            class_setter.addClass(pgel, makeResponsiveClass(newClass), options.classSorter);
                            setMPValueOnFields(type, dir, size, dirValue);
                        }
                    }

                    return value;
                }
            }

            return field;
        })
    }

    pgFrameworkHelpers.getDevicesPositionControlFactory = function (control_id, func) {
        var positionValues = ['0', 'auto'];

        var field_def, field, full_field_key;

        var forSize = function (f) {
            /* options.sizes */ ['all'].forEach(function (size) {
                f(size, size == options.size_for_all ? '' : size + ':', options.prefix + control_id + '.' + size);
            })
        }

        return function () {
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            //will keep helper functions and vars private
            var updateFieldsValues = function (size, size_token, field_key) {
                var pgel = c.obj;
                var size_base = (size_token ? size + ':' : '');

                var class_setter = pgel.getPage().classSetter;

                for (var i = 3; i < posDirections.length; i++) {
                    var pos = posDirections[i];
                    var pos_field_key = field_key + '.' + pos;

                    var value = c.values[pos_field_key];
                    if (!value) {
                        for (var j = 0; j < 3; j++) {
                            var dir = posDirections[j];

                            var insetClassPref = size_base + 'inset-' + (dir ? dir + '-' : '');
                            for (var k = 0; k < positionValues.length; k++) {
                                var posValue = positionValues[k];
                                if (class_setter.hasClass( pgel, makeResponsiveClass(insetClassPref + posValue))) {
                                    if (isAllowedToOverwritePos(pos, dir)) {
                                        var classValue = size_base + pos + '-' + posValue;

                                        var pgField = positionObjs[size][pos];
                                        pgField && pgField.setValue(classValue);

                                        c.values[pos_field_key] = classValue;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            var hasValue = function (field_key) {
                for (let j = 3; j < posDirections.length; j++) {
                    var direction = posDirections[j];

                    var full_field_key = field_key + '.' + direction;
                    if (c.values[full_field_key]) return true;
                }

                return false;
            }

            //Register subfields.
            c.onDefine = function () {
                forSize(function (size, size_token, field_key) {
                    for (var i = 3; i < posDirections.length; i++) {
                        var dir = posDirections[i];

                        field_def = func(size, size_token, dir);
                        full_field_key = field_key + '.' + dir;
                        c.registerInputField(full_field_key, field_def);
                    }
                });
            }

            //Show control. Return the control $el.
            c.onShow = function () {
                var $div = $('<div class="crsa-css-gui tw-position"></div>');

                forSize(function (size, size_token, field_key) {
                    positionObjs[size] = {};
                    var $container = $('<div class="position-container text-center" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="tw-device-size">' + size + '</span>').appendTo($container);

                    for (var i = 3; i < posDirections.length; i++) {
                        var dir = posDirections[i];

                        field_def = func(size, size_token, dir);
                        full_field_key = field_key + '.' + dir;
                        field = c.showInputField($container, full_field_key, field_def);
                        field.$field.addClass('crsa-field-' + dir + ' crsa-small-input crsa-align-center-input crsa-inline-container');
                        positionObjs[size][dir] = field;
                    }

                    var $positionOverallBorder = $('<div class="position-overall-border"></div>').appendTo($container);

                    var $positoinPaddingTop = $('<div class="position-padding-top"></div>').appendTo($positionOverallBorder);
                    var $positoinPaddingTop = $('<div class="position-padding-left"></div>').appendTo($positionOverallBorder);
                    var $positionBorder = $('<div class="position-border"></div>').appendTo($positionOverallBorder);
                    var $positionPaddingRight = $('<div class="position-padding-right"></div>').appendTo($positionOverallBorder);
                    var $positionPaddingRight = $('<div class="position-padding-bottom"></div>').appendTo($positionOverallBorder);

                    var $equalIcon = positionActiveBtn[size] = $('<div class="change-all-icon"><i class="icon-same"></i></div>').appendTo($positionBorder);
                    $equalIcon.data('frKey', '{direction}');
                    $equalIcon.data('type', '');
                    $equalIcon.data('directions', posDirections.slice(3, posDirections.length));
                    $equalIcon.data('valueObject', positionObjs);
                    $equalIcon.on('click', toggleIconActivation);
                })

                return $div;
            }

            //Called when control is recycled and new values are set
            c.onSetValues = function (values) {
                var none_or_all = true;

                var pgel = this.obj;
                var class_setter = pgel.getPage().classSetter;


                forSize(function (size, size_token, field_key) {
                    // Check position
                    updateFieldsValues(size, size_token, field_key);

                    var $equalIcon = positionActiveBtn[size];
                    $equalIcon.removeClass('active');

                    for (var i = 0; i < posPoints.length; i++) {
                        var className = getClassNameForPosition('', size, posPoints[i]);
                        if (class_setter.hasClass(pgel, makeResponsiveClass(className))) {
                            $equalIcon.addClass('active');
                        }
                    }

                    var $field = c.$element.find('> [data-device-size="' + size + '"]');
                    if (hasValue(field_key)) {
                        $field.addClass('tw-has-value');
                        if (size != options.size_for_all) {
                            none_or_all = false;
                        }
                    } else {
                        $field.removeClass('tw-has-value');
                    }
                })
                if (none_or_all) {
                    c.$element.addClass('tw-none-or-all');
                } else {
                    c.$element.removeClass('tw-none-or-all');
                }
            }

            return c;
        }
    }

    pgFrameworkHelpers.getPositionControl = function (parent_section_id) {
        return pgFrameworkHelpers.getDevicesPositionControlFactory('position.dir', function (size, size_token, direction_token) {
            var getAllOptions = function () {
                var base = (size_token ? `${size}:` : '') + direction_token + '-';
                return getOptionsPosition(base, direction_token);
            }

            var field = {
                type : 'select',
                name : null,
                // placeholder : direction_token[0],
                // show_empty : true,
                options : getAllOptions,
                action : 'custom',
                draggable_list : true,
                parent_section_id: parent_section_id,
                async_options_function: true,
                fill_list_on_open: true,
                more_prop: {
                    size: size,
                    direction: direction_token
                },
                preview_class: true,
                responsive_class: function(val, pgel) {
                    //dummy, just return
                    //a hack to get prefix for class preview
                    if(val === '___dummy___') {
                        return makeResponsiveClass(val, pgel);
                    }
                    return val;
                },
                on_preview_class: function($el, fn, value, values, pgel) {
                    $el.getPage = function() {
                        return {
                            classSetter: pgDefaultClassSetter
                        }
                    }
                    var old_value = field.get_value($el, fn, values, field);
                    field.set_value($el, value, values, old_value, 'input', field);
                },
                get_value: function(pgel, fn, values, fv) {
                    var class_setter = pgel.getPage().classSetter;

                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (class_setter.hasClass(pgel, makeResponsiveClass( optionsValue[i].key ))) return optionsValue[i].key;
                    }

                    var size = fv.more_prop.size,
                        direction = fv.more_prop.direction;

                    for (var i = 0; i < 3; i++) {
                        var pos = posDirections[i];

                        if (isAllowedToOverwritePos(direction, pos)) {
                            for (var j = 0; j < posPoints.length; j++) {
                                var point = posPoints[j];
                                var className = getClassNameForPosition(pos, size, point);
                                if (class_setter.hasClass(pgel, makeResponsiveClass(className))) {
                                    return getClassNameForPosition(direction, size, point);
                                }
                            }
                        }
                    }

                    return '';
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var size = fdef.more_prop.size,
                        direction = fdef.more_prop.direction;

                    var class_setter = pgel.getPage().classSetter;

                    var num = getNumFromPositionValue(direction, value);

                    var $icon = positionActiveBtn[size];
                    var isActive = $icon.hasClass('active');

                    var valuesNum = [];
                    for (var i = 0; i < posDirections.length; i++) {
                        var dir = posDirections[i];

                        for (var j = 0; j < posPoints.length; j++) {
                            var point = posPoints[j];
                            var className = makeResponsiveClass(getClassNameForPosition(dir, size, point));
                            if (class_setter.hasClass(pgel, className)) {
                                valuesNum[dir] = point;

                                var dirs = getPosDirectionsFor(dir);
                                for (var k = 0; k < dirs.length; k++) {
                                    valuesNum[dirs[k]] = point;
                                }
                            }
                        }
                    }

                    var top    = direction == 'top'    ? num  : valuesNum['top'];
                    var left   = direction == 'left'   ? num  : valuesNum['left'];
                    var right  = direction == 'right'  ? num  : valuesNum['right'];
                    var bottom = direction == 'bottom' ? num  : valuesNum['bottom'];

                    var removeDirections = {
                        '': valuesNum[''],
                        'x': valuesNum['x'],
                        'y': valuesNum['y'],
                        'top': valuesNum['top'],
                        'left': valuesNum['left'],
                        'right': valuesNum['right'],
                        'bottom': valuesNum['bottom'],
                    };
                    var addDirections = {};
                    if (isActive || (top && top == left && left == right && right == bottom)) {
                        addDirections[''] = isActive ? num : top;
                    }
                    else {
                        if (top && top == bottom) {
                            addDirections['y'] = top;
                        }
                        else {
                            addDirections['top'] = top;
                            addDirections['bottom'] = bottom;
                        }
                        if (left && left == right) {
                            addDirections['x'] = left;
                        }
                        else {
                            addDirections['left'] = left;
                            addDirections['right'] = right;
                        }
                    }

                    for (var dir in removeDirections) {
                        var dirValue = removeDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForPosition(dir, size, dirValue);
                            class_setter.removeClass(pgel, makeResponsiveClass(newClass));
                        }
                    }

                    for (var dir in addDirections) {
                        var dirValue = addDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForPosition(dir, size, dirValue);
                            class_setter.addClass(pgel, makeResponsiveClass(newClass), options.classSorter);
                            setPosValueOnFields(dir, size, dirValue);
                        }
                    }


                    return value;
                }
            }

            return field;
        });
    }

    pgFrameworkHelpers.getDevicesTransformControlFactory = function (control_id, clsPrefix, func) {
        var forSize = function (f) {
            /* options.sizes */['all'].forEach(function (size) {
                f(size, size == options.size_for_all ? '' : size + ':', options.prefix + control_id);
            })
        }

        return function () {
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            c.onDefine = function () {
                forSize(function (size, size_token, field_key) {
                    for (var type in transformTypes) {
                        var directions = transformTypes[type].directions;

                        for (var i = 0; i < directions.length; i++) {
                            var dir = directions[i];
                            var field_def = func(size, size_token, type, dir);
                            var full_field_key = `${field_key}.${type}${(dir ? '.' + dir : '')}.${size}`;
                            c.registerInputField(full_field_key, field_def);
                        }
                    }
                });
            }

            c.onShow = function () {
                var $div = $('<div class="crsa-css-gui tw-transform"></div>');

                forSize(function (size, size_token, field_key) {
                    var $container = $('<div class="transform-container" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="tw-device-size">' + size + '</span>').appendTo($container);


                    var $tabsCont = $('<div class="transform-tabs-container"></div>').appendTo($container);

                    transformObjs[size] = {};
                    for (var type in transformTypes) {
                        transformObjs[size][type] = {};
                        var directions = transformTypes[type].directions;

                        var $typeCont = $('<div class="' + type + '-container tw-device-control"></div>').appendTo($tabsCont);
                        // if (type != "translate") pg$Hide($typeCont);

                        for (var i = 0; i < directions.length; i++) {
                            var dir = directions[i];
                            var field_def = func(size, size_token, type, dir);
                            var full_field_key = `${field_key}.${type}${(dir ? '.' + dir : '')}.${size}`;
                            transformObjs[size][type][dir] = c.showInputField($typeCont, full_field_key, field_def);
                        }
                    }
                })

                return $div;
            }

            c.onSetValues = function (values) {

            }

            return c;
        }
    }

    pgFrameworkHelpers.getTransformControl = function (prefix, parent_section_id) {
        var prefix = prefix || ""
        var clsPrefix = prefix ? prefix + ":" : "";

        return pgFrameworkHelpers.getDevicesTransformControlFactory(prefix + 'transform', clsPrefix, function (size, size_token, type, direction_token) {
            var getAllOptions = function () {
                // var size_t = (size_token ? `${size}:` : '') + clsPrefix;
                var base = type + '-' + (direction_token ? direction_token + '-' : '');
                return getOptionsTransform(base, type, direction_token);
            }

            // var dirName = direction_token ? direction_token[0].toUpperCase() : crsaCamelize(type);
            var name = type == 'translate' ? 'tran.' : type;
            name = crsaCamelize(name) + ' ' + (direction_token ? direction_token[0].toUpperCase() : '');

            var field = {
                type : 'select',
                name: name,
                // icon: 'icon-Transform_' + dirName,
                // placeholder : direction_token[0],
                // show_empty : true,
                options : getAllOptions,
                action : 'custom',
                draggable_list : true,
                parent_section_id: parent_section_id,
                async_options_function: true,
                fill_list_on_open: true,
                more_prop: {
                    size: size,
                    type: type,
                    direction: direction_token,
                    prefix: prefix,
                    clsPrefix: clsPrefix,
                },
                get_value: function(pgel, fn, values, fv) {
                    var class_setter = pgel.getPage().classSetter;

                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (class_setter.hasClass(pgel, makeResponsiveClass(optionsValue[i].key))) return optionsValue[i].key;
                    }

                    var size = fv.more_prop.size;
                    var type = fv.more_prop.type;
                    var dir = fv.more_prop.direction;
                    var prefix = fv.more_prop.prefix;

                    if (type == "scale") {
                        var points = transformTypes[type].points;
                        for (var i = 0; i < points.length; i++) {
                            var point = points[i];
                            var size_t = (size == options.size_for_all ? '' : size + ':');
                            var className = size_t + clsPrefix + type + '-' + point;
                            if (class_setter.hasClass(pgel, makeResponsiveClass(className))) {
                                var newClass = size_t + clsPrefix + type + '-' + dir + '-' + point;
                                var full_field_key = `${options.prefix}${prefix}transform.${type}${(dir ? '.' + dir : '')}.${size}`;
                                values[full_field_key] = newClass;
                                return newClass;
                            }
                        }
                    }

                    return '';
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var class_setter = pgel.getPage().classSetter;

                    class_setter.removeClass(pgel, makeResponsiveClass(oldValue));

                    var size = fdef.more_prop.size;
                    var type = fdef.more_prop.type;
                    var direction = fdef.more_prop.direction;

                    if (type == "scale") {
                        var points = transformTypes[type].points;
                        var directions = transformTypes[type].directions;
                        var num = getNumFromScaleValue(`${type}-${direction}`, value);

                        var valuesNum = [];
                        var allDirections = ['', ...directions];
                        for (var i = 0; i < allDirections.length; i++) {
                            var dir = allDirections[i];

                            for (var j = 0; j < points.length; j++) {
                                var point = points[j];
                                var className = makeResponsiveClass(getClassNameForScale(type, dir, size, point, clsPrefix));
                                if (class_setter.hasClass(pgel, className)) {
                                    valuesNum[dir] = point;

                                    var dirs = getScaleDirectionsFor(dir);
                                    for (var k = 0; k < dirs.length; k++) {
                                        valuesNum[dirs[k]] = point;
                                    }
                                }
                            }
                        }

                        var x = direction == 'x' ? num  : valuesNum['x'];
                        var y = direction == 'y' ? num  : valuesNum['y'];

                        x = x == undefined ? null : x;
                        y = y == undefined ? null : y;

                        var removeDirections = {
                            '': valuesNum[''],
                            'x': valuesNum['x'],
                            'y': valuesNum['y'],
                        };
                        var addDirections = {};
                        if (x && x == y) {
                            addDirections[''] = x;
                        }
                        else {
                            addDirections['x'] = x;
                            addDirections['y'] = y;
                        }

                        for (var dir in removeDirections) {
                            var dirValue = removeDirections[dir];

                            if (dirValue != null) {
                                var newClass = getClassNameForScale(type, dir, size, dirValue, clsPrefix);
                                class_setter.removeClass(pgel, makeResponsiveClass(newClass));
                            }
                        }

                        for (var dir in addDirections) {
                            var dirValue = addDirections[dir];

                            if (dirValue != null) {
                                var newClass = getClassNameForScale(type, dir, size, dirValue, clsPrefix);
                                class_setter.addClass(pgel, makeResponsiveClass(newClass), options.classSorter);
                            }
                        }
                    }
                    else {
                        class_setter.addClass(pgel, makeResponsiveClass(value), options.classSorter);
                    }
                    return value;
                }
            }

            return field;
        });
    }

    pgFrameworkHelpers.getDevicesBorderControlFactory = function (control_id, func) {
        var forSize = function (f) {
            /* options.sizes */['all'].forEach(function (size) {
                f(size, size == options.size_for_all ? '' : size + ':', options.prefix + control_id);
            })
        }

        return function () {
            var c = new PgCustomPropertyControl(options.prefix + control_id);

            c.onDefine = function () {
                forSize(function (size, size_token, field_key) {
                    for (var type in borderTypes) {
                        var directions = borderTypes[type].directions;

                        for (var i = 0; i < directions.length; i++) {
                            var dir = directions[i];
                            var field_def = func(size, size_token, type, dir);
                            var full_field_key = `${field_key}.${type}${(dir ? '.' + dir : '')}.${size}`;
                            c.registerInputField(full_field_key, field_def);
                        }
                    }
                });
            }

            c.onShow = function () {
                var $div = $('<div class="crsa-css-gui tw-border"></div>');

                forSize(function (size, size_token, field_key) {
                    borderObjs[size] = {};
                    borderBtn[size] = {};
                    var $container = $('<div class="border-radius-container text-center" data-device-size="' + size + '"></div>').appendTo($div);
                    var $labelSize = $('<label/>', { 'class': "size-label" }).html('<span class="tw-device-size">' + size + '</span>').appendTo($container);

                    var directions = null;
                    var $controlContainer = $('<div class="control-container"></div>').appendTo($container);

                    var type = 'rounded';
                    borderObjs[size][type] = {};
                    directions = borderTypes[type].directions;

                    for (var i = 5; i < directions.length; i++) {
                        var dir = directions[i];
                        var field_def = func(size, size_token, type, dir);
                        var full_field_key = `${field_key}.${type}${(dir ? '.' + dir : '')}.${size}`;
                        var field = borderObjs[size][type][dir] = c.showInputField($controlContainer, full_field_key, field_def);

                        var directionName = "";
                        var iconClass = "";
                        if (dir == 'tl') {
                            directionName = "top-left";
                            iconClass = 'icon-round_lu';
                        }
                        else if (dir == 'tr') {
                            directionName = "top-right";
                            iconClass = "icon-round_ru";
                        }
                        else if (dir == 'bl') {
                            directionName = "bottom-left";
                            iconClass = "icon-round-ld";
                        }
                        else if (dir == 'br') {
                            directionName = "bottom-right";
                            iconClass = "icon-round_rd";
                        }

                        field.$field.addClass('crsa-radius-field crsa-small-input crsa-align-left-input crsa-inline-container has-icon-before crsa-field-border-' + directionName + '-radius');
                        $('<i class="' + iconClass + '"></i>').insertBefore(field.$field.find('input'));
                    }

                    $('<div class="border-radius-border"></div>').appendTo($controlContainer);

                    type = 'border';
                    borderObjs[size][type] = {};
                    directions = borderTypes[type].directions;

                    for (var i = 1; i < directions.length; i++) {
                        var dir = directions[i];
                        var field_def = func(size, size_token, type, dir);
                        var full_field_key = `${field_key}.${type}${(dir ? '.' + dir : '')}.${size}`;
                        var field = borderObjs[size][type][dir] = c.showInputField($controlContainer, full_field_key, field_def);

                        var directionName = "";
                        if (dir == 't') directionName = "top";
                        else if (dir == 'l') directionName = "left";
                        else if (dir == 'r') directionName = "right";
                        else if (dir == 'b') directionName = "bottom";
                        field.$field.addClass('crsa-border-field crsa-small-input crsa-align-left-input crsa-field-border-' + directionName);
                    }

                    var $changeAllCont = $('<div class="change-all-container"></div>').appendTo($container);

                    type = 'border';
                    directions = borderTypes[type].directions;

                    var $widthIcon = borderBtn[size][type] = $('<div data-type="width" class="change-all-icon"><i class="icon-same"></i></div>').appendTo($changeAllCont);
                    $widthIcon.data('frKey', '{type}');
                    $widthIcon.data('sdKey', '{direction}');
                    $widthIcon.data('type', type);
                    $widthIcon.data('directions', directions.slice(1));
                    $widthIcon.data('valueObject', borderObjs);
                    $widthIcon.on('click', toggleIconActivation);

                    type = 'rounded';
                    directions = borderTypes[type].directions;

                    var $roundedIcon = borderBtn[size][type] = $('<div class="change-all-border-raduis"><i class="icon-round_lu"></i><i class="icon-same"></i></div>').appendTo($changeAllCont);
                    $roundedIcon.data('frKey', '{type}');
                    $roundedIcon.data('sdKey', '{direction}');
                    $roundedIcon.data('type', type);
                    $roundedIcon.data('directions', directions.slice(5));
                    $roundedIcon.data('valueObject', borderObjs);
                    $roundedIcon.on('click', toggleIconActivation);
                });

                return $div;
            }

            c.onSetValues = function (values) {
                var pgel = this.obj;
                var class_setter = pgel.getPage().classSetter;

                forSize(function (size, size_token, field_key) {
                    for (const type in borderTypes) {
                        var $icon = borderBtn[size][type];
                        $icon.removeClass('active');

                        var points = borderTypes[type].points;
                        for (var i = 0; i < points.length; i++) {
                            var className = getClassNameForBorder(type, '', size, points[i]);
                            if (class_setter.hasClass(pgel, makeResponsiveClass(className))) {
                                $icon.addClass('active');
                            }
                        }
                    }
                });
            }

            return c;
        }

    }

    pgFrameworkHelpers.getBorderControl = function (parent_section_id) {
        return pgFrameworkHelpers.getDevicesBorderControlFactory('border', function (size, size_token, type, direction_token) {

            var getAllOptions = function () {
                var size_t = (size_token ? `${size}:` : '');
                var base = type + (direction_token ? '-' + direction_token : '');
                var propDir = getBorderDirectionProp(type, direction_token);
                return getOptionsBorder(base, type, propDir);
            }

            var field = {
                type : 'select',
                name: "",
              //  placeholder : direction_token[0],
                show_empty : true,
                options : getAllOptions,
                action : 'custom',
                can_make_change_action: 'apply_class',
                draggable_list : true,
                async_options_function: true,
                fill_list_on_open: true,
                parent_section_id: parent_section_id,
                more_prop: {
                    size: size,
                    type: type,
                    direction: direction_token
                },
                responsive_class: function(val, pgel) {
                    //dummy, just return
                    //a hack to get prefix for class preview
                    if(val === '___dummy___') {
                        return makeResponsiveClass(val, pgel);
                    }
                    return val;
                },
                get_value: function(pgel, fn, values, fv) {
                    var class_setter = pgel.getPage().classSetter;

                    var optionsValue = fv.options();
                    for (var i = 0; i < optionsValue.length; i++) {
                        if (class_setter.hasClass(pgel, makeResponsiveClass( optionsValue[i].key ))) return optionsValue[i].key;
                    }

                    var type = fv.more_prop.type;
                    var size = fv.more_prop.size;
                    var dir = fv.more_prop.direction;

                    var size_t = (size_token ? `${size}:` : '');
                    var dirBase = (dir ? '-' + dir : '');

                    var field_key = `${options.prefix}border.${type}${(dir ? '.' + dir : '')}.${size}`;

                    var points = borderTypes[type].points;
                    var directions = borderTypes[type].directions;

                    var start = 0;
                    var end = directions.length - 1;

                    if (type == 'border') end = 1;
                    else if (type == 'rounded') end = 5;

                    for (var i = start; i < end; i++) {
                        var currDir = directions[i];
                        var currDirBase = (currDir ? '-' + currDir : '');

                        for (var j = 0; j < points.length; j++) {
                            var point = points[j];
                            var pointBase = (point ? '-' + point : '');
                            var mainBase = size_t + type;
                            var className = mainBase + currDirBase + pointBase;
                            if (class_setter.hasClass(pgel, makeResponsiveClass( className )) && isAllowedToOverwriteBor(dir, currDir)) {
                                var usedClassName = mainBase + dirBase + pointBase;
                                values[field_key] = usedClassName;
                                return usedClassName;
                            }
                        }
                    }

                    return '';
                },
                set_value: function(pgel, value, values, oldValue, eventType, fdef) {
                    var class_setter = pgel.getPage().classSetter;

                    class_setter.removeClass(pgel, makeResponsiveClass( oldValue ));

                    var type = fdef.more_prop.type;
                    var size = fdef.more_prop.size;
                    var direction = fdef.more_prop.direction;

                    var points = borderTypes[type].points;
                    var directions = borderTypes[type].directions;

                    var num = getNumFromBorderValue(`${type}-${direction}`, value, points);

                    var $icon = borderBtn[size][type];
                    var isActive = $icon.hasClass('active');

                    var valuesNum = [];
                    for (var i = 0; i < directions.length; i++) {
                        var dir = directions[i];

                        for (var j = 0; j < points.length; j++) {
                            var point = points[j];

                            var className = makeResponsiveClass(getClassNameForBorder(type, dir, size, point));
                            if (class_setter.hasClass(pgel, className)) {
                                valuesNum[dir] = point;

                                var dirs = getBorderDirectionsFor(type, dir);
                                for (var k = 0; k < dirs.length; k++) {
                                    valuesNum[dirs[k]] = point;
                                }
                            }
                        }
                    }

                    var removeDirections = {};
                    var addDirections = {};
                    if (type == 'border') {
                        var top    = direction == 't' ? num : valuesNum['t'];
                        var left   = direction == 'l' ? num : valuesNum['l'];
                        var right  = direction == 'r' ? num : valuesNum['r'];
                        var bottom = direction == 'b' ? num : valuesNum['b'];

                        removeDirections = {
                            '': valuesNum[''],
                            't': valuesNum['t'],
                            'l': valuesNum['l'],
                            'r': valuesNum['r'],
                            'b': valuesNum['b'],
                        };
                        if (isActive || ((top || top == "") && top == left && left == right && right == bottom)) {
                            addDirections[''] = isActive ? num : top;
                        }
                        else {
                            addDirections['t'] = top;
                            addDirections['l'] = left;
                            addDirections['r'] = right;
                            addDirections['b'] = bottom;
                        }
                    }
                    else if (type == 'rounded') {
                        var topleft     = direction == 'tl' ? num : valuesNum['tl'];
                        var topright    = direction == 'tr' ? num : valuesNum['tr'];
                        var bottomleft  = direction == 'bl' ? num : valuesNum['bl'];
                        var bottomright = direction == 'br' ? num : valuesNum['br'];

                        var directionSet = {};

                        var canAddDirection = function (direction) {
                            if (directionSet[direction]) return false;

                            return true;
                        }

                        var directionAdded = function (direction) {
                            delete addDirections[direction];
                            directionSet[direction] = true;
                        }

                        removeDirections = {
                            '': valuesNum[''],
                            't': valuesNum['t'],
                            'l': valuesNum['l'],
                            'r': valuesNum['r'],
                            'b': valuesNum['b'],
                            'tl': valuesNum['tl'],
                            'tr': valuesNum['tr'],
                            'bl': valuesNum['bl'],
                            'br': valuesNum['br'],
                        };

                        if (isActive || (topleft && topleft == topright && topright == bottomleft && bottomleft == bottomright)) {
                            addDirections[''] = isActive ? num : topleft;
                        }
                        else {
                            // Top
                            if (topleft && topleft == topright) {
                                addDirections['t'] = topleft;
                                directionAdded('tl');
                                directionAdded('tr');
                            }
                            else {
                                if (canAddDirection('tl')) addDirections['tl'] = topleft;
                                if (canAddDirection('tr')) addDirections['tr'] = topright;
                            }

                            // Left
                            if (topleft && topleft == bottomleft) {
                                addDirections['l'] = topleft;
                                directionAdded('tl');
                                directionAdded('bl');
                            }
                            else {
                                if (canAddDirection('tl')) addDirections['tl'] = topleft;
                                if (canAddDirection('bl')) addDirections['bl'] = bottomleft;
                            }

                            // right
                            if (topright && topright == bottomright) {
                                addDirections['r'] = topright;
                                directionAdded('tr');
                                directionAdded('br');
                            }
                            else {
                                if (canAddDirection('tr')) addDirections['tr'] = topright;
                                if (canAddDirection('br')) addDirections['br'] = bottomright;
                            }

                            // bottom
                            if (bottomleft && bottomleft == bottomright) {
                                addDirections['b'] = bottomleft;
                                directionAdded('bl');
                                directionAdded('br');
                            }
                            else {
                                if (canAddDirection('bl')) addDirections['bl'] = bottomleft;
                                if (canAddDirection('br')) addDirections['br'] = bottomright;
                            }
                        }
                    }

                    for (var dir in removeDirections) {
                        var dirValue = removeDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForBorder(type, dir, size, dirValue);
                            class_setter.removeClass(pgel, makeResponsiveClass(newClass));
                        }
                    }

                    for (var dir in addDirections) {
                        var dirValue = addDirections[dir];

                        if (dirValue !== undefined) {
                            var newClass = getClassNameForBorder(type, dir, size, dirValue);
                            class_setter.addClass(pgel, makeResponsiveClass(newClass), options.classSorter);
                            setBorderValueOnFields(type, dir, size, dirValue);
                        }
                    }

                    return value;
                }
            }

            return field;
        });
    }

    pgFrameworkHelpers.init();

    return pgFrameworkHelpers;
}


var TwCustomizer = function(rc, sizes_icons, getClassInfo, getColorsForColorPicker, getTailwindColor, settings) {

    // init options
    settings = settings || {};
    var onTailwindUpdated = typeof settings['onTailwindUpdated'] == 'function' ? settings['onTailwindUpdated'] : function() {};


    var _this = this;

    var has_data = false;
    var using_defaults = true;

    this.has_data = function() {
        return has_data;
    }

    var map = {}
    var list = []

    var colors = {}

    var color_rows = {}
    var color_options = {}
    var options = {};
    var screens = []
    var pseudo_states = []
    var screen_sizes = {}
    var screen_defs = [];
    var filter_func_called = {}

    this.getTheme = function() {

        if(settings.onGetTWVersion() >= 3) {
            pinegrow.showAlert(`<p>Customizing Tailwind controls works by inspecting the full generated Tailwind CSS stylesheet. This does not work on Tailwind CSS 3 and higher, because there only the minimal stylesheet with used classes is generated.</p><p>That said, you can still work with your custom Tailwind CSS theme using the <a href="https://pinegrow.com/docs/tailwind/tailwind-css-design-panel/" class="external">Design panel for Tailwind CSS</a> to visually customize colors, fonts and other properties.</p>
`, 'Customization is not supported on Tailwind CSS 3+');
            return;
        }

        if(!pinegrow.getCurrentProject()) {
            pinegrow.showQuickError('Please open the page as project to use customized Tailwind visual controls.');
            return;
        }

        var page = pinegrow.getSelectedPage();
        var sheet = page.hasStylesheet(/tailwind/);

        if(sheet && false) {
            _this.update(page, sheet);
        } else {
            var fields = {};

            var sheets = page.getAllDOMStylesheets();

            if(sheets.length === 0) {
                pinegrow.showQuickError('The page has no stylesheets to inspect!');
            }

            var pi = pinegrow.getCurrentProjectInfo();
            var map = {}

            var p = pinegrow.getCurrentProject();

            var def_value = pi.getSetting('tw-custom-stylesheet');
            if(def_value) {
                def_value = pi.makeUrlAbsolute(def_value);
            }

            var def_in_list = false;

            var list = sheets.map(function(s) {
                var url = pinegrow.getOriginalUrl(s.href);

                map[url] = s;

                if(!def_value && s.href.match(/tailwind/)) {
                    def_value = url;
                }

                if(def_value && url === def_value) {
                    def_in_list = true;
                }

                var key = url;

                if(p) {
                    url = p.getRelativeUrl(url);
                }

                return {
                    key: key,
                    name: crsaGetSummaryStr(url, 40)
                }
            })

            if(!def_in_list) def_value = null;

            fields.url = {
                name: 'Stylesheet',
                type: 'select',
                show_empty: true,
                options: list,
                validate: function(obj, prop, value) {
                    return value ? null : 'Please select the stylesheet with Tailwind classes.'
                },
                value: def_value
            }

            pinegrow.showQuickDialog(`Customize Tailwind visual controls`, 'Where to look for Tailwind classes?', `<p>Please select the stylesheet that contains Tailwind CSS rules so that Tailwind visual controls can be adjusted to your Tailwind theme.</p><p>This operation can take a couple of seconds to complete. Learn more about <a href="https://pinegrow.com/docs/tailwind/customized-themes/" class="external">TailwindCSS customization</a>.</p>`, fields, 'Cancel', 'Customize visual controls', null, function(values) {

                pi.setSetting('tw-custom-stylesheet', pi.makeUrlRelative(values.url));
                pi.save();

                var sheet = map[values.url];
                _this.update(page, sheet);
            })
        }
    }

    var getCustomizedClasses = function(page, sheet) {
        var p = new CrsaProfile();

        var s = '';

        has_data = false;

        map = {}
        list = []
        colors = {}
        color_rows = {}
        color_options = {}
        options = {}
        screens = ['all'];
        pseudo_states = [];
        screen_sizes = {
            all: {
                name: 'ALL',
                name_short: 'ALL',
                w: 0,
                size: 'all',
                slug: ''
            }
        }
        filter_func_called = {}

        var has_colors = 0;

        try {

            var re = /\.(\\3[0-9]+\s[a-z0-9\\\:\-\/]+|[a-z0-9\\\:\-\/]+)/gi;
            var pre = /([a-z0-9])(:[a-z:\-]+)/;
            var sre = /[\.\\]/g;
            var getcre = /^text\-([a-z]+)\-([0-9]+)$/;
            var not_col = ['opacity'];
            var def_colors = ['white', 'black', 'current', 'transparent'];

            var doRules = function (rules) {

                if (rules) {
                    for (var i = 0; i < rules.length; i++) {
                        var r = rules[i];

                        if (r.type === CSSRule.STYLE_RULE) {
                            var sel = r.selectorText;
                            //console.log(sel)
                            sel = sel.replace('\\.', '--dot--');

                            var m = sel.match(re);
                            if (m) {
                                for (var j = 0; j < m.length; j++) {
                                    if (m[j].indexOf('.pg-') < 0) {
                                        //console.log(m[j]);
                                        var cls = m[j].replace(/\\3([0-9]+)\s?/, '$1');
                                        cls = cls.replace(pre, '$1');

                                        cls = cls.replace(sre, '');
                                        cls = cls.replace('--dot--', '.');

                                        cls = cls.replace(/^dark\:/, '');
                                        cls = cls.replace(/\:dark\:/, ':');

                                        if (!(cls in map)) {
                                            map[cls] = true;

                                            var props = r.cssText;
                                            var bidx = props.indexOf('{');
                                            if (bidx >= 0) {
                                                props = props.substr(bidx);
                                            }
                                            list.push({
                                                class: cls,
                                                rule: r,
                                                props: props
                                            });

                                            var b = cls.split(':');
                                            if (b.length === 3) {
                                                //console.log(b[0]);
                                                //mq - pseudo - cls combo
                                                if (screens.indexOf(b[0]) < 0 && cls.indexOf('dark:') < 0) {
                                                    screens.push(b[0]);

                                                    screen_sizes[b[0]] = {
                                                        name: b[0].toUpperCase(),
                                                        name_short: b[0].toUpperCase(),
                                                        w: 600, //default,
                                                        size: b[0],
                                                        slug: '-' + b[0]
                                                    };

                                                    if (r.parentRule && r.parentRule.type === CSSRule.MEDIA_RULE) {
                                                        var mqm = r.parentRule.media.mediaText.match(/min\-width\:\s?([0-9]+)px/);
                                                        if (mqm) {
                                                            var w = parseInt(mqm[1]);
                                                            screen_sizes[b[0]].w = w;
                                                            screen_sizes[b[0]].name += `<span style="opacity:0.7;"> / ${w}px and up</span>`
                                                        }
                                                    }
                                                }
                                                if (pseudo_states.indexOf(b[1]) < 0) {
                                                    pseudo_states.push(b[1]);
                                                }
                                            }
                                        }

                                    }
                                }

                            }
                        }
                        if (r.cssRules) {
                            doRules(r.cssRules);
                        }
                    }
                }
            }
            if (sheet) {
                doRules(sheet.cssRules);
            }

            screen_defs = [];

            for (var i = 0; i < screens.length; i++) {
                if (i < screens.length - 1) {
                    screen_sizes[screens[i]].w_range = [screen_sizes[screens[i]].w, screen_sizes[screens[i + 1]].w - 1];
                } else {
                    screen_sizes[screens[i]].w_range = [screen_sizes[screens[i]].w, 9999];
                }
                if (sizes_icons.length > i) {
                    screen_sizes[screens[i]].icon = sizes_icons[i];
                } else {
                    screen_sizes[screens[i]].icon = 'icon-Lg';
                }
                screen_defs.push(screen_sizes[screens[i]]);
            }

            list.forEach(function (d) {
                var cls = d.class;
                var r = d.rule;
                var ci = getClassInfo(cls, screens);
                s += ci.class + ' ';

                if (ci.size === 'all') {
                    var cm = ci.class.match(getcre);

                    if (cls.startsWith('text-')) {
                        var c = r.style.getPropertyValue('color');

                        if (c) {
                            var col = null;
                            var shade = 'noshade';

                            var a = ci.class.split('-');
                            a.shift();

                            if (a.length > 1) {
                                shade = a.pop();
                            }
                            col = a.join('-');

                            if (def_colors.indexOf(col) >= 0 && shade === 'noshade') {

                            } else if (not_col.indexOf(col) < 0) {

                                c = c.replace('var(--text-opacity)', '1').replace('var(--tw-text-opacity)', '1');

                                if (!colors[col]) {
                                    colors[col] = {}
                                    has_colors++;
                                }
                                colors[col][shade] = c;
                                //console.log(`${ci.class} = ${c}`);
                            }
                        }
                    }
                }
                delete d.rule;
            })

            if(list.length > 20 && has_colors > 1) {
                has_data = true;
                using_defaults = false;
            }

            p.show('TW GET CLASSES');
        } catch(err) {

        }
        //console.log(s);
        //console.log(colors, screens, screen_sizes, pseudo_states);
    }

    this.tailwindIsBeingUpdating = function () {
        onTailwindUpdated();
    }

    this.save = function(save_defaults) {
        if(this.has_data() && (!using_defaults || save_defaults)) {
            var d = {
                //list: list,
                options: options,
                colors: colors,
                screens: screens,
                pseudo_states: pseudo_states,
                screen_sizes: screen_sizes,
                screen_defs: screen_defs
            }
            if(!save_defaults) {
                var pi = pinegrow.getCurrentProjectInfo();
                pi.saveFile('tailwind.theme.json', JSON.stringify(d));
            } else {
                var def_file = settings.framework.getResourceFile('resources/tailwind.theme.json');
                require('fs').writeFileSync(def_file, JSON.stringify(d), {encoding: 'utf8'});
            }
        }
    }

    this.load = function(only_defaults) {
        has_data = false;

        var json;

        if(!only_defaults) {
            var pi = pinegrow.getCurrentProjectInfo();
            json = pi.loadFile('tailwind.theme.json');
        }

        var doWithJson = function() {
            if(json) {
                try {
                    options = {}
                    list = []

                    var d = typeof json === 'string' ? JSON.parse(json) : json;
                    if(d.options) {
                        options = d.options;
                    } else {
                        list = d.list;
                    }
                    colors =  d.colors;
                    screens = d.screens;
                    pseudo_states = d.pseudo_states;
                    screen_sizes = d.screen_sizes;
                    screen_defs = d.screen_defs;
                    has_data = true;
                    filter_func_called = {}

                    _this.tailwindIsBeingUpdating();

                    rc.reregister();

                    if(!using_defaults) {
                        console.log('TailwindCSS theme info loaded from _pginfo.');
                        pinegrow.showQuickMessage('Using custom TailwindCSS controls.');
                    } else {
                        console.log('Default TailwindCSS theme info loaded.');
                    }

                    pinegrow.dispatchEvent('on_page_frameworks_changed', null);

                } catch(err) {

                }
            }
        }

        if(!json) {
            var def_file = settings.framework.getResourceFile('resources/tailwind.theme.json');
            try {
                if(pgIsLive()) {
                    settings.framework.twRemoteResources.getContent('tailwind.theme.json');
                } else {
                    json = require('fs').readFileSync(def_file, {encoding: 'utf8'});
                }
                using_defaults = true;
            } catch(err) {
                console.error('Unable to load default TW theme from ' + def_file);
            }
        } else {
            using_defaults = false;
        }

        if(!json) {
            if(isApp()) {
                var def_file = settings.framework.getResourceFile('resources/tailwind.theme.json');
                try {
                    json = require('fs').readFileSync(def_file, {encoding: 'utf8'});
                    using_defaults = true;
                } catch (err) {
                    console.error('Unable to load default TW theme from ' + def_file);
                }
            } else {
                var def_file = settings.framework.getResourceUrl('resources/tailwind.theme.json');

                $.getJSON( def_file, function( data ) {
                    json = data;
                    doWithJson();
                });

                return;
            }
        } else {
            using_defaults = false;
        }
        doWithJson();
    }

    this.useDefaults = function() {
        has_data = false;
        rc.reregister();
        pinegrow.dispatchEvent('on_page_frameworks_changed', null);
    }

    this.restoreDefaults = function() {
        this.useDefaults();
        var pi = pinegrow.getCurrentProjectInfo();
        pi.deleteFile('tailwind.theme.json');
        this.load(true);
        pinegrow.showQuickMessage('Visual controls now use default TailwindCSS values.');
    }

    this.update = function(page, sheet) {
        getCustomizedClasses(page, sheet);

        if(this.has_data()) {
            this.tailwindIsBeingUpdating();

            rc.reregister();
            this.save();

            pinegrow.dispatchEvent('on_page_frameworks_changed', null);
            pinegrow.refreshCurrentProjectDelayed(true);

            pinegrow.showQuickMessage('Tailwind visual controls updated!');
        } else {
            pinegrow.showQuickError('Could not get the information about the TailwindCSS theme from the selected stylesheet.')
        }
    }

    this.getScreens = function() {
        return screens;
    }

    this.getPseudoStates = function() {
        return pseudo_states;
    }

    this.getScreenSizeDefs = function() {
        return screen_defs;
    }

    this.getClasses = function(prefix, size, pseudo, is_color, props) {
        if(list.length === 0) {
            pgIsDev() && console.error('TW Customizer: getClasses called on empty list', prefix, size, pseudo, is_color, prop)
        }
        var rec = /\-[0-9]+$/;

        if(props) {
            if(!Array.isArray(props)) {
                props = [props];
            }
            for (var i = 0; i < props.length; i++) {
                if (typeof props[i] === 'string') {
                    props[i] += ':';
                }
            }
        }

        var sw = prefix;
        if(pseudo) {
            sw = pseudo + ':' + sw;
        }
        if(size) {
            sw = size + ':' + sw;
        }
        var r = [];
        for(var i = 0; i < list.length; i++) {
            var shouldCheck = false;
            var cls = list[i].class;
            if (sw) {
                if (cls.startsWith(sw)) {
                    shouldCheck = true;
                }
            }
            else {
                var ci = getClassInfo(cls, screens);
                if (ci.size == 'all' && ci.state == "") {
                    shouldCheck = true;
                }
            }

            if (shouldCheck) {
                if(is_color) {
                    if(!cls.match(rec)) {
                        continue;
                    }
                }
                if(props) {
                    var skip = false;
                    for(var pi = 0; pi < props.length; pi++) {
                        var prop = props[pi];
                        if(typeof prop === 'string') {
                            if(list[i].props.indexOf(prop) < 0) {
                                skip = true;
                                break;
                            }
                        } else {
                            if(!list[i].props.match(prop)) {
                                skip = true;
                                break;
                            }
                        }
                    }
                    if(skip) continue;
                }
                r.push(list[i].class);
            }
        }
        return r;
    }

    this.getColorsForColorPicker = function(base) {
        //var str = '';

        if(!color_rows[base]) {
            var r = [
                [
                    {
                        'key': base + 'current', 'name': 'current',
                        'color': getTailwindColor('current')
                    },
                    {
                        'key': base + 'transparent', 'name': 'transparent',
                        'color': getTailwindColor('transparent')
                    },
                    {
                        'key': base + 'black', 'name': 'black',
                        'color': getTailwindColor('black')
                    },
                    {
                        'key': base + 'white', 'name': 'white',
                        'color': getTailwindColor('white')
                    }
                ]
            ]

            $.each(colors, function (col, shades) {
                var row = [];
                $.each(shades, function (shade, colval) {
                    var cd = {
                        key: base + col + (shade === 'noshade' ? '' : ('-' + shade)),
                        name: (shade === 'noshade') ? col : col + '-' + shade,
                        color: colval
                    }
                    row.push(cd)
                    //str += `'${cd.key}': '${cd.color}',
//`;
                })
                r.push(row);
            })
            color_rows[base] = r;
        }

        //console.log('TW COLORS', str);
        return color_rows[base];
    }

    this.getColorOptions = function(base) {
        if(!color_options[base]) {
            var r = [];
            getColorsForColorPicker(base).forEach(function (row) {
                row.forEach(function (def) {
                    r.push(def)
                })
            })
            color_options[base] = r;
        }
        return color_options[base];
    }

    this.getOptions = function(base, prop, prop_re, add_minus, filter_func) {
        var prop_key = '';
        if(prop) {
            if(Array.isArray(prop)) {
                for(var i = 0; i < prop.length; i++) {
                    prop_key += '_' + prop[i];
                }
            } else {
                prop_key = prop;
            }
        }
        var key = base + prop_key;
        if(!options[key]) {
            var r = []

            if(add_minus) {
                this.getClasses('-' + base, null, null, null, prop_re || prop).reverse().forEach(function(cls) {
                    r.push({
                        key: cls,
                        name: cls.replace(base, '')
                    })
                })
            }

            this.getClasses(base, null, null, null, prop_re || prop).forEach(function(cls) {
                r.push({
                    key: cls,
                    name: cls.replace(base, '')
                })
            })

            if(filter_func) {
                filter_func(r);
            }

            options[key] = r;
        }
        if(filter_func) {
            if(!filter_func_called[key]) {
                filter_func(options[key]);
                filter_func_called[key] = true;
            }
        }
        return options[key];
    }
}
