/**
 * Created by Matjaz on 11/20/13.
 */

var PgApiCodeEditorConnector = function() {

  var getPageForUrl = function(url) {
      var page = pinegrow.getCrsaPageByUrl(url);
      if(page) return page;

      /*var project = pinegrow.getCurrentProject();
      if(project) {
          page = project.getBackgroundPageForUrl(url);
      }*/
      return page;
  }

  var syntax_error_timer = null;

  var abort_on_errors = pgf.kids ? false : true;

  this.codeChangedInEditor = function(url, code, editor, undoData) {
      var page = getPageForUrl(url);
      if(page) {
          //page.undoStack.add('External change');
          if(pgf.kids) {
              var parser = new pgParser();
              parser.assignIds = true;
              parser.parse(code);
              rootNode = parser.rootNode;

              var old_html = page.sourceNode.toStringWithIds(true);

              page.sourceNode.removeWithoutEvents();

              page.sourceNode = rootNode;

              //manually add event
              var pgEvent = new PgNodeEvent(page.sourceNode, 'setCode', null, {
                  oldValue: old_html
              }, page);
              var undoEvent = new PgUndoEvent(pgEvent);
              pinegrow.undoManager.getHistoryForPage(page).addEvent(undoEvent);

              pinegrow.changeManager.didChange(page.sourceNode, 'Code edit', {
                  source: editor,
                  undoCombineKey: 'code_edit',
                  undoSourceId: url,
                  undoData: undoData
              });

              page.refresh(null, true /* silent_multi_refresh */);
          } else {
              var status = page.applyChangesToSource(code, abort_on_errors /* abort on error */, false, false, 'editor');
              if (syntax_error_timer) {
                  clearTimeout(syntax_error_timer);
                  syntax_error_timer = null;
              }
              if (status.errors.length && abort_on_errors) {
                  // page.undoStack.remove(); //code was not applied
                  if (editor == 'external_editor') {
                      syntax_error_timer = setTimeout(function () {
                          pinegrow.showQuickMessage('Code modified in external editor has syntax errors.', 3000, true);
                          syntax_error_timer = null;
                      }, 3000);
                  } else {
                      editor.setErrors(status.errors);
                  }
              } else {
                  if (status.errors.length && !abort_on_errors) {
                      //code has errors but we applied it anyway
                  }
                  status.replace_selected.forEach(function (d) {
                      pinegrow.selectedElements.replace(d.old_pgel, d.new_pgel);
                  })
                  pinegrow.changeManager.didChange(page.sourceNode, 'Code edit', {
                      source: editor,
                      undoCombineKey: 'code_edit',
                      undoSourceId: url,
                      undoData: undoData
                  });

                  if (!status.updated && status.whole_changed) {
                      if (status.stylesheets_changed) {
                          page.refreshWithUpdateStylesheets();
                      } else {
                          page.refresh();
                      }
                  } else {
                      if (status.stylesheets_changed) {
                          page.updateStylesheetsList();
                      }
                  }
                  if (editor && editor.setErrors) {
                      editor.setErrors(null);
                  }
              }
          }
      } else {
          var cs = null;
          if (crsaIsFileUrl(url)) {
              url = pgDecodeUrl(url);
              var filePath = crsaMakeFileFromUrl(url);
              cs = pinegrow.getStylesheetByFileName(filePath);
          }
          else {
              cs = pinegrow.stylesheets.getByUrl(url, true);
          }

          if(cs) {
              if(!cs.sourceStylesheet) {
                  cs.setNeedFormatting(true);
                  cs.setSource(code, function () {
                      cs.loaded = true;
                      pinegrow.dispatchEvent('on_css_source_changed', null, {list: [cs], eventType: 'editor', editor: editor});
                      $('body').trigger('crsa-stylesheets-properties-changed');
                      pinegrow.selectedElements.repositionMenus();
                  });
              } else {
                  pinegrow.showQuickMessage('This is a compiled stylesheets. Please edit its SASS or LESS source file.')
              }
          } else {
              var project = pinegrow.getCurrentProject();
              if(project) {
                  var bck_page = project.getBackgroundPageForUrl(url);
                  if(bck_page) {
                      if(bck_page.onlyCode) {
                          bck_page.setCode(code);
                          bck_page.setPageChanged(true);
                      }
                  }
              }
              /*
              //perhaps page in project, not yet open in PG
              var project = pinegrow.getCurrentProject();
              if(project && project.isUrlInProject(url)) {
                  var file = crsaMakeFileFromUrl(url);
                  page = new CrsaPage().fromFile(file, !pinegrow.isFileEditable(file) / only code if not editable /);
                  project.addBackgroundPage(page);
                  page.applyChangesToSource(code);
              }
              */
          }
      }
  }

  this.elementSelectedInEditor = function(url, element_path) {
      if(pinegrow.isFileEditable(crsaMakeFileFromUrl(url))) {
          pinegrow.openOrShowPage(url, function(page) {
              var node = page.sourceNode.getNodeFromPath(element_path, true /* closest valid */);
              if(node) {
                  //if(!page.visible) pinegrow.pageTabs.showPage(page, true);
                  node.scrollTo();
                  pinegrow.selectElement(node);
              }
          }, true, true);
      }
  }

  this.fileSavedInEditor = function(url, code) {
      var page = getPageForUrl(url);
      if(page) {
          page.changed = false;
          page.refreshPageChangedStatus(true);
          pinegrow.showQuickMessage('<b>' + page.name + '</b> was saved in external editor.');
          pinegrow.dispatchEvent('on_page_loaded_from_external', page);
      } else {
          url = pgDecodeUrl(url);
          var cs = pinegrow.getStylesheetByFileName(crsaMakeFileFromUrl(url));
          if(cs) {
              cs.lastSavedAt = (new Date()).getTime();
              cs.fileSavedInEditor();
              pinegrow.getAllPages().forEach(function(page) {
                  page.refreshPageChangedStatus(true);
              })
              pinegrow.showQuickMessage('<b>' + cs.name + '</b> was saved in external editor.');
              pinegrow.dispatchEvent('on_css_loaded_from_external', pinegrow.stylesheets.getPagesForStylesheet(cs), cs);
          }
      }
  }

  this.refreshPage = function(url) {
      var page = getPageForUrl(url);
      if(page && page.openedInEditor) {
          page.refresh();
      }
  }
}

var PgApiCodeEditors = function(connector) {

    var editors = [];
    var active_editor = null;
    var last_id = 0;

    var _this = this;

    var ignore_select_after_refresh = false;
    var ignore_select_timer = null;

    var ignoreSelectsForAWhile = function() {
        ignore_select_after_refresh = true;
        if(ignore_select_timer) {
            clearTimeout(ignore_select_timer);
        }
        ignore_select_timer = setTimeout(function() {
            ignore_select_after_refresh = false;
            ignore_select_timer = null;
        }, 2000);
    }

    var findEditorsByUrl = function(url) {
        if (crsaIsFileUrl(url)) url = crsaRemoveUrlParameters(url);

        return editors.filter(function(editor) {
            return pgDecodeUrl(editor.url) == pgDecodeUrl(url);
        })
    }

    var findEditorById = function(id) {
        for(var i = 0; i < editors.length; i++) {
            if(editors[i].id == id) return editors[i];
        }
        return null;
    }

    this.isUrlOpen = function(url) {
        var is = false;
        findEditorsByUrl(url).forEach(function(editor) {
            is = editor;
        })
        return is;
    }

    this.showEditor = function(editor) {
        var sm = pgPageCodeEditorsStack;
        if(!sm.isEditorShown(editor.view)) {
            sm.showEditor(editor.view);
        }
    }

    var ignore_code_for = {};

    this.ignoreCodeForUrl = function(url) {
        if(url) {
            ignore_code_for[url] = true;
        }
    }

    this.getCodeForUrl = function(url) {
        if(ignore_code_for[url]) {
            delete ignore_code_for[url];
            return null;
        }

        var code = null;

        findEditorsByUrl(url).forEach(function(editor) {
            code = editor.getCode();
        })
        return code;
    }

    this.closeEditor = function(editor) {
        var idx = editors.indexOf(editor);
        if(idx >= 0) {
            editors.splice(idx, 1);
        }
    }

    this.open = function(url, code, done, not_html, win, let_monaco_handle_undo, view_reference) {
        if(ignore_code_for[url]) delete ignore_code_for[url];

        var existing_editors = findEditorsByUrl(url);
        var editor = new PgInternalCodeEditor(++last_id, url, existing_editors.length ? existing_editors[0] : null, not_html, win, let_monaco_handle_undo);

        editor.view = view_reference;
        editors.push(editor);

        var first_change = true;

        editor.onChange = function(editor, undoData) {
            var current_code = editor.getCode();

            if(first_change && current_code === code) {
                //this event was triggered by the editor.setCode statement below, so we can ignore it
                return;
            }
            first_change = false;
            ignoreSelectsForAWhile();
            connector.codeChangedInEditor(editor.url, current_code, editor, undoData);

            last_code_received_from_external[editor.url] = {code: current_code};

            //Sync editors
            syncChangedCode(editor.url, current_code, editor);

            if(editor.view) {
                editor.view.setModified(true);
            }
        }

        editor.onElementSelectedInEditor = function(editor, path) {
            connector.elementSelectedInEditor(editor.url, path);
        }

        editor.setCode(code);

        if(done) done(editor.id);

        return editor;
    }

    this.getFocusedEditor = function() {
        for(var i = 0; i < editors.length; i++) {
            if(editors[i].isFocused && editors[i].isFocused()) return editors[i];
        }
        return null;
    }


    var syncChangedCode = function(url, code, source_editor) {
        findEditorsByUrl(url).forEach(function(e) {
            if(e != source_editor) {
                e.setCode(code);
            }
        })
    }


    var last_code_received_from_external = {};

    this.isCodeForUrlSameInExternalEditor = function(url, code, done) {
        setTimeout(function() {
            if(url in last_code_received_from_external) {
                done(last_code_received_from_external[url].code === code);
            } else {
                done(false);
            }
        }, (numberOfConnectedClients > 0 && last_code_received_from_external[url] && last_code_received_from_external[url].external) ? 1000 : 0);
    }

    this.show = function(editor_id, done) {

    }

    this.setCode = function (url, code) {
        last_code_received_from_external[url] = {code: code};
    }

    this.codeChanged = function(url, code, source, done, info) {
        findEditorsByUrl(url).forEach(function(editor) {
            if(editor !== source) {
                if (typeof code == 'function') code = code();
                editor.setCode(code, info, true /* skip do change call */);
            }
        })
        if(editorServer && source !== 'external_editor') {
            if (typeof code == 'function') code = code();
            editorServer.emit('codeChanged', {url: url, code: code});
            last_code_received_from_external[url] = {code: code};
        }
        if(done) done();
    }

    this.cssCodeChanged = function(url, code, source, info, cs) {
        findEditorsByUrl(url).forEach(function(editor) {
            if(editor !== source) {
                if (typeof code == 'function') code = code();
                editor.setCode(code, info, true /* skip do change call */);
                if(editor.view && cs) {
                    editor.view.setModified(cs.changed);
                }
            }
        })
        if(editorServer && source !== 'external_editor') {
            if (typeof code == 'function') code = code();
            editorServer.emit('codeChanged', {url: url, code: code});
            last_code_received_from_external[url] = {code: code};
        }
    }

    this.onPageSaved = function(url) {
        if (editorServer) {
            editorServer.emit('codeSaved', {url: url});
        }
    }

    this.onCssSaved = function(url) {
        if (editorServer) {
            editorServer.emit('codeSaved', {url: url});
        }
    }

    var broadcastOpenFiles = function() {
        var list = [];
        var getAllImportedFiles = function (stylesheet) {
            stylesheet.getAllImportedFiles(function (importedFile) {
                if (importedFile.localFile) {
                    list.push(crsaMakeUrlFromFile(importedFile.getLocalFileName()));
                    return true;
                }
                return false;
            });
        }

        pinegrow.getAllPages().forEach(function(page) {
            if(page.url) list.push(page.url);
        })
        pinegrow.getStylesheets().forEach(function(cs) {
            if (cs.localFile) list.push(crsaMakeUrlFromFile(cs.getLocalFileName()));

            // add source file
            if (cs.sourceStylesheet) {
                var sourceStylesheet = cs.sourceStylesheet;
                if (sourceStylesheet.localFile) list.push(crsaMakeUrlFromFile(sourceStylesheet.getLocalFileName()));

                // imported files
                getAllImportedFiles(sourceStylesheet);
            }
            else {
                // css imported files
                getAllImportedFiles(cs);
            }
        })
        if(editorServer) {
            editorServer.emit('listOfOpenFiles', {list: list});
        }
    }

    var last_remote_selected_paths = {};

    this.elementWasSelectedInPreview = function(url, element_path, info) {
        findEditorsByUrl(url).forEach(function(editor) {
            editor.elementWasSelectedInPreview(element_path);
        })
        var source = (info && info.source) ? info.source : null;
        if(editorServer && source != 'external_editor') {

          if((last_remote_selected_paths[url] || null) != element_path) {
            editorServer.emit('elementSelectedInPreview', {url: url, path: element_path});
          }
          last_remote_selected_paths[url] = null;
        }
    }

    this.close = function(editor_id, done) {

    }


    var editorServer = null;
    var numberOfConnectedClients = 0;

    var onCodeChanged = function (p, data) {
        if (data && (!data.eventType || data.eventType != 'editor') && data.list) {
            data.list.forEach(function (cs) {
                if (cs && cs.localFile) {
                    _this.codeChanged(crsaMakeUrlFromFile(cs.getLocalFileName()), cs.getFilteredSource());
                }
            })
        }
    }

    $('body').one('pinegrow-ready', function(e, pinegrow) {
        pinegrow.codeEditors = codeEditors;

        pinegrow.addEventHandler('on_api_server_created', function(server) {

            if(!pgf.editor_sync) return;

            editorServer = server.addEndPoint('editor', function(socket) {
                socket.on('elementSelectedInEditor', function(data) {
                    ignoreSelectsForAWhile();
                    last_remote_selected_paths[data.url] = data.path;
                    connector.elementSelectedInEditor(data.url, data.path);
                })
                socket.on('codeChangedInEditor', function(data) {
                    syncChangedCode(data.url, data.code);
                    last_code_received_from_external[data.url] = {code: data.code, external: true};
                    connector.codeChangedInEditor(data.url, data.code, 'external_editor');
                })
                socket.on('fileSavedInEditor', function(data) {
                    connector.fileSavedInEditor(data.url);
                })
                socket.on('requestParserModule', function() {
                    if(pgIsLive()) {

                        fetch('lib/crsa/pg-parser.js')
                            .then(response => {
                                if (!response.ok) {
                                    throw new Error('Network response was not OK');
                                }
                                return response.text();
                            })
                            .then(code => {
                                socket.emit('parserModule', {code: code});
                            })
                            .catch(error => {
                                console.error('requestParserModule: Unable to get pg-parser.js', error);
                            });

                    } else {
                        var fs = require('fs');
                        var file = crsaMakeFileFromUrl(pgMakeFileUrl(crsaGetBaseForUrl(window.location.href) + '/lib/crsa/pg-parser.js'));
                        var code = fs.readFileSync(file, {encoding: 'utf8'});

                        socket.emit('parserModule', {code: code});
                    }
                })
                socket.on('openFile', function(data) {
                    ignoreSelectsForAWhile();
                    pinegrow.openOrShowPage(data.url, function(page) {

                    }, true, true);
                })

                socket.on('refreshPage', function(data) {
                    ignoreSelectsForAWhile();
                    connector.refreshPage(data.url);
                })

                socket.on('disconnect', function(reason, details) {
                    numberOfConnectedClients--;
                    pinegrow.showQuickMessage('External code editor disconnected.');
                })

                socket.on('error', function(a, b, c) {
                    console.log(a, b, c)
                })

                broadcastOpenFiles();

                numberOfConnectedClients++;

                pinegrow.showQuickMessage('External code editor connected.');
                //
            })

        });

        pinegrow.addEventHandler('on_css_loaded', function(pagenull, cs) {
            if(cs.url) {
                _this.cssCodeChanged(cs.url, function () {
                    return cs.filterSource(cs.getSource());
                }, null, null, cs)
            }
        });

        pinegrow.addEventHandler('on_imported_css_loaded', function (pagenull, cs) {
            if (cs.url) {
                _this.cssCodeChanged(cs.url, function () {
                    return cs.filterSource(cs.getSource());
                }, null, null, cs)
            }
        });

        pinegrow.addEventHandler('on_page_loaded_from_external', function(page) {
            if(page.url) {
                _this.codeChanged(page.url, function() {
                    return page.getSource();
                }, 'external_editor')
            }
        })

        pinegrow.addEventHandler('on_css_stylesheet_locked_changed', function(pagenull, cs, locked) {
            if(cs.url) {
                findEditorsByUrl(cs.url).forEach(function(editor) {
                    if(editor.setReadOnly) editor.setReadOnly(locked);
                })
            }
        })

        pinegrow.addEventHandler('on_page_changes_done', function(page, info) {
            var source = (info && info.source) ? info.source : null;
            _this.codeChanged(page.url, function() {
                return page.getSource()
            }, source, null, info);
        })

        pinegrow.addEventHandler('on_elements_collapsed_expanded', function(page) {
            var info = {};
            _this.codeChanged(page.url, function() {
                return page.getSource()
            }, null, null, info);
        })

        pinegrow.addEventHandler('on_css_stylesheet_changed', function(pagenull, cs, info) {
            if(cs.inline && cs.inlineSheetPgel) {
                var page = cs.inlineSheetPgel.getPage();
                if(page) {
                    var source = (info && info.source) ? info.source : null;
                    _this.codeChanged(page.url, function() {
                        return page.getSource()
                    }, source, null, {});
                }
            }
        })

        pinegrow.addEventHandler('on_css_file_saved', function(pagenull, cs) {
            if(cs.url) {
                _this.onCssSaved(cs.url);
                findEditorsByUrl(cs.url).forEach(function(editor) {
                    if(editor.view) editor.view.setModified(false);
                })
            }
        })

        pinegrow.addEventHandler('on_page_saved', function(page) {
            _this.onPageSaved(page.url);
        })

        pinegrow.addEventHandler('on_selected_elements_changed', function(page, pages, info) {
            if(info?.sourceType && info?.sourceType === 'undo') return;

            var pgel = pinegrow.selectedElements.getLastSelected();
            if(pgel && !ignore_select_after_refresh) {
                var page = pgel.getPage();
                if(page && pgel) {
                    _this.elementWasSelectedInPreview(page.url, pgel.getPath(), info);
                }
            }
        })

        // pinegrow.addEventHandler('on_selected_elements_changed', function(page, pages, info) {
        //     if(info.sourceType && info.sourceType == 'undo') return;
        //
        //     var pgel = pinegrow.selectedElements.getLastSelected();
        //     if(pgel && !ignore_select_after_refresh) {
        //         var page = pgel.getPage();
        //         if(page && pgel) {
        //             _this.elementWasSelectedInPreview(page.url, pgel.getPath(), info);
        //         }
        //     }
        // });

        pinegrow.addEventHandler('on_css_source_changed', onCodeChanged);

        $('body').on('crsa-rule-selected', function(e, data) {
            if(!pgf.kids && pgPageCodeEditorsStack.highlight_css_rule && pgPageCodeEditorsStack.isEditorShown() && data.rule && data.rule.selector && data.source && data.source instanceof PgStylesPane) {

                pinegrow.revealCSSRuleInCodeEditor(data.rule);
                /*
                var cs = pgStylesheetHelper.getStylesheetFromNode(data.rule);
                if(cs && cs.url) {
                    var editor = pinegrow.codeEditors.isUrlOpen(cs.url);
                    if(editor) {
                        if(!pgPageCodeEditorsStack.isEditorShown(editor.view)) {
                            pgPageCodeEditorsStack.showEditor(editor.view);
                        }
                        editor.search(cs.getSourceForRules(data.rule));
                    }
                }*/
            }
        });
    });

    $('body')
        .on('crsa-variables-changed crsa-stylesheets-changed crsa-rule-changed-in-editor', onCodeChanged)
        .on('crsa-page-closed crsa-page-loaded crsa-page-saved-as crsa-stylesheets-changed', function() {
            broadcastOpenFiles();
        })
}

var codeEditors = new PgApiCodeEditors(new PgApiCodeEditorConnector());




/*
var io = require('socket.io')(40001);
var remoteEditor = io
  .of('/editor')
  .on('connection', function (socket) {
      console.log('server got connection');

      socket.on('thanks', function(data) {
          console.log('GOT THANKS!');
      })
    socket.emit('elementSelected', {
        that: 'only'
      , '/chat': 'will get'
    });
    remoteEditor.emit('a message', {
        everyone: 'in'
      , '/chat': 'will get'
    });
  })




  */



var CrsaCodeEdit = function() {

    this.currentPage = null;
    this.currentCrsaPage = null;

    this.currentCodePage = null;

    var old_head_value = '';
    var old_body_value = '';
    var update_timer = null;
    var tree_update_timer = null;
    var parse_timer = null;
    var $html = null;
    var doc;
    var code_has_errors;
    var codeEditor;
    var mirror;
    var $textEditWrapper = $('#textedit_wrapper');
    var $textEdit = $('#textedit');
    var $bar = $('#textedit_bar');
    var $select = $bar.find('select.edit-css-select');

    var $live_refresh = $bar.find('.live-update');
    $live_refresh.tooltip({container: 'body', placement: 'bottom', title: 'Auto refresh page view. If disabled - or if the code has syntax errors - click on Refresh! (CMD + R) to refresh page view.', trigger: 'hover'});
    //$live_refresh.hide();

    var live_refresh = true;
    var live_refresh_off_reason = null;

    if(pinegrow.getSetting('code-live-update','1') == '1') {
        $live_refresh.attr('checked', 'checked');
    }

    var isLiveRefresh = function() {
        return live_refresh && $live_refresh.is(':checked');
    }

    var last_syntax_error_message = 0;

    var setLiveRefresh = function(val, reason, always_show) {
        var str;
        var showmsg = true;
        switch(reason) {
            case 'page':
                str = 'Can\'t auto refresh. Whole page changed.';
                break;
            case 'error':
                str = 'Can\'t auto refresh. Code has syntax errors.';
                var t = (new Date()).getTime();
                if(t - last_syntax_error_message > 30*1000) {
                    last_syntax_error_message = t;
                } else {
                    showmsg = false;
                }
                break;
            case 'slow':
                str = 'Auto refresh takes too long.';
                showmsg = false;
                break;
            default:
                str = '';
                break;
        }
        if(!val) {
            if((live_refresh || always_show) && showmsg) {
                crsaQuickMessage(" " + str + " <b>Refresh (CMD + R)</b> manually.", 3000);
            }
            $live_refresh.parent().addClass('manual');
            if(!live_refresh_off_reason || reason == 'page') {
                live_refresh_off_reason = reason;
            }
        } else {
            $live_refresh.parent().removeClass('manual');
            live_refresh_off_reason = null;
        }
        live_refresh = val;
    }

    this.refreshBeforeSaveIfNeeded = function() {
        if(code_mode == 'html' && this.isInEdit(this.currentPage) && (has_changes || live_refresh_off_reason == 'page')) {
            this.refreshPreview();
            has_changes = false;
        }
    }

    $live_refresh.on('change', function() {
        pinegrow.setSetting('code-live-update', isLiveRefresh() ? '1' : '0');
        if(isLiveRefresh()) {
            crsaQuickMessage('Auto refresh is enabled.');
        } else {
            crsaQuickMessage('Auto refresh is disabled.');
            showNotice('Click on <b>Refresh!</b> or press <b>CMD + R</b> to manually refresh page view after you make changes.', 'Auto update is disabled', 'auto-refresh');
        }
    });

    var $wrap = $bar.find('.wrap');
    $wrap.on('change', function() {
        var checked = $wrap.is(':checked');
        pinegrow.setSetting('code-wrap', checked ? '1' : '0');
        mirror.setOption('lineWrapping', checked);
    });
    if(pinegrow.getSetting('code-wrap','1') == '1') {
        $wrap.attr('checked', 'checked');
    }



    var code_ignore_change = false;
    var code_mode;
    var selected_cs = null;
    var needsUpdate = false;
    var source_synced = false;
    var force_refresh = false;
    var has_changes = false;

    var _this = this;

    var script_changes_msg_shown = false;

    var refresh_html_on_select = false;

    var first_rule_updates = true;

    var id_to_markers = {};

    var setCodeMode = function(mode) {
        code_mode = mode;
        $bar.find('.mode-button').removeClass('active-mode');
        $bar.find('.edit-' + mode).addClass('active-mode');

        if(mode == 'js') {
            $bar.find('.mode-button').hide();
            $bar.find('.edit-refresh').hide();


        } else {
            $bar.find('.mode-button').show();
            $bar.find('.edit-refresh').show();
        }
    }

    setCodeMode('html');

    this.findAndHideIds = function(mirror, id_to_markers) {
        var tmp = [];
        var re = /data-pg-id="([0-9]+)"/g;
        var doc = mirror.getDoc();
        var num = doc.lineCount();
        for(var i = 0; i < num; i++) {
            var line = doc.getLine(i);
            if(line) {
                re.lastIndex = 0;
                var a;
                while ((a = re.exec(line)) !== null)
                {
                    var idx = re.lastIndex - a[0].length - 1;
                    var id = a[1] + "";
                    tmp.push({line: i, ch: idx, id: id, len: a[0].length + 1});
                }
            }
        }
        try {
        if(id_to_markers) {
            $.each(id_to_markers, function(id, marker) {
                marker.clear();
            })
        }
        } catch(err) {}

        id_to_markers = {};
        for(var i = 0; i < tmp.length; i++) {
           id_to_markers[tmp[i].id] = doc.markText({line: tmp[i].line, ch: tmp[i].ch}, {line: tmp[i].line, ch: tmp[i].ch + tmp[i].len}, {collapsed: true, inclusiveLeft: false, inclusiveRight: true});
        }
        return id_to_markers;
    }

    this.findAndHideCollapsed = function(mirror) {
        var tmp = [];
        var re = /data-pg-collapsed/g;
        var doc = mirror.getDoc();
        var num = doc.lineCount();
        for(var i = 0; i < num; i++) {
            var line = doc.getLine(i);
            if(line) {
                re.lastIndex = 0;
                var a;
                while ((a = re.exec(line)) !== null)
                {
                    var idx = re.lastIndex - a[0].length - 1;
                    var id = a[1] + "";
                    tmp.push({line: i, ch: idx, id: id, len: a[0].length + 1});
                }
            }
        }

        for(var i = 0; i < tmp.length; i++) {
            doc.markText({line: tmp[i].line, ch: tmp[i].ch}, {line: tmp[i].line, ch: tmp[i].ch + tmp[i].len}, {collapsed: true, inclusiveLeft: false, inclusiveRight: true});
        }
    }

    var scroll_to_selection = true;

    this.editorSizeChanged = function() {
        if(mirror) mirror.refresh();
    }

    this.selectElementInCodeMirror = function($el, mirror, id_to_markers) {
        var pgel = getElementPgNode($el);
        if(pgel) {
            var pgid = pgel.getId();

            //console.log(pgel.toDebug());
            var source = pgel.toStringWithIds(true, pinegrow.getFormatHtmlOptions());
            var source_lines = source.split("\n");

            pgid = pgid + "";
            if(id_to_markers[pgid]) {
                var marker = id_to_markers[pgid];
                var pos = marker.find();
                var doc = mirror.getDoc();
                var start_ch = 0;

                if(pos) {
                    var start_line = pos.from.line;
                    var li = pos.from.line;
                    var start = pos.from.ch;
                    while(li >= 0) {
                        var line = doc.getLine(li);
                        if(line && line.length) {
                            if(li != pos.from.line) start = line.length-1;
                            var idx = line.lastIndexOf("<", start);
                            if(idx >= 0) {
                                start_ch = idx;
                                start_line = li;
                                break;
                            }
                        }
                        li--;
                    }
                    /*  doc.markText({line: start_line, ch: start_ch}, {line: pos.from.line + source_lines.length - 1, ch: source_lines[source_lines.length-1].length + (source_lines.length == 1 ? start_ch : 0)},
                     {

                     }); */
                  /*  if(!crsaIsInEdit()) {
                        mirror.focus();
                    }*/

                    // mirror.setCursor({line: start_line, ch: start_ch});

                    doc.setSelection(
                        {line: start_line, ch: start_ch},
                        {line: pos.from.line + source_lines.length - 1, ch: source_lines[source_lines.length-1].length + (source_lines.length == 1 ? start_ch : start_ch)},
                        {scroll: scroll_to_selection});

                }
            }
        }
    }

    this.selectElementInEditor = function($el) {
        if(code_mode == 'html') {
            if(refresh_html_on_select) {
                if(!has_changes) {
                    var si = mirror.getScrollInfo();
                    updateCodeDisplay();
                    mirror.scrollTo(si.left, si.top);
                } else {
                    refresh_html_on_select = false;
                }
            }
            this.selectElementInCodeMirror($el, mirror, id_to_markers);
        }
        scroll_to_selection = true;
    }

    pinegrow.addEventHandler('on_element_selected', function(page, pgel, info) {
        if(info && info.sourceType && info.sourceType == 'undo') return;
        if(!_this.currentPage) return;
        //if(!isLiveRefresh()) return;

        if(pgel) {
            setTimeout(function() {
                var $el = pgel.get$DOMElement();
                _this.selectElementInEditor($el);
            }, 100); //delay so that we're called after crsa-page-changed event
        }
    });

    var getPgIdOfElementAtPos = function(mirror, pos, posOut) {
        var doc = mirror.getDoc();
        var pgid = null;
        var tag = null;
        var cur = {line: pos.line, ch: pos.ch};
        while(cur.line >= 0 && !tag) {
            tag = CodeMirror.findMatchingTag(mirror, cur, 0);
            cur.ch--;
            if(cur.ch < 0) {
                cur.line--;
                if(cur.line >= 0) {
                    var line = doc.getLine(cur.line).length;
                    cur.ch = line.length ? line.length-1 : 0;
                }
            }
        }
        if(tag && tag.open) {
            var start_line = tag.open.from.line;
            var li = start_line;
            var start = tag.open.from.ch;
            var line = doc.getLine(li);
            while(typeof line == 'string') {

                var idx = line.indexOf('data-pg-id', start);
                if(idx < 0) {
                    li++;
                    line = doc.getLine(li);
                    start = 0;
                } else {
                    line = line.substr(idx);
                    var m = line.match(/data\-pg\-id="([0-9]+)"/);
                    if(m) {
                        pgid = m[1];
                        if(posOut) {
                            posOut.line = cur.line;
                            posOut.ch = cur.ch;
                            posOut.tag = tag;
                        }
                        break;
                    }
                }
            }
        }
        return pgid;
    }

    var parse_interval = 100;
    var last_update_stylesheets_msg = 0;
    var last_errors_msg = 0;

    var onChange = function(changeObj) {
        if(code_ignore_change || !mirror) return;

        //if(!isLiveRefresh() && !force_refresh) return;

        var cp = getCrsaPageForIframe(_this.currentPage);

        var doc = mirror.getDoc();
        var v = doc.getValue();

        //console.log(v);
        var ms = (new Date()).getTime();

        has_changes = true;

        if(code_mode == 'html') {

            if(isLiveRefresh() || force_refresh || live_refresh_off_reason == 'error' || live_refresh_off_reason == 'page') {
                var orig_cic = code_ignore_change;

                var ms = (new Date()).getTime();
                code_ignore_change = true;
                var ret = cp.applyChangesToSource(v, !force_refresh);
                code_ignore_change = orig_cic;
                var took = (new Date()).getTime() - ms;

                if(took > 150) {
                    setLiveRefresh(false, "slow");

                }
                //console.log('Auto update took ' + took + 'ms');

                refresh_html_on_select = true;

                if(ret.updated) {
                    if(!isLiveRefresh() && live_refresh_off_reason == 'error') {
                        //there was error but now is gone
                        setLiveRefresh(true);
                        cp.setSyntaxErrorIndicator(false);
                        //crsaQuickMessage("Auto refreshed!");
                    }
                    if(ret.stylesheets_changed || ret.scripts_changed) {
                        if(ms - last_update_stylesheets_msg > 10000) {
                            crsaQuickMessage("<b>Refresh (CMD + R)</b> to update scripts and stylesheets list.");
                            last_update_stylesheets_msg = ms;
                        }
                    }
                    has_changes = false;

                } else if(ret.changes && !ret.update) {
                    //whole document is changed
                    setLiveRefresh(false, "page");
                    has_changes = false;

                } else if(ret.errors && !ret.update) {
                    //we have errors
                    setLiveRefresh(false, 'error', ms - last_errors_msg > 10000);
                    cp.setSyntaxErrorIndicator(true);
                    last_errors_msg = ms;
                }
            } else if(live_refresh_off_reason == 'slow') {
                if(update_timer) {
                    clearTimeout(update_timer);
                }
                update_timer = setTimeout(function() {
                    update_timer = null;
                    setLiveRefresh(true);
                    onChange();
                }, 1000);
            }
            return;
        } else if(code_mode == 'js') {
            if(_this.currentCodePage) {
                _this.currentCodePage.applyChangesToSource(v);
                updateNameForCodePage();
            }
        } else if(code_mode == 'css') {
            if(selected_cs) {
                if(update_timer) {
                    clearTimeout(update_timer);
                    update_timer = null;
                }
                update_timer = setTimeout(function() {
                    has_changes = false;
                    selected_cs.forgetSourceErrors();
                    selected_cs.setSource(v, function() {
                        var gerror = selected_cs.getError();
                        if(gerror) {
                            console.log(gerror);

                            //codeEditor.getSession().addGutterDecoration(selected_cs.less_error.line, 'crsa-code-error');
                            code_has_errors = true;

                            /*codeEditor.getSession().setAnnotations([{
                                row: gerror.line - 1,
                                column: gerror.column,
                                text: "Syntax error",
                                type: "error" // also warning and information
                            }]);*/
                        } else {
                            if(code_has_errors) {
                              //  codeEditor.getSession().clearAnnotations();
                                code_has_errors = false;
                            }
                        }

                        $('body').trigger('crsa-stylesheets-changed');
                    });
                }, 750);
            }
        }
        needsUpdate = true;
    };


    var getHead = function(html) {
        var i = html.search(/<\/head>/i);
        if(i >= 0) {
            return html.substr(0, i);
        }
        return '';
    }

    var getBody = function(html) {
        var m = html.match(/<body[^>]*>([\s\S]*)<\/body>/mi);
        // console.log(m);
        return m ? m[1] : null;
    }

    var getHtml = function(html) {
        var m = html.match(/<html[^>]*>([\s\S]*)<\/html>/mi);
        // console.log(m);
        return m ? m[1] : null;
    }

    var getStylesheets = function() {
        if(!_this.currentPage) return [];
        var cp = getCrsaPageStylesForPage(_this.currentPage);
        var r = [];
        if(!cp) return r;
        $.each(cp.getAllCrsaStylesheets(), function(i, cs) {
            if(!cs.inline && cs.loaded) {
                r.push(cs);
            }
        });
        return r;
    }

    var showSelectedCss = function() {
        setCodeMode('css');
        var name = $select.val();

        code_mode = 'css';

        var getSelectedCsFromAllImportedFiles = function (cs) {
            var forAll = function (importedFiles) {
                for (var i = 0; i < importedFiles.length; i++) {
                    var importedFile = importedFiles[i];
                    (importedFile);
                    if (importedFile.name == name) selected_cs = importedFile

                    if (importedFile.importedFiles.length > 0) {
                        return forAll(importedFile.importedFiles);
                    }
                }
            }

            return forAll(cs.importedFiles);
        }

        var css_list = getStylesheets();
        $.each(css_list, function(i, cs) {
            if (cs.compiledFile) cs = cs.sourceStylesheet;

            if(cs.name == name) {
                selected_cs = cs;
                return false;
            }

            getSelectedCsFromAllImportedFiles(cs);
        });
        if(selected_cs) {
            updateCodeDisplay();
        }
    }

    $select.on('change', function(e) {
        setTimeout(function() {
            showSelectedCss();
        }, 10);
    });

    this.pageChanged = function(page, html_code) {
        if(code_ignore_change) return;
        setTimeout(function() {
            var si = mirror.getScrollInfo();
            updateCodeDisplay(code_mode == 'html' ? html_code : null);
            mirror.scrollTo(si.left, si.top);
        }, 1);
    }

    var code_ignore_change_timeout = null;

    var updateName = function(str) {
        $bar.find('.file-name').html(str);
    }

    var updateNameForCodePage = function() {
        if(_this.currentCodePage) {
            updateName(_this.currentCodePage.name + (_this.currentCodePage.hasChanges() ? ' *' : ''));
        }
    }

    var updateCodeDisplay = function(html_code) {

        if(code_ignore_change_timeout) {
            clearTimeout(code_ignore_change_timeout);
            code_ignore_change_timeout = null;
        }
        code_ignore_change = true;
        var str = '';
        switch(code_mode) {
            case 'css':
                str = selected_cs ? selected_cs.filterSource(selected_cs.getSource(), true) : '';

                var mode = 'text/css';
                if (selected_cs) {
                    if (selected_cs.fileType == "scss") mode = 'text/x-scss';
                    else if (selected_cs.fileType == "less") mode = 'text/x-less';
                }

                mirror.setOption('mode', mode);

                mirror.getDoc().setValue(str);
                code_ignore_change = false;
                updateName(selected_cs ? crsaGetNameFromUrl(selected_cs.url) : '- none -')
                break;
            case 'js':
                if(_this.currentCodePage) {
                    var ext = crsaGetExtFromUrl(_this.currentCodePage.url);
                    var mode = null;

                    if(pinegrow.isFileEditable(_this.currentCodePage.localFile)) {
                        mode = 'application/x-httpd-php';
                    }
                    var auto = {
                        'php,php5': 'application/x-httpd-php',
                        'js': 'text/javascript',
                        'json' : 'application/json',
                        'css,less,scss' : 'text/css'
                    }
                    $.each(auto, function(key, m) {
                        if(key.split(',').indexOf(ext) >= 0) {
                            mode = m;
                            return false;
                        }
                    })
                    if(mirror) {
                        mirror.setOption('mode', mode);
                        //CodeMirror.autoLoadMode(mirror, mode);
                    }
                }
                mirror.getDoc().setValue(_this.currentCodePage ? _this.currentCodePage.code : '');
                code_ignore_change = false;
                updateNameForCodePage();

                break;
            default:
                updateName(_this.currentCrsaPage.tab_name);
                if(html_code) {
                    str = html_code;
                } else {
                    //var body = getIframeBody(_this.currentPage.get(0));
                    //str = pinegrow.formatHtml(body.outerHTML);
                    str = getCrsaPageForIframe(_this.currentPage).sourceNode.toStringWithIds(true, pinegrow.getFormatHtmlOptions());

                   // str = removeCrsaClassesFromHtml(str);
                   // str = pinegrow.formatHtml(str);
                }
                //source_synced = false;
                refresh_html_on_select = false;

                //old_head_value = getHead(str);
                old_body_value = getBody(str);
                mirror.operation(function() {
                    mirror.setOption('mode', 'application/x-httpd-php');
                    mirror.getDoc().setValue(str);
                    id_to_markers = _this.findAndHideIds(mirror, id_to_markers);

                    code_ignore_change_timeout = setTimeout(function() {
                        code_ignore_change = false;
                        code_ignore_change_timeout = null;
                    }, 100);
                });

                break;
        }

        //codeEditor.getSession().getSelection().clearSelection();

    }

    $('body').on('crsa-page-selected', function(e, crsaPage) {
        if(mirror && code_mode != 'js' && crsaPage && crsaPage != _this.currentCrsaPage) {
            _this.exitEdit(false, true);
            _this.editCode(crsaPage.$iframe, null, _this.currentCodePage);
        }
    })

    this.exitEdit = function(update, dont_close_window) {
        this.refreshBeforeSaveIfNeeded();

        if(edit_win && !dont_close_window) {
            edit_win.hide();
            //edit_win.close(true);
            //edit_win = null;
        }

        var page = this.currentPage;
        this.currentPage = null;
        this.currentCrsaPage = null;
        this.currentCodePage = null;
        $('body').append($textEditWrapper);
        $textEditWrapper.hide();
        $textEdit.html('');

        pinegrow.code_editors.unregister(mirror);

        mirror = null;

        if(tree_update_timer) {
            clearTimeout(tree_update_timer);
            tree_update_timer = null;
        }
        /* if(parse_timer) {
            clearTimeout(parse_timer);
            parse_timer = null;
        }*/

        if(update && code_mode != 'js') {
            //this.refreshPreview();
            //$.fn.crsa('updateStructureAndWireAllElemets', page);
            $('body').trigger('crsa-stylesheets-changed');

            //do we need this
            /*
            var cp = getCrsaPageForIframe(page);
            cp.addCrsaStyles();
            cp.runScripts();
            didMakeChange(page);
            */
        }
        $.fn.crsa('resizeChrome');

    }


    this.isInEdit = function(page) {
        if(page && !page instanceof CrsaPage) {
            debugger;
            console.log('isInEdit page not CP');
            return false;
        }
        if(this.currentCrsaPage == page) return true;
        if(this.currentCodePage && !page) return true;
        return false;
    }

    this.getExternalWindow = function() {
        return edit_win;
    }

    this.refreshPreview = function() {
        if(!has_changes) {
            this.currentCrsaPage.refresh();
            setLiveRefresh(true);
            return;
        }

        force_refresh = true;
        onChange(null);
        force_refresh = false;
        crsaQuickMessage('Page view was refreshed.');
    }

    var last_undo_rec_time = 0;

    var onBarClick = function(event) {
        var $b = $(event.delegateTarget);

        if($b.hasClass('edit-win')) {
            _this.showInExternalWindow = !_this.showInExternalWindow;
            $b.find('i').removeClass('fa-expand fa-compress').addClass(_this.showInExternalWindow ? 'fa-compress' : 'fa-expand');
            var cp = _this.currentCrsaPage;
            var codep = _this.currentCodePage;
            _this.exitEdit();
            _this.editCode(cp ? cp.$iframe : null, null, codep);

            if(!_this.showInExternalWindow && edit_win) {
                edit_win.close(true);
                edit_win = null;
            }

        } else if($b.hasClass('edit-done')) {

            _this.exitEdit(true);

        } else if($b.hasClass('edit-refresh')) {
            if(event.target != $live_refresh.get(0)) {

                _this.refreshPreview();
            } else {
                return;
            }
        } else {
            _this.refreshBeforeSaveIfNeeded();
            if($b.hasClass('edit-html')) {
                setCodeMode('html');
            } else if($b.hasClass('edit-js')) {
                setCodeMode('js');
            } else if($b.hasClass('edit-css')) {
                showSelectedCss();
            }

            updateCodeDisplay();
        }
        last_undo_rec_time = 0;

        event.preventDefault();
    }

    $bar.find('a, button').on('click', function(event) {
        onBarClick(event);
    });

    this.selectClickedElement = function(mirror, event) {
        var pos = mirror.coordsChar({left: event.pageX, top: event.pageY}, "page");
        this.selectElementAtCursorPosition(mirror, pos);
    }

    this.selectElementAtCursorPosition = function(mirror, pos) {
        console.log(pos)
        if(pos) {
            var pgel = null;
            var posOut = {};
            var pgid;
            do {
                pgid = getPgIdOfElementAtPos(mirror, pos, posOut);
                if(pgid) {
                    pgel = getPgNodeByPgId(pgid);
                    pos.line = posOut.line;
                    pos.ch = posOut.ch - 1;
                    if(pos.ch < 0) {
                        pos.line--;
                        if(pos.line > 0) {
                            var line = mirror.getDoc().getLine(pos.line);
                            if(line && line.length) {
                                pos.ch = line.length-1;
                            } else {
                                pos.ch = 0;
                            }
                        }
                    }
                }
            }
            while(pgid && posOut.line >= 0 && pgel && pgel.tagName == 'br');

            if(pgel) {
                setTimeout(function() {
                    scroll_to_selection = false;
                    pinegrow.selectElement(pgel, 'code_click');
                    pgel.scrollTo();
                }, 10);
            }
        }
    }

    this.setFontSize = function(size, font, doc) {
        //debugger;
        // if(!doc) {
        //     /*if(edit_win && edit_win.window.document) {
        //         this.setFontSize(size, font, edit_win.window.document);
        //     }*/
        //     doc = document;
        // }
        // var $head = $(doc.head);
        // var $s = $head.find('#code-size-style');
        // var ff = (font && font.length) ? 'font-family:' + font + ';' : '';
        // var css = '<style id="code-size-style">.CodeMirror pre { font-size:' + size + ';' + ff + '}</style>';
        //
        // if($s.length == 0) {
        //     $head.append(css);
        // } else {
        //     $s.replaceWith($(css));
        // }
        pinegrow.code_editors.forEachEditor(function(mirror) {
            mirror.updateOptions({fontSize: size})
            mirror.updateOptions({fontFamily: font == "" ? "monospace" : font})
        })
    }

    ///////
    var edit_win = null;
    var edit_win_manager = null;


    this.showInExternalWindow = false;

    /////
    this.editCode = function($iframe, done, codePage) {

        /*
        var page = pinegrow.getSelectedPage();
        codeEditors.open(page.url, page.getCode(), function(editor_id) {

        })

        return;/////
        */

        if(!this.showInExternalWindow) {
            this.showEditor($iframe, null, codePage);
            if(done) done();


        } else {
            pinegrow.stats.using('edit.pagecodewin');
            if(edit_win) {
                this.showEditor($iframe, edit_win, codePage);
                edit_win.show();
                edit_win.focus();
                if(done) done();
            } else {
                var gui = require('nw.gui');

                edit_win = gui.Window.open('empty.html', {
                    focus: true,
                    toolbar: false,
                    title: 'Pinegrow - Edit code'
                });

                edit_win_manager = new pgWindowManager(edit_win, 'editCode', function(win) {
                })

                edit_win.on('close', function () {
                    _this.exitEdit(true);
                    edit_win.close(true);
                    edit_win = null;
                });


                edit_win.on('loaded', function () {
                    _this.setFontSize( pinegrow.getSetting('code-size', '12px'), pinegrow.getSetting('code-font', '') );
                    _this.showEditor($iframe, edit_win, codePage);
                    $(edit_win.window.document).on('keydown', function(e) {
                        return $.fn.crsa('processKeydownEvent', e);
                    });
                    if(done) done();
                });

                var main = gui.Window.get();
                main.on('focus', function() {
                    if(edit_win) edit_win.setAlwaysOnTop(true);
                })
                main.on('blur', function() {
                    if(edit_win) edit_win.setAlwaysOnTop(false);
                })


            }
        }
    }

    this.closeExternalWindow = function() {
        if(edit_win) {
            edit_win.close(true);
            edit_win = null;
        }
    }

    var edit_code_msg_shown = false;

    this.isEditingCodeOnlyPage = function(url) {
        return this.currentCodePage && code_mode == 'js' && (!url || url == this.currentCodePage.url);
    }

    this.isEditorFocused = function() {
        if($(document.activeElement).closest('#textedit').length) return true;
        //console.log(document.activeElement);
        return false;
    }

    pinegrow.addEventHandler('on_page_saved', function(page) {
        if(code_mode == 'js' && _this.currentCodePage == page) {
            updateNameForCodePage();
        }
    })

    this.showEditor = function($iframe, win, codePage) {
        pinegrow.stats.using('edit.pagecode');
        /*
        if((this.currentPage && !this.isInEdit($iframe))) {
            this.exitEdit(true);
        }
        */
        if(mirror) {
            this.exitEdit(true);
        }

        this.currentPage = $iframe;
        this.currentCrsaPage = getCrsaPageForIframe($iframe);

        this.currentCodePage = codePage;

        if($iframe) {
            var b = getIframeBody(this.currentPage[0]);
            doc = getIframeDocument(this.currentPage[0]);
            $html = $(doc).find('> html');

            if(!canMakeChange(getElementPgNode($html), 'edit_code')) return;
        } else {
            doc = null;
            $html = null;
        }

        if(win) {
            var $win_body = $(win.window.document.body);
            $textEditWrapper.css('height','auto').css('width','auto').css('top',0).css('right',0).css('left',0).css('bottom',0).appendTo($win_body.find('#win-content'));
        } else {
            $('body').append($textEditWrapper);
        }


        var $mirror_el = null;

        //pgSetCodeMirrorDocumentAndWindow(edit_win.window.document, edit_win.window);

        //replace all window. with cm.window or this.cm.window

        setCodeMode(codePage ? 'js' : 'html');

        var undo_depth = 0;

        this.editorHandlesUndo = false;

        var keys = {
            "Ctrl-Q": function(cm){ cm.foldCode(cm.getCursor()); },
            "Ctrl-Space": "autocomplete"
        };

        if(code_mode != 'js') {
            keys["Ctrl-Z"] = function(cm) {};
        } else {
            undo_depth = 100;
            this.editorHandlesUndo = true;
        }

        mirror = CodeMirror(function(elt) {
            $textEdit.append($(elt).css('height','100%'));
            $mirror_el = $(elt);
        }, {
            pgWindow: edit_win ? edit_win.window : window,
            mode: "text/html",
            autoFocus: true,
            theme: pinegrow.getSetting('code-theme-cm', pinegrow.defaults.codeTheme),
            indentUnit: parseInt(pinegrow.getSetting('html-indent-size', '4')),
            tabSize: parseInt(pinegrow.getSetting('html-indent-size', '4')),
            lineWrapping: pinegrow.getSetting('code-wrap','1') == '1',
            lineNumbers: true,
            matchBrackets: true,
            autoCloseBrackets: true,
            autoCloseTags: true,
            matchTags: true,
            undoDepth: undo_depth,
            extraKeys: keys,
            foldGutter: true,
            gutters: ["CodeMirror-linenumbers", "CodeMirror-foldgutter"],
            profile: 'xhtml',
            hintOptions : {
                completeSingle: false
            }
        });

        setTimeout(function() {
            mirror.focus();
        }, 100);


        setLightDarkClassForEditor($mirror_el, pinegrow.getSetting('code-theme-cm', pinegrow.defaults.codeTheme));

        showAutoCompleteOnInput(mirror);

        pinegrow.code_editors.register(mirror);

        $mirror_el.attr('data-has-codemirror','').data('codeMirrorInstance', mirror);

        if(pinegrow.getSetting('editor-emmet', pgf.default_emmet_setting) === 'true') {
            emmetCodeMirror(mirror);
        }


        if(isApp()) {
            $mirror_el.on('copy', function() {
                var selectedText = mirror.getDoc().getSelection();

                if(selectedText && selectedText.length) {

                    setTimeout(function() {
                        var gui = require('nw.gui');
                        var clipboard = gui.Clipboard.get();
                        selectedText = pgRemovePgIdsFromCode(selectedText);
                        //console.log(selectedText);
                        clipboard.set(selectedText, 'text');
                    }, 50);
                }
            });
        }

        mirror.on('contextmenu', function(instance, event) {
            _this.selectClickedElement(mirror, event, _this.currentCrsaPage);
            return false;
        });

        var on_change_timer = null;

        mirror.on('change', function(instance, changeObj) {
            if(on_change_timer) {
                clearTimeout(on_change_timer);
            }
            if(code_mode == 'js') {
                onChange(changeObj);
            } else if(code_mode == 'css' && code_ignore_change) {
                onChange(changeObj);
            } else {
                on_change_timer = setTimeout(function() {
                    onChange(changeObj);
                    on_change_timer = null;
                }, 200);
            }
        });

        last_undo_rec_time = 0;//(new Date()).getTime();

        mirror.on('beforeChange', function(instance, change) {
            if(!code_ignore_change && code_mode == 'html') {
                if(change.text && change.update) {
                    var a = [];
                    var len = 0;
                    for(var i = 0; i < change.text.length; i++) {
                        len += change.text[i].length;
                        a.push(change.text[i].replace(/\sdata-pg-id="[0-9]+"/g,''));
                    }
                    change.update(null, null, a);

                    var ms = (new Date()).getTime();
                    if(ms - last_undo_rec_time > 5000) {
                        last_undo_rec_time = ms;
                        if(code_mode == 'css' && selected_cs) { // How code_mode == 'html and 'css'
                            // selected_cs.setChanged(true); //otherwise undo will not record
                        }
                        willMakeChange(_this.currentPage, "Edit code");
                    }
                }
                if(!change.update) {
                    //undo or redo
                    console.log('UNDO in edit code');
                }
            }
        });

        //codeEditor.setTheme("ace/theme/" + pinegrow.getSetting('code-theme', 'xcode'));

        $textEditWrapper.show();

        $textEditWrapper.find('.tree-resizer').remove();

        var $resizer = $('<div/>', {class: 'tree-resizer'}).appendTo($textEditWrapper)
            .on('mousedown', function(e) {
                var startLeft = $textEdit.position().left;
                var startWidth = $textEdit.width();
                e.preventDefault();
                $.fn.crsapages('showOverlays');
                var bar_h = $bar.height();
                $('body')
                    .on('mousemove.editResizer', function(m) {
                        var topbar_h = 38;
                        var y = m.pageY;
                        var all = $(window).height();
                        var h = all - y;
                        var percent = h * 100 / all;

                        if(h < 80) h = 80;
                        if(h > all - 80) h = all - 80;

                        var wy = all - h;

                        $textEditWrapper.css('top', wy + 'px').css('height', h + 'px');
                        //$bar.css('top', wy + 'px');

                        localStorage.crsaCodeEditHeight = h;
                    })
                    .on('mouseup.editResizer', function(e) {
                        e.preventDefault();
                        $('body').off('.editResizer');
                        $.fn.crsapages('showOverlays', true);
                        $(window).trigger('resize');

                    });
            });//.css('top', 0);

        if(win) {
            $resizer.hide();
        } else {
            $resizer.show();
        }

        $.fn.crsa('resizeChrome');

        code_has_errors = false;

        var css_list = getStylesheets();
        $select.html('');

        var addOptionToSelect = function (cs) {
          $('<option/>', {value: cs.name}).html(cs.name).appendTo($select);
        }

        var addOptionAllImportedFiles = function (cs) {
            var forAll = function (importedFiles) {
                for (var i = 0; i < importedFiles.length; i++) {
                    var importedFile = importedFiles[i];
                    addOptionToSelect(importedFile);

                    if (importedFile.importedFiles.length > 0) {
                        forAll(importedFile.importedFiles);
                    }
                }
            }

            forAll(cs.importedFiles);
        }

        $.each(css_list, function(i, cs) {
            if (cs.compiledFile) cs = cs.sourceStylesheet;

            addOptionToSelect(cs);
            addOptionAllImportedFiles(cs);
        });
        if(selected_cs) $select.val(selected_cs.name);

        if(codePage) {
            setCodeMode('js');
        }

        //willMakeChange(_this.currentPage, "Edit code");
        updateCodeDisplay();

        if(!edit_code_msg_shown) {
            showNotice("<p><b>New in Pinegrow: Open code editor in external window!</b><p>Click on the <i style=\"color:#333;\" class=\"fa fa-expand\"></i> icon in the code editor toolbar to open the editor in an external window. That\'s especially useful if you have multiple screens.</p><p><b>Code view and Page view are in sync</b></p><p>When you select an element on the page it gets highlighted in the code. When you make a change on the page the code gets updated.</p><p><b>Right click</b> on the element in the code to select the element on the page.</p><p>Use <b>Page -&gt; Refresh (CMD + R)</b> to refresh the view if it gets messed up during editing. No need to save the page first.</p>", "A note about code view", "code-view-4");
            edit_code_msg_shown = true;
        }

        if(!isLiveRefresh()) {
            crsaQuickMessage("Auto refresh is disabled. Press <b>CMD + R</b> to refresh.", 3000);
        }

        if(code_mode == 'html') {
            var selected = pinegrow.getSelectedElement();
            if(selected) {
                this.selectElementInEditor(selected.get$DOMElement());
            }
        }

        script_changes_msg_shown = false;
    }
}
