var PagePgStylesheet = function (page, pgStylesheet, parentPgStylesheet, compiledStylesheet) {
    this.page = page;
    this.pgStylesheet = pgStylesheet;
    this.parentPgStylesheet = parentPgStylesheet;
    this.compiledStylesheet = compiledStylesheet;
}

var CrsaPageStyles = function(cp) {
    //this.$iframe = $(iframe);
    var stylesheetKeyObj = {};
    this.stylesheets = [];
    this.default_ignore = [/fonts\.googl/i];
    this.default_ignore_checked = false;
    this.framework_ignore_checked = {};

    var active_rules = null;
    var _this = this;

    this.getStylesheetInfo = function (cs) {
        if (stylesheetKeyObj[cs.url]) {
            return {
                parentPgStylesheet: stylesheetKeyObj[cs.url].parentPgStylesheet,
                compiledStylesheet: stylesheetKeyObj[cs.url].compiledStylesheet,
            };
        }
        return null;
    }

    this.isStylesheetUsed = function (cs, newCs) {
        if (cs) {
            return stylesheetKeyObj[cs.url] ? true : false;
        }
        else {
            return this.has(newCs) ? true : false;
        }
    }

    this.addStylesheetAs = function (csObj, parentCS, compiledCs) {
        // TODO: fix adding same stylesheet more than once
        // var cs = stylesheetKeyObj[csObj.url];
        // if (cs) {
        //     pinegrow.showQuickError('This file is already imported, this will couse some issues');
        //     if (cs != csObj.pgStylesheet) console.log('Same URL with diffrent cs, Shouldn\'t happen');
        //     return;
        // }

        stylesheetKeyObj[csObj.url] = new PagePgStylesheet(cp, csObj.pgStylesheet, parentCS, compiledCs);
    }

    this.updateStylesheetUrl = function (oldUrl, newUrl, cs) {
        if (stylesheetKeyObj[oldUrl]) {
            if (stylesheetKeyObj[oldUrl].pgStylesheet == cs) {
                stylesheetKeyObj[newUrl] = stylesheetKeyObj[oldUrl];
                delete stylesheetKeyObj[oldUrl];
            } else {
                throw "Error: same url with different stylesheet";
            }
        }
    }

    this.removeUsedStylesheet = function () {
        stylesheetKeyObj = {};
    }

    this.getByRegExp = function(re) {
        for(var i = 0; i < this.stylesheets.length; i++) {
            if(this.stylesheets[i].url && this.stylesheets[i].url.match(re)) {
                return this.stylesheets[i];
            }
        }
        return null;
    }

    this.getSourceStylesheets = function() {
        /*
        return this.stylesheets.map(function(cs) {
            return cs.sourceStylesheet ? cs.sourceStylesheet : cs;
        })
        */
        var list = [];
        for(var i = 0; i < this.stylesheets.length; i++) {
            var cs = this.stylesheets[i];
            if(cs.sourceStylesheet) {
                cs = cs.sourceStylesheet;
            }
            list.push(cs);
            cs.getImportedStylesheets(list);
        }
        return list;
    }

    this.getList = function(options) {
        return pgcssh.filterStylesheetsList(this.getSourceStylesheets(), options);
    }

    var getIgnoreFiles = function() {
        var r = _this.default_ignore;
        $.each(cp.frameworks, function(i,f) {
            r = r.concat(f.ignore_css_files);
        });
        return r;
    }


    var getOne$Iframe = function() {
        var all = cp.getPageView$Iframes();
        if(all.length) {
            return all[0].get$Iframe();
        }
        return null;
    }

    this.destroy = function() {
        this.stylesheets = [];
        stylesheetKeyObj = {};
        pinegrow.removeEventHandler('on_page_frameworks_changed', onFrameworksChanged);
    }

    this.attachTo = function(cs, pgLink, done, replacePgel, first_css) {
        var idx = this.stylesheets.indexOf(cs);
        if(idx >= 0) {
            if(done) done();
            return;
        }
        this.stylesheets.push(cs);
        cs.attachTo(cp, pgLink, function() {
            if(done) done();
        }, replacePgel, first_css)
        return;
        /*
        this.stylesheets.push(cs);
        cp.getPageView$Iframes().forEach(function($iframe) {
            cs.attachTo($iframe.get(0), sheet, done);
        })*/
    }

    this.has = function(cs) {
        return this.stylesheets.indexOf(cs) >= 0;
    }

    this.detachFrom = function(cs) {
        cs.detachFrom(cp);
        var idx = this.stylesheets.indexOf(cs);
        if (idx >= 0) {
            this.removeStylesheetByKey(cs);
            this.stylesheets.splice(idx, 1);
        }

        /*
        cp.getPageView$Iframes().forEach(function($iframe) {
            cs.detachFrom($iframe.get(0));
        })*/

    }

    this.removeStylesheetByKey = function (cs) {
        var removeStylesheet = function (currCS) {
            delete stylesheetKeyObj[currCS.url];

            currCS.forEachStylesheet(function (stylesheet) {
                removeStylesheet(stylesheet);
            }, true);

            var sourceStylesheet = currCS.sourceStylesheet;
            if (sourceStylesheet) removeStylesheet(sourceStylesheet);
        }

        removeStylesheet(cs);
    }

    this.removeAllInlineSheets = function() {
        for (var i = 0; i < this.stylesheets.length; i++) {
            var cs = this.stylesheets[i];
            if (cs.inline) {
                this.stylesheets.splice(i,1);
                i--;
                pinegrow.stylesheets.removeCrsaStylesheet(cs);
                cs.unload();
            }
        }
    }

    this.removeAllExclusiveSheets = function() {
        stylesheetKeyObj = {};
        for (var i = 0; i < this.stylesheets.length; i++) {
            var cs = this.stylesheets[i];
            if (pinegrow.stylesheets.getAllSheetsForPages(cs).length <= 1) {
                this.stylesheets.splice(i,1);
                i--;
                pinegrow.stylesheets.removeCrsaStylesheet(cs);
                cs.unload();
            }
        }
    }

    this.updateIgnoreStatuses = function() {

        var check = function(f) {
            var list = null;
            if(!f) {
                if(!_this.default_ignore_checked) {
                    list = _this.default_ignore;
                    _this.default_ignore_checked = true;
                }
            } else {
                if(!_this.framework_ignore_checked[f.key]) {
                    list = f.lock_css_files || f.ignore_css_files;
                    _this.framework_ignore_checked[f.key] = true;
                }
            }
            if(list && list.length) {
                _this.stylesheets.forEach(function (cs) {
                    for (var n = 0; n < list.length; n++) {
                        if (cs.url && cs.url.match(list[n])) {
                            cs.setIgnore(true);
                            cs.setLocked(true, 'Stylesheet is a framework resource.', true);
                            break;
                        }
                    }
                })
            }
        }

        check(); //check default
        cp.frameworks.forEach(function(f) { //check frameworks
            check(f);
        })
    }

    this.loadAllStylesheets = function($iframe, done, include_dynamic) {

        var onDone = function() {
            if(done) done();

            pinegrow.dispatchEvent('on_page_stylesheets_loaded', cp, $iframe);
        }

        if(!pgf.edit_css) {
            onDone();
            return;
        }

        var count = 0;
        var done_timeout = null;
        var doc = $iframe.get(0).contentDocument || $iframe.get(0).contentWindow.document;

        var loading_list = [];

        for (var i = 0; i < doc.styleSheets.length; i++) {
            var sheet = doc.styleSheets[i];
            var ignore = false;
            var pgOwnerNode = null;

            if (sheet.ownerNode) {
                if (sheet.ownerNode.id == 'crsa-inline-styles') {
                    continue;
                }
                pgOwnerNode = getElementPgNode($(sheet.ownerNode), true /* dont create dyn elements */);

                if (!include_dynamic && !pgOwnerNode) {
                    continue;
                }

            }

            var sheet_url = sheet.href;

            var data_url = sheet.ownerNode.getAttribute('data-pg-style-url');

            if(data_url) {
                sheet_url = data_url;
            }

            var ignoreFiles = getIgnoreFiles();
            for (var n = 0; n < ignoreFiles.length; n++) {
                if (sheet_url && sheet_url.match(ignoreFiles[n])) {
                    ignore = true;
                    break;
                }
            }

            if (sheet_url) {
                if (sheet_url.indexOf('data:') === 0) continue;

                var originalUrl = pinegrow.getOriginalUrl(sheet_url);
                var cs = pinegrow.stylesheets.getStylesheetByKey(originalUrl);
                if (cs) {
                    console.log('PS: ss found, reusing ' + cs.name);
                    this.attachTo(cs, pgOwnerNode);
                    pinegrow.stylesheets.addCrsaStylesheet(cs);

                    pinegrow.stylesheets.stylsheetUsed({
                        url: originalUrl,
                        pgStylesheet: cs
                    });

                    var sourceStylesheet = cs.sourceStylesheet;
                    if (sourceStylesheet) {
                        pinegrow.stylesheets.stylsheetUsed({
                            url: sourceStylesheet.url,
                            pgStylesheet: sourceStylesheet
                        }, 'source', cs);
                    }
                } else if(isApp() || pgOwnerNode) {
                    console.log('PS: loading ' + sheet_url);

                    cs = PgCSSStylesheet();

                    this.attachTo(cs, pgOwnerNode);
                    pinegrow.stylesheets.addCrsaStylesheet(cs);

                    cs.ignore = ignore;

                    var loadFile = function (cs, href, loadDone) {
                        //console.log('loading ' + cs.url)
                        //loading_list.push(href);

                        cs.load(href, function (cs) {
                            if (loadDone) loadDone(cs);
                            count--;
                            console.log('done ' + cs.url + ' ' + count);

                            /*
                            var idx = loading_list.indexOf(pinegrow.getProxyUrl(cs.url));
                            if(idx >= 0) {
                                loading_list.splice(idx, 1);
                            }
                            console.log('Remaining ', loading_list)
                            */

                            if (count == 0) {
                                onDone();
                                count--;
                            }
                        }, true /* force loading */)
                    }

                    var loadAllFiles = function (csList) {
                        for (var i = 0; i < csList.length; i++) {
                            count++;
                            var cs = csList[i].cs;
                            loadFile(cs, csList[i].href, function (cs) {
                                // if (!cs.getError()) {
                                //     var newList = cs.getImportedFiles();
                                //     if (newList.length > 0) loadAllFiles(newList);
                                // }
                            });
                        }
                    }

                    pinegrow.stylesheets.stylsheetUsed({
                        url: originalUrl,
                        pgStylesheet: cs
                    });
                    loadAllFiles([{cs: cs, href: sheet_url}]);
                }
            } else {
                if(isApp() || pgOwnerNode) {
                    console.log('loading inline css');
                    count++;

                    cs = PgCSSStylesheet();
                    this.attachTo(cs, pgOwnerNode);
                    pinegrow.stylesheets.addCrsaStylesheet(cs);
                    cs.loadInline(sheet, function () {
                        count--;
                        if (count == 0) {
                            onDone();
                            count--;
                        }
                    });
                }
            }

        }
        if (count == 0) onDone();
    }

    this.getAllActiveRules = function(skip_list) {
        if (active_rules == null || true) {
            active_rules = [];
            for (var i = 0; i < this.stylesheets.length; i++) {
                var cs = this.stylesheets[i];
                if (skip_list && skip_list.indexOf(cs) >= 0) continue;
                if (cs.isEnabledForPage(cp)) {
                    active_rules = active_rules.concat(cs.getAllRules());
                }
            }
        }
        return active_rules;
    }

    this.getAllCrsaStylesheets = function() {
        return this.stylesheets;
    }

    this.reorder = function(crsa_styles_list, done) {
        if (crsa_styles_list.length > 0) {
            var c = 0;

            for (var n = 0; n < crsa_styles_list.length; n++) {
                var cs = crsa_styles_list[n];

                cs.detachFrom(cp);

                c++;

                cs.attachTo(cp, null, function () {
                    c--;
                    if (c <= 0) {
                        this.stylesheets = crsa_styles_list;
                        if (done) done();
                    }
                });
            }

        } else {
            if (done) done();
        }
    }

    var onFrameworksChanged = function(page) {
        if(page == cp) {
            _this.updateIgnoreStatuses();
        }
    }

    pinegrow.addEventHandler('on_page_frameworks_changed', onFrameworksChanged);
}