var CrsaSelectBox = function ($container, options) {
    var SELECT_ITEM_CLASS = 'selected-item';

    var enabled = false; //Disabled by MT, due to issues in multi win

    var _this = this;
    var oldSelectBoxPos = {left: 0, top: 0};

    var dragged_cpyel = null,
        dragged_$liTarget = null,
        dragged_$target = null;

    var containerW = parseInt($container.css('width'), 10),
        containerH = parseInt($container.css('height'), 10);

    var dragging_mode_active = false;
    var mouse_down = false;
    var $selectBox = $('<div class="select-box"></div>').appendTo($('body'));

    var $dummy_input = $('#crsa-dummy-field');
    var selectedItemsWithBox = [];

    var doObjectsCollide = function ($item, $selection) {
        var aTop = $item.offset().top;
        var aLeft = $item.offset().left;
        var aH = parseInt($item.css('height'), 10);

        var bTop = $selection.offset().top;
        var bLeft = $selection.offset().left;
        var bH = parseInt($selection.css('height'), 10);

        if ((aTop < bTop && aTop + aH > bTop) ||
            (aTop >= bTop && aTop < bTop + bH)) return true;
        else return false;
    }

    var clearAllSelection = function () {
        $container.find('.' + SELECT_ITEM_CLASS).removeClass(SELECT_ITEM_CLASS);
    }

    var clearItemSelectionByBox = function () {
        for (var i = 0; i < selectedItemsWithBox.length; i++) {
            selectedItemsWithBox[i].removeClass(SELECT_ITEM_CLASS);
        }
        selectedItemsWithBox = [];
    }

    var selectItemsInBetween = function ($frListItem, $lsListItem, dir) {
        var $item = $frListItem;
        clearItemSelectionByBox();

        var selectAllElm = function (getElementCallback, getContainerCallback, childPos, forceIgnoreUpdateOverParent) {
            var getItemOverParent = function ($item) {
                while(true) {
                    var $overParent = $item.closest('.' + options.overParentClassName);
                    var $nextParent = getElementCallback($overParent);

                    if ($nextParent.length > 0) {
                        var $childParent = $nextParent.find(options.selectorToFindChild + ':' + childPos + '-child');
                        if ($childParent.length > 0) {
                            return $childParent;
                        }
                        $item = $nextParent;
                    }
                    else {
                        return [];
                    }
                }
            }

            var containerJustPicked = false;
            if ($item.is($container)) {
                var $arr = $container.find('> .' + options.overParentClassName);
                $item = getContainerCallback($arr);
                if (!$item || $item.length == 0) return;
                containerJustPicked = true;
            }

            if (!$item.hasClass(options.itemClass)) {
                var $overParent = $item.closest('.' + options.overParentClassName);
                if (!forceIgnoreUpdateOverParent) {
                    if (!containerJustPicked) $overParent = getElementCallback($overParent);
                }

                if ($overParent.length > 0) {
                    $item = $overParent.find(options.selectorToFindChild + ':' + childPos + '-child');
                }
            }

            var breakNextTime = false;
            while ($item.length > 0) {
                if ($item.hasClass(options.itemClass)) {
                    if (doObjectsCollide($item, $selectBox))  {
                        selectedItemsWithBox.push($item);
                        var $parentItem = $item.parent().closest('.' + options.itemClass);

                        if (!$parentItem.hasClass(SELECT_ITEM_CLASS)) {
                            $item.addClass(SELECT_ITEM_CLASS);
                        }
                    }
                    else break;
                }
                var $prevItem = getElementCallback($item);

                if ($prevItem.length == 0) {
                    $prevItem = getItemOverParent($item);
                }
                $item = $prevItem;

                if (breakNextTime) break;
                if ($item[0] == $lsListItem[0]) breakNextTime = true;
            }
        }

        if (dir == "up") {
            selectAllElm(function ($el) {
                return $el.prev();
            }, function ($arr) {
                for (var i = $arr.length - 1; i >= 0; i--) {
                    var $p = $($arr[i]);
                    if (doObjectsCollide($p, $selectBox)) {
                        return $p;
                    }
                }
            }, 'last');
        }
        else if (dir == "down") {
            selectAllElm(function ($el) {
                return $el.next();
            }, function ($arr) {
                for (var i = 0; i < $arr.length; i++) {
                    var $p = $($arr[i]);
                    if (doObjectsCollide($p, $selectBox)) {
                        return $p;
                    }
                }
            }, 'first', true);
        }
    }

    var onPageChangeing = function(e) {
        if (onPageChangeing == null) return;

        clearAllSelection();
    }

    var mousemove = function(event) {
        if (mouse_down) {
            _this.startSelect();
            if (dragging_mode_active) {
                var w = Math.abs(oldSelectBoxPos.left - event.pageX);
                var h = Math.abs(oldSelectBoxPos.top - event.pageY);

                $selectBox.css({
                    'width': w,
                    'height': h
                });
                if (event.pageX <= oldSelectBoxPos.left && event.pageY >= oldSelectBoxPos.top) {
                    $selectBox.css({
                        'left': event.pageX
                    });
                } else if (event.pageY <= oldSelectBoxPos.top && event.pageX >= oldSelectBoxPos.left) {
                    $selectBox.css({
                        'top': event.pageY
                    });
                } else if (event.pageY < oldSelectBoxPos.top && event.pageX < oldSelectBoxPos.left) {
                    $selectBox.css({
                        'left': event.pageX,
                        "top": event.pageY
                    });
                }

                var dir = '';
                if (oldSelectBoxPos.top > event.pageY) dir = 'up';
                else if (oldSelectBoxPos.top < event.pageY) dir = 'down';

                var $lsListItem = $(event.target).closest('li');
                if (!$lsListItem.hasClass(options.itemClass)) $lsListItem = [];

                if (!event.metaKey && !event.shiftKey) {
                    clearAllSelection();
                }

                selectItemsInBetween(dragged_$liTarget, $lsListItem, dir);
                return false;
            }
        }
    }

    var mousedown = function(event) {
        if (event.which == 3) return;

        dragged_$target = $(event.target);
        if (options.isAllowed) {
            if (!options.isAllowed(dragged_$target)) {
                return;
            }
        }

        mouse_down = true;

        if (!dragged_$target.is($container)) {
            dragged_$liTarget = dragged_$target.closest('li');
        }
        else {
            dragged_$liTarget = dragged_$target;
        }
        oldSelectBoxPos = {
            left: event.pageX,
            top: event.pageY
        }
        $selectBox.css(oldSelectBoxPos);

        containerW = parseInt($container.css('width'), 10);
        containerH = parseInt($container.css('height'), 10);

        if (event.metaKey || event.shiftKey) {
            var $target = $(event.target);
            var $li = $target.closest('li');
            if ($li.hasClass(options.itemClass)) {
                if ($li.hasClass(SELECT_ITEM_CLASS)) {
                    $li.removeClass(SELECT_ITEM_CLASS);
                }
                else {
                    var $parentItem = $li.parent().closest('.' + options.itemClass);

                    if (!$parentItem.hasClass(SELECT_ITEM_CLASS)) {
                        $li.addClass(SELECT_ITEM_CLASS);
                    }
                }
            }
        }
        else {
            clearAllSelection();
        }
    }

    var mouseup = function(event) {
        selectedItemsWithBox = [];
        mouse_down = false;

        if (dragging_mode_active) {
            $selectBox.css({
                display: 'none',
                width: 0,
                height: 0
            });
            dragging_mode_active = false;

            return false;
        }
    }

    var eventDragStart = function(event) {
        if (options.isDraggableElm) {
            if (!options.isDraggableElm(dragged_$target)) {
                _this.startSelect();
                dragged_cpyel = this.cloneNode(true);
                dragged_cpyel.style.visibility = "hidden";
                document.body.appendChild(dragged_cpyel);
                event.dataTransfer.setDragImage(dragged_cpyel, 0, 0);
                event.dataTransfer.effectAllowed = "none";
            }
        }
    }

    var eventDragEnd = function(event) {
        mouseup(event);

        if (dragged_cpyel) {
            if (options.isDraggableElm) {
                document.body.removeChild(dragged_cpyel);
                dragged_cpyel = null;
                dragged_$target = null;
            }
        }
    }


    if(enabled) {
        $container.get(0).addEventListener('mousedown', mousedown);

        document.body.addEventListener('mousemove', mousemove);

        document.body.addEventListener('mouseup', mouseup);

        $container.get(0).addEventListener('dragstart', eventDragStart, true);

        $container.get(0).addEventListener('dragover', mousemove, true);

        $container.get(0).addEventListener('dragend', eventDragEnd, true);
        $('body').on('crsa-page-changing', onPageChangeing);
    }




    this.remove = function () {
        if(enabled) {
            document.body.removeEventListener('mousemove', mousemove);
            document.body.removeEventListener('mouseup', mouseup);

            $container.get(0).removeEventListener('mousedown', mousedown);
            $container.get(0).removeEventListener('dragstart', eventDragStart, true);
            $container.get(0).removeEventListener('dragover', mousemove, true);
            $container.get(0).removeEventListener('dragend', eventDragEnd, true);

            $('body').off('crsa-page-changing', onPageChangeing);
        }

        onPageChangeing = null;
    }

    this.startSelect = function () {
        dragging_mode_active = true;
        $selectBox.show();
    }
}