/**
 * Created by Matjaz on 12/6/13.
 */

var path = require('path');
var fs = require('fs');

var cloneObject = function(obj) {
    return $.extend({}, obj);
}

var extendObject = function( subClass, superClass ) {
    subClass.prototype = Object.create( superClass.prototype );
    subClass.prototype.contructor = subClass;
}


var pgSplitKeyEqualsValueString = function(str) {
    var r = [];
    str.split("\n").forEach(function(line) {
        line = $.trim(line);
        if(line.length) {
            var a = line.split('=');
            if (a.length === 2) {
                r.push({one: a[0].trim(), two: a[1].trim()});
            } else if(a.length === 1) {
                r.push({one: a[0].trim(), two: a[0].trim()});
            }
        }
    })
    return r;
}

var pgMakeSlug = function(name, allow_dot, space_replace) {
    if(!name) return name;

    name = name.toLowerCase();
    if(space_replace === undefined) space_replace = '_';

    if(allow_dot) {
        name = name.replace(/[^\w \.]+/g,'')
    } else {
        name = name.replace(/[^\w ]+/g,'')
    }
    return name.replace(/ +/g,space_replace);
}


function pgGetHttpRequest(url, done, fail) {
    var r = require('request');
    var page = pinegrow.getSelectedPage();

    r.get({
        url: url, encoding: null, headers: {
            'User-Agent': navigator.userAgent,
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,text/css,image/webp,*/*;q=0.8',
            'Accept-Encoding': 'identity',
            'Accept-Language': 'en-US,en;q=0.8',
            'Cache-Control': 'max-age=0',
            'Connection': 'close',
            'Cookie': '',
            'Referer': page ? page.url : ''
        }
    }, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            done(body)
        } else {
            fail();
        }
    });
}

function pgGetCookieByName(name) {
    const value = `; ${document.cookie}`;
    const parts = value.split(`; ${name}=`);
    if (parts.length === 2) return parts.pop().split(';').shift();
}

function pgSetCookie(name, value, ms) {
    var expires = "";
    if (ms) {
        var date = new Date();
        date.setTime(date.getTime() + (ms));
        expires = "; expires=" + date.toUTCString();
    }
    document.cookie = name + "=" + (value || "")  + expires + "; path=/";
}

function pgRemoveCookie(name) {
    document.cookie = name +'=; Path=/; Expires=Thu, 01 Jan 1970 00:00:01 GMT;';
}

var CrsaProfile = function(enab, str) {
    str = str || '';
    var enabled = pgIsDev();//false;//true;
    var start_ms;
    var elapsed = 0;
    var paused = false;
    var last = 0;

    if(enabled) {
        start_ms = (new Date()).getTime();
        last = 0;
    }

    this.pause = function() {
        elapsed += (new Date()).getTime() - start_ms;
        paused = true;
    }

    this.resume = function() {
        start_ms = (new Date()).getTime();
        paused = false;
    }

    this.show = function(name) {
        if(enabled) {
            var elapsed_ms = paused ? elapsed : elapsed + (new Date()).getTime() - start_ms;
            console.log(str + name + ' took ' + elapsed_ms + ' ms, ' + (elapsed_ms - last) + 'ms this step');
            last = elapsed_ms;
        }
    }
}

var pgDummyTextarea = document.createElement("textarea");

function pgDecodeHtml(html) {
    pgDummyTextarea.innerHTML = html;
    return pgDummyTextarea.value;
}

function pgUrlToDisplayString(url) {
    return decodeURI(url);
}

function pgAddProxyLinksToStyleAttribute(str) {
    if(!str) return str;

    var project = pinegrow.getCurrentProject();

    return str.replace(/url\('([^']*)'\)/g, function(m, a, b, c) {
        var url = a;
        if(!crsaIsRemoteUrl(a) && !a.startsWith('data:')) {
            url = pinegrow.getProxyUrl(a);
        }
        if(project) {
            url = project.mapSourceUrlToDOMUrl(url);
        }
        return 'url(\'' + url + '\')';
    })
}

function pgRemoveProxyLinksFromStyleAttribute(str) {
    if(!str) return str;

    var project = pinegrow.getCurrentProject();

    return str.replace(/url\('([^']*)'\)/g, function(m, a, b, c) {

        if(project) {
            a = project.mapDOMUrlToSourceUrl(a);
        }

        a = pinegrow.getOriginalUrl(a);

        return 'url(\'' + a + '\')';
    })
}

function pgAddProxyLinksToAttr(attr, str) {
    switch(attr) {
        case 'style':
            return pgAddProxyLinksToStyleAttribute(str);
        case 'href':
        case 'src':
            return pinegrow.getProxyUrl(str);
    }
    return str;
}

function crsaEvaluateShowIfString(str, values, fdef, obj) {
    var ands = str.split('&&');
    if(ands.length > 1) {
        var r = true;
        ands.forEach(function(ex) {
            r = r && crsaEvaluateShowIfString(ex, values);
        })
        if(fdef && fdef.on_evaluate_show_if) {
            r = fdef.on_evaluate_show_if(obj, r, values, fdef);
        }
        return r;
    }
    var not = str.startsWith('!');
    if(not) {
        str = str.substr(1);
    }
    var r = false;

    var a = str.split('==');
    if(a.length == 1) {
        r = values[a[0]] || false;
    } else if(a.length == 2) {
        var tokens = a[1].split('||');
        r = tokens.indexOf(values[a[0]] || '' ) >= 0;
    }
    if(not) r = !r;
    if(fdef && fdef.on_evaluate_show_if) {
        r = fdef.on_evaluate_show_if(obj, r, values, fdef);
    }
    return r;
}


function crsaRemoveScripts(str, replace_with) {
    if(str && str.indexOf('<?') >= 0) {
        str = str.replace(/<\?.*?\?>/g, replace_with || '');
    }
    return str;
}

function crsaPopulateSectionClosedStatus(sec_def) {
    if(!sec_def.hasOwnProperty('closed')) {
        var k = 'sec_open_' + sec_def.key;
        if(localStorage[k]) {
            sec_def.closed = localStorage[k] != '1'; //not open
        } else if('default_closed' in sec_def) {
            sec_def.closed = sec_def.default_closed;
        } else {
            sec_def.closed = true;
        }
    }
}


function pgParseCSSValue(val) {
    let r = {
        type: 'value',
        calc: null,
        variable: null,
        value: null,
        unit: null,
        important: null,
        unknown: null,
        original: val,
        original_without_important: val
    }

    if(val !== null && val !== '') {
        if(typeof val !== 'string') val = val.toString();

        val = val.trim();

        if(val.endsWith('!important')) {
            r.important = true;
            val = val.replace('!important', '').trim();
            r.original_without_important = val;
        }

        if(val.startsWith('calc(')) {
            r.calc = val;
            r.type = 'calc';
        } else {
            var m = val.match(/^var\((\-\-.*)\)$/);
            if(m) {
                r.variable = m[1];
                r.type = 'var';
            } else {
                var m = val.match(/^\-\-[a-z0-9\-]+$/i);
                if(m) {
                    r.variable = val;
                    r.type = val;
                } else {
                    var info = crsaSplitCSSValue(val, true);
                    if (info) {
                        r.value = info.value;
                        r.unit = info.unit;
                        r.type = 'value';
                    } else {
                        r.unknown = r.original;
                        r.type = 'unknown';
                    }
                }
            }
        }
    }

    return r;
}

function pgGetCSSValueFromParsedInfo(r) {
    //if(r.unknown) return r.original;

    var val = '';

    if(r.calc) {
        val = r.calc;
    } else if(r.variable) {
        val = `var(${r.variable})`;
    } else if(r.value !== null) {
        val = `${r.value}${r.unit}`;
    } else {
        val = r.original_without_important;
    }
    if(r.important) {
        val += ' !important';
    }
    return val;
}

function crsaSplitCSSValue(v, no_add_px) {
    if(v === null) return v;
    if(typeof v !== 'string') v += no_add_px ? '' : 'px';

    var m = v.match(/^(\-?[0-9\.]+)([a-z\%]*)$/);
    return m ? {value: parseFloat(m[1]), unit: m[2].toLowerCase() || (no_add_px ? '' : 'px')} : null;
}

function crsaOneUpCSSValue(cv, down) {
    if(typeof down == 'undefined') down = 1;
    var breakpoints_units = {
        'px': 1,
        'em': 0.0625,
        'rem': 0.0625,
        'pt': 1
    }
    if(cv) {
        var a = cv.unit in breakpoints_units ? breakpoints_units[cv.unit] : 1;
        return {value: cv.value + down * a, unit: cv.unit};
    }
    return cv;
}

var crsaConvertCSSValueToPx_el = null;

function crsaConvertCSSValueToPx(cv, $iframe, $el, prop) {
    var s = typeof cv != 'object' ? cv + "" : (cv.value + cv.unit);
    if(s.match(/^[0-9\.]*$/)) s += 'px';

    var d;
    var $d;
    var orig_style = -1;
    if($el) {
        d = $el.get(0);

        //paddings, margins interfere when setting % sizes
        /*
        orig_style = $el.attr('style');
        $el.css({
            margin: '0!important',
            padding: '0!important'
        })

         */

    } else if(!$iframe) {
        if(!crsaConvertCSSValueToPx_el) {
            crsaConvertCSSValueToPx_el = $('#crsa-dummy-div').get(0);
        }
        d = crsaConvertCSSValueToPx_el;
    } else {
        var body = $(getIframeBody($iframe.get(0)));
        $d = body.find('#crsa-dummy-div');
        if($d.length == 0) {
            $d = $('<div id="crsa-dummy-div" style="position:absolute;top:-1000px;"></div>');
            body.append($d);
        }
        d = $d.get(0);
    }

    var px = Length.toPx(d, s, prop);
    if($iframe) $d.remove();
    if(orig_style !== -1) {
        if(!orig_style) {
            $el.get(0).removeAttribute('style');
        } else {
            $el.attr('style', orig_style);
        }
    }
    return px;
}

function pgConvertCSSUnit(value, origin_unit, dest_unit, $el, prop) {
    if(value == 0) return 0;

    var dest_unit_in_px = crsaConvertCSSValueToPx('1' + dest_unit, null, $el, prop);

    if(dest_unit_in_px === 0) {
        prop = null;
        dest_unit_in_px = crsaConvertCSSValueToPx('1' + dest_unit, null, $el, prop);
    }
    var origin_unit_in_px = crsaConvertCSSValueToPx('1' + origin_unit, null, $el, prop);

    //1cm = 20px
    //convert 10px -> cm

    //dest_px = 20px
    //orig_px = 1px

    var dest_value = value * (origin_unit_in_px / dest_unit_in_px);

    dest_value = Math.round((dest_value + Number.EPSILON) * 1000) / 1000;

    if(dest_unit === 'px') {
        dest_value = Math.round(dest_value);
    }
    /*
        var rounded = Math.round(dest_value);
        if(Math.abs(rounded - dest_value) < 0.002) {
            dest_value = rounded;
        }
    */
    //10 * (1/20)
    return dest_value;
}

function pgByteArrayToString(bytes) {
    var CHUNK_SIZE = 8*1024;
    if (bytes.length <= CHUNK_SIZE)
        return String.fromCharCode.apply(null, bytes);
    var str = '';
    for (var i = 0; i < bytes.length; i += CHUNK_SIZE)
        str += String.fromCharCode.apply(null, bytes.slice(i, i+CHUNK_SIZE));
    return str;
}

function pgRevealElement($el, done) {
    if($el.length === 0) {
        done && done();
        return;
    }
    var $body = $el.closest('body');

    var pos = $el.get(0).getBoundingClientRect();
    var $r = $('<div class="pg-docs-revealer"></div>').appendTo($body);

    $body.addClass('pg-docs-revealer-animating');

    var bw = $body.width();
    var bh = $body.height();
    var bbw = Math.max(bw, bh) * 0.8;

    $r.css({
        width: bbw + 'px',
        height: bbw + 'px',
        top: (bh/2 - bbw/2) + 'px',
        left: (bw/2 - bbw/2) + 'px'
    }).addClass('animate');

    var w = Math.min(pos.width, pos.height);
    var x = pos.left + pos.width/2 - w/2;
    var y = pos.top + pos.height/2 - w/2;

    $r.get(0).addEventListener('transitionend', function(e) {
        $r.remove();
        $body.removeClass('pg-docs-revealer-animating');
        done && done();
    })

    setTimeout(function() {
        $r.css({
            left: x + 'px',
            top: y + 'px',
            width: w + 'px',
            height: w + 'px',
            opacity: 0.7
        })

    }, 10)
}

function pgAreArraysEqual(a, b) {
    if(a.length !== b.length) return false;
    for(var i = 0; i < a.length; i++) {
        if(a[i] !== b[i]) return false;
    }
    return true;
}

function diffStrings(s1, s2) {
    //if(s1.length != s2.length) return s1.length > s2.length ? s1.length : s2.length;
    for(var i = 0; i < s1.length; i++) {
        if(i >= s2.length) return i;
        if(s1.charAt(i) != s2.charAt(i)) return i;
    }
    return -1;
}

function describeDiffStrings(s1, s2) {
    var i = diffStrings(s1, s2);
    if(i < 0) {
        return 'Strings are EQUAL';
    } else {
        var si = i - 20;
        if(si < 0) si = 0;
        return 'Strings are DIFFERENT at ' + i + ' s1[' + s1.length + ']="' + escapeHtmlCode(s1.substr(si, 25)) + '" s2[' + s2.length + ']="' + escapeHtmlCode(s2.substr(si, 25)) + '"';
    }
}

function crsaGetInlineStylePropertyValue(style, prop, is_url) {
    if(style) {
        var re = new RegExp('(^|;\\s*)' + escapeRegExp(prop) + '\\:\\s*' + (is_url ? 'url\\(([^\\)]*)\\)' : '([^\\;]*)') + '\\;?')
        var m = style.match(re);
        if(m) {
            var url = is_url ? m[2].replace(/['"]/g, '') : m[2];
            return url;
        }
    }
    return null;
}

//'(^|;|\\W)' +


function crsaSetInlineStylePropertyValue(style, prop, value, is_url, add_important, safe_mode) {
    style = style || '';
    if(!style.startsWith(';')) style = ';' + style;
    if(!style.endsWith(';')) style += ';';

    //;color:red;
    var re = new RegExp( ';\\s*' + escapeRegExp(prop) + '\\:\\s*' + (is_url ? 'url\\(([^\\)]*)\\)' : '([^\\;]*)') + '\\;', 'g');

    if(value !== null && value !== '' && value !== undefined) {
        var css = prop + ':' + (is_url ? 'url(\'' + value + '\')' : value) + (add_important ? ' !important' : '') + ';';
        if(style.match(re)) {
            style = style.replace(re, ';' + css);
        } else {
            style += css;
        }
    } else {
        style = style.replace(re, ';');
    }
    if(style.startsWith(';')) style = style.substring(1);
    if(style === ';') style = '';

    return style.trim();
}

/*
if(pgIsDev()) {
    console.log('crsaSetInlineStylePropertyValue test');

    var style = `background-position: top;`;

}

 */

function pgTransformSrcsetUrls(srcset, func) {
    if(srcset !== null && srcset.length) {
        var items = [];
        srcset.split(',').forEach(function(item) {
            item = $.trim(item);
            var a = item.split(/\s+/);
            a[0] = func(a[0], a[1]);
            items.push(a.join(' '));
        })
        return items.join(', ');
    }
    return srcset;
}

function escapeRegExp( value ) {
    if(typeof value != 'string') value = String(value);
    return value.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g, "\\$&");
}

function escapeHtmlCode(t) {
    return typeof t === 'string' ? t.replace(/&/g,'&amp;').replace(/</g,'&lt;').replace(/>/g,'&gt;') : '';
}

var cssHelperLink = document.createElement('a');

var crsaAppendQueryToUrl = function(url, queries) {
    cssHelperLink.href = url;
    if(cssHelperLink.search) cssHelperLink.search += '&';
    cssHelperLink.search += queries.join('&');
    return cssHelperLink.href;
}

var pgDecodeUrl = function(url) {
    cssHelperLink.href = url;
    return cssHelperLink.href;
}
/*
 ['href','protocol','host','hostname','port','pathname','search','hash'].forEach(function(k) {
 console.log(k+':', a[k]);
 });
 */
var crsaMakeLinkRelativeTo = function(link, parent, ignore_too_many_backpaddles) {
    var re = /^[a-z:]*\/+/i;

    var m_link = link.match(re);
    if(!m_link) return link;
    var link_prot = m_link[0];

    var m_parent = parent.match(re);
    if(!m_parent) return link;
    var parent_prot = m_parent[0];

    if(link_prot.toLowerCase() != parent_prot.toLowerCase()) return link;

    var a = link.replace(re, '').split('/');
    var p = parent.replace(re, '').split('/');

    var i = 0;
    while(i < a.length && i < p.length && a[i] == p[i]) {
        i++;
    }
    if(i == 0) {
        //cant make relative
        return link;
    } else {
        a.splice(0, i);
        if(p.length > i+1) {
            for(var n = 0; n < p.length - i - 1; n++) {
                a.unshift('..');
            }
        }
        var rel = a.join('/');
        if(false && !ignore_too_many_backpaddles && rel.indexOf('../..') >= 0) {
            return link; //too many backpaddles
        }
        return rel;
    }
}

function crsaGetAbsoluteUrlForRelativeUrlAndParentUrl(relative_url, parent_url) {
    if(crsaIsAbsoluteUrl(relative_url)) return relative_url;
    return require('url').resolve(parent_url, relative_url);
}

function crsaMakeUrlAbsolute(url) {
    if(url.match(/^[a-z]*:\/\//i)) return url;
    return crsaMakeUrlFromFile(crsaGetAppDir() + url);
}

function crsaDuplicateName(orig_name, local_file, skip) {
    var a = orig_name.split('.');
    var i = a.length > 1 ? a.length - 2 : a.length-1;
    var c = 0;
    var found;
    var name;
    var localBase = local_file ? crsaGetFileDir(local_file) : null;
    var fs = isApp() ? require('fs') : null;

    do {
        a[i] = a[i].replace(/\([0-9]+\)$/i, '');
        a[i] = c == 0 ? a[i] : a[i] + '(' + c + ')';
        name = a.join('.');
        found = false;
        if(skip) {
            if(skip.indexOf(name) >= 0) found = true;
        }
        if(!found && fs && localBase) {
            var newFile = localBase + name;
            if(fs.existsSync(newFile)) {
                found = true;
            }
        }
        c++;
    }
    while(found);

    return name;
}


function crsaGetFileDir(file) {
    if(isApp()) {
        var path = require('path');
        var dir = path.dirname(file);
        if(dir.charAt(dir.length-1) != path.sep) dir += path.sep;
        return dir;
    }
    var a = file.split('/');
    var dir = '';
    if(a.length > 1) {
        a.splice(a.length-1,1);
        dir = a.join('/');
    }
    if(dir.charAt(dir.length-1) != '/') dir += '/';
    return dir;
}

function crsaGetAppDir() {
    if(isApp()) {
        return crsaGetFileDir(crsaMakeFileFromUrl(pgMakeFileUrl(window.location.href)));
    } else {
        return null;
    }
}

function crsaIsFileOrDir(path, fs) {
    if(!isApp()) return false;

    try {
        if(!fs) fs = require('fs');
        var stat = fs.statSync(path);
        if(stat.isFile()) return 'file';
        if(stat.isDirectory()) return 'dir';
        return null;
    } catch(err) {
        return null;
    }
}

function crsaGetDir(file, path, fs) {
    if(!path) path = require('path');
    if(crsaIsFileOrDir(file, fs) == 'dir') return file;
    return path.dirname(file);
}

function crsaCopyFileSync(fs, source, dest, filter_func, with_backup) {
    if(with_backup && !filter_func) {
        filter_func = function(str) {
            return str;
        }
    }
    if(!fs) fs = require('fs');
    if(filter_func) {
        var text = fs.readFileSync(source, {encoding: 'utf8'});
        text = filter_func(text);
        if(with_backup) {
            crsaWriteFileWithBackup(fs, dest, text, 'utf8')
        } else {
            fs.writeFileSync(dest, text, {encoding: 'utf8'});
        }
    } else {
        var data = fs.readFileSync(source);
        fs.writeFileSync(dest, data);
    }
}

function crsaCopyFile(fs, source, target, cb) {
    var cbCalled = false;

    var rd = fs.createReadStream(source);
    rd.on("error", function(err) {
        done(err);
    });
    var wr = fs.createWriteStream(target);
    wr.on("error", function(err) {
        done(err);
    });
    wr.on("close", function(ex) {
        done();
    });
    rd.pipe(wr);

    function done(err) {
        if (!cbCalled) {
            cb(err);
            cbCalled = true;
        }
    }
}

function crsaCopyFolder(source, dest, done, quick, overwrite_existing) {
    //var ncp = require('ncp').ncp;
    var path = require('path');
    var fs = require('fs');

    var fse = require('fs-extra')

    //ncp.limit = 4;

    fse.copy(source, dest, {filter: function(filename) {

        if(quick) {
            var rel_name = filename.replace(source, '');
            var dest_name = path.join(dest, rel_name);
            var dest_exists = false;
            try {
                var stat_dest = fs.statSync(dest_name);
                var stat_source = fs.statSync(filename);
                //exists
                dest_exists = stat_source.isFile() /* && stat_dest.mtime >= stat_source.mtime */;
            } catch(err) {}

            var oe = typeof overwrite_existing == 'function' ? overwrite_existing(rel_name) : overwrite_existing;

            //console.log('copy ' + filename);

            if(dest_exists && !oe) {
                //console.log('NCP SKIP ' + filename);

                return false;
            }
        }
        //console.log('NCP ' + filename);
        return true;
    }}, function (err) {
        if(done) done(err);
    });
}

function crsaCopyFolderSync(source, dest, quick, overwrite_existing) {
    //var ncp = require('ncp').ncp;
    var path = require('path');
    var fs = require('fs');

    var fse = require('fs-extra')

    //ncp.limit = 4;

    fse.copySync(source, dest, {filter: function(filename) {

            if(quick) {
                var rel_name = filename.replace(source, '');
                var dest_name = path.join(dest, rel_name);
                var dest_exists = false;
                try {
                    var stat_dest = fs.statSync(dest_name);
                    var stat_source = fs.statSync(filename);
                    //exists
                    dest_exists = stat_source.isFile() /* && stat_dest.mtime >= stat_source.mtime */;
                } catch(err) {}

                var oe = typeof overwrite_existing == 'function' ? overwrite_existing(rel_name) : overwrite_existing;

                //console.log('copy ' + filename);

                if(dest_exists && !oe) {
                    //console.log('NCP SKIP ' + filename);

                    return false;
                }
            }
            //console.log('NCP ' + filename);
            return true;
        }
    });
}

function crsaWriteFileWithBackup(fs, file, data, enc) {

    if(!fs) fs = require('fs');
    var path = require('path');

    if(crsaIsFileOrDir(file, fs) == 'file') {

        if(pgf.use_backups && pinegrow.getSetting('backup', '1') == '1') {
            var path = require('path');
            var dir = crsaGetFileDir(file) + '_pgbackup';

            var project = pinegrow.getCurrentProject();
            if(project && project.isFileInProject(file)) {
                var rel = project.getRelativePath(file);
                dir = path.dirname(path.join(project.getDir(), '_pgbackup', rel));
            }
            if(!crsaIsFileOrDir(dir, fs)) {
                //fs.mkdirSync(dir);
                crsaCreateDirs(dir, fs);
            }
            //dir += path.sep;
            var name = getPageName(file);
            var a = name.split('.');
            var mark = '_' + Math.floor((new Date()).getTime() / 1000);
            if(a.length > 1) {
                a[a.length-2] += mark;
            } else {
                a[0] += mark;
            }
            var bckfile = path.join(dir, a.join('.'));
            crsaCopyFileSync(fs, file, bckfile);

            fs.readdir(dir, function(err, files) {
                if(!err) {
                    var ext = path.extname(name);
                    var base = path.basename(name, ext);
                    var re = new RegExp('^' + escapeRegExp(base + '_') + '([0-9]+)' + escapeRegExp(ext) + '$');
                    var list = [];
                    for(var i = 0; i < files.length; i++) {
                        var m;
                        if(m = files[i].match(re)) {
                            list.push(parseInt(m[1]));
                        }
                    }
                    list.sort();
                    var max = parseInt(pinegrow.getSetting('backup-revisions', '20'));
                    if(!max) max = 9999999;

                    if(list.length > max) {

                        console.log('Deleting ' + (list.length - max) + ' ' + name + ' old backup files...');

                        var doDelete = function(err) {
                            if(!err && list.length > max) {
                                var file = path.join(dir, base + '_' + list.shift() + ext);
                                fs.unlink(file, doDelete);
                            }
                        }
                        doDelete();
                    }
                }
            })
        }
    } else {
        var paths = crsaGetFileDir(file);
        crsaCreateDirs(paths, fs);
    }
    //console.log('writing file ' + file);
    fs.writeFileSync(file, data, enc);
}

function crsaGetUrlParameters (url) {
    var a;
    var res = '';

    a = url.split('?');
    var params = a[1];

    a = a[0].split('#');
    var hashes = a[1];

    if (params) res  = '?' + params;
    if (hashes) res += '#' + hashes;

    return res;
}

function crsaRemoveUrlParameters(url) {
    var a = url.split(/[\?\#]/);
    return a.length > 0 ? a[0] : url;
}

function pgRemoveUrlParameter(url, param) {
    return url.replace(new RegExp(`\\??\\&?${escapeRegExp(param)}(=[^\\&\\#$\\n]*)?`, 'g'), '');
}

function crsaRemoveUrlHash(url) {
    return url.replace(/\#.*/, '');
}

function crsaGetUrlHash(url) {
    var m = url.match(/\#(.*)/);
    return m ? m[1] : null;
}

function crsaCreateDirs(path, fs) {
    if(crsaIsFileOrDir(path, fs)) return false;
    mkdirpsync(path);
    return true;
}

function crsaCreateFile(file) {
    var fs = require('fs');
    if(crsaIsFileOrDir(file, fs)) return false;

    return fs.writeFileSync(file, "");
}

function pgMakeFileUrl(url_or_file) {
    //console.log('url', url_or_file);

    if(url_or_file.startsWith('chrome-extension:') || url_or_file.startsWith('https://pglive') || url_or_file.startsWith('https://pinegrow.com/liveapp') /* support remote edit.html */) {
        url_or_file = url_or_file.replace(/chrome\-extension\:\/\/[^\/]*/, '');
        url_or_file = url_or_file.replace('https://pglivedev:8890', '');
        url_or_file = url_or_file.replace('https://pinegrow.com/liveapp', '');

        var cwd = process.cwd();
        var url = crsaMakeUrlFromFile(cwd) + url_or_file;
        return url;
    } else {
        if(url_or_file.startsWith('file:')) {
            return url_or_file;
        } else {
            return crsaMakeUrlFromFile(url_or_file);
        }
    }
}

function pgGetAppFilePath(file) {
    return crsaMakeFileFromUrl(pgMakeFileUrl(crsaGetBaseForUrl(window.location.href) + file));
}

function pgDeleteFileOrFolder(currentPath, just_content) {
    var path = require('path');
    var fs = require('fs');

    if(currentPath.length < 10 || currentPath.split(path.sep).length < 3) {
        throw 'Not sure it is a good idea to delete this folder.';
    }

    if (fs.lstatSync(currentPath).isDirectory()) {
        fs.readdirSync(currentPath).forEach(function (currnetFile, index) {
            var curPath = path.join(currentPath, currnetFile);
            if (fs.lstatSync(curPath).isDirectory()) { // recurse
                pgDeleteFileOrFolder(curPath);
            } else { // delete file
                fs.unlinkSync(curPath);
            }
        });
        if(!just_content) {
            fs.rmdirSync(currentPath);
        }
    }
    else {
        fs.unlinkSync(currentPath);
    }
}

function pgDeleteFolderContent(path) {
    try {
        pgDeleteFileOrFolder(path, true);
        return true;
    }
    catch(err) {
        pinegrow.showAlert(`<p>The folder ${path} could not be emptied.</p><p>${err}</p>`);
    }
    return false;
}



/*



 */

function crsaMakeUrlFromFile(file) {
    var f = 'file://';
    if(isApp()) {
        if(file.match(/^(file|http|https)+\:\/\//)) {
            return file; // already a url
        }
        var path = require('path');
        if(file.match(/^[a-z]\:/i)) {
            //win, c://ppp
            file = '/' + file
        } else if(file.startsWith('\\\\')) {
            //win, //sfp/aaa/aaa
            file = file.substr(2);
        }
    }

    var newURL = encodeURI(file.replace(/\\/g, "/")).replace(/\'/g, '%27');
    //    newURL = encodeURIComponent(newURL).replace(/%2F/g, "/");

    return f + newURL;
}

function crsaMakeFileFromUrl(url, skip_remove_params) {
    if (!url) return '';
    if(isApp()) {
        var path = require('path');
        var f = path.sep == '/' ? 'file://' : 'file://';

        if(!skip_remove_params) {
            url = crsaRemoveUrlParameters(url);
        }
        if(!url.startsWith(f)) {
            //relative url
            return path.normalize(decodeURI(url.replace(/\//g, path.sep)));
        }
        url = url.replace(f, '');
        if(path.sep == '\\') {
            //win stuff
            // wsl file path
            if (url.startsWith('wsl%24')) {
                url = url.replace('wsl%24', 'wsl$');
            }

            if(url.startsWith('/')) {
                //win, c://ppp
                url = url.substr(1);
            } else {
                //win, \\spf\sdsd
                url = '//' + url;
            }
        }
        var newPath = decodeURI(url.replace(/\//g, path.sep).replace(/\%27/g, '\''));
        //    newPath = decodeURIComponent(newPath);

        return path.normalize(newPath);
    } else {
        return crsaRemoveUrlParameters(url).replace('file://', '');
    }
}

function pgDebugCanvas(c) {
    c.setAttribute('style', 'position:absolute;top:00px;left:00px;z-index:100;');
    $('body').append(c);
    $(c).on('click', function() {
        $(this).remove();
    })
}

function pgDebugPos(pos) {
    var $c = $('<div/>');
    $c.attr('style', `position:absolute;top:${pos.top}px;left:${pos.left}px;width:${pos.width}px;height:${pos.height}px;z-index:100;background:rgba(255,0,0,0.5);`);
    $('body').append($c);
    $c.on('click', function() {
        $(this).remove();
    })
}

function pgGetColorWithEnoughContrast(c, a, b) {
    if(chroma.contrast(c, a) > 4.5) {
        return a;
    } else {
        return b;
    }
}

function crsaIsFileUrl(url) {
    return url && (url.indexOf('file:') == 0);
}

function crsaCleanUpUrl (url) {
    if(!url) return url;
    url = crsaRemoveUrlParameters(url);
    if (url[url.length-1] == '/')
        url = url.substr(0, url.length-1);

    return url;
}

function crsaGetExtFromUrl(url) {
    url = crsaCleanUpUrl(url);

    var name = crsaGetNameFromUrl(url);
    if (name.indexOf('.') > -1) {
        var nameArr = name.split('.');
        return nameArr[nameArr.length-1].toLowerCase();
    }
    return null;
}

function crsaGetNameFromUrl(url, def) {
    url = crsaRemoveUrlParameters(url);
    var s = url.split(/[\\\/]/);
    var n = s.pop();
    while(n.length == 0 && s.length) {
        n = s.pop();
    }
    if(n.length == 0) n = (def || '');
    var a = n.split('?');
    return a.length > 1 ? a[0] : n;
}

function crsaGetBaseForUrl(url) {
    if(!url) return url;
    url = crsaRemoveUrlParameters(url);
    var a = url.split(/[\/]/);
    a.pop();
    return a.join('/');
}

function crsaGeneralCombineUrlWith(url, doc) {
    if(doc.length == 0) return url;
    while (doc.startsWith('../')) {
      url = crsaGetBaseForUrl(url);
      doc = doc.substr(3, doc.length);
    }
    if(doc.charAt(0) != '/') {
        doc = '/' + doc;
    }
    return url + doc;
}

function crsaCombineCurrentUrlWith(url, doc) {
    return crsaGeneralCombineUrlWith(url, doc);
}

function crsaExtractDomain(url) {
    var domain, protocol;
    //find protocol (http, ftp, etc.) and get domain
    if (url.indexOf("://") > -1) {
        protocol = url.split('/')[0];
        domain = url.split('/')[2];
    }
    else {
        domain = url.split('/')[0];
    }

    //find & remove port number
    domain = domain.split(':')[0];

    return (protocol ? protocol + '//' : '' ) + domain;
}

function crsaCombineUrlWith(url, doc) {
    if(!url) return doc;
    if (crsaIsAbsoluteUrl(url) && doc[0] == '/') {
        url = crsaExtractDomain(url);
    }
    else {
        url = crsaGetBaseForUrl(url);
    }
    return crsaGeneralCombineUrlWith(url, doc);
}

function crsaGetUrlToSmartCombine(url, doc) {
    var newUrl = doc;

    var docBase = crsaGetBaseForUrl(newUrl);
    var urlBase = crsaGetBaseForUrl(url);
    while (true) {
        var docPosition = urlBase.indexOf(docBase);
        if (docPosition > -1) {
            if (urlBase.substr(docPosition, urlBase.length) == docBase) {
                doc = doc.substr(newUrl.length, doc.length);
                if (doc == "") doc = crsaGetNameFromUrl(newUrl);
                break;
            }
        }
        var docArr = newUrl.split('/');
        if (docArr.length == 1) break;
        docArr.pop();
        newUrl = docArr.join('/');
    }
    return doc;
}

function crsaIsAbsoluteUrl(url) {
    return !!url && (url.indexOf('://') >= 0 || url.indexOf('//') === 0);
}

function crsaGetNumberAndUnit(str) {
    var matched = str.match(/^([0-9\.]+)(.*)/);
    if (matched && matched.length > 1) {
        var number = matched[1];
        var unit = matched[2];

        if (number.startsWith('.')) {
            number = '0' + number;
        }

        return {
            text: number + unit,
            number: number,
            unit: unit
        }
    }

    return null;
}

function crsaIsRemoteUrl(url) {
    return url && crsaIsAbsoluteUrl(url) && url.startsWith('http');
}

async function pgFetchRemoteContent(url, options) {
    if(crsaIsFileUrl(url)) {
        url = pinegrow.getProxyUrl(url);
    }

    const response = await fetch(url, options);

    if (!response.ok) {
        const message = `An error has occured: ${response.status}`;
        throw new Error(message);
    }

    var data = await response.arrayBuffer();
    return data;
}

function pgCopyRemoteUrlToLocalPath(url, dest_path, done) {
    var fs = require('fs');
    var path = require('path');

    var saveData = function(data) {
        try {
            crsaCreateDirs(path.dirname(dest_path), fs);
            fs.writeFileSync(dest_path, data);
            console.log('Save remote file ' + url + ' -> ' + dest_path);
        } catch(err) {
            console.log('Could not save ' + dest_path, err);
        }
    }

    pgFetchRemoteContent(url).catch(function (error) {
        console.log('Error getting remote file ' + url, error);
        if (done) done();
    }).then(function (data) {
        try {
            if (data === undefined) throw 'Error getting remote file ' + url;
            saveData(new Buffer(data));
        } catch (err) {
            console.log('Error getting remote file ' + url, err);
        }
        if (done) done();
    });

/*
    try {
        var data = pgFetchRemoteContent(url);

        if (data) {
            saveData(new Buffer(data));
        } else {
            console.log('Error getting remote file ' + url);
        }
        if (done) done();
        return true;
    } catch(err) {
        console.log('Error getting remote file ' + url, err);
        if (done) done();
        return false;
    }*/

}

function crsaIsFileExist (file_path) {
    var fs = require('fs');
    return fs.existsSync(file_path);
}

function crsaRemoveExtFromUrl (fileUrl) {
    var filename = crsaGetNameFromUrl(fileUrl);
    var filenameArr = filename.split('.');
    if(filenameArr.length === 1) return filename;
    filenameArr.splice(filenameArr.length-1, 1);
    return filenameArr.join('.');
}

function crsaGetUrlWithoutExt (url) {
    return crsaGetBaseForUrl(url) + '/' + crsaRemoveExtFromUrl(url);
}

function crsaGetCaretPosition(editableDiv) {
    var caretStartPos = 0,
        caretEndPos = 0,
        sel, range;
    if (window.getSelection) {
        sel = window.getSelection();
        if (sel.rangeCount) {
            range = sel.getRangeAt(0);
            if (range.commonAncestorContainer.parentNode == editableDiv) {
                caretStartPos = range.startOffset;
                caretEndPos = range.endOffset;
            }
        }
    } else if (document.selection && document.selection.createRange) {
        range = document.selection.createRange();
        if (range.parentElement() == editableDiv) {
            var tempEl = document.createElement("span");
            editableDiv.insertBefore(tempEl, editableDiv.firstChild);
            var tempRange = range.duplicate();
            tempRange.moveToElementText(tempEl);
            tempRange.setEndPoint("EndToEnd", range);
            caretStartPos = caretEndPos = tempRange.text.length;
        }
    }
    return [caretStartPos, caretEndPos];
}

function crsaSetCaretPosition(editableDiv, start, end) {
    end = end || start;

    var textNode = editableDiv.firstChild;
    var range = document.createRange();
    range.setStart(textNode, start);
    range.setEnd(textNode, end);
    var sel = window.getSelection();
    sel.removeAllRanges();
    sel.addRange(range);
}

function crsaSplitAndTrimList(cls) {
    return (cls && cls.length) ? cls.split(',').map(function(cls_item) {
        return $.trim(cls_item);
    }) : [];
}

function crsaGetSummaryStr (str, n, front) {
    n = n || 40;
    if (str.length <= n) return str;
    if(front) {
        return str.substr(0, n) + '...';
    } else {
        return '...' + str.substr(str.length - n, n);
    }
    //MT2MHD: the middle part is more importnt than beginning
    if (str.length < n*2) return str;
    return str.substr(0, n) + '...' + str.substr(str.length - n, str.length-1);
}
/*
var test = [
    {link: 'http://cnn.com/image.jpg', parent: 'http://cnn.com/page.html'},
    {link: 'http://bbc.com/image.jpg', parent: 'http://cnn.com/page.html'},
    {link: 'http://cnn.com/images/image.jpg', parent: 'http://cnn.com/page.html'},
    {link: 'http://cnn.com/page.html/image.jpg', parent: 'http://cnn.com/page.html'},
    {link: 'http://cnn.com/image.jpg', parent: 'http://cnn.com/pages/sub/page.html'},
    {link: 'file:///Users/Matjaz/images/image.jpg', parent: 'file:///Users/Matjaz/pages/sub/page.html'},
    {link: 'file:///Users/Matjaz/images/image.jpg', parent: 'http://Users/Matjaz/pages/sub/page.html'}
]

for(var i = 0; i < test.length; i++) {
    console.log(test[i].link + ' to ' + test[i].parent + ' = ' + crsaMakeLinkRelativeTo(test[i].link, test[i].parent));
}
*/
function crsaGetRelativeDistanceForUrls (existUrl, currentUrl, callback) {
    var path = require('path');

    var currentDirArr = currentUrl.split('/');
    var existDirArr = existUrl.split('/');

    var newCrrent = currentDirArr.slice(0);
    var newExist = existDirArr.slice(0);
    for(var i = 0; i < currentDirArr.length; i++) {
        if (currentDirArr[i] == existDirArr[i]) {
            newCrrent.splice(newCrrent.indexOf(currentDirArr[i]), 1);
            newExist.splice(newExist.indexOf(existDirArr[i]), 1);
        }
        else break;
    }
    var newCrrentStr = newCrrent.join('/');
    var newExistStr = newExist.join('/');

    var newRelativeUrl = newCrrentStr.replace(newExistStr, '');
    if (newRelativeUrl == newCrrentStr) {
        newRelativeUrl = newExistStr.replace(newCrrentStr, '');

        var finalRelativeUrl;
        if (newRelativeUrl[newRelativeUrl.length-1] == "/")
            finalRelativeUrl = newRelativeUrl + newCrrentStr;
        else
            finalRelativeUrl = newRelativeUrl;

        if (newRelativeUrl == newExistStr) {
            var newPathArr = newCrrentStr.split('/');
            if (callback) callback(newPathArr.length - 1, finalRelativeUrl);
        }
        else {
            if (callback) callback(0, finalRelativeUrl);
        }
    }
    else {
        var newPathArr = newRelativeUrl.split('/');
        var distance = -1 * (newPathArr.length - 1);
        if (callback) callback(distance, null);
    }
}

var splitCssValue = function(v, comma_as_token, opening_re, closing_re, ignore_empty) {
    if(!opening_re) opening_re = /\(/;
    if(!closing_re) closing_re = /\)/;

    var r = [];
    if(!v || v.length == 0) return r;
    v = v + " \n";
    var token = '';
    var in_exp = false;
    var operands = "+-/*";
    var space = false;
    var par_level = 0;
    var quote_level = 0;

    for(var i = 0; i < v.length; i++) {
        var ch = v.charAt(i);
        if(ch == ' ') {
            space = true;
        } else {
            if(comma_as_token && ch == ',' && par_level == 0 && quote_level == 0) {
                r.push($.trim(token));
                space = false;
                token = '';
                r.push(',');
                continue;
            } else if(space && par_level == 0 && quote_level == 0) {
                token = $.trim(token);
                if(!ignore_empty || token.length) {
                    r.push(token);
                }
                space = false;
                token = '';
            }
            if(ch.match(opening_re)) {
                par_level++;
            } else if(ch.match(closing_re)) {
                par_level--;
            } else if(ch == '"' || ch == '\'') {
                quote_level = quote_level == 0 ? 1 : 0;
            }
        }
        token = token + ch;
    }
    token = $.trim(token);
    if(token.length > 0 && token != "\n") {
        r.push(token);
    }
    return r;
}

/*
 var tests = ["", "10px, 20px", "10+20 20px", "@a+@b @c-@d", "(10 + 20 + 30) 50", "func(10 20 (40 + 40), 40) 30px 40px 50px"];
 $.each(tests, function(i,t) {
 console.log(t);
 console.log(splitCssValue(t, true));
 });
*/

var getUrlFromCssUrlValue = function(url) {
    url = url.replace(/[\'\"]/g, '');
    var m = url.match(/url\(([^\)]*)\)/i);
    return m ? m[1] : url;
}

var getRand = function(min, max) {
    return Math.round(Math.random() * (max - min)) + min;
}

var pgGetRandomItem = function(list) {
    var idx = getRand(0, list.length - 1);
    return list[idx];
}

var pgKeepScrollPosition = function($content, func) {
    var st = $content.scrollTop();
    func();
    $content.scrollTop(st);
}

//console.log(getUrlFromCssUrlValue("url('aaaaaaaa')"));

var CssBackgroundParser = function(css) {

    var _this = this;

    this.parse = function(bg) {
        this.css = bg;

        this.color = '';
        this.image = '';
        this.repeat = '';
        this.attachment = '';
        this.position = '';
        this.important = false;

        //background:#ffffff url('img_tree.png') no-repeat right top;

        var def = ['transparent', 'none', 'repeat', 'scroll', '0%', '0%'];

        var idx = -1;
        var a = splitCssValue(bg, true);
        a.push(',');

        var bcka = [];
        for(var i = 0; i < a.length; i++) {
            if(a[i].match(/\!important/i)) {
                this.important = true;
                continue;
            } else if(a[i] == ',') {
                if(this.color.length > 0) this.color += ', ';
                this.color += bcka.length >= 1 ? bcka[0] : def[0];

                if(this.image.length > 0) this.image += ', ';
                this.image += bcka.length >= 2 ? bcka[1] : def[1];

                if(this.repeat.length > 0) this.repeat += ', ';
                this.repeat += bcka.length >= 3 ? bcka[2] : def[2];

                if(this.attachment.length > 0) this.attachment += ', ';
                this.attachment += bcka.length >= 4 ? bcka[3] : def[3];

                if(this.position.length > 0) this.position += ', ';
                this.position += bcka.length >= 5 ? bcka[4] : def[4];
                this.position += ' '
                this.position += bcka.length >= 6 ? bcka[5] : def[5];

                bcka = [];
            } else {
                bcka.push(a[i]);
            }
        }
        this.color = apply_important(this.color);
        this.image = apply_important(this.image);
        this.repeat = apply_important(this.repeat);
        this.attachment = apply_important(this.attachment);
        this.position = apply_important(this.position);
    }

    var apply_important = function(prop)
    {
        return prop + (_this.important ? ' !important' : '');
    }

    this.getValue = function(prop) {
        switch(prop.toLowerCase()) {
            case 'background-color': return this.color;
            case 'background-image': return this.image;
            case 'background-repeat': return this.repeat;
            case 'background-attachment': return this.attachment;
            case 'background-position': return this.position;
            default: return null;
        }
    }

    if(css) this.parse(css);
}

/*
tests = ["#ffffff url('img_tree.png') no-repeat scroll right top,#123456 url('img.png') no-repeat !important"];

var cp = new CssBackgroundParser();

$.each(tests, function(i,t) {
    cp.parse(t);
    console.log(t);
    console.log(cp.color, cp.image, cp.repeat, cp.attachment, cp.position);
});
*/

function crsaAddCancelSearch($input, style) {
    var $cancel = $('<a/>', {class: 'cancel-search', href: '#'}).html('&times;').appendTo($input.parent()).on('click', function(e) {
        e.preventDefault();
        $input.val('').trigger('input');
    });
    $input.on('input', function() {
        var val = $input.val();
        if(val && val.length) {
            $cancel.show();
            $input.addClass('has-value');
        } else {
            $cancel.hide();
            $input.removeClass('has-value');
        }
    })
    $cancel.hide();
    if(style) $cancel.attr('style', style);
    if ($input.val() != '') $input.addClass('has-value');
}

function crsaShowDate(date) {
    var monthNames = [
        "January", "February", "March",
        "April", "May", "June", "July",
        "August", "September", "October",
        "November", "December"
    ];
    var day = date.getDate();
    var monthIndex = date.getMonth();
    var year = date.getFullYear();

    return day + '. ' + monthNames[monthIndex] + ' ' + year;
}







/*
 * JavaScript MD5 1.0.1
 * https://github.com/blueimp/JavaScript-MD5
 *
 * Copyright 2011, Sebastian Tschan
 * https://blueimp.net
 *
 * Licensed under the MIT license:
 * http://www.opensource.org/licenses/MIT
 *
 * Based on
 * A JavaScript implementation of the RSA Data Security, Inc. MD5 Message
 * Digest Algorithm, as defined in RFC 1321.
 * Version 2.2 Copyright (C) Paul Johnston 1999 - 2009
 * Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet
 * Distributed under the BSD License
 * See http://pajhome.org.uk/crypt/md5 for more info.
 */

/*jslint bitwise: true */
/*global unescape, define */

(function ($) {
    'use strict';

    /*
     * Add integers, wrapping at 2^32. This uses 16-bit operations internally
     * to work around bugs in some JS interpreters.
     */
    function safe_add(x, y) {
        var lsw = (x & 0xFFFF) + (y & 0xFFFF),
            msw = (x >> 16) + (y >> 16) + (lsw >> 16);
        return (msw << 16) | (lsw & 0xFFFF);
    }

    /*
     * Bitwise rotate a 32-bit number to the left.
     */
    function bit_rol(num, cnt) {
        return (num << cnt) | (num >>> (32 - cnt));
    }

    /*
     * These functions implement the four basic operations the algorithm uses.
     */
    function md5_cmn(q, a, b, x, s, t) {
        return safe_add(bit_rol(safe_add(safe_add(a, q), safe_add(x, t)), s), b);
    }
    function md5_ff(a, b, c, d, x, s, t) {
        return md5_cmn((b & c) | ((~b) & d), a, b, x, s, t);
    }
    function md5_gg(a, b, c, d, x, s, t) {
        return md5_cmn((b & d) | (c & (~d)), a, b, x, s, t);
    }
    function md5_hh(a, b, c, d, x, s, t) {
        return md5_cmn(b ^ c ^ d, a, b, x, s, t);
    }
    function md5_ii(a, b, c, d, x, s, t) {
        return md5_cmn(c ^ (b | (~d)), a, b, x, s, t);
    }

    /*
     * Calculate the MD5 of an array of little-endian words, and a bit length.
     */
    function binl_md5(x, len) {
        /* append padding */
        x[len >> 5] |= 0x80 << (len % 32);
        x[(((len + 64) >>> 9) << 4) + 14] = len;

        var i, olda, oldb, oldc, oldd,
            a =  1732584193,
            b = -271733879,
            c = -1732584194,
            d =  271733878;

        for (i = 0; i < x.length; i += 16) {
            olda = a;
            oldb = b;
            oldc = c;
            oldd = d;

            a = md5_ff(a, b, c, d, x[i],       7, -680876936);
            d = md5_ff(d, a, b, c, x[i +  1], 12, -389564586);
            c = md5_ff(c, d, a, b, x[i +  2], 17,  606105819);
            b = md5_ff(b, c, d, a, x[i +  3], 22, -1044525330);
            a = md5_ff(a, b, c, d, x[i +  4],  7, -176418897);
            d = md5_ff(d, a, b, c, x[i +  5], 12,  1200080426);
            c = md5_ff(c, d, a, b, x[i +  6], 17, -1473231341);
            b = md5_ff(b, c, d, a, x[i +  7], 22, -45705983);
            a = md5_ff(a, b, c, d, x[i +  8],  7,  1770035416);
            d = md5_ff(d, a, b, c, x[i +  9], 12, -1958414417);
            c = md5_ff(c, d, a, b, x[i + 10], 17, -42063);
            b = md5_ff(b, c, d, a, x[i + 11], 22, -1990404162);
            a = md5_ff(a, b, c, d, x[i + 12],  7,  1804603682);
            d = md5_ff(d, a, b, c, x[i + 13], 12, -40341101);
            c = md5_ff(c, d, a, b, x[i + 14], 17, -1502002290);
            b = md5_ff(b, c, d, a, x[i + 15], 22,  1236535329);

            a = md5_gg(a, b, c, d, x[i +  1],  5, -165796510);
            d = md5_gg(d, a, b, c, x[i +  6],  9, -1069501632);
            c = md5_gg(c, d, a, b, x[i + 11], 14,  643717713);
            b = md5_gg(b, c, d, a, x[i],      20, -373897302);
            a = md5_gg(a, b, c, d, x[i +  5],  5, -701558691);
            d = md5_gg(d, a, b, c, x[i + 10],  9,  38016083);
            c = md5_gg(c, d, a, b, x[i + 15], 14, -660478335);
            b = md5_gg(b, c, d, a, x[i +  4], 20, -405537848);
            a = md5_gg(a, b, c, d, x[i +  9],  5,  568446438);
            d = md5_gg(d, a, b, c, x[i + 14],  9, -1019803690);
            c = md5_gg(c, d, a, b, x[i +  3], 14, -187363961);
            b = md5_gg(b, c, d, a, x[i +  8], 20,  1163531501);
            a = md5_gg(a, b, c, d, x[i + 13],  5, -1444681467);
            d = md5_gg(d, a, b, c, x[i +  2],  9, -51403784);
            c = md5_gg(c, d, a, b, x[i +  7], 14,  1735328473);
            b = md5_gg(b, c, d, a, x[i + 12], 20, -1926607734);

            a = md5_hh(a, b, c, d, x[i +  5],  4, -378558);
            d = md5_hh(d, a, b, c, x[i +  8], 11, -2022574463);
            c = md5_hh(c, d, a, b, x[i + 11], 16,  1839030562);
            b = md5_hh(b, c, d, a, x[i + 14], 23, -35309556);
            a = md5_hh(a, b, c, d, x[i +  1],  4, -1530992060);
            d = md5_hh(d, a, b, c, x[i +  4], 11,  1272893353);
            c = md5_hh(c, d, a, b, x[i +  7], 16, -155497632);
            b = md5_hh(b, c, d, a, x[i + 10], 23, -1094730640);
            a = md5_hh(a, b, c, d, x[i + 13],  4,  681279174);
            d = md5_hh(d, a, b, c, x[i],      11, -358537222);
            c = md5_hh(c, d, a, b, x[i +  3], 16, -722521979);
            b = md5_hh(b, c, d, a, x[i +  6], 23,  76029189);
            a = md5_hh(a, b, c, d, x[i +  9],  4, -640364487);
            d = md5_hh(d, a, b, c, x[i + 12], 11, -421815835);
            c = md5_hh(c, d, a, b, x[i + 15], 16,  530742520);
            b = md5_hh(b, c, d, a, x[i +  2], 23, -995338651);

            a = md5_ii(a, b, c, d, x[i],       6, -198630844);
            d = md5_ii(d, a, b, c, x[i +  7], 10,  1126891415);
            c = md5_ii(c, d, a, b, x[i + 14], 15, -1416354905);
            b = md5_ii(b, c, d, a, x[i +  5], 21, -57434055);
            a = md5_ii(a, b, c, d, x[i + 12],  6,  1700485571);
            d = md5_ii(d, a, b, c, x[i +  3], 10, -1894986606);
            c = md5_ii(c, d, a, b, x[i + 10], 15, -1051523);
            b = md5_ii(b, c, d, a, x[i +  1], 21, -2054922799);
            a = md5_ii(a, b, c, d, x[i +  8],  6,  1873313359);
            d = md5_ii(d, a, b, c, x[i + 15], 10, -30611744);
            c = md5_ii(c, d, a, b, x[i +  6], 15, -1560198380);
            b = md5_ii(b, c, d, a, x[i + 13], 21,  1309151649);
            a = md5_ii(a, b, c, d, x[i +  4],  6, -145523070);
            d = md5_ii(d, a, b, c, x[i + 11], 10, -1120210379);
            c = md5_ii(c, d, a, b, x[i +  2], 15,  718787259);
            b = md5_ii(b, c, d, a, x[i +  9], 21, -343485551);

            a = safe_add(a, olda);
            b = safe_add(b, oldb);
            c = safe_add(c, oldc);
            d = safe_add(d, oldd);
        }
        return [a, b, c, d];
    }

    /*
     * Convert an array of little-endian words to a string
     */
    function binl2rstr(input) {
        var i,
            output = '';
        for (i = 0; i < input.length * 32; i += 8) {
            output += String.fromCharCode((input[i >> 5] >>> (i % 32)) & 0xFF);
        }
        return output;
    }

    /*
     * Convert a raw string to an array of little-endian words
     * Characters >255 have their high-byte silently ignored.
     */
    function rstr2binl(input) {
        var i,
            output = [];
        output[(input.length >> 2) - 1] = undefined;
        for (i = 0; i < output.length; i += 1) {
            output[i] = 0;
        }
        for (i = 0; i < input.length * 8; i += 8) {
            output[i >> 5] |= (input.charCodeAt(i / 8) & 0xFF) << (i % 32);
        }
        return output;
    }

    /*
     * Calculate the MD5 of a raw string
     */
    function rstr_md5(s) {
        return binl2rstr(binl_md5(rstr2binl(s), s.length * 8));
    }

    /*
     * Calculate the HMAC-MD5, of a key and some data (raw strings)
     */
    function rstr_hmac_md5(key, data) {
        var i,
            bkey = rstr2binl(key),
            ipad = [],
            opad = [],
            hash;
        ipad[15] = opad[15] = undefined;
        if (bkey.length > 16) {
            bkey = binl_md5(bkey, key.length * 8);
        }
        for (i = 0; i < 16; i += 1) {
            ipad[i] = bkey[i] ^ 0x36363636;
            opad[i] = bkey[i] ^ 0x5C5C5C5C;
        }
        hash = binl_md5(ipad.concat(rstr2binl(data)), 512 + data.length * 8);
        return binl2rstr(binl_md5(opad.concat(hash), 512 + 128));
    }

    /*
     * Convert a raw string to a hex string
     */
    function rstr2hex(input) {
        var hex_tab = '0123456789abcdef',
            output = '',
            x,
            i;
        for (i = 0; i < input.length; i += 1) {
            x = input.charCodeAt(i);
            output += hex_tab.charAt((x >>> 4) & 0x0F) +
                hex_tab.charAt(x & 0x0F);
        }
        return output;
    }

    /*
     * Encode a string as utf-8
     */
    function str2rstr_utf8(input) {
        return unescape(encodeURIComponent(input));
    }

    /*
     * Take string arguments and return either raw or hex encoded strings
     */
    function raw_md5(s) {
        return rstr_md5(str2rstr_utf8(s));
    }
    function hex_md5(s) {
        return rstr2hex(raw_md5(s));
    }
    function raw_hmac_md5(k, d) {
        return rstr_hmac_md5(str2rstr_utf8(k), str2rstr_utf8(d));
    }
    function hex_hmac_md5(k, d) {
        return rstr2hex(raw_hmac_md5(k, d));
    }

    function md5(string, key, raw) {
        if (!key) {
            if (!raw) {
                return hex_md5(string);
            }
            return raw_md5(string);
        }
        if (!raw) {
            return hex_hmac_md5(key, string);
        }
        return raw_hmac_md5(key, string);
    }

    if (typeof define === 'function' && define.amd) {
        define(function () {
            return md5;
        });
    } else {
        $.md5 = md5;
    }
}(this));


// scales the canvas by (float) scale < 1
// returns a new canvas containing the scaled image.
function downScaleCanvas(cv, scale) {
    if (!(scale < 1) || !(scale > 0)) throw ('scale must be a positive number <1 ');
    var sqScale = scale * scale; // square scale = area of source pixel within target
    var sw = cv.width; // source image width
    var sh = cv.height; // source image height
    var tw = Math.ceil(sw * scale); // target image width
    var th = Math.ceil(sh * scale); // target image height
    var sx = 0, sy = 0, sIndex = 0; // source x,y, index within source array
    var tx = 0, ty = 0, yIndex = 0, tIndex = 0; // target x,y, x,y index within target array
    var tX = 0, tY = 0; // rounded tx, ty
    var w = 0, nw = 0, wx = 0, nwx = 0, wy = 0, nwy = 0; // weight / next weight x / y
    // weight is weight of current source point within target.
    // next weight is weight of current source point within next target's point.
    var crossX = false; // does scaled px cross its current px right border ?
    var crossY = false; // does scaled px cross its current px bottom border ?
    var sBuffer = cv.getContext('2d').
        getImageData(0, 0, sw, sh).data; // source buffer 8 bit rgba
    var tBuffer = new Float32Array(4 * sw * sh); // target buffer Float32 rgb
    var sR = 0, sG = 0,  sB = 0; // source's current point r,g,b
    // untested !
    var sA = 0;  //source alpha

    for (sy = 0; sy < sh; sy++) {
        ty = sy * scale; // y src position within target
        tY = 0 | ty;     // rounded : target pixel's y
        yIndex = 4 * tY * tw;  // line index within target array
        crossY = (tY != (0 | ty + scale));
        if (crossY) { // if pixel is crossing botton target pixel
            wy = (tY + 1 - ty); // weight of point within target pixel
            nwy = (ty + scale - tY - 1); // ... within y+1 target pixel
        }
        for (sx = 0; sx < sw; sx++, sIndex += 4) {
            tx = sx * scale; // x src position within target
            tX = 0 |  tx;    // rounded : target pixel's x
            tIndex = yIndex + tX * 4; // target pixel index within target array
            crossX = (tX != (0 | tx + scale));
            if (crossX) { // if pixel is crossing target pixel's right
                wx = (tX + 1 - tx); // weight of point within target pixel
                nwx = (tx + scale - tX - 1); // ... within x+1 target pixel
            }
            sR = sBuffer[sIndex    ];   // retrieving r,g,b for curr src px.
            sG = sBuffer[sIndex + 1];
            sB = sBuffer[sIndex + 2];
            sA = sBuffer[sIndex + 3];

            if (!crossX && !crossY) { // pixel does not cross
                // just add components weighted by squared scale.
                tBuffer[tIndex    ] += sR * sqScale;
                tBuffer[tIndex + 1] += sG * sqScale;
                tBuffer[tIndex + 2] += sB * sqScale;
                tBuffer[tIndex + 3] += sA * sqScale;
            } else if (crossX && !crossY) { // cross on X only
                w = wx * scale;
                // add weighted component for current px
                tBuffer[tIndex    ] += sR * w;
                tBuffer[tIndex + 1] += sG * w;
                tBuffer[tIndex + 2] += sB * w;
                tBuffer[tIndex + 3] += sA * w;
                // add weighted component for next (tX+1) px
                nw = nwx * scale
                tBuffer[tIndex + 4] += sR * nw; // not 3
                tBuffer[tIndex + 5] += sG * nw; // not 4
                tBuffer[tIndex + 6] += sB * nw; // not 5
                tBuffer[tIndex + 7] += sA * nw; // not 6
            } else if (crossY && !crossX) { // cross on Y only
                w = wy * scale;
                // add weighted component for current px
                tBuffer[tIndex    ] += sR * w;
                tBuffer[tIndex + 1] += sG * w;
                tBuffer[tIndex + 2] += sB * w;
                tBuffer[tIndex + 3] += sA * w;
                // add weighted component for next (tY+1) px
                nw = nwy * scale
                tBuffer[tIndex + 4 * tw    ] += sR * nw; // *4, not 3
                tBuffer[tIndex + 4 * tw + 1] += sG * nw; // *4, not 3
                tBuffer[tIndex + 4 * tw + 2] += sB * nw; // *4, not 3
                tBuffer[tIndex + 4 * tw + 3] += sA * nw; // *4, not 3
            } else { // crosses both x and y : four target points involved
                // add weighted component for current px
                w = wx * wy;
                tBuffer[tIndex    ] += sR * w;
                tBuffer[tIndex + 1] += sG * w;
                tBuffer[tIndex + 2] += sB * w;
                tBuffer[tIndex + 3] += sA * w;
                // for tX + 1; tY px
                nw = nwx * wy;
                tBuffer[tIndex + 4] += sR * nw; // same for x
                tBuffer[tIndex + 5] += sG * nw;
                tBuffer[tIndex + 6] += sB * nw;
                tBuffer[tIndex + 7] += sA * nw;
                // for tX ; tY + 1 px
                nw = wx * nwy;
                tBuffer[tIndex + 4 * tw    ] += sR * nw; // same for mul
                tBuffer[tIndex + 4 * tw + 1] += sG * nw;
                tBuffer[tIndex + 4 * tw + 2] += sB * nw;
                tBuffer[tIndex + 4 * tw + 3] += sA * nw;
                // for tX + 1 ; tY +1 px
                nw = nwx * nwy;
                tBuffer[tIndex + 4 * tw + 4] += sR * nw; // same for both x and y
                tBuffer[tIndex + 4 * tw + 5] += sG * nw;
                tBuffer[tIndex + 4 * tw + 6] += sB * nw;
                tBuffer[tIndex + 4 * tw + 7] += sA * nw;
            }
        } // end for sx
    } // end for sy

    // create result canvas
    var resCV = document.createElement('canvas');
    resCV.width = tw;
    resCV.height = th;
    var resCtx = resCV.getContext('2d');
    var imgRes = resCtx.getImageData(0, 0, tw, th);
    var tByteBuffer = imgRes.data;
    // convert float32 array into a UInt8Clamped Array
    var pxIndex = 0; //
    for (sIndex = 0, tIndex = 0; pxIndex < tw * th; sIndex += 4, tIndex += 4, pxIndex++) {
        tByteBuffer[tIndex] = Math.ceil(tBuffer[sIndex]);
        tByteBuffer[tIndex + 1] = Math.ceil(tBuffer[sIndex + 1]);
        tByteBuffer[tIndex + 2] = Math.ceil(tBuffer[sIndex + 2]);
        tByteBuffer[tIndex + 3] = Math.ceil(tBuffer[sIndex + 3]);
    }
    // writing result to canvas.
    resCtx.putImageData(imgRes, 0, 0);
    return resCV;
}


var crsaResizeImage = function(file, w, h, new_file) {
    var fs = require('fs');
    var tempCanvas = document.createElement('canvas');
    tempCanvas.width = w;
    tempCanvas.height = h;
    var ctx = tempCanvas.getContext('2d');
    var img = new Image();
    img.src = crsaMakeUrlFromFile(file);
    img.onload = function () {
        ctx.drawImage(this, 0, 0, w, h);

        var smallCanvas = downScaleCanvas(tempCanvas, 0.5);

        var image = smallCanvas.toDataURL('image/png');
        image = image.replace('data:image/png;base64,', '');
        var buffer = new Buffer(image, 'base64');
        fs.writeFileSync(new_file, buffer);
    };
}

var pg_is_safari = navigator.userAgent.indexOf("Safari") > -1 && navigator.userAgent.indexOf('Chrome') < 0;

var pgScrollToWithoutAnimation = function(elem, dim, val) {
    var win;
    if ( elem === window ) {
        win = elem;
    } else if ( elem.nodeType === 9 ) {
        win = elem.defaultView;
    }
    if ( win ) {
        if(pg_is_safari) {
            win.scrollTo(dim === 'left' ? val : win.pageXOffset, dim === 'top' ? val : win.pageYOffset);
        } else {
            win.scrollTo({
                top: dim === 'top' ? val : win.pageYOffset,
                left: dim === 'left' ? val : win.pageXOffset,
                behavior: 'auto'
            })
        }
        //win[dim === 'top' ? 'scrollY' : 'scrollX'] = val;
    } else {
        elem[dim === 'top' ? 'scrollTop' : 'scrollLeft' ] = val;
    }
}

var crsaScrollToElementInIframe = function($iframe, $el) {
    var $contents = $iframe.contents();
    var h = $iframe.get(0).clientHeight;
    var w = $iframe.get(0).clientWidth;
    var st = $contents.scrollTop();
    var ofs = $el.offset();
    var y = ofs.top;
    var x = ofs.left;
    var sl = $contents.scrollLeft();
    if(y < st || y > st + h*0.9) {
        if(y >= h/3.0) y -= Math.floor(h/3.0);
        //pgScrollToWithoutAnimation($contents.get(0), 'top', y);
        $contents.scrollTop(y);
    }
    if(x < sl || x > sl + w*0.9) {
        if(x >= w/3.0) x -= Math.floor(w/3.0);
        //pgScrollToWithoutAnimation($contents.get(0), 'left', y);
        $contents.scrollLeft(x);
    }
}


function takeElementPhoto ($el, filename, comp) {
    var fs = require('fs');

    pinegrow.takePhotoOfElement($el, 400.0, filename, function() {
        pinegrow.showQuickMessage('Screenshot has been taken.');
        pinegrow.cache_breaker++;

        var gui = require('nw.gui');
        gui.App.clearCache();

        comp.preview_image = comp.type + '.png';

        pinegrow.frameworksChanged();
    });
}

// for dev only
function addTakePreviewImage (f) {
    f.on_build_actions_menu = function(page, menus, pgel, for_all_selected) {
        menus.push({
            label: "Take preview image",
            kbd: null,
            action: function($el) {

                var selectedComp = null;
                for (var comp in f.component_types) {
                    var currentComp = f.component_types[comp];

                    if (currentComp.selector) {
                        var selector = currentComp.selector;
                        if (typeof(currentComp.selector) == "function") {
                            selector = currentComp.selector($el);
                        }
                        else {
                            selector = pgel.is(selector)
                        }

                        if (selector) {
                            selectedComp = currentComp;
                            break;
                        }
                    }
                }

                if (selectedComp) {
                    var filename = crsaCombineCurrentUrlWith(crsaCleanUpUrl(f.getImagePreviewBaseUrl()), selectedComp.type) + '.png';
                    takeElementPhoto ($el, crsaMakeFileFromUrl(filename), selectedComp);
                }
                else {
                    pinegrow.showQuickMessage('Component not found!');
                }
            }
        });
    };
}


// Override the ui autocompelete filter
$.ui.autocomplete.filter = function (array, term) {
    var matcher = new RegExp("^" + $.ui.autocomplete.escapeRegex(term), "i");
    return $.grep(array, function (value) {
        return matcher.test(value.label || value.value || value);
    });
};

function useSelete2 () {
    $('select').select2({
        allowClear: true
    });
}

function getUrlWithoutProxy (url) {
    if(!pinegrow.httpServer) {
        return url;
    }
    return pinegrow.getOriginalUrl(url);
    /*
    var proxyUrl = pinegrow.httpServer.url + '/';
    var proxyRe = new RegExp(escapeRegExp(proxyUrl), 'g');

    return url.replace(proxyRe, '');*/
}



// CSS helpers
function styleToObject (style) {
    if (style) {
        styleArr = style.split(';');
        var styleObj = {};
        for (var i = 0; i < styleArr.length; i++) {
            if (styleArr[i] != "") {
                var keyVal = styleArr[i].split(':');
                var key = keyVal[0].trim();
                keyVal.splice(0, 1);

                styleObj[key] = keyVal.join(':').trim();
            }
        }
        return styleObj;
    }
    else {
        return {};
    }
}

function objectToStyle (styleObj) {
    var style = "";
    for (var key in styleObj) {
        if (key && styleObj[key]) {
            style += key + ":" + styleObj[key] + ";";
        }
    }
    return style;
}

function getNewStyle (currentStyle, newStyle, oldStyle) {
    var currentStyleObj = styleToObject(currentStyle);
    var newStyleObj = styleToObject(newStyle);
    var oldStyleObj = styleToObject(oldStyle);

    for (var key in oldStyleObj) {
        var value = currentStyleObj[key];
        if (value) delete currentStyleObj[key]
    }

    for (var key in newStyleObj) {
        currentStyleObj[key] = newStyleObj[key];
    }

    return objectToStyle(currentStyleObj);
}

function getStyleFor (attr, value) {
    return attr + ":" + value + ";";
}

function setStyle (pgel, newStyle, oldStyle) {
    if (!pgel) return;
    var currentStyle = pgel.getAttr('style');
    pgel.setAttr('style', getNewStyle(currentStyle, newStyle, oldStyle));
}

function crsaCopyToClipboard(text) {
    copyCodeToClipboard(text, true);
    return;

  var $temp = $("<input>");
  $("body").append($temp);
  $temp.val(text).select();
  document.execCommand("copy");
  $temp.remove();
}

var crsaSelectText = function(element) {
    var doc = document;
    if (doc.body.createTextRange) {
       var range = document.body.createTextRange();
       range.moveToElementText(element);
       range.select();
    }
    if (doc.body.createTextRange) {
        var range = document.body.createTextRange();
        range.moveToElementText(element);
        range.select();
    } else if (window.getSelection) {
        var selection = window.getSelection();
        var range = document.createRange();
        range.selectNodeContents(element);
        selection.removeAllRanges();
        selection.addRange(range);
    }
}

var crsaSetCaretPlacement = function (element, atStart) {
    if (typeof window.getSelection != "undefined" && typeof document.createRange != "undefined") {
        var range = document.createRange();
        range.selectNodeContents(element);
        range.collapse(atStart);
        var sel = window.getSelection();
        sel.removeAllRanges();
        sel.addRange(range);
    } else if (typeof document.body.createTextRange != "undefined") {
        var textRange = document.body.createTextRange();
        textRange.moveToElementText(element);
        textRange.collapse(atStart);
        textRange.select();
    }
}


function crsaGetNextElement (currElement, selector) {
    var elements = $(selector);
    var currentIndex = elements.index(currElement);
    return elements[currentIndex + 1];
}

function crsaGetPrevElement (currElement, selector) {
    var elements = $(selector);
    var currentIndex = elements.index(currElement);
    return elements[currentIndex - 1];
}

function crsaCamelize(str) {
  return str.replace(/^([a-z])/, function(letter, index) {
    return index == 0 ? letter.toLowerCase() : letter.toUpperCase();
  }).replace(/\s+/g, '');
}

var crsaHtmlEncodeHelper = document.createElement("div");

function crsaHtmlEncode(s) {
    crsaHtmlEncodeHelper.innerText = s;
    return crsaHtmlEncodeHelper.innerHTML;
}

// TODO: enhance imports using this function
function crsaGetUrlAndParamsFromImportNEW(importParams) {
    var allImporters = [];
    var importers = importParams.split(',');
    for (var i = 0; i < importers.length; i++) {
        var imp = importers[i].trim();

        var impPath = "",
            impParams = null;


        var spaceIndex = imp.indexOf(' ');
        var length = imp.length;
        if (spaceIndex > -1) {
            impPath = imp.substr(0, spaceIndex);
            impParams = imp.substr(spaceIndex + 1, length);
        }
        else {
            impPath = imp;
        }

        var usedUrl = false;
        if (impPath.indexOf("url") > -1) {
            usedUrl = true;
            impPath = impPath.replace(/(?:url\()(.*)(?:\))/, '$1');
        }

        impPath = impPath.replace(/[\'\"]/g, '');

        allImporters.push({
            usedUrl: usedUrl,
            path: impPath,
            params: impParams
        })
    }

    return allImporters;
}

function crsaGetUrlAndParamsFromImport(importParams) {
    var importArr = [];
    var spaceIndex = importParams.indexOf(' ');
    var length = importParams.length;
    if (spaceIndex > -1) {
        importArr[0] = importParams.substr(0, spaceIndex);
        importArr[1] = importParams.substr(spaceIndex + 1, length);
    }
    else {
        importArr[0] = importParams;
    }

    importArr[0] = importArr[0].replace(/(?:url\()(.*)(?:\))/, '$1');
    importArr[0] = importArr[0].replace(/[\'\"]/g, '');

    return importArr;
}

var colorToString = function(color) {
    if(color === null) return '';
    if (color.getAlpha() != 1.0) {
        return color.toRgbString();
    } else {
        var c = color.toString();
        return c;
    }
}

function pgRemoveMultiselects($el) {
    $el.find('.crsa-input[data-autocomplete]').each(function (i, elm) {
        var autoComplete = $(elm).data('pg-autocomplete');
        if (autoComplete) {
            autoComplete.remove(true);
            $(elm).data('pg-autocomplete', autoComplete = null);
        }
    });
}

function pgRemoveSpectrum($el, selector) {
    selector = selector || '.crsa-input-color-picker';
    $el.find(selector).each(function(i, elm) {
        var pgSpectrumHelper = $(elm).data('pg-spectrum-helper');

        if (pgSpectrumHelper) {
            pgSpectrumHelper.remove();
            pgSpectrumHelper = null;
            $(elm).data('pg-spectrum-helper', null);
        }

        $(elm).spectrum("destroy");
    });
}

function pgRemoveFilePicker($el, selector) {
    selector = selector || '.crsa-pick-file';
    $el.find(selector).each(function(i, elm) {
        $(elm).unbind("click");
    });
}

function pgSetDataOfObjsTo($elArr, attr, value) {
    $elArr.each(function (i, elm) {
        $(elm).data(attr, value);
    });
}

function slideToggle(el, callback) {
    var $el = $(el), height = $el.data("originalHeight"), visible = $el.is(":visible");
    bShow = !visible;
    if (bShow == visible) return false;
    if (!height) {
        height = $el.show().height();
        $el.data("originalHeight", height);
        if (!visible) $el.hide().css({height: 0});
    }

    if (bShow) {
        $el.show().animate({height: height}, {
            duration: 250, complete: function () {
                $el.height('auto');
                if (callback) callback();
            }
        });
    } else {
        $el.animate({height: 0}, {
            duration: 250, complete: function () {
                $el.hide();
                if (callback) callback();
            }
        });
    }
}

function crsaGetFuzzySearchRegxFrom(str) {
    var ignoreSpcialChar = function (char) {
        return '\\' + char;
    }

    var regexStr = '.*';
    for (var i = 0; i < str.length; i++) {
        var currChar = str[i];
        if (currChar == '$') currChar = ignoreSpcialChar(currChar);

        regexStr += (currChar + '.*');
    }
    return new RegExp(regexStr);
}

function getCodeMirrorMode (ext) {
    if (ext == "css") {
        return 'text/css';
    } else if (ext = "scss") {
        return 'text/x-scss';
    } else if (ext = "less") {
        return 'text/x-less';
    }

    return 'text';
}

// End CSS helpers

function pgE(tag, attrs, html)
{
    var e = document.createElement(tag);
    if(attrs)
    {
        for(var attr in attrs)
        {
            e.setAttribute(attr, attrs[attr]);
        }
    }
    if(html) {
        e.innerHTML = html;
    }
    return e;
}

function pg$E(tag, attrs, html) {
    return $(pgE(tag, attrs, html));
}

//Length
//https://raw.githubusercontent.com/heygrady/Units/master/Length.js

(function(window, document, undefined){
    "use strict";

// create a test element
    var testElem = document.createElement('test'),
        docElement = document.documentElement,
        defaultView = document.defaultView,
        getComputedStyle = defaultView && defaultView.getComputedStyle,
        computedValueBug,
        runit = /^(-?[\d+\.\-]+)([a-z]+|%)$/i,
        convert = {},
        conversions = [1/25.4, 1/2.54, 1/72, 1/6],
        units = ['mm', 'cm', 'pt', 'pc', 'in', 'mozmm'],
        i = 6; // units.length

// add the test element to the dom
    docElement.appendChild(testElem);

// test for the WebKit getComputedStyle bug
// @see http://bugs.jquery.com/ticket/10639
    if (getComputedStyle) {
        // add a percentage margin and measure it
        testElem.style.marginTop = '1%';
        computedValueBug = getComputedStyle(testElem).marginTop === '1%';
    }

// pre-calculate absolute unit conversions
    while(i--) {
        convert[units[i] + "toPx"] = conversions[i] ? conversions[i] * convert.inToPx : toPx(testElem, '1' + units[i]);
    }

// remove the test element from the DOM and delete it
    docElement.removeChild(testElem);
    testElem = undefined;

// convert a value to pixels
    function toPx(elem, value, prop, force) {
        // use width as the default property, or specify your own
        prop = prop || 'width';

        var style,
            inlineValue,
            ret,
            unit = (value.match(runit)||[])[2],
            conversion = unit === 'px' ? 1 : convert[unit + 'toPx'],
            rem = /r?em/i;

        if (conversion || rem.test(unit) && !force) {
            // calculate known conversions immediately
            // find the correct element for absolute units or rem or fontSize + em or em
            elem = conversion ? elem : unit === 'rem' ? docElement : prop === 'fontSize' ? elem.parentNode || elem : elem;

            // use the pre-calculated conversion or fontSize of the element for rem and em
            conversion = conversion || parseFloat(curCSS(elem, 'fontSize'));

            // multiply the value by the conversion
            ret = parseFloat(value) * conversion;
        } else {
            // begin "the awesome hack by Dean Edwards"
            // @see http://erik.eae.net/archives/2007/07/27/18.54.15/#comment-102291

            // remember the current style
            style = elem.style;
            inlineValue = style[prop];

            // set the style on the target element
            try {
                style[prop] = value;
            } catch(e) {
                // IE 8 and below throw an exception when setting unsupported units
                return 0;
            }

            // read the computed value
            // if style is nothing we probably set an unsupported unit
            ret = !style[prop] ? 0 : parseFloat(curCSS(elem, prop));

            // reset the style back to what it was or blank it out
            style[prop] = inlineValue !== undefined ? inlineValue : null;
        }

        // return a number
        return ret;
    }

// return the computed value of a CSS property
    function curCSS(elem, prop) {
        var value,
            pixel,
            unit,
            rvpos = /^top|bottom/,
            outerProp = ["paddingTop", "paddingBottom", "borderTop", "borderBottom"],
            innerHeight,
            parent,
            i = 4; // outerProp.length

        if (getComputedStyle) {
            // FireFox, Chrome/Safari, Opera and IE9+
            value = getComputedStyle(elem)[prop];
        } else if (pixel = elem.style['pixel' + prop.charAt(0).toUpperCase() + prop.slice(1)]) {
            // IE and Opera support pixel shortcuts for top, bottom, left, right, height, width
            // WebKit supports pixel shortcuts only when an absolute unit is used
            value = pixel + 'px';
        } else if (prop === 'fontSize') {
            // correct IE issues with font-size
            // @see http://bugs.jquery.com/ticket/760
            value = toPx(elem, '1em', 'left', 1) + 'px';
        } else {
            // IE 8 and below return the specified style
            value = elem.currentStyle[prop];
        }

        // check the unit
        unit = (value.match(runit)||[])[2];
        if (unit === '%' && computedValueBug) {
            // WebKit won't convert percentages for top, bottom, left, right, margin and text-indent
            if (rvpos.test(prop)) {
                // Top and bottom require measuring the innerHeight of the parent.
                innerHeight = (parent = elem.parentNode || elem).offsetHeight;
                while (i--) {
                    innerHeight -= parseFloat(curCSS(parent, outerProp[i]));
                }
                value = parseFloat(value) / 100 * innerHeight + 'px';
            } else {
                // This fixes margin, left, right and text-indent
                // @see https://bugs.webkit.org/show_bug.cgi?id=29084
                // @see http://bugs.jquery.com/ticket/10639
                value = toPx(elem, value);
            }
        } else if ((value === 'auto' || (unit && unit !== 'px')) && getComputedStyle) {
            // WebKit and Opera will return auto in some cases
            // Firefox will pass back an unaltered value when it can't be set, like top on a static element
            value = 0;
        } else if (unit && unit !== 'px' && !getComputedStyle) {
            // IE 8 and below won't convert units for us
            // try to convert using a prop that will return pixels
            // this will be accurate for everything (except font-size and some percentages)
            value = toPx(elem, value) + 'px';
        }
        return value;
    }

// expose the conversion function to the window object
    window.Length = {
        toPx: toPx
    };
}(this, this.document));


if(!Object.values) {
    Object.values = function(obj) {
        var vals = [];

        for (var key in obj) {
            if (obj.hasOwnProperty(key)) {
                vals.push(obj[key]);
            }
        }
        return vals;
    }
}


var crsaGetPositionOverElementCenterPosition = function(w, h, pos_left, pos_top, zoom) {
    if(zoom === undefined) zoom = 1.0;

    var center_w = 100;//w * 0.4;
    var center_h = 100;//h * 0.4;
    var scale = 1;

    if(w < 110 && h < 110) {
        center_w = 20;
        center_h = 20;
    }

    /*
    if(center_w < 80) {
        center_w = w * 0.8;
    }

    if(center_h < 80) {
        center_h = h * 0.8;
    }
    */

    var center_x_start = pos_left + (w - center_w)/2;
    var center_x_end = center_x_start + center_w;

    var center_y_start = pos_top + (h - center_h)/2;
    var center_y_end = center_y_start + center_h;

    return {
        w: center_w,
        h: center_h,
        top: center_y_start,
        left: center_x_start,
        bottom: center_y_end,
        right: center_x_end
    }
}

var crsaIsDOMElementDescendantOf = function(node, parent) {
    while(node && node.nodeType === 1) {
        if(node === parent) return true;
        node = node.parentNode;
    }
    return null;
}

var crsaGetPositionOverElement = function(doc, x, y, page, prev_info, ignore_collapsed, in_subcall) {

    var arg_x = x;
    var arg_y = y;

    var dd_options = pinegrow.drag_drop_options;

    var pel = doc.elementFromPoint(x, y);

    if(!pel) return null;

    if(pel.ownerSVGElement) {
        var svg = pel.closest('svg');
        if(svg.getAttribute('data-pg-collapsed') !== 'false') {
            pel = svg;
        }
    }
    //console.log(pel)

    if(pel.classList.contains('pinegrow-ui-helper')) {
        pel = pel.closest(':not(.pinegrow-ui-helper)')
    }


    if(!pel) return null;

    var $pel = $(pel);

    if(pel.tagName === 'HTML') $pel = $pel.find('body');

    if($pel.length === 0) return null;

    var collapsed = false;

    var visible_col_pgel = getElementPgNode($pel);

    var redo_info = false;

    var getElementVisibility = function(pgel) {
        var pgid = pgel.getId();

        if(pgid) {
            var tree_data = page.tree && page.tree.getTreeDataItemForPgId(pgid);
            //console.log(cloneObject(tree_data))
            if(tree_data && 'hidden' in tree_data) {
                if(tree_data.hidden) {
                    return 'hidden';
                } else if(tree_data.collapsed) {
                    return 'collapsed';
                } else {
                    return 'normal';
                }
            } else {
                var status = 'normal';

                var node = pgel;

                while(node.parent) {
                    if(node.hasAttribute('data-pg-collapsed')) {
                        status = node === pgel ? 'collapsed' : 'hidden';
                    }
                    node = node.parent;
                }
                return status;
            }
        }
        return 'normal';
    }

    if(dd_options.force_parent) {
        $pel = dd_options.force_parent.get$DOMElement();
        visible_col_pgel = dd_options.force_parent;

    } else if(!ignore_collapsed /* && page && page.tree */) {
        var col_pgel = visible_col_pgel;

        var in_uncollapsed_list = false;

        /*
        if(prev_info && prev_info.uncollapsed) {
            var clean_list = [];
            //console.log('UNCOLLAPSED')
            prev_info.uncollapsed.forEach(function(n) {
                //console.log(n)
                if(visible_col_pgel.isDescendantOf(n) || visible_col_pgel === n) {
                    clean_list.push(n);
                }
            })
            prev_info.uncollapsed = clean_list;
        }
        */

        if(col_pgel) {
            //check if collapsed in tree
            visible_col_pgel = col_pgel;

            var tree_data = null;
            var collapsed_pgel = null;

            var v = null;

            while(true) {
                v = getElementVisibility(visible_col_pgel);

//                console.log(visible_col_pgel.getName(), v)

                if(v === 'normal') {
                    break;
                }

                if(v === 'collapsed') {
                    collapsed_pgel = visible_col_pgel;
                    break;
                }

                /*
                var pgid = visible_col_pgel.getId();

                if (pgid) {
                    tree_data = page.tree.getTreeDataItemForPgId(pgid);
                } else {
                    break;
                }
*/
                /*
                if(tree_data && tree_data.collapsed) {
                    collapsed_pgel = visible_col_pgel;
                }

                if(prev_info && prev_info.uncollapsed) {
                    if(prev_info.uncollapsed.indexOf(visible_col_pgel) >= 0 ) {
                        in_uncollapsed_list = true;
                        break;
                    }
                }
                */

                /*
                if (!tree_data || !tree_data.hidden) {
                    break;
                }
                */

                if(!visible_col_pgel.parent) break;
                visible_col_pgel = visible_col_pgel.parent;
            }

            //console.log(visible_col_pgel, in_uncollapsed_list, tree_data ? tree_data.collapsed : false);

            if(collapsed_pgel === visible_col_pgel) {
                visible_col_pgel = visible_col_pgel.parent;
                collapsed = true;
            }
            $pel = visible_col_pgel.get$DOMElement();

        } else {
            //console.log('no col_pgel', $pel);
        }
    } else if(!ignore_collapsed) {
        //try with el collapse info

        visible_col_pgel.withSelfAndParents(function(node) {
            if(node.hasAttribute('data-pg-collapsed')) {
                visible_col_pgel = node.parent;
            }
        })
    }

    var pos = $pel.offset();

    x = x + doc.scrollingElement.scrollLeft;
    y = y + doc.scrollingElement.scrollTop;

    var info = null;
    var grid_info = null;

    var dm = pinegrow.getDragMenu();

    if(pgf.edit_css) {
        grid_info = new PgCSSGridInfo($pel, true);
        if (dm) {
            dm.setOverGrid(grid_info.isGrid(), grid_info.$el)
        }


        if (pinegrow.keyState.cmdCtrl) {
            //grid drop
            //try to find closest grid element
            if (grid_info.isGrid()) {
                $pel = grid_info.$el;

                var grid_x = x - grid_info.rect.left - doc.scrollingElement.scrollLeft;
                var grid_y = y - grid_info.rect.top - doc.scrollingElement.scrollTop;
                var grid_col = grid_info.getColForXOld(grid_x);
                var grid_row = grid_info.getRowForYOld(grid_y);

                var cell = grid_info.getCellRect(grid_col, grid_row);

                info = {
                    $e: $pel, grid: true, rect: cell, position: 'inside-bottom', grid_position: {
                        col: grid_col,
                        row: grid_row
                    }
                };
                //console.log(info, grid_info.rect);
                return info;
            }
            /*
             //handle grid drop
             var pels = window.getComputedStyle($pel.get(0));
             if(pels.display == 'grid' || pels.display == 'inline-grid') {
             //we're over empty grid cell
             grid_info = grid_info || new PgCSSGridInfo($pel);
             var grid_x = x - grid_info.rect.left - doc.scrollingElement.scrollLeft;
             var grid_y = y - grid_info.rect.top - doc.scrollingElement.scrollTop;
             var grid_col = grid_info.getColForX(grid_x);
             var grid_row = grid_info.getRowForY(grid_y);

             var cell = grid_info.getCellRect(grid_col, grid_row);

             info = {$e: $pel, grid: true, rect: cell, position: 'inside-bottom', grid_position: {
             col: grid_col,
             row: grid_row
             }};
             //console.log(info);
             return info;
             }*/
        }
    }

    //console.log($pel.get(0))

    //check if mouse over any el child margins

    var $children = $pel.children();

    $children.each(function(i, e) {

        if(e.classList.contains('pinegrow-ui-helper')) return true;

        var $e = $(e);

        var pos = $e.offset();

        var w = $e.innerWidth();
        var h = $e.innerHeight();

        //console.log(e.nodeName);

        if(w > 0 && h > 0 && e.nodeName !== 'SCRIPT') {

            //console.log('Visible')

            var pt = parseInt($e.css('padding-top'));
            var pb = parseInt($e.css('padding-bottom'));
            var mt = parseInt($e.css('margin-top'));
            var mb = parseInt($e.css('margin-bottom'));
            var ml = parseInt($e.css('margin-left'));
            var mr = parseInt($e.css('margin-right'));

            var rect = {x1: pos.left - ml, y1: pos.top - mt};
            rect.x2 = pos.left + w + mr;
            rect.y2 = pos.top + h + mb;

            //console.log(`${x} ${y}`, rect)

            if (x >= rect.x1 && x <= rect.x2 && y >= rect.y1 && y <= rect.y2) {
                //console.log('in margin')
                info = {$e: $e};

                $pel = $e;

                if (y < pos.top) {
                    info.position = 'top';
                } else if (y > pos.top + h) {
                    info.position = 'bottom';
                } else if (x < pos.left) {
                    info.position = 'left';
                } else if(x > pos.left + w) {
                    info.position = 'right';
                } else {
                    if (y < rect.y1 + pt + h / 2) {
                        info.position = 'top';
                    } else {
                        info.position = 'bottom';
                    }
                }

                if(false && collapsed && !in_subcall) {

                    var center_i = crsaGetPositionOverElementCenterPosition(w, h, pos.left, pos.top);

                    if(x > center_i.left && x < center_i.right && y > center_i.top && y < center_i.bottom) {
                        //it's at the center of collapsed element
                        //console.log('at the center')
                        info.center = true;

                        var t = (new Date()).getTime();

                        info.center_time = t;

                        info.over_center = true;

                        if(prev_info) {
                            if(prev_info.center) {
                                info.center_time = prev_info.center_time;

                                if(t - info.center_time > 1000) {
                                    //uncollapse this element
                                    var this_pgel = getElementPgNode($e);

                                    if(this_pgel) {
                                        info.uncollapsed = prev_info.uncollapsed || [];
                                        var idx = info.uncollapsed.indexOf(this_pgel);
                                        if (idx < 0) {
                                            info.uncollapsed.push(this_pgel);
                                        }
                                        collapsed = false;

                                        info = crsaGetPositionOverElement(doc, arg_x, arg_y, page, info, true, true);
                                    }
                                }
                            }
                        }
                    }
                }

                //console.log(e, w, h)

                return false;
            }
        }
    })

    //console.log(info)

    var pgel = getElementPgNode($pel);

    if(!info) {
        var pt = parseInt($pel.css('padding-top'));
        var pb = parseInt($pel.css('padding-bottom'));
        var pl = parseInt($pel.css('padding-left'));
        var pr = parseInt($pel.css('padding-right'));

        var cw = $pel.width();
        var ch = $pel.height();

        var w = pl + cw + pr;
        var h = pt + ch + pb;

        var v_spacer = h_spacer = 15;

        if(pgel && pgel.tagName === 'img') {
            v_spacer = h_spacer = 50;
        }

        //v_spacer = h / 4;
        //h_spacer = w / 4;

        if(h < v_spacer * 4) {
            v_spacer = h / 4;
        }

        if(w < h_spacer * 4) {
            h_spacer = w / 4;
        }

        info = info || {}


        info.$e = $pel;
        info.position = 'inside-top';

        if (y - pos.top > h / 2) {
            info.position = 'inside-bottom';
        }

        if(pel.tagName !== 'BODY') {

            if (y > pos.top + pt && y < pos.top + pt + ch && x > pos.left + pl && x < pos.left + pl + cw) {
                info.over_text = true;
            }

            if (y < pos.top + v_spacer) {
                info.position = 'top';
            } else if (y > pos.top + h - v_spacer) {
                info.position = 'bottom';
            } else if (x < pos.left + h_spacer) {
                info.position = 'left';
            } else if (x > pos.left + w - h_spacer) {
                info.position = 'right';
            }
        }
    }

    if(prev_info) {
        if(!info.uncollapsed && prev_info.uncollapsed) {
            info.uncollapsed = prev_info.uncollapsed;
        }
    }

    info.collapsed = collapsed;

    if(pgel) {
        info.target_pgel = pgel;
    }

    return info;
}

function crsaSplitCharsInArray(str) {
    var a = [];
    var i = 0;
    while(i < str.length) {
        var ch = str[i];
        if(ch == '&') {
            var idx = str.indexOf(';', i);
            if(idx < 0) idx = str.length - 1;
            a.push(str.substring(i, idx + 1));
            i = idx + 1;
        } else {
            a.push(ch);
            i++;
        }
    }
    return a;
}

function crsaGetCharArraySubstring(a, start, end) {
    var s = '';
    if(end > a.length) end = a.length;
    for(var i = start; i < end; i++) {
        s += a[i];
    }
    return s;
}

var crsaGetFreshMatchedRulesOLD = function (win, el, pseudo) {
    pseudo = pseudo || '';
    return win.getMatchedCSSRules(el, pseudo);
}


var crsaGetFreshMatchedRules = function (pgel, done) {
    $.fn.crsa('getMatchedCSSRules', pgel, function (params, inherited, inline) {
        if (done) done(params, inherited, inline);
    });
}

function crsaCombinedTwoCSSList (first, second) {
    first  = first  || [];
    second = second || [];

    var arr = [];

    for (var i = 0; i < first.length; i++)  {
        arr.push(first[i]);
    }
    for (var i = 0; i < second.length; i++) {
        var found = false;
        for (var j = 0; j < arr.length; j++) {
            if (second[i] == arr[j]) {
                found = true;
                break;
            }
        }

        if (!found) arr.push(second[i]);
    }

    return arr;
}

var crsaGetAllUsedNodes = function (cs, node, skipRecursive) {
    var nodesName = [];
    var nodes = [];

    var pgcssh = pinegrow.pgPostCSSHelper;

    var resolveUsed = function (value) {
        var varReg = new RegExp('\\' + cs.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');
        var m = value.match(varReg);

        if (m) {
            m.forEach(function (used_var) {
                var newNode = pgcssh.findNodeInside('decl', 'prop', used_var, cs.getRootNode());
                if (newNode && !skipRecursive) {
                    resolveUsed(newNode.value);
                }

                if (newNode && nodesName.indexOf(used_var) == -1) {
                    nodesName.push(used_var);
                    nodes.push(newNode);
                }
            })
        }
    }

    var isVariable = pgcssh.isVariable(node);
    if (isVariable) nodesName.push(node.prop);

    resolveUsed(node.value);

    if (isVariable) nodes.push(node);

    return {
        names: nodesName,
        nodes: nodes
    }
}

function getStylePropWithoutImportant(propValue) {
    if (!propValue) return propValue;

    var important = "!important";
    var spectrumVal = propValue;
    if (spectrumVal.endsWith(important)) {
        spectrumVal = spectrumVal.replace(new RegExp('\\s*\\' + important), '');
    }

    return spectrumVal;
}

function pgGetActive$DOMElement($node, on_invalid) {
    var node = $node.get(0);
    if(node.ownerDocument.defaultView) return $node;
    var pgid = node.getAttribute('data-pg-id');
    if(!pgid) {
        return null;
    }
    var pgel = getPgNodeByPgId(pgid);
    if(!pgel) {
        return null;
    }
    var $el = pgel.get$DOMElement();
    if(!$el) {
        return null;
    }
    if(on_invalid) on_invalid();
    return $el;
}

var getNewFontsUrl = function (url, name, action_type) {
    if (action_type == 'remove') {
        var newFontsArray = [];

        var matchedAllFonts = url.match(/family=(.*)/);
        if (matchedAllFonts && matchedAllFonts.length > 0) {
            var fontsArray = matchedAllFonts[1].split('|');

            for (var i = 0; i < fontsArray.length; i++) {
                var font = fontsArray[i];
                var matched = font.match(/([a-zA-Z0-9\+]*)/);
                if (matched && matched.length > 0) {
                    var fontName = matched[1].replace(/\+/g, ' ');
                    if (fontName != name) {
                        newFontsArray.push(font);
                    }
                }
            }
        }

        if (newFontsArray.length > 0) {
            return "https://fonts.googleapis.com/css?family=" + newFontsArray.join('|');
        }
        else {
            return null;
        }
    }
    else if (action_type == 'add') {
        return url + "|" + name;
    }

    return url;
}

var getFontNameFromURL = function (url) {
    var allFonts = [];

    var matchedAllFonts = url.match(/family=(.*)/);
    if (matchedAllFonts && matchedAllFonts.length > 0) {
        var fontsArray = matchedAllFonts[1].split('|');

        for (var i = 0; i < fontsArray.length; i++) {
            var font = fontsArray[i];
            var matched = font.match(/([a-zA-Z0-9\+]*)/);
            if (matched && matched.length > 0) {
                var fontName = matched[1].replace(/\+/g, ' ');
                var fontOptions = {};

                var fontOptionsStr = font.replace(matched[1], '');
                if (fontOptionsStr.length > 0) {
                    var varientsIndex = fontOptionsStr.indexOf(':');
                    var moreOptionsIndex = fontOptionsStr.indexOf('&');

                    if (varientsIndex > -1) {
                        var varientsStr = fontOptionsStr.substr(varientsIndex, moreOptionsIndex > -1 ? moreOptionsIndex : fontOptionsStr.length);

                        fontOptions['varients'] = varientsStr.slice(1).split(',');
                        fontOptionsStr = fontOptionsStr.replace(varientsStr, '');
                    }

                    if (fontOptionsStr.length > 0 && moreOptionsIndex > -1) {
                        fontOptionsStr = fontOptionsStr.slice(1);
                        var restOptions = fontOptionsStr.split('&');

                        for (var j = 0; j < restOptions.length; j++) {
                            var option = restOptions[j];
                            var keyValue = option.split('=');
                            if (keyValue[0] == "subset") {
                                fontOptions[keyValue[0]] = keyValue[1].split(',');
                            }
                            else {
                                fontOptions[keyValue[0]] = keyValue[1];
                            }
                        }
                    }
                }

                allFonts.push({
                    name: fontName,
                    options: fontOptions
                });
            }
        }
    }

    return allFonts;
}

// var getFontNameFromURL = function (url) {
//     var matched = url.match(/family=([a-zA-Z\+]*)/);
//     if (matched && matched.length > 0) {
//         return matched[1].replace(/\+/g, ' ');
//     } else {
//         return null;
//     }
// }

var getFontNameAndOptions = function (url) {
    var fonts = [];
    var fontsArray = getFontNameFromURL(url);

    for (var i = 0; i < fontsArray.length; i++) {
        var name = fontsArray[i].name;
        if (name) {
            var fontObj = {
                url: url,
                name: name
            }

            if (fontsArray[i].options) {
                if (fontsArray[i].options.varients) {
                    fontObj['included_varients'] = fontsArray[i].options.varients;
                }
                if (fontsArray[i].options.display) {
                    fontObj['included_display'] = fontsArray[i].options.display;
                }
                if (fontsArray[i].options.subset) {
                    fontObj['included_subset'] = fontsArray[i].options.subset;
                }
            }

            fonts.push(fontObj);
        }
    }

    return fonts;
}

var getAttachedGoogleFonts = function (page) {
    var existFonts = [];
    pinegrow.goThroughAllForAttribute(page, 'href', function (el, attr, url) {
        var matched = url.match(/fonts.googleapis.com\/css/);
        if (el.is('link') && (matched && matched.length > 0)) {
            existFonts = [...existFonts, ...getFontNameAndOptions(url)];
        }
    });

    return pinegrow.pgGoogleFontManager.getFontOptions(existFonts);
}

function pgShow(node, d) {
    node.style.display = d || 'block';
}

function pgHide(node) {
    node.style.display = 'none';
}

function pg$Show($node, d) {
    $node.each(function(i,e) {
        e.style.display = d || 'block';
    })
}

function pg$Hide($node) {
    $node.each(function(i,e) {
        e.style.display = 'none';
    })
}

function pg$CSS($node, style_obj) {
    $node.each(function(i,e) {
        for(const prop in style_obj) {
            e.style[prop] = style_obj[prop];
        }
    })
}

function pg$PositionAt($node, x, y) {
    $node.each(function(i,e) {
        pgPositionAt(e, x, y);
    })
}

function pgPositionAt(e, x, y) {
    e.pgCurrentPosition = {x: x, y: y};
    e.style.transform = `translate3d(${x}px, ${y}px, 0)`;
}

function pg$Closest($node, sel) {
    const n = $node.get(0).closest(sel);
    return n ? $(n) : $();
}

function pgFindAll(node, selector) {
    return node.querySelectorAll(selector);
}

function pg$FindAll($node, selector) {
    return $node.get(0).querySelectorAll(selector);
}

function pg$Flash($e, animation, delay) {
    animation = animation || 'flash';
    if($e._pg_flash_timer) {
        clearTimeout($e._pg_flash_timer);
    }
    $e.addClass(animation);
    flash_timer = setTimeout(function() {
        $e.removeClass(animation);
        $e._pg_flash_timer = null;
    }, delay || 1000);
}

function pgGetDOMElementName(node, get_cls, get_text, info) {

    if(get_cls === undefined) get_cls = true;
    if(get_text === undefined) get_text = true;

    var search = node.tagName.toLowerCase();

    var name = '<nametag class="dyn">' + search + '</nametag>';

    var t;
    if(get_text) {
        t = '';
        var text_done = false;
        //debugger;
        if(node.tagName == 'META') {
            var meta_name = node.getAttribute('name');
            var meta_content = node.getAttribute('content');
            if(meta_name) {
                t = meta_name + '="' + (meta_content || '') + '"';
            } else {
                if(node.attributes.length) {
                    t = node.attributes[0].nodeName + '="' + node.attributes[0].nodeValue + '"';
                }
            }
            text_done = true;
        } else if(node.tagName == 'LINK') {
            t = pinegrow.getOriginalUrl((node.getAttribute('href') || ''));
            text_done = true;
        }  else if(node.tagName == 'SCRIPT') {
            var src = node.getAttribute('src');
            if(src) {
                t = pinegrow.getOriginalUrl(src);
                text_done = true;
            } else {
                t = node.innerText || '';
            }
        }
        if(!text_done) {
            for(var ci = 0; ci < node.childNodes.length; ci++) {
                if(node.childNodes[ci].nodeType == 3) {
                    t += $.trim(node.childNodes[ci].textContent);
                    if(t.length) {
                        t = t.substring(0,30);
                        break;
                    }
                }
            }
        }
        if(t.length) t = t.replace(/&/g,'&amp;').replace(/</g,'&lt;').replace(/>/g,'&gt;');

        if(node.tagName === 'A') {
            var href = pinegrow.getOriginalUrl((node.getAttribute('href') || ''));
            if(href) {
                var lt = crsaGetSummaryStr(href, 30);
                if(t.length) {
                    lt += ' | ' + t;
                }
                t = lt;
            }
        }
    }

    var id = node.getAttribute('id');
    if(id) {
        id = escapeHtmlCode(id);
        name += '<nameid>' + "#" + id + '</nameid>';
        search += id;
    }

    if(get_cls) {
        var cls_count = 0;
        var first_cls = null;
        var all_cls = '';

        if(node.classList.length > 0) {
            for(var i = 0; i < node.classList.length; i++) {
                if(node.classList[i].startsWith('crsa') || node.classList[i].startsWith('pgc-') || node.classList[i].startsWith('pg-')) continue;
                cls_count++;
                var cls = escapeHtmlCode(node.classList[i]);
                if(cls_count === 1) first_cls = cls;
                all_cls += '.' + cls;

            }
            if(cls_count > 0) {
                search += ':' + all_cls;
                name += '<nameclass' + (cls_count > 1 ? ' title="' + all_cls  + '"' : '') + '>' + '.' + first_cls + (cls_count > 1 ? '...' : '') + '</nameclass>';
            }

        }

    }
    if(t) {
        name += ' | ' + '<nametext>' + t + '</nametext>';
        search += '|' + t;
    }

    if(info) info.search_string = search;
    return name;
}

function pgDynElementMessage(msg) {
    var m = '<p><b>' + msg + '</b></p><p>Dynamic elements are HTML elements created by Javascript code on your page. These elements are not part of the HTML code of the page and so can\'t be edited through Pinegrow.</p><p>You have to edit the relevant Javascript code in order to change dynamic elements.</p>';
    pinegrow.showNotice(m, msg, 'dyn-element-operation', function(shown) {
        if(!shown) {
            pinegrow.showQuickMessage(msg, 4000, false, 'error');
        }
    })
}

function pgGetElementClasses(el) {
    var r = [];
    var re = /^(crsa\-|medium\-editor|pg\-node|pgc\-|pg\-show\-outline|pinegrow\-ui\-|pg\-internal\-|pghelper\-|pg\-state\-|wp\-mark\-areas)/;
    el.classList.forEach(function(cls) {
        if(!cls.match(re)) {
            r.push(cls);
        }
    })
    return r;
}

function getCSSFunctionNameAndParams(str) {
    var startBracketIndex = str.indexOf('(');
    var endBracketIndex = str.indexOf(')');
    var fnName = str;
    var params = null;
    if (startBracketIndex > -1) {
        fnName = str.substring(0, startBracketIndex);
        params = str.substring(startBracketIndex + 1, endBracketIndex);
    }

    return {
        name: fnName,
        params: params
    }
}


function pgStringToNumberRange(str, min, max) {
    var n = 0;
    for(var i = 0; i < str.length; i++) {
        n += str.charCodeAt(i);
    }
    return (n % (max - min)) + min;
}

function pgScrollToItemInContainer($i, $container, always, animate, done, offset) {
    if(always === undefined) always = false;
    if(offset === undefined) offset = 0;

    //var pos = $i.position();
    var top = $container.scrollTop();
    //pos.top += top;

    var iRect = $i.get(0).getBoundingClientRect();
    var cRect = $container.get(0).getBoundingClientRect();

    var pos = {}
    pos.top = (iRect.top - cRect.top) + top;

    var h = cRect.height;//$container.height();
    var ih = iRect.height;//$i.height();

    var vis_top = top;
    var vis_bottom = top + h;

    var el_top = pos.top + offset;
    var el_bottom = el_top + ih;

    if(!always && el_top >= vis_top && el_bottom <= vis_bottom) {
        if(done) done(false);
        return; //fully visible
    }
    if(animate) {
        $container.stop().animate({
            scrollTop: el_top
        }, typeof animate === 'number' ? animate : 150, function() {
            if(done) done(true);
        })
    } else {
        $container.scrollTop(el_top);
        if(done) done(true);
    }

    /*

    if(pos.top - top > h - ih * 2) {
        $container.scrollTop(pos.top - ih);
    } else if(pos.top < top) {
        $container.scrollTop(pos.top - h + ih*2);
    }*/
}

function pgGetUniqueShortNamesForUrls(list) {
    var h = list.map(function(url) {
        var items = url.split('/');
        var last = items.pop();
        while(last.length == 0 && items.length) {
            last = items.pop();
        }
        return {
            items: items,
            name: last
        }
    })

    var expandName = function(h) {
        if(h.items.length) {
            h.name = h.items.pop() + '/' + h.name;
            return true;
        } else {
            return false;
        }
    }

    while(true) {

        var same_map = {};

        var expanded = false;
        var all_unique = true;

        for(var i = 0; i < h.length; i++) {

            if(!same_map[h[i].name]) {
                same_map[h[i].name] = [h[i]];
            } else {
                if(same_map[h[i].name].length == 1) {
                    expanded = expanded || expandName(same_map[h[i].name][0]);
                }
                same_map[h[i].name].push(h[i]);
                expanded = expanded || expandName(h[i]);
                all_unique = false;
            }
        }

        if(all_unique || !expanded) {
            return h.map(function(h) {
                return h.name;
            })
        }
    }
}

function pgMakeStringArrayUnique(a) {
    var map = {};
    var r = [];
    a.forEach(function(d) {
        if(d in map) {

        } else {
            r.push(d);
            map[d] = true;
        }
    })
    return r;
}


function mkdirpsync (p, opts, made) {
    /*
     "author": {
     "name": "James Halliday",
     "email": "mail@substack.net",
     "url": "http://substack.net"
     }
     */
    var _0777 = parseInt('0777', 8);

    if (!opts || typeof opts !== 'object') {
        opts = { mode: opts };
    }

    var mode = opts.mode;
    var xfs = opts.fs || fs;

    if (mode === undefined) {
        mode = _0777 & (~process.umask());
    }
    if (!made) made = null;

    p = path.resolve(p);

    try {
        xfs.mkdirSync(p, mode);
        made = made || p;
    }
    catch (err0) {
        switch (err0.code) {
            case 'ENOENT' :
                made = mkdirpsync(path.dirname(p), opts, made);
                mkdirpsync(p, opts, made);
                break;

            // In the case of any other error, just see if there's a dir
            // there already.  If so, then hooray!  If not, then something
            // is borked.
            default:
                var stat;
                try {
                    stat = xfs.statSync(p);
                }
                catch (err1) {
                    throw err0;
                }
                if (!stat.isDirectory()) throw err0;
                break;
        }
    }
    return made;
};

function getTrimed(source) {
    if (source) {
        return source.trim();
    }
    return null;
}

function pgGeneratePassword(len) {
    var possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+-$";
    len = len || 6;
    s = '';
    for(var i = 0; i < len; i++) {
        s += possible.charAt(Math.floor(Math.random() * possible.length));
    }
    return s;
}

var pgGenerateUsername = function() {
    var w1 = [
        'Space', 'Super', 'Cool', 'Quick', 'Smart', 'Kind', 'Clever', 'Amazing', 'Awesome', 'Good', 'Brainy', 'Blue', 'Red', 'Green', 'Yellow', 'White', 'Big', 'Slow', 'Full', 'Moving', 'Golden', 'Soft', 'Speedy', 'Loud', 'Quiet', 'Strong', 'Lovely', 'Kind'
    ]
    var w2 = [
        'Cookie', 'Star', 'Rocket', 'Moon', 'Sun', 'Flower', 'Bird', 'Duck', 'Rabbit', 'Bee', 'Ship', 'Cloud', 'Hill', 'Tree', 'Puppy', 'Cat', 'Bear', 'Dove', 'River', 'Stone', 'Gold', 'Grass', 'Ant', 'Light', 'Dance', 'Lightning', 'Alien', 'Dog', 'Sea', 'Car', 'giant', 'Crab', 'Coder', 'Cricket', 'Boat', 'Jet'   ]

    return w1[ getRand(0, w1.length - 1)] + w2[ getRand(0, w2.length - 1)] + getRand(100, 999);
}

function pgSaveOriginalStyle($el) {
    $el.attr('data-pg-original-style', $el.attr('style') || '')
}

function pgRestoreOriginalStyle($ch, except_prop) {
    var style = $ch.attr('data-pg-original-style');
    if(style !== null && style !== undefined) {
        if(style) {
            if(except_prop) {
                var re = new RegExp( escapeRegExp(except_prop) + '\\:\\s*' + '([^\\;]*)' + '\\;?', 'g');
                style = style.replace(re, '');
            }
            $ch.attr('style', style);
        } else {
            $ch.removeAttr('style');
        }
        $ch.removeAttr('data-pg-original-style');
    }
}

function pgForgetOriginalStyle($ch) {
    $ch.removeAttr('data-pg-original-style');
}

var PgDOMHelper = function() {

    var getWindow = function(node) {
        return node.defaultView;
    }

    this.closestWithClass = function(node, cls) {
        while(node.nodeType === 1) {
            if(node.classList.contains(cls)) return node;
            node = node.parentNode;
        }
        return null;
    }

    this.getScrollLeft = function(node) {
        return node.pageXOffset || node.scrollLeft;// getWindow(node).pageXOffset;
    }

    this.getScrollTop = function(node) {
        return node.pageYOffset || node.scrollTop;// getWindow(node).pageYOffset;
    }

    this.getWidth = function(node) {
        return node.clientWidth;
    }

    this.getHeight = function(node) {
        return node.clientHeight;
    }

    this.getOffset = function(elem) {
        var box = elem.getBoundingClientRect();
        var docElem = elem.ownerDocument.documentElement;
        var win = getWindow( elem.ownerDocument );
        return {
            top: box.top  + ( win.pageYOffset || docElem.scrollTop )  - ( docElem.clientTop  || 0 ),
            left: box.left + ( win.pageXOffset || docElem.scrollLeft ) - ( docElem.clientLeft || 0 )
        };
    }

    this.findChildByTag = function(node, tag) {
        for(var i = 0; i < node.children.length; i++) {
            if(node.children[i].nodeType === 1 && node.children[i].tagName === tag) {
                return node.children[i];
            }
        }
        return null;
    }
}

var pgMeasureTextWidth_measureTextWidthCanvasContext = $('<canvas></canvas>').get(0).getContext("2d");
pgMeasureTextWidth_measureTextWidthCanvasContext.font = '13px -apple-system,BlinkMacSystemFont,"Segoe UI",Roboto,Oxygen-Sans,Ubuntu,Cantarell,"Helvetica Neue",sans-serif';

var pgMeasureTextWidth = function (text) {
    return pgMeasureTextWidth_measureTextWidthCanvasContext.measureText(text).width;
}

var pgUpdateGeneratedSectionContent = function(content, section, code, position) {
    if(!position) position = 'append';

    var re = new RegExp('\\/\\*\\s*Pinegrow generated ' + section + ' Begin\\s*\\*\\/[\\s\\S]*?\\/\\*\\s*Pinegrow generated ' + section + ' End\\s*\\*\\/', 'ig');

    if(!content.match(re)) {
        if(position === 'none') return content;

        var add_str = `
/* Pinegrow generated ${section} Begin */

/* Pinegrow generated ${section} End */
`;
        if(position === 'append') {
            content += add_str;
        } else if(position === 'prepend') {
            content = add_str + content;
        }
    }
    content = content.replace(re, '/* Pinegrow generated ' + section + ' Begin */\n' + code + '\n    /* Pinegrow generated ' + section + ' End */');

    return content;
}

var pgUpdateGeneratedSection = function(file, section, code) {
    var fs = require('fs');
    var content = '';

    try {
        content = fs.readFileSync(file, {encoding: 'utf8'}) || '';
    } catch(err) {}

    content = pgUpdateGeneratedSectionContent(content, section, code);

    var dir = require('path').dirname(file);

    if(!crsaIsFileOrDir(dir, fs)) {
        crsaCreateDirs(dir, fs);
    }

    crsaWriteFileWithBackup(fs, file, content, 'utf8');
}

class PgPreloadedResources {
    constructor(base) {
        this.base = base;
        this.map = {};
    }

    preload(list) {
        var _this = this;
        list.forEach(function(item) {
            var url = _this.base + item;
            if(url.endsWith('.php')) url += 'dis.html';

            var version = Date.now();

            fetch(url + '?version=' + version).then(function(response) {
                return response.text().then(function(text) {
                    if(item.endsWith('.phpdis.html')) {
                        item = item.replace('dis.html', '');
                    }
                    _this.map[item] = text;
                    console.log('Got remote resource file ' + url);
                });
            }).catch(function (error) {
                console.log('Error getting remote resource file ' + url, error);
            });
        })
    }

    getContent(url) {
        return this.map[url] || null;
    }
}