"use strict";

(function() {
    var pgIdCount = 0;
    var pgObjectCount = 0;

    var pgIdRegExp = new RegExp('\\s*data\\-pg\\-id="([0-9]+)"', 'i');
    var pgIdRegExpReplace = new RegExp('\\s*data\\-pg\\-id="[0-9]+"', 'ig');
    var pgWhiteSpaceRegExp = new RegExp('^[\\s\\n]*$');
    var pgSelectorCache = {};

    var pgDontFormat = ['a', 'p', 'span', 'bdo', 'em', 'strong', 'dfn', 'code', 'samp', 'kbd', 'var', 'cite', 'abbr', 'acronym', 'q', 'sub', 'sup', 'tt', 'i', 'b', 'big', 'small', 'u', 's', 'strike', 'font', 'ins', 'del', 'pre', 'address', 'dt', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'br', 'textarea'];//, '?php', '?='];

    var pgInlineTags = ['a', 'span', 'em', 'strong', 'code', 'kbd', 'var', 'cite', 'abbr', 'acronym', 'sub', 'sup', 'i', 'b', 'big', 'small', 'u', 's', 'strike', 'font', 'ins', 'del', 'br'];


    var pgDontFormatIfAllChildrenNonFormat = ['a', 'td', 'th'];

    var pgAutoClosedTags = {
        'html': [],
        'head': ['body'],
        'body': [],
        'p': ['p', 'address', 'article', 'aside', 'blockquote', 'div', 'dl', 'fieldset', 'footer', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'header', 'hr', 'menu', 'nav', 'ol', 'pre', 'section', 'table', 'ul'],
        'dd': ['dd'],
        'dt': ['dt'],
        'option': ['option', 'optgroup'],
        'optgroup': ['optgroup'],
        'thead': ['tbody', 'tfoot'],
        'th': ['td', 'th'],
        'tbody': ['tfoot'],
        'tr': ['tr'],
        'td': ['td', 'th'],
        'tfoot': ['tbody'],
        'colgroup': [],
        'li': ['li']
    }
    var pgSingleTags = window.pgSingleTags = ['area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input', 'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr', '!', /* '?php', '?=', */ '!doctype'];


    var pgScripts = {
        '<?php': {type: 'php', name: 'PHP', close: '?>', name_tag: '&lt;?php'},
        '<?xml': {type: 'xml', name: 'XML', close: '?>', name_tag: '&lt;?xml'},
        '<?=': {type: 'php', name: 'PHP', close: '?>', name_tag: '&lt;?='},
        '<?': {type: 'php', name: 'PHP', close: '?>', name_tag: '&lt;?'},
        '<%': {type: 'asp', name: 'ASP &amp; ERB', close: '%>', name_tag: '&lt;%'}
    }

    var pgScriptsByTag = {};
    var pgScriptsList = [];
    var pgScriptsRegExp = null;
    var pgScriptsTagsRegExp = null;
    var pgScriptsStartsWithRegExp = null;
    var pgScriptsStartsWithCloseRegExp = null;

    var pgAttrScripts = [];//[['<?php', '?>'], ['<?=', '?>']];

    var pgParserEscapeRegExp = window.pgParserEscapeRegExp = function (value) {
        if (typeof value != 'string') value = String(value);
        return value.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g, "\\$&");
    }


    var pgPrepareScripts = function () {
        var re = '';
        var re_tags = '';
        var re_sw = '^';
        var re_sw_close = '^';

        for (var key in pgScripts) {
            if (pgScripts.hasOwnProperty(key)) {
                var def = pgScripts[key];
                def.open = key;
                pgScriptsList.push(def);

                var tag = def.open.replace('<', '');
                pgScriptsByTag[tag] = def;

                pgDontFormat.push(tag);
                pgSingleTags.push(tag);

                re += (re.length ? '|' : '') + '(' + pgParserEscapeRegExp(def.open) + ')';
                re_tags += (re_tags.length ? '|' : '') + '(' + pgParserEscapeRegExp(tag) + ')';
                re_sw += (re_sw.length !== 1 ? '|' : '') + '(' + pgParserEscapeRegExp(def.open) + ')';
                re_sw_close += (re_sw_close.length !== 1 ? '|' : '') + '(' + pgParserEscapeRegExp(def.close) + ')';

                pgAttrScripts.push([key, def.close]);
            }
        }


        pgScriptsRegExp = new RegExp(re, 'i');
        pgScriptsTagsRegExp = new RegExp(re_tags, 'i');
        pgScriptsStartsWithRegExp = new RegExp(re_sw, 'ig');
        pgScriptsStartsWithCloseRegExp = new RegExp(re_sw_close, 'ig');
    }

    pgPrepareScripts();

    var pgHasScript = function (str) {
        return str.match(pgScriptsRegExp);
    }

    var pgIsScriptTag = function (tag) {
        return tag.match(pgScriptsTagsRegExp);
    }

    var pgGetScriptInfo = function (opening_tag) {
        if (opening_tag[0] != '<') return null;
        return pgScripts[opening_tag] || null;
    }

    var pgGetScriptInfoByTag = function (tag) {
        return pgScriptsByTag[tag] || null;
    }

    var pgTestAtIndex = function (str, re, i) {
        re.lastIndex = i;
        var m = re.test(str);
        return m;
    }

    var pgMatchAtIndex = function (str, re, i) {
        re.lastIndex = i;
        var m = re.exec(str);
        return m;
    }


    var pgAttributeEncode_re = /"/g;
    var pgAttributeDecode_re = /&quot;/g;


    var pgEncodeAttribute = window.pgEncodeAttribute = function (str) {
        if (str === null) return null;
        //if(pgHasScript(str)) return str;
        if (str.indexOf('<?') >= 0) return str;
        return str.replace(pgAttributeEncode_re, '&quot;');
    }

    var pgDecodeAttribute = window.pgDecodeAttribute = function (str) {
        if (typeof str != 'string') return str;
        return str.replace(pgAttributeDecode_re, '"');
    }

    var pgParserException = window.pgParserException = function (msg, obj) {
        this.message = msg;
        this.name = "pgParserException";
        this.reference = obj;

        this.toString = function () {
            return this.message + ' - ' + escapeHtmlCode(this.reference);
        }
    }

    var pgParserSourceProblem = window.pgParserSourceProblem = function (node, $el, skip_locked) {
        this.message = "Can not edit source node.";
        this.name = "pgParserSourceProblem";
        this.node = node;
        this.$el = $el;

        var list = [];

        this.add = function (obj_type, obj, action, reason, msg) {
            if (!reason) reason = 'dynamic';
            list.push({
                obj_type: obj_type,
                obj: obj instanceof pgParserNode ? obj.getName() : obj,
                action: action,
                reason: reason,
                msg: msg
            });
        }

        if (node && !skip_locked) {
            var locked = pinegrow.isElementLocked(node);
            if (locked) {
                this.add('element', $el ? getElementName($el) : "node", "change", 'custom', locked);
            }
        }

        this.ok = function () {
            return list.length === 0;
        }

        this.toString = function () {
            if (list.length === 0) return null;
            var s = '<ul>';
            for (var i = 0; i < list.length; i++) {
                var msg = '';
                var p = list[i];
                switch (p.obj_type) {
                    case 'class':
                        msg = 'Class <b>' + p.obj + '</b>';
                        break;

                    case 'element':
                        msg = 'Element <b>' + p.obj + '</b>';
                        break;
                }

                var reason;
                switch (p.reason) {
                    default:
                        reason = 'because it was either: <b>added by a script</b>, <b>removed from the page</b> or <b>changed in the code editor without refreshing the page</b>';
                }
                if (p.msg) {
                    reason = p.msg;
                }

                switch (p.action) {
                    case 'remove':
                        msg += ' can\'t be <b>removed</b>';
                        break;
                    case 'add':
                        msg += ' can\'t be <b>added</b>';
                        break;
                    case 'change':
                        msg += ' can\'t be <b>changed</b>';
                        break;
                    case 'find':
                        msg += ' does\'t <b>exist in source</b>';
                        break;
                    case 'duplicate':
                        msg += ' cant\'t be <b>duplicated</b>';
                        break;
                }

                msg += ' ' + reason + '.';

                s += '<li>' + msg + '</li>';
            }
            s += '</ul>';
            return s;
        }

    }

    var pgParserNodeCatalogue = function () {

        var dict = {};

        this.clear = function () {
            dict = {};
        }

        this.add = function (node) {
            var id = node.getId();
            if (id) {
                dict[id] = node;
            }
        }

        this.get = function (id) {
            if (id in dict) {
                return dict[id];
            }
            return null;
        }

        this.remove = function (node) {
            var id = node.getId();
            if (id && (id in dict)) {
                if (node.objectCount !== dict[id].objectCount) {
                    return;
                }
                delete dict[id];
            }
        }

        this.logStat = function () {
            var count = 0;
            for (var k in dict) {
                if (dict.hasOwnProperty(k)) {
                    ++count;
                }
            }
            console.log('pgParserNodeCatalogue - # objects: ' + count);
        }
    }

    var pgParserNodeCatalogueInstance = window.pgParserNodeCatalogueInstance = new pgParserNodeCatalogue();

    /*
     if(!window.global.pgGlobalParserNodeCatalogueInstance) {
     window.global.pgGlobalParserNodeCatalogueInstance = new pgParserNodeCatalogue();
     } else {
     console.log("Using global pgParser Node Catalogue");
     }
     pgParserNodeCatalogueInstance = window.global.pgGlobalParserNodeCatalogueInstance;
     */

    /*
     pgCreateNodeFromHtml creates pgParserNode from the source html

     LIMITATION: only one element with subelements can be created with one call to this function.

     Examples:
     var pgel = pgCreateNodeFromHtml('<p>Hi</p>'); //this will work

     var pgel = pgCreateNodeFromHtml('<p>Hi</p><p>Bye</p>'); //this will FAIL because two nodes are being created

     var pgel = pgCreateNodeFromHtml('<div><p>Hi</p><p>Bye</p></div>'); //will WORK because there is only one root element

     To create more nodes simply call this function for each node:

     var pgel1 = pgCreateNodeFromHtml('<p>Hi</p>');
     var pgel2 = pgCreateNodeFromHtml('<p>Bye</p>');

     var pgdest = ... //element that will be replaced with pgel1 and pgel2
     pgdest.replaceWith(pgel1);
     pgel2.insertAfter(pgel2);

     */
    var pgGetAttributesStringFilterOutPgIds = window.pgGetAttributesStringFilterOutPgIds = function (node, name, value, quote) {
        if (name === 'data-pg-id') return null;
        if (value === null) return name;
        return name + '=' + quote + value + quote;
    }

    var pgCreateNodeFromHtml = window.pgCreateNodeFromHtml = function (html, assign_ids, multi) {
        //create node tree from html code
        var r = [];
        if (assign_ids === undefined) assign_ids = true;
        var p = new pgParser();
        p.assignIds = false;
        p.parse(html);

        if ((!multi && p.rootNode.children.length !== 1) || p.rootNode.children.length === 0) {
            throw new pgParserException("Create node from html failed. Only one node can be created.", html);
        }
        while (p.rootNode.children.length > 0) {
            var i = 0;
            var node = p.rootNode.children[i];
            if (assign_ids) {
                node.assignIdAndAddToCatalog(true);
                node.removePgIdAttributes();
            }
            node.detach();
            node.emitEvent("nodeCreated", {html: html});
            if (!multi) return node;
            r.push(node);
        }
        return r;
    }

    var pgCreateNodeSimple = window.pgCreateNodeSimple = function (html, p) {
        //create node tree from html code
        var r = [];
        p = p || new pgParser();
        p.assignIds = false;
        p.parse(html);

        if ((p.rootNode.children.length !== 1) || p.rootNode.children.length === 0) {
            throw new pgParserException("Create node from html failed. Only one node can be created.", html);
        }
        return p.rootNode.children[0];
    }

    var pgCreate = window.pgCreate = pgCreateNodeFromHtml;

    var pgCreateDocumentNode = function (nodeCatalogue) {
        var rootNode = new pgParserNode;
        rootNode.tagName = 'document';
        rootNode.rootNode = true;
        rootNode.nodeCatalogue = nodeCatalogue ? nodeCatalogue : pgParserNodeCatalogueInstance;
        rootNode.document = rootNode;
        return rootNode;
    }

    var startsWithCaseInsensitive = function (token, source, position) {
        //return source.startsWith(token, i);
        var len = source.length;
        for (var n = 0; n < token.length; n++) {
            if ((position + n >= len) || token.charAt(n) !== source.charAt(position + n).toLowerCase()) return false;
        }
        return true;
    }

    var startsWithCaseInsensitiveScript = function (source, position) {
        for (var i = 0; i < pgScriptsList.length; i++) {
            if (startsWithCaseInsensitive(pgScriptsList[i].open, source, position)) {
                return pgScriptsList[i];
            }
        }
        return null;
    }

    const pgAttr = 'data-pg-id'

    var pgParserNode = window.pgParserNode = function (html) {

        this.tagName = null;
        this.tagNameOriginal = null;
        this.textNode = false;
        this.rootNode = false;
        this.attributes = '';
        this.closingTag = null;
        this.closed = false;
        this.selfClosed = false;
        this.content = null;
        this.singleTag = false;
        this.script = false;
        this.isElementOrScript = false;
        this.script_info = null;
        this.scriptClosed = false;
        this.children = [];
        this.parent = null;
        this.comment = false;
        this.isElement = false;
        this.pgId = null;
        this.hasPgIdAttr = false;
        this.objectCount = ++pgObjectCount;
        this.data = {};
        this.isDeleted = false;
        this.formatHtml = false;

        this.attrList = null;
        this.attrListChanged = false;

        this.nodeCatalogue = pgParserNodeCatalogueInstance;
        this.document = null;
        this.clonedFromPgId = null;

        this.$el = null; //jQuery element

        this.startStringIndex = -1;
        this.endStringIndex = -1;

        this.outputOnly = null;
        this.dynamic = false;

        var _this = this;
    }

    //pgParserNode methods
    pgParserNode.prototype.setDocument = function (document) {
        this.walkSelfAndChildren(function (node) {
            node.document = document;
            node.catchedPage = null;
            return true;
        });
    }

    pgParserNode.prototype.getPage = function () {

        if(this.document && this.document.emulatedPage) return this.document.emulatedPage;

        if(!this.catchedPage) {
            var p = typeof pinegrow != 'undefined' ? pinegrow.getCrsaPageOfPgParserNode(this) : null;
            if (!p && this.removedFromPage) {
                p = this.removedFromPage;
            }
            this.catchedPage = p;
        }
        return this.catchedPage;
    }


    pgParserNode.prototype.getParent = function () {
        return this.parent;
    }

    pgParserNode.prototype.getParentsBySelector = function (sel) {
        var r = [];
        this.withParents(function (p) {
            if (p.isSelector(sel)) {
                r.push(p)
            }
        })
        return r;
    }

    pgParserNode.prototype.shouldHaveId = function () {
        //    if(this.script && (this.tagName == '?php' || this.tagName == '?=')) return true;
        if (this.script && pgIsScriptTag(this.tagName)) return true;
        return this.isElement && !this.script && !this.textNode && !this.comment;
    }

    pgParserNode.prototype.getByPgId = function (id) {
        var cat = this.nodeCatalogue ? this.nodeCatalogue : (this.document ? this.document.nodeCatalogue : null);
        if (cat) return cat.get(id);
        return null;
    }
    pgParserNode.prototype.withParents = function (func) {
        var pel = this.parent;
        while (pel) {
            func(pel);
            pel = pel.parent;
        }
    }

    pgParserNode.prototype.withSelfAndParents = function (func) {
        func(this);
        this.withParents(func);
    }

    pgParserNode.prototype.walk = function (func, depth) {
        depth = depth || 0;

        var len = this.children.length;
        for (var i = 0; i < len; i++) {
            var node = this.children[i];
            var r = func(node, depth);
            if (len !== this.children.length) {
                //children array changed during iteration
                var idx = this.children.indexOf(node);
                if (idx >= 0) {
                    i = idx;
                } else {
                    i--; //node was deleted
                }
                len = this.children.length;
            }
            if (!r) {
                return;
            } else if (r === 'skip_children') {
                continue;
            }
            node.walk(func, depth + 1);
        }
    }

    pgParserNode.prototype.walkSelfAndChildren = function (func) {
        if (!func(this, 0)) return;
        this.walk(func, 1);
    }

    pgParserNode.prototype.makeListByWalkingSelfAndChildren = function () {
        var list = [];
        this.walkSelfAndChildren(function (node) {
            if (node.isElementOrScript) list.push(node);
            return true;
        })
        return list;
    }

    pgParserNode.prototype.updatePositionInformation = function () {
        var c = 0;
        var _this = this;
        this.document.walkSelfAndChildren(function (node) {
            c++;
            node.positionInDocument = c;
            return true;
        })
    }

    pgParserNode.prototype.getPositionInDocument = function (fast) {
        if (fast && this.hasOwnProperty('positionInDocument')) return this.positionInDocument;

        var c = 0;
        var _this = this;
        this.document.walkSelfAndChildren(function (node) {
            c++;
            if (node === _this) return false;
            return true;
        })
        this.positionInDocument = c;
        return c;
    }

    pgParserNode.prototype.findOne = function (sel, stop_func) {
        var r = this.find(sel, true, false, stop_func);
        return r.length ? r[0] : null;
    }

    pgParserNode.prototype.findMaxItems = function (sel, max_items, stop_func, filter_func) {
        return this.find(sel, false, false, stop_func, max_items, filter_func);
    }

    var pgSelectorCacheCounter = 0;

    pgParserNode.prototype.find = function (sel, single, only_children, stop_func, max_items, filter_func) {
        pgSelectorCacheCounter++;

        if (pgSelectorCacheCounter > 100) {
            pgSelectorCache = {};
            pgSelectorCacheCounter = 0;
        }

        if (!max_items) max_items = 9999999999;

        var sel_list = sel.split(',');
        var results = [];
        for (var i = 0; i < sel_list.length; i++) {
            var path = sel_list[i].replace(/\s*\>\s*/g, ' >');
            path = path.split(' ');

            var tmppath = [];
            for (var j = 0; j < path.length; j++) {
                if (path[j].length > 0) tmppath.push(path[j]);
            }
            path = tmppath;

            var _this = this;

            if (path.length === 1 && $.trim(path[0]) === 'self') {
                results.push(this);
            } else {
                var c = 0;

                this.walk(function (node) {
                    node.positionInDocument = c++;

                    if (stop_func) {
                        if (stop_func(node)) return 'skip_children';
                    }
                    if (node.isElementOrScript && node.isTo(path, _this)) {
                        //if(!filter_func || filter_func(node)) {
                        if (results.indexOf(node) < 0) {
                            results.push(node);
                            if (max_items <= results.length) {
                                return false;
                            }

                        }
                        if (single) return false;
                        // }
                    }

                    return only_children ? 'skip_children' : true;
                });
            }
        }
        if (sel_list.length > 1) {
            //sort according to page order
            var col = new PgCollection();
            col.setList(results);
            col.orderByDocumentPosition(true);
            results = col.getList();
            col.setList(null);
        }
        return results;
    }

    pgParserNode.prototype.findIncludingSelf = function (sel, single, stop_func) {
        if (single) {
            if (this.isSelector(sel)) return this;
            return this.findOne(sel, stop_func);
        } else {
            var r = this.find(sel, false, false, stop_func);
            if (this.isSelector(sel)) r.unshift(this);
            return r;
        }
    }

    pgParserNode.prototype.findOneIncludingSelf = function (sel) {
        return this.findIncludingSelf(sel, true);
    }

    pgParserNode.prototype.findOneWithAttrValue = function (attr, value, tag) {
        tag = tag || '';
        var list = this.find(tag + '[' + attr + ']');
        for (var i = 0; i < list.length; i++) {
            if (list[i].getAttr(attr) === value) return list[i];
        }
        return null;
    }

    pgParserNode.prototype.findWithAttrValue = function (attr, value, tag) {
        tag = tag || '';
        var r = [];
        var list = this.find(tag + '[' + attr + ']');
        for (var i = 0; i < list.length; i++) {
            if (list[i].getAttr(attr) === value) r.push(list[i]);
        }
        return r;
    }

    pgParserNode.prototype.findWithMultipleAttrValues = function (list) {
        var r = [];
        this.walkSelfAndChildren(function (node) {
            if (node.isElement) {
                for (var i = 0; i < list.length; i++) {
                    if (list[i].value !== null) {
                        if (node.getAttribute(list[i].name) !== list[i].value) {
                            return true;
                        }
                    } else {
                        if (!node.hasAttribute(list[i].name)) {
                            return true;
                        }
                    }
                }
                r.push(node)
            }
            return true;
        })
        return r;
    }

    pgParserNode.prototype.goThroughAllUrls = function (func, just_this_element) {
        var attrs = ['src', 'href', 'action', 'data-src', 'srcset'];

        var task = function(node) {
            if (node.isElement) {
                attrs.forEach(function (a) {
                    var url = node.getAttr(a);
                    var new_url = null;
                    if (url) {
                        if (a === 'srcset') {
                            new_url = pgTransformSrcsetUrls(url, function (url) {
                                return func(url, a, node);
                            })
                        } else {
                            new_url = func(url, a, node);
                        }
                        if (new_url !== url) {
                            node.setAttr(a, new_url);
                        }
                    }
                })
            }
        }

        if(just_this_element) {
            task(this);
        } else {
            this.walkSelfAndChildren(function (node) {
                task(node);
                return true;
            })
        }
    }

    pgParserNode.prototype.rebaseAllUrls = function (base_url, filter) {
        //var node_url = require('url');

        this.goThroughAllUrls(function (url) {
            var new_url;
            if (filter) {
                new_url = filter(url);
                if (new_url !== url) return new_url;
            }
            new_url = url;

            try {
                new_url = new URL(url, base_url).href;
            } catch (err) {

            }
            //console.log(`${base_url} + ${url} = ${new_url}`)
            return new_url;//node_url.resolve(base_url, url);
        })
    }

    pgParserNode.prototype.mapLinksToNewLocation = function(old_url, new_url, only_if_exists) {

        var old_base_url = crsaGetBaseForUrl(old_url);
        var new_base_url = crsaGetBaseForUrl(new_url);

        if(old_base_url === new_base_url) {
            return false;
            //nothing to do
        }

        var changed = false;

        this.goThroughAllUrls(function(url, attr, pgel) {
            if(!url || crsaIsAbsoluteUrl(url) || url.startsWith('/') || url.startsWith('#')) {
                return url;

            } else {
                var abs_url = crsaGetAbsoluteUrlForRelativeUrlAndParentUrl(url, old_url);
                var rel_url = crsaMakeLinkRelativeTo(abs_url, new_url, true /* ignore too many backpaddles */);

                if(only_if_exists && rel_url !== url) {
                    if(!crsaIsFileExist( crsaMakeFileFromUrl(abs_url) )) {
                        rel_url = url; //dont change if it doesn't exist
                    }
                }
                changed = changed || (rel_url !== url);
                return rel_url;
            }
        }, true)

        return changed;
    }

    pgParserNode.prototype.validateTree = function (strict) {
        var invalid = [];
        this.walkSelfAndChildren(function (node) {
            if (!node.validate(strict)) invalid.push(node);
            return true;
        });
        return invalid;
    }

    var PG_CHAR_GREATER = '>'.charCodeAt(0);

    var pgNodeIsToHelperIsChildOf = function (parent, current, selector, base_selector) {
        var children = parent.children.filter(function (child) {
            return child.tagName !== "text"
        });
        if (children.length === 0) return false;
        if (selector === "first-child") {
            if (children[0] !== current) return false;
        }
        else if (selector === "last-child") {
            if (children[children.length - 1] !== current) return false;
        }
        else {
            var match = selector.match(/nth-child\(([0-9]*)\)/);
            if (match) {
                var childIndex = parseInt(match[1]);
                if (childIndex > children.length) {
                    return false;
                } else if (children[childIndex - 1] !== current) {
                    return false;
                }
            } else {
                var match = selector.match(/nth-of-type\(([0-9]*)\)/);
                if (match) {
                    //children = parent.children.filter(function(child) { return child.tagName === current.tagName });
                    children = parent.find('>' + base_selector);
                    var childIndex = parseInt(match[1]);
                    if (childIndex > children.length) {
                        return false;
                    } else if (children[childIndex - 1] !== current) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    var pgNodeIsToHelperGetSepArray = function (sel) {
        if (sel.indexOf(':') < 0) return false;
        var separatorArr = sel.split(':');
        if (separatorArr.length > 1) {
            return [true, separatorArr[0], separatorArr[1]];
        }
        else {
            return false;
        }
    }

    pgParserNode.prototype.isTo = function (path, node) {

        var currNode = this;
        var idx = path.length - 1;
        var next_immediate = true; //should next level match

        while (currNode && currNode !== node) {
            var last_immediate = false;
            if (idx >= 0) {
                var me = path[idx];
                if (me.charCodeAt(0) === PG_CHAR_GREATER) {
                    last_immediate = true;
                    me = me.substr(1);
                }
                var separatorArr = pgNodeIsToHelperGetSepArray(me);
                var selector = null;
                if (separatorArr !== false) {
                    me = separatorArr[1];
                    selector = separatorArr[2];
                }

                if (currNode.isSelector(me)) {
                    if (selector) {
                        if (!pgNodeIsToHelperIsChildOf(currNode.parent, currNode, selector, me)) return false;
                    }

                    idx--;
                } else {
                    if (next_immediate) return false;
                }
            } else {
                if (next_immediate) return false;
            }

            next_immediate = last_immediate;
            currNode = currNode.parent;
        }

        if (idx >= 0) {
            //still have unchecked parts, but parent chain is finished
            return false;
        }

        if (currNode) return true;
        else return false;

        // var _this = this;
        // var me = path[path.length-1];
        // var immediate = false;
        // if(me.charCodeAt(0) === PG_CHAR_GREATER) {
        //     immediate = true;
        //     me = me.substr(1);
        // }
        // var separatorArr = getSepArray(me);
        // var selector;
        // if (separatorArr[0]) {
        //     me = separatorArr[1];
        //     selector = separatorArr[2];
        // }
        // if(this.isSelector(me)) {
        //     if(path.length < 2) {
        //         if (selector) {
        //             if (!isChildOf(this.getParent(), this, selector)) return false;
        //         }
        //         return true;
        //     }
        //     var idx = path.length - 2;
        //     var parent = this.getParent();
        //     var current = this;
        //     while(parent && parent !== node) {
        //         var sel = path[idx];
        //         var next_immediate = false;
        //         if(sel.charCodeAt(0) === PG_CHAR_GREATER) {
        //             next_immediate = true;
        //             sel = sel.substr(1);
        //         }
        //         var pSepArray = getSepArray(sel);
        //         var pSelector;
        //         if (pSepArray[0]) {
        //             sel = pSepArray[1];
        //             pSelector = pSepArray[2];
        //         }
        //         if(parent.isSelector(sel)) {
        //             if (selector) {
        //                 if (!isChildOf(parent, this, selector)) return false;
        //             }
        //             idx--;
        //             if(idx <= 0) {
        //                 if (pSelector) {
        //                     current = parent;
        //                     parent = current.getParent();
        //                     if (!isChildOf(parent, current, pSelector)) return false;
        //                 }
        //                 return true;
        //             }
        //         } else {
        //             if(immediate) return false;
        //         }
        //         parent = parent.getParent();
        //         immediate = next_immediate;
        //         selector = undefined;
        //     }
        //     return false;
        // } else {
        //     return false;
        // }
    }

    pgParserNode.prototype.is = function (sel) {
        var sel_array = sel.split(',');

        for (var i = 0; i < sel_array.length; i++) {
            var fixedEl = sel_array[i].trim();
            fixedEl = fixedEl.replace(/\s*\>\s*/g, ' >');
            var find_sel = fixedEl.split(' ');
            if (this.isTo(find_sel, this.document)) {
                return true;
            }
        }
        return false;
    }

    pgParserNode.prototype.parseSelector = function (sel) {
        var sel_orig = sel;
        if (!pgSelectorCache[sel_orig]) {
            var m = sel.match(/\[([^\]]+)\]/);
            var attr = null;
            var attrValue = null;
            if (m) {
                attr = m[1];
                sel = sel.replace(m[0], '');

                var attrArr = attr.split('=');
                attr = attrArr[0];
                if (attrArr.length > 1) {
                    attrValue = attrArr[1].replace(/(\"|\')/g, '');
                }
            }

            var id_m = sel.match(/#[^\s\.]+/);
            var tag_m = sel.match(/^[^\s\.#]+/);

            pgSelectorCache[sel_orig] = {
                id: id_m ? id_m[0].replace('#', '') : null,
                tag: tag_m ? tag_m[0] : null,
                classes: sel.match(/\.[^\s\.#]+/g),
                attr: attr,
                attrValue: attrValue
            }
        }
        return pgSelectorCache[sel_orig];
    }

    pgParserNode.prototype.isSelector = function (sel, parsed_sel) {
        if (!parsed_sel) {
            if (sel.indexOf(',') >= 0) {
                //handle multiple , separated selectors
                var list = crsaSplitAndTrimList(sel);
                for (var i = 0; i < list.length; i++) {
                    if (this.isSelector(list[i])) return true;
                }
                return false;
            }
        }
        parsed_sel = parsed_sel || this.parseSelector(sel);

        if (parsed_sel.id) {
            if (this.getElementId() !== parsed_sel.id) return false;
        }
        if (parsed_sel.tag) {
            if (parsed_sel.tag.length > 0 && parsed_sel.tag !== '*') {
                if (parsed_sel.tag !== this.tagName) return false;
            }
        }
        if (parsed_sel.classes) {
            for (var i = 0; i < parsed_sel.classes.length; i++) {
                if (!this.hasClass(parsed_sel.classes[i].replace('.', ''))) return false;
            }
        }
        if (parsed_sel.attr) {
            if (this.hasAttr(parsed_sel.attr)) {
                if (parsed_sel.attrValue !== null && this.getAttr(parsed_sel.attr) !== parsed_sel.attrValue) return false;
            } else {
                return false;
            }
        }
        return true;
    }


    pgParserNode.prototype.closest = function (sel, stop_at) {
        var _this = this;
        var sel_list = crsaSplitAndTrimList(sel);
        var parsed_sels = sel_list.map(function (item) {
            return _this.parseSelector(item)
        });
        ;
        var r = this;

        var isAny = function () {
            for (var i = 0; i < sel_list.length; i++) {
                if (r.isSelector(sel_list[i], parsed_sels[i])) return true;
            }
            return false;
        }

        //    while(!r.isSelector(sel, parsed_sel)) {
        while (!isAny()) {
            if (!r.parent) return null;
            if (stop_at && stop_at === r) return null;
            r = r.parent;
        }
        return r;
    }


    //Methods with operations
    pgParserNode.prototype.addChild = function (node) {

        // this.withEmitEvent(function() {
        this.children.push(node);
        node.parent = this;
        node.setDocument(this.document);
        // }, 'addChild', node)

        //this.emitEvent('addChild', node);
        /*
         var o = this.makeOperation('addChild', node,
         function(o) { //undo
         o.getElement().removeChild(o.getObject());
         },
         function(o) { //redo
         o.getElement().addChild(o.getObject());
         }
         );
         */
    }

    pgParserNode.prototype.removeChild = function (node) {

        this.withEmitEvent(function () {
            var idx = this.getChildPos(node);
            if (idx >= 0) {
                this.children.splice(idx, 1);
            }
            node.parent = null;
            node.setDocument(null);

        }, 'removeChild', node);

        /*
         var o = this.makeOperation('removeChild', node,
         function(o) { //undo
         o.getElement().addChild(o.getObject());
         },
         function(o) { //redo
         o.getElement().removeChild(o.getObject());
         }
         );
         */
    }

    pgParserNode.prototype.removeChildWithoutEvent = function (node) {

        var idx = this.getChildPos(node);
        if (idx >= 0) {
            this.children.splice(idx, 1);
        }
        node.parent = null;
        node.setDocument(null);
    }

    pgParserNode.prototype.replaceTag = function (tag) {
        var pgel = null;
        var oldHtml = this.getEmitEvents() ? this.toStringWithIds() : null;
        var pgid = this.getId();

        this.withoutEvents(function () {
            this.tagName = tag;
            this.tagNameOriginal = tag;
            if (this.closingTag) {
                this.closingTag = tag;
            }
            pgel = this.clone(null, false);
            this.replaceWith(pgel);
        })
        pgel.update(oldHtml, pgid);
        return pgel;

        /*
         var o = this.makeOperation('replaceTag', {tag: tag, currentTag: this.tagName, currentTagOriginal: this.tagNameOriginal, currentClosingTag: this.closingTag},
         function(o) { //undo
         var pgel = o.getElement();
         pgel.tagName = o.data.currentTag;
         pgel.tagNameOriginal = o.data.currentTagOriginal;
         pgel.closingTag = o.data.currentClosingTag;
         },
         function(o) { //redo
         o.getElement().replaceTag(o.data.tag);
         }
         );
         */

        /*
         this.withEmitEvent(function() {
         this.tagName = tag;
         this.tagNameOriginal = tag;
         if(this.closingTag) {
         this.closingTag = tag;
         }

         }, 'replaceTag', {tag: tag});

         return this;
         */
    }

    pgParserNode.prototype.replaceWith = function (node, detach) {
        if (!this.parent) {
            //keep track anyway
            this.replacedWith = node;
            if (!this.replacedWithChain) {
                this.replacedWithChain = []
            }
            this.replacedWithChain.push(node);
            return this;
        } //nothing to do

        var page = this.getPage();

        this.withEmitEvent(function () {
            node.insertBefore(this);
            this.replacedWith = node;
            if (!this.replacedWithChain) {
                this.replacedWithChain = []
            }
            this.replacedWithChain.push(node);

            node.data = this.data; //copy data object to new node
            if (detach) {
                this.detach();
            } else {
                this.remove();
            }
        }, 'replaceWith', node, {
            detach: detach,
            objectHtml: node.toStringWithIds(),
            objectDomHtml: node.getViewHtml(page),
            oldValue: function () {
                return this.toStringWithIds(true);
            }
        }, page);

        return this;
    }

    pgParserNode.prototype.replaceContentWithContentOf = function (node, detach) {
        this.withEmitEvent(function () {
            while (this.children.length) {
                if (detach) {
                    this.children[0].detach();
                } else {
                    this.children[0].remove();
                }
            }
            while (node.children.length) {
                this.append(node.children[0]);
            }

        }, 'replaceContentWithContentOf', node, {detach: detach});
    }

    pgParserNode.prototype.replaceContentWithElement = function (node, detach) {
        this.withEmitEvent(function () {
            while (this.children.length) {
                if (detach) {
                    this.children[0].detach();
                } else {
                    this.children[0].remove();
                }
            }
            this.append(node);
        }, 'replaceContentWithElement', node, {detach: detach});
    }

    pgParserNode.prototype.replaceWithContentOf = function (node, detach) {
        this.withEmitEvent(function () {
            while (node.children.length) {
                node.children[0].insertBefore(this);
            }
            if (detach) {
                this.detach();
            } else {
                this.remove();
            }

        }, 'replaceWithContentOf', node, {detach: detach});
    }

    pgParserNode.prototype.detag = function () {
        this.withEmitEvent(function () {
            if (this.parent) {
                while (this.children.length) {
                    this.children[0].insertBefore(this);
                }
                this.remove();
            }
        }, 'detag', null, {
            oldValue: function () {
                return {
                    parentHtml: this.parent ? this.parent.html(null, true) : null,
                    parentId: this.parent ? this.parent.getId() : null
                };
            }
        }, this.getPage());
    }

    pgParserNode.prototype.getHeadingLevel = function () {
        if (this.tagName) {
            var m = this.tagName.match(/^h([0-9])+$/i);
            if (m) return parseInt(m[1]);
        }
        return 0;
    }

    pgParserNode.prototype.getHeadingLevelOfSection = function () {
        var node = this;
        while (true) {
            var level = node.getHeadingLevel();
            if (level > 0) return level;

            var prev = node.prev();
            if (prev) {
                node = prev;
            } else {
                node = node.parent;
            }
            if (!node) return 0;
        }
    }

    //will add_level to all headings
    pgParserNode.prototype.rebaseHeadingLevel = function (add_level) {
        this.walkSelfAndChildren(function (node) {
            var l = node.getHeadingLevel();
            if (l > 0) {
                var tag = 'h' + (l + add_level);
                node.tagName = tag;
                node.tagNameOriginal = tag;
                if (node.closingTag) {
                    node.closingTag = tag;
                }
            }
            return true;
        })
    }

    pgParserNode.prototype.callDefHandlers = function (handler, a, b, c, d, cp) {
        cp = cp || this.getPage();
        if (!cp) return;
        var defs = cp.getAllTypes(null, this);
        defs.forEach(function (def) {
            if (handler in def) {
                def[handler](a, b, c, d);
            }
        })
        return defs;
    }

    pgParserNode.prototype.callDefHandlersOnParents = function (handler, a, b, c, d, cp) {
        cp = cp || this.getPage();
        if (!cp) return;

        this.withParents(function (p) {
            p.callDefHandlers(handler, a === -1 ? p : a, b, c, d);
        })
    }

    pgParserNode.prototype.getUniqueIdAttribute = function (pref, prev) {
        return getUniqueId(pref, this.getPage(), prev);
    }

    pgParserNode.prototype.clone = function (dont_asign_id, copy_ids, no_children) {
        var node = new pgParserNode();
        node.tagName = this.tagName;
        node.tagNameOriginal = this.tagNameOriginal;
        node.textNode = this.textNode;
        node.rootNode = this.rootNode;
        node.attributes = this.getAttributesString();
        node.closingTag = this.closingTag;
        node.closed = this.closed;
        node.selfClosed = this.selfClosed;
        node.content = this.content;
        node.singleTag = this.singleTag;
        node.script = this.script;
        node.isElementOrScript = this.isElementOrScript;
        node.script_info = this.script_info;
        node.scriptClosed = this.scriptClosed;
        node.comment = this.comment;
        node.outputOnly = this.outputOnly;
        node.formatHtml = this.formatHtml;

        if (!dont_asign_id) {
            node.document = this.document;
        } else {
            node.document = node;
        }

        node.nodeCatalogue = this.nodeCatalogue || pgParserNodeCatalogueInstance;
        node.isElement = this.isElement;

        node.attrList = null;
        node.hasPgIdAttr = this.hasPgIdAttr;
        node.data = this.data;

        if (!dont_asign_id && node.shouldHaveId()) {
            node.assignId();
            node.nodeCatalogue.add(node);
        }
        node.clonedFromPgId = this.getId();

        if(this.dynamic) {
            node.page = this.page || null;
            node.dynamic = true;
        }

        if (copy_ids) {
            node.pgId = node.clonedFromPgId;
        }


        if (!no_children) {
            for (var n = 0; n < this.children.length; n++) {
                node.addChild(this.children[n].clone(dont_asign_id, copy_ids));
            }
        }

        return node;
    }

    pgParserNode.prototype.setData = function (key, value) {
        this.data[key] = value;
    }

    pgParserNode.prototype.getData = function (key, default_value) {
        return this.data[key] || default_value || null;
    }

    pgParserNode.prototype.getClonedFromPgEl = function () {
        if (this.clonedFromPgId) {
            return getPgNodeByPgId(this.clonedFromPgId);
        }
        return null;
    }

    pgParserNode.prototype.removeAllChildren = function () {
        this.withEmitEvent(function () {
            while (this.children.length) {
                this.children[0].remove();
            }

        }, 'removeAllChildren');
    }

    pgParserNode.prototype.html = function (html, withIds, formatOptions, dont_assign_ids) {
        if (typeof html == 'undefined' || html === null) {
            if (this.script || this.textNode) return this.content;
            if (this.comment) return this.attributes;
            if (this.tagName === 'script' || this.tagName === 'php' || this.tagName === 'style') return this.content;

            var s = '';
            for (var i = 0; i < this.children.length; i++) {
                s += this.children[i].toString('', true, withIds, null, formatOptions ? true : false, formatOptions);
            }
            return s;
        }
        var oldValue = null;

        if (this.getEmitEvents()) {
            oldValue = this.html(null, true);
        }
        this.withoutEvents(function () {
            if (this.script) {
                this.content = html;
            } else if (this.tagName === 'script' || this.tagName === 'php' || this.textNode || this.tagName === 'style') {
                this.content = html;
            } else if (this.comment) {
                this.attributes = html;
            } else {

                var p = new pgParser();
                if (dont_assign_ids) {
                    p.assignIds = false;
                }
                p.parse(html);
                while (this.children.length) {
                    this.children[0].remove();
                }
                for (var i = 0; i < p.rootNode.children.length; i++) {
                    this.addChild(p.rootNode.children[i]);
                }
            }
        })
        /*, 'html', {
         html: html,
         domHtml: this.getViewInnerHtml(),
         oldValue: function() {
         return this.html(null, true)
         }});*/

        if (this.getEmitEvents()) {
            this.updateInner(oldValue);
        }

        return this;
    }

    pgParserNode.prototype.text = function (html, immediate) {
        if (typeof html == 'undefined' || html === null) {
            var s = '';
            for (var i = 0; i < this.children.length; i++) {
                if (this.children[i].tagName === 'text') {
                    s += this.children[i].toString('', true, false);
                } else if (!immediate) {
                    s += this.children[i].text();
                }
            }
            return s;
        }
        this.html(html); //will emit events
    }

    pgParserNode.prototype.indentText = function (text, indent, level, html_options) {
        if (indent === null) {
            if (!html_options) html_options = pinegrow.getFormatHtmlOptions();
            if (level === null) level = this.getNestingLevel();
            indent = pinegrow.getHtmlIndentForLevel(level, html_options);
        }

        var lines = text.split("\n");
        var s = '';

        for (var i = 0; i < lines.length; i++) {
            if (i > 0) s += '\n';
            s += (i === 0) ? lines[i] : indent + lines[i];
        }
        return s;
    }


    pgParserNode.prototype.isDescendantOf = function (ancestor) {
        var p = this.parent;
        while (p != null) {
            if (p === ancestor) {
                return true;
            }
            p = p.parent;
        }
        return false;
    }

    pgParserNode.prototype.contains = function (desc) {
        return desc.isDescendantOf(this);
    }

    pgParserNode.prototype.getChildPos = function (child, skip_text) {
        if (!skip_text) return this.children.indexOf(child);
        var idx = -1;
        for (var i = 0; i < this.children.length; i++) {
            if (this.children[i].isElementOrScript) {
                idx++;
                if (this.children[i] === child) return idx;
            }
        }
        return -1;
    }

    pgParserNode.prototype.index = function () {
        if (!this.parent) return 0;
        return this.parent.getChildPos(this, true);
    }

    pgParserNode.prototype.getChildAtPos = function (pos, skip_text) {
        var idx = -1;
        for (var i = 0; i < this.children.length; i++) {
            if (!skip_text || this.children[i].isElementOrScript) {
                idx++;
            }
            if (idx === pos) return this.children[i];
        }
        return null;
    }

    pgParserNode.prototype.getFirstChild = function (incl_text) {
        if (incl_text === undefined) incl_text = false;
        return this.getChildAtPos(0, !incl_text)
    }

    pgParserNode.prototype.isChildOf = function (node) {
        return this.parent === node;
    }

    pgParserNode.prototype.isEmpty = function () {
        if(this.children.length === 0) return true;
        for(var i = 0; i < this.children.length; i++) {
            if(this.children[i].isElementOrScript) return false;
        }
        return this.text().trim().length === 0;
    }

    pgParserNode.prototype.isTextContentEmpty = function () {
        return this.textNode && this.content.match(pgWhiteSpaceRegExp);
    }

    pgParserNode.prototype.hasChildren = function (all) {
        for (var i = 0; i < this.children.length; i++) {
            if (this.children[i].isElementOrScript || all) {
                return true;
            }
        }
        return false;
    }

    pgParserNode.prototype.getChildren = function (all) {
        var r = [];
        for (var i = 0; i < this.children.length; i++) {
            if (this.children[i].isElementOrScript || all) {
                r.push(this.children[i]);
            }
        }
        return r;
    }

    pgParserNode.prototype.getLinkElement = function () {
        if(this.tagName === 'a') return this;
        var l = this.getChildren();
        if(l.length === 1 && l[0].tagName === 'a') {
            return l[0];
        } else if(this.parent && this.parent.tagName === 'a') {
            return this.parent;
        }
        return null;
    }

    pgParserNode.prototype.hasTextOnly = function (all) {
        for (var i = 0; i < this.children.length; i++) {
            if (!this.children[i].textNode) {
                return false;
            }
        }
        return this.children.length > 0;
    }

    pgParserNode.prototype.findElementWithTextOnly = function () {
        var text_pgel = null;

        this.walkSelfAndChildren(function (node) {
            if (node.hasTextOnly()) {
                text_pgel = node;
                return false;
            }
            return true;
        })

        return text_pgel;
    }

    pgParserNode.prototype.isTheOnlyChild = function (all) {
        return this.parent && this.parent.getChildren(all).length === 1;
    }

    pgParserNode.prototype.next = function (include_text) {
        if (!this.parent) return null;
        var idx = this.parent.getChildPos(this) + 1;
        var next = null;
        while (idx < this.parent.children.length && !next) {
            if (this.parent.children[idx].isElementOrScript || include_text) {
                next = this.parent.children[idx];
                break;
            }
            idx++;
        }
        return next;
    }

    pgParserNode.prototype.prev = function (include_text) {
        if (!this.parent) return null;
        var idx = this.parent.getChildPos(this) - 1;
        var prev = null;
        while (idx >= 0 && !prev) {
            if (this.parent.children[idx].isElementOrScript || include_text) {
                prev = this.parent.children[idx];
                break;
            }
            idx--;
        }
        return prev;
    }

    pgParserNode.prototype.prevWithSelector = function (sel) {
        var prev = this.prev();
        while (prev) {
            if (prev.isSelector(sel)) return prev;
            prev = prev.prev();
        }
        return prev;
    }

    pgParserNode.prototype.nextWithSelector = function (sel) {
        var next = this.next();
        while (next) {
            if (next.isSelector(sel)) return next;
            next = next.next();
        }
        return next;
    }

    pgParserNode.prototype.first = function (include_text) {
        var idx = 0;
        while (idx < this.children.length) {
            if (this.children[idx].isElementOrScript || include_text) {
                return this.children[idx];
                break;
            }
            idx++;
        }
        return null;
    }

    pgParserNode.prototype.last = function (include_text) {
        var idx = this.children - 1;
        while (idx >= 0) {
            if (this.children[idx].isElementOrScript || include_text) {
                return this.children[idx];
                break;
            }
            idx--;
        }
        return null;
    }

    pgParserNode.prototype.indexToElementIndex = function (idx) {
        var el_idx = -1;
        for (var i = 0; i < this.children.length; i++) {
            if (this.children[i].shouldHaveId()) {
                el_idx++;
                if (idx === i) break;
            } else if (idx === i) {
                el_idx++;
                break;
            }

        }
        return el_idx === -1 ? 0 : el_idx;
    }

    pgParserNode.prototype.insertAtIndex = function (parent, idx, skip_text) {
        //console.log('BBB ', idx, skip_text);

        this.withEmitEvent(function () {
            var use_idx = idx;

            if (parent === this.parent && !skip_text) {
                var i = this.parent.children.indexOf(this);
                if (i >= 0 && i < idx) use_idx--;
            }
            this.detach();

            if (skip_text && use_idx > 0) {
                var num_non_text = 0;
                var nidx = use_idx;
                for (var i = 0; i < parent.children.length; i++) {
                    if (!parent.children[i].shouldHaveId()) {
                        nidx++;
                    } else {
                        num_non_text++;
                        if (use_idx === num_non_text) break;
                    }
                }
                use_idx = nidx;
            }
            if (use_idx >= parent.children.length) {
                parent.children.push(this);
            } else {
                parent.children.splice(use_idx, 0, this);
            }
            this.parent = parent;
            this.setDocument(this.parent.document);

        }, 'insertAtIndex', parent, {
            index: idx,
            element_index: parent.indexToElementIndex(idx),
            skip_text: skip_text
        }, parent.getPage());

        return this;
    }

    pgParserNode.prototype.insert = function (dest, before) {
        if (!dest.parent) return;//throw new pgParserException("Object has no parent", dest);
        var idx = dest.parent.getChildPos(dest);
        if (idx < 0) {
            throw new pgParserException("Node is not in parent child list", dest);
        } else {
            this.withEmitEvent(function () {
                if (before) {
                    this.insertAtIndex(dest.parent, idx);
                } else {
                    this.insertAtIndex(dest.parent, idx + 1);
                    /*
                     var next = dest.next();
                     if(!next) {
                     dest.parent.append(this);
                     } else {
                     idx = dest.parent.getChildPos(next);
                     this.insertAtIndex(dest.parent, idx);
                     }*/
                }
            }, 'insert', dest, {before: before}, dest.getPage());
        }

        return this;
    }

    pgParserNode.prototype.detach = function () {
        this.withEmitEvent(function () {
            if (this.parent) {
                this.parent.removeChild(this);
            }
        }, 'detach', null, {}, this.getPage());
    }

    //Moves children of this into pgel
    pgParserNode.prototype.moveChildrenToElement = function (pgel, prepend, trim) {
        if (this === pgel || pgel.isDescendantOf(this)) return;

        if (this.getEmitEvents()) {
            var pgel_old = pgel.html(null, true);
            var this_old = this.html(null, true);
        }

        var _this = this;

        pgel.withoutEvents(function () {
            if(trim) {
                //trim front
                while(_this.children.length && _this.children[0].textNode && _this.children[0].content.trim().length === 0) {
                    _this.children[0].remove();
                }

                //trim back
                for(var i = _this.children.length - 1; i >= 0; i--) {
                    if(_this.children[i].textNode && _this.children[i].content.trim().length === 0) {
                        _this.children[i].remove();
                    } else {
                        break;
                    }
                }

                //trim first
                if(_this.children.length && _this.children[0].textNode) {
                    _this.children[0].content = _this.children[0].content.trimLeft();
                }
                if(_this.children.length && _this.children[_this.children.length - 1].textNode) {
                    _this.children[_this.children.length - 1].content = _this.children[_this.children.length - 1].content.trimRight();
                }
            }
            if (prepend) {
                var last = null;
                while (_this.children.length) {
                    var node = _this.children[0];
                    if (!last) {
                        pgel.prepend(node);
                    } else {
                        node.insertAfter(last);
                    }
                    last = node;
                }
            } else {
                while (_this.children.length) {
                    var node = _this.children[0];
                    pgel.append(node);
                }
            }
        })
        if (this.getEmitEvents()) {
            pgel.updateInner(pgel_old);
            this.updateInner(this_old);
        }
    }

    pgParserNode.prototype.getPositionInfo = function () {
        var r = null;

        if (this.parent && !this.parent.rootNode && this.parent.getId()) {
            r = {
                parent: this.parent.getId(),
                pos: this.parent.getChildPos(this)
            }
            var after = this.prev(true);
            if (after) r.afterId = after.getId();

            var before = this.next(true);
            if (before) r.beforeId = before.getId();
        }
        return r;
    }

    pgParserNode.prototype.placeAtPositionInfo = function (pos) {
        if (pos) {
            var parent = pgParserNodeCatalogueInstance.get(pos.parent);
            if (parent) {
                if (pos.afterId) {
                    var after = pgParserNodeCatalogueInstance.get(pos.afterId);
                    if (after && parent === after.parent) {
                        this.insertAfter(after);
                        return;
                    }
                }
                if (pos.beforeId) {
                    var before = pgParserNodeCatalogueInstance.get(pos.beforeId);
                    if (before && parent === before.parent) {
                        this.insertBefore(before);
                        return;
                    }
                }
                this.insertAtIndex(parent, pos.pos);
            }
        }
    }

    pgParserNode.prototype.removeWithoutEvents = function (sub, ids) {
        if (!ids) ids = [];
        if (!sub) this.removedFromPage = this.getPage();
        if (ids) {
            var id = this.getId();
            if (id) ids.push(id);
        }

        //this.detach(); this triggers events!
        if (this.parent) {
            //this.parent.removeChild(this);
            this.parent.removeChildWithoutEvent(this);
        }

        if (this.nodeCatalogue) this.nodeCatalogue.remove(this);
        this.isDeleted = true;
        while (this.children.length) {
            this.children[0].removeWithoutEvents(true, ids);
        }
    }

    pgParserNode.prototype.remove = function (sub, ids) {
        if (!ids) ids = [];
        if (!this.nonElementChangesContainer(this.parent, function () {
            this.removeWithoutEvents(sub, ids);
        })) {
            this.withEmitEvent(function () {

                this.removeWithoutEvents(sub, ids);

            }, 'remove', null, {
                removedIds: ids,
                oldValue: function () {
                    return {
                        position: this.parent ? this.parent.getChildPos(this) : null,
                        html: this.toStringWithIds(),
                        parentId: this.parent ? this.parent.getId() : null,
                        positionInfo: this.getPositionInfo()
                    }
                }
            }, this.getPage());
        }
    }

    pgParserNode.prototype.insertBefore = function (dest) {
        if (!this.nonElementChangesContainer(dest.parent, function () {
            this.insert(dest, true);
        })) {
            this.withEmitEvent(function () {
                this.insert(dest, true);
            }, 'insertBefore', dest, {
                oldValue: function () {
                    return {position: this.getPositionInfo()}
                }
            }, dest.getPage());
        }
        return this;
    }

    pgParserNode.prototype.insertAfter = function (dest) {
        if (!this.nonElementChangesContainer(dest.parent, function () {
            this.insert(dest, false);
        })) {
            this.withEmitEvent(function () {
                this.insert(dest, false);
            }, 'insertAfter', dest, {
                oldValue: function () {
                    var d = {
                        position: this.getPositionInfo()
                    }
                    if (!this.isElement) {
                        d.destHtml = dest.parent ? dest.parent.html(null, true) : null;
                        d.parentHtml = this.parent ? this.parent.html(null, true) : null;
                    }
                    return d;
                }
            }, dest.getPage());
        }
        return this;
    }

    pgParserNode.prototype.nonElementChangesContainer = function (dest, func) {
        if (!this.isElement && this.getEmitEvents() && dest) {
            var oldValue = dest.html(null, true);
            this.withoutEvents(function () {
                func.call(this);
            })
            dest.updateInner(oldValue);
            return true;
        }
        return false;
    }

    pgParserNode.prototype.appendPrependWithoutEvent = function (dest, prepend) {
        this.detach();
        if (prepend) {
            dest.children.unshift(this);
        } else {
            dest.children.push(this);
        }
        this.parent = dest;
        this.setDocument(this.parent.document);
    }

    pgParserNode.prototype.appendPrepend = function (dest, prepend) {

        if (!this.nonElementChangesContainer(dest, function () {
            this.appendPrependWithoutEvent(dest, prepend);
        })) {
            this.withEmitEvent(function () {

                this.appendPrependWithoutEvent(dest, prepend);

            }, 'appendPrepend', dest, {
                prepend: prepend, oldValue: function () {
                    var d = {
                        position: this.getPositionInfo()
                    }
                    if (!this.isElement) {
                        d.destHtml = dest.html(null, true);
                        d.parentHtml = this.parent ? this.parent.html(null, true) : null;
                    }
                    return d;

                }
            }, dest.getPage());
        }

        return this;
    }

    pgParserNode.prototype.indent = function () {
        var prev = this.prev();
        if (!prev || !prev.canHaveChildren()) return false;
        this.appendTo(prev);
        return true;
    }

    pgParserNode.prototype.outdent = function () {
        var prev = this.prev();
        var next = this.next();
        if (next || !this.parent) return false;
        this.insertAfter(this.parent);
        return true;
    }

    pgParserNode.prototype.canHaveChildren = function () {
        return this.isElement && !this.singleTag && this.tagName !== 'script';
    }

    pgParserNode.prototype.canDuplicate = pgParserNode.prototype.canDelete = function () {
        return this.canHaveSiblings();
    }

    pgParserNode.prototype.canHaveSiblings = function () {
        return ['html', 'body', 'head'].indexOf(this.tagName) < 0;
    }

    pgParserNode.prototype.canMove = function() {
        return this.canHaveSiblings();
    }

    pgParserNode.prototype.canDelete = function() {
        return this.canHaveSiblings();
    }

    pgParserNode.prototype.isCollapsed = function() {
        if (this.tagName == 'head') return !this.getPage().show_head_in_tree;
        var val = this.getAttribute('data-pg-collapsed', -1);
        if(val !== -1) {
            return val !== 'false';
        }
        return this.tagName === 'svg'; //collapsed by default
    }

    /*

     0 a
     1 b
     2 c
     3 d

     p = 2

     C
     a
     b
     d

     */
    pgParserNode.prototype.appendTo = function (dest) {
        return this.appendPrepend(dest, false); //will emit event
    }

    pgParserNode.prototype.prependTo = function (dest) {
        return this.appendPrepend(dest, true); //will emit event
    }

    pgParserNode.prototype.append = function (node) {
        node.appendTo(this); //will emit event
    }

    pgParserNode.prototype.prepend = function (node) {
        node.prependTo(this); //will emit event
    }

    pgParserNode.prototype.getClasses = function () {
        var classes = this.getAttr('class');
        if (classes) {
            //var a = classes.split(/\s/);
            var a = splitCodeAware(classes, ' ');
            var map = {};
            var r = [];
            a.forEach(function (cls) {
                if (cls.length && !map[cls]) {
                    r.push(cls);
                    map[cls] = true;
                }
            })
            return r;
        }
        return [];
    }

    pgParserNode.prototype.hasClass = function (cls) {
        var list = this.getClasses();
        var idx = list.indexOf(cls);
        if (idx >= 0) return true;
        return false;
    }

    pgParserNode.prototype.addClass = function (cls, class_sorter) {
        this.withEmitEvent(function () {
            var list = this.getClasses();
            var idx = list.indexOf(cls);
            if (idx < 0) {
                list.push(cls);
                if(class_sorter) {
                    list = class_sorter(list);
                } else {
                    list.sort();
                }
                this.setAttr('class', list.join(' '));
            }
        }, 'addClass', {class: cls});
        //this.emitEvent('addClass', {class: cls});
    }

    pgParserNode.prototype.removeClass = function (cls, class_sorter) {
        this.withEmitEvent(function () {
            var list = this.getClasses();
            var r = [];
            for (var i = 0; i < list.length; i++) {
                if (list[i] !== cls) {
                    r.push(list[i]);
                }
            }
            if (r.length) {
                if(class_sorter) {
                    list = class_sorter(list);
                } else {
                    list.sort();
                }
                this.setAttr('class', r.join(' '));
            } else {
                this.removeAttr('class');
            }
        }, 'removeClass', {class: cls, hadClass: this.hasClass(cls)});
    }

    pgParserNode.prototype.removeClassesStartingWith = function (cls) {
        var list = this.getClasses();
        var r = [];
        for (var i = 0; i < list.length; i++) {
            if (list[i].startsWith(cls)) {
                r.push(list[i]);
            }
        }
        var _this = this;
        r.forEach(function (cls) {
            _this.removeClass(cls);
        })
    }

    pgParserNode.prototype.canAddClass = function (cls) {
        return true;
    }

    pgParserNode.prototype.canRemoveClass = function (cls) {
        //if source node doesn't have this class then we can't remove it as it was added dynamically
        return this.hasClass(cls);
    }

    //console.log(_this.attributes);


    var indexOfCodeAware = function (str, what, from) {
        //return str.indexOf(what, from);
        var scripts = pgAttrScripts;
        var len = str.length;
        var idx = -1;
        var n = from || 0;
        while (n < len) {
            var ch = str.charAt(n);
            if (ch === '<') {
                for (var si = 0; si < scripts.length; si++) {
                    if (startsWithCaseInsensitive(scripts[si][0], str, n)) {
                        //script starts
                        var end_idx = str.indexOf(scripts[si][1], n + scripts[si][0].length);
                        if (end_idx < 0) {
                            return idx;
                        }
                        //found end
                        n = end_idx + scripts[si][1].length;
                    }
                }
            }
            if (str.startsWith(what, n)) {
                return n;
            }
            n++;
        }
        return idx;
    }

    var splitCodeAware = function (str, split_str) {
        var r = [];
        var i = 0;
        while (i < str.length) {
            var next = indexOfCodeAware(str, split_str, i);
            if (next < 0) next = str.length;
            if (next - i > 0) {
                var item = str.substr(i, next - i);
                if (item.length) {
                    r.push(item);
                }
            }
            i = next + split_str.length;
        }
        return r;
    }


    pgParserNode.prototype.getAttrList = function () {
        if (this.attrList) return this.attrList;

        var _this = this;

        this.attrList = [];
        this.attrListChanged = false;

        if (this.script || this.comment || !this.attributes || this.attributes.length === 0) return this.attrList;

        var i = 0;
        var in_name = true;
        var in_value = false;
        var cur_name = '';
        var cur_value = '';
        var in_quote = null;

        var attrReadUntil = function (end_ch, ignore) {
            //var idx = _this.attributes.indexOf(end_ch, i);

            var idx = scripts_possible ? indexOfCodeAware(_this.attributes, end_ch, i) : _this.attributes.indexOf(end_ch, i);

            do {
                if (idx < 0) {
                    i = _this.attributes.length;
                    return null;
                } else {
                    /*
                     if(ignore) {
                     var ign_idx = _this.attributes.indexOf(ignore, i);
                     if(ign_idx >= 0 && idx >= ign_idx && idx + end_ch.length <= ign_idx + ignore.length) {
                     i = idx + end_ch.length;
                     continue;
                     }
                     }
                     */
                    var s = _this.attributes.substr(i, idx - i);
                    i = idx + end_ch.length;
                    return s;
                }
            }
            while (i < _this.attributes.length);
            return '';
        }

        var scripts_possible = this.attributes.indexOf('<') >= 0;

        while (i < this.attributes.length) {
            var ch = this.attributes.charAt(i);
            if (ch === ' ' || ch === "\t" || ch === "\n" || ch === "\r" || ch === '/' || ch === '>' || ch === '=') {
                //space
                if (cur_name.length > 0) {
                    //name done

                    var has_equal = false;
                    if (i < this.attributes.length - 1) {
                        while (i < this.attributes.length && (ch === ' ' || ch === '=')) {
                            if (ch === '=') has_equal = true;
                            i++;
                            ch = this.attributes.charAt(i);
                        }

                    }

                    if (has_equal) {
                        var which_quote = "\"";

                        in_value = true;
                        if (ch === '"' || ch === '\'') {
                            in_quote = ch;
                            which_quote = ch;
                        } else {
                            in_quote = null;
                        }
                        var val = '';
                        if (in_quote) {
                            i++;
                            val = attrReadUntil(in_quote); //disable escaping quotes, '\\' + in_quote);
                        } else {

                            if (i < this.attributes.length - 1) {
                                //var next_ch;
                                do {
                                    val += ch;
                                    i++;
                                    ch = this.attributes.charAt(i);
                                    //next_ch = this.attributes.length < i + 1 ? this.attributes.charAt(i+1) : null;
                                }
                                while (i < this.attributes.length && (ch !== ' ' && ch !== '>' && (ch !== '/' || i + 1 < this.attributes.length)));
                            }
                        }
                        this.attrList.push({name: cur_name, value: val, quote: which_quote, name_lc: cur_name.toLowerCase()});
                    } else {
                        i--;
                        this.attrList.push({name: cur_name, value: null, name_lc: cur_name.toLowerCase()});
                    }
                    cur_name = '';
                }
            } else {
                //nonspace
                cur_name += ch;//keep the case .toLowerCase();

                var si = pgGetScriptInfo(cur_name);

                if (si) {
                    i++;
                    var r = attrReadUntil(si.close);
                    cur_name += r + si.close;
                    if (r.length) i--;
                }
                /*
                 if(cur_name === '<?php' || cur_name === '<?=') {
                 i++;
                 var r = attrReadUntil('?>');
                 cur_name +=  r + '?>';
                 if(r.length) i--;
                 }*/
            }
            i++;
        }
        if (cur_name.length) {
            this.attrList.push({name: cur_name, value: null, name_lc: cur_name.toLowerCase()});
        }
        //console.log(this.attrList);
        return this.attrList;
        // console.log(this.attrList);
    }


    pgParserNode.prototype.hasAttr = pgParserNode.prototype.hasAttribute = function (attr) {
        var list = this.getAttrList();
        if (!list) return null;
        var r = this.findAttrsInList(attr, list);
        if (r.length) {
            return true;
        } else {
            return false;
        }
    }

    pgParserNode.prototype.attr = function (attr, value) {
        if (value === undefined) {
            return this.getAttr(attr);
        } else {
            return this.setAttr(attr, value)
        }
    }

    pgParserNode.prototype.findAttributesStartingWith = function (attr) {
        var list = this.getAttrList();
        if (!list) return [];
        attr = attr.toLowerCase();
        var r = [];
        for (var i = 0; i < list.length; i++) {
            if (list[i].name_lc.startsWith(attr)) {
                r.push(list[i]);
            }
        }
        return r;
    }

    pgParserNode.prototype.findAttrsInList = function (attr, list, single) {
        var r = [];
        attr = attr.toLowerCase();
        for (var i = 0; i < list.length; i++) {
            if (attr === list[i].name_lc) {
                r.push(list[i]);
            }
        }
        return r;
    }

    pgParserNode.prototype.appendToAttribute = function (attr, value, sep) {
        var c = this.getAttr(attr);
        if (typeof c === 'string') {
            c += `${sep}${value}`;
        } else {
            c = value;
        }
        this.setAttr(attr, c);
    }

    pgParserNode.prototype.setAttr = pgParserNode.prototype.setAttribute = function (attr, value, skip_encode, force_quote, event_info) {
        this.withEmitEvent(function () {
            if (value !== null) {
                switch (typeof value) {
                    case 'undefined':
                        value = null;
                        break;
                    case 'string':
                    case 'null':
                        break;
                    default:
                        value = value.toString();
                        break;
                }
            }
            var list = this.getAttrList();
            var r = this.findAttrsInList(attr, list);
            if (r.length) {
                var quote = force_quote || r[r.length - 1].quote || '"';
                if (!skip_encode && quote === '"') value = pgEncodeAttribute(value);
                r[r.length - 1].value = value;
                r[r.length - 1].quote = quote;
                r[r.length - 1].name = attr;
                r[r.length - 1].name_lc = attr.toLowerCase();
            } else {
                if (!skip_encode && force_quote !== `'`) value = pgEncodeAttribute(value);
                list.push({name: attr, value: value, quote: force_quote || null, name_lc: attr.toLowerCase()});
            }
            this.attrListChanged = true;

            if (attr === 'data-pg-id' && value == this.getId()) {
                this.hasPgIdAttr = true;
            }

        }, 'setAttr', null, {
            attr: attr,
            value: value,
            quote: force_quote,
            domValue: pgAddProxyLinksToAttr(attr, value),
            oldValue: function () {
                var r = this.findAttrsInList(attr, this.getAttrList());
                if (r.length) {
                    return cloneObject(r[r.length - 1]);
                } else {
                    return null;
                }
            },
            info: event_info
        });

        return value;
    }

    pgParserNode.prototype.getAttr = pgParserNode.prototype.getAttribute = function (attr, val_for_none) {
        if (val_for_none === undefined) val_for_none = null;
        var list = this.getAttrList();
        if (!list) return val_for_none;
        var r = this.findAttrsInList(attr, list);
        if (r.length) {
            var quote = r[r.length - 1].quote || '"';
            if (quote === '"') {
                return pgDecodeAttribute(r[r.length - 1].value);
            } else {
                return r[r.length - 1].value;
            }
        } else {
            return val_for_none;
        }
    }

    pgParserNode.prototype.removeAttr = pgParserNode.prototype.removeAttribute = function (attr) {
        this.withEmitEvent(function () {
            var list = this.getAttrList();
            attr = attr.toLowerCase();
            this.attrList = [];
            for (var i = 0; i < list.length; i++) {
                if (list[i].name_lc !== attr) {
                    this.attrList.push(list[i]);
                }
            }
            if (list.length !== this.attrList.length) {
                this.attrListChanged = true;
            }
            if (attr === 'data-pg-id') {
                this.hasPgIdAttr = false;
            }

        }, 'removeAttr', null, {
            attr: attr, oldValue: function () {
                var r = this.findAttrsInList(attr, this.getAttrList());
                if (r.length) {
                    return cloneObject(r[r.length - 1]);
                } else {
                    return null;
                }
            }
        });
    }

    pgParserNode.prototype.siblings = function (selector) {
        var arr = [];
        if (selector && selector.length > 0) {
            arr = this.parent.find(selector);
        }
        else {
            arr = this.parent.children;
        }

        var index = arr.indexOf(this);
        if (index > -1) arr.splice(index, 1);
        return arr;
    }

    pgParserNode.prototype.siblingOne = function (selector) {
        var siblings = this.siblings(selector);
        return siblings.length ? siblings[0] : null;
    }

    pgParserNode.prototype.removeAttrIfStartsWith = function (str, in_children, in_subcall, exclude_re) {

        if (this.isElement) {

            var strs = typeof str === 'string' ? [str] : str;
            var first_found = -1;

            var has_hits = false;

            if (!this.attrListChanged) {
                if (this.attributes.length) {
                    for (var j = 0; j < strs.length; j++) {
                        if (this.attributes.indexOf(strs[j]) >= 0) {
                            has_hits = true;
                            break;
                        }
                    }
                }
            } else if (this.getAttrList().length) {
                has_hits = true; //search the list
            }
            if (has_hits) {
                var list = this.getAttrList();

                for (var i = 0; i < list.length; i++) {
                    for (var j = 0; j < strs.length; j++) {
                        if (list[i].name.startsWith(strs[j]) && (!exclude_re || !list[i].name.match(exclude_re))) {
                            first_found = i;
                            break;
                        }
                    }
                    if (first_found >= 0) break;
                }
                if (first_found >= 0) {
                    this.attrList.splice(first_found, 1);

                    for (var i = first_found; i < this.attrList.length; i++) {
                        for (var j = 0; j < strs.length; j++) {
                            if (list[i].name.startsWith(strs[j]) && (!exclude_re || !list[i].name.match(exclude_re))) {
                                this.attrList.splice(i, 1);
                                i--;
                                break;
                            }
                        }
                    }
                    this.attrListChanged = true;
                }
            }
        }
        if (in_children) {
            this.walk(function (node) {
                if (node.isElement) {
                    node.removeAttrIfStartsWith(str, false, true, exclude_re);
                }
                return true;
            })
        }
        if (!in_subcall) {
            this.emitEvent('removeAttrIfStartsWith', {starts_with: str, in_children: in_children});
        }

        //}}}}}}}}}

    };

    pgParserNode.prototype.removeAttrIfStartsWithStringWithEvents = function(str) {
        var _this = this;
        this.getAttrList().slice(0).forEach(function(a) {
            if(a.name.startsWith(str)) {
                _this.removeAttribute(a.name);
            }
        })
    }

    pgParserNode.prototype.removePinegrowAttributes = function () {
        this.removeAttrIfStartsWith('data-pg-', false, false, /^data\-pg\-ia/);
    }

    pgParserNode.prototype.getAttributesString = function (filter_func) {
        if (!this.attrListChanged && !filter_func) {
            return this.attributes;
        }
        var o = [];
        var list = this.getAttrList();
        if (filter_func) {
            for (var i = 0; i < list.length; i++) {
                var str = filter_func(this, list[i].name, list[i].value, list[i].quote || '"');
                if (str && str.length) {
                    o.push(str);
                }
            }
        } else {
            for (var i = 0; i < list.length; i++) {
                if (list[i].value === null) {
                    o.push(list[i].name);
                } else {
                    var quote = list[i].quote || '"';
                    o.push(list[i].name + '=' + quote + list[i].value + quote);
                }
            }
        }
        return o.length > 0 ? ' ' + o.join(' ') : '';
    }

    pgParserNode.prototype.getInlineStylePropertyValue = function (prop, is_url) {
        var v = crsaGetInlineStylePropertyValue(this.getAttr('style'), prop, is_url);
        if (is_url) {
            v = pinegrow.getOriginalUrl(v);
        }
        return v;
    }

    pgParserNode.prototype.setInlineStylePropertyValue = function (prop, value, is_url, add_important, safe_mode) {
        var style = crsaSetInlineStylePropertyValue(this.getAttr('style'), prop, value, is_url, add_important, safe_mode);
        if (style) {
            this.setAttr('style', style);
        } else {
            this.removeAttr('style');
        }
    }

    pgParserNode.prototype.transformSrcsetUrls = function (func) {
        var srcset = this.getAttr('srcset');
        if (srcset !== null) {
            this.setAttr('srcset', pgTransformSrcsetUrls(srcset, func));
        }
    }

    pgParserNode.prototype.getElementId = function () {
        return this.getAttr('id');
    }


    pgParserNode.prototype.getId = function () {
        if (this.pgId === null && this.attributes) {
            var m = this.attributes.match(pgIdRegExp);
            if (m) {
                this.pgId = m[1];
                this.hasPgIdAttr = true;
            } else {
                this.pgId = 0;
                this.hasPgIdAttr = false;
            }
        }
        return this.pgId !== 0 ? this.pgId : null;
    }

    pgParserNode.prototype.assignId = function (prefix) {
        if (this.hasPgIdAttr) {
            //var len = this.attributes.length;
            this.attributes = this.attributes.replace(pgIdRegExpReplace, '');
            //if(len != this.attributes.length) {

            //}
            this.hasPgIdAttr = false;
        }
        this.pgId = prefix ? prefix + '_' + (++pgIdCount) : ++pgIdCount;
        return this.pgId;
    }

    pgParserNode.prototype.getOrAssignId = function () {
        var id = this.getId();
        if (!id) id = this.assignId();
        return id;
    }

    pgParserNode.prototype.removePgIdAttributes = function () {
        this.walkSelfAndChildren(function (node) {
            if (node.isElementOrScript) {
                node.removeAttribute('data-pg-id');
            }
        })
    }

    pgParserNode.prototype.assignIdAndAddToCatalog = function (do_subnodes) {
        var _this = this;
        if (do_subnodes) {
            this.walkSelfAndChildren(function (node) {
                if (node.shouldHaveId()) {
                    var id = node.getOrAssignId();
                    _this.nodeCatalogue.add(node);
                }
                return true;
            })
        } else {
            if (this.shouldHaveId()) {
                var id = this.getOrAssignId();
                _this.nodeCatalogue.add(this);
            }
        }
    }

    pgParserNode.prototype.get$DOMElement = function ($html) {
        if (this.$el) return this.$el;
        if (!$html) {
            var page = pinegrow.getCrsaPageOfPgParserNode(this);
            if (page) {
                $html = page.get$Html();
            }
        }
        if ($html) {
            var id = this.getId();
            var html = $html.get(0);
            if (html.getAttribute('data-pg-id') == id) return $html;
            var sel = '[data-pg-id="' + id + '"]';
            var el = html.querySelector(sel);
            if (el) return $(el);
        }
        return null;
    }

    pgParserNode.prototype.getDOMElement = function ($html) {
        var $el = this.get$DOMElement($html);
        return $el ? $el.get(0) : null;
    }

    pgParserNode.prototype.get$DOMElements = function (page, except_pv) {
        var _this = this;
        page = page || this.getPage();
        if (!page) return [];
        var r = [];
        page.getPageViews().forEach(function (pv) {
            if (except_pv && except_pv === pv) return;
            var $el = _this.get$DOMElement(pv.get$Html());
            if ($el) {
                $el.pgPageView = pv;
                r.push($el);
            }
        })
        return r;
    }


    pgParserNode.prototype.mapIdsToDomElement = function (node) {
        var id = this.getId();
        if (id) {
            node.setAttribute("data-pg-id", id);
        }
        var dom_i = 0;
        var pg_i = 0;

        if (node.hasChildNodes()) {
            for (var i = 0; i < this.children.length; i++) {
                if (!this.children[i].shouldHaveId()) continue;

                while (dom_i < node.childNodes.length && node.childNodes[dom_i].nodeType !== 1) {
                    dom_i++;
                }
                if (dom_i < node.childNodes.length) {
                    this.children[i].mapIdsToDomElement(node.childNodes[dom_i]);
                }
                dom_i++;
            }
        }
    }

    pgParserNode.prototype.isAutoClosedTag = function () {
        return pgAutoClosedTags[this.tagName] ? true : false;
    }

    let _allTags = new Map();

    function setupHtmlTags(tagArr) {
        try {
            for (const key in tagArr) {
                let isInt = parseInt(key);
                let isStr = typeof tagArr === "string";

                switch (key.constructor) {
                    case Object:
                        setupHtmlTags(key);
                        break;
                    case Array:
                        if (!_allTags.has(key)) {
                            _allTags.set(key, key)
                            setupHtmlTags(tagArr[key]);
                        }
                        break;
                    case String:
                        if( !Number.isInteger(isInt)  ){
                            let tag =key.replace("<","")
                            _allTags.set(tag, tag)
                        }

                        break;
                }

                if(tagArr[key] && !isStr){
                    if(typeof tagArr[key] == "string"){
                        let tag =tagArr[key].replace("<","")
                        _allTags.set(tag, tag)
                    }else{
                        setupHtmlTags(tagArr[key]);
                    }
                }
            }
        }catch (e) {
            console.log(e)
        }

    }

    // setupHtmlTags(["path", "!--", "button", "svg", "canvas"])
    setupHtmlTags([...pgSingleTags, ...pgDontFormatIfAllChildrenNonFormat,...pgDontFormat, pgAutoClosedTags, ...pgSingleTags])

    let allTags =[..._allTags.values(),"path", "!--", "button", "svg", "canvas","title","script" ]

    // added this to check if this value is a valid html5 tag
    /**@param {boolean|null} checkClosing **/
    pgParserNode.prototype.isValidTag = function (checkClosing =null) {
        return allTags.includes(checkClosing ? this.closingTag : this.tagName)
    }

    pgParserNode.prototype.isSingleTag = function () {
        return pgSingleTags.indexOf(this.tagName) >= 0;
    }

    pgParserNode.prototype.isProperlyClosed = function () {
        return this.validate();
    }

    pgParserNode.prototype.validate = function (strict) {
        if (!this.rootNode && this.tagName !== 'text' && (!(this.closed || pgAutoClosedTags[this.tagName]) || (this.closingTag && this.tagName !== this.closingTag.toLowerCase()))) {
            return false;
        }
        if(strict && this.attrList) {
            //check errors that usually happen during typing the code when tags are not yet closed
            for(var i = 0; i < this.attrList.length; i++) {
                //having < in attr is a sign of improperly closed tag (unless it is a script such as <?php ...
                if(this.attrList[i].name.startsWith('<') && !pgHasScript(this.attrList[i].name)) {
                    return false;
                }
            }
        }
        return true;
    }

    pgParserNode.prototype.getNestingLevel = function () {
        var level = 0;
        var n = this;
        while (n.parent) {
            level++;
            n = n.parent;
        }
        return level;
    }

    pgParserNode.prototype.getOpeningTag = function () {
        if (this.isElement) {
            return '<' + this.tagName + '' + this.getAttributesString(pgGetAttributesStringFilterOutPgIds) + '>';
        }
        return null;
    }

    pgParserNode.prototype.getClosingTag = function () {
        if (this.isElement) {
            return '</' + (this.closingTag ? this.closingTag : this.tagName) + '>';
        }
        return null;
    }

    pgParserNode.prototype.findChildNodeAtSourceIndex = function (index) {
        for (var i = 0; i < this.children.length; i++) {
            if (index >= this.children[i].startStringIndex && (i === this.children.length - 1 || index < this.children[i + 1].startStringIndex)) return this.children[i];
        }
        return null;
    }

    pgParserNode.prototype.findNodeAtSourceIndex = function (index) {
        var node = this.findChildNodeAtSourceIndex(index);
        if (node) {
            var child = node.findChildNodeAtSourceIndex(index);
            while (child && !child.textNode) {
                node = child;
                child = node.findChildNodeAtSourceIndex(index);
            }
        }
        return node;
    }

    pgParserNode.prototype.getPath = function (precise) {
        if (precise === undefined) precise = false;
        var p = this;
        var path = '';
        while (p.parent) {
            path = p.parent.getChildPos(p, !(p === this && precise)) + (path.length ? ',' : '') + path;
            p = p.parent;
        }
        return path;
    }

    pgParserNode.prototype.getRelativePath = function (to, precise) {
        if (precise === undefined) precise = false;
        var my_path = this.getPath(precise);
        var to_path = to.getPath();
        if (my_path.startsWith(to_path)) {
            return my_path.replace(to_path + ',', '');
        }
        return null;
    }

    pgParserNode.prototype.getNodeFromPath = function (path, return_last_valid) {
        if (path === '' || path === ',') return this;
        var a = path.split(',');
        var node = this;
        for (var i = 0; i < a.length; i++) {
            node = node.getChildAtPos(parseInt(a[i]), true);
            if (!node) return return_last_valid ? node : null;
        }
        return node;
    }

    pgParserNode.prototype.insertAtPath = function (node, path, precise) {
        if (precise === undefined) precise = false;
        var a = path.split(',');
        var index = parseInt(a.pop());
        var parent = this.getNodeFromPath(a.join(','));
        if (parent) {
            node.insertAtIndex(parent, index, !precise);
            return true;
        } else {
            return false;
        }
    }

    pgParserNode.prototype.getUniqueSelectors = function (level, max_level, context_pgel, return_first) {
        var debug = false;
        level = level || 0;

        var use_class = null;
        if (max_level === undefined) max_level = 2;
        if (debug === undefined) debug = false;

        if (!context_pgel) {
            //var page = this.getPage();
            context_pgel = this.document;
        }

        var _this = this;

        var isUnique = function (sel) {
            var list = context_pgel.find(sel);
            return list.length === 1 && _this === list[0];
        }

        var paths = []

        var pgid = this.getAttr('data-pg');
        if (pgid) {
            var sel = `[data-pg="${pgid}"]`;
            if (isUnique(sel)) {
                paths.push(sel);
                if (return_first) return paths;
            }
        }

        var id = this.getAttribute('id')
        if (id) {
            if (isUnique('#' + id) && !id.match(/^[0-9]/)) {
                paths.push('#' + id);
                if (return_first) return paths;
            }
        }

        var use_tag = '';

        var tryed = []

        var main_tags = ['h1', /*'h2', 'h3', 'h4', 'h5', 'h6',*/ 'html', 'body', 'head', 'header', 'main', 'footer', 'article', 'nav'];

        var strong_tags = main_tags.concat(['h2', 'h3', 'h4', 'h5', 'h6', 'section'])

        var is_main_tag = false;

        if (strong_tags.indexOf(this.tagName) >= 0) {
            //use_tag = this.tagName;
            tryed.push(this.tagName);

            is_main_tag = true;

            if (isUnique(this.tagName)) {
                paths.push(this.tagName);
                if (return_first) return paths;
            }
        }

        if (use_class && this.hasClass(use_class)) {

            tryed.push('.' + use_class);

            if (isUnique('.' + use_class)) {
                paths.push('.' + use_class);
                if (return_first) return paths;
            }
            if (use_tag) {
                tryed.push(use_tag + '.' + use_class);
                if (isUnique(use_tag + '.' + use_class)) {
                    paths.push(use_tag + '.' + use_class);
                    if (return_first) return paths;
                }
            }
        } else {
            var classes = this.getClasses();
            if (classes.length <= 3) {
                for (var i = 0; i < classes.length; i++) {
                    var cls = classes[i];
                    tryed.push(use_tag + '.' + cls);
                    if (isUnique(use_tag + '.' + cls)) {
                        paths.push(use_tag + '.' + cls);
                        if (return_first) return paths;
                    }
                }
            }
        }

        if (!is_main_tag) {
            tryed.push(this.tagName);

            if (isUnique(this.tagName)) {
                paths.push(this.tagName);
                if (return_first) return paths;
            }
        }

        if (level < max_level) {
            if (paths.length === 0) {
                //no sels found so far, try with parent
                var parent = null;
                if (this.parent /* && this.parent !== context_pgel */) {
                    parent = this.parent;

                    var parent_paths;
                    if (parent === context_pgel) {
                        parent_paths = [''];
                    } else {
                        parent_paths = this.parent.getUniqueSelectors(level + 1, max_level, context_pgel, true /* return_first */);
                    }

                    if (parent_paths.length) {

                        for (var ii = 0; ii < parent_paths.length; ii++) {
                            var pp = parent_paths[ii];

                            for (var iii = 0; iii < tryed.length; iii++) {
                                var sel = tryed[iii];

                                var s = `${pp} > ${sel}`.trim();
                                if (isUnique(s)) {
                                    paths.push(s);
                                    if (return_first) return paths;
                                }
                            }
                        }
                        if (paths.length === 0) {
                            //look for unique classes
                            var min_c = 9999;
                            var min_cls = null;

                            this.getClasses().forEach(function (cls) {
                                var c = _this.parent.find('>.' + cls);
                                if (min_cls === null || min_c > c) {
                                    min_c = c;
                                    min_cls = cls;
                                }
                            })

                            //try n-th based
                            var sel = _this.tagName;
                            var all = parent.find('>' + sel);
                            var idx = all.indexOf(this);

                            /*
                            //This is not supported with nth-type
                            if (min_cls) {
                                var all_wc = parent.find('>' + _this.tagName + '.' + min_cls);
                                if (all_wc.length < all.length) {
                                    all = all_wc;
                                    sel = _this.tagName + '.' + min_cls;
                                    idx = all.indexOf(this);
                                }
                            }
                            */
                            if (idx >= 0) {
                                idx++;
                                for (var ii = 0; ii < parent_paths.length; ii++) {
                                    var pp = parent_paths[ii];

                                    if (all.length === 1) {
                                        var s = `${pp} > ${sel}`.trim();
                                    } else {
                                        var s = `${pp} > ${sel}:nth-of-type(${idx})`.trim();
                                    }
                                    //console.log(s)
                                    if (isUnique(s)) {
                                        paths.push(s);
                                        if (return_first) return paths;
                                    } else {
                                        if (debug) console.error('Should be unique: ', s, _this)
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        if (paths.length && paths[0].indexOf(' ') < 0) {
            //we have a single selector
            var sel = paths[0];
            if (sel.indexOf('#') >= 0) {
                //id, ok
            } else if (sel.indexOf('data-pg=') >= 0) {
                //data-pg attr, ok
            } else {
                //hm, getting tricky, just tag or class or both
                var force = false;
                if (sel.indexOf('.') < 0 && main_tags.indexOf(this.tagName) < 0) {
                    force = true;
                }

                var isStrong = function (p) {
                    return p.hasAttr('id') || p.hasAttr('data-pg') || main_tags.indexOf(p.tagName) >= 0;
                }

                if (this.parent && (force || isStrong(this.parent))) {
                    paths.unshift(this.parent.getUniqueSelectors(0, 5, null, true) + ' > ' + sel);
                } else if (this.parent && this.parent.parent && isStrong(this.parent.parent)) {
                    paths.unshift(this.parent.parent.getUniqueSelectors(0, 5, null, true) + ' ' + sel);
                }
            }
        }
        return paths;
    }

    pgParserNode.prototype.getUniqueSelector = function(base_pgel) {
        var sels = this.getUniqueSelectors(0, 9999, base_pgel, true);
        if (sels.length) {
            return sels[0];
        }
        return null;
    }

    pgParserNode.prototype.getPositionInSource = function () {
        var pos = {start: this.startStringIndex, end: -1};
        var next = this.next(true);
        if (next) {
            pos.end = next.startStringIndex;
        } else if (this.parent) {
            pos.end = this.parent.getPositionInSource().end;
            var ct = this.parent.getClosingTag();
            if (ct) pos.end -= ct.length;
        }
        return pos;
    }

    pgParserNode.prototype.getScriptIndent = function () {
        if (!this.script || !this.content) return 0;
        //  1 - indent next line
        //  -1 - indent back this line
        // -2 - indent back this, indent next
        // 0
        if (this.content.match(/^\s(if|while|foreach|for)/) && this.content.indexOf(':') > 0) {
            return 1;
        } else if (this.content.match(/^\s(endif|endwhile|endforeach|endfor)/)) {
            return -1;
        } else if (this.content.match(/^\selse/)) {
            return -2;
        } else {
            //count { and }
            var open_c = 0;
            var close_c = 0;
            for (var i = 0; i < this.content.length; i++) {
                var ch = this.content.charCodeAt(i);
                if (ch === 123) {
                    open_c++;
                } else if (ch === 125) {
                    close_c++;
                }
            }
            if (open_c === close_c) {
                return 0;
            } else if (open_c > close_c) {
                return 1;
            } else {
                return -1;
            }
        }

        return 0;
    }

    pgParserNode.prototype.changeToPhp = function (code) {
        this.tagName = "?php";
        this.tagNameOriginal = "?php";
        this.script = true;
        this.script_info = pgScripts['<?php'];
        this.isElement = false;
        this.isElementOrScript = true;
        this.textNode = false;
        this.singleTag = true;
        this.closed = true;
        this.scriptClosed = true;
        this.closingTag = null;
        this.content = code;
        this.selfClosed = false;
    }


    pgParserNode.prototype.toString = pgParserNode.prototype.toStringNEW = function (pref, showTextNodes, withIds, func, format_html, options, use_pref, only_content) {
        pref = pref || '';

        //format_html = true;//force always because wo format is broken
        format_html = /*this.hasAttr('df') ||*/ format_html || this.formatHtml;

        //pref = '';
        if (!options) {
            options = {
                indent: '    ',
                php_ids: false,
                assign_missing_ids: true,
                remove_collapsed: false,
                remove_hidden: false
            }
        }

        if (typeof use_pref == 'undefined') use_pref = format_html;//true; MTCH

        var nl = ''
        var enl = '';
        var epref = pref;

        var s = '';

        var pgId = null;
        var pgIdAdd = '';
        var pgIdPrepend = '';

        var assign_missing_ids = 'assign_missing_ids' in options ? options.assign_missing_ids : true;
        var indent = options.indent;

        this.attrListChanged = this.isElement; //force attr parsing to eliminate empty spaces etc
        if (this.attrListChanged) {
            this.attributes = this.getAttributesString(function (node, name, value, quote) {
                if (!withIds && name === 'data-pg-id') return null;
                if (name === 'data-pg-collapsed' && options.remove_collapsed) return null;
                if (name === 'data-pg-hidden' && options.remove_hidden) return null;

                var str;
                if (value === null) {
                    str = name;
                } else {
                    str = name + '=' + quote + value + quote;
                }
                if (func) {
                    str = func(node, str, 'attribute', name, value, quote);
                }
                return str;
            });
            this.attrListChanged = false;
        }

        var attrs = this.attributes;

        if (this.script) {
            attrs = this.indent_text ? pgReindentText(this.content, pref + indent, pref) : this.content;
        }

        pgId = this.getId();

        var format = format_html && pgDontFormat.indexOf(this.tagName) < 0 && !this.selfClosed;

        if (format && (pgDontFormatIfAllChildrenNonFormat.indexOf(this.tagName) >= 0 || pgSingleTags.indexOf(this.tagName) >= 0)) {
            format = false;
            for (var i = 0; i < this.children.length; i++) {
                if (!this.children[i].textNode && pgDontFormat.indexOf(this.children[i].tagName) < 0) {
                    format = true;
                    break;
                }
            }
        }

        options.last_format = format;

        var orig_pref = pref;

        var php_ids = options.php_ids || false;

        if (format) {
            nl = "\n";
            enl = '\n';
        } else {
            //pref = '';
        }
        if (pref.length) enl = '\n';

        if (this.rootNode) {
            indent = '';
        }

        if (pgId) {
            if (!withIds) {
                attrs = attrs.replace(pgIdRegExp, '');
            } else {
                if (!this.hasPgIdAttr) {
                    pgIdAdd = ' ' + pgAttr + '="' + pgId + '"';
                }
            }
        }
        if (pgId === null && withIds && this.isElement && assign_missing_ids) {
            pgId = this.assignId();
            pgIdAdd = ' ' + pgAttr + '="' + pgId + '"';
        }
        if (pgIdAdd && pgIsScriptTag(this.tagName)) {
            if (php_ids) pgIdPrepend = ' /*' + pgIdAdd + '*/';
            pgIdAdd = '';
        }

        if (showTextNodes && this.textNode) {
            //if(this.content == "\n") return this.content;
            if (format && this.content.match(pgWhiteSpaceRegExp)) {
                //           if(this.content.indexOf("\n") >= 0) return '\n';
                var nidx = this.content.indexOf("\n");

                ////mtexp
                //added this to correctly handle <p>aaa\n<a...>. Was no space between aaa and <a> otherwise
                if (nidx === 0 && this.content.length > 1 && !options.parent_format) {
                    return ' ';
                }
                ///end

                if (nidx >= 0) return '';
                if (this.content.length) return ' ';
                return '';
            }
            return this.content;
        }

        var content = '';
        var just_text_or_empty = true;
        var prev_is_formatted = false;
        var is_text = false;
        var prev_is_text = false;

        for (var i = 0; i < this.children.length; i++) {
            if (!showTextNodes && this.children[i].textNode) continue;
            just_text_or_empty = just_text_or_empty && this.children[i].textNode;

            var script_indent = this.children[i].getScriptIndent();

            if (script_indent === -1 || script_indent === -2) {
                pref = pref.substr(0, pref.length - options.indent.length);
            }
            options.parent_format = format;

            var use_pref_on_child = format;
            prev_is_text = is_text;
            is_text = this.children[i].textNode;

            if(prev_is_text && !is_text && pgInlineTags.indexOf(this.children[i].tagName) >= 0) {
                //this element is inline and before is a text, dont use pref
                use_pref_on_child = false;
                //options.parent_format = false;
            }

            /*
            if(is_text && this.children[i].content === "\n") {
                content += "\n";
                continue;
            }*/

            var c = this.children[i].toStringNEW(pref + indent, showTextNodes, withIds, func, this.tagName === 'pre' ? false : format_html, options, use_pref_on_child);
            var child_format = options.last_format;

            if (withIds && pgIsScriptTag(this.children[i].tagName) && php_ids) {
                var cpgId = this.children[i].getId();
                c = pref + indent + '<!--[start-pg-id:' + cpgId + ']-->\n' + c + '\n' + pref + indent + '<!--[end-pg-id:' + cpgId + ']-->';
            }
            if (script_indent === 1 || script_indent === -2) {
                pref = pref + options.indent;
            }

            if (format) {
                if (c.length) {

                    if (is_text) {
                        if (c.length) {
                            var br = '';
                            if (c.charAt(0) === "\n") {
                                c = c.replace(/^\n\s*/, "");
                                prev_is_formatted = true;
                                br = "\n";
                            }
                            if (prev_is_formatted) {
                                c = br + pref + indent + c;
                                if (i + 1 < this.children.length) {
                                    if (!this.children[i + 1].textNode && pgDontFormat.indexOf(this.children[i + 1].tagName) < 0) {
                                        //next node is element that should be formatted
                                        c = c + "\n";
                                    }
                                } else {
                                    c = c + "\n";
                                }
                            }
                        }
                    } else {
                        if ((!this.rootNode || content.length) && use_pref_on_child) {
                            //if pref was used on child, add nl, otherwise just append inline
                            c = "\n" + c;
                        }
                    }
                    prev_is_formatted = child_format;
                } else {
                    if (i + 1 === this.children.length) {
                        //last one
                        c = "\n" + c;
                    }
                }

            }
            content += c;

        }
        if (format && this.children.length && !just_text_or_empty && content.length && !this.rootNode) {
            if (content.charAt(content.length - 1) !== "\n") {
                content = content + "\n";
            }
        }

        if (only_content) return content;

        if (format && just_text_or_empty) {
            nl = '';
            epref = '';
        }
        enl = '';

        options.last_format = format;

        if (/*attrs.length && */ func) {
            attrs = func(this, attrs, 'attrs');
        }

        if (!this.rootNode) {
            if (this.outputOnly === null || this.outputOnly === 'opening') {
                s += (use_pref ? pref : "") + '<' + (this.tagNameOriginal ? this.tagNameOriginal : this.tagName) + pgIdPrepend + attrs + pgIdAdd + (this.selfClosed ? '/' : '');
                if (this.script) {

                    //if(this.tagName == '?php' || this.tagName == '?=') {
                    if (this.script_info) {
                        if (this.scriptClosed) {
                            s += this.script_info.close;
                        }
                    } else {
                        s += '>';
                    }
                } else {
                    s += '>';
                }
            }
        } else {
            return s + content;
        }

        if (this.content && !this.script) {
            if (this.tagName === 'php') {
                content += pgReindentText(this.content, pref);// + indent);
            } else {
                content += this.content;
            }
        }
        if (func) content = func(this, content, 'content');

        if (format) {
            s += content;
        } else {
            s += content;
        }

        if (!this.singleTag) {
            if (this.closingTag && (this.outputOnly === null || this.outputOnly === 'closing')) {
                s += (format ? epref : '') + '</' + this.closingTag + '>' + enl;
            } else {
                s += enl;
            }
        } else {
            s += enl;
        }
        if (func) s = func(this, s, 'node');
        return s;
    }

    pgParserNode.prototype.toStringOriginal = function (format_html, options, func, only_content) {
        if (typeof pinegrow != 'undefined') {
            if (format_html && format_html !== 'force') {
                format_html = pinegrow.autoFormatHtml;
            }
        }
        return this.toString(null, true, false, func, format_html, options, null, only_content)
    }

    pgParserNode.prototype.toStringWithIds = function (format_html, options, func, only_content) {
        if (typeof pinegrow != 'undefined') {
            if (format_html && format_html !== 'force') {
                format_html = pinegrow.autoFormatHtml;
            }
        }
        return this.toString(null, true, true, func, format_html, options, null, only_content)
    }

    pgParserNode.prototype.toStringContent = function (format_html, options) {
        var html = '';
        if (this.tagName === 'style' || this.tagName === 'script') {
            html = this.content;
        } else {
            html = this.toStringOriginal(format_html, options, null, true);
        }
        html = pgReindentText(html);
        html = $.trim(html);
        return html;
    }

    pgParserNode.prototype.toDebug = function () {
        var s = '';

        if (this.textNode) {
            s = '[TEXT] |' + this.content.replace(/\n/g, '\\n').replace(/ /g, '_') + '|';
        } else if (!this.rootNode) {
            s = (this.tagNameOriginal ? this.tagNameOriginal : this.tagName);
            if (this.closed) s += ' /' + this.closingTag;
        }

        if (this.children.length) {
            s += '<ul>';
            for (var i = 0; i < this.children.length; i++) {
                s += this.children[i].toDebug();
            }
            s += '</ul>';
        }
        if (this.rootNode) return s;

        return '<li>' + s + '</li>';
    }

    pgParserNode.prototype.getTextContent = function () {
        var c = this.content.replace(/^\n\s*/, '');
        /*if(this.content.match(/^[\s\n]*$/)) {
         return '';
         }*/
        return c;
        return $.trim(this.content);
    }

    pgParserNode.prototype.containsServerSideScripts = function (check_attrs) {
        var has = false;
        this.walkSelfAndChildren(function (pgel) {
            if (check_attrs) {
                var list = pgel.getAttrList();
                for (var i = 0; i < list.length; i++) {
                    if (list[i].value && pgHasScript(list[i].value)) {
                        has = true;
                        return false;
                    }
                }
            }
            if (pgel.script_info || pgel.tagName === 'php') {
                has = true;
                return false;
            } else {
                return true;
            }
        })
        return has;
    }

    pgParserNode.prototype.canConvertToJade = function () {
        var can = true;
        this.walkSelfAndChildren(function (pgel) {
            if (
                (pgel.script && pgel.script_info) ||
                (pgel.comment && pgel.attributes.indexOf('[if') >= 0) ||
                (pgel.tagName === 'style' && pgel.content && pgel.content.length > 0) ||
                (pgel.tagName === 'script' && pgel.content && pgel.content.length > 0)) {
                can = false;
                return false;
            } else {
                return true;
            }
        })
        return can;
    }

    pgParserNode.prototype.toJade = function (indent) {
        if (!indent) indent = '';
        var s = '';
        var spacer = pinegrow.getFormatHtmlOptions().indent;

        if (this.textNode) {
            var t = this.getTextContent();
            if (t.length == 0) return null;
            s = '| ' + t;
        } else if (this.comment) {
            var a = $.trim(this.attributes.replace(/\-\-$/, ''));
            if (a.startsWith('[if')) {
                /*var re = /\[(.*)\]>([\s\S]*)\<!\[endif\]/;
                 var m = a.match(re);
                 if(m) {
                 var p = new pgParser();
                 p.assignIds = true;
                 p.parse($.trim(m[2]));
                 a = m[1] + p.rootNode.toJade(indent + spacer);
                 }*/
                s = pgReindentText(this.toStringOriginal(true), indent, '');
            } else {
                s += '//' + a;
            }
        } else if (!this.rootNode) {
            s = this.tagName;
            var id = this.getAttr('id');
            if (id) {
                s += '#' + id;
            }
            var classes = this.getClasses();
            if (classes.length) {
                s += '.' + classes.join('.');
            }

            var attrs = this.getAttrList();
            if (attrs.length) {

                var a = '';
                for (var i = 0; i < attrs.length; i++) {
                    if (attrs[i].name === 'id' || attrs[i].name === 'class') continue;
                    if (['data-pg-collapsed', 'data-pg-hidden', 'data-pg-id'].indexOf(attrs[i].name) >= 0) continue;

                    if (a.length > 0) a += ', ';
                    a += attrs[i].name;
                    if (attrs[i].value !== null) {
                        a += '="' + attrs[i].value + '"';
                    }
                }
                if (a.length) {
                    s += '(' + a + ')';
                }
            }

            if ((this.tagName === 'script' || this.tagName === 'style') && this.content.length) {
                s += '.' + pgReindentText(this.content, indent + spacer);
            } else {
                if (this.children.length === 1 && this.children[0].textNode) {
                    s += ' ' + this.children[0].getTextContent();
                } else {
                    for (var i = 0; i < this.children.length; i++) {
                        var jade = this.children[i].toJade(indent + spacer);
                        if (jade !== null) {
                            s += "\n" + jade;
                        }
                    }
                }
            }
        } else {
            for (var i = 0; i < this.children.length; i++) {
                var jade = this.children[i].toJade(indent);
                if (jade !== null) {
                    s += "\n" + jade;
                }
            }
        }
        return indent + s;
    }

    pgParserNode.prototype.findLastNonScriptChild = function () {
        var firstScript = null;
        for (var i = this.children.length - 1; i >= 0; i--) {
            if (this.children[i].isElement) {
                if (this.children[i].tagName === 'script') {
                    firstScript = this.children[i];
                } else {
                    break;
                }
            }
        }
        if (firstScript) {
            return firstScript.prev();
        }
        return null;
    }

    var pgParserNodeGetNameAttrs = ['src', 'url', 'href'];

    pgParserNode.prototype.getPlainTextName = function (obj) {
        return this.getName(obj).replace(/<[^><]+>/g, '');
    }

    pgParserNode.prototype.getName = function (obj) {
        if (this.isElementOrScript) {
            var max_text_len = (obj && obj.max_text_len) ? obj.max_text_len : 20;
            var short = false;
            var action_tag = null;
            var classes_filter = null;

            var num_classes = 1;
            if (obj) {
                num_classes = obj.num_classes || num_classes;
                short = obj.short || short;
                action_tag = obj.action_tag || null;
                classes_filter = obj.classes_filter || null;
            }

            var r = '';
            var search_str = '';
            var name;
            var dname = null;

            if (this.isElement) {
                name = this.tagName;
            } else if (this.script_info) {
                name = this.script_info.name_tag;
            }

            if (obj) {
                if (obj.name) {
                    r += '<namedef>' + obj.name + '&nbsp;</namedef>';
                    search_str = obj.name + ':';
                } else if (obj.def) {
                    dname = obj.def.name;
                    if (obj.def.display_name) {
                        dname = obj.def.display_name === 'tag' ? this.tagName : obj.def.display_name(this, obj.def);
                    }
                    //if (dname !== name) {
                    r += (obj.def.namedef_class ? `<namedef class="${obj.def.namedef_class}">` : '<namedef>') + dname + '&nbsp;</namedef>';
                    search_str = dname + ':';
                    //}
                }
            }

            if (this.script_info) {
                r = '<namedef>' + name + '</namedef>';
                search_str += name;
                r += '<nametext> | ' + escapeHtmlCode(crsaGetSummaryStr(this.content, 20)) + '</nametext>';
                search_str += this.content;
                return r;
            }

            search_str += this.tagName;

            if(dname === null) {
                r += '<nametag>' + name + '</nametag>';
            } else {
                if(dname === name) {
                    r = '<nametag>' + name + '</nametag>';
                } else if(name === dname.toLowerCase()) {
                    //keep just dname
                } else {
                    r += '<nametag>' + name + '</nametag>';
                }
            }

            var id = this.getAttr('id');
            if (id) {
                if (id.length > 15) {
                    r += `<nameid title="#${id}">#${shortenString(id, 15)}</nameid>`;
                } else {
                    r += '<nameid>#' + id + '</nameid>';
                }
                search_str += '-#' + id;
            }
            var classes = this.getClasses();

            var used_i = -1;
            var all_classes = '';
            var add_more_classes = false;

            if(classes_filter) {
                var ci = classes_filter(this);
                r += ci.html;
                search_str += ci.search_str;
                if(ci.all_classes) {
                    r += '<nameclass title="' + ci.all_classes + '">...</nameclass>';
                }
            } else {
                if (num_classes > 0) {
                    for (var i = 0; i < classes.length; i++) {
                        if (classes[i] === 'pg-empty-placeholder') continue;
                        if(classes[i].indexOf('<') >= 0) continue; //php classes like <?php echo $cls; ?>

                        used_i++;
                        if (used_i < num_classes) {
                            r += '<nameclass>.' + (classes[i]) + '</nameclass>';
                        } else if (used_i === num_classes && classes.length > num_classes) {
                            add_more_classes = true;
                            //break;
                        }
                        search_str += ':.' + classes[i];
                        all_classes += (all_classes.length ? ('.' + classes[i]) : classes[i])
                    }
                    if (add_more_classes) {
                        r += '<nameclass title="' + all_classes + '">...</nameclass>';
                    }
                }
            }
            if (action_tag) {
                r += ' <action>' + action_tag + '</action>';
            }
            if (!short) {
                var t = '';
                if (this.tagName === 'meta') {
                    var meta_name = this.getAttr('name');
                    var meta_content = this.getAttr('content');
                    if (meta_name) {
                        t = meta_name + '="' + (meta_content || '') + '"';
                    } else {
                        var attrs = this.getAttrList();

                        if (attrs.length) {
                            t = attrs[0].name + '="' + attrs[0].value + '"';
                        }
                    }
                    if (t.length) {
                        r += '<nametext> | ' + t + '</nametext>';
                    }
                    search_str += ':' + t;

                } else if (this.singleTag || ((this.tagName === 'script' || this.tagName === 'style') && !this.content)) {
                    for (var a = 0; a < pgParserNodeGetNameAttrs.length; a++) {
                        var v = this.getAttr(pgParserNodeGetNameAttrs[a]);
                        if (v) {
                            t = shortenString(v, max_text_len, '...', true);
                            r += '<nametext> | ' + t + '</nametext>';
                            search_str += ':' + v;
                            break;
                        }
                    }
                } else if (this.tagName === 'script' || this.tagName === 'style') {
                    t = $.trim(this.content);
                    if (t.length) {
                        if (t.length > max_text_len) {
                            t = t.substr(0, max_text_len) + '...';
                        }
                        t = t.replace(/\n/g, ' ');
                        r += '<nametext> | ' + t + '</nametext>';
                        search_str += ':' + t;
                    }
                } else if (this.tagName === 'a') {
                    var lt = '';
                    var href = this.getAttribute('href');
                    t = $.trim(this.text(null, true).replace(/\n/g, ' '));
                    var max_href_len = max_text_len * 2;
                    if (t.length) {
                        if (t.length > max_text_len) {
                            t = t.substr(0, max_text_len) + '...';
                        }
                    }

                    if (href) {
                        lt = crsaGetSummaryStr(href, max_href_len) + ' | ';
                        search_str += ':' + href;
                    }
                    if (t.length) {
                        lt += t;
                        search_str += ':' + t;
                    }
                    r += '<nametext> | ' + lt + '</nametext>';
                } else {
                    t = $.trim(this.text(null, true).replace(/\n/g, ' '));
                    if (t.length) {
                        if (t.length > max_text_len) {
                            t = t.substr(0, max_text_len) + '...';
                        }
                        r += '<nametext> | ' + t + '</nametext>';
                        search_str += ':' + t;
                    }
                }
            }
            if (obj) {
                obj.search_string = search_str;
            }
            if(this.dynamic) {
                r += '<namedyn title="This element is created by Javascript code. Any changes done to the element will not be saved to the source code.">dyn</namedyn>';
            }
            return r;
        } else if (this.textNode) {
            var t = this.content.substr(0, max_text_len);
            if (obj) {
                obj.search_string = t;
            }
            return '[TEXT] ' + t;
        }
        return 'Element';
    }

    pgParserNode.prototype.getViewHtml = function (page) {
        page = page || pinegrow.getCrsaPageOfPgParserNode(this);
        return page ? page.getViewHTMLForElement(this) : this.toStringWithIds();
    }

    pgParserNode.prototype.getViewInnerHtml = function (page) {
        page = page || pinegrow.getCrsaPageOfPgParserNode(this);
        return page ? page.getViewInnerHTMLForElement(this) : this.html();
    }


    pgParserNode.prototype.toIncrementalDOM = function () {
        var elementOpenStart = IncrementalDOM.elementOpenStart,
            elementOpenEnd = IncrementalDOM.elementOpenEnd,
            elementVoid = IncrementalDOM.elementVoid,
            elementClose = IncrementalDOM.elementClose,
            attr = IncrementalDOM.attr,
            text = IncrementalDOM.text;

        var _this = this;

        return function () {

            if (_this.isElement) {
                var args = [];
                var list = _this.getAttrList();

                //console.log(args);

                elementOpenStart(_this.tagName, null /*_this.getId() */, null, args);

                for (var i = 0; i < list.length; i++) {
                    attr(list[i].name, list[i].value || '');
                }


                elementOpenEnd();

                for (var i = 0; i < _this.children.length; i++) {
                    _this.children[i].toIncrementalDOM()();
                }
                elementClose(_this.tagName);

            } else if (_this.textNode) {
                text(_this.content, pgDecodeHtml);
            }
        }
    }

    pgParserNode.prototype.toIncrementalDOMInner = function () {
        var _this = this;

        return function () {
            for (var i = 0; i < _this.children.length; i++) {
                _this.children[i].toIncrementalDOM()();
            }
        }
    }

    pgParserNode.prototype.patchDOMElementWithIncrementalDOM = function ($el) {
        var node = $el.get(0);
        if (node.nodeName.toLowerCase() !== this.tagName) {
            $el.replaceWith(this.toStringWithIds());
        } else {
            //see if attrs differ
            var has = [];
            this.getAttrList().forEach(function (attr) {
                if ($el.attr(attr.name) !== attr.value) {
                    $el.attr(attr.name, attr.value);
                }
                has.push(attr.name.toLowerCase());
            })
            //remove unused
            var remove = [];
            for (var i = 0; i < node.attributes.length; i++) {
                var attrib = node.attributes[i];
                if (attrib.specified) {
                    if (has.indexOf(attrib.name.toLowerCase()) < 0) {
                        remove.push(attrib.name);
                    }
                }
            }
            remove.forEach(function (attr) {
                $el.removeAttr(attr);
            })

            IncrementalDOM.patch(node, this.toIncrementalDOMInner());
        }
    }

    pgParserNode.prototype.isHidden = function () {
        return this.hasAttr('data-pg-hidden');
    }

    pgParserNode.prototype.hide = function () {
        if (!this.isHidden()) {
            this.withEmitEvent(function () {
                this.setAttr('data-pg-hidden');
            }, 'hide', null, {
                domOnly: false
            });
        }
    }

    pgParserNode.prototype.show = function () {
        if (this.isHidden()) {
            this.withEmitEvent(function () {
                this.removeAttr('data-pg-hidden');
            }, 'show', null, {
                domOnly: false
            });
        }
    }

    //trigger dom view update
    pgParserNode.prototype.update = function (oldHtml, domPgId /* which element was updated */) {
        this.withEmitEvent(function () {

        }, 'update', null, {
            domHtml: this.getViewHtml(),
            domPgId: domPgId,
            oldValue: oldHtml
        });
    }

    pgParserNode.prototype.updateInner = function (oldHtml) {
        this.withEmitEvent(function () {

        }, 'updateInner', null, {
            domHtml: this.getViewInnerHtml(),
            oldValue: oldHtml
        });
    }

    //pgParserNode DOM events



    pgParserNode.prototype.executeScript = function (code) {

        this.withEmitEvent(function () {

        }, 'executeScript', null, {
            domOnly: true,
            code: code
        });
    }

    pgParserNode.prototype.scrollTo = function () {
        this.withEmitEvent(function () {

        }, 'scrollTo', null, {
            domOnly: true
        });
    }

    pgParserNode.prototype.selected = function () {
        this.withEmitEvent(function () {

        }, 'selected', null, {
            domOnly: true
        });
    }

    pgParserNode.prototype.deselected = function () {
        this.withEmitEvent(function () {

        }, 'deselected', null, {
            domOnly: true
        });
    }

    pgParserNode.prototype.removeFromDOM = function () {
        this.withEmitEvent(function () {

        }, 'removeFromDOM', null, {
            domOnly: true
        });
    }
    //End pgParserNode methods


    var pgParser = window.pgParser = function () {

        var source = null;

        this.singleTags = pgSingleTags;
        this.closingTags = pgAutoClosedTags;

        var _this = this;

        this.assignIds = true;
        this.idPrefix = null;

        var tokenGetTagNameState = 0;

        var tokenGetTagNameRe = /[a-z0-9\?!:\-\=\%\_\.]/i; //Added . to support React

        var CHAR_EXCLAMATION = '!'.charCodeAt(0);
        var CHAR_SLASH = '/'.charCodeAt(0);
        var CHAR_CLOSETAG = '>'.charCodeAt(0);
        var CHAR_OPENTAG = '<'.charCodeAt(0);
        var CHAR_QUOTE = '"'.charCodeAt(0);
        var CHAR_SINGLEQUOTE = '\''.charCodeAt(0);
        var CHAR_EQUAL = '='.charCodeAt(0);
        var CHAR_PERCENT = '%'.charCodeAt(0);

        var tokenGetTagName = function (ch, s, ch_code) {
            if (s.length === 0) {
                tokenGetTagNameState = 0;
                if (ch_code === CHAR_SLASH) return true; //only allow / as the first char of tag name. otherwise <br/> is wrongly parsed
            }
            if (tokenGetTagNameState === 2) {
                return false; //prev char was = or %, this could be <%miskolin
            }
            if (tokenGetTagNameState === 0) {
                if (ch_code === CHAR_EXCLAMATION /* '!' */) {
                    tokenGetTagNameState = 1;
                } else if (ch_code === CHAR_EQUAL || ch_code === CHAR_PERCENT) {
                    tokenGetTagNameState = 2; //next char will be false - ends script tags like <%miskolin %>
                    return true;
                }
            }
            if (tokenGetTagNameState === 1) {
                // we have a !xxx tag
                if (s === '!--') return false; //we have a comment
            }
            return ch.match(tokenGetTagNameRe);
        }

        var tokenGetTagAttributes = function (ch, s, ch_code) {
            return ch_code !== CHAR_CLOSETAG; //'>';
        }
        this.replaceExistingIds = false;
        this.replaceExistingExternalIds = false;

        this.nodeCatalogue = pgParserNodeCatalogueInstance;

        this.rootNode = null;

        this.getNode = function (id) {
            return this.nodeCatalogue ? this.nodeCatalogue.get(id) : null;
        }

        this.parse = function (s, done, chunkDone) {

            source = s;
            var _this = this;

            var nodeLevels = [];

            this.rootNode = new pgParserNode;
            this.rootNode.tagName = 'document';
            this.rootNode.rootNode = true;
            this.rootNode.nodeCatalogue = this.nodeCatalogue;
            this.rootNode.document = this.rootNode;

            if (this.assignIds) {
                this.rootNode.assignId();
                if (this.nodeCatalogue) this.nodeCatalogue.add(this.rootNode);
            }

            this.replacedIds = [];

            //nodeLevels.push(this.rootNode);

            var len = source.length;
            var currentNode = this.rootNode;
            var node = null;

            var i = 0;


            var readToken = function (func, parse_quotes, attr_mode) {
                var s = '';
                var start_script = '<?php';
                var start_script2 = '<?=';
                var end_script = '?>';

                var in_script = false;
                do {
                    var si = startsWithCaseInsensitiveScript(source, i);
                    //if(startsWithCaseInsensitive(start_script, source, i) || startsWithCaseInsensitive(start_script2, source, i)) {
                    if (si) {
                        //if(pgTestAtIndex(source, pgScriptsStartsWithRegExp, i)) {
                        in_script = true;
                        end_script = si.close;
                    } else if (startsWithCaseInsensitive(end_script, source, i)) {
                        in_script = false;
                        s += source.substr(i, end_script.length);
                        i += end_script.length;
                        if (i >= len) break;
                        continue;
                    }
                    var ch = source.charAt(i);
                    var ch_code = source.charCodeAt(i);

                    if (in_script) {
                        s += ch;
                    } else {
                        if (!func(ch, s, ch_code)) {
                            return s;
                        } else {
                            s += ch;
                            var attr_html_mode = false; //allow <>, dont auto fix open attrs
                            if (parse_quotes && (ch_code === CHAR_QUOTE || ch_code === CHAR_SINGLEQUOTE)) {
                                var quote = ch;
                                var quote_code = ch_code;
                                var escaped = false;
                                i++;
                                while (i < len) {
                                    ch = source.charAt(i);
                                    ch_code = source.charCodeAt(i);
                                    if (ch_code === quote_code && !escaped) {
                                        s += ch;
                                        //i++;
                                        break;
                                    }
                                    if (attr_mode && ch_code === CHAR_OPENTAG) {
                                        attr_html_mode = true; //looks like <tags> in attr value
                                        //perhaps a script tag?
                                        var asi = startsWithCaseInsensitiveScript(source, i);
                                        if (asi) {
                                            var idx = source.indexOf(asi.close, i);
                                            if (idx >= i) {
                                                s += source.substr(i, idx - i + asi.close.length);
                                                i = idx + asi.close.length;
                                                attr_html_mode = false;
                                                continue;
                                            }
                                        }
                                    } else if (attr_mode && ch_code === CHAR_CLOSETAG && !attr_html_mode) {
                                        //missing "?
                                        var fix = true;
                                        var qidx = source.indexOf(quote, i);
                                        if (qidx >= 0) {
                                            var qstr = source.substr(qidx, 2);
                                            if (qstr === `${quote}>` || qstr === `${quote} ` || qstr === `${quote}/`) {
                                                fix = false; //looks ok
                                            }
                                        }
                                        if (fix) return s + quote;
                                    }
                                    /*
                                     if(!escaped && ch == '\\') {
                                     //escaped = true; disable escaping quotes
                                     } else {
                                     escaped = false;
                                     }
                                     */
                                    s += ch;
                                    i++;
                                }
                            }
                        }
                    }
                    i++;
                }
                while (i < len);
                return s;
            }

            var readUntilIncluding = function (tok) {
                var attrs = null;
                var end = source.indexOf(tok, i);
                if (end < 0) {
                    attrs = source.substr(i, len - i);
                    i = len - 1;
                } else {
                    attrs = source.substr(i, end - i + tok.length - 1);
                    i = end + tok.length - 1;
                }
                return attrs;
            }

            var isPrefixedId = function (id) {
                if (typeof id !== 'string') return false;
                var c = id.charCodeAt(0);
                return !(c >= 48 && c <= 57);
            }


            var doBatch = function () {
                var start_ms = (new Date()).getTime();
                var chunk_max_time = 50;
                var chunks_pi = 100;
                var m = 0;

                var abort_check_count = 0;

                var shouldAbort = function () {
                    abort_check_count++;
                    if (abort_check_count % 100 === 0) {
                        if ((new Date()).getTime() - start_ms > 5000 || nodeLevels.length > 1000) {
                            _this.rootNode.remove();
                            _this.rootNode = null;
                            throw "Parsing the HTML source if taking too much time. Either it is too large, it includes too many syntax errors or it is a messy legacy HTML.";
                            if (done) done();
                            return true;
                        }
                    }
                    return false;
                }

                while (i < len) {

                    if (shouldAbort()) return;

                    var idx = source.indexOf('<', i);
                    if (idx < 0) {
                        //only text till the end
                        var textNode = new pgParserNode();
                        textNode.tagName = 'text';
                        textNode.content = source.substr(i);
                        textNode.closingTag = textNode.tagName;
                        textNode.textNode = true;
                        textNode.document = _this.rootNode;

                        if (textNode.content.length) {
                            currentNode.addChild(textNode);
                        }
                        textNode.startStringIndex = i;
                        i = len;
                    } else {
                        if (idx > i) {
                            var textContent = source.substr(i, idx - i).split("\n");
                            var text_start_idx = i;
                            for (var ti = 0; ti < textContent.length; ti++) {
                                var textNode = new pgParserNode();
                                textNode.tagName = 'text';
                                textNode.content = ((ti > 0) ? "\n" : "") + textContent[ti];
                                textNode.closingTag = textNode.tagName;
                                textNode.textNode = true;
                                textNode.document = _this.rootNode;
                                textNode.startStringIndex = text_start_idx;
                                text_start_idx += textNode.content.length;

                                if (textNode.content.length) {
                                    currentNode.addChild(textNode);
                                }
                            }

                            //  console.log('text = "' + textNode.content + '"');
                        }
                        i = idx + 1;

                        var tag = readToken(tokenGetTagName);
                        var tagLower = tag.toLowerCase();
                        var tag_idx = i - tag.length - 1;

                        var attrs;
                        var comment = false;
                        var script_is_closed = false;

                        var si = null;

                        if (tag.charCodeAt(0) === CHAR_EXCLAMATION /* '!' */ && tag.indexOf('!--') == 0) {
                            attrs = readUntilIncluding('-->');
                            comment = true;
                            // } else if(tagLower === '?php' || tagLower === '?=') {
                        } else if (si = pgGetScriptInfoByTag(tagLower)) {
                            attrs = readUntilIncluding(si.close);
                            if (attrs.length && attrs.endsWith(si.close[0])) {
                                attrs = attrs.substr(0, attrs.length - 1);
                                script_is_closed = true;
                            }
                        } else {
                            attrs = readToken(tokenGetTagAttributes, true, true);
                        }

                        var selfClosedWithSlash = false;

                        if (attrs.length && attrs.charCodeAt(attrs.length - 1) === CHAR_SLASH /* '/' */) {
                            selfClosedWithSlash = true;
                            attrs = attrs.substr(0, attrs.length - 1);
                        }
                        var isSingleTag = _this.singleTags.indexOf(tag.toLowerCase()) >= 0 || selfClosedWithSlash;

                        if (comment) {
                            isSingleTag = true;
                        }
                        i++; //skip >

                        //console.log('<' + tag + attrs + '>');

                        if (tag.charCodeAt(0) !== CHAR_SLASH /* '/' */) {
                            //opening tag
                            node = new pgParserNode();
                            node.tagName = tagLower;
                            node.tagNameOriginal = tag;
                            //si = pgGetScriptInfoByTag(tagLower);
                            if (si) {
                                //if(tagLower === '?php' || tagLower === '?=') {
                                node.script = true;
                                node.script_info = si;
                                node.isElementOrScript = true;
                                node.content = attrs;
                                attrs = null;
                                node.scriptClosed = script_is_closed;
                            } else {
                                node.script = false;
                            }
                            node.attributes = attrs;
                            node.comment = comment;
                            node.document = _this.rootNode;
                            node.selfClosed = selfClosedWithSlash;

                            node.isElement = !node.script && !node.comment && node.tagName.charCodeAt(0) !== CHAR_EXCLAMATION /* '!' */;
                            node.isElementOrScript = node.isElement || node.script;
                            node.startStringIndex = tag_idx;

                            //node.getAttrList();

                            //MTCH
                            //if node is autoclosing tag, check if it is currently in such tag. if so close that tag and add this one after it
                            //if(_this.closingTags[node.tagName].indexOf(node.children[ii].tagName) >= 0) {
                            ////

                            currentNode.addChild(node);

                            if (isSingleTag) {
                                node.singleTag = true;
                                node.closed = true;
                                node.endStringIndex = i;
                            } else {
                                nodeLevels.push(currentNode);
                                currentNode = node;
                            }

                            if (tagLower === 'script' || tagLower === 'style' || tagLower === 'php' || tagLower === 'textarea') {
                                node.content = readUntilIncluding('</' + tag + '>');
                                node.closingTag = tag;
                                if (node.content.length >= tag.length + 2) {
                                    node.content = node.content.substr(0, node.content.length - tag.length - 2);
                                }
                                node.closed = true;
                                currentNode = nodeLevels.length ? nodeLevels.pop() : currentNode;
                                i++;
                            }

                            var id = node.getId();

                            if (_this.assignIds && node.shouldHaveId()) {
                                if (!id) {
                                    id = node.assignId(_this.idPrefix);
                                } else if (_this.replaceExistingIds) {
                                    if (_this.replaceExistingExternalIds) {
                                        var nid = node.assignId(_this.idPrefix);
                                        //_this.replacedIds.push({old: id, new: nid});
                                        id = nid;
                                    } else {
                                        if (!isPrefixedId(id)) {
                                            var nid = node.assignId(_this.idPrefix);
                                            id = nid;
                                        }
                                    }
                                } else {
                                    var id_num = parseInt(id);
                                    if (!isNaN(id_num) && id_num > pgIdCount) {
                                        pgIdCount = id_num;
                                    }
                                }
                                if (_this.nodeCatalogue) _this.nodeCatalogue.add(node);
                            }
                            node.getAttrList();
                        } else {
                            //closing tag
                            var closedTag = tag.replace('/', '').toLowerCase();
                            if (currentNode.tagName === closedTag) {
                                currentNode.closingTag = tag.replace('/', '');
                                currentNode.closed = true;
                                currentNode.endStringIndex = i;
                                currentNode = nodeLevels.length ? nodeLevels.pop() : currentNode;
                            } else if(node && node.singleTag && closedTag === node.tagName ) {
                                //closing tag of a single tag... this doesn't work perfectly if any text is between the tags
                                node.closingTag = tag.replace('/', '');
                                node.closed = true;
                                node.endStringIndex = i;
                                node.singleTag = false;
                            } else {
                                var closedLevel = -1;
                                for (var idx = nodeLevels.length - 1; idx >= 0; idx--) {

                                    if (nodeLevels[idx].tagName === closedTag) {
                                        closedLevel = idx;
                                        nodeLevels[idx].closingTag = tag.replace('/', '');
                                        nodeLevels[idx].closed = true;

                                        //nodeLevels.splice(idx, nodeLevels.length - idx);
                                        break;
                                    }
                                }


                                if (closedLevel >= 0) {

                                    while (nodeLevels.length > closedLevel + 0) {

                                        if (shouldAbort()) return;

                                        var node = nodeLevels.pop();

                                        //var node = nodeLevels.shift();

                                        node.closed = true;

                                        if (_this.closingTags[node.tagName] && node.children.length) {
                                            var ii = 0;
                                            var last = node;
                                            while (ii < node.children.length) {

                                                if (shouldAbort()) return;

                                                if (_this.closingTags[node.tagName].indexOf(node.children[ii].tagName) >= 0) {
                                                    last = node.children[ii].insertAfter(last);
                                                } else {
                                                    ii++;
                                                }
                                            }
                                        }
                                    }
                                    currentNode = nodeLevels.length ? nodeLevels.pop() : currentNode;
                                    //currentNode = nodeLevels.length ? nodeLevels.pop() : currentNode;

                                } else {
                                    currentNode.closingTag = tag.replace('/', '');
                                    currentNode.closed = true;
                                    currentNode = nodeLevels.length ? nodeLevels.pop() : currentNode;
                                }
                            }
                        }
                    }

                    m++;
                    if (m % 10 === 0 && done) {
                        var elapsed_ms = (new Date()).getTime() - start_ms;
                        if (elapsed_ms >= chunk_max_time) {
                            if (chunkDone) chunkDone();
                            setTimeout(doBatch, 10);
                            break;
                        }
                    }
                }
                if (i >= len && done) done();
            }
            doBatch();
        }

        this.find = function (sel) {
            return this.rootNode.find(sel);
        }

        this.toStringOriginal = function (format_html, options, func) {
            return this.rootNode.toStringOriginal(format_html, options, func);
        }

        this.toStringWithIds = function (format_html, options, func) {
            return this.rootNode.toStringWithIds(format_html, options, func);
        }

        this.validate = function (strict) {
            return this.rootNode.validateTree(strict);
        }
    }


    var PgNodeEvent = window.PgNodeEvent = function (pgel, operation, obj_or_data, data, page) {

        this.pgId = pgel.getId();
        if (pgel.textNode) {
            this.text = pgel.content;
        } else if(!this.pgId && pgel.shouldHaveId()) {
            if(pgIsDev()) {
                console.error('PgNodeEvent: no pgId');
            }
        }
        this.operation = operation;
        this.data = null;
        this.objectId = null;
        this.eventType = "PgNodeEvent";
        this.page = page;

        if (obj_or_data) {
            if (obj_or_data instanceof pgParserNode) {
                this.objectId = obj_or_data.getId();
            } else {
                this.data = obj_or_data;
            }
        }

        if (data) this.data = data;
        if (!this.data) this.data = {};
    }


    PgNodeEvent.prototype.toDebugString = function () {
        var s = this.eventType + ' "' + this.operation + '" (' + this.pgId + ', ' + this.objectId + ', {';
        if (this.data) {
            //debugger;
            $.each(this.data, function (k, v) {
                s += k + ': "' + v + '", ';
            })
        }
        s += '})';
        return s;
    }

    PgNodeEvent.prototype.getPgel = function () {
        return getPgNodeByPgId(this.pgId);
    }


    var pgParserNode_emit_events = true;

    pgParserNode.prototype.emitEvent = function (operation, object, data, page) {
        if (!pgParserNode_emit_events) return null;
        if (object && object instanceof pgParserNode && !object.getId()) return null;
        if (!page) page = this.getPage();
        if (!page) return null;
        var o = new PgNodeEvent(this, operation, object, data, page);
        if (pinegrow.event_conductor) pinegrow.event_conductor.emit(o);
        return o;
    }

    pgParserNode.prototype.getEmitEvents = function () {
        return pgParserNode_emit_events;
    }

    pgParserNode.prototype.withEmitEvent = function (func, operation, object, data, page) {
        var orig = pgParserNode_emit_events;
        pgParserNode_emit_events = false;
        if (orig && data && typeof data.oldValue == 'function') {
            data.oldValue = data.oldValue.call(this);
        }
        func.call(this);
        pgParserNode_emit_events = orig;
        return this.emitEvent(operation, object, data, page);
    }

    pgParserNode.prototype.withoutEvents = function (func) {
        var orig = pgParserNode_emit_events;
        pgParserNode_emit_events = false;
        func.call(this);
        pgParserNode_emit_events = orig;
    }


    var PgEventConductor = window.PgEventConductor = function () {

        var count = 0;
        var debug = false;

        var list = [];
        var local_subscribers = [];

        this.emit = function (event) {
            event.eventId = count++;
            //list.push(event);
            if (debug) console.log('Emitted: ' + event.toDebugString());
            for (var i = 0; i < local_subscribers.length; i++) {
                local_subscribers[i](event);
            }
        }

        this.addLocalSubscriber = function (handler, position) {
            if (local_subscribers.indexOf(handler) < 0) {
                if (position < local_subscribers.length) {
                    local_subscribers.splice(position, 0, handler);
                } else {
                    local_subscribers.push(handler);
                }
            }
        }

        this.removeLocalSubscriber = function (handler) {
            var idx = local_subscribers.indexOf(handler);
            if (idx >= 0) {
                local_subscribers.splice(idx, 1);
            }
        }
    }

    var PgEventHandler = window.PgEventHandler = function (port, position) {

        var debug = false;
        var _this = this;

        this.handle_dom_events = true;

        if (port === undefined) port = 0;
        if (position === undefined) position = 9999;

        this.on_event = null;

        var received = function (event) {
            if (debug) console.log('Received: ' + event.toDebugString());
            if (_this.on_event) {
                if (_this.handle_dom_events || !event.data.domOnly) {
                    _this.on_event(event);
                }
            }
        }

        this.start = function () {
            if (port === 0) {
                pinegrow.event_conductor.addLocalSubscriber(received, position);
            }
        }

        this.stop = function () {
            if (port === 0) {
                pinegrow.event_conductor.removeLocalSubscriber(received);
            }
        }

        this.destroy = function () {
            this.stop();
        }
    }

    var PgNodeEventDOMPlayer = window.PgNodeEventDOMPlayer = function ($iframe, pageView) {

        var operations = {};

        var html = null;

        var getDocument = function () {
            return html.ownerDocument;
        }

        var getBody = function () {
            return html.ownerDocument.body;
        }


        var getElement = function (pgid, event) {
            if (!pgid) {
                //text node?
                if (event.text) {
                    return $(document.createTextNode(event.text));
                }
            }
            var sel = '[data-pg-id="' + pgid + '"]';
            if (html.getAttribute('data-pg-id') == pgid) return $(html);
            var el = html.querySelector(sel);
            if (el) return $(el);

            //new element?
            var pgel = pgParserNodeCatalogueInstance.get(pgid);
            if (pgel) {
                var page = pinegrow.getCrsaPageOfPgParserNode(pgel);
                var html_code = page.getViewHTMLForElement(pgel);

                //is this a svg node?
                if(['svg'].indexOf(pgel.tagName) >= 0) {
                    html_code = html_code.replace(/\sdata\-pg\-collapsed(=\"[^\"]\")?/g, '');
                    var doc = html.ownerDocument;
                    var doc2 = (new DOMParser()).parseFromString(html_code, "text/xml");
                    return $(doc.adoptNode(doc2.documentElement));
                } else {
                    return $(html_code);
                }
            }

            return null;
        }

        this.play = function (event, html_el) {
            if (event.data.sourcePageView && event.data.sourcePageView === pageView) {
                //DOM for this pageview is already updated, most likely this operation comes from inline editing
                return;
            }
            //console.log(event);

            try {

                html = html_el;
                var $el = getElement(event.data.domPgId || event.pgId, event);
                var $obj;
                if (event.data.objectHtml || event.data.objectDomHtml) {
                    $obj = $(event.data.objectDomHtml || event.data.objectHtml);
                } else {
                    $obj = event.objectId ? getElement(event.objectId) : null;
                }

                if (!$el || $el.length === 0) {
                    //ups
                    //console.log('Element ' + event.pgId + ' not found');
                    //Probably not op for this view
                    return;
                }

                /*
                 if(event.operation === 'setAttr') {
                 var pgel = getElementPgNode($el);

                 var comp_pgel = pgel ? pgel.closest('welcomecomp') : null;
                 if(comp_pgel) {
                 var page = pinegrow.getCrsaPageOfPgParserNode(comp_pgel);
                 var html_code = page.getViewHTMLForElement(comp_pgel);
                 var $comp = getElement(comp_pgel.getId(), {});
                 $comp.replaceWith(html_code);
                 done = true;
                 }
                 }*/

                if (event.operation in operations) {
                    operations[event.operation](event, $el, $obj);
                    refreshDisplay();
                } else {
                    throw "PgNodeEventDOMPlayer: operation " + event.operation + ' not implemented!';
                }
            } catch(err) {
                console.log('domPlayerError', err);
            }
        }

        this.destroy = function() {

        }

        var refreshDisplayTimer = null;

        var refreshDisplay = function () {
            return;
            //try without this
            if (refreshDisplayTimer) {
                clearTimeout(refreshDisplayTimer);
            }
            refreshDisplayTimer = setTimeout(function () {
                var $contents = $(html.ownerDocument);
                var o = $contents.scrollTop();
                $html.find('>body').redraw();
                $contents.scrollTop(o);
                refreshDisplayTimer = null;
            }, 200);
        }

        //update the view of the element
        operations.update = function (e, $el) {
            var p = new pgParser();
            p.assignIds = false;
            p.parse(e.data.domHtml || e.data.html || "");

            var html = p.rootNode.children[0].toStringWithIds();
            html = pageView.emulateCSSFeaturesInHTML(html);

            $el.replaceWith(html);

            //p.rootNode.children[0].patchDOMElementWithIncrementalDOM($el); //todo30 - breaks things on syntax errors in html
            return;

            $el.replaceWith($(e.data.domHtml || e.data.html));
        }

        operations.updateInner = function (e, $el) {
            var html = e.data.domHtml || e.data.html || "";
            html = pageView.emulateCSSFeaturesInHTML(html);
            $el.html(html);
        }

        operations.addClass = function (e, $el) {
            $el.addClass(e.data.class);
        }

        operations.removeClass = function (e, $el) {
            $el.removeClass(e.data.class);
        }

        operations.setAttr = function (e, $el, $obj) {
            var val = e.data.domValue || e.data.value || "";
            if (e.data.attr === 'style') {
                val = pageView.emulateCSSFeatures(val);
            }

            if(e.data.attr === 'src') {
                if(val && val.startsWith('/')) {
                    var project = pinegrow.getCurrentProject();
                    if(project && !pgf.root_urls) {
                        val = project.getUrl() + val;
                        val = pinegrow.getProxyUrl(val);
                    }
                }
            }
            $el.attr(e.data.attr, val);

            if (e.data.attr === 'src' && $el.get(0).nodeName === 'IMG') {
                $el.off('.pg').one('load.pg', function (e) {
                    pinegrow.selectedElements.repositionMenusDelayed();
                })
            }
        }

        operations.removeAttr = function (e, $el, $obj) {
            $el.removeAttr(e.data.attr);
        }

        operations.replaceTag = function (e, $el, $obj) {
            var attrs = {};

            $.each($el.get(0).attributes, function (idx, attr) {
                attrs[attr.nodeName] = attr.nodeValue;
            });
            var $n;
            $el.replaceWith(function () {
                $n = $("<" + e.data.tag + "/>", attrs).append($el.contents());
                return $n;
            });
        }

        operations.replaceWith = function (e, $el, $obj) {
            if ($el.is('body')) {
                //special case
                var code = e.data.objectDomHtml || e.data.objectHtml;

                code = pageView.emulateCSSFeaturesInHTML(code);

                var p = new pgParser();
                p.assignIds = false;
                p.nodeCatalogue = null;
                p.parse(code);
                var bodyel = p.rootNode.children[0];

                var remove = [];
                $.each($el.prop("attributes"), function () {
                    remove.push(this.name);
                });
                remove.forEach(function (name) {
                    $el.removeAttr(name);
                })

                bodyel.getAttrList().forEach(function (a) {
                    $el.attr(a.name, a.value || '');
                })

                $el.get(0).innerHTML = bodyel.html(null, true);

                if (e.data.detach) {
                    console.log('player.replaceWith WARNING: Can not handle detach for body');
                }

            } else {
                $obj.insertBefore($el);
                if (e.data.detach) {
                    $el.detach();
                } else {
                    $el.remove();
                }
            }
        }

        operations.replaceContentWithContentOf = function (e, $el, $obj) {
            if (e.data.detach) {
                $el.contents().detach();
            } else {
                $el.contents().remove();
            }
            $el.append($obj.contents());
        }

        operations.replaceContentWithElement = function (e, $el, $obj) {
            if (e.data.detach) {
                $el.contents().detach();
            } else {
                $el.contents().remove();
            }
            $el.append($obj);
        }

        operations.detag = function (e, $el, $obj) {
            $el.contents().insertBefore($el);
            $el.remove();
        }

        operations.removeAllChildren = function (e, $el, $obj) {
            $el.contents().remove();
        }

        operations.html = function (e, $el, $obj) {
            $el.html(e.data.domHtml || e.data.html);
        }

        operations.remove = function (e, $el, $obj) {
            $el.remove();
        }

        operations.removeFromDOM = function (e, $el, $obj) {
            $el.remove();
        }

        operations.detach = function (e, $el, $obj) {
            $el.detach();
        }

        operations.appendPrepend = function (e, $el, $obj) {
            if (e.data.prepend) {
                $obj.prepend($el);
            } else {
                $obj.append($el);
            }
        }

        operations.insertBefore = function (e, $el, $obj) {
            $el.insertBefore($obj);
        }

        operations.insertAfter = function (e, $el, $obj) {
            $el.insertAfter($obj);
        }

        operations.insertAtIndex = function (e, $el, $obj) {
            var parent = $obj.get(0);
            var $children = $obj.children();
            var $before = e.data.element_index < $children.length ? $($children.get(e.data.element_index)) : null;
            if ($before) {
                $el.insertBefore($before);
            } else {
                $obj.append($el);
            }
        }

        var temp_id_count = 0;

        operations.executeScript = function (e, $el, $obj) {
            var scr = getDocument().createElement('script');
            scr.async = false;

            var code = e.data.code;

            var a = null;

            var re = /__THIS_ELEMENT__/g;

            if(code.match(re)) {
                a = `data-pg-temp-id-${temp_id_count++}`;
                $el.attr(a, '');
                code = code.replace(re, `document.querySelector('[${a}]')`);
            }
            scr.text = code;
            getBody().appendChild(scr);

            setTimeout(function () {
                scr.parentNode.removeChild(scr);
                if(a) {
                    $el.removeAttr(a);
                }
            }, 500);
        }

        operations.scrollTo = function (e, $el, $obj) {
            crsaScrollToElementInIframe($iframe, $el);
        }

        operations.selected = function (e, $el, $obj) {

        }

        operations.deselected = function (e, $el, $obj) {

        }

        operations.hide = function (e, $el, $obj) {
            $el.attr('data-pg-hidden', '');
        }

        operations.show = function (e, $el, $obj) {
            $el.removeAttr('data-pg-hidden', '');
        }

        operations.removeAttrIfStartsWith = function () {
            //dummy
        }
    }


    /////Shortlived node cached info

    var PgNodeInfo = function (pgel) {

        this.pgel = pgel;
        this.$el = null;
        this.$els = null;
        this.page = null;
        this.def = null;
        this.defs = null;
    }

    PgNodeInfo.prototype.get$El = function () {
        if (!this.$el) {
            this.$el = this.pgel.get$DOMElement(this.getPage().get$Html());
        }
        return this.$el;
    }

    PgNodeInfo.prototype.get$Els = function () {
        if (!this.$els) {
            this.$els = this.pgel.get$DOMElements(this.getPage());
        }
        return this.$els;
    }

    PgNodeInfo.prototype.getPage = function () {
        if (!this.page) {
            this.page = this.pgel.getPage();
        }
        return this.page;
    }

    PgNodeInfo.prototype.getMainType = function () {
        if (!this.def) {
            this.def = this.getPage().getMainType(this.get$El(), this.pgel); //todo30 - $el type?
        }
        return this.def;
    }

    PgNodeInfo.prototype.getAllTypes = function () {
        if (!this.defs) {
            this.$el = this.getPage().getAllTypes(this.get$El(), this.pgel); //todo30 - $el type?
        }
        return this.defs;
    }

    pgParserNode.prototype.getInfo = function () {
        return new PgNodeInfo(this);
    }

    ////end PgNodeInfo


    if (typeof module !== 'undefined') {
        module.exports = {
            pgParser: pgParser,
            pgParserNode: pgParserNode
        }
    }


})();

