window.PgAllStylesheets = function () {
    var stylesheetsByKey = {};
    var crsaStylesheets = [];
    var _this = this;
    var csIndex = 1;

    /* Preload some node modules to speed up page loading */
    var preload = true;
    if(preload) {
        require('postcss');
        // require('postcss-safe-parser');
        // require('autoprefixer');
        // require('stylefmt');
    }
    /* end preload */

    this.getByUrl = function(url, ignore_source) {
        var r = this.findCrsaStylesheetForUrl(url, null, ignore_source);
        return r.length ? r[0] : null;
    }

    this.findCrsaStylesheetForUrl = function (url, ignore_ext, dont_map_to_source) {

        // url = crsaRemoveUrlParameters(url);
        if(pinegrow.httpServer) {
            url = pinegrow.httpServer.getOriginalUrl(url);
        }

        //broken
        var removeExt = function(url) {
            if(!url) return url;
            var a = url.split('.');
            if(a.length > 1) a.pop();
            return a.join('.');
        }

        ignore_ext = false;
        if (ignore_ext) url = removeExt(url);

        var add = function(cs) {
            (r.indexOf(cs) < 0) && r.push(cs)
        }

        var r = [];
        var forAll = function (stylesheets) {
            for (var i = 0; i < stylesheets.length; i++) {
                var stylesheet = stylesheets[i];

                if (!stylesheet.url) continue;

                // var stylesheetUrl = crsaRemoveUrlParameters(stylesheet.url);

                var csurl = ignore_ext ? removeExt(stylesheet.url) : stylesheet.url;
                var sourceStylesheet = stylesheet.sourceStylesheet;

                if(!dont_map_to_source) {
                    if(csurl === url || (sourceStylesheet && sourceStylesheet.url === url)) {
                        if(sourceStylesheet) {
                            stylesheet = sourceStylesheet;
                        }
                        add(stylesheet);
                    }
                } else {
                    if(csurl === url) {
                        add(stylesheet);
                    } else if(sourceStylesheet && sourceStylesheet.url === url) {
                        add(sourceStylesheet);
                    }
                }



                if (sourceStylesheet) stylesheet = sourceStylesheet;
                if (stylesheet.importedFiles && stylesheet.importedFiles.length) {
                    forAll(stylesheet.importedFiles);
                }
            }
        }
        forAll(Object.values(stylesheetsByKey));
        //forAll(crsaStylesheets);
        return r;
    }

    this.findCrsaInlineStylesheet = function (pgel) {
        for (var i = 0; i < crsaStylesheets.length; i++) {
            if (crsaStylesheets[i].inline) {
                if (pgel == crsaStylesheets[i].inlineSheetPgel) return crsaStylesheets[i];
            }
        }

        return null;
    }

    this.findCrsaInlineStylesheetUsingSelector = function (selector) {
        var fileName = '';

        var matchedPgFile = selector.match(/\.pg-file-(.*)/);
        if (matchedPgFile && matchedPgFile.length > 1) {
            fileName = matchedPgFile[1];
        }

        if (fileName) {
            for (var i = 0; i < crsaStylesheets.length; i++) {
                if (fileName == crsaStylesheets[i].getUniqueName()) return crsaStylesheets[i];
            }
        }

        return null;
    }

    this.findCrsaStylesheetForUniqueName = function (uniqueName) {

        var isTheSameStylesheet = function (stylesheet) {
            if (stylesheet.getUniqueName() == uniqueName) return true;
            else                                          return false;
        }

        var forAll = function (stylesheets) {
            for (var i = 0; i < stylesheets.length; i++) {
                var stylesheet = stylesheets[i];
                if (isTheSameStylesheet(stylesheet)) {
                    return stylesheet;
                }
                else if (stylesheet.sourceStylesheet) {
                    stylesheet = stylesheet.sourceStylesheet;
                    if (isTheSameStylesheet(stylesheet)) {
                        return stylesheet;
                    }
                }

                if (stylesheet.importedFiles && stylesheet.importedFiles.length) {
                    stylesheet = forAll(stylesheet.importedFiles);
                    if (stylesheet) return stylesheet;
                }
            }
            return null;
        }

        var cs = forAll(Object.values(stylesheetsByKey));
        //this doesnt find inline cs
        if(!cs) {
            cs = forAll(crsaStylesheets);
        }
        return cs;
        //return forAll(crsaStylesheets);
    }

    this.getCrsaStylesheetByUniqueName = function (index) {
        for (var i = 0; i < crsaStylesheets.length; i++) {
            var cs = crsaStylesheets[i];
            if (cs.sourceStylesheet) cs = cs.sourceStylesheet;
            if (cs.getUniqueName() == index) return cs;
        }
        return null;
    }

    this.editCode = function(cs) {
        if(!isApp()) {
            return pinegrow.openFileInCodeEditor(cs.url, cs.url);
        } else if(cs.localFile) {
            return pinegrow.openFileInCodeEditor(cs.localFile, cs.url);
        } else {
            pinegrow.showQuickError('Remote and inline stylesheets can\'t be edited in the code editor.');
        }
    }

    this.getStylesheetForHTMLElement = function(node) {
        //link or style
        if(node.tagName == 'LINK') {
            if(node.sheet && node.sheet.href) {
                return this.getByUrl(node.sheet.href);
            }
        } else if(node.tagName == 'STYLE') {
            var data_url = node.getAttribute('data-pg-style-url');

            if(data_url) {
                return this.getByUrl(data_url);
            }
            var pgel = getElementPgNode($(node), true /* dont create dyn elements */);
            if(pgel) {
                return this.findCrsaInlineStylesheet(pgel);
            }
        }
        return null;
    }

    this.getAllSheeetsForPage = function (page) {
        var doc = page.$iframe.get(0).contentDocument || page.$iframe.get(0).contentWindow.document;
        var length = doc.styleSheets.length;
        var styleSheets = doc.styleSheets;

        var sheets = [];
        for(var i = 0; i < length; i++) {
            var sheet = styleSheets[i];
            if (sheet.cssRules == null) continue;

            var pgStylesheets = this.findCrsaStylesheetForUrl(sheet.href);
            if (pgStylesheets && pgStylesheets.length > 0) sheets.push(pgStylesheets[0]);
        }

        return sheets;
    }

    this.recomplileAllSheetsInPage = function (page, done) {
        var doc = page.$iframe.get(0).contentDocument || page.$iframe.get(0).contentWindow.document;
        var length = doc.styleSheets.length;
        var styleSheets = doc.styleSheets;

        var stylesheetLoaded = 0;
        var onRegenerated = function () {
            stylesheetLoaded--;
            if (stylesheetLoaded == 0) {
                if (done) done();
            }
        }

        for (var i = 0; i < length; i++) {
            var sheet = styleSheets[i];
            if (sheet.cssRules == null) continue;

            var allStylesheets = this.findCrsaStylesheetForUrl(sheet.href);
            for (var j = 0; j < allStylesheets.length; j++) {
                var pgStylesheet = allStylesheets[j];

                if (pgStylesheet && !pgStylesheet.ignore) {
                    stylesheetLoaded++;
                    pgStylesheet.regenerateAndSetCssSource(onRegenerated);
                }
            }
        }
    }

    this.isStylesheetUsed = function (cs) {
        var pages = pinegrow.getAllPages();
        for (var i = 0; i < pages.length; i++) {
            var page = pages[i];
            var csInfo = this.getStylesheetInfo(cs, page);
            if (csInfo) return true;
        }

        return false;
    }

    this.stylsheetUsed = function (csObj, asStr, forCS) {
        var parentCS = null;
        var compiledCs = null;
        if (asStr == 'source') {
            compiledCs = forCS;
        }
        else if (asStr == 'child') {
            parentCS = forCS;
        }

        pinegrow.getAllPages().forEach(function (page) {
            if (page.stylesheets && page.stylesheets.isStylesheetUsed(forCS, csObj.pgStylesheet)) {
                page.stylesheets.addStylesheetAs(csObj, parentCS, compiledCs);
            }
        })
    }

    this.updateUsedStylesheet = function (oldUrl, newUrl, cs) {
        pinegrow.getAllPages().forEach(function (page) {
            if (page.stylesheets.isStylesheetUsed(cs)) {
                page.stylesheets.updateStylesheetUrl(oldUrl, newUrl, cs);
            }
        })
    }

    this.removeUsedStylesheet = function (cs, page) {
        if (!page) return;

        page.stylesheets.removeStylesheetByKey(cs);
    }

    this.getStylesheetInfo = function (cs, page) {
        if (page && page.stylesheets) {
            return page.stylesheets.getStylesheetInfo(cs);
        }

        return null;
    }

    this.addStylesheetByKey = function (url, cs) {
        if (url) stylesheetsByKey[url] = cs;
    }

    this.getStylesheetByKey = function (url) {
        return stylesheetsByKey[url];
    }

    this.updateStylesheetByKey = function (oldUrl, newUrl, cs) {
        if (stylesheetsByKey[oldUrl]) {
            if (stylesheetsByKey[oldUrl] == cs) {
                stylesheetsByKey[newUrl] = stylesheetsByKey[oldUrl];
                delete stylesheetsByKey[oldUrl];
            }
            else {
                throw "Error: same url with different stylesheet";
            }
        }
    }

    this.removeStylesheetByKey = function (url) {
        if (stylesheetsByKey[url]) {
            this.removeCrsaStylesheet(stylesheetsByKey[url]);
            delete stylesheetsByKey[url];
        }
    }

    this.getLastIndex = function (increase) {
        var index = csIndex;
        if (increase) csIndex++;

        return index + '';
    }

    this.addCrsaStylesheet = function (cs) {
        var index = crsaStylesheets.indexOf(cs);
        if (index < 0) {
            if (!cs.index) {
                cs.index = this.getLastIndex(true);
            }

            if (crsaStylesheets.indexOf(cs) < 0) {
                crsaStylesheets.push(cs);
            }
        }
    }

    this.removeCrsaStylesheet = function (cs) {
        var index = crsaStylesheets.indexOf(cs);
        if (index > -1) {
            crsaStylesheets.splice(index, 1);
            cs.remove();
        }
    }

    this.isPageHasStylesheet = function (page, pgStylesheet) {
        return page.stylesheets ? page.stylesheets.has(pgStylesheet) : false;
        /*
         var sheets = this.getAllSheeetsForPage(page);
         for (var i = 0; i < sheets.length; i++) {
         var currPgStylesheet = sheets[i];
         if (currPgStylesheet.compiledStylesheet) {
         currPgStylesheet = currPgStylesheet.compiledStylesheet;
         }

         if (currPgStylesheet.getUniqueName() == pgStylesheet.getUniqueName()) return true;
         }

         return false;*/
    }

    this.removeAllUnusedStylesheets = function() {
        var unused = crsaStylesheets.filter(function(cs) {
            return pinegrow.stylesheets.getPagesForStylesheet(cs).length === 0;
        })
        unused.forEach(function(cs) {
            cs.remove();
        })
    }

    this.getAll = function (options) {
        return pgcssh.filterStylesheetsList(crsaStylesheets, options);
    }

    this.getAllAsSelectOptions = function(options, do_short) {
        options = options || {};
        options.inline = false;
        options.ignored = false;
        options.loaded = true;

        var l = this.getAll(options);
        return l.map(function(cs) {
            var short = crsaGetNameFromUrl(cs.url);
            return {name: short, key: cs.url, html: (short + do_short ? '' : (' <small>' + cs.url + '</small>'))};
        })
    }

    this.getAllCrsaStylesheets = function (options) {
        return crsaStylesheets;
    }

    this.crsaStylesheetsLength = function () {
        return crsaStylesheets.length;
    }

    this.getCrsaStylesheetById = function (index) {
        return crsaStylesheets[index];
    }

    this.getAllSheetsForPages = function (pgStylesheet) {
        var pages = [];

        $.each(pinegrow.getAllPages(), function (i, page) {
            var doc = page.$iframe.get(0).contentDocument || page.$iframe.get(0).contentWindow.document;

            for(var i = 0; i < doc.styleSheets.length; i++) {
                var sheet = doc.styleSheets[i];

                if (pgStylesheet.inline) {
                    var $el = pgStylesheet.inlineSheetPgel.get$DOMElement();
                    if($el && $el[0] == sheet.ownerNode) pages.push(sheet);
                }
                else {
                    if (_this.isStylesheetUrlEqual(pgStylesheet.url, sheet.href)) {
                        pages.push(sheet);
                    } else if(!isApp()) {
                        var url = sheet.ownerNode.getAttribute('data-pg-style-url');
                        if(url && _this.isStylesheetUrlEqual(pgStylesheet.url, url)) {
                            pages.push(sheet);
                        }
                    }
                }
            }
        });

        return pages;
    }

    //Matjaz change: return pages instead of count of pages
    this.getPagesForStylesheet = function (pgStylesheet) {
        var list = [];

        $.each(pinegrow.getAllPages(), function (i, page) {
            var csInfo = pinegrow.stylesheets.getStylesheetInfo(pgStylesheet, page);
            if (csInfo && csInfo.compiledStylesheet) pgStylesheet = csInfo.compiledStylesheet;

            var doc = page.$iframe.get(0).contentDocument || page.$iframe.get(0).contentWindow.document;

            for(var i = 0; i < doc.styleSheets.length; i++) {
                var sheet = doc.styleSheets[i];

                if (_this.isStylesheetUrlEqual(pgStylesheet.url, sheet.href)) {
                    list.push(page);
                } else if(!isApp()) {
                    var url = sheet.ownerNode.getAttribute('data-pg-style-url');
                    if(url && _this.isStylesheetUrlEqual(pgStylesheet.url, url)) {
                        list.push(page);
                    }
                }
            }
        });

        return list;
    }

    this.isStylesheetUrlEqual = function (firstSSUrl, secondSSUrl) {
        if (firstSSUrl && secondSSUrl) {
            if(pinegrow.httpServer) {
                return pinegrow.httpServer.getOriginalUrl(firstSSUrl) === pinegrow.httpServer.getOriginalUrl(secondSSUrl);
            } else {
                return firstSSUrl === secondSSUrl;
            }
        }
        else {
            return false;
        }
    }

    this.findCssRules = function(rules, node) {
        if (node.type == "rule") {
            sel = normalizeSelector(node.selector);
        }
        else if (node.type == "atrule") {
            sel = node.params;
        }
        var list = [];

        var forAll = function (rules, sel) {
            if(rules) {
                $.each(rules, function(i, r) {
                    if (node.type == "rule" && r.type ==  4) {
                        if (node.parent.params == r.media[0]) {
                            forAll(r.cssRules, sel);
                        }
                    }
                    else {
                        if (node.type == "atrule" && r.type ==  4) {
                            if (node.params == r.media[0]) {
                                list.push(r);
                            }
                        }
                        else if (pgStylesheetHelper.getNormalizedSelectorForStylesheetRule(r) == sel) {
                            list.push(r);
                        }
                    }
                })
            }
        }

        forAll(rules, sel);
        return list;
    }

    this.updateNodeInStylesheet = function (node, source) {
        var pgStylesheet = pgStylesheetHelper.getStylesheetFromNode(node);

        $.each(pinegrow.getAllPages(), function (i, page) {
            var doc = page.$iframe.get(0).contentDocument || page.$iframe.get(0).contentWindow.document;

            for(var i = 0; i < doc.styleSheets.length; i++) {
                var sheet = doc.styleSheets[i];

                if (!_this.isStylesheetUrlEqual(pgStylesheet.url, sheet.href)) continue;

                var searchList = sheet.cssRules;
                if (searchList && searchList.length == 1 && searchList[0].media) {
                    searchList = searchList[0].cssRules || [];
                }
                var css_rules = _this.findCssRules(searchList, node);
                if(css_rules.length == 1) {
                    css_rules[0].style.cssText = source.replace("\n","");
                    console.log('single rule changed');
                }
            }
        });
    }

    this.getIncludedPgStylesheetParent = function (pgStylesheet, page) {
        var parentPgStylesheet = pgStylesheet;
        var parentPgStylesheetList = [];
        while (parentPgStylesheet != null) {
            parentPgStylesheetList.unshift(parentPgStylesheet);
            var csInfo = pinegrow.stylesheets.getStylesheetInfo(parentPgStylesheet, page);
            if (csInfo) {
                parentPgStylesheet = csInfo.parentPgStylesheet;
            }
            else {
                parentPgStylesheet = null;
            }
        }
        return parentPgStylesheetList;
    }

    this.searchForSheet = function (sheet, parentList) {
        var pgStylesheetList = pinegrow.stylesheets.findCrsaStylesheetForUrl(sheet.href);
        if (pgStylesheetList.length < 1) return;

        var found = false;
        while (!found) {
            var cssRules = sheet.cssRules;
            var sheetChanged = false;
            for (var j = 0; j < cssRules.length; j++) {
                if (cssRules[j].type == 3) {
                    if (pgStylesheetList[0].getFullUrlFor(cssRules[j].href) == parentList[0].url) {
                        sheet = cssRules[j].styleSheet;
                        sheetChanged = true;
                        parentList.splice(0, 1);
                        if (parentList.length == 0) found = true;
                        break;
                    }
                }
            }
            if (!found && !sheetChanged) break;
        }
        if (found) return sheet;
        else return null;
    }

    this.canInsertRule = function (rule) {
        if ((/*rule.type == "atrule" &&*/ rule.name == "import") ||
            (rule.type == "comment")) return false;

        return true;
    }

    var addPgStatesToSelector = function (selector) {
        var selArr = selector.split(',');
        var selArrLength = selArr.length;
        var pgCssStates = ['active', 'hover', 'focus', 'visited'];

        for (var i = 0; i < selArrLength; i++) {
            var currSel = selArr[i];
            if (currSel.indexOf('.pg-state-') > -1) {
                selArr.splice(i, 1);
                selArrLength--;
                continue;
            }

            var newSel = currSel;
            for (var j = 0; j < pgCssStates.length; j++) {
                var state = pgCssStates[j];

                var stateReg = new RegExp(':' + state, 'g');
                var matched = newSel.match(stateReg);
                if (matched && matched.length > 0) {
                    newSel = newSel.replace(stateReg, '.pg-state-' + state).trim();
                }
            }
            if (newSel != currSel) {
                // selArr.push(newSel);
                selArr.splice(i + 1, 0, newSel);
                selArrLength++;
                i++;
            }
        }

        return selArr.join(',');
    };

    var shouldAddPlaceholder = function (rule) {
        return pgcssh.shouldAddPlaceholder(rule);
    }

    var addEmptyPlaceholderForEmptyRule = function(code, addPlaceholder) {
        return pgcssh.addEmptyPlaceholderForEmptyRule(code, addPlaceholder);
    }

    //add --empty to rules in server

    var getRuleSourceWithFullStuff = function (pgStylesheet, node, with_ids) {
        if(with_ids === undefined) with_ids = true;
        if (node.type == "rule") {
            var addPlaceholder = shouldAddPlaceholder(node);
            var code = pgStylesheet.getSourceForRules(node, with_ids);
            code = pgAddProxyLinksToStyleAttribute(code);
            code = addEmptyPlaceholderForEmptyRule(code, addPlaceholder);

            //code = addPgStatesToSelector

            return code;
        }
        else if (node.type == "atrule" && (node.name == "media" || node.name == "supports")) {
            var nodes = node.nodes;
            var body = '';
            if(nodes) {
                for (var i = 0; i < nodes.length; i++) {
                    body += getRuleSourceWithFullStuff(pgStylesheet, nodes[i], with_ids);
                }
            }
            return '@' + node.name + ' ' + node.params + ' {' + body + ' }';
        }
        else {
            return pgStylesheet.getSourceForRules(node, with_ids);
        }
    }


    this.doWithStylesheet = function(pgStylesheet, func) {
        $.each(pinegrow.getAllPageViews(), function (pvi, pv) {
            var doc = pv.getDocument();
            if(doc == null) return;
            var length = doc.styleSheets.length;
            var styleSheets = doc.styleSheets;

            if (!pv.ssMap) {
                pv.ssMap = new WeakMap();
            }

            var sheet;
            for (var i = 0; i < styleSheets.length; i++) {
                sheet = styleSheets[i];

                try {
                    if (sheet.cssRules == null) continue;
                }
                catch(err) {
                    continue;
                }

                var page = pv.getPage();
                if (pgStylesheet.isImportedFor(page)) {
                    var parentPgStylesheetList = _this.getIncludedPgStylesheetParent(pgStylesheet, page);
                    if (!_this.isStylesheetUrlEqual(parentPgStylesheetList[0].url, sheet.href)) continue;


                    parentPgStylesheetList.splice(0, 1);
                    if (parentPgStylesheetList.length > 1 && !pgStylesheet.canHasImportMoreThanOneLevel()) {
                        break;
                    }
                    else {
                        var sheet = _this.searchForSheet(sheet, parentPgStylesheetList);
                        if (sheet == null) break;
                    }
                }
                else {
                    if (!_this.isStylesheetUrlEqual(pgStylesheet.url, sheet.href)) continue;
                }
                var first_apply = false;

                var map = pv.ssMap.get(sheet);
                if(!map) {
                    map = [];
                    pv.ssMap.set(sheet, map);
                    first_apply = true;
                }

                func(sheet, pv, map, first_apply);
            }
        });
    }

    var emulateCSSFeatures = function(code, pv) {
        return pv.emulateCSSFeatures(code);
    }

    this.applyToPagesSingle = function(pgStylesheet, rule) {
        var code = getRuleSourceWithFullStuff(pgStylesheet, rule);
        // var addPlaceholder = shouldAddPlaceholder(rule);
        // var code = pgStylesheet.getSourceForRules(rule, true);
        //     code = pgAddProxyLinksToStyleAttribute(code);
        //     code = addEmptyPlaceholderForEmptyRule(code, addPlaceholder);

        var pgid = '.pg-node-id-' + rule['pg-node-id'];

        var ss_list = [];

        var findAndUpdate = function(cssRules, parent, map, pv) {
            if(cssRules) {
                for(var i = 0; i < cssRules.length; i++) {

                    if(cssRules[i].selectorText && cssRules[i].selectorText.indexOf(pgid) >= 0) {
                        //found it!
                        code = emulateCSSFeatures(code, pv)
                        parent.deleteRule(i);
                        parent.insertRule(code, i);

                        if(map) {
                            map[i] = code;
                        }
                        return true;
                    }
                    if(cssRules[i].cssRules) {
                        if(findAndUpdate(cssRules[i].cssRules, cssRules[i], null, pv)) {
                            return true;
                        }
                    }
                }
            }
        }

        var do_whole_apply = false;

        if(!isApp() && pgStylesheet.loaded_from_project_content) {
            browserForAllStylesheets(pgStylesheet, function(e, pv) {
                if(e) {
                    findAndUpdate(e.sheet.cssRules, e.sheet, null, pv);
                    ss_list.push(e.sheet);
                }
            })
        } else {
            this.doWithStylesheet(pgStylesheet, function (sheet, pv, map) {
                if(!sheet._pg_applied_to_page_with_ids) {
                    do_whole_apply = true;
                } else {
                    findAndUpdate(sheet.cssRules, sheet, map, pv);
                    ss_list.push(sheet);
                }
            })
        }

        if(do_whole_apply) {
            this.applyToPages(pgStylesheet, function() {}, null);
        } else {
            pinegrow.dispatchEvent('on_css_applied_to_pages', null, pgStylesheet, ss_list, false);
        }
    }

    var addPseudoStatesToCSSSource = function(src) {
        //add pseudo states, quick & dirty
        if(src) {
            src = src.replace(/([^\}])([\.\#\:\>\-\_\(\)a-z0-9\s\t\,]*)(\{)/ig, function (m, s1, s2, s3) {
                if (s2[0] === '@') return m;
                return s1 + addPgStatesToSelector(s2.trim()) + s3;
            })
        }
        return src;
    }

    this.applyToPagesQuick = function (pgStylesheet, src) {

        if (pgStylesheet.inline) {
            var dom_source = pgStylesheet.getSourceForRules(null, true);
            dom_source = pgAddProxyLinksToStyleAttribute(dom_source);
            pgStylesheet.inlineSheetPgel.get$DOMElements().forEach(function ($el) {
                $el.get(0).innerHTML = dom_source;
            })
            return;
        }

        var is_first_apply = false;
        var ss_list = [];

        this.doWithStylesheet(pgStylesheet, function (sheet, pv, map, first_apply) {
            src = addPseudoStatesToCSSSource(src.toString());
            src = '@supports(display:block) {' + src + '}';
            while (sheet.cssRules.length) {
                sheet.deleteRule(0);
            }
            try {
                sheet.insertRule(src, 0);
                //console.log(src);
            } catch (err) {
                //if(debug) {
                //console.log(err);
                //}
            }

            is_first_apply = is_first_apply || first_apply;
            ss_list.push(sheet);
        })

        pinegrow.dispatchEvent('on_css_applied_to_pages', null, pgStylesheet, ss_list, is_first_apply);
    }

    var browserForAllStylesheets = function(pgStylesheet, func) {
        //var dom_source = pgStylesheet.getSourceForRules(null, true);

        pinegrow.getAllPageViews().forEach(function (pv) {
            var doc = pv.getDocument();
            if (doc == null) return;
            var e = doc.querySelector(`[data-pg-style-url="${pgStylesheet.url}"]`);

            func(e, pv);
        })
    }

    this.applyToPages = function (pgStylesheet, done, only_page) {

        var ignore_imports = pgStylesheet.ignore_imports_in_apply || false;
        var rules = pgStylesheet.getAllRules();

        var ss_list = [];
        var is_first_apply = false;

        var debug = false;//pgStylesheet.url.indexOf('style.css') > 0;

        if(!isApp() && pgStylesheet.loaded_from_project_content) {
            var dom_source = pgStylesheet.getSourceForRules(null, true);
            dom_source = pgAddProxyLinksToStyleAttribute(dom_source);

            browserForAllStylesheets(pgStylesheet, function(e, pv) {
                var code = emulateCSSFeatures(dom_source, pv);
                if(e) {
                    ss_list.push(e.sheet);
                    //we include a helper func that will get the proper sheet, used in pg.insight
                    e.sheet._pg_get_sheet = function() {
                        var doc = pv.getDocument();
                        if (doc === null) return null;
                        var e = doc.querySelector(`[data-pg-style-url="${pgStylesheet.url}"]`);
                        return e ? e.sheet : null;
                    }
                    e.innerHTML = code;
                    //on FF sheet becomes null for a while after assigning innerHTML?
                }
            })
            pinegrow.dispatchEvent('on_css_applied_to_pages', null, pgStylesheet, ss_list, is_first_apply);
            done && done();
            return;
        }

        var smart = false;//true;

        if(pgStylesheet.inline) {
            var dom_source = pgStylesheet.getSourceForRules(null, true);
            dom_source = pgAddProxyLinksToStyleAttribute(dom_source);

            pgStylesheet.inlineSheetPgel.get$DOMElements().forEach(function($el) {
                var pv = getPageViewOfElement($el);
                var code = dom_source;
                if(pv) {
                    code = emulateCSSFeatures(code, pv);
                }
                var style_node = $el.get(0);
                style_node.innerHTML = code;
                ss_list.push(style_node.sheet);
            })
            pinegrow.dispatchEvent('on_css_applied_to_pages', null, pgStylesheet, ss_list, is_first_apply);
            done && done();
            return;
        }

        var wrapperFunc = isApp() ? pinegrow.devTools.withDisabledCSS : function(func, done) { func(); done && done() };

        wrapperFunc(function() {

            _this.doWithStylesheet(pgStylesheet, function(sheet, pv, map, first_apply) {
                if(only_page && pv.getPage() !== only_page) return;

                if(pgStylesheet.ignore && !pgStylesheet.changed) {
                    //no need to apply ignored and not-changed stylesheets
                    //pv.reloadPage(null, null, true);
                    return;
                }

                var with_ids = true;

                if(pgStylesheet.ignore) with_ids = false;

                sheet._pg_applied_to_page_with_ids = true;

                ss_list.push(sheet);

                debug && console.log(`Applying styling of ${pgStylesheet.url} to pv ${pv.deviceWidth}`);

                var p = new CrsaProfile();
                var rule_i = 0;
                var changed = 0;

                var use_supports = false;

                is_first_apply = is_first_apply || first_apply;

                if(false && debug) {
                    for (var i = 0; i < sheet.cssRules.length; i++) {
                        console.log(i, sheet.cssRules[i])
                    }
                }

                if(use_supports) {
                    var src = '';
                    for(var j = 0; j < rules.length; j++) {
                        src += getRuleSourceWithFullStuff(pgStylesheet, rules[j], with_ids);
                    }
                    src = '@supports(display:block) {' + src + '}';
                    while(sheet.cssRules.length) {
                        sheet.deleteRule(0);
                    }
                    try {
                        sheet.insertRule(src, 0);
                    } catch(err) {
                        //if(debug) {
                        console.log(err);
                        //}
                    }
                    return;
                }


                 //Disable because it doesn't work well with @imports
                var code = '';

                for(var j = 0; j < rules.length; j++) {
                    if(rules[j].type === 'atrule' && rules[j].name === 'import') {
                        //this will cause the old way to apply the page
                        if(!ignore_imports) {
                            code = '@import';
                            break;
                        }
                    } else {
                        if(rules[j].type !== 'comment') {
                            var ruleSource = getRuleSourceWithFullStuff(pgStylesheet, rules[j], with_ids);
                            code += emulateCSSFeatures(ruleSource, pv);
                        }
                    }
                }

                if(isApp() && code.indexOf('@import') < 0) {
                    //we can do a quick set
                    if (pinegrow.devTools.setStylesheetContent(pgStylesheet.url, code, pv)) {
                        return;
                    }
                }

                //test delete all
                /*
                console.log('BEFORE sheet.cssRules.length', sheet.cssRules.length);
                for (var i = 0; i < sheet.cssRules.length; i++) {
                    console.log(sheet.cssRules[i])
                    if(i === 0) break;
                }

                while(sheet.cssRules.length) {
                    sheet.deleteRule(0);
                }

                 */

                for(var j = 0; j < rules.length; j++) {
                    var ruleSource = getRuleSourceWithFullStuff(pgStylesheet, rules[j], with_ids);

                    ruleSource = emulateCSSFeatures(ruleSource, pv);
                    // var addPlaceholder = shouldAddPlaceholder(rules[j]);
                    // var ruleSource = pgStylesheet.getSourceForRules(rules[j], true);
                    //     ruleSource = pgAddProxyLinksToStyleAttribute(ruleSource);
                    //     ruleSource = addEmptyPlaceholderForEmptyRule(ruleSource, addPlaceholder);

                    if (!_this.canInsertRule(rules[j])) continue;

                    while(sheet.cssRules.length > rule_i && sheet.cssRules[rule_i].type == 3) {
                        rule_i++;
                    }

                    if(false && debug) {
                        console.log(rule_i, ruleSource);
                    }

                    try {
                        var rule;
                        if (sheet.cssRules.length > rule_i) {
                            rule = sheet.cssRules[rule_i];

                            var oldSource = smart ? map[rule_i] : null;

                            if(!oldSource || oldSource != ruleSource) {
                                if(false && debug) {
                                    console.log('Chnaged rule ' + ruleSource);
                                }
                                if (_this.canInsertRule(rules[j])) {
                                    rule = sheet.cssRules[sheet.insertRule(ruleSource, rule_i)];
                                    sheet.removeRule(rule_i + 1);
                                    //map.delete(rule);
                                    if(smart) {
                                        map[rule_i] = ruleSource;
                                    }

                                    //rule._pgRuleSource = ruleSource;
                                    changed++;
                                    rule_i++;

                                } else {
                                    changed = changed;
                                }
                            } else {
                                rule_i++;
                            }
                        } else {
                            //if (_this.canInsertRule(rules[j])) {
                            rule_i = sheet.insertRule(ruleSource, sheet.cssRules.length);
                            rule = sheet.cssRules[rule_i];
                            if(smart) {
                                map[rule_i] = ruleSource;
                            }
                            rule_i++;
                            changed++;
                            // }
                        }
                    } catch(err) {
                        if(debug) {
                           // console.log(err);
                        }
                    }
                }

                //if(rule_i == 0) rule_i = 1;
                //rule_i = rules.length;
                while(rule_i < sheet.cssRules.length) {
                    var r = sheet.cssRules[rule_i];
                    if(r.type !== 3) {
                        map.splice(rule_i, 1);
                        sheet.deleteRule(rule_i);
                    } else {
                        rule_i++;
                    }
                    changed++;
                }

                if(false) {
                    console.log('sheet.cssRules.length', sheet.cssRules.length);
                    for (var i = 0; i < sheet.cssRules.length; i++) {
                        console.log(sheet.cssRules[i])
                        if(i === 0) break;
                    }
                }
                p.show('CSS APPLY - SMART - Changed ' + changed);
                /*
                 var index = 0
                 while (sheet.cssRules && sheet.cssRules.length > index) {
                 if (sheet.cssRules[index].type != 3) {
                 if(sheet.removeRule) {
                 sheet.removeRule(index);
                 } else if(sheet.deleteRule) {
                 sheet.deleteRule(index);
                 }
                 }
                 else {
                 index++;
                 }
                 }
                 p.show('CSS APPLY - Delete rules - ' + sheet.href);

                 for(var j = 0; j < rules.length; j++) {
                 var ruleSource = pgStylesheet.getSourceForRules(rules[j]);
                 try {
                 if (_this.canInsertRule(rules[j])) {
                 var rule = sheet.cssRules[sheet.insertRule(ruleSource, sheet.cssRules.length)];
                 rule._pgRuleSource = ruleSource;
                 }
                 } catch(err) {
                 // console.log(err);
                 }
                 }
                 p.show('CSS APPLY - Add rules - ' + sheet.href);
                 */

            });
            pinegrow.dispatchEvent('on_css_applied_to_pages', null, pgStylesheet, ss_list, is_first_apply);
        }, done)
    }

    // Dons't work
    this.findImportedRuleFor = function (sheet, pgStylesheet) {
        var rules = sheet.cssRules;
        for (var i = 0; i < rules.length; i++) {
            if (rules[i].href == pgStylesheet.relativeUrl) {
                return rules[i];
            }
        }
    }

    // Doesn't work
    this.findImportedAttachmentFor = function (sheet, pgStylesheet) {
        return this.findImportedRuleFor(sheet, pgStylesheet).styleSheet;
    }

    this.findAttachmentFor = function(pgStylesheet, page) {
        var iframe = page.$iframe.get(0);
        if (pgStylesheet.inline) {
            return pgStylesheet.getInlineSheet(iframe);
        }

        var imported = pgStylesheet.isImportedFor(page);
        var pgParentStylesheet = pgStylesheet;
        var csInfo = null;

        csInfo = pinegrow.stylesheets.getStylesheetInfo(pgParentStylesheet, page);
        if (csInfo && csInfo.parentPgStylesheet) pgParentStylesheet = csInfo.parentPgStylesheet;

        csInfo = pinegrow.stylesheets.getStylesheetInfo(pgParentStylesheet, page);
        if (csInfo && csInfo.compiledStylesheet) pgParentStylesheet = csInfo.compiledStylesheet;

        var doc = getIframeDocument(iframe);
        for (var i = 0; i < doc.styleSheets.length; i++) {
            var sheet = doc.styleSheets[i];
            if (this.isStylesheetUrlEqual(pgParentStylesheet.url, sheet.href)) {
                if (!imported) return sheet;
                // else return this.findImportedAttachmentFor(sheet, pgStylesheet); // Doesn't work
            }
        }
        return null;
    }

    this.isEnabledForPage = function(pgStylesheet, page) {
        var ss = this.findAttachmentFor(pgStylesheet, page);
        return ss ? ss.disabled == false : false;
    }

    this.setEnabledForPage = function(pgStylesheet, page, enabled) {
        var ss = this.findAttachmentFor(pgStylesheet, page);
        if(ss) ss.disabled = !enabled;
    }
}



//CSS Change observer

$('body').one('pinegrow-ready', function() {

    var refresh_timer = null;

    var refresh = function() {
        if(refresh_timer) {
            clearTimeout(refresh_timer);
        }
        refresh_timer = setTimeout(function() {
            pinegrow.getAllPages().forEach(function(page) {
                page.refreshPageChangedStatus();
            })
            refresh_timer = null;
        }, 500)
    }

    var onCSSChanged = function(cs, data) {
        refresh();

        //push change to external editors
        var editor = (data && data.editor) ? data.editor : null;
        if(cs.url && (!data || !data.editor || data.editor != 'external_editor')) {
            codeEditors.cssCodeChanged(cs.url, function() {return cs.filterSource(cs.getSource())}, editor, data, cs);
        }
    }

    pinegrow.addEventHandler('on_css_source_changed', function (p, info) {
        if(info && info.list) {
            info.list.forEach(function(cs) {
                onCSSChanged(cs, info);
            })
        }
    });

    var setCsCode = function (stylesheet) {
        codeEditors.setCode(stylesheet.url, stylesheet.source);
    }

    pinegrow.addEventHandler('on_css_loaded_from_external', function(pagenull, cs) {
        setCsCode(cs);
    });

    pinegrow.addEventHandler('on_css_loaded', function(pagenull, cs) {
        var forAll = function (importedFiles) {
            for (var i = 0; i < importedFiles.length; i++) {
                var importedFile = importedFiles[i];
                setCsCode(importedFile);

                if (importedFile.importedFiles.length > 0) {
                    forAll(importedFile.importedFiles);
                }
            }
        }

        setCsCode(cs);
        forAll(cs.importedFiles);

        refresh();
    });

    pinegrow.addEventHandler('on_imported_css_loaded', function (pagenull, cs) {
        setCsCode(cs);

        refresh();
    });

    pinegrow.addEventHandler('on_css_node_added', function(pagenull, data) {
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_node_deleted', function(pagenull, data) {
        onCSSChanged(data.stylesheet, data);
    })

    pinegrow.addEventHandler('on_css_node_updated', function(pagenull, data) {
        onCSSChanged(data.stylesheet, data);
    })

    $('body').on('crsa-rule-changed-in-editor', function (data) {
        refresh();
    })

})