window.PgCompilableStylesheet = function () {
    var pgStylesheet = new window.PgStylesheet();
    pgStylesheet.compiledCssCode = null;
    pgStylesheet.compilationError = null;
    pgStylesheet.includeSourceMap = false;
    pgStylesheet.sourceFormated = false;
    pgStylesheet.ignoredImportedFiles = [];

    pgStylesheet.dontAddIds = false;

    var titleRegx = new RegExp(pgStylesheet.GROUP_TITLE_MARGIN + '\\s(.*)\\s' + pgStylesheet.GROUP_TITLE_MARGIN);
    var pgPrevStylesheetMap = {};

    var id_count = 0;

    var getTreeNodeId = function() {
        return ++id_count;
    }

    var getRelativeHrefPair = function (importValue, _this, node) {
        var fileRelativeUrl = _this.getRelativeUrlFromImported(importValue, node, true);
        if (_this.skipImportFile(fileRelativeUrl)) {
            return {
                relativeUrl: fileRelativeUrl,
                href: fileRelativeUrl
            }
        }
        else {
            if(_this.onMapImportRelativeUrl) {
                fileRelativeUrl = _this.onMapImportRelativeUrl(fileRelativeUrl);
            }
            var href = crsaCombineUrlWith(_this.url, fileRelativeUrl);
            var filePath = crsaMakeFileFromUrl(href);

            if (isApp() && !crsaIsFileExist(filePath)) {
                fileRelativeUrl = _this.getRelativeUrlFromImported(importValue, node);
                href = crsaCombineUrlWith(_this.url, fileRelativeUrl);
            }

            return {
                relativeUrl: fileRelativeUrl,
                href: href
            }
        }
    }

    pgStylesheet.isCompilableStylesheet = function () {
        return true;
    }

    pgStylesheet.canEditRuleSelector = function () {
        return true;
    }

    pgStylesheet.canAddChangeMediaQuery = function () {
        return true;
    }

    pgStylesheet.resetTreeNodeId = function() {
        id_count = 0;
    }

    pgStylesheet.lastNodeId = function () {
        return id_count;
    }

    pgStylesheet.smartGetAllUsedNodes = function (node, skipRecursive, nodeValue) {
        nodeValue = nodeValue || null;

        var nodesName = [];
        var nodes = [];

        var pgcssh = pinegrow.pgPostCSSHelper;
        var varReg = new RegExp('\\' + this.getDeclCharachter() + '[a-z0-9_\\-]*', 'ig');

        var findNodeBefore = function (node, value, searchAll, importedStylesheet, calledImported) {
            var children = [],
                length = -1,
                currPgStylesheet = null;

            if (searchAll) {
                var rootNode = importedStylesheet.getRootNode();
                currPgStylesheet = importedStylesheet;
                children = rootNode.nodes;
                length = rootNode.nodes.length;
            }
            else {
                var topLvlNode = pgcssh.getTopLevelNodeFor(node);
                var rootNode = topLvlNode.parent;
                var children = rootNode.nodes;
                var topLvlNodeIndex = children.indexOf(topLvlNode);

                currPgStylesheet = pgcssh.getStylesheetForNode(rootNode);
                length = topLvlNodeIndex || children.length;
            }

            for (var i = length - 1; i > -1; i--) {
                var currNode = children[i];
                if (pgcssh.isVariable(currNode)) {
                    var prop = 'prop';
                    var prefix = '';

                    if (currNode.type == 'atrule') {
                        prefix = '@';
                        prop = 'name';
                    }
                    if (prefix + currNode[prop] == value) return currNode;
                }
                else if (pgcssh.getNodeType(currNode) == 'import') {
                    var newImportedStylesheet = currPgStylesheet.getImportedStylesheetFromRelativeUrl(currNode.params);
                    if (currPgStylesheet.getUniqueName() != newImportedStylesheet.getUniqueName()) {
                        if (newImportedStylesheet.name != "custom.scss") { // TODO get custom file name
                            var currNode = findNodeBefore(null, value, true, newImportedStylesheet, true);
                            if (currNode) return currNode;
                        }
                    }
                }
            }

            var page = pinegrow.getSelectedPage();
            if (!calledImported && currPgStylesheet.isImportedFor(page)) {
                var csInfo = pinegrow.stylesheets.getStylesheetInfo(currPgStylesheet, page);
                if (csInfo && csInfo.parentPgStylesheet) {
                    var parentPgStylesheet = csInfo.parentPgStylesheet;
                    if (parentPgStylesheet.name != "_custom.scss") { // TODO get custom file name
                        var currNode = findNodeBefore(null, value, true, parentPgStylesheet);
                        if (currNode) return currNode;
                    }
                }
            }

            return null;
        }

        var resolveUsed = function (value, mainNode) {
            var m = value.match(varReg);

            if (m) {
                m.forEach(function (used_var) {
                    var newNode = findNodeBefore(mainNode, used_var);
                    if (newNode && !skipRecursive) {
                        resolveUsed(newNode.value, newNode);
                    }

                    if (newNode && nodesName.indexOf(used_var) == -1) {
                        nodesName.push(used_var);
                        nodes.push(newNode);
                    }
                })
            }
        }

        var isVariable = pgcssh.isVariable(node);
        if (isVariable) nodesName.push(pgcssh.getVariableProp(node));

        if (!nodeValue) {
            if (pgcssh.getNodeType(node) == 'decl') {
                nodeValue = node.value;
            }
            else if (pgcssh.getNodeType(node) == 'atrule') {
                nodeValue = node.params;
            }
        }

        resolveUsed(nodeValue, node);

        if (isVariable) nodes.push(node);

        return {
            names: nodesName,
            nodes: nodes
        }
    }

    var delayed = new PgDelayedTasks();
    var updateCompiledDelay = 500;

    pgStylesheet.updateCompiledCode = function (css, done, fastDelayed) {
        //wait for changes to finish before updating
        // delayed.run('updateCompiled', function() {
        this.compiledStylesheets.forEach(function (compiledStylesheet) {
            var p = new CrsaProfile(true, 'UPDATE COMPILED ');

            compiledStylesheet.setSource(css.toString(), function () {
                p.show('DONE');

                compiledStylesheet.getSourceToSave(function (error, source) {
                    if (error) source = '';
                    pinegrow.codeEditors.cssCodeChanged(compiledStylesheet.url, function () {
                        return source;
                    }, null, null, compiledStylesheet);
                });

                if (done) done();

                p.show('AFTER DONE');
            }, false, false, fastDelayed);
        });
        // }, updateCompiledDelay)


    }

    pgStylesheet.addIdsToNodes = function(nodes) {
        var vars = this.varGroups;
        this.processNodes(nodes);
        this.varGroups = vars;
    }

    pgStylesheet.processNodes = function (allNodes) {
        var _this = this;
        var currentGroup = [];

        var addNewGroup = function (newGroup) {
            _this.varGroups.push(newGroup);
        }

        var addToAllNodes = function (nodes, nested) {
            for (var i=0; i < nodes.length; i++) {
                var justFound = false;
                // Detect Formatting
                if (pinegrow.autoFormatCss == 'auto-detect') {
                    pgcssf.detectFormatting(nodes[i], nested, i, _this.formattingOptions);
                }

                if (nodes[i].type == "rule" && !nodes[i].ruleWithoutBody && !nodes[i].params) {
                    if (!nodes[i].nodes && !nodes[i].empty) nodes[i].nodes = [];
                    if (!_this.canAddRuleIds(nodes[i], _this.atruleWithNoChild)) continue;

                    /*
                    if (nodes[i].nodes.length == 0) {
                        nodes[i].nodes.push({
                            type: 'decl',
                            prop: '--empty-placeholder',
                            value: '0px',
                            parent: nodes[i],
                            raws: {
                                before: '\n',
                                between: ': '
                            }
                        });
                        _this.emptyNodes.push(nodes[i]);
                    } */

                    var id = getTreeNodeId() + '';
                    nodes[i]['pg-node-id'] = id;
                    nodes[i]['pg-main-file'] = _this.getUniqueName();

                    if (_this.fileType == 'less') {
                        var lastNode = nodes[i].last;
                        if (lastNode && lastNode.empty && lastNode.type == 'rule') {
                            lastNode.raws.semicolon = true;
                        }
                    }

                    if(!_this.dontAddIds) {
                        if (_this.isImported()) {
                            nodes[i]['pg-file'] = _this.name;
                            nodes[i]['pg-file-name'] = _this.getUniqueName();
                        }
                    }
                }
                else if (nodes[i].type == "rule" && nodes[i].params) {
                    if (nodes[i].raws.hasOwnProperty('semicolon')) {
                        nodes[i].raws.semicolon = true;
                    }
                    else {
                        nodes[i].after = ';' + nodes[i].after;
                    }
                }
                else if (nodes[i].type == "import") {
                    nodes[i].raws.semicolon = true;
                }
                else if (nodes[i].type == "decl") {
                    var parent = nodes[i].parent;
                    if (parent.type == 'root' && nodes[i].prop[0] == _this.getDeclCharachter()) {
                        currentGroup.push(nodes[i]);

                        // if (currentGroup == null) {
                        //     if (varsGroupTitle == null) {
                        //         while (groupNumbers.indexOf(groupIndex) > -1) groupIndex++;

                        //         var name = "Unnamed Group " + groupIndex;
                        //         var text = _this.GROUP_TITLE_MARGIN + ' ' + name + ' ' + _this.GROUP_TITLE_MARGIN;
                        //         var node = pinegrow.pgPostCSSHelper.createNode('comment', text, null);
                        //         // pinegrow.pgPostCSSHelper.insertNodeBefore(node, nodes[i]);
                        //         groupNumbers.push(groupIndex);
                        //         groupIndex++;
                        //         // i++;

                        //         varsGroupTitle = {
                        //             name: name,
                        //             node: node
                        //         };
                        //     }

                        //     currentGroup = {
                        //         name: varsGroupTitle.name,
                        //         titleNode: varsGroupTitle.node,
                        //         vars: []
                        //     };

                        //     addNewGroup(currentGroup);
                        // }
                        // currentGroup.vars.push(nodes[i]);
                    }

                    // if (i > 0) {
                    //     if (nodes[i - 1].type == "rule" && nodes[i - 1].params) {
                    //         nodes[i].before = '\n\n';
                    //     }
                    // }
                }

                // else if (nodes[i].type == "comment") {
                //     var matched = nodes[i].text.match(titleRegx);
                //     if (matched && matched.length > 1) {
                //         var nodeTitle = matched[1];

                //         var groupUnnamedMatched = matched[1].match(/Unnamed Group\s(.*)/);
                //         if (groupUnnamedMatched && groupUnnamedMatched.length > 1) {
                //             var nodeNameId = parseInt(groupUnnamedMatched[1]);
                //             if (groupNumbers.indexOf(nodeNameId) > -1) {
                //                 while (groupNumbers.indexOf(groupIndex) > -1) groupIndex++;
                //                 nodeTitle = 'Unnamed Group ' + groupIndex;
                //                 nodes[i].text = _this.GROUP_TITLE_MARGIN + ' ' + nodeTitle + ' ' + _this.GROUP_TITLE_MARGIN;
                //                 groupNumbers.push(groupIndex);
                //                 groupIndex++;
                //             }
                //             else {
                //                 groupNumbers.push(nodeNameId);
                //             }
                //         }

                //         varsGroupTitle = {
                //             name: nodeTitle,
                //             node: nodes[i]
                //         };
                //         currentGroup = {
                //             name: varsGroupTitle.name,
                //             titleNode: varsGroupTitle.node,
                //             vars: []
                //         };
                //         addNewGroup(currentGroup);
                //         justFound = true;
                //     }
                // }

                // if (!justFound && currentGroup != null && (nodes[i].type != "decl" || nodes[i].prop[0] != _this.getDeclCharachter())) {
                //     varsGroupTitle = null;
                //     currentGroup = null;
                // }

                if (currentGroup.length > 0  && ((nodes[i].type != 'decl' || nodes[i].prop[0] != _this.getDeclCharachter()) || i + 1 == nodes.length)) {
                    addNewGroup(currentGroup);
                    currentGroup = [];
                }

                if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                    addToAllNodes(nodes[i].nodes, true);
                }
            }
        }

        this.formattingOptions = {};

        this.emptyNodes = [];
        this.varGroups = [];
        addToAllNodes(allNodes, false);
        return allNodes;
    }

    pgStylesheet.addSemicolonToSource = function (source) {
        source = source.trim();

        var cssLength = source.length;
        if (cssLength > 0 && source[cssLength - 1] != ';' && source[cssLength - 1] != '}') {
            var addSemicolon = true;

            var matched = source.match(/(?!\n).*$/);
            if (matched.length > 0) {
                var lastLine = matched[0];
                var lineLength = lastLine.length;
                if ((lastLine[0] == '/' && lastLine[1] == '/') ||
                    (lastLine[lineLength - 2] == '*' && lastLine[lineLength - 1] == '/')) {
                    addSemicolon = false;
                }
            }

            if (addSemicolon) source += ';';
        }

        return source;
    }

    pgStylesheet.addCompiledStylesheet = function (cs) {
        this.compiledStylesheets.push(cs);
        cs.compiledFile = true;
    }

    pgStylesheet.importer = function (request, prev, done) {
        if(request) {
            var importedStylesheet = null,
                buff = "";

            var reqUrl = request;
            if (this.fileType == 'scss') {
                reqUrl = reqUrl.replace(/\-pg[0-9]+/, '');
            }

            if (prev == "stdin") {
                importedStylesheet = this.getImportedStylesheetFromRelativeUrl(reqUrl);
            }
            else {
                importedStylesheet = pgPrevStylesheetMap[prev].getImportedStylesheetFromRelativeUrl(reqUrl);
            }

            if (importedStylesheet) {
                //buff = importedStylesheet.getSourceForRules(null, true);
                buff = importedStylesheet.getSourceForRules(null, this.dontAddIds ? false : true);
                pgPrevStylesheetMap[request] = importedStylesheet;
            }

            if (done) done({ contents: buff });
            return { contents: buff };
        } else {
            if (done) done();
            return { contents: '' };
        }
    }

    pgStylesheet.renderingFinished = function (error, result, done) {
        if (error) {
            if (done) done(error);
            return;
        }
        if (done) done(null, result);
    }

    pgStylesheet.setSourceForRules = function(node, done, skipUpdatingRule, fastDelayed) {
        this.setNeedFormatting(true);
        var newSource = this.getSourceForRules(node);
        this.setSource(newSource, done, skipUpdatingRule, false, fastDelayed);
    }

    pgStylesheet.removeAllImportFiles = function () {
        var importedFiles = this.importedFiles;
        for (var i = 0; i < importedFiles.length; i++) {
            importedFiles[i].remove();
            importedFiles[i] = null;
        }
        this.importedFiles = [];
    }

    pgStylesheet.reload = function(done, saveCss) {
        pinegrow.codeEditors.ignoreCodeForUrl(this.url);

        var count = 0;
        var allDone = function () {
            count--;
            if (count <= 0) {
                if (done) done();
            }
        }

        var doneReloading = function (parentStylesheet) {
            if (saveCss) {
                parentStylesheet.saveCompiledFile(allDone);
            }
            else {
                allDone();
            }
        }

        var _this = this;
        if (this.isImported()) {
            this.loadFromUrl(this.url, function () {
                _this.setChanged(false);
                var mainParentStylesheets = _this.getMainParnetStylesheets();
                mainParentStylesheets.forEach(function (mainParentStylesheet) {
                    count++;
                    mainParentStylesheet.regenerateAndSetCssSource(function () {
                        doneReloading(mainParentStylesheet);
                    });
                });
            });
        }
        else if (this.compiledStylesheets.length > 0) {
            var compiledStylesheets = this.compiledStylesheets;

            this.removeAllImportFiles();
            this.compiledStylesheets = [];

            compiledStylesheets.forEach(function (compiledStylesheet, i) {
                if (i == 0) {
                    _this.load(_this.url, function () {
                        _this.setChanged(false);
                        doneReloading(_this);
                    }, true, false, compiledStylesheet);
                }
                else {
                    _this.addCompiledStylesheet(compiledStylesheet);
                }
            })
        }
    }

    pgStylesheet.getImportedStylesheetFromRelativeUrl = function (reqUrl, importedArray) {
        reqUrl = crsaGetUrlAndParamsFromImport(reqUrl)[0];
        var importedFiles = importedArray || this.importedFiles;
        var relativeUrls = [];
        relativeUrls.push(
            this.getFullUrlFor(reqUrl),
            this.getFullUrlFor(reqUrl, true),
        );

        for (var i = 0; i < importedFiles.length; i++) {
            if (relativeUrls.indexOf(importedFiles[i].url) > -1) {
                return importedFiles[i];
            }
        }
    }

    pgStylesheet.setSource = function(source, done, skipUpdatingRule, skipRegenerateCss, fastDelayed) {
        var _this = this;
        this.forgetSourceErrors();

        this.sourceBeingCurrentlyParsed = source;

        this.parseSource(source, function (parsingError, parsedSource, rules) {
            _this.sourceBeingCurrentlyParsed = null;

            if (parsingError) {
                _this.codeError = parsingError;
                if (getTrimed(_this.parsedSource) != getTrimed(source)) {
                    _this.source = _this.parsedSource = source;
                    if (_this.loaded) _this.setChanged(true);
                }
                if (done) done();
            }
            else {
                var compileNow = function () {
                    if(skipRegenerateCss) {
                        done && done();
                        return;
                    }
                    _this.regenerateCssFromCompilableSource(function (compilationError, renderResult) {
                        if (compilationError) {
                            _this.compilationError = compilationError;
                            if (done) done();
                        } else {
                            _this.compilationError = null;
                            // pinegrow.stylesheets.applyToPagesQuick(_this.compiledStylesheet, renderResult.css);
                            _this.updateCompiledCode(renderResult.css, done, fastDelayed);
                            // TODO find what to do with map (use renderResult.map)
                        }

                        _this.ignore = tmpIgnore;
                    })
                }

                _this.codeError = null;
                if (getTrimed(_this.parsedSource) != getTrimed(parsedSource)) {
                    _this.source = _this.parsedSource = parsedSource;
                    if (_this.loaded) _this.setChanged(true);
                }
                if (!skipUpdatingRule) {
                    _this.rootNode = rules;
                    _this.rootNode.pgStylesheet = _this;
                }

                if (_this.source == '') {
                    _this.removeAllImportFiles();
                    if (_this.isImported()) {
                        var count = 0;
                        var allDone = function (params) {
                            count--;
                            if (count <= 0) {
                                if (done) done(params);
                            }
                        }


                        if (skipRegenerateCss) {
                            if (done) done();
                        }
                        else {
                            _this.forEachParentCS(function (cs) {
                                count++;
                                cs.regenerateAndSetCssSource(allDone);
                            }, true);
                        }
                    }
                    else {
                        _this.updateCompiledCode('', done);
                    }
                    return;
                }

                var tmpIgnore = _this.ignore;
                _this.ignore = false;
                _this.loadAllImportedFiles(function () {
                    if (_this.getError()) {
                        _this.ignore = tmpIgnore;
                        if (done) done();
                    }
                    else {
                        if (_this.isImported()) {
                            var count = 0;
                            var allDone = function (params) {
                                count--;
                                if (count <= 0) {
                                    if (done) done(params);
                                }
                            }

                            if (skipRegenerateCss) {
                                if (done) done();
                            }
                            else {
                                if (_this.compiledStylesheets.length > 0) {
                                    compileNow();
                                }

                                _this.forEachParentCS(function (cs) {
                                    count++;
                                    cs.regenerateAndSetCssSource(allDone);
                                }, true);
                            }

                            _this.ignore = tmpIgnore;
                        }
                        else {
                            compileNow();
                        }
                    }
                });
            }
        }, true);
    }

    pgStylesheet.regenerateAndSetCssSource = function(done, process_rules, rootNode) {
        rootNode = rootNode || this.getRootNode();

        this.source = this.getSourceForRules(rootNode);
        this.setSource(this.source, done);
    }

    pgStylesheet.getSource = function() {
        //var p = new CrsaProfile();
        if(this.codeError || !this.rootNode) {
            //rules are probably invalid
            return this.source;
        }
        if (!this.source || !this.sourceFormated) {
            this.source = this.getSourceForRules();
            if (this.changedOnce) this.sourceFormated = true;
        }
        //p.show('Get source ' + this.name);
        return this.source;
    }

    pgStylesheet.getStylesheetSource = function (done) {
        if (!this.source) this.source = this.getSourceForRules();
        if (done) done(this.source);
    }

    pgStylesheet.getCssSource = function(done, on_error) {
        if (this.isImported()) {
            done('');
            return;
        }
        if (this.compiledCssCode != null) {
            done(this.compiledCssCode);
        }
        else {
            this.regenerateCssFromCompilableSource(function (error, result) {
                if (!error) {
                    done(result.css.toString());
                } else {
                    on_error(error)
                }
            })
        }
    }

    pgStylesheet.getImportUrlIfExist = function (node) {
        if (node.type == "atrule" && node.name == "import") {
            return node.params;
        }

        return null;
    }

    pgStylesheet.updateImportedListOrder = function () {
        return; // Not used
        var importedList = this.importedFiles;
        var newImportedList = [];
        var nodes = this.rootNode.nodes;
        for (var i = 0; i < nodes.length; i++) {
            var node = nodes[i];
            var importValue = this.getImportUrlIfExist(node);
            if (importValue) {
                var relativeUrl = this.getRelativeUrlFromImported(importValue);

                for (var j = 0; j < importedList.length; j++) {
                    if (importedList[j].relativeUrl == relativeUrl) {
                        newImportedList.push(importedList[j]);
                        importedList.splice(j, 1);
                        break;
                    }
                }
            }
        }

        this.importedFiles = newImportedList;
    }

    pgStylesheet.ignoreImportedFile = function (node, ignore) {
        var importedFile = this.getImportedStylesheetFromRelativeUrl(node.params);
        if (importedFile) {
            this.ignoredImportedFiles.push(importedFile);
            this.setRecursiveImportsError(null);
        }
    }

    pgStylesheet.removeImportedFile = function (node) {
        var importValue = this.getImportUrlIfExist(node);
        if (importValue) {
            var fileRelativeUrl = crsaGetUrlAndParamsFromImport(importValue)[0];
            var importedStylesheet = this.getImportedStylesheetFromRelativeUrl(fileRelativeUrl);
            var index = this.importedFiles.indexOf(importedStylesheet)
            if (index > -1) {
                this.importedFiles.splice(index, 1);
                found = true;
            }
            else {
                this.setRecursiveImportsError(null);
            }
        }
    }

    pgStylesheet.addImportedFile = function (node, done) {
        var importValue = this.getImportUrlIfExist(node);
        if (this.skipImportFile(importValue)) {
            if (done) done();
            return;
        }

        var relativeHrefPair = getRelativeHrefPair(importValue, this, node);

        var fileRelativeUrl = relativeHrefPair.relativeUrl;
        var href = relativeHrefPair.href;

        var recursiveImports = this.checkRecursiveImports(href, true);
        if (!recursiveImports) {
            var cs = pinegrow.stylesheets.getStylesheetByKey(href);

            if (cs) {
                if (done) done();
            }
            else {
                cs = this.getNewStylesheet();
                cs.nakedUrl = importValue.replace(/[\"\']/g, '');
                cs.index = pinegrow.stylesheets.getLastIndex(true);
                cs.dontAddIds = this.dontAddIds;

                // this.importedFiles.push(cs);
                var newFilePath = crsaMakeFileFromUrl(href);


                var loadFile = function () {
                    pinegrow.showQuickMessage('Loading ' + fileRelativeUrl + ' file...');
                    // setTimeout(function () {
                        cs.loadFromUrl(href, function() {
                            console.log('done ' + cs.url);
                            pinegrow.dispatchEvent('on_css_loaded', null, cs);
                            if(done) done();
                        });
                    // }, 1000);
                }

                if (!crsaIsFileExist(newFilePath)) {
                    crsaCreateDirs(crsaGetDir(newFilePath));
                    fs.writeFileSync(newFilePath, "")
                    console.log("Imported file created!");
                    loadFile();
                }
                else {
                    loadFile();
                }
            }

            cs.addParentStylesheet(this);
            if (this.importedFiles.indexOf(cs) < 0) {
                pinegrow.stylesheets.stylsheetUsed({
                    url: href,
                    pgStylesheet: cs
                }, 'child', this);
            }
        }
        else {
            if (done) done(recursiveImports);
        }
    }

    // Not used
    pgStylesheet.getImportedFileFromIgnoreList = function (fileRelativeUrl, remove) {
        var ignoreImportedStylesheet = this.getImportedStylesheetFromRelativeUrl(fileRelativeUrl, this.ignoredImportedFiles);

        if (ignoreImportedStylesheet) {
            var index = this.ignoredImportedFiles.indexOf(ignoreImportedStylesheet);
            if (index > -1) {
                if (remove) this.ignoredImportedFiles.splice(index, 1);
            }
            return ignoredStylesheet;
        }

        return null;
    }

    pgStylesheet.getImportedFiles = function () {
        if (this.ignore) return [];

        var _this = this;
        var list = [];
        var doTask = function (rules) {
            if (rules) {
                for(var i = 0; i < rules.length; i++) {
                    var rule = rules[i];
                    var importValue = _this.getImportUrlIfExist(rule);

                    if (importValue) {
                        // var importValue = rules[i].params;

                        var relativeHrefPair = getRelativeHrefPair(importValue, _this, rule);
                        var fileRelativeUrl = relativeHrefPair.relativeUrl;
                        if (_this.skipImportFile(fileRelativeUrl)) continue;

                        // var fileRelativeUrl = _this.getRelativeUrlFromImported(importValue);
                        // if (crsaIsAbsoluteUrl(fileRelativeUrl)) continue;

                        // var cs = _this.getImportedFileFromIgnoreList(fileRelativeUrl, !skipResetingImportedFiles);
                        var href = relativeHrefPair.href;
                        var recursiveImports = _this.checkRecursiveImports(href, true);
                        if (recursiveImports) return;

                        var cs = pinegrow.stylesheets.getStylesheetByKey(href);

                        // var href = crsaCombineUrlWith(_this.url, fileRelativeUrl);
                        if (!cs) {
                            cs = _this.getNewStylesheet();
                            cs.nakedUrl = importValue.replace(/[\"\']/g, '');
                            cs.index = pinegrow.stylesheets.getLastIndex(true);
                            cs.dontAddIds = _this.dontAddIds;
                        }

                        cs.addParentStylesheet(_this);
                        list.push({ cs: cs, href: href });

                        if (_this.importedFiles.indexOf(cs) < 0) {
                            pinegrow.stylesheets.stylsheetUsed({
                                url: href,
                                pgStylesheet: cs
                            }, 'child', _this);
                        }
                    }

                    if (rule.nodes && rule.nodes.length > 0) {
                        doTask(rule.nodes);
                    }
                }
            }
        }

        doTask(_this.rootNode.nodes);
        return list;
    }

    pgStylesheet.load = function (url, done, force, skipFileExistence, mainPgStylesheet) {
        this.resetError();

        var _this = this;
        this.addCompiledStylesheet(mainPgStylesheet);

        // this.outFile = this.compiledStylesheet.url;
        var csurl = this.setNakedUrl(url);

        var doneasync = function(a, b, c) {
            if(done) {
                setTimeout(function() {
                    done && done(_this, a, b, c);
                    pinegrow.dispatchEvent('on_css_loaded', null, _this);
                }, 1);
            }
        }

        if(this.ignore && !force) {
            doneasync();
            return;
        }

        var fs = require('fs');
        if(isApp() && crsaIsFileUrl(csurl)) {
            try {
                var fileName = crsaMakeFileFromUrl(csurl, true);

                var str = pinegrow.codeEditors.getCodeForUrl(csurl);

                if(str === null) {
                    str = fs.readFileSync(fileName, {encoding: "utf8"});
                }
                console.log('loaded ' + csurl);

                this.setSource(str, function() {
                    if(_this.getError()) {
                        console.log('invalid ' + _this.fileType + ' ' + csurl, _this.getError());
                        _this.loaded = true;
                        doneasync();
                    } else {
                        _this.loaded = true;
                        console.log('loaded ' + _this.fileType + ' ' + csurl);
                        doneasync();
                    }
                    _this.setLocalFile(fileName);
                });

            } catch(err) {
                this.setSource('', doneasync);
                if (!skipFileExistence) this.error = err.message;
            }
        } else {
            var withData = function(data) {
                _this.setSource(data, function() {
                    if(_this.codeError) {
                        console.log('invalid ' + _this.fileType + ' ' + csurl, _this.codeError);
                    } else {
                        _this.loaded = true;
                        console.log('loaded ' + _this.fileType + ' ' + csurl);
                        doneasync();
                    }
                });
            }

            var project = pinegrow.getCurrentProject();
            if(project.hasContentForUrl(csurl)) {
                project.getContentOfUrl(csurl, withData);
            } else {
                $.ajax({
                    url: csurl,
                    data: null,
                    dataType: 'text'
                }).done(function (data) {
                    console.log('loaded ' + csurl);
                    _this.setLoadingFailed(false);
                    withData(data);
                }).fail(function () {
                    var message = 'Failed loading ' + csurl;
                    console.log(message);
                    _this.error = message;
                    _this.setLoadingFailed(true);
                });
            }
        }
    }

    pgStylesheet.loadFromUrl = function (url, done, force, skipFileExistence) {
        this.resetError();

        var _this = this;
        var csurl = this.setNakedUrl(url);

        var doneasync = function(a, b, c) {
            if(done) {
                setTimeout(function() {
                    done(_this, a, b, c);
                    pinegrow.dispatchEvent('on_imported_css_loaded', null, _this);
                }, 1);
            }
        }

        if(this.ignore && !force) {
            doneasync();
            return;
        }

        var fs = require('fs');
        if(isApp() && crsaIsFileUrl(csurl)) {
            try {
                var fileName = crsaMakeFileFromUrl(csurl, true);
                var str = fs.readFileSync(fileName, {encoding: "utf8"});
                console.log('loaded ' + csurl);

                this.setSource(str, function() {
                    if(_this.getError()) {
                        _this.loaded = true;
                        console.log('invalid ' + _this.fileType + ' ' + csurl);
                        doneasync();
                    } else {
                        _this.loaded = true;
                        console.log('loaded ' + _this.fileType + ' ' + csurl);
                        doneasync();
                    }
                    _this.setLocalFile(fileName);
                }, false, this.isImported());

            } catch(err) {
                this.setSource('', doneasync, false, this.isImported());
                if (!skipFileExistence) this.error = err.message;
            }
        } else {
            $.ajax({
                url: csurl,
                data: null,
                dataType: 'text'
            }).done(function(data) {
                    console.log('loaded ' + csurl);
                    _this.setLoadingFailed(  false);
                    _this.setSource(data, function() {
                        if(_this.codeError) {
                            console.log('invalid ' + _this.fileType + ' ' + csurl);
                        } else {
                            _this.loaded = true;
                            console.log('loaded ' + _this.fileType + ' ' + csurl);
                            doneasync();
                        }
                    }, false, _this.isImported());
                }).fail(function() {
                    var message = 'Failed loading ' + csurl;
                    console.log(message);
                    _this.error = message;
                    _this.setLoadingFailed( true);
                });
        }
    }

    pgStylesheet.getPageStylesheetFrom = function (page) {
        var csInfo = pinegrow.stylesheets.getStylesheetInfo(this, page);
        if (csInfo && csInfo.compiledStylesheet) {
            var compiledStylesheet = csInfo.compiledStylesheet;
            if (compiledStylesheet) {
                return pinegrow.stylesheets.findAttachmentFor(compiledStylesheet, page);
            }
        }

        return null;
    }

    pgStylesheet.saveCompiledFile = function (done) {
        var page = pinegrow.getSelectedPage();
        var csInfo = pinegrow.stylesheets.getStylesheetInfo(this, page);
        if (csInfo && csInfo.compiledStylesheet) {
            var compiledStylesheet = csInfo.compiledStylesheet;
            if (compiledStylesheet) {
                compiledStylesheet.save(compiledStylesheet.localFile, done);
            }
        }
    }

    pgStylesheet.save = function(path, done, fs, filter_func, fileWriter) {
        var _this = this;
        if(!fs) fs = require('fs');
        var shouldUpdateProject = false;
        var count = 0;

        var allDone = function () {
            if (count == 0) {
                if (shouldUpdateProject) pinegrow.refreshCurrentProjectDelayed(true);
                if (done) done();
            }
        }

        var saveCompiledCss = function (compiledStylesheet) {
            var cssPath = compiledStylesheet.localFile;
            if (!crsaIsFileExist(cssPath)) shouldUpdateProject = true;
            compiledStylesheet.save(cssPath, null, fs, filter_func, fileWriter);
        }

        var saveStylesheet = function (cs, filePath) {
            var pathLib = require('path');
            filePath = pathLib.resolve(filePath);
            if (!crsaIsFileExist(filePath)) shouldUpdateProject = true;

            var save = function() {
                setTimeout(function() {
                    try{
                        var csCode = cs.getSource();
                        csCode = _this.filterSource(csCode);

                        if(fileWriter) {
                            fileWriter.writeFile(filePath, csCode, function() {
                                cs.lastSavedAt = (new Date()).getTime();
                                console.log('saved ' + cs.fileType + ' path ' + filePath);
                                cs.setChanged(false);
                            })
                        } else {
                            crsaWriteFileWithBackup(fs, filePath, csCode, "utf8");
                            cs.lastSavedAt = (new Date()).getTime();
                            console.log('saved ' + cs.fileType + ' file ' + filePath);
                            cs.setChanged(false);
                        }

                        //only change name if file changed. avoid ?params bug
                        if(crsaIsFileUrl(cs.url) && cs.localFile != filePath) {
                            //leave remote css urls intact
                            cs.setUrl(crsaMakeUrlFromFile(filePath));
                        }

                        pinegrow.dispatchEvent('on_css_file_saved', null, cs);

                        count--;
                        allDone();
                    }
                    catch(err) {
                        done(err);
                    }
                }, 10);
            }

            cs.getCssSource(function(css) {
                save();
            }, function(error) {
                //for now ignore this. cp.save is doing syntax errors checking on cs before saving it. So this should never be reached anyway.
                return;


                var source = cs.getSource();
                //cs can't be compiled to css due to error, what to do?
                if(!cs.codeError && source !== null) {
                    //source is ok and parsable, only css can't be generated, should be ok to save
                    save();
                } else {
                    //syntax errors in source code?
                    if(cs.codeError && source !== null) {
                        //ask the user
                        pinegrow.showAlert(`<p>Stylesheet <b>${cs.url}</b> has syntax errors: ${cs.codeError}</p><p>Do you want to save the changes anyway?</p>`, `Syntax errors in ${cs.name}`, `Don't save`, 'Save', function() {
                            //cancel
                            count--;
                            allDone();
                        }, function() {
                            //save
                            save();
                        })
                    } else {
                        pinegrow.showAlert(`<p>Stylesheet <b>${cs.url}</b> can't be saved.`, `Syntax errors in ${cs.name}`, null, 'OK');
                        count--;
                        allDone();
                    }
                }
            });
        }
        count++;
        saveStylesheet(this, path);

        var saveAllImportedFiles = function (cs) {
            var forAll = function (importedFiles, parentPath) {
                for (var i = 0; i < importedFiles.length; i++) {
                    count++;
                    var importedFile = importedFiles[i];
                    parentPath = parentPath || path;

                    saveStylesheet(importedFile, importedFile.localFile);

                    if (importedFile.importedFiles.length > 0) forAll(importedFile.importedFiles, importedFile.localFile);
                }
            }

            forAll(cs.importedFiles);
        }

        if (this.importedFiles.length > 0) saveAllImportedFiles(this);

        var page = pinegrow.getSelectedPage();
        if (!this.isImportedFor(page)) {
            var csInfo = pinegrow.stylesheets.getStylesheetInfo(this, page);
            if (csInfo && csInfo.compiledStylesheet) {
                var compiledStylesheet = csInfo.compiledStylesheet;
                saveCompiledCss(compiledStylesheet);
            }
        } else {
            //imported
            var list = pinegrow.stylesheets.getIncludedPgStylesheetParent(this, page);
            if (list.length) {
                var topcs = list[0];
                var csInfo = pinegrow.stylesheets.getStylesheetInfo(topcs, page);
                if (csInfo) {
                    if (csInfo.compiledStylesheet && csInfo.compiledStylesheet.changed) {
                        saveCompiledCss(csInfo.compiledStylesheet);
                    }
                }
            }
        }
    }

    pgStylesheet.regenerateRules = function (done) {
        this.parseSource(this.source, function (error, postcssSource, rules) {
            if (error) {
                done({nodes: []});
            }
            else {
                done(rules);
            }
        }, true);
    }

    pgStylesheet.getRuleById = function () {
        // TODO implement this
    }

    pgStylesheet.getRuleSelector = function (rule, done) {
        var node = rule;
        var selector = "";
        while (node.type != "root") {
            if (node.type == "rule") {
                var nodeSelector = node.selector;

                if (nodeSelector[0] == '&') {
                    selector = nodeSelector.replace('&', '') + selector;
                }
                else {
                    selector = " " + nodeSelector + selector;
                }
                node = node.parent;
            }
            else {
                break;
            }
        }

        return selector.trim();
    }

    pgStylesheet.isVarInUse = function (node) {
        var _this = this;

        var regx = new RegExp('(\\' + node.prop + ')');

        var forAll = function (nodes) {
            for (var i = 0; i < nodes.length; i++) {
                var currentNode = nodes[i];
                if (currentNode.type == "decl") {
                    var matched = currentNode.value.match(regx);
                    if (matched && matched.length > 1) return true;
                }

                if (currentNode.nodes && currentNode.nodes.length > 0) {
                    if (forAll(currentNode.nodes)) return true;
                }
            }
        }

        var parent = node.parent;
        if (node.value.toLowerCase().indexOf('!global')) parent = this.getRootNode();

        // var isInUsed = forAll(parent.nodes) ? true : false;
        if (!isInUsed && this.isImported()) {
            var skipFirst = true;
            this.forEachParentCS(function (cs) {
                if (!skipFirst) {
                    isInUsed = forAll(currentStylesheet.getRootNode().nodes) ? true : false;
                    if (!isInUsed && cs.isImported()) {
                        return false;
                    }
                }
                else {
                    skipFirst = false;
                }
            });
        }
        return isInUsed
    }

    pgStylesheet.forgetCachedSource = function() {
        this.source = null;
        this.compilationError = null;
    }

    pgStylesheet.forgetSourceErrors = function() {
        this.codeError = null;
        this.compilationError = null;
        this.recursiveImports = null;
    }

    pgStylesheet.updateNodeStylesheetWithSource = function(node, source, done) {
        var _this = this;

        var updateToNewSource = function (newNode) {
            _this.setSourceForRules(pinegrow.pgPostCSSHelper.getRootNodeFor(newNode), function () {
                if (done) done(newNode);
            });
        }

        if (node.type == "decl") {
            var values = source.split(':');
            //node.prop = values[0].trim();
            pgcssh.setNodeProperty(node, 'prop', values[0].trim())
            //node.value = values[1].trim();
            pgcssh.setNodeProperty(node, 'value', values[1].trim())
            updateToNewSource(node);
        }
        else {
            this.parseSource(source, function (err, postcssSource, rootNode) {
                if (err) {
                    if (done) done();
                    return;
                }

                var newNode = pinegrow.pgPostCSSHelper.replaceNodeWith(node, rootNode.nodes[0]);
                newNode.raws.before = node.raws.before;
                updateToNewSource(newNode);
            }, true);
        }
    }

    pgStylesheet.renameRuleSelector = function (rule, newSelector, done, edited_by) {
        if(!this.canEdit(false, true)) {
            if(done) done();
            return;
        }

        if (pinegrow.pgPostCSSHelper.renameRuleSelector(rule, newSelector)) {
            this.forgetCachedSource();
            pinegrow.dispatchEvent('on_css_node_updated', null, {node: rule, stylesheet: this, edited_by: edited_by});
            this.setSourceForRules(pinegrow.pgPostCSSHelper.getRootNodeFor(rule), done, true);
        }
    }

    pgStylesheet.resetError = function () {
        this.error = null;
        this.codeError = null;
        this.compilationError = null;
        this.recursiveImports = null;
    }

    pgStylesheet.getError = function () {
        if(this.compilationError) {
            var e = this.compilationError;
            try {
                var file = '';
                if (e.file) file = e.file.replace(/\-pg[0-9]+/, '');

                return `${e.message} in file <b>${file}</b> at ${e.line}:${e.column}`;
            } catch(err) {}
        }
        return this.error || this.codeError || this.compilationError || this.recursiveImports;
    }

    pgStylesheet.filter = function (node) {
        var obj = this.superFilter(node);
        if (obj) return obj;

        return {
            haveTo: "doNothing"
        };
    }

    pgStylesheet.removeFromCompiledStylesheet = function (cs) {
        var index = this.compiledStylesheets.indexOf(cs);
        if (index > -1) {
            this.compiledStylesheets.splice(index, 1);
        }
    }

    pgStylesheet.remove = function (skipRemovingCompiled) {
        if (pinegrow.stylesheets.isStylesheetUsed(this)) return;
        this.removeSuper();

        for (var key in pgPrevStylesheetMap) {
            pgPrevStylesheetMap[key] = null;
        }
        pgPrevStylesheetMap = {};


        this.compiledStylesheets.forEach(function (compiledStylesheet) {
            if (compiledStylesheet) {
                if (!skipRemovingCompiled) compiledStylesheet.remove();

                if (compiledStylesheet) {
                    compiledStylesheet.sourceStylesheet = null;
                    compiledStylesheet = null;
                }
            }
        });
        this.ignoredImportedFiles = [];
        this.compiledStylesheets = [];
        this.parentPgStylesheets = [];

        if (this.varsPane) {
            this.varsPane.remove();
            this.varsPane = null;
        }

        delayed.destroy();
    }

    return pgStylesheet;
}
