"use strict";

var PgCSSActiveRulesEditor = function (selectedElement, $container) {
    $container.html('');
    if (!selectedElement) {
        $('<div class="alert alert-info">There is no element selected</div>').appendTo($container);
        return;
    }
    this.ready = true;

    var lastParentPath = null;
    var extraRuleElId = '-active-css';
    var cssStates = ['active', 'hover', 'focus', 'visited'];
    var activeDecl = {};
    var rulesDistance = {};
    var moreSheets = {};

    var _this = this;

    var $psudoContainer = $('<div class="pseudo-container"></div>');

    if(!pgf.kids) {
        $psudoContainer.appendTo($container);
    }

    var $list_container = $('<div class="rules-list"></div>').appendTo($container);

    var currDistance = 0;
    var stylePaneLastIndex = 1;

    var elementPgStylesheet = null;

    var toRuleIconShown = false;
    var $toRuleIcon = null;

    var stylesheets = [];
    var filesArr = [];

    var activeEditorRemoved = false;

    for (var i = 0; i < cssStates.length; i++) {
        var cssState = cssStates[i];
        this[cssState] = false;

        var $input  = $('<div class="pseudo-state" data-css-state="' + cssState + '"><label> :' + cssState + '</label><input type="checkbox"></div>').appendTo($psudoContainer);//.find('input')
        //var $input_chk = $input.find('input');

        $input.click(function (e) {
            var $this = $(this);
            var currCssState = $this.data('cssState');

            if (!_this[currCssState]) {
                _this[currCssState] = true;
                //$input_chk.prop("checked", true);
                $this.addClass('checked');
            } else {
                _this[currCssState] = false;
                //$input_chk.prop("checked", false);
                $this.removeClass('checked');
            }

            _this.removeStylesheetObj();
            togglePgStateClass(currCssState, _this[currCssState]);
            pinegrow.pgActiveRules.updateMatchedRules(true);
            // _this.show(null, null, current_selected_rule);
        });

        var $el = selectedElement.get$DOMElement();
        if ($el && $el.hasClass('pg-state-' + cssState)) {
            _this[cssState] = true;
            //$input_chk.prop("checked", true);
            $input.addClass('checked');
        }
        addTooltip($input, 'Show selected element in <b>:' + cssState + '</b> state.');
    }


    //Collapse all toolbar icon
    var toolbar = new PgButtonToolbar(true);
    toolbar.$element.appendTo($psudoContainer);

    toolbar.addSection('tools');

    var uncollapseAll = pinegrow.getSetting('active-uncollapse-all', "0") === '1';

    if(pgf.kids || !isApp()) {
        uncollapseAll = true;
    }

    var uncollapseAllButton = new PgButtonToolbarItem('icon-down', 'Always expand all rules');
    uncollapseAllButton.setActive(uncollapseAll);

    if(!pgf.kids || !isApp()) {
        toolbar.addItem('tools', uncollapseAllButton);
    }

    uncollapseAllButton.onClick = function(item) {
        uncollapseAll = !this.getActive();
        this.setActive(uncollapseAll);
        pinegrow.setSetting('active-uncollapse-all', uncollapseAll ? "1" : "0");
        pinegrow.selectedElements.reselect();
        pinegrow.showMode('Uncollapse rules', uncollapseAll);
    }

    var rememberRuleForElement = pinegrow.getSetting('active-remember-for-element', "1") === '1';

    var rememberButton = new PgButtonToolbarItem('icon-Pin', 'Remember the selected CSS rule for each element');
    rememberButton.setActive(rememberRuleForElement);

    if(!pgf.kids || !isApp()) {
        toolbar.addItem('tools', rememberButton);
    }

    rememberButton.onClick = function(item) {
        rememberRuleForElement = !this.getActive();
        this.setActive(rememberRuleForElement);
        pinegrow.setSetting('active-remember-for-element', rememberRuleForElement ? "1" : "0");
        pinegrow.showMode('Remember the selected CSS rule for each element', rememberRuleForElement);
    }

    // var $cmRules = $('<div class="crsa-cm-rules"></div>').insertBefore($container);

    var onNewRuleCreated = function (elementSelector, selectedFileUrl, mediaParams, done) {
        for (var i = 0; i < stylesheets.length; i++) {
            if (stylesheets[i].url == selectedFileUrl) {
                var stylesheet = stylesheets[i];

                if(!stylesheet.canEdit()) {
                    return;
                }

                var insertRule = false;
                var rootNode = stylesheet.getRootNode();
                var addNodeTo = rootNode;

                var ruleNode = pinegrow.pgPostCSSHelper.createNode("rule", elementSelector, null, null, stylesheet);
                //var ruleNode = pinegrow.pgPostCSSHelper.createNode("rule", elementSelector);
                var ruleIndices = [rootNode.nodes.length];
                var toCheckIndices = [rootNode.nodes.length];

                var parentNode = ruleNode;
                if (mediaParams) {
                    var lastNode = rootNode.last;
                    if (lastNode && lastNode.params && lastNode.params == mediaParams) {
                        addNodeTo = lastNode;
                        ruleIndices = [rootNode.nodes.length - 1];
                        toCheckIndices = [rootNode.nodes.length - 1, lastNode.nodes.length];
                        insertRule = true;
                    }
                    else {
                        var mediaNode = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams, null, stylesheet);
                        //var mediaNode = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams);
                        pinegrow.pgPostCSSHelper.insertNodeInside(ruleNode, mediaNode);
                        parentNode = mediaNode;
                        // ruleIndices.push(0);
                        toCheckIndices.push(0);
                    }
                }

                pinegrow.pgPostCSSHelper.insertNodeAt(parentNode, addNodeTo, addNodeTo.nodes.length);

                stylesheet.setSourceForRules(rootNode, function () {
                    if (done) done(stylesheet, ruleIndices, toCheckIndices, insertRule);
                });
                break;
            }
        }
    }

    var onSelectorMakerCreate = function(maker) {
        var elementSelector = maker.getSelector();
        var selectedFileUrl = maker.getStylesheet();

        var add_classes = maker.getAddClasses();
        var save_inline = maker.getSaveInline();

        var doTask = function() {

            if (add_classes.length) {
                var col = pinegrow.getCollectionForElement(selectedElement);
                var api = new PgApi();
                api.addClass(col, add_classes);
            }

            onNewRuleCreated(elementSelector, selectedFileUrl, maker.getMediaQuery(), function (stylesheet, ruleIndices, toCheckIndices, insertRule) {
                var node = pinegrow.pgPostCSSHelper.getNodeByIndices(stylesheet.getRootNode(), ruleIndices);
                var nodeToCheck = pinegrow.pgPostCSSHelper.getNodeByIndices(stylesheet.getRootNode(), toCheckIndices);
                if (insertRule) {
                    pinegrow.dispatchEvent('on_css_node_added', null, {node: nodeToCheck, stylesheet: stylesheet});
                }
                else {
                    pinegrow.dispatchEvent('on_css_node_added', null, {node: node, stylesheet: stylesheet});
                }

                var target_rule = null;

                if (node.type == 'atrule') {
                    target_rule = node.nodes[node.nodes.length - 1];
                }
                else {
                    target_rule = node;
                }

                if(save_inline) {
                    try {
                        var $stylesheet = $('.style-attribute-style-pane');
                        var cs_inline = $stylesheet.data('cs');
                        var rule = cs_inline.getRootNode().nodes[0];

                        var r = target_rule;

                        var cs = pgcssh.getStylesheetForNode(r);

                        var last_decl = 0;

                        for(var i = 0; i < rule.nodes.length; i++) {
                            if(rule.nodes[i].type == 'decl') {
                                last_decl = i;
                            }
                        }
                        for(var i = 0; i < rule.nodes.length; i++) {
                            if(rule.nodes[i].type == 'decl') {
                                cs.ruleValueChanged(target_rule, rule.nodes[i].prop, rule.nodes[i].value, null, false, 'gui', i == last_decl, null);
                            }
                        }

                        var col = pinegrow.getCollectionForElement(selectedElement);
                        col.forEachElement(function (pgel) {
                            pgel.removeAttr('style');
                        });
                        pinegrow.changeManager.didChange(col, 'Remove style attribute');

                    } catch(err) {}
                }


                selectRule(target_rule);

                if (!activeEditorRemoved) {
                    var containers = createContainers(stylesheet, false, 1);
                    var $stylePaneContainer = containers.$stylePaneContainer;

                    addNodePath(stylesheet, node, containers.$currRuleList, true);
                    showRulesFor(stylesheet, node, nodeToCheck, $stylePaneContainer, {uncollapse: true});

                    $stylePaneContainer.find('li.crsa-node').addClass('node-not-active');

                    pinegrow.pgActiveRules.updateMatchedRules(false, function () {
                        var matched_rules = pinegrow.pgActiveRules.matched_rules;
                        if (current_matched_rules.length != matched_rules.length) active_out_dated = true;
                        current_matched_rules = pinegrow.pgActiveRules.matched_rules;
                    });

                    updateClassesWithoutRules();

                    createRuleView && createRuleView.setSelector('');
                }

                pinegrow.showQuickMessage('Rule <b>' + elementSelector + '</b> created in <b>' + crsaGetNameFromUrl(selectedFileUrl) + '</b>.')
            });
        }

        if(selectedFileUrl) {
            doTask();
        } else {
            pinegrow.getDefaultStylesheet(selectedElement.getPage(), function(cs) {

                if(cs) {
                    selectedFileUrl = cs.url;
                    doTask();
                } else {
                    pinegrow.showQuickError('Please choose or add stylesheet for the new CSS rule.');
                }
            }, true /* show error */)
        }
    }

    this.getSelectedElement = function () {
        return selectedElement;
    }

    this.openCreateNewRuleDialog = function(selector, media, stylesheet, select_class, simple, $target) {

        fillStylesheetsArray();

        var opts = {
            stylesheets: filesArr,
            $position: $target || $createNewRule,
            selector_value: selector,
            stylesheet_value: stylesheet,
            media_value: media,
            select_class: select_class,
            simple: simple
        }

        var selector_maker = new PgQuickSelectorMaker(selectedElement, null, opts);

        selector_maker.onCreate = onSelectorMakerCreate;
    }

    var $addRuleContainer = $('<div class="pg-active-add-rule"></div>').appendTo($list_container);

    var $createNewRule;
    var createRuleView;

    var selectedElementPage = selectedElement.getPage();

    this.showCreateCssRule = function(sel) {
        if(!createRuleView) return;

        createRuleView.setSelector(sel);
        $list_container.scrollTop(0);
        createRuleView.onShown(); //will focus field
        addTooltip(createRuleView.getInput().$field, 'Input the selector and press ' + pgShowKbdCombo('Enter') + '', {
            trigger: 'manual'
        })
    }

    pinegrow.getDefaultStylesheet(selectedElementPage, function(defaultCs, detected) {
        if(defaultCs || true) {
            createRuleView = new PgSelectorMakerView(selectedElement, null, {
                simple: true,
                labels: false,
                media_query: false,
                stylesheets: false,
                compact: true,
                save_inline: true
            })

            if(!defaultCs) {
                detected = true; //force dialog
            }

            createRuleView.view.setAlwaysVisible(true);

            if(detected) {
                createRuleView.onCreate = function(maker) {
                    var selectDefault = new PgSelectDefaultStylesheet(selectedElementPage, defaultCs);
                    selectDefault.show(function(cs, canceled) {
                        if(!canceled && cs) {
                            onSelectorMakerCreate(maker);
                        }
                    })

                }
            } else {
                createRuleView.onCreate = onSelectorMakerCreate;
            }

            createRuleView.onEmptySelector = function() {
                $('.create-dialog').trigger('click');
            }

            if(!pgf.kids) {
                $addRuleContainer.append(createRuleView.view.get$Element());
            } else {
                var $name = $(`<div class="pg-show-properties-name" style="margin-top: -5px;margin-bottom: 10px;"><span>Style for</span> <name>${pinegrow.selectedElements.getName({short: true, num_classes: 3})}</name></div>`).prependTo($addRuleContainer);
            }

            var $button = $addRuleContainer.find('button.create');

            var def_name = defaultCs ? crsaGetNameFromUrl(defaultCs.url) : 'default stylesheet';

            addTooltip($button, 'Create a new CSS rule in <b>' + def_name + '</b>. Input the selector and press ' + pgShowKbdCombo('Enter') + ' or click Create. Click on ... to open selector builder and choose media query and stylesheet.')

            $createNewRule = $('<button class="pg-button create-dialog">' + (pgf.kids ? 'Create new CSS rule' : '...') + '</button>');
            pgf.kids ? $addRuleContainer.append($createNewRule) : $createNewRule.insertAfter($button);

            if(pgf.kids && pgf.kids_with_transitions) {
                var pgel = pinegrow.selectedElements.getLastSelected();
                var $el = pgel.get$DOMElement();

                if($el) {
                    var grid_info = new PgCSSGridInfo($el);
                    var is_grid = grid_info.isGrid();

                    var $editCSSGrid = $(`<button class="pg-button edit-grid" style="margin-left:8px;">${is_grid ? 'Edit grid' : 'Create grid'}</button>`).appendTo($addRuleContainer);
                    $editCSSGrid.on('click', function (e) {
                        e.preventDefault();
                        pinegrow.showCSSGridControl(pinegrow.selectedElements.getLastSelected(), $editCSSGrid);
                    })

                    if (is_grid) {
                        $editCSSGrid.addClass('pg-button-primary');
                        addTooltip($editCSSGrid, 'Edit the CSS grid layout of the selected element.');
                    } else {
                        addTooltip($editCSSGrid, 'Create a CSS grid layout for the selected element.');
                    }
                }
            }

            if(false && pgf.kids) {
                //var $show_all = $(`<label class="pg-show-all-rules-checkbox"><input type="checkbox" ${pinegrow.pgActiveRules.show_all_rules ? ' checked' : ''}> Show all rules</label>`).appendTo($addRuleContainer);

                var $show_all = $('<i class="pg-show-all-rules-checkbox icon icon-267"></i>').appendTo($addRuleContainer);
                !pinegrow.pgActiveRules.show_all_rules && $show_all.addClass('active');

                addTooltip($show_all, 'Show only CSS rules that are active for the selected element (otherwise all CSS rules are shown).');
                $show_all.on('click', function(e) {
                    e.preventDefault();
                    //pinegrow.pgActiveRules.setShowAllRules($show_all.find('input').is(':checked'));
                    pinegrow.pgActiveRules.setShowAllRules(!pinegrow.pgActiveRules.show_all_rules);
                    pinegrow.selectedElements.reselect();
                })
            }


        } else {
            $createNewRule = $('<button class="pg-button create-dialog">Create new rule</button>').appendTo($addRuleContainer);
        }

        $createNewRule.on('click', function(e) {
            e.preventDefault();

            _this.openCreateNewRuleDialog(createRuleView ? createRuleView.getSelector() : null, null, null, null, false, $(e.target));
        })

        addTooltip($addRuleContainer.find('button.create-dialog'), 'Create new CSS rule with selector builder.')
    })

    var $classesWithoutRules = $('<div class="classes-whithout-rules"><p>Classes without rules: <classlist></classlist></p></div>');

    if(!pgf.kids) {
        $classesWithoutRules.appendTo($list_container);
    }

    $classesWithoutRules.on('click','a', function(e) {
        e.preventDefault();
        var cls = $(this).attr('data-class');
        _this.openCreateNewRuleDialog('.' + cls, null, null, null, true /* simple */, $(e.target));
    })

    pg$Hide($classesWithoutRules);

    var updateClassesWithoutRules = function() {
        //disable because it shows classes that have rules but are not active on the current media query
        return;

        var pgel = selectedElement;
        var rules = current_matched_rules;

        var $ul = $classesWithoutRules.find('classlist');
        var html = '';
        var all = [];
        for(var im = 0; im < rules.length; im++) {
            if(rules[im].text) {
                var list = pgcssh.getClassesFromSelector(rules[im].text);
                list.forEach(function(cls) {
                    if(all.indexOf(cls) < 0) {
                        all.push(cls);
                    }
                })
            }
        }
        pgel.getClasses().forEach(function(cls) {
            if(all.indexOf(cls) < 0) {
                if(html.length) html += ', ';
                html += '<a href="#" data-class="' + cls + '">' + cls + '</a>';
            }
        })

        if(html.length) {
            $ul.get(0).innerHTML = html;
            pg$Show($classesWithoutRules);
        } else {
            pg$Hide($classesWithoutRules);
        }

        addTooltip($ul.find('a'), 'Click to create CSS rule for this class.');
    }

    var $ruleList = $('<div class="all-active-rule-list"></div>').appendTo($list_container);

    //$cmRules.append($container);

    var $moreStylesheet = $('<div class="more-stylesheet"></div>').appendTo($list_container);
    var $sheetList = $('<ul></ul>').appendTo($moreStylesheet);

    var stylePaneList = [];

    var selectRule = function (rule) {
        // var rule_obj = getObjectFromRule(rule);
        $.fn.crsa('selectRule', rule);
        onRuleSelected(rule);
    }

    var getDocumentForSelectedPage = function () {
        return pinegrow.getSelectedPage().$iframe.contents();
    }

    var togglePgStateClass = function (state, add) {
        var el = selectedElement.get$DOMElement().get(0);
        if (add) {
            $(el).addClass('pg-state-' + state);
        } else {
            $(el).removeClass('pg-state-' + state);
        }
    }

    var getFilenameId = function (pgStylesheet) {
        return pgStylesheet.getUniqueName();
    }

    var convertStatesToClass = function (selector) {
        var states = ['hover', 'active', 'focus', 'visited'];

        var newSelector = selector;
        for (var i = 0; i < states.length; i++) {
            newSelector = newSelector.replace(new RegExp(':' + states[i], 'g'), '.pg-state-' + states[i]);
        }

        return newSelector;
    }

    var canShowNodeBySelector = function (selector, $el) {
        var found = true;
        var selArr = selector.split(',');
        for (var i = 0; i < selArr.length; i++) {
            var sel = selArr[i].trim();

            sel = sel.replace(/&/g, '*');
            if (sel[0] == '>') {
                sel = sel.substr(1, sel.length).trim();
            }

            var stateFound = false;
            for (var j = 0; j < cssStates.length; j++) {
                var state = ':' + cssStates[j];
                if (sel.indexOf(state) > -1) {
                    stateFound = true;
                    if (sel.length == sel.indexOf(state) + state.length) {
                        if (state == ":active") {
                            if (_this.active)  return true;
                            else               found = false;
                        }
                        if (state == ":hover") {
                            if (_this.hover)   return true;
                            else               found = false;
                        }
                        if (state == ":focus") {
                            if (_this.focus)   return true;
                            else               found = false;
                        }
                        if (state == ":visited") {
                            if (_this.visited) return true;
                            else               found = false;
                        }
                    }
                    else {
                        var cleanSel = sel.replace(/>\s*/g, '>').replace(/ +(?= )/g,'').trim();
                        var cleanSel = convertStatesToClass(cleanSel);
                        var newSelArr = cleanSel.split(' ');

                        var $currEl = $el;
                        for (var k = newSelArr.length - 2; k >= 0; k--) {
                            if (newSelArr[k + 1][0] == '>' && $currEl) {
                                if ($currEl.parent().length === 0 || !$currEl.parent().is(newSelArr[k])) {
                                    found = false;
                                    break;
                                }
                            }
                            else {
                                $currEl = $currEl.closest(newSelArr[k]);
                                if ($currEl.length == 0) {
                                    found = false;
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            try {

                if (!stateFound && $el[0].matches(sel)) {
                    return true;
                }
            } catch(err) {}
        }

        if (found) return true;
        else       return false;
    }

    var canShowNode = function (node, $el) {
        if (!node.selector) return true;

        return canShowNodeBySelector(node.selector, $el)
    }

    var getRuleElId = function (node, pgStylesheet, stylePaneIndex, nodeId, by) {
        if (node.type != "rule") return '';

        nodeId = nodeId || parseInt(pgStylesheet.getNodeId(node));
        by = by || 0;
        var ruleId = nodeId + by;
        return 'pg-' + getFilenameId(pgStylesheet) + '-rule-id-' + ruleId + stylePaneIndex;
    }

    var addDeclToActiveDecl = function (node, pgStylesheet, stylePaneIndex) {
        var parentDistance = node.parent.distance;
        var nodes = node.nodes;

        for (var i = 0; i < nodes.length; i++) {
            var currNode = nodes[i];

            if (currNode.type == "decl") {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(currNode);
                var importantNode = currNode.important || currNode.value.endsWith('!important') || false;
                var parentElId = getRuleElId(node, pgStylesheet, stylePaneIndex);

                if (activeDecl[currNode.prop]) {
                    var existNode = activeDecl[currNode.prop];
                    if (importantNode) {
                        if (existNode.nodes[existNode.activeIndex].important) {
                            activeDecl[currNode.prop].nodes.push({
                                distance: activeDecl[currNode.prop].nodes.length,
                                important: importantNode,
                                nodeIndices: nodeIndices,
                                parentElId: parentElId,
                                uniqueName: pgStylesheet.getUniqueName()
                            });
                        }
                        else {
                            activeDecl[currNode.prop].nodes.push({
                                distance: activeDecl[currNode.prop].nodes.length,
                                nodeIndices: nodeIndices,
                                parentElId: parentElId,
                                important: currNode.important,
                                uniqueName: pgStylesheet.getUniqueName()
                            });
                            activeDecl[currNode.prop].activeIndex = activeDecl[currNode.prop].nodes.length - 1;
                        }
                    }
                    else {
                        activeDecl[currNode.prop].nodes.push({
                            distance: activeDecl[currNode.prop].nodes.length,
                            nodeIndices: nodeIndices,
                            important: importantNode,
                            parentElId: parentElId,
                            uniqueName: pgStylesheet.getUniqueName()
                        });

                        var activeNodeObj = existNode.nodes[existNode.activeIndex];
                        if (parentElId == activeNodeObj.parentElId) {
                            activeDecl[currNode.prop].activeIndex = activeDecl[currNode.prop].nodes.length - 1;
                        }
                    }
                }
                else {
                    activeDecl[currNode.prop] = {
                        activeIndex: 0,
                        nodes: [{
                            distance: 0,
                            nodeIndices: nodeIndices,
                            parentElId: parentElId,
                            important: importantNode,
                            uniqueName: pgStylesheet.getUniqueName()
                        }]
                    }
                }
            }
        }
    }

    var getNode$el = function (node, pgStylesheet, stylePaneIndex) {
        var id = getRuleElId(node, pgStylesheet, stylePaneIndex);
        return getNode$elById(id);
    }

    var getNode$elById = function (id) {
        if (id) id = "#" + id;
        var nodes$el = $container.find(id)
        return nodes$el.length > 0 ? $(nodes$el[0]) : $();
    }

    var findDeclElFromRuleEl = function ($ruleEl, nodeIndices) {
        return $ruleEl.find('[data-node-indices="' + nodeIndices.join(',') + '"]');
    }

    var getDeclIndices = function (node) {
        return pinegrow.pgPostCSSHelper.getNodeIndices(node);
    }

    var getStylesheetFromParentId = function (parentElId) {
        var $ruleEl = getNode$elById(parentElId);
        var $csStylesheet = $ruleEl.closest('.crsa-stylesheet')
        var cs = $csStylesheet.data('cs');
        return cs;
    }

    var setOverridenDecl = function () {
        if(pgf.kids || pinegrow.pgActiveRules.show_all_rules) return;

        for (var activeProp in activeDecl) {
            var nodes = activeDecl[activeProp].nodes;
            for (var i = 0; i < nodes.length; i++) {
                if (i  == activeDecl[activeProp].activeIndex) continue;

                var overriden = nodes[i];
                if (overriden.parentElId) {
                    var $declEl = findDeclElFromRuleEl(getNode$elById(overriden.parentElId), overriden.nodeIndices);
                    $declEl.addClass('overriden-node');
                }
            }
        }
    }

    var findNodeIndexInOveridden = function (overriden, node, pgStylesheet, stylePaneIndex) {
        var parentRuleElId = getRuleElId(node.parent, pgStylesheet, stylePaneIndex);
        for (var i = 0; i < overriden.length; i++) {
            if (overriden[i].parentElId == parentRuleElId) return i;
        }
    }

    var getDistanceBetween = function (frNode, sdNode, uniqueName) {
        if(!frNode || !sdNode) return 0;
        var parentDistance = getRuleDistance(frNode.parent) - getRuleDistance(sdNode.parent);
        if (parentDistance != 0) return parentDistance;

        var frIndices = getDeclIndices(frNode);
        var sdIndices = getDeclIndices(sdNode);

        for (var i = 0; i < frIndices.length; i++) {
            var distance = frIndices[i] - sdIndices[i];
            if (distance != 0) return distance;
        }

        return 0;
    }

    var updateNodeDistance = function (nodes) {
        for (var i = 0; i < nodes.length; i++) {
            nodes[i].distance = i;
        }
    }

    var getNodeDistance = function (node, activeDeclNodes) {
        var currentNodeParentDistance = getRuleDistance(node.parent);

        var getDistance = function (obj, node) {
            var activeNodeParentDistance = getNodeParentDistanceFromIndices(obj.uniqueName, obj.nodeIndices);
            // var activeNodeParentDistance = getRuleDistance(node);

            if (currentNodeParentDistance < activeNodeParentDistance) return obj.distance;
            else if (currentNodeParentDistance != activeNodeParentDistance) return null;

            var currNodeIndices = obj.nodeIndices;
            var nodeIndices = getDeclIndices(node);

            var currNodeLength = false;
            var length = nodeIndices.length;
            if (currNodeIndices.length > nodeIndices.length) {
                currNodeLength = true;
                length = currNodeIndices.length;
            }

            for (var j = 0; j < length; j++) {
                var currNodeIndex = currNodeIndices[j];
                var nodeIndex = nodeIndices[j];

                if (nodeIndex > currNodeIndex) return obj.distance;
                else if (nodeIndex != currNodeIndex) break;
            }

            return null;
        }

        var nodes = activeDeclNodes.nodes;
        for (var i = 0; i < nodes.length; i++) {
            var distance = getDistance(nodes[i], node);
            if (distance != null) return distance;
        }

        return nodes.length;
    }

    var setNewOverridenNode = function (prop, frNodeImportant, sdNodeImportant, nodeIndex, activeObj, hasMorePriority) {
        if (hasMorePriority) {
            if (frNodeImportant == false || sdNodeImportant == true) {
                useNodeAsCurrentNode(prop, nodeIndex, activeObj);
                return true;
            }
        }
        else {
            if (!(frNodeImportant == true || sdNodeImportant == false)) {
                useNodeAsCurrentNode(prop, nodeIndex, activeObj);
                return true;
            }
        }
        return false;
    }

    var addNewNodeTo = function (node, activeDeclNodes, pgStylesheet, stylePaneIndex) {
        var nodes = activeDeclNodes.nodes;
        var parentElId = getRuleElId(node.parent, pgStylesheet, stylePaneIndex);
        var importantNode = node.important || node.value.endsWith('!important') || false;

        var distance = getNodeDistance(node, activeDeclNodes);

        if (activeDeclNodes.activeIndex >= distance) activeDeclNodes.activeIndex++;

        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        nodes.splice(distance, 0, {
            important:  importantNode,
            distance: distance,
            nodeIndices: nodeIndices,
            parentElId: parentElId
        });

        setNodeElActive(parentElId, nodeIndices, false);
        updateNodeDistance(nodes);
        return distance;
    }

    var setNodeElActive = function (parentElId, nodeIndices, active) {
        if(pgf.kids) return;
        var $declEl = findDeclElFromRuleEl(getNode$elById(parentElId), nodeIndices);

        if (active) $declEl.removeClass('overriden-node');
        else        $declEl.addClass('overriden-node');
    }

    var useNodeAsCurrentNode = function (prop, nodeIndex, oldActiveObj) {
        var activeDeclObj = activeDecl[prop];

        activeDeclObj.activeIndex = nodeIndex;
        var newActiveObj = activeDeclObj.nodes[nodeIndex];

        setNodeElActive(oldActiveObj.parentElId, oldActiveObj.nodeIndices, false);
        setNodeElActive(newActiveObj.parentElId, newActiveObj.nodeIndices, true);
    }

    var findNodeFromObjs = function (nodeIndices, objs, pgStylesheet) {
        for (var i = 0; i < objs.length; i++) {
            var activeNodeInices = objs[i].nodeIndices.join(',');
            var changedNodeIndices = nodeIndices.join(',');

            if (activeNodeInices == changedNodeIndices) return { nodeObj: objs[i], index: i };
        }

        return null;
    }

    var findClosestNodeIndex = function (objs) {
        for (var i = 0; i < objs.length; i++) {
            if (objs[i].important) return i;
        }

        return 0;
    }

    var isInsertedNode = function (node, pgStylesheet, stylePaneIndex) {
        var insertedRule = false;
        var parent = node.parent;

        while (parent && parent.type != 'root') {
            var parentElId = getRuleElId(parent, pgStylesheet, stylePaneIndex);
            if (parentElId) {
                var $ruleEl = getNode$elById(parentElId);

                if ($ruleEl.hasClass('node-not-active')) {
                    insertedRule = true;
                    break;
                }
            }

            parent = parent.parent;
        }

        return insertedRule;
    }

    var isNodeAlreadyAdded = function (node, pgStylesheet) {
        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
        var activeDeclObj = activeDecl[node.prop];
        if (activeDeclObj) {
            return findNodeFromObjs(nodeIndices, activeDeclObj.nodes, pgStylesheet) ? true : false;
        }
        else {
            return false;
        }
    }

    var toggleConvertToRuleIcon = function (pgStylesheet, show) {
        if (pgStylesheet.elementStyle) {
            var rootNode  = pgStylesheet.getRootNode();
            var firstNode = rootNode.nodes[0];
            var declNodes = firstNode.nodes;

            if (show) {
                if (declNodes.length > 0) {
                    $toRuleIcon.show();
                    toRuleIconShown = true;

                    createRuleView && createRuleView.setSaveInline(true);
                    /*
                     addTooltip($toRuleIcon, "Save the inline style into a proper CSS rule.", {
                     trigger: 'manual'
                     })*/
                    setTimeout(function() {
                        if($toRuleIcon && $toRuleIcon.is(':visible')) {
                            $toRuleIcon.addClass('animate');
                        }
                    }, 500)

                    $toRuleIcon.closest('.style-attribute-style-pane').addClass('has-inline-style');
                }
            }
            else {
                if (declNodes.length == 1) {
                    $toRuleIcon.hide();
                    $toRuleIcon.removeClass('animate');
                    toRuleIconShown = false;
                    $toRuleIcon.closest('.style-attribute-style-pane').removeClass('has-inline-style');
                    createRuleView && createRuleView.setSaveInline(false);
                }
            }
        }
    }

    var findClassesFromSelector = function (selector) {
        var parsedSelector = {
            definer: '',
            classes: [],
            id: ''
        };

        var type = 'definer';
        var currentToken = '';
        var addToken = function (token, tokenType) {
            if (tokenType == 'class') {
                parsedSelector.classes.push(token);
            }
            else if (tokenType == 'id') {
                parsedSelector.id = token;
            }
            else if (tokenType == 'definer') {
                parsedSelector.definer = token;
            }
            currentToken = '';
        }

        var i = 0;
        while (i < selector.length) {
            if (selector[i] == '.') {
                addToken(currentToken, type);
                type = 'class';
            }
            else if (selector[i] == '#') {
                addToken(currentToken, type);
                type = 'id';
            }
            else {
                currentToken += selector[i];
            }
            i++;
        }

        addToken(currentToken, type);

        return parsedSelector;
    }

    var newRuleAdded = function (node, pgStylesheet, stylePaneIndex, info) {
        var parentElId = getRuleElId(node, pgStylesheet, stylePaneIndex);
        var $ruleEl = getNode$elById(parentElId);
        if(!$ruleEl) return;
        $ruleEl.addClass('node-not-active');

        var addClassesAndId = true;
        var selector = pgStylesheet.filterSource(node.selector);
        if (selector.indexOf(',') > -1) return;

        // var selectorObj = findClassesFromSelector(selector);
        // if (selectorObj.definer) {
        //     var $el = selectedElement.get$DOMElement();
        //     if (!$el[0].matches(selectorObj.definer)) addClassesAndId = false;
        // }

        // if (addClassesAndId) {
        //     if(info && info.operation && info.operation == 'ignore') {
        //         //rule was un-hiden, no need to do anything
        //     } else {
        //         //better to use convenient PgApi, it handles mulyiple selected elements and creates undo points. Also takes multiple classes. Check out the api.addClass to see everything that it does.
        //         //selectedElement.addClass(selectorObj.classes.join(' '));
        //         var api = new PgApi();
        //         api.addClass(null /* all selected */, selectorObj.classes, {
        //             source: 'elementStyle'
        //         });

        //         /* dangerous, easy to duplicate id or replace existing one
        //         if (selectorObj.id) {
        //             selectedElement.attr('id', selectorObj.id);
        //         }*/
        //     }
        // }
    }

    var overridenRuleObj = { key: '', val: '' };
    var overridenDeclObj = {};
    var onNodeIndicesChanged = function (node, pgStylesheet, oldNodeIndices, newNodeIndices, pgStylePane, ulOpenForRule) {
        if (!node) return;
        newNodeIndices = newNodeIndices || pinegrow.pgPostCSSHelper.getNodeIndices(node);

        var changeChildrenIndices = function (nodes) {
            for (var i = 0; i < nodes.length; i++) {
                if (nodes[i].type == 'decl') {
                    onNodeIndicesChanged(nodes[i], pgStylesheet, oldNodeIndices.concat(i), newNodeIndices.concat(i), pgStylePane);
                }
            }
        }

        var newDeclIndicesChanged = function (currOverridenDeclObj, oldNode, newNode, parentElId) {
            if (newNode) {
                currOverridenDeclObj.key = newNodeIndices;
                currOverridenDeclObj.val = newNode;
                currOverridenDeclObj.parentElId = parentElId;
            }
            else {
                currOverridenDeclObj.key = '';
                currOverridenDeclObj.val = '';
                currOverridenDeclObj.parentElId = -1;
            }

            var currentNodeObj = activeDeclObj.nodes[oldNode.index];
            currentNodeObj.nodeIndices = newNodeIndices;
        }

        if (node.type == "rule") {
            var parentElId = getRuleElId(node, pgStylesheet, pgStylePane.getIndex());
            var $ruleEl = getNode$elById(parentElId);
            if ($ruleEl.length === 0 || $ruleEl.hasClass('node-not-active')) return;

            var newNodeDistanceKey = pgStylesheet.getUniqueName() + '-' + newNodeIndices,
                oldNodeDistanceKey = pgStylesheet.getUniqueName() + '-' + oldNodeIndices;

            if (overridenRuleObj.key == oldNodeIndices) {
                var newVal = overridenRuleObj.val;

                if (rulesDistance[newNodeDistanceKey]) {
                    overridenRuleObj.key = newNodeIndices;
                    overridenRuleObj.val = rulesDistance[newNodeDistanceKey];
                }
                else {
                    overridenRuleObj.key = '';
                    overridenRuleObj.val = '';
                }

                rulesDistance[newNodeDistanceKey] = newVal;
            }
            if (rulesDistance[oldNodeDistanceKey]) {

                if (rulesDistance[newNodeDistanceKey]) {
                    overridenRuleObj.key = newNodeIndices;
                    overridenRuleObj.val = rulesDistance[newNodeDistanceKey];
                }

                rulesDistance[newNodeDistanceKey] = rulesDistance[oldNodeDistanceKey];
                delete rulesDistance[oldNodeDistanceKey];
            }

            if (!ulOpenForRule) changeChildrenIndices(node.nodes);
        }
        else if (node.type == "decl") {
            var parentElId = getRuleElId(node.parent, pgStylesheet, pgStylePane.getIndex());
            var $ruleEl = getNode$elById(parentElId);
            if ($ruleEl.length === 0 || $ruleEl.hasClass('node-not-active')) return;

            var activeDeclObj = activeDecl[node.prop];

            if (!overridenDeclObj[node.prop]) overridenDeclObj[node.prop] = { key: '', val: '', parentElId: -1 };
            var currOverridenDeclObj = overridenDeclObj[node.prop];

            var oldNode, newNode;
            if (currOverridenDeclObj.key.toString() == oldNodeIndices.toString() &&
                currOverridenDeclObj.parentElId == parentElId) {
                oldNode = currOverridenDeclObj.val;
                newNode = findNodeFromObjs(newNodeIndices, activeDeclObj.nodes, pgStylesheet);
            }
            else {
                oldNode = findNodeFromObjs(oldNodeIndices, activeDeclObj.nodes, pgStylesheet);
                newNode = findNodeFromObjs(newNodeIndices, activeDeclObj.nodes, pgStylesheet);
            }

            newDeclIndicesChanged(currOverridenDeclObj, oldNode, newNode, parentElId);
        }
    }

    var onNodeAdded = function (node, pgStylesheet, pgStylePane, info) {
        if (node == null) return;

        if (node.type == 'rule') {
            newRuleAdded(node, pgStylesheet, pgStylePane.getIndex(), info);
        }
        else if (node.type == 'decl') {
            var nodeParent = node.parent;
            if (!nodeParent || nodeParent.type != 'rule') return;

            if (!toRuleIconShown) toggleConvertToRuleIcon(pgStylesheet, true);

            if (isInsertedNode(node, pgStylesheet, pgStylePane.getIndex())) return;
            if (isNodeAlreadyAdded(node, pgStylesheet)) return;
            var importantNode = node.important || node.value.endsWith('!important') || false;
            var activeDeclNodes = activeDecl[node.prop];

            if (activeDeclNodes) {
                var nodeIndex = addNewNodeTo(node, activeDeclNodes, pgStylesheet, pgStylePane.getIndex());
                var activeObj = activeDeclNodes.nodes[activeDeclNodes.activeIndex];

                var currentNodeParentDistance = getRuleDistance(node.parent);
                var activeNodeParentDistance = getNodeParentDistanceFromIndices(activeObj.uniqueName, activeObj.nodeIndices);

                if (currentNodeParentDistance == activeNodeParentDistance) {
                    var $oldActiveDecl = findDeclElFromRuleEl(getNode$elById(activeObj.parentElId), activeObj.nodeIndices);
                    // var activeNodeInices = $oldActiveDecl.attr('data-node-indices').split(',');

                    var activeNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, activeObj.nodeIndices);

                    var nodeDistance = getDistanceBetween(activeNode, node, activeObj.uniqueName);
                    if (nodeDistance != 0) {
                        setNewOverridenNode(node.prop, activeObj.important, importantNode, nodeIndex, activeObj, nodeDistance < 0);
                    }
                }
                else {
                    setNewOverridenNode(node.prop, activeObj.important, importantNode, nodeIndex, activeObj, currentNodeParentDistance < activeNodeParentDistance);
                }
            }
            else {
                var parent = pinegrow.pgPostCSSHelper.closestNode(node, 'rule');
                if (parent.type == 'rule') {
                    var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
                    var importantNode = node.important || node.value.endsWith('!important') || false;
                    var parentElId = getRuleElId(parent, pgStylesheet, pgStylePane.getIndex());

                    activeDecl[node.prop] = {
                        activeIndex: 0,
                        nodes: [{
                            distance: 0,
                            nodeIndices: nodeIndices,
                            parentElId: parentElId,
                            important: importantNode,
                            uniqueName: pgStylesheet.getUniqueName()
                        }]
                    }

                    setNodeElActive(parentElId, nodeIndices, true);
                }
            }
        }
    }

    var useTheClosestNode = function (prop, activeDeclObj, activeObj) {
        if (activeDeclObj.nodes.length == 0) return;

        var index = findClosestNodeIndex(activeDeclObj.nodes);
        useNodeAsCurrentNode(prop, index, activeObj);
    }

    var getRuleDistance = function (rule) {
        var pgStylesheet = pinegrow.pgPostCSSHelper.getStylesheetForNode(rule),
            nodeIndices  = pinegrow.pgPostCSSHelper.getNodeIndices(rule);

        return rulesDistance[pgStylesheet.getUniqueName() + '-' + nodeIndices];
    }

    var getNodeParentDistance = function (node) {
        return getRuleDistance(node);
    }

    var getNodeParentDistanceFromIndices = function (uniqueName, nodeIndices) {
        return rulesDistance[uniqueName + '-' + nodeIndices.slice(0, nodeIndices.length-1)];
    }

    var onNodeChanged = function (node, pgStylesheet, oldValue, propName, pgStylePane) {
        if (node == null || node.parent.newNode) return;

        if (node.type == 'decl') {
            if (isInsertedNode(node, pgStylesheet, pgStylePane.getIndex())) return;
            if (propName == "prop") {
                if (node[propName] != oldValue != "" && oldValue != "") {
                    if (node[propName] != "") {
                        var newValue = node[propName];

                        node[propName] = oldValue;
                        onNodeBeforeDeleted(node, pgStylesheet, pgStylePane);
                        node[propName] = newValue;
                        onNodeAdded(node, pgStylesheet, pgStylePane);
                    }
                    return;
                }
            }

            if(oldValue !== undefined && oldValue != node[propName]) {
                node.pgOldValue = {};
                node.pgOldValue[propName] = oldValue;
            }

            var activeDeclObj = activeDecl[node.prop];
            if(!activeDeclObj) return;

            var activeObj = activeDeclObj.nodes[activeDeclObj.activeIndex];

            var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
            var parentElId = getRuleElId(node.parent, pgStylesheet, pgStylePane.getIndex());
            var nodeObjIndexPair = findNodeFromObjs(nodeIndices, activeDeclObj.nodes, pgStylesheet);

            //if (!nodeObjIndexPair) debugger;
            if (!nodeObjIndexPair) return;

            var $declEl = findDeclElFromRuleEl(getNode$el(node.parent, pgStylesheet, pgStylePane.getIndex()), nodeIndices);

            var hadImportant = nodeObjIndexPair.nodeObj.important;
            var importantNode = node.important || node.value.endsWith('!important') || false;
            nodeObjIndexPair.nodeObj.important = importantNode;

            if ($declEl.hasClass('overriden-node')) {
                if (importantNode) {
                    if (activeObj.important) {
                        var activeDeclCS = getStylesheetFromParentId(activeObj.parentElId);

                        // var activeNode = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, activeObj.nodeIndices);
                        var activeNode = pinegrow.pgPostCSSHelper.getNodeByIndices(activeDeclCS.rootNode, activeObj.nodeIndices);
                        var disBetweenNodes = getDistanceBetween(activeNode, node);
                        if (disBetweenNodes > 0) {
                            useNodeAsCurrentNode(node.prop, nodeObjIndexPair.index, activeObj);
                        }
                    }
                    else {
                        useTheClosestNode(node.prop, activeDeclObj, activeObj);
                    }
                }
            }
            else if (activeObj.important) {
                useTheClosestNode(node.prop, activeDeclObj, activeObj);
            }
            else if (hadImportant) {
                useTheClosestNode(node.prop, activeDeclObj, activeObj);
            }
        }
        else if (node.type == 'rule') {
            var nodes = node.nodes
            if (nodes) {
                for (var i = 0; i < nodes.length; i++) {
                    var currNode = nodes[i];

                    onNodeBeforeDeleted(currNode, pgStylesheet, pgStylePane);
                    onNodeAdded(currNode, pgStylesheet, pgStylePane);
                }
            }

            newRuleAdded(node, pgStylesheet, pgStylePane.getIndex(), { operation: "ignore" });
        }
        else if (node.type == 'atrule') {
            if (node.name == 'media' || node.name == 'supports') {
                var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);

                // var nodeSelector  = '.' + pgStylesheet.getUniqueName() + '-container' + extraRuleElId;
                var nodeSelector  = '.' + pgStylesheet.getUniqueName() + '-container' + pgStylePane.getIndex();
                var $parentPath = $(nodeSelector).prev('.parent-path');

                var $mediaEl = $parentPath.find('[data-node-indices="' + nodeIndices + '"]');
                if ($mediaEl.length > 0) {
                    $mediaEl.html(node.params);
                }
            }
        }
    }

    var onNodeBeforeDeleted = function (node, pgStylesheet, pgStylePane) {
        if (node == null || node.type != 'decl') return;

        var activeDeclObj = activeDecl[node.prop];
        if (activeDeclObj) {
            if (toRuleIconShown) toggleConvertToRuleIcon(pgStylesheet, false);

            var activeObj = activeDeclObj.nodes[activeDeclObj.activeIndex];

            var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
            var parentElId = getRuleElId(node.parent, pgStylesheet, pgStylePane.getIndex());
            var nodeObjIndexPair = findNodeFromObjs(nodeIndices, activeDeclObj.nodes, pgStylesheet);

            if (nodeObjIndexPair) {
                activeDeclObj.nodes.splice(nodeObjIndexPair.index, 1);
                updateNodeDistance(activeDeclObj.nodes);

                if (activeDeclObj.nodes.length == 0) {
                    delete activeDecl[node.prop];
                }
                else {
                    useTheClosestNode(node.prop, activeDeclObj, activeObj);
                }
            }
        }
    }

    var onEditFinish = function (node, pgStylesheet, pgStylePane) {
        // on edit finish
    }

    var onNodeReplaced = function (oldNode, newNode, pgStylesheet, pgStylePane) {
        if (!oldNode || !newNode) return;

        if (oldNode.type == "decl") {
            var activeDeclObj = activeDecl[oldNode.prop];

            if (activeDeclObj) {
                var oldNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(oldNode);
                var parentElId = getRuleElId(oldNode.parent, pgStylesheet, pgStylePane.getIndex());
                var nodeObjIndexPair = findNodeFromObjs(oldNodeIndices, activeDeclObj.nodes, pgStylesheet);
                if (nodeObjIndexPair) {
                    if (oldNode.prop != newNode.prop) {
                        onNodeBeforeDeleted(oldNode, pgStylesheet, pgStylePane);
                        onNodeAdded(newNode, pgStylesheet, pgStylePane);
                    }
                }
            }
        }
        else if (oldNode.type == "rule") {
            if (oldNode.nodes) {
                var nodes = oldNode.nodes;
                for (var i = 0; i < nodes.length; i++) {
                    var node = nodes[i];

                    onNodeBeforeDeleted(node, pgStylesheet, pgStylePane);
                    onNodeAdded(node, pgStylesheet, pgStylePane);
                }
            }

            newRuleAdded(newNode, pgStylesheet, pgStylePane.getIndex(), { operation: "ignore" });
        }
    }

    var onRuleIdChanged = function (rule, pgStylesheet, newId, oldId, pgStylePane) {
        if (!rule || !rule.nodes || rule.nodes.length == 0) return;
        var parentElId = getRuleElId(rule, pgStylesheet, pgStylePane.getIndex(), newId);

        for (var i = 0; i < rule.nodes.length; i++) {
            var childNode = rule.nodes[i];

            if (childNode.type == "decl") {
                var activeDeclObj = activeDecl[childNode.prop];
                var childNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(childNode);
                var currParentElId = oldId || getRuleElId(childNode.parent, pgStylesheet, pgStylePane.getIndex());

                if(activeDeclObj) {
                    var nodeObjIndexPair = findNodeFromObjs(childNodeIndices, activeDeclObj.nodes, pgStylesheet);
                    if (nodeObjIndexPair) {
                        nodeObjIndexPair.nodeObj.parentElId = parentElId;
                    }
                }
            }
        }
    }

    var isAllowedInsertNode = function ($li, pgStylePane) {
        return true;
    }

    var addSeeMoreStylesheetCSS = function () {
        var showStylesheet = function (e) {
            e.preventDefault();
            var $target = $(e.target);
            var uniqueName = $target.attr('data-stylesheet-name');
            var sheetItem = moreSheets[uniqueName];

            sheetItem.stylesheet.loadAndShowInActive(function() {
                pinegrow.selectedElements.reselect();
                pinegrow.dispatchEvent('on_css_source_changed');
            });

            /*
            sheetItem.stylesheet.ignore = false;
            sheetItem.stylesheet.reload(function () {
                // TODO Enhance this
                //$('body').trigger('crsa-element-selected');
                pinegrow.selectedElements.reselect();
                pinegrow.dispatchEvent('on_css_source_changed');
                // _this.remove();
                // _this.show();
                // delete moreSheets[uniqueName];
            });

             */
        }

        for (var sheetName in moreSheets) {
            var $listItem = $('<li><a href="#" data-stylesheet-name="' + sheetName + '">Show ' + moreSheets[sheetName].nodesNo + ' more rule' + (moreSheets[sheetName].nodesNo > 1 ? 's' : '') + ' from ' + moreSheets[sheetName].name + '</a></li>').appendTo($sheetList);
            $listItem.find('> a').click(showStylesheet);
        }
    }

    var onMediaParamsClicked = function (e) {
        var target = e.target;
        var $target = $(target);

        if (!$target.is(':focus')) {
            $target.data('old-text', $target.html());
            $target.attr('contenteditable', true);
            crsaSelectText(target);
        }
    }

    var onMediaParamsKeydown = function (e) {
        var $target = $(e.target);

        var code = e.keyCode || e.which;
        if (code == 13 || code == 9) {
            $target.blur();
            return false;
        }
        else if (code == 27) {
            $target.attr('contenteditable', false);
            $target.html($target.data('old-text'));
            return false;
        }
    }

    var onMediaParamsBlured = function (e) {
        var target = e.target;
        var $target = $(target);

        var $rootNodeEl = $target.closest('.parent-path').next('.sass-root-node');
        var rootNode = $rootNodeEl.data('node');

        var pgStylesheet = rootNode.pgStylesheet;
        var node = pinegrow.pgPostCSSHelper.getNodeByIndices(pgStylesheet.rootNode, $target.attr('data-node-indices').split(','));

        node.params = $target.text();

        pinegrow.dispatchEvent('on_css_node_updated', null, {node: node, stylesheet: pgStylesheet});
        $target.removeAttr('contenteditable');

        pgStylesheet.stylePane.recompileCode();
    }

    var onNodeUncollapsed = function () {
        setOverridenDecl();
    }

    var current_selected_rule = null;
    if(pinegrow.activeElementStyleWasSelected === undefined) {
        pinegrow.activeElementStyleWasSelected = false;
    }

    var onRuleSelected = function(node, $li) {
        current_selected_rule = node;
        pinegrow.dispatchEvent('on_css_rule_selected_in_active', null, node, $li);
        pinegrow.activeElementStyleWasSelected = node && node.selector && node.selector == 'Style attribute';
    }

    var convertStyleToRule = function (e) {
        e.preventDefault();
        var $target = $(e.target);
        var $stylesheet = $target.closest('.crsa-stylesheet');
        var cs = $stylesheet.data('cs');
        var rule = cs.getRootNode().nodes[0];
        var elm = cs.element;

        var active_rules = pinegrow.pgActiveRules.getActiveNodes();

        var copyToNewRule = function() {

            fillStylesheetsArray();

            if (filesArr.length > 0) {

                var onCreate = function (elementSelector, selectedFile, mediaParams) {
                    for (var i = 0; i < stylesheets.length; i++) {
                        if (stylesheets[i].url == selectedFile) {
                            var stylesheet = stylesheets[i];

                            var rootNode = stylesheet.getRootNode();
                            var addNodeTo = rootNode;

                            var ruleNode = rule;
                            pinegrow.pgPostCSSHelper.removeNodeWidthRegexFrom("comment", "text", /(pg-uncollapsed:(?:true|false))/, ruleNode);
                            //pinegrow.pgPostCSSHelper.removeNodeWidthRegexFrom("decl", "prop", '--empty-placeholder', ruleNode);
                            var ruleIndices = [rootNode.nodes.length];

                            var parentNode = ruleNode;
                            if (mediaParams) {
                                var lastNode = rootNode.last;
                                if (lastNode.params && lastNode.params == mediaParams) {
                                    addNodeTo = lastNode;
                                    ruleIndices = [rootNode.nodes.length - 1, lastNode.nodes.length];
                                }
                                else {
                                    var mediaNode = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams, null, stylesheet);
                                    //var mediaNode = pinegrow.pgPostCSSHelper.createNode("atrule", "media", mediaParams);
                                    pinegrow.pgPostCSSHelper.insertNodeInside(ruleNode, mediaNode);
                                    parentNode = mediaNode;
                                    ruleIndices.push(0);
                                }
                            }

                            stylesheet.setNodeNeedFormatting(parentNode, true, true);
                            pinegrow.pgPostCSSHelper.insertNodeAt(parentNode, addNodeTo, addNodeTo.nodes.length);

                            stylesheet.setSourceForRules(rootNode, function () {
                                var col = pinegrow.getCollectionForElement(elm);
                                col.forEachElement(function (pgel) {
                                    pgel.removeAttr('style');
                                    pgel.setData('selectedCssRule', pgcssh.getGlobalNodeId(rule));
                                });
                                pinegrow.changeManager.didChange(col, 'Remove style attribute');

                                pinegrow.dispatchEvent('on_css_node_added', null, {node: rule, stylesheet: stylesheet});

                                $.fn.crsa('selectRule', rule, _this);

                                pinegrow.selectedElements.repositionMenusDelayed();
                                //pinegrow.reselectElement();
                            });
                        }
                    }
                }

                var opts = {
                    stylesheets: filesArr,
                    $position: $stylesheet,
                    title: 'Save the inline style into a proper CSS rule'
                }

                var selector_maker = new PgQuickSelectorMaker(elm, null, opts);

                selector_maker.onCreate = function (maker) {
                    var elementSelector = maker.getSelector();
                    var selectedFile = maker.getStylesheet();

                    rule.selector = elementSelector;

                    var col = pinegrow.getCollectionForElement(elm);

                    var add_classes = maker.getAddClasses();
                    if (add_classes.length) {
                        var api = new PgApi();
                        api.addClass(col, add_classes);
                    }

                    onCreate(elementSelector, selectedFile, maker.getMediaQuery());
                }
            }
        }

        var add_to_rules = pinegrow.pgActiveRules.getShownNodes().filter(function(r) {
            var cs = pgcssh.getStylesheetForNode(r);
            if(cs && !cs.canEdit(true)) return false;
            return true;
        })

        if(add_to_rules.length == 0) {
            copyToNewRule();
        } else {
            var actions = [
                {
                    label: 'Create new rule...',
                    action: function() {
                        copyToNewRule();
                    }
                },
                {
                    type: 'divider'
                },
                {
                    type: 'header',
                    label: 'Add to existing rule:'
                }
            ]

            add_to_rules.forEach(function(r) {

                var name = pgcssh.getRuleDisplayName(r);

                name += ' <small>' + pgcssh.getStylesheetForNode(r).name + '</small>';

                actions.push(
                    {
                        label: name,
                        action: function() {
                            var cs = pgcssh.getStylesheetForNode(r);

                            var last_decl = 0;

                            for(var i = 0; i < rule.nodes.length; i++) {
                                if(rule.nodes[i].type == 'decl') {
                                    last_decl = i;
                                }
                            }

                            var col = pinegrow.getCollectionForElement(elm);
                            col.forEachElement(function (pgel) {
                                pgel.removeAttr('style');
                            });

                            for(var i = 0; i < rule.nodes.length; i++) {
                                if(rule.nodes[i].type == 'decl') {
                                    cs.ruleValueChanged(r, rule.nodes[i].prop, rule.nodes[i].value, null, false, 'gui', i == last_decl, null);
                                }
                            }

                            pinegrow.changeManager.didChange(col, 'Remove style attribute');

                            //Needs this to properly update
                            pinegrow.dispatchEvent('on_css_node_updated', null, {node: r, stylesheet: cs});

                            $.fn.crsa('selectRule', r, _this);
                            pinegrow.reselectElement();



                        }
                    }
                )
            })

            var menu = new CrsaContextMenu(actions, $target);
            menu.showForElement($target);
        }
        return false;
    }

    var fillStylesheetsArray = function () {
        var selectedPage = pinegrow.getSelectedPage();
        var pageStylesheets = selectedPage.getStylesheets();
        var allSheets = pageStylesheets.getAllCrsaStylesheets();

        stylesheets = [];
        for (var i = 0; i < allSheets.length; i++) {
            stylesheets = stylesheets.concat(pgStylesheetHelper.getAllStylesheetsFor(allSheets[i]));
        }

        filesArr = [];
        for (var i = 0; i < stylesheets.length; i++) {
            if(stylesheets[i].url && stylesheets[i].canEdit(true)) {
                filesArr.push(stylesheets[i].url);
            }
        }
    }

    var createContainers = function (pgStylesheet, prepend, insertAt) {
        var $stylePaneContainer = null;

        if (insertAt) $stylePaneContainer = $('<div class="crsa-stylesheet"></div>').insertAfter($ruleList.find('> .crsa-stylesheet:nth-child(' + insertAt +')'));
        else if (prepend)  $stylePaneContainer = $('<div class="crsa-stylesheet"></div>').prependTo($ruleList);
        else               $stylePaneContainer = $('<div class="crsa-stylesheet"></div>').appendTo($ruleList);

        var $currRuleList = $('<div class="rule-list active-rule-container"></div>').appendTo($stylePaneContainer);
        $stylePaneContainer.data('cs', pgStylesheet);

        return {
            $stylePaneContainer: $stylePaneContainer,
            $currRuleList: $currRuleList
        }
    }

    var showRulesFor = function (pgStylesheet, node, nodeToCheck, $stylePaneContainer, options) {
        var stylePaneIndex = extraRuleElId + '-' + stylePaneLastIndex;

        options = options || {};
        var allOptions = $.extend({}, {
            'uncollapse': uncollapseAll,
            'created-by': 'active_css',
            'uncollapse-on-click': false,
            'extra-rule-el-id': stylePaneIndex,
            'inversed-insert': true,
            'skip-showing-rule': true,
            'on-node-indices-changed': onNodeIndicesChanged,
            'on-node-added': onNodeAdded,
            'on-node-changed': onNodeChanged,
            'on-before-node-deleted': onNodeBeforeDeleted,
            'on-edit-finish': onEditFinish,
            'on-node-replaced': onNodeReplaced,
            'on-rule-id-changed': onRuleIdChanged,
            'on-node-uncollapsed': onNodeUncollapsed,
            'on-rule-selected': onRuleSelected,
            'is-allowed-insert-node': isAllowedInsertNode,
            'handle-element-selected' : false,
            'show-all': true //show all nodes, avoids dom measure cycle
        }, options);

        allOptions['read-only'] = pgf.kids || false;

        rulesDistance[pgStylesheet.getUniqueName() + '-' + pinegrow.pgPostCSSHelper.getNodeIndices(nodeToCheck)] = currDistance;
        addDeclToActiveDecl(nodeToCheck, pgStylesheet, stylePaneIndex);

        var stylePane = new PgStylesPane(pgStylesheet, $stylePaneContainer, $stylePaneContainer.closest('.rules-list'), [node], allOptions);
        stylePane.show();

        if (node.type == "atrule") {
            stylePane.onlyShowNode(nodeToCheck);
            // var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(nodeToCheck);
            // var $atruleLi = $stylePaneContainer.find('> .rule-list > ul > li.crsa-atrule');
            // var $ul = $atruleLi.find('> ul');
            // $ul.find('> li:not([data-node-indices="' + nodeIndices.join(',') + '"])').remove();
        }
        else {
            stylePane.updateShownNodes();
        }
        stylePaneList.push(stylePane);

        currDistance++;
        stylePaneLastIndex++;

        new PgCSSRuleControls($stylePaneContainer, stylePane, {
            'withoutDargDrop': true,
            'skip-move-item': true,
            'by': 'active_css',
            'skip-enable-item' : pgf.kids || false
        });

        return stylePane;
    }

    var addNodePath = function (pgStylesheet, node, $container, forceAddingPath) {
        var parent = node.parent;
        var i = 1;

        var parentPath = '';
        while (parent && parent.type != "root") {
            if (parent.type == "rule" && parent.selector) {
                parentPath = parent.selector + ' / '  + parentPath;
            }
            else if (i != 1 && parent.type == "atrule" && (parent.name == "media" || parent.name == 'supports')) {
                var parentNodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(parent);
                parentPath = '<span class="media-params" data-node-indices="' + parentNodeIndices.join(',') + '">' + parent.params + '</span>'  + parentPath;
            }
            parent = parent.parent;
            i++;
        }

        parentPath =  pgStylesheet.name + (parentPath.length ? ' / ' : '')  + parentPath;

        if (lastParentPath != parentPath || forceAddingPath) {

            var parent_path_html = parentPath;

            parent_path_html += '<i class="cmd open icon  icon-align-left"></i>';

            if(pgStylesheet.localFile || (!isApp() && pgStylesheet.url)) {
                parent_path_html += '<i class="cmd edit-code icon icon-KODA"></i>';
            }

            parent_path_html += '<i class="cmd ignore icon icon-hide"></i>';

            if(!pgStylesheet.inline) {
                parent_path_html += '<i class="cmd lock icon icon-Lock_closed"></i>';
            }

            var $pathContainer = $('<div class="parent-path">' + parent_path_html + '</div>');

            if(!pgf.kids) {
                $pathContainer.appendTo($container);
            }

            $pathContainer.find('> .media-params')
                .off('click', onMediaParamsClicked)
                .on('click', onMediaParamsClicked)

                .off('keydown', onMediaParamsKeydown)
                .on('keydown', onMediaParamsKeydown)

                .off('blur', onMediaParamsBlured)
                .on('blur', onMediaParamsBlured);

            var $code = $pathContainer.find('.edit-code');
            $code.on('click', function(e) {
                e.preventDefault();
                pinegrow.stylesheets.editCode(pgStylesheet);
            })

            if($code.length) {
                addTooltip($code, 'Edit stylesheet code.')
            }

            var $open = $pathContainer.find('.open');
            $open.on('click', function(e) {
                e.preventDefault();
                if(_this.parentView) {
                    _this.parentView.setActiveCs(pgStylesheet);
                }
            })
            addTooltip($open, 'Open the stylesheet in a separate tab.')

            var $ignore = $pathContainer.find('.ignore');
            $ignore.on('click', function(e) {
                e.preventDefault();
                pgStylesheet.ignoreInActive();
            })
            addTooltip($ignore, 'Ignore the stylesheet in Active tab.')

            if(!pgStylesheet.inline) {
                var $lock = $pathContainer.find('.lock');
                $lock.on('click', function (e) {
                    e.preventDefault();
                    if (pgStylesheet.canEdit(true)) {
                        pgStylesheet.setLocked(true, '');
                        pinegrow.selectedElements.reselect();
                        pinegrow.showQuickMessage('Stylesheet is locked.');
                    } else {
                        pgStylesheet.unlockWithAsk();
                    }
                })
                if (pgStylesheet.canEdit(true)) {
                    addTooltip($lock, 'Click to lock the stylesheet.')
                } else {
                    $lock.addClass('pg-css-locked icon-Lock_open').removeClass('icon-Lock_closed');
                    addTooltip($lock, `The stylesheet is locked. ${pgStylesheet.lockedReason || ''} Click to unlock.`)
                }
            }



            lastParentPath = parentPath;
        }
    }

    this.updateLayout = function () {
        for (var i = 0; i < stylePaneList.length; i++) {
            stylePaneList[i].updateShownNodes();
        }
    }

    var shown_nodes = [];

    // NOT USED
    this.didMatchedRulesChange = function() {
        //var seletedDocument = getDocumentForSelectedPage();
        //var win = seletedDocument.get(0).defaultView;
        //var el = selectedElement.get$DOMElement().get(0);
        // var matched_rules = crsaGetFreshMatchedRules();
        var getSelectedElement = function () {
            var seletedDocument = getDocumentForSelectedPage();
            var win = seletedDocument.get(0).defaultView;
            var $el = selectedElement.get$DOMElement();
            if (!win || !$el) return null;
            return {
                'win': win,
                'el': $el.get(0)
            }
        }


        var obj = getSelectedElement();
        if(!obj) {
            current = [];
            return true;
        }
        var matched_rules = crsaGetFreshMatchedRules(obj.win, obj.el);
        matched_rules = crsaCombinedTwoCSSList(crsaGetFreshMatchedRules(obj.win, obj.el, 'after'),  matched_rules);
        matched_rules = crsaCombinedTwoCSSList(crsaGetFreshMatchedRules(obj.win, obj.el, 'before'), matched_rules);

        if(!current_matched_rules) return true;

        var current = [];
        for(var i = 0; i < current_matched_rules.length; i++) {
            current.push(current_matched_rules[i]);
        }

        for(var i = 0; i < matched_rules.length; i++) {
            var idx = current.indexOf(matched_rules[i]);
            if(idx < 0) {
                return true;
            }
            current.splice(idx, 1);
        }
        if(current.length == 0) {
            return false;
        } else {
            return true;
        }
    }

    var current_matched_rules = [];
    var active_out_dated = false;

    this.setActiveOutDated = function() {
        active_out_dated = true;
    }

    this.show = function (select_rule, rule_list, pgel_selected_css_rule) {

        if(!rememberRuleForElement && pgel_selected_css_rule !== 'none') pgel_selected_css_rule = null;

        $ruleList.html('');
        $sheetList.html('');
        lastParentPath = '';

        shown_nodes = [];
        stylePaneLastIndex = 1;

        //var seletedDocument = getDocumentForSelectedPage();
        //var win = seletedDocument.get(0).defaultView;
        //var el = selectedElement.get$DOMElement().get(0);

        if(!selectedElement) return;

        var isNodeIsPlaceholderNode = function (nodes) {
            /* if (nodes.length == 1) {
             var oneNode = nodes[0];
             if (oneNode.type == 'decl' && oneNode.prop == '--empty-placeholder') return true;
             }   */
            return false;
        }

        var nodes_to_show = null;

        var doMatched = function () {
            for (var i = 0; i < nodes_to_show.length; i++) {
                var node_to_show = nodes_to_show[i];

                var node = node_to_show.node;
                var pgStylesheet = node_to_show.pgStylesheet;

                shown_nodes.push(node);

                var containers = createContainers(pgStylesheet);
                addNodePath(pgStylesheet, node, containers.$currRuleList);

                var moreOptions = {
                    'uncollapse-all': uncollapseAll,
                    'uncollapse': uncollapseAll
                };

                var nodeToDisplay = node;
                var parent = node.parent;
                if (parent && parent.type === "atrule" && (parent.name === "media" || parent.name === 'supports')) {
                    nodeToDisplay = parent;
                    moreOptions.uncollapse = true;
                }

                var $stylePaneContainer = containers.$stylePaneContainer;
                var stylePane = showRulesFor(pgStylesheet, nodeToDisplay, node, $stylePaneContainer, moreOptions);
            }
        }

        var doNonElementStyleStuff = function() {
            doMatched();

            setOverridenDecl();
            addSeeMoreStylesheetCSS();

            if(pgel_selected_css_rule === 'none') {
                return true;
            }

            var selectFirst = function() {
                if(shown_nodes.length == 1) {
                    selectRule(shown_nodes[0])
                } else if(shown_nodes[0].nodes.length) {
                    selectRule(shown_nodes[0])
                } else if(pinegrow.activeElementStyleWasSelected) {
                    selectRule(shown_nodes[0])
                } else if(shown_nodes[1].selector && shown_nodes[1].selector.startsWith('*')) {
                    selectRule(shown_nodes[0])
                } else {
                    selectRule(shown_nodes[1])
                }
            }

            if(shown_nodes.length) {
                if(pgel_selected_css_rule === 'inline') {
                    selectRule(shown_nodes[0]);

                } else if(pgel_selected_css_rule && shown_nodes.indexOf(pgel_selected_css_rule) >= 0 && !pgel_selected_css_rule.selector.startsWith('*')) {
                    //select the rule that was stored in pgel as last selected for that pgel
                    selectRule(pgel_selected_css_rule);
                } else if (select_rule) {
                    var r = select_rule.data || select_rule;
                    if (shown_nodes.indexOf(r) >= 0) {
                        selectRule(r);
                    } else {
                        selectFirst();
                        //selectRule(shown_nodes[0]);
                    }
                } else {
                    selectFirst();
                    //selectRule(shown_nodes[0]);
                }
                return true; //rule selected
            }
            return false; //rule not selected
        }

        var callback = function (active_rules) {
            moreSheets = active_rules.more_sheets;
            current_matched_rules = active_rules.matched_rules;

            nodes_to_show = active_rules.nodes_to_show;

            fillStylesheetsArray();

            updateClassesWithoutRules();

            if (!rule_list) {

                elementPgStylesheet = PgCSSStylesheet();
                elementPgStylesheet.elementStyle = true
                elementPgStylesheet.element = selectedElement;
                elementPgStylesheet.index = 'style';

                elementPgStylesheet.load(null, function () {
                    if (!elementPgStylesheet) {
                        //Object was removed while it was loading. Nothing to do, bye! Hope no leaks are left.
                        return;
                    }
                    var containers = createContainers(elementPgStylesheet, true);
                    var $stylePaneContainer = containers.$stylePaneContainer;

                    $stylePaneContainer.addClass('style-attribute-style-pane');

                    $toRuleIcon = $('<i class="icon icon-plus-square to-rule-icon"></i>');
                    $toRuleIcon.appendTo($stylePaneContainer).hide();

                    $toRuleIcon.on('click', convertStyleToRule);

                    addTooltip($toRuleIcon, 'Save inline style to a proper CSS rule.');

                    var all = elementPgStylesheet.getAllRules();

                    if (all && all.length) {
                        var node = all[0];

                        shown_nodes.push(node);
                        if (node && node.nodes && node.nodes.length > 0) {
                            if (!isNodeIsPlaceholderNode(node.nodes)) {
                                $toRuleIcon.show();
                                toRuleIconShown = true;
                                $toRuleIcon.closest('.style-attribute-style-pane').addClass('has-inline-style');
                                createRuleView.setSaveInline(true);
                            }
                        }

                        var stylePane = showRulesFor(elementPgStylesheet, node, node, $stylePaneContainer, {
                            uncollapse: pgf.kids ? true : true, //(selectedElement.getAttribute('style') || '').length > 0,
                            auth: {
                                can_add_rule: false,
                                can_edit_rule: false
                            }
                        });

                        $stylePaneContainer.find('li.crsa-rule').addClass('selected');

                        if (!doNonElementStyleStuff()) {
                            //rule was not selected
                            selectRule(node);
                        }
                    }
                });
            } else {
                doNonElementStyleStuff();
            }
        }

        activeDecl = {};
        if (active_out_dated) {
            active_out_dated = false;
            pinegrow.pgActiveRules.updateMatchedRules(true, null, {rule_list: rule_list});
        }
        else {
            var active_rules = pinegrow.pgActiveRules.getActiveNodes();
            if (active_rules) {
                callback({
                    more_sheets: active_rules.more_nodes,
                    matched_rules: active_rules.matched_rules,
                    nodes_to_show: active_rules.nodes_to_show
                });
            }
        }
        // else {
        //     pinegrow.stylesheets.getAllMatchedPgRules(selectedElement, rule_list, function (nodes_obj) {
        //         callback({
        //             more_sheets: nodes_obj.more_sheets,
        //             matched_rules: nodes_obj.matched_rules,
        //             nodes_to_show: nodes_obj.nodes_to_show
        //         });
        //         active_out_dated = false;
        //     });
        // }
    }

    this.removeStylesheetObj = function() {
        if (elementPgStylesheet) {
            elementPgStylesheet.remove();
            elementPgStylesheet = null;
        }
    }

    this.remove = function () {
        this.removeStylesheetObj();

        if(stylePaneList) {
            for (var i = 0; i < stylePaneList.length; i++) {
                stylePaneList[i].remove();
                stylePaneList[i] = null;
            }

            stylePaneList = null;
        }

        if(createRuleView) {
            createRuleView.view.destroy();
            createRuleView = null;
        }
        stylesheets = [];
        stylePaneList = [];
        activeDecl = {};
        moreSheets = {};

        _this = null;


        $toRuleIcon = null;
        $ruleList.html('');
        $sheetList.html('');

        activeEditorRemoved = true;
    }
}
