var PgCSSRuleControls = function ($container, stylePane, options) {
    // options
    if (!options) options = {};
    options['$scrollParent'] = options['$scrollParent'] || null;
    options['withoutDargDrop'] = options['withoutDargDrop'] == true ? true : false;
    options['by'] = options['by'] || '';

    options['skip-move-item']   = options['skip-move-item']   || false;
    options['skip-delete-item'] = options['skip-delete-item'] || false;
    options['skip-code-edit']   = options['skip-code-edit']   || false;
    options['skip-enable-item'] = options['skip-enable-item'] || false;

    // Adding elements
    var rule_menu_HTML = '<div class="rule-menu">';
    if (!options['skip-enable-item']) rule_menu_HTML += '<i class="rule-menu-icon icon icon-see crsa-cm-enable"></i>';
    if (!options['skip-code-edit']) rule_menu_HTML += '<i class="rule-menu-icon icon icon-KODA crsa-cm-code"></i>';
    if (!options['skip-delete-item']) rule_menu_HTML += '<i class="rule-menu-icon icon icon-bin crsa-cm-remove"></i>';
    if (!options['skip-move-item']) rule_menu_HTML += '<i class="rule-menu-icon icon icon-Pin6 crsa-cm-move"></i>';

    rule_menu_HTML += '<i class="rule-menu-icon menuicon icon icon-repeat-Y"></i></div>';

    var $rule_menu = $(rule_menu_HTML);

    if (options.withoutDargDrop) {
        $container.addClass('not-dragable');
    }

    // dragging vars
    var dragged_$target = null;
    var dragged_rule = null;
    var dragged_$el = null;
    var dragged_$placeholder = null;
    var dragged_insert_after_rule = null;

    var editRule = function(r) {
        // var rule_obj = getObjectFromRule(r);
        //if(stylePane.pgStylesheet && !stylePane.pgStylesheet.canEdit()) return;
        $.fn.crsa('selectRule', r);
    }

    var getRuleForDiv = function($li, cs) {
        if (!$li.attr('data-node-indices')) return null;
        if (!cs) cs = $li.closest('.crsa-stylesheet').data('cs');
        return pinegrow.pgPostCSSHelper.getNodeByIndices(cs.getRootNode(), $li.attr('data-node-indices').split(','));
    }

    var getPreviousRule = function($el) {
        var $prev = $el.prev();
        if ($prev.length) {
            if($prev.is('.crsa-node')) {
                return getRuleForDiv($prev);
            }
        } else {
            return 'first';
        }
    }

    var editRuleSource = function (rule, $target) {
        rule = pgcssh.getRuleForNode(rule);
        var q = new PgQuickCSSRuleCodeEdit(rule, $target);

        if(!stylePane.pgStylesheet.canEdit(true)) {
            pinegrow.showQuickError('Opened in read-only mode because the stylesheet is locked.');
        }
        /*
        var nodeIndices = pinegrow.pgPostCSSHelper.getNodeIndices(rule),
                     cs = pgStylesheetHelper.getStylesheetFromNode(rule);
        $.fn.crsa('editRuleSource', { nodeIndices: nodeIndices, pgStylesheet: cs });
        */
    }

    var showContextMenuForRule = function($li, only_classes, event) {

        if(pgf.kids) return;

        var node = getRuleForDiv($li);

        var locked = !stylePane.pgStylesheet.canEdit(true);

        var cs = pgcssh.getStylesheetForNode(node);
        var r = pgcssh.getRuleForNode(node);

        var $menu = new CrsaContextMenu(null, $li);

        var createCSSelector = function($li, suffix) {
            suffix = suffix || '';
            var options = pinegrow.stylesheets.getAllAsSelectOptions(null, true);
            //options.unshift({name: '[current stylesheet]', key:'current'})

            var cslist = pinegrow.getSelectedPage().getAttachedStylesheets(
                {
                    loaded: false,
                    ignored: true,
                    imported: true,
                    inline: false,
                    locked: false
                })

            var options = cslist.map(function (cs) {
                var short = crsaGetNameFromUrl(cs.url)
                return {name: short + (cs.locked ? ' &#128274;' : ''), key: cs.url};
            })

            var in_input = new PgInputField('stylesheet', {
                type: 'select',
                show_empty: true,
                name: 'in',
                without_text: false,
                system_field: true,
                default_value: pgcssh.getUrlOfStylesheetForOverrides(null, suffix),
                options: options
            });

            $li.addClass('has-field');
            in_input.$field.addClass('field-in-menu');

            in_input.appendTo($li);

            //if(options.stylesheet_value) {
            //  in_input.setValue(options.stylesheet_value)
            //}

            in_input.on('change', function(e) {
                var v = in_input.getValue();
                if(v) {
                    var cs = pinegrow.stylesheets.getByUrl(v);
                    if(cs && !cs.canEdit(true)) {
                        pinegrow.showQuickError('The selected stylesheet is locked.');
                        in_input.setValue('');
                        return;
                    }
                }
                pgcssh.setUrlOfStylesheetForOverrides(v, null, suffix);
                if(v) {
                    //$li.find('>a').trigger('click');
                }
            })
        }

        var withSelectedCS = function(func, suffix) {
            suffix = suffix || '';
            var cs;
            var csurl = pgcssh.getUrlOfStylesheetForOverrides(null, suffix);
            if(!csurl) {
                pinegrow.showQuickMessage('Please select the destination stylesheet.', 4000, false, 'error');
                return;
            }
            if (csurl == 'current') {
                cs = stylePane.pgStylesheet;
            } else {
                cs = pinegrow.stylesheets.getByUrl(csurl);
            }
            if(cs) {
                if(!cs.canEdit()) return;

                cs.ignore = false;
                if(cs.loaded) {
                    func(cs);
                } else {
                    cs.load(cs.url, function() {
                        func(cs);
                    }, true /* force loading ignored cs */);
                }
            } else {
                pinegrow.showAlert('Stylesheet with ' + csurl + ' not found or not loaded!', 'Error');
            }
        }

        var se = pinegrow.getSelectedElement();

        var getSelectableRule = function(node) {
            if(node.type == 'rule') return node;

            for(var i = 0; i < node.nodes.length; i++) {
                var rule = getSelectableRule(node.nodes[i]);
                if(rule) return rule;
            }
            return null;
        }

        var duplicate = function(cs, r, node, pos) {
            if(!cs.canEdit()) return;
            if(!stylePane) return;
            node = node || r;

            if (pgcssh.isMedia(node.parent)) node = node.parent;
            var position = pos || (cs === stylePane.pgStylesheet ? node : 'append');
            var api = new PgApi();
            api.addCSSRuleToStylesheet(r, cs, position, {
                action: 'Duplicate ' + pgcssh.getNodeDisplayName(node)
            }, function(error, new_nodes) {
                if(!error) {
                    if(new_nodes.length) {
                        var rule = getSelectableRule(new_nodes[0]);

                        var show_cs = true;
                        if(!stylePane || !stylePane.pgStylesheet) {
                            //if(rule && pinegrow.pgActiveRules.isRuleShown(rule)) {
                                show_cs = false;
                           // }
                        }
                        if(show_cs &&  cs !== stylePane.pgStylesheet) {
                            //show this cs
                            pinegrow.activeCSSView.setActiveCs(cs);
                        }
                        if(rule) {
                            editRule(rule);
                        }
                    }
                }
            })
        }

        var override = function(cs, r) {
            if(!stylePane) return;
            if(!cs.canEdit()) return;
            var node = r;
            if (pgcssh.isMedia(node.parent)) node = node.parent;

            var api = new PgApi();
            var code = pgcssh.getSelectorWithoutPgId(r.selector) + ' {}';

            var position = cs === stylePane.pgStylesheet ? node : 'append';
            api.addCSSRuleToStylesheet(code, cs, position, {
                action: 'Override ' + pgcssh.getNodeDisplayName(r)
            }, function (error, new_nodes) {
                if (!error) {
                    if (new_nodes.length) {
                        var rule = getSelectableRule(new_nodes[0]);

                        var show_cs = true;
                        if(!stylePane || !stylePane.pgStylesheet) {
                            //if(rule && pinegrow.pgActiveRules.isRuleShown(rule)) {
                            show_cs = false;
                            // }
                        }
                        if(show_cs &&  cs !== stylePane.pgStylesheet) {
                            //show this cs
                            pinegrow.activeCSSView.setActiveCs(cs);
                        }
                        if(rule) {
                            editRule(rule);
                        }
                    }
                }
            })
        }

        var duplicateNonRule = function (cs, node, position) {
            var rule_code = cs.getSourceForRules(node);
            rule_code = cs.filterSource(rule_code, true);

            duplicate(cs, rule_code, node, position);
        }

        if(locked) {
            $menu.add({
                label: 'This stylesheet is locked.',
                action: function() {}
            })
        }
        if(r) {

            if (!only_classes) {
                if (r.selector && !locked) {
                    $menu.add("Edit...", null, function () {
                        editRule(r);
                    });
                }
                $menu.add("Edit code...", null, function () {
                    editRuleSource(r, $li);
                });

                if(cs && cs.url) {
                    $menu.add("Reveal in code", null, function () {
                        pinegrow.revealCSSRuleInCodeEditor(r)
                    });
                }

            }
        }

        var isVariable = pgcssh.isVariable;

        if (pgcssh.getNodeType(node) == 'decl' && r == node.parent) {
            $menu.add("", null, null, 'divider');
            $menu.add("!important", null, function () {
                if (cs.canEdit(false, false)) {
                    var oldValue = node.value;
                    var value = oldValue;
                    if (value.endsWith('!important') || node.important) {
                        node.value = value.replace(/\s*!important/, '');
                        node.important = false;
                    } else {
                        node.important = true;
                    }

                    var nodeIndices = pgcssh.getNodeIndices(node);
                    var rootNode = pgcssh.getRootNodeFor(node);

                    cs.setSourceForRules(rootNode, function () {
                        var updatedNode = pgcssh.getNodeByIndices(cs.getRootNode(), nodeIndices);
                        pinegrow.dispatchEvent('on_css_node_updated', null, {
                            node: updatedNode,
                            stylesheet: cs,
                            edited_by: 'context_menu',
                            oldValue: pgcssh.getNodeOldValues(updatedNode),
                            undoCombineKey: 'ruleValueChanged_' + updatedNode.prop
                        });
                    }, false, true /* fastDelayed */ );
                }
            }, null, null, node.value.endsWith('!important') || node.important);
        }

        if (r) {
            $menu.add("", null, null, 'divider');
            $menu.add("Duplicate the whole rule", null, null, 'header');

            if(!locked) {
                $menu.add("Duplicate", null, function () {
                    duplicate(stylePane.pgStylesheet, r);
                });
            }

            $menu.add("<span>Duplicate</span>", null, function () {
                withSelectedCS(function (cs) {
                    duplicate(cs, r);
                })
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li);
            });
        } else if (!isVariable(node)) {
            $menu.add("", null, null, 'divider');

            var cs_suffix = '-media';
            $menu.add("<span>Duplicate</span>", null, function () {
                withSelectedCS(function (cs) {
                    duplicateNonRule(cs, node);
                }, cs_suffix)
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li, cs_suffix);
            });
        }
        else {
            $menu.add("", null, null, 'divider');
            //$menu.add("Duplicate the whole rule", null, null, 'header');

            if(!locked) {
                $menu.add("Duplicate", null, function () {
                    duplicateNonRule(stylePane.pgStylesheet, node);
                });
            }

            var isVarNode = isVariable(node);

            var label = isVarNode ? 'Customize' : 'Duplicate';
            var cs_suffix = isVarNode ? '-vars' : '';

            $menu.add("<span>" + label + "</span>", null, function () {
                withSelectedCS(function (cs) {
                    var api = new PgApi();

                    api.overrideCSSVariable(node, cs, function(new_nodes) {
                        if(new_nodes.length) {
                            if(!stylePane || cs !== stylePane.pgStylesheet) {
                                //show this cs
                                pinegrow.activeCSSView.setActiveCs(cs);
                            }
                            editRule(new_nodes.pop());
                        }
                    })
                }, cs_suffix)
            }, null, null, null, function ($li) {
                //onShow
                createCSSelector($li,  cs_suffix);
            });
        }

        if (pgcssh.getNodeType(node) == 'decl' || pgcssh.getNodeType(node) == 'atrule') {

            if (cs.isCompilableStylesheet() && !locked) {
                // var fr_lvl_nodes_obj = crsaGetAllUsedNodes(cs, node, true);
                var fr_lvl_nodes_obj = cs.smartGetAllUsedNodes(node, true);
                var fr_lvl_nodes = fr_lvl_nodes_obj.nodes;

                if (fr_lvl_nodes.length > 0) {
                    $menu.add("", null, null, 'divider');
                    $menu.add("Edit variables", null, null, 'header');

                    for (var i = 0; i < fr_lvl_nodes.length; i++) {
                        var curr_node = fr_lvl_nodes[i];

                        var prop = pgcssh.getVariableProp(curr_node)
                        $menu.add("Edit " + prop, null, function (pgel, e, $li, data) {
                            if (!data || !data.node) return;

                            var element_node = data.node;
                            var $element = $('<div class="var-dialog-content"></div>');
                            var view = new PgVarDialogView($element);

                            var nodeStylesheet = pgcssh.getStylesheetForNode(element_node);

                            var parents = nodeStylesheet.getMainParnetStylesheets();
                            var nodes = [];

                            parents.forEach(function(pcs) {
                                var list = pcs.smartGetAllUsedNodes(element_node);
                                if(list) {
                                    nodes = nodes.concat(list.nodes);
                                }
                            })

                            //console.log(nodes);

                            // var used_nodes = crsaGetAllUsedNodes(nodeStylesheet, curr_node);
                            //var used_nodes = nodeStylesheet.smartGetAllUsedNodes(element_node);
                            //var nodes = used_nodes.nodes;

                            var varsPane = new PgStylesheetVars(nodeStylesheet, $element, {
                                useInsertBorder: false
                            });
                            varsPane.show(nodes);

                            view.setStylesheetVar(varsPane);
                            var win = new PgQuickWindow('Edit ' + pgcssh.getVariableProp(element_node), view, 'vars-editor', 380, 400);
                            // win.showAtRandomPosition();
                            win.showFor$El($li);

                            var onCssSourceChanged = function (p, info) {
                                if (info) {
                                    var list = info.list;
                                    for (var i = 0; i < list.length; i++) {
                                        var cslist = list[i];
                                        if (varsPane.isPgStylesheetExists(cslist)) {
                                            win.close();
                                            break;
                                        }
                                    }
                                }
                            }

                            pinegrow.addEventHandler('on_css_source_changed', onCssSourceChanged);

                            win.onDestroy = function() {
                                pinegrow.removeEventHandler('on_css_source_changed', onCssSourceChanged);
                            }

                        }, null, {node: curr_node});
                    }
                }
            }

        }

        if(r && r.selector) {

            $menu.add("", null, null, 'divider');
            $menu.add("Create empty rule with this selector", null, null, 'header');

            if(!locked) {
                $menu.add("<span>Override</span>", null, function () {
                    override(stylePane.pgStylesheet, r);
                }, null, null, null, function ($li) {
                });
            }

            $menu.add("<span>Override</span>", null, function () {
                withSelectedCS(function(cs) {
                    override(cs, r);
                })
            }, null, null, null, function($li) {
                //onShow
                createCSSelector($li);
            });
        }

        if(!locked) {
            $menu.add("", null, null, 'divider');
            $menu.add("Delete", null, function () {
                if (!stylePane.pgStylesheet.canEdit(false, true)) return;
                stylePane.removeNode($li);
            });
        }

        if(r) {
            var classes = r.selector ? pgcssh.getClassesFromSelector(r.selector) : [];

            if (classes.length) {
                if (se) {
                    var remove = [];
                    var add = [];

                    $.each(classes, function (i, cls) {
                        if (se.hasClass(cls)) {
                            remove.push(cls);
                        } else {
                            add.push(cls);
                        }
                    })

                    if (add.length) {
                        $menu.add("", null, null, 'divider');
                        $menu.add("Add class to&nbsp;&nbsp;<b>" + pinegrow.selectedElements.getName() + "</b>", null, null, 'header');

                        add.forEach(function (cls) {
                            $menu.add('+ ' + cls, null, function () {
                                var api = new PgApi();
                                api.addClass(null, cls)
                            });
                        })
                    }

                    var all = se.getClasses();
                    var other = [];
                    /*
                     if (all) {
                     $.each(all, function(i, cls) {
                     if(remove.indexOf(cls) < 0) {
                     other.push(cls);
                     }
                     });
                     }*/

                    if (remove.length > 0 || other.length > 0) {
                        $menu.add("", null, null, 'divider');
                        $menu.add("Remove class from&nbsp;&nbsp;<b>" + pinegrow.selectedElements.getName() + "</b>", null, null, 'header');
                        $.each(remove, function (i, cls) {
                            $menu.add('- ' + cls, null, function () {
                                var api = new PgApi();
                                api.removeClass(null, cls)
                            });
                        })
                        $.each(other, function (i, cls) {
                            if (cls.indexOf('crsa-') != 0) {
                                $menu.add('- ' + cls, null, function () {
                                    var api = new PgApi();
                                    api.removeClass(null, cls)
                                });
                            }
                        })
                    }
                } else {
                    $menu.add("Select element first, then assign class to it.", null, null, 'header');
                }
            }
        }

        if (!locked && r && r.nodes != undefined) {
            $menu.add("", null, null, 'divider');
            $menu.add("Add rule inside...", null, function() {
                //if(!stylePane.pgStylesheet.canEdit()) return;
                stylePane.uncollapseLi($li);
                stylePane.addNewNodePlaceholderTo($li, "rule");
            });
            $menu.add("Add rule after...", null, function() {
                //if(!stylePane.pgStylesheet.canEdit()) return;
                stylePane.addNewNodePlaceholderAfter($li, "rule");
            });
            $menu.add("Add rule before...", null, function() {
                //if(!stylePane.pgStylesheet.canEdit()) return;
                stylePane.addNewNodePlaceholderBefore($li, "rule");
            });
        }

        var $menu_ul = $menu.showAt(event.pageX, event.pageY, $li.closest('.crsa-cm-rules'));
        return $menu_ul;
    }

    var contextmenuTriggered = function(event) {
        var $target = $(event.target);
        if ($target.closest('.has-options').length) {
            var $li = $target.closest('.has-options');
            showContextMenuForRule($li, false, event);
            event.stopPropagation();
        }

        event.preventDefault();
    }

    var containerClicked = function(event) {
        var stop = false;
        var $target = $(event.target);

        pgPreview.hide();

        if ($target.closest('.crsa-cm-code').length) {
            var $li = $target.closest('li');
            var r = getRuleForDiv($li);

            if(pgf.kids) {
                pinegrow.revealCSSRuleInCodeEditor(r);
            } else {
                editRuleSource(r, $li);
            }
            stop = true;
        } else if ($target.closest('.crsa-cm-remove').length) {
            var $li = $target.closest('.has-options');
            // var r = getRuleForDiv($li);
            // var cs = pinegrow.pgPostCSSHelper.getRootNodeFor(r).pgStylesheet;

            // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Remove CSS rule", false, cs);
            if(stylePane.pgStylesheet.canEdit(false, true)) {
                stylePane.removeNode($li);
            }

            stop = true;
        } else if ($target.closest('.crsa-cm-enable').length) {
            var $closest = $target.closest('.crsa-cm-enable');
            var ignore = false;
            if ($closest.hasClass('icon-see')) ignore = true;

            if(stylePane.pgStylesheet.canEdit()) {

                stylePane.toggleCodeIgnoring($closest, ignore, function (changed) {
                    if (changed) {
                        if (ignore) {
                            $closest.removeClass('icon-see').addClass('icon-hide');
                        }
                        else {
                            $closest.removeClass('icon-hide').addClass('icon-see');
                        }
                    }
                    // pinegrow.dispatchEvent('on_css_source_changed', null, { list: [cs], by: options.by });
                });
            }
            stop = true;
        }

        if (stop) {
            event.preventDefault();
            event.stopPropagation();
        }
    }

    $container.get(0).addEventListener("contextmenu", contextmenuTriggered, true);

    $container.get(0).addEventListener('click', containerClicked, true);

    $container.on('mouseenter', '.has-options > div', function (event) {

        var $target = $(event.target);

        var $li = $target.closest('.has-options');

        if(!pgf.kids) {
            $li.closest('li.hoverd').removeClass('hoverd');
            $li.addClass('hoverd');
        }
        var $eyeIcon = $rule_menu.find('.icon-see, .icon-hide');
        var r = getRuleForDiv($li);
        if (!r) return;
        var cs = pgStylesheetHelper.getStylesheetFromNode(r);

        if ($li.hasClass('crsa-rule')) {
            if (r) {
                if (r.selector) {
                    try {
                        var $els = pinegrow.getSelectedPage().activeView.get$Body().parent().find(cs.getRuleSelector(r)).not('[class*="pinegrow-ui-helper"]');

                        $.fn.crsa('highlightElement', $els);
                    } catch(err) {
                        // do nothing
                    }
                }
            }
        }

        if ($li.hasClass('has-options')) {
            if (r) {
                var commented = pinegrow.pgPostCSSHelper.isCommented(r);
                if (commented) $eyeIcon.removeClass('icon-see').addClass('icon-hide');
                else $eyeIcon.removeClass('icon-hide').addClass('icon-see');
            }
            $rule_menu.detach();
            var $div = $li.find('> div').not('.pg-tree-quick-insert-line');
            if($div.length) {
                $($div.get(0)).append($rule_menu);
            }
        }
    });

    $container.on('mouseleave', '.has-options > div', function (event) {
        var $target = $(event.target);

        var $li = $target.closest('.has-options');
        $li.removeClass('hoverd');
        if ($li.hasClass('has-options')) {
            $rule_menu.detach();
        }
        $.fn.crsa('highlightElement', null);
    });

    $container.on('mouseover', '.crsa-cm-code', function(event) {
        return;//

        var $target = $(event.target);
        var $li = $target.closest('.has-options');

        if ($li.hasClass('has-options')) {
            var r = getRuleForDiv($li);

            if (r) {
                var cs = pgStylesheetHelper.getStylesheetFromNode(r);
                var $el = $('<div/>');
                var $pre = $('<pre/>').html(cs.filterSource(cs.getHtmlSourceForRule(r))).appendTo($el);
                $('<p>Click on <i class="fa fa-code"></i> to edit rule code.<br />Click on <span>.Selector</span> to edit rule properties.</p>').appendTo($el);

                pgPreview.show($li, $el, 'cm-preview');

                /*$.fn.crsa('showPreview', $li, $el, 'cm-preview', function(w) {
                    return getPreviewPosition(w, $li.closest('#crsa-left-plane, .panel,.lm_content'));
                });*/
            }
        }
    });

    $container.on('mouseout', '.crsa-cm-code', function(event) {
        return;

        var $target = $(event.target);
        var $li = $target.closest('.has-options');

        if ($li.hasClass('has-options')) {
            pgPreview.hide();
        }
    });

    var containerMousedownTrigger = function(event) {
        dragged_$target = $(event.target);
    }

    $container.get(0).addEventListener('mousedown', containerMousedownTrigger);

    if (!options.withoutDargDrop /* && stylePane.pgStylesheet.canEdit(true)*/) {
        $container.get(0).addEventListener('dragstart', function(event) {
            var $target = $(event.target);

            if (dragged_$target.is('.crsa-cm-move')) {
                var pgStylesheet = $target.closest('.crsa-stylesheet').data('cs');
                // getCrsaPageForIframe($.fn.crsa('getSelectedPage')).undoStack.add("Rearrenge CSS rule", false, pgStylesheet);

                if ($target.is('.crsa-node')) {
                    dragged_rule = getRuleForDiv($target);
                    dragged_$el = $target;
                }

                setTimeout(function() {
                    dragged_$el.css('opacity', 0.4).hide();
                }, 100);

                $.fn.crsa('hidePreview')
            }
        }, true);

        $container.get(0).addEventListener('dragover', function(event) {
            if (dragged_$target && dragged_$target.is('.crsa-cm-move')) {
                var $target = $(event.target);

                if ($target.is('.crsa-cm-placeholder')) {
                    var $p = $target.closest('.crsa-atrule');
                    if ($p.is('.crsa-atrule') && $target.index() == $p.find('> div').children().length - 1) {

                        if (event.offsetX < 25) {
                            dragged_$placeholder.insertAfter($p);
                        } else {
                            dragged_$placeholder.appendTo($p.find('> div'));
                        }
                    }
                } else if ($target.closest('.crsa-node').length) {
                    $target = $target.closest('.crsa-node');

                    if ($target.get(0) != dragged_$el.get(0)) {
                        if (!dragged_$placeholder) {
                            dragged_$placeholder = $('<div class="crsa-cm-placeholder"></div>');
                        }

                        var h = $target.height();
                        if(event.offsetY < h / 2) {
                            dragged_$placeholder.insertBefore($target);
                            dragged_insert_after_rule = getPreviousRule($target);
                        } else {
                            dragged_$placeholder.insertAfter($target);
                            dragged_insert_after_rule = getRuleForDiv($target);
                        }
                    }
                }
                event.preventDefault();
            }
        }, true);

        $container.get(0).addEventListener('dragleave', function(event) {
            var $target = $(event.target);

            if ($target.closest('.crsa-cm-placeholder').length) {
                // do nothing
            } else if ($target.closest('.crsa-node').length) {
                // do nothing
            }
        }, true);

        var eventDragend = function(event) {
            if (dragged_$target && dragged_$target.is('.crsa-cm-move')) {
                var $target = $(event.target);

                if ($target.closest('.crsa-node').length) {
                    if (dragged_$placeholder != null) {
                        var cs = $target.closest('.crsa-stylesheet').data('cs');
                        var $el = dragged_$placeholder.prev();

                        var newCS = dragged_$placeholder.closest('.crsa-stylesheet').data('cs');

                        var addedToNewCS = false;
                        if (cs != newCS) addedToNewCS = true;

                        stylePane.moveNodeFromTo(dragged_$el, dragged_$placeholder, addedToNewCS, newCS);
                    }

                    if (dragged_$el) {
                        dragged_$el.css('opacity', 1.0);
                        dragged_$el.show();
                    }
                }

                dragged_$el = null;
                dragged_rule = null;
                if (dragged_$placeholder) dragged_$placeholder.remove();
                dragged_$placeholder = null;
            }
            dragged_$target = null;
        }

        $container.get(0).addEventListener('dragend', eventDragend, true);

        $container.get(0).addEventListener('drop', function(event) {
            var $target = $(event.target);

            if($target.closest('.crsa-node').length) {
                // do nothing
            }
            event.stopPropagation();
        }, true);
    }

    // prevent drag scroll
    if (options.$scrollParent) {
        options.$scrollParent.on('scroll', function() {
            if (dragged_$el) {
                $(this).parent(0).scrollLeft(0);
            }
        });
    }

    this.remove = function () {
        stylePane = null;

        $container.get(0).removeEventListener("contextmenu", contextmenuTriggered, true);
        $container.get(0).removeEventListener('click', containerClicked, true);
        $container.get(0).removeEventListener('dragend', eventDragend, true);
        $container.get(0).removeEventListener('mousedown', containerMousedownTrigger);
    }
}