window.PgCSSStylesheet = function () {
    var pgStylesheet = new window.PgStylesheet();
    pgStylesheet.fileType = "css";
    pgStylesheet.inline = false;
    pgStylesheet.inlineSheetPgel = null;
    pgStylesheet.sourceToSave = null;

    var delayedTasks = new PgDelayedTasks();
    pgStylesheet.hasDelayedUpdate = false;

    var canCleanTheCode   = false;

    var allPostcssLib        = require('postcss');
    var pgStringifiers       = allPostcssLib.postcssPG;
    var pgStringifier        = pgStringifiers.pgStringify;
    var pgStringifierWithIds = pgStringifiers.pgStringifyWithIds;

    //workaround for @charset missing ; at the end
    /*
    pgStringifier.prototype.bodyOriginal = pgStringifier.prototype.body;

    pgStringifier.prototype.body = function(node) {
        this.bodyOriginal(node);

    }

     */


    function getAutoprefixerBrowsers() {
        var browsersStr = pinegrow.getSetting('compilable-autoprefixer-browsers', pinegrow.defaults.autoprefixer, false);
        if (browsersStr && browsersStr.length) {
            var browsers = browsersStr.split(',');

            var filteredBrowsers = browsers.map(function (browser) {
                return browser.trim();
            });

            if (filteredBrowsers.length) return filteredBrowsers;
        }

        return [];
    }

    pgStylesheet.isCompilableStylesheet = function () {
        return false;
    }

    pgStylesheet.getAutocompleteArr = function () {
        return [];
    }

    pgStylesheet.getInlineSheet = function(iframe) {
        if(!this.inlineSheetPgel) return null;
        if(!iframe) {
            var page = this.inlineSheetPgel.getPage();
            if(page) iframe = page.$iframe.get(0);
        }

        if(!iframe || !iframe.contentDocument) return null;
        var $html = $(iframe.contentDocument.documentElement);
        var domel = this.inlineSheetPgel.getDOMElement($html);
        if(domel && domel.sheet) return domel.sheet;
        return null;
    }

    pgStylesheet.elementStyle = false;
    pgStylesheet.element = null;
    pgStylesheet.styleRuleName = "Style attribute";

    var id_count = 0;
    var pgCssStates = ['active', 'hover', 'focus', 'visited'];

    var postcss = allPostcssLib.postcss;
    var cssSyntax = allPostcssLib.postcssSafeParser;
    var autoprefixer = allPostcssLib.autoprefixer;


    var getTreeNodeId = function() {
        return ++id_count;
    }
    var resetTreeNodeId = function() {
        id_count = 0;
    }

    var isRuleEmpty = function (rule) {
        if (rule.nodes.length == 0) return true;
        return false;
    }

    var addPgIdToNode = function (allNodes, _this) {
        var addToAllNodes = function (nodes, nested) {
            for (var i=0; i < nodes.length; i++) {
                // Detect Formatting
                if (pinegrow.autoFormatCss == 'auto-detect') {
                    pgcssf.detectFormatting(nodes[i], nested, i, _this.formattingOptions);
                }

                if (nodes[i].type == "rule") {
                    if (!nodes[i].nodes) nodes[i].nodes = [];
                    if (!_this.canAddRuleIds(nodes[i], _this.atruleWithNoChild)) continue;

                    nodes[i]['pg-node-id'] = getTreeNodeId();
                    nodes[i]['pg-main-file'] = _this.getUniqueName();

                    if (_this.isImported() || _this.inline) {
                        nodes[i]['pg-file'] = _this.getUniqueName();
                    }

                    var selector = nodes[i].selector;
                    if (selector.indexOf('.pg-state-') > -1 || selector.indexOf('.pg-node-id-') > -1) canCleanTheCode = true;
                }

                if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                    addToAllNodes(nodes[i].nodes, true);
                }
            }
        }

        _this.formattingOptions = {};
        _this.emptyNodes = [];
        addToAllNodes(allNodes, false);
        return allNodes;
    }

    var addPgIdToNodeUsingProp = function (allNodes, _this) {
        var addToAllNodes = function (nodes) {
            for (var i = 0; i < nodes.length; i++) {
                if (nodes[i].type == "rule") {
                    if (!nodes[i].nodes) nodes[i].nodes = [];
                    if (nodes[i].nodes.length == 0) {
                        _this.emptyNodes.push(nodes[i]);
                    }

                    var childrenNode = nodes[i].nodes;
                    var length = childrenNode.length;
                    var lastNode = childrenNode[length - 1];

                    if (lastNode && lastNode.type == "decl" && lastNode.prop == "pg-id") {
                        nodes[i]['pg-node-id'] = lastNode.value;
                        childrenNode.splice(length - 1, 1);
                        length--;


                        lastNode = childrenNode[length - 1];
                        if (lastNode && lastNode.type == "decl" && lastNode.prop == "pg-main") {
                            nodes[i]['pg-main-file'] = lastNode.value.replace(/[\"\']/g, '');
                            childrenNode.splice(length - 1, 1);
                            length--;
                        }


                        lastNode = childrenNode[length - 1];
                        if (lastNode && lastNode.type == "decl" && lastNode.prop == "pg-file") {
                            nodes[i]['pg-file'] = lastNode.value.replace(/[\"\']/g, '');
                            childrenNode.splice(length - 1, 1);
                            length--;
                        }
                    }
                }
                else if (nodes[i].type == "decl") {
                    if (nodes[i].prop == "pg-id") {
                        nodes.splice(i, 1);
                        i--;
                    }
                    else if (nodes[i].prop == "pg-main") {
                        nodes.splice(i, 1);
                        i--;
                    }
                    else if (nodes[i].prop == "pg-file") {
                        nodes.splice(i, 1);
                        i--;
                    }
                }

                if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                    addToAllNodes(nodes[i].nodes);
                }
            }
        }

        _this.emptyNodes = [];
        addToAllNodes(allNodes);
        return allNodes;
    }

    var removePgStateFromSelector = function (selector) {
        var selArr = selector.split(',');

        for (var i = 0; i < selArr.length; i++) {
            var currSel = selArr[i];

            var matched = currSel.match('pg-state-');
            if (matched && matched.length > 0) {
                selArr.splice(i, 1);
                i--;
            }
        }

        return selArr.join(',');
    }

    pgStylesheet.showCanCleanTheCodeAlert = function () {
        var _this = this;
        var cleanPgSelector = function () {
            var cleanAllNodes = function (nodes) {
                for (var i=0; i < nodes.length; i++) {
                    if (nodes[i].type == "rule") {
                        var selector = nodes[i].selector;
                        var selectorArr = selector.split(',');
                        var newSelector = [];
                        for (var j = 0; j < selectorArr.length; j++) {
                            if (selectorArr[j].indexOf('.pg-state-') < 0 &&
                                selectorArr[j].indexOf('.pg-node-id-') < 0)
                                newSelector.push(selectorArr[j]);
                        }

                        nodes[i].selector = newSelector.join(',');
                    }

                    if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                        cleanAllNodes(nodes[i].nodes);
                    }
                }
            }

            var rootNode = _this.rootNode;
            cleanAllNodes(rootNode.nodes);
            canCleanTheCode = false;

            _this.source = _this.getSourceForRules(rootNode);

            var path = crsaMakeFileFromUrl(_this.url);
            _this.save(path, function(err) {
                if (err) {
                    showAlert('Error: ' + err, 'Can\'t save this file');
                }
            });
        }
        pinegrow.showAlert(
            '<p>This file contains some special Pinegrow CSS selectors that help Pinegrow keep track of active CSS rules. Under certain conditions, previous versions of Pinegrow left these temporary helpers in the stylesheet code. This issue is now fixed, but some Pinegrow selectors are still left in this stylesheet.</p><p>Do you want to remove the Pinegrow selectors from the stylesheet and then save it?</p>',
            'Clean <b>' + this.name + '</b>', "No", "Yes",
            function () {
                // on No
                // do nothing
            },
            function () {
                // on Yes
                cleanPgSelector();
            });
    }

    pgStylesheet.canEditRuleSelector = function () {
        if (this.elementStyle) return false;
        else                   return true;
    }

    pgStylesheet.canAddChangeMediaQuery = function () {
        if (this.elementStyle) return false;
        else                   return true;
    }

    pgStylesheet.lastNodeId = function () {
        return id_count;
    }

    pgStylesheet.reload = function(done) {
        pinegrow.codeEditors.ignoreCodeForUrl(this.url);

        if (!this.inline) {
            this.load(this.url, done, true);
        }
        else {
            if (done) done(this);
        }
    }

    pgStylesheet.callPostcss = function (source, onDone, onError) {
        if (!source) source = "";
        var parsingPlugin = [];

        if (pinegrow.getSetting('disable-formating', "0") == "0") {
            // parsingPlugin.push(stylefmt);
        }

        if (this.sourceStylesheet) {
            if (pinegrow.getSetting('compilable-use-autoprefixer', '1') == '1') {
                var browsers = getAutoprefixerBrowsers();

                parsingPlugin.push(autoprefixer({
                    browsers: browsers
                }));
            }
        }

        var syntaxRes = postcss(parsingPlugin).process(source, { parser: cssSyntax });

        if (this.elementStyle) {
            //use sync mode for style attr to avoid flicker in Active panel
            try {
                onDone(syntaxRes.result)
            } catch (err) {
                onError('Error parsing inline style');
            }

        } else {
            syntaxRes.then(function (postcssRes) {
                onDone(postcssRes)
            });
        }

        if (syntaxRes.error) {
            var currError = syntaxRes.error;
            if (currError.reason && currError.line) {
                onError(currError.reason + ":" + currError.line);
            } else {
                onError(currError.message);
            }
        }
    }

    pgStylesheet.parseSource = function (source, done) {
        this.resetError();
        var _this = this;

        var onDone = function(postcssRes) {
            resetTreeNodeId();

            var rules = postcssRes.root;
            if (_this.compiledFile) {
                addPgIdToNodeUsingProp(rules.nodes, _this);
            }
            else {
                addPgIdToNode(rules.nodes, _this);
            }
            var postcssSource = _this.getSourceForRules(rules);

            if (done) done(null, postcssSource, rules);
        }

        var onError = function(parsingError) {
            _this.codeError = parsingError;
            _this.source = source;
            if (done) done(parsingError);
        }

        this.sourceToSave = null;
        this.recursiveImports = null;
        this.callPostcss(source, onDone, onError);
    }

    pgStylesheet.addIdsToNodes = function(nodes) {
        if (this.compiledFile) {
            addPgIdToNodeUsingProp(nodes);
        }
        else {
            addPgIdToNode(nodes, this);
        }
    }

    pgStylesheet.updateSingleRule = function(rule, done, prop, value) {
        this.setChanged(true);
        this.setNeedFormatting(true);
        if(this.elementStyle || this.inline || !rule['pg-node-id']) {
            //cant update single
            if(this.elementStyle && prop && false) {
                //should just change the prop if multiple elements are selected
                var col = pinegrow.getCollection();

                col.forEachElement(function(pgel) {
                    pgel.setInlineStylePropertyValue(prop, value, false, false);
                })
                pinegrow.changeManager.didChange(col, 'Set style ' + prop, {
                    undoCombineKey: 'setAttr_style_' + prop,
                    source: 'elementStyle'
                });

                done && done();

            } else {
                this.setSourceForRules(this.rootNode, done);
            }
        } else {
            this.source = null; //invalidate current source
            // var code = pgcssh.getSource(rule);
            // pinegrow.stylesheets.applyToPagesSingle(this, rule, code);
            pinegrow.stylesheets.applyToPagesSingle(this, rule);
            if(done) done();
        }
    }

    pgStylesheet.setAllCssText = function(styles, done) {
        // if (!styles) return;

        var _this = this;
        if(!this.rootNode) {
            this.parseSource(styles, function (error, parsedSource, rules) {
                if (!error) {
                    _this.rootNode = rules;
                    _this.rootNode.pgStylesheet = _this;
                    _this.setAllCssText(parsedSource);
                }
            });
            return;
        }

        //this is bad. pg-node-is will go to source. what happens when we save the cs?
        //styles should always be without ids
        this.source = styles;

        if (this.elementStyle) {
            var col = pinegrow.getCollectionForElement(this.element);

            styles = styles
                .replace(this.styleRuleName, '')
                // .replace(/\,\s*\.pg-node-id-[0-9]*/g, '')
                // .replace(/\n?\s*\/\*\spg-uncollapsed:true\s\*\//g, '')
                .replace(/[\{\}]/g, '');

            styles = this.filterSource(styles);

            var changed = false;
            var dyn_notice = false;

            col.forEachElement(function(pgel) {
                if(styles != (pgel.getAttr('style') || '')) {

                    if (styles.length) {
                        pgel.setAttr('style', styles);
                    } else {
                        pgel.removeAttr('style');
                    }
                    changed = true;
                    if(pgel.dynamic && !pgel.dynamicStyleWarned) {
                        dyn_notice = true;
                        pgel.dynamicStyleWarned = true;
                    }
                }
            })
            if(changed) {
                pinegrow.changeManager.didChange(col, 'Set style attribute', {
                    undoCombineKey: 'setAttr_style',
                    source: 'elementStyle'
                });
            }
            if(dyn_notice) {
                pinegrow.showQuickError(`WARNING: Inline style on dynamic elements can't be saved.`);
            }
        }
        else if (this.inline) {
            if (this.inlineSheetPgel) {
                var source = this.source;
                var page = this.inlineSheetPgel.getPage();
                if(page) {
                    //here we just need to do source = styles with ids
                    var domSource = _this.getSourceForRules(_this.rootNode, true);

                    //handle undo states in crsa.undo.js
                    pinegrow.undoManager.getHistoryForPage(page).ignore(function() {
                        //todo, have to solve this a bit differently... will do it tomorrow
                        _this.inlineSheetPgel.html(source);
                        //console.log(source)
                    })
                }
            }
        }
        else {
            pinegrow.stylesheets.applyToPages(this);
        }
    }

    pgStylesheet.setSource = function(source, done, skipUpdatingRule, skipUpdatingCssText, fastDelayed) {
        var _this = this;

        fastDelayed = false; //disable

        //fast cs don't care about ids and parsing rules
        // @TODO: deleting var in custom.scss will add pg-file & pg-id to css rules
        if(this.fast || fastDelayed) {
            this.source = source;
            this.sourceToSave = source;

            if(fastDelayed) {
                this.hasDelayedUpdate = true;
                pinegrow.stylesheets.applyToPagesQuick(_this, source);

                delayedTasks.run('delayedUpdate', function() {
                    _this.setSource(source);
                }, 2000)
            }
            if(done) done();

            return;
        }

        this.hasDelayedUpdate = false;
        delayedTasks.cancel('delayedUpdate');

        skipUpdatingRule = skipUpdatingRule || false;

        if (this.inline) source = this.filterSource(source);

        this.sourceBeingCurrentlyParsed = source;

        this.parseSource(source, function (parsingError, parsedSource, rules) {
            _this.sourceBeingCurrentlyParsed = null;

            if (parsingError) {
                _this.codeError = parsingError;
                if (done) done();
            }
            else {
                _this.codeError = null;
                if (getTrimed(_this.parsedSource) != getTrimed(parsedSource) || _this.source === null) {
                    _this.source = _this.parsedSource = parsedSource;
                    if (_this.loaded) _this.setChanged(true);
                }
                if (!skipUpdatingRule) {
                    _this.rootNode = rules;
                    _this.rootNode.pgStylesheet = _this;
                }
                // _this.updateImportedListOrder();

                if(!skipUpdatingCssText) {
                    _this.setAllCssText(_this.source);
                }

                _this.loadAllImportedFiles(done);
            }
        });
    }

    pgStylesheet.getSourceToSave = function (done) {
        if (this.sourceToSave != null) {
            done(null, this.sourceToSave);
            return;
        }

        var _this = this;
        var source = this.getSource();

        var remove_empty = pinegrow.getSetting('compilable-remove-empty-rules', '1') == '1';
        var filter_nodes = remove_empty; // || otherThings

        var filterNodes = function (root) {
            var forAllNodes = function (nodes) {
                for (var i = 0; i < nodes.length; i++) {
                    if (pgcssh.getNodeType(nodes[i]) == "rule") {
                        if (remove_empty) {
                            if (isRuleEmpty(nodes[i])) {
                                nodes[i].remove();
                                i--;
                            }
                        }
                    }
                    else if (pgcssh.getNodeType(nodes[i]) == 'import') {
                        if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                            forAllNodes(nodes[i].nodes);
                        }
                    }
                }
            }

            forAllNodes(root.nodes);
            return root;
        }

        var onDone = function (postcssRes) {
            var root = postcssRes.root;
                root = filterNodes(root);

            var newSource = _this.getSourceForRules(root);
            _this.sourceToSave = newSource;
            if (done) done(null, newSource);
        }

        var onError = function (error) {
            if (done) done(error);
        }

        if (!source || source.length == 0) {
            if (done) done(null, "");
            return;
        }

        if (filter_nodes) {
            this.callPostcss(source, onDone, onError);
        }
        else {
            if (done) done(null, source);
        }
    }

    pgStylesheet.getSource = function() {
        if (!this.source) {
            this.source = this.getSourceForRules();
        }
        return this.source;
    }

    pgStylesheet.getSourceForRules = function(node, withIds, forceFormatting) {
        node = node || this.getRootNode();
        if (node) {
            if (this.elementStyle) {
                this.formattingOptions = {
                    'indent': '',
                    'before-first-nested-decl': '',
                    'before-decl-after-decl': ' ',
                    'before-nested-decl-after-decl': ' ',
                    'after-decl': '',
                    'after-rule': '',
                    'rule-between': ''
                }
            }

            if (forceFormatting || (!withIds && !this.compiledFile)) {
                this.updateNodeFormatting(node, forceFormatting);
            }

            var stringifier = pgStringifier;
            if (withIds) {
                stringifier = pgStringifierWithIds;
            }
            var code = pinegrow.pgPostCSSHelper.getSource(node, stringifier, withIds);
            if(node.type === 'atrule' && node.name === 'charset') {
                if(!code.endsWith(';')) {
                    code += ';';
                }
            }
            return code;
        }
        else {
            return '';
        }
    }

    pgStylesheet.setSourceForRules = function(node, done, skipUpdatingRule) {
        this.setNeedFormatting(true);
        var newSource = this.getSourceForRules(node);
        this.setSource(newSource, done, skipUpdatingRule);
    }

    pgStylesheet.regenerateAndSetCssSource = function (done, process_rules, rootNode) {
        if(typeof process_rules == 'undefined') process_rules = true;

        rootNode = rootNode || this.getRootNode();
        var css = this.getSourceForRules(rootNode);
        if(process_rules) {
            this.setSource(css, done);
        } else {
            this.setAllCssText(css);
            setTimeout(function() {
                if(done) done();
            }, 10);
        }
    }

    pgStylesheet.updateNodeWithSource = function (node, source, done) {
        var _this = this;
        this.parseSource(source, function (err, postcssSource, rootNode) {
            if (err) {
                if (done) done(error);
                return;
            }
            if (pinegrow.pgPostCSSHelper.getRootNodeFor(node) == _this.getRootNode()) {
                var newNode = pinegrow.pgPostCSSHelper.replaceNodeWithAndUpdatePgId(node, rootNode.nodes[0]);
                if (done) done(null, newNode);
            }
            else {
                var indices = pinegrow.pgPostCSSHelper.getNodeIndices(node);
                var newRootNode = pinegrow.pgPostCSSHelper.getNodeByIndices(_this.getRootNode(), indices);
                var newNode = pinegrow.pgPostCSSHelper.replaceNodeWithAndUpdatePgId(newRootNode, rootNode.nodes[0]);
                if (done) done(null, newNode);
            }
        });
    }

    pgStylesheet.updateNodeStylesheetWithSource = function(node, source, done) {
        var _this = this;

        this.parseSource(source, function (err, postcssSource, rootNode) {
            if (err) {
                if (done) done(err);
                return;
            }

            var newNode = rootNode.nodes[0];
            pinegrow.pgPostCSSHelper.replaceNodeWith(node, newNode);
            newNode.raws.before = node.raws.before;
            _this.setSourceForRules(pinegrow.pgPostCSSHelper.getRootNodeFor(newNode), function (error) {
                if (done) done(error, newNode);
            });
        });
    }

    pgStylesheet.regenerateRules = function (done) {
        this.parseSource(this.source, function (error, postcssSource, rules) {
            if (error) {
                done({nodes: []});
            }
            else {
                done(rules);
            }
        });
    }

    pgStylesheet.getRuleById = function () {
        // TODO implement this
    }

    pgStylesheet.getRuleSelector = function (rule) {
        return rule.selector;
    }

    pgStylesheet.getCssSource = function (done, withIds) {
        if(this.fast) {
            done(this.source);
            return;
        }

        if (!this.source || withIds) {
            done(this.getSourceForRules(null, withIds));
        }
        else {
            done(this.source);
        }
    }

    pgStylesheet.updateImportedListOrder = function () {
        return; // Not used
        var importedList = this.importedFiles;
        var newImportedList = [];
        var nodes = this.rootNode.nodes;
        for (var i = 0; i < nodes.length; i++) {
            var node = nodes[i];
            if (node.type == "atrule") {
                var relativeUrl = crsaGetUrlAndParamsFromImport(nodes[i].params)[0];
                for (var j = 0; j < importedList.length; j++) {
                    if (importedList[j].relativeUrl == relativeUrl) {
                        newImportedList.push(importedList[j]);
                        importedList.splice(j, 1);
                        break;
                    }
                }
            }
            else {
                if (node.type != "comment") break;
            }
        }

        this.importedFiles = newImportedList;
    }

    pgStylesheet.getIndexForSourceStylesheet = function () {
        return 'source-' + this.index;
    }

    pgStylesheet.getImportedStylesheetFromRelativeUrl = function (reqUrl, importedArray) {
        reqUrl = crsaGetUrlAndParamsFromImport(reqUrl)[0];
        var importedFiles = importedArray || this.importedFiles;
        var relativeUrls = [];
        relativeUrls.push(
            this.getFullUrlFor(reqUrl),
        );

        for (var i = 0; i < importedFiles.length; i++) {
            if (relativeUrls.indexOf(importedFiles[i].url) > -1) {
                return importedFiles[i];
            }
        }
    }

    pgStylesheet.ignoreImportedFile = function (node, ignore) {
        var importedStylesheet = this.getImportedStylesheetFromRelativeUrl(node.params);
        if (importedStylesheet) {
            importedStylesheet.ignore = ignore;
            this.setRecursiveImportsError(null);
        }
    }

    pgStylesheet.removeImportedFile = function (node) {
        var importedStylesheet = this.getImportedStylesheetFromRelativeUrl(node.params);
        var index = this.importedFiles.indexOf(importedStylesheet);
        if (index > -1) {
            this.importedFiles.splice(index, 1);
        }
        else {
            this.setRecursiveImportsError(null);
        }
    }



    pgStylesheet.addImportedFile = function (node, done) {
        var fileRelativeUrl = crsaGetUrlAndParamsFromImport(node.params)[0];
        var href = crsaCombineUrlWith(this.url, fileRelativeUrl);
        var cs = pinegrow.stylesheets.getStylesheetByKey(href);
        var fileCreated = false;

        var allDone = function (recursiveImports) {
            if (fileCreated) pinegrow.refreshCurrentProjectDelayed(true);
            if (done) done(recursiveImports);
        }

        var loadFile = function () {
            cs.load(href, function () {
                console.log('done ' + cs.url);
                pinegrow.dispatchEvent('on_css_loaded', null, cs);
                allDone();
            });
        }

        var recursiveImports = this.checkRecursiveImports(href, true);
        if (!recursiveImports) {
            if (cs) {
                allDone();
            }
            else {
                cs = PgCSSStylesheet();
                cs.loaded = false;
                cs.index = pinegrow.stylesheets.getLastIndex(true);
                var fileUrl = crsaMakeFileFromUrl(href);

                if (!crsaIsFileExist(fileUrl)) {
                    fileCreated = true;
                    crsaCreateDirs(crsaGetFileDir(fileUrl));
                    fs.writeFile(fileUrl, "", function(err) {
                        if(err) {
                            return console.log(err);
                        }

                        console.log("Imported file created!");
                        loadFile();
                    });
                }
            }
        }
        else {
            allDone(recursiveImports);
        }

        cs.addParentStylesheet(this);
        if (this.importedFiles.indexOf(cs) < 0) {
            pinegrow.stylesheets.stylsheetUsed({
                url: href,
                pgStylesheet: cs
            }, 'child', this);
        }
        if (!cs.loaded) {
            loadFile();
        }
    }



    pgStylesheet.getImportedFiles = function () {
        if (this.ignore) return [];

        var _this = this;
        var doTask = function () {
            var list = [];
            var rules = _this.rootNode.nodes;
            if(rules) {
                for(var i = 0; i < rules.length; i++) {
                    if(rules[i].type == "atrule" && rules[i].name == "import") {
                        var fileRelativeUrl = crsaGetUrlAndParamsFromImport(rules[i].params)[0];
                        if (crsaIsAbsoluteUrl(fileRelativeUrl)) continue;

                        var href = crsaCombineUrlWith(_this.url, fileRelativeUrl);
                        var recursiveImports = _this.checkRecursiveImports(href, true);
                        if (recursiveImports) return;

                        var cs = pinegrow.stylesheets.getStylesheetByKey(href);

                        if (!cs) {
                            cs = PgCSSStylesheet();
                            cs.index = pinegrow.stylesheets.getLastIndex(true);
                        }

                        cs.addParentStylesheet(_this);
                        list.push({cs: cs, href: href});

                        if (_this.importedFiles.indexOf(cs) < 0) {
                            pinegrow.stylesheets.stylsheetUsed({
                                url: href,
                                pgStylesheet: cs
                            }, 'child', _this);
                        }
                    }
                }
            }
            return list;
        }

        return doTask();
    }

    pgStylesheet.loadFromUrl = function (url, done) {
        this.load(url, done, false, true);
    }

    pgStylesheet.load = function (url, done, force, skipCheckingCompilables, skipFileExistence, newFile) {
        this.resetError();

        /*
        if(url.indexOf('shortcodes') > 0) {
            debugger;
        }*/

        var _this = this;
        var nakedUrl = this.setNakedUrl(url);

        var doneasync = function(a, b, c) {
            var allDone = function () {
                _this.appliedToPage = true;
                if (canCleanTheCode) _this.showCanCleanTheCodeAlert();

                if (done) {
                    if (_this.elementStyle) {
                        done(_this, a, b, c);
                        pinegrow.dispatchEvent('on_css_loaded', null, _this);
                    } else {
                        setTimeout(function () {
                            if (newFile) _this.setChanged(true);
                            var changeChecked = function () {
                                done(_this, a, b, c);
                                pinegrow.dispatchEvent('on_css_loaded', null, _this);
                            }

                            if (_this.compiledFile) {
                                _this.getSourceToSave(function (error, source) {
                                    if (!error) {
                                        if (getTrimed(_this.compiledFileSource) == getTrimed(source)) {
                                            _this.setChanged(false);
                                        }
                                        else {
                                            _this.setChanged(true);
                                        }
                                    }
                                    changeChecked();
                                });
                            }
                            else {
                                changeChecked();
                            }
                        }, 1);
                    }
                }
            }

            if (!_this.appliedToPage && _this.compiledFile && (_this.sourceStylesheet.changedOnce || _this.usedExistedCompiled)) {
                var compiledFileSource = _this.compiledFileSource;
                _this.sourceStylesheet.regenerateAndSetCssSource(function () {
                    if (getTrimed(compiledFileSource) != getTrimed(_this.source)) {
                        _this.setChanged(true);
                    }
                    allDone();
                });
            }
            else {
                allDone();
            }
        }

        var loadFile = function () {
            if(_this.ignore && !force) {
                doneasync();
                return;
            }

            var fs = require('fs');

            var mappedFile = pinegrow.getMappedRemoteFileForUrl(nakedUrl, 'css');

            if(mappedFile) {
                if(crsaIsFileExist(mappedFile)) {
                    nakedUrl = crsaMakeUrlFromFile(mappedFile);
                }
                _this.mappedRemote = true;
            }


            if(isApp() && crsaIsFileUrl(nakedUrl)) {
                try {
                    var fileName = crsaMakeFileFromUrl(nakedUrl);

                    var str;

                    //check if url open in code editor and use that source
                    str = pinegrow.codeEditors.getCodeForUrl(url);

                    if(str === null) {
                        if (crsaIsFileExist(fileName)) {
                            str = fs.readFileSync(fileName, {encoding: "utf8"});
                        }
                    }

                    console.log('loaded css ' + url);

                    if (!_this.compiledFile) {
                        _this.setSource(str, function() {
                            _this.setChanged(false);
                            _this.loaded = true;
                            doneasync();

                            _this.setLocalFile(fileName);
                        });
                    }
                    else {
                        _this.compiledFileSource = str;
                        doneasync();
                    }

                } catch(err) {
                    _this.setSource('', doneasync);
                    if (!skipFileExistence) _this.error = err.message;
                }
            } else {

                var onDone = function(data) {
                    console.log('loaded css ' + url);
                    _this.setLoadingFailed(false);

                    if(mappedFile) {
                        _this.setLocalFile(mappedFile);
                    }

                    if (!_this.compiledFile) {
                        _this.setSource(data, function() {
                            _this.loaded = true;
                            doneasync();
                        });
                    }
                    else {
                        _this.compiledFileSource = data;
                        _this.loaded = true;
                        doneasync();
                    }
                }

                var onFail = function(is_ajax) {
                    var message = 'Failed loading css ' + url;
                    console.log(message);

                    if(is_ajax && false) {
                        console.log('Retrying...');

                        //Try with referer
                        try {
                            pgGetHttpRequest(url,
                                function (body) {
                                    console.log(body);
                                    onDone(body);
                                },
                                function () {
                                    onFail();
                                }
                            )
                        } catch(err) {
                            onFail();
                        }

                    } else {
                        _this.loaded = true;
                        _this.setLoadingFailed( true);
                        _this.setSource('', doneasync);
                        _this.error = message;
                    }
                }

                var project = pinegrow.getCurrentProject();

                if(project && project.loaded_from_backend && project.hasContentForUrl(url)) {
                    //we are in the browser
                    if(project) {
                        _this.loaded_from_project_content = true;
                        project.getContentOfUrl(url, function(data) {
                            if(data === null) {
                                onFail(true);
                            } else {
                                onDone(data);
                            }
                        })
                    }

                } else {

                    $.ajax({
                        url: url,
                        data: null,
                        dataType: 'text'
                    }).done(function (data) {
                        onDone(data);

                    }).fail(function () {
                        onFail(true);

                    });
                }
            }
        }

        var loadStyle = function () {
            var style = _this.element.getAttr('style') || '';

            style = _this.styleRuleName + '{' + style + '}';
            var is_done = false;
            _this.setSource(style, function() {
                if(!is_done) {
                    //sometimes this callback gets called twice. prevent it
                    is_done = true;
                    _this.loaded = true;
                    doneasync();
                }
            }, false, true /* dont update el style */);
        }

        if (this.elementStyle && this.element) {
            loadStyle();
        }
        else if (!skipCheckingCompilables && (!this.sourceStylesheet || !this.sourceStylesheet.loaded)) {
            var sourceCs = null, sourceUrl = null;

            var projectInfo = this.getProjectInfo();
            var compiledStylesheet = isApp() ? projectInfo.getSetting('compiledStylesheet') : null;

            var relative_url = projectInfo.makeUrlRelative(this.url);

            var compiledSSSet = (compiledStylesheet && (compiledStylesheet[this.url] || compiledStylesheet[relative_url]));

            if (compiledSSSet) {
                if(compiledStylesheet[relative_url]) {
                    sourceUrl = projectInfo.makeUrlAbsolute(compiledStylesheet[relative_url]);
                } else {
                    sourceUrl = compiledStylesheet[this.url];
                }
            }

            if (!sourceUrl && crsaIsFileUrl(this.url)) {
                var currExt = crsaGetExtFromUrl(this.url);
                var cleanUrl = crsaRemoveUrlParameters(this.url);
                var urlWithoutExt = currExt ? (cleanUrl.substr(0, cleanUrl.length - currExt.length)) : cleanUrl;

                var scssUrl = urlWithoutExt + 'scss';
                var lessUrl = urlWithoutExt + 'less';

                if (crsaIsFileExist(crsaMakeFileFromUrl(scssUrl))) {
                    sourceUrl = scssUrl;
                }
                else if (crsaIsFileExist(crsaMakeFileFromUrl(lessUrl))) {
                    sourceUrl = lessUrl;
                }
            }

            var useStylesheetSource = this.getUseStylesheetSource(nakedUrl);

            if (sourceUrl) {
                var fileExt = crsaGetExtFromUrl(sourceUrl);

                var mainFileName = crsaGetNameFromUrl(nakedUrl);
                var sourceFileName = crsaGetNameFromUrl(sourceUrl);

                var loadSourceFile = function () {
                    var sourceCs = pinegrow.stylesheets.getStylesheetByKey(sourceUrl);
                    if (sourceCs) {
                        _this.sourceStylesheet = sourceCs;
                        sourceCs.addCompiledStylesheet(_this);
                        _this.loaded = sourceCs.loaded;
                        _this.ignore = sourceCs.ignore;
                        _this.usedExistedCompiled = true;
                        loadFile();
                    }
                    else {
                        var fileExt = crsaGetExtFromUrl(sourceUrl);

                        if (fileExt == "scss") {
                            sourceCs = PgSCSSStylesheet();
                        }
                        else if (fileExt == "less") {
                            sourceCs = PgLESSStylesheet();
                        }
                        if(sourceCs) {
                            sourceCs.index = _this.getIndexForSourceStylesheet();
                        }

                        if (sourceCs) {
                            _this.compiledFile = true;
                            _this.sourceStylesheet = sourceCs;
                            sourceCs.ignore = _this.ignore;

                            sourceCs.load(sourceUrl, function () {
                                _this.loaded = sourceCs.loaded;
                                if(sourceCs.getError() && !sourceCs.loaded) {
                                    //Could not load this file
                                    pinegrow.showQuickError(`Could not load ${sourceCs.url}.`);
                                    _this.compiledFile = false;
                                    _this.sourceStylesheet = null;
                                } else {
                                    _this.stopWatchingFileForChanges();
                                }
                                loadFile();
                            }, force, null, _this);
                        }
                    }

                    pinegrow.stylesheets.stylsheetUsed({
                        url: sourceUrl,
                        pgStylesheet: sourceCs
                    }, 'source', _this);
                }

                if (useStylesheetSource == "true" || compiledSSSet) {
                    loadSourceFile();
                }
                else if (useStylesheetSource == null) {
                    pinegrow.showAlert('Do you want to use <b>' + sourceFileName + '</b> as a source file for <b>' + mainFileName + '</b> and compile it with SASS/LESS?', 'Use SASS/LESS Source File', 'No', 'Yes',
                        function () {
                            // No
                            _this.setUseStylesheetSource("false");
                            loadFile();
                        },
                        function () {
                            // Yes
                            _this.setUseStylesheetSource("true");
                            loadSourceFile();
                        });
                }
                else {
                    loadFile();
                }
            }
            else {
                loadFile();
            }
        }
        else {
            loadFile();
        }
    }

    pgStylesheet.getPageStylesheetFrom = function (page) {
        var iframe = page.$iframe.get(0);
        var found_ss = null;
        if (this.inline) {
            found_ss = this.getInlineSheet(iframe);
        }
        else {
            found_ss = pinegrow.stylesheets.findAttachmentFor(this, page);
        }

        return found_ss
    }

    pgStylesheet.duplicate = function(done) {
        var n = PgCSSStylesheet();

        n.name = this.name.replace(/\.css($|[\?\/])/i, ".copy.css");

        var a = this.url.split("/");
        if(a.length > 0) {
            a.splice(a.length-1,1);
        } else {
            a = [""];
        }
        n.url = a.join("/") + "/" + n.name;

        n.loaded = this.loaded;
        n.ignore = this.ignore;

        n.setSource(this.getSource(), function() {
            if(done) done(n);
        });
        return n;
    }

    pgStylesheet.detachedFromSourceStylesheet = function () {
        this.compiledFile = false;
        this.sourceStylesheet = null;
    }

    pgStylesheet.filter = function (node) {
        return {
            haveTo: "doNothing"
        };
    }

    pgStylesheet.getDeclCharachter = function () {
        return null;
    }

    pgStylesheet.saveCompiledCss = function (path, done, fs, filter_func, fileWriter) {
        var _this = this;
        this.getSourceToSave(function (error, cssSource) {
            if (error) {
                console.log(error);
                return;
            }

            cssSource = _this.filterSource(cssSource);
            if (filter_func) cssSource = filter_func(cssSource);

            if (fileWriter) {
                fileWriter.writeFile(path, cssSource, function () {
                    _this.lastSavedAt = (new Date()).getTime();
                    console.log('saved css path ' + path);
                    _this.setChanged(false);
                    if (done) done();
                })
            } else {
                crsaWriteFileWithBackup(fs, path, cssSource, "utf8");
                _this.lastSavedAt = (new Date()).getTime();
                console.log('saved css path ' + path);
                _this.setChanged(false);
                if (done) done();
            }
        });
    }

    pgStylesheet.save = function(path, done, fs, filter_func, fileWriter) {
        if (this.compiledFile) {
            this.saveCompiledCss(path, done, fs, filter_func, fileWriter);
            return;
        }

        var _this = this;
        if(!fs) fs = require('fs');
        var shouldUpdateProject = false;
        var count = 0;

        var allDone = function () {
            if (count == 0) {
                if (shouldUpdateProject) pinegrow.refreshCurrentProjectDelayed(true);
                if (done) done();
            }
        }

        var saveStylesheet = function (cs, filePath) {
            var pathLib = require('path');
            filePath = pathLib.resolve(filePath);
            if (!crsaIsFileExist(filePath)) shouldUpdateProject = true;

            cs.getCssSource(function(css) {
                setTimeout(function() {
                    try{
                        css = _this.filterSource(css);
                        if(filter_func) css = filter_func(css);

                        if(fileWriter) {
                            fileWriter.writeFile(filePath, css, function() {
                                cs.lastSavedAt = (new Date()).getTime();
                                console.log('saved css path ' + filePath);
                                cs.setChanged(false);
                            })
                        } else {
                            crsaWriteFileWithBackup(fs, filePath, css, "utf8");
                            cs.lastSavedAt = (new Date()).getTime();
                            console.log('saved css path ' + filePath);
                            cs.setChanged(false);
                        }

                        //Here we had a bug with cs with ?params in the name,
                        if(crsaIsFileUrl(cs.url) && cs.localFile != filePath) {
                            //leave remote css urls intact
                            cs.setUrl(crsaMakeUrlFromFile(filePath));
                        }

                        pinegrow.dispatchEvent('on_css_file_saved', null, cs);

                        count--;
                        allDone();
                    }
                    catch(err) {
                        done(err);
                    }
                }, 10);
            });
        }
        count++;
        saveStylesheet(this, path);

        var importedFiles = this.importedFiles;
        if (importedFiles.length > 0) {
            p = require('path');
            for (var i = 0; i < importedFiles.length; i++) {
                count++;
                var cs = importedFiles[i];
                saveStylesheet(cs, cs.localFile);
            }
        }
    }

    pgStylesheet.setUseStylesheetSource = function (useStylesheet) {
        var project_info = this.getProjectInfo();

        var useStylesheetSourceObj = project_info.getSetting('useSourceStylesheet');
        if (!useStylesheetSourceObj) useStylesheetSourceObj = {};

        var relative_url = project_info.makeUrlRelative(this.url);

        delete useStylesheetSourceObj[this.url]
        useStylesheetSourceObj[relative_url] = useStylesheet;
        project_info.setSetting('useSourceStylesheet', useStylesheetSourceObj);
        project_info.save();
    }

    pgStylesheet.getUseStylesheetSource = function () {
        var project_info = this.getProjectInfo();
        var useStylesheetSourceObj = project_info.getSetting('useSourceStylesheet');

        if (useStylesheetSourceObj) {
            var relative_url = project_info.makeUrlRelative(this.url);
            return useStylesheetSourceObj[relative_url] || useStylesheetSourceObj[this.url] || null;
        }

        return null;
    }

    pgStylesheet.setSourceStylesheet = function (sourceStylesheet) {
        this.compiledFile = true;
        this.sourceStylesheet = sourceStylesheet;

        var project_info = this.getProjectInfo();
        var allCompiledStylesheet = isApp() ? project_info.getSetting('compiledStylesheet') : null;
        if (!allCompiledStylesheet) allCompiledStylesheet = {};

        var relative_url = project_info.makeUrlRelative(this.url);
        var relative_source_url = project_info.makeUrlRelative(sourceStylesheet.url);

        delete allCompiledStylesheet[this.url];
        allCompiledStylesheet[relative_url] = relative_source_url;

        project_info.setSetting('compiledStylesheet', allCompiledStylesheet);

        var useStylesheetSourceObj = project_info.getSetting('useSourceStylesheet');
        if (!useStylesheetSourceObj) useStylesheetSourceObj = {};

        delete useStylesheetSourceObj[this.url];
        delete useStylesheetSourceObj[relative_url];
        project_info.setSetting('useSourceStylesheet', useStylesheetSourceObj);

        project_info.save();
    }

    pgStylesheet.resetSourceStylesheet = function () {
        this.detachedFromSourceStylesheet();

        var project_info = this.getProjectInfo();
        var allCompiledStylesheet = isApp() ? project_info.getSetting('compiledStylesheet') : null;
        if (!allCompiledStylesheet) allCompiledStylesheet = {};

        var relative_url = project_info.makeUrlRelative(this.url);

        delete allCompiledStylesheet[this.url];
        allCompiledStylesheet[relative_url] = '';
        project_info.setSetting('compiledStylesheet', allCompiledStylesheet);

        var useStylesheetSourceObj = project_info.getSetting('useSourceStylesheet');
        if (!useStylesheetSourceObj) useStylesheetSourceObj = {};

        delete useStylesheetSourceObj[this.url];
        useStylesheetSourceObj[relative_url] = "false";
        project_info.setSetting('useSourceStylesheet', useStylesheetSourceObj);

        project_info.save();
    }

    pgStylesheet.resetError = function () {
        this.error = null;
        this.codeError = null;
        this.recursiveImports = null;
    }

    pgStylesheet.getError = function() {
        return this.error || this.codeError || (this.sourceStylesheet ? this.sourceStylesheet.getError() : null) || this.recursiveImports;
    }

    pgStylesheet.forgetCachedSource = function() {
        this.source = null;
    }

    pgStylesheet.renameRuleSelector = function (rule, newSelector, done, edited_by) {
        if(!this.canEdit(false, true)) {
            if(done) done();
            return;
        }
        // if (rule['pg-node-id']) {
        //     newSelector = this.filterSource(newSelector) + ", .pg-node-id-" + rule['pg-node-id'];
        // }
        if (pinegrow.pgPostCSSHelper.renameRuleSelector(rule, newSelector)) {
            this.forgetCachedSource();
            pinegrow.dispatchEvent('on_css_node_updated', null, {node: rule, stylesheet: this, edited_by: edited_by});
            this.setSourceForRules(pinegrow.pgPostCSSHelper.getRootNodeFor(rule), done, true);
        }
    }

    pgStylesheet.canHasImportMoreThanOneLevel = function () {
        return false;
    }

    pgStylesheet.stringifyNode = function (node) {
        return this.filterSource(pinegrow.pgPostCSSHelper.getSource(node));
    }

    pgStylesheet.loadInline = function(sheet, done) {
        var doneasync = function(a,b,c) {
            setTimeout(function() {
                done(a,b,c);
                pinegrow.dispatchEvent('on_css_loaded', null, _this);
            },1);
        }

        if(!sheet.ownerNode) {
            if(done) doneasync();
            return;
        }
        var pgel = getElementPgNode($(sheet.ownerNode));
        if(!pgel) {
            if(done) doneasync();
            return;
        }
        this.name = '(inline)';
        var $ss = $(sheet.ownerNode);
        var css = pgel.html();//$ss.html();
        this.inline = true;
        this.inlineSheetPgel = pgel;

        var _this = this;

        this.setSource(css, function() {
            _this.loaded = true;
            pinegrow.stylesheets.applyToPages(_this, function() {
                doneasync();
            });
        });
    }

    pgStylesheet.remove = function () {
        if (pinegrow.stylesheets.isStylesheetUsed(this)) return;

        delayedTasks.destroy();
        if (this.parentPgStylesheets.length > 0) this.parentPgStylesheets = [];
        if (this.sourceStylesheet) {
            this.sourceStylesheet.removeFromCompiledStylesheet(this);
            this.sourceStylesheet.remove(true);
            this.sourceStylesheet = null;
        }

        pgStylesheet = null;
        this.removeSuper();
    }

    return pgStylesheet;
}
