window.PgLESSStylesheet = function () {
    var pgStylesheet = window.PgCompilableStylesheet();
    pgStylesheet.fileType = "less";

    var less = require('less');


    var allPostcssLib = require('postcss');
    var postcss = allPostcssLib.postcss;
    var lessSyntax = allPostcssLib.postcssLESS;
    var lessStringify = allPostcssLib.postcssLESS.stringify;
    var lessStringifyWithIds = allPostcssLib.postcssLESS.stringifyWithIds;

    const LESS_FUNCTIONS = ['rgb', 'rgba', 'argb', 'hsl', 'hsla', 'hsv', 'hsva', 'boolean', 'red', 'green', 'blue', 'color', 'alpha', 'luma', 'luminance', 'dimension', 'convert', 'data-uri', 'unit', 'get-unit', 'svg-gradient', 'hue', 'saturation', 'lightness', 'hsvhue', 'hsvsaturation', 'hsvvalue', 'length', 'extract', 'ceil', 'floor', 'percentage', 'round', 'sqrt', 'abs', 'sin', 'asin', 'cos', 'acos', 'tan', 'atan', 'pi', 'pow', 'mod', 'min', 'max', 'isnumber', 'isstring', 'iscolor', 'iskeyword', 'isurl', 'ispixel', 'isem', 'ispercentage', 'isunit', 'isruleset', 'saturate', 'desaturate', 'lighten', 'darken', 'fadein', 'fadeout', 'fade', 'spin', 'mix', 'tint', 'shade', 'greyscale', 'contrast', 'multiply', 'screen', 'overlay', 'softlight', 'hardlight', 'difference', 'exclusion', 'average', 'negation', 'image-size', 'image-width', 'image-height', 'convert', 'data-uri', 'default'];

    pgStylesheet.canAddRuleIds = function (rule) {
        var canAdd = this.canAddRuleIdsSuper(rule);
        if (canAdd) {
            if (rule.empty) return false;
            if (rule.type == 'rule') {
                var selector = rule.selector.replace(/\s/g, '');
                if (selector.indexOf('(')) {
                    var selectors = selector.split(',');
                    for (var i = 0; i < selectors.length; i++) {
                        var sel = selectors[i];
                        if (sel.indexOf('(') > -1 && sel.indexOf(':') < 0) return false;
                    }
                }
            }
        }

        return canAdd;
    }

    pgStylesheet.getAutocompleteArr = function () {
        return LESS_FUNCTIONS;
    }

    pgStylesheet.isBuitInFunction = function (functionName) {
        return LESS_FUNCTIONS.indexOf(functionName) > -1;
    }

    pgStylesheet.skipImportFile = function (path) {
        return crsaIsAbsoluteUrl(path) ||
            path.indexOf(this.getDeclCharachter()) > -1 ||
            path.indexOf("@{") > -1;
    }

    pgStylesheet.getRelativeUrlFromImported = function (url, node) {
        if (url) url = url.trim();
        var filePath = crsaGetUrlAndParamsFromImport(url)[0];

        if (this.skipImportFile(filePath)) return filePath;

        var ext = '';
        if (!filePath.endsWith('.less')) ext = '.less';

        var importRelativeUrl = crsaGetBaseForUrl(filePath) + '/' + crsaGetNameFromUrl(filePath) + ext;
        if (importRelativeUrl[0] == '/') {
            importRelativeUrl = importRelativeUrl.substr(1, importRelativeUrl.length);
        }
        return importRelativeUrl;
    }

    pgStylesheet.parseSource = function (source, done, addNodeId) {
        var parsingPlugin = [];
        var _this = this;
        if (pinegrow.getSetting('disable-formating', "0") == "0") {
            // parsingPlugin.push(stylefmt);
        }

        var syntaxRes = postcss(parsingPlugin).process(source, { syntax: lessSyntax, stringifier: lessStringify });
        syntaxRes.then(function (postcssRes) {
            var rootNode = postcssRes.root;
            _this.resetTreeNodeId();

            if (addNodeId) {
                _this.processNodes(rootNode.nodes);
                var postcssSource = pinegrow.pgPostCSSHelper.getSource(rootNode, lessStringify);
                postcssSource = _this.addSemicolonToSource(postcssSource);
            }

            _this.sourceFormated = false;
            if (done) done(null, postcssSource, rootNode);
        });

        if (syntaxRes.error) {
            var parsingError = syntaxRes.error.reason + ":" + syntaxRes.error.line;
            this.codeError = parsingError;
            if (done) done(parsingError);
        }
    }

    pgStylesheet.getSourceForRules = function (node, withIds, forceFormatting) {
        node = node || this.getRootNode();

        if(!node) return '';

        if (forceFormatting || !withIds) {
            this.updateNodeFormatting(node, forceFormatting);
        }

        var sourceWithId = null;
        if (withIds) sourceWithId = pinegrow.pgPostCSSHelper.getSource(node, lessStringifyWithIds);
        // Call the normal Stringifier to remove params from nodes
        var source = pinegrow.pgPostCSSHelper.getSource(node, lessStringify);

        if (withIds) return sourceWithId;
        else return source;
    }

    pgStylesheet.regenerateCssFromLessSource = function (done) {
        var _this = this;
        var source = this.getSourceForRules(null, true);
        if (source == "") {
            if (done) done(null, { css: "" });
            return;
        }

        var options = {
            importer: function (request, prev, done) {
                return _this.importer(request, prev, done);
            },
            paths: [ crsaGetFileDir(this.localFile) ],
            relativeUrls: true,
            filename: this.name
        }

        if (this.includeSourceMap) {
            if (this.outFile) {
                options['sourceMap'] = true;
                options['outFile'] = this.outFile;
            }
            else {
                throw "You should enter the out file in order to generate source map";
            }
        }

        this.recursiveImports = null;
        less.render(source, options, function (error, result) {
            _this.renderingFinished(error, result, done);
        });
    }

    pgStylesheet.regenerateCssFromCompilableSource = function (done) {
        this.regenerateCssFromLessSource(done);
    }

    pgStylesheet.getNewStylesheet = function () {
        return PgLESSStylesheet();
    }

    pgStylesheet.getDeclCharachter = function () {
        return "@";
    }

    pgStylesheet.stringifyNode = function (node) {
        return this.addSemicolonToSource(this.filterSource(pinegrow.pgPostCSSHelper.getSource(node, lessStringify)));
    }

    pgStylesheet.superFilter = function (node) {
        return null;
    }

    return pgStylesheet;
}