window.PgSpectrumHelper = function ($input, options, no_check) {
    var $body = $('body');
    var $colorPickerOverlay = null;
    var _this = this;

    var win_ctx_map = new WeakMap();

    $input.data('pg-spectrum-helper', this);

    options = options || {}
    options['mouse-moved']  = options['mouse-moved'] || null;
    options['color-picked'] = options['color-picked'] || null;

    function rgbToHex(r, g, b) {
        if (r > 255 || g > 255 || b > 255)
            throw "Invalid color component";
        return ((r << 16) | (g << 8) | b).toString(16);
    }

    var getNewColor = function (color) {
        if (color.toRgb()['a'] == 1) return color.toHexString();
        else return color.toRgbString();
    }

    var colorPicked = function (e) {
        if (colorPicked == null) return;

        e.preventDefault();
        $(this).data('color-picked', true);

        var $el = $(this).data('selected-jq-element');
        if ($el && $el.length > 0) {
            if (options['color-picked']) options['color-picked']($el);
            $(this).data('selected-jq-element', null);
        }

        $el.spectrum("hide");
        return false;
    }

    var last_x = 0;
    var last_y = 0;

    var delayed = new PgDelayedTasks();

    var onMoved = function(e) {
        if(e.pageX === last_x && e.pageY === last_y) return;

        last_x = e.pageX;
        last_y = e.pageY;

        delayed.run('move', function() {
            pickColorUnderMouse(e)
        }, 200)
    }

    var pickColorUnderMouse = function (e) {
        var el = e.target;
        var $el = $(el);
        var offset = $el.offset();
        var pos = { x: e.pageX - offset.left, y: e.pageY - offset.top };


        var ctx = win_ctx_map.get(e.view);
        if(!ctx) return;

        //console.log(pos)

        var imageData = ctx.getImageData(pos.x, pos.y, 1, 1).data;
        var hex = "#" + ("000000" + rgbToHex(imageData[0], imageData[1], imageData[2])).slice(-6);

        // $input.val(hex).change();
        $input.spectrum('set', hex);

        if (options['mouse-moved']) options['mouse-moved']($input);
        var spectrum = $input.spectrum('get');

        //if (spectrum.isValid()) $colorPickerOverlay.data('color-value', getNewColor(spectrum));
    }

    var pickColorOverlayCLicked = function (e) {
        delayed.runAllNow();
        //console.log('drag end')
        removeOverlay();
        pick_color = false;
        //$colorPickerOverlay.data('color-picked', true);
    }

    var pick_color = false;
    var drag_ghost = null;

    var mode = 'click';

    $input.on('show.spectrum', function (e, tinycolor) {
        pick_color = false;
        var spectrum = $input.spectrum('get');
        var $c = $input.spectrum('container');
        var $buttons = $c.find('.sp-button-container');
        var $pick = $buttons.find('.pg-sp-pick-color');
        if($pick.length == 0) {
            $pick = $('<div class="pg-sp-pick-color" draggable="true"><i class="icon icon-pipet"></i></div>');

            if(isApp()) $pick.prependTo($buttons);

            var $system_pick = $('<div class="pg-sp-pick-color-system"><input type="color"/></div>');

            if(!pgf.kids) {
                $system_pick.prependTo($buttons);
            }

            $pick_input = $system_pick.find('input');

            $pick_input.on('input change', function(e) {
                var val = $pick_input.val();
                $input.spectrum('set', val);
                if (options['mouse-moved']) options['mouse-moved']($input);
            })

            const multi = 'MULTIPLE WINDOWS OPEN: ';

            if(pgGetAppWindows().length > 1) mode = 'drag';

            addTooltip($pick, mode === 'click' ? 'Click to pick a color from anywhere in Pinegrow window.' : multi + '<b>Drag the picker icon</b> over Pinegrow windows to pick the color.');

            addTooltip($system_pick, 'Click to open system color picker.')

            var onDragOver = function(e) {

            }

            if(mode !== 'click') {
                $pick.on('click', function (e) {
                    e.preventDefault();
                    pinegrow.showQuickError(multi + 'Drag the picker icon over Pinegrow windows to pick the color.');
                })
            }

            $pick.on(mode === 'click' ? 'click' : 'dragstart', function (e) {
                //$pick.on('click', function (e) {
                //e.preventDefault();
                pick_color = !pick_color;

                pinegrow.showMode('Pick a color from the screen', pick_color);

                if(pick_color) {
                    $pick.addClass('active');

                    var img = document.createElement("img");

                    img.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
                    $pick.closest('body').append(img);
                    drag_ghost = img;

                    e.originalEvent.dataTransfer && e.originalEvent.dataTransfer.setDragImage(img, 0, 0);

                    pgGetAppWindows().forEach(function(win) {
                        $(win.window).on(mode === 'click' ? 'mousemove' : 'dragover', onMoved);

                        pinegrow.takePhotoOfUI(function (canvas, data, datauri) {

                            if(!pick_color) return; //dragend before image was captured

                            var $replacerEl = $input.siblings('.sp-replacer');
                            if (!no_check && ($replacerEl.length == 0 || !$replacerEl.hasClass('sp-active'))) return;

                            var ctx = canvas.getContext('2d');

                            win_ctx_map.set(win.window, ctx);

                            var $colorPickerOverlay = $('<div class="color-picker-overlay close"></div>');

                            $colorPickerOverlay.appendTo(win.window.document.body);

                            /*
                            $colorPickerOverlay
                                .off('click', colorPicked)
                                .on('click', colorPicked);
*/
                            $colorPickerOverlay.removeClass('close');
                            $colorPickerOverlay.data('selected-jq-element', $input);

                            //$colorPickerOverlay[0].addEventListener("mousemove", pickColorUnderMouse, false);
                            //$colorPickerOverlay[0].addEventListener("click", pickColorOverlayCLicked, false);
                        }, true, win);
                    })

                    $($pick.get(0).ownerDocument).one(mode === 'click' ? 'click' : 'dragend', pickColorOverlayCLicked);
                } else {
                    removeOverlay();
                    $pick.removeClass('active');
                }
            })

        } else {
            $pick.removeClass('active');
        }
    });

    var removeOverlay = function() {
        pgGetAppWindows().forEach(function(win) {
            $(win.window.document.body).find('.color-picker-overlay').remove();
        })
        if(drag_ghost) {
            $(drag_ghost).remove();
            drag_ghost = null;
        }
        return;

        if ($colorPickerOverlay) {
            $colorPickerOverlay.addClass('close');

            //$colorPickerOverlay[0].removeEventListener("mousemove", pickColorUnderMouse, false);
            //$colorPickerOverlay[0].removeEventListener("click", pickColorOverlayCLicked, false);

            if (pick_color && $colorPickerOverlay.data('color-picked')) {
                $(this).spectrum('set', $colorPickerOverlay.data('color-value'));
                $colorPickerOverlay.data('color-value', '');
                $colorPickerOverlay.data('color-picked', false);
            }
        }
    }

    $input.on('hide.spectrum', function (e, tinycolor) {
        removeOverlay();

        pgGetAppWindows().forEach(function (win) {
            $(win.window).off(mode === 'click' ? 'mousemove' : 'dragover', onMoved);
        });
    });

    this.remove = function () {
        if ($colorPickerOverlay && $colorPickerOverlay.length > 0) {
            $colorPickerOverlay.off('click', colorPicked);
            $colorPickerOverlay.remove();
        }

        _this = null;
        colorPicked = null;
    }
}
