window.PgStylesheetHelper = function () {

    var autocompleteSCSSProp = null;

    $.getJSON( "scss-prop.json", function( data ) {
        autocompleteSCSSProp = data;
    });

    var checkIfStylesheetLoaded = function (doc, cssnum, stylesheetLoadedCallback, timeoutCallback, done) {
        var interval_id = setInterval( function() {
                if (doc.styleSheets.length > cssnum) {
                    try {
                        clearInterval( interval_id );
                        clearTimeout( timeout_id );

                        if (stylesheetLoadedCallback) {
                            if (stylesheetLoadedCallback()) {
                                if (done) done(true);
                            }
                        }
                        else {
                            if (done) done(true);
                        }
                    } catch( e ) {} finally {}
                }
            }, 10 ),
            timeout_id = setTimeout( function() {
                clearInterval( interval_id );
                clearTimeout( timeout_id );

                if (timeoutCallback) timeoutCallback();
                if (done) done(false);
            }, 10000 );
    }

    var getStylesheetFromTag = function (stylesheets, tag) {
        for (var i = 0; i < stylesheets.length; i++) {
            if (stylesheets[i].ownerNode == tag) return stylesheets[i];
        }
    }

    return {
        getStylesheetFromNode: function (node) {
            var root = pinegrow.pgPostCSSHelper.getRootNodeFor(node);
            return root.pgStylesheet;
        },
        getNormalizedSelectorForStylesheetRule: function (rule) {
            if(!rule.selectorText) return null;
            if(!rule.crsaNormalizedSelector) {
                rule.crsaNormalizedSelector = normalizeSelector(rule.selectorText);
            }
            return rule.crsaNormalizedSelector;
        },
        loadStyleSheet: function(page, path, fn, replacePgel, first_css) {//csspgdone
            var doc = page.getDocument();

            var pgHead = page.sourceNode.findOne('head');
            if(!pgHead) return null;
            var code = '<link href="' + path + '" rel="stylesheet" type="text/css">';
            // var pgLink = pgCreate(code);
            var pgLink = pgCreateNodeFromHtml(code);

            var cssnum = doc.styleSheets.length - 1;
            checkIfStylesheetLoaded(doc, cssnum,
                function () {               // stylesheetLoaded callback
                    return pgLink;
                }, function () {            // timeout callback
                    pgLink.remove();
                }, function (success) {     // done callback
                    if (success) fn(pgLink);
                    else         fn(null);
                });

            if (replacePgel) {
                replacePgel.replaceWith(pgLink);
            } else {
                var insert_before = null;
                if(first_css) {
                    insert_before = pgHead.findOne('link[rel="stylesheet"]');
                }
                if(insert_before) {
                    pgLink.insertBefore(insert_before);
                } else {
                    pgHead.append(pgLink);
                }
            }

            return pgLink; // return the link node;
        },
        loadInlineStylesheet: function (page, fn, replacePgel) {//csspgdone

            var pgHead = page.sourceNode.findOne('head');
            if(!pgHead) return null;
            var code = '<style></style>';
            var pgStyleTag = pgCreate(code);

            var doc = page.activeView.getDocument();

            var cssnum = doc.styleSheets.length;
            checkIfStylesheetLoaded(doc, cssnum, null, null,
                function (success) {     // done callback
                    var styleTag = pgStyleTag.getDOMElement();
                    //if (success) fn(getStylesheetFromTag(doc.styleSheets, styleTag));
                    if (success) fn(pgStyleTag)
                    else         fn(null);
                });

            if (replacePgel) {
                replacePgel.replaceWith(pgStyleTag);
            } else {
                pgHead.append(pgStyleTag);
            }

            return pgStyleTag;
        },
        getAllStylesheetsFor: function (pgStylesheet) {
            var stylesheets = [];
            var forAll = function (stylesheet) {
                if (stylesheet.ignore && (!stylesheet.sourceStylesheet || stylesheet.sourceStylesheet.ignore)) return;

                if (stylesheet.sourceStylesheet) stylesheet = stylesheet.sourceStylesheet;
                stylesheets.push(stylesheet);

                var importedFiles = stylesheet.importedFiles;
                for (var i = 0; i < importedFiles.length; i++) {
                    forAll(importedFiles[i]);
                }
            }

            forAll(pgStylesheet);
            return stylesheets;
        },
        getSCSSAutocomleteProps: function () {
            return autocompleteSCSSProp;
        },
        getParentStylesheetBySelector: function (selectorList, page) {
            var selectors = selectorList.selectors;

            for (var i = 0; i < selectors.length; i++) {
                var selector = selectors[i].text;

                var matched = selector.match(/pg-main-(.*)/);
                if (matched && matched.length > 0) {
                    var index = matched[1];
                    var mainUniqueNameCS = pinegrow.stylesheets.findCrsaStylesheetForUniqueName(index);
                    return mainUniqueNameCS.getMainParnetStylesheetFor(page);
                }
            }

            return null;
        },
        getImportedStylesheetBySelector: function (pgStylesheet, selectorList) {
            var selectorText = selectorList.text;
            var matched = selectorText.match(/pg-file-(.*)/);
            if (matched && matched.length > -1) {
                return pgStylesheet.getImportedStylesheetsByUniqueName(matched[1]);
            }

            return null;
        }
    }
}

window.PgPostCSSHelper = function () {
    var allPostcssLib = require('postcss');
    var postcss = allPostcssLib.postcss;
    var cssSyntax = allPostcssLib.postcssSafeParser;

    var fillRawsInfos = function (helperNode, stylesheet) {
        var newNode = {
            raws: {}
        };
        if (helperNode && helperNode.raws && helperNode.type !== 'root') {
            newNode.raws.before = helperNode.raws.before;
            if (helperNode.nodes && helperNode.nodes.length == 0) {
                newNode.raws.before = newNode.raws.before.replace(/\n\n/, '\n');
                newNode.raws.before += "  ";
            }
        }
        else {
            newNode.raws.before = "\n";
        }
        if (stylesheet) stylesheet.setNodeNeedFormatting(newNode, true, false);
        return newNode;
    }

    return {
        getNodePositionInSource: function(node) {
            var r = {
                start: {
                    ch: 0,
                    line: 0
                },
                end: {
                    ch:0,
                    line:0
                }
            }
            var cs = pgcssh.getStylesheetForNode(node);

            if(!node.parent) return r;

            node.pg_highlight = true;
            var css = cs.getSourceForRules();

            //node.parent.nodes.splice(idx + 2, 1);
            //node.parent.nodes.splice(idx, 1);

            //console.log(css)
            delete node.pg_highlight;

            var start = '/*PG_START*/';
            var end = '/*PG_END*/';

            var has_start = false;

            var lines = css.split("\n");
            for(var i = 0; i < lines.length; i++) {
                var idx;

                if(!has_start) {
                    idx = lines[i].indexOf(start);
                    if (idx >= 0) {
                        r.start.ch = idx + 1;
                        r.start.line = i + 1;
                        has_start = true;
                    }
                }
                if(has_start) {
                    idx = lines[i].indexOf(end);
                    if(idx >= 0) {
                        r.end.line = i + 1;
                        r.end.ch = r.start.line === i + 1 ? idx - start.length : idx + 1;
                        break;
                    }
                }
            }
            return r;
        },
        getSource: function (rule, stringifier, withIds) {
            return rule.toString(stringifier);
        },
        wrapNodeWithRoot: function (nodes) {
            return postcss.root({ nodes: nodes });
        },
        parsedCSSSourceFromText: function (source, done) {
            var syntaxRes = postcss().process(source, { parser: cssSyntax });
            syntaxRes.then(function (postcssRes) {
                var postcssSource = postcssRes.css;
                var rules = postcssRes.root;

                if (done) done(null, postcssSource, rules);
            });

            if (syntaxRes.error) {
                var parsingError = syntaxRes.error.reason + ":" + syntaxRes.error.line;
                if (done) done(parsingError);
            }
        },
        isCommented: function (node) {
            return (node.type == "comment" || node.ignore);
        },
        getSelector: function (rule) {
            var getNormalizedSelectorForCrsaRule = function(rule) {
                if(!rule.selector) return "";
                if(!rule.crsaNormalizedSelector) {
                    rule.crsaNormalizedSelector = normalizeSelector(rule.selector);
                }
                return rule.crsaNormalizedSelector;
            }

            return getNormalizedSelectorForCrsaRule(rule);
        },
        forAllNodes: function (node, callback) {
            if (!callback) return;
            callback(node);

            var forAll = function (nodes) {
                for (var i = 0; i < nodes.length; i++) {
                    var node = nodes[i];
                    callback(node);

                    if (node.nodes && node.nodes.length > 0) {
                        forAll(node.nodes);
                    }
                }
            }

            if (node.nodes) forAll(node.nodes);
        },
        getNodeFromText: function (text, helperNode, type, stylesheet) {
            if (!text || text.length == 0) return null;

            var node = fillRawsInfos(helperNode, stylesheet);

            text = text.trim();
            var postcssNode;

            var topLvlNode = helperNode && helperNode.type == 'root';

            var hasParams = text.indexOf('(') > -1 && text.indexOf('(') < text.indexOf(':');
            var isRule = (text.indexOf('.') > -1 && text.indexOf(':') < 0) || text[text.length-1] == '{' || (topLvlNode && text[0] != '$' && text[0] != '@');
            var isComment = text.substr(0, 2) == '/*' || text.substr(text.length - 2, text.length) == "*/";

            if ((text[0] == '@' && (text.indexOf(':') < 0 || hasParams)) || type == "atrule") {
                if (!type || text[0] == '@') text = text.substr(1, text.length).trim();
                node.name = text;
                node.params = "";
                node.raws.afterName = " ";
                node.raws.after = "\n";
                postcssNode = postcss.atRule(node);
            } else if (isComment || type == "comment") {
                postcssNode = postcss.comment({
                    text: text.substr(2, text.length - 4).trim()
                });
            } else if (isRule || type == "rule") {
                if (text[text.length-1] == '{') text = text.substr(0, text.length-1).trim();
                node.selector = text;
                node.raws.between = " ";
                node.raws.after = node.raws.before.replace(/\n\n/, '\n');
                node.nodes = [];
                postcssNode = postcss.rule(node);
            } else {
                var ruleChoosed = false;
                if (text.indexOf(':') == -1) {
                    if (helperNode) {
                        if (helperNode.type == "rule" && helperNode.nodes.length > 0) {
                            postcssNode = pinegrow.pgPostCSSHelper.createNode('rule', text, null, helperNode, stylesheet);
                            ruleChoosed = true;
                        }
                        else if (helperNode.type != "decl" && helperNode.type != "root") {
                            var helperNodeSiblings = helperNode.parent.nodes;
                            var helperNodeIndex = helperNodeSiblings.indexOf(helperNode);
                            var newHelperNode = helperNodeSiblings[helperNodeIndex + 1];
                            if (newHelperNode && newHelperNode.type == "rule") {
                                postcssNode = pinegrow.pgPostCSSHelper.createNode('rule', text, null, helperNode, stylesheet);
                                ruleChoosed = true;
                            }
                        }
                    }
                }

                if (!ruleChoosed) {
                    node.prop = text;
                    if (text[text.length-1] == ':') {
                        //node.prop = text.substr(0, text.length-1).trim();
                        pgcssh.setNodeProperty(node, 'prop', text.substr(0, text.length-1).trim())
                    }
                    //node.value = "";
                    pgcssh.setNodeProperty(node, 'value', "")
                    node.raws.after = "\n";
                    node.raws.between = ": ";
                    postcssNode = postcss.decl(node);
                }
            }

            return postcssNode;
        },
        copyNodeRawsTo: function (copyTo, copyFrom) {
            copyTo.raws.before = copyFrom.raws.before;
            return copyTo;
        },
        getNodeIndices: function (node) {
            if (!node) return null;

            var indices = [];
            var currentNode = node;
            while (currentNode.type != "root") {
                if (currentNode.parent) {
                    indices.unshift(currentNode.parent.nodes.indexOf(currentNode));
                    currentNode = currentNode.parent;
                }
                else {
                    break;
                }
            }
            return indices;
        },
        getNodeByIndices: function (root, indices) {
            var currentNode = root;
            for (var i = 0; i < indices.length && currentNode; i++) {
                if (currentNode && currentNode.nodes) {
                    currentNode = currentNode.nodes[parseInt(indices[i])];
                }
                else {
                    currentNode = null;
                    break;
                }
            }
            return currentNode;
        },
        getNodeLevel: function (node) {
            var level = 0;
            var currentNode = node;
            while (currentNode.parent && currentNode.parent.type != "root" ) {
                level++;
                currentNode = currentNode.parent;
            }
            return level;
        },
        getNodeLevelFromIndices: function (nodeIndices) {
            var level = nodeIndices.length - 1
            return level < 0 ? null : level;
        },
        getLastDeclIndex: function (parent) {
            var nodes = parent.nodes;
            var i = -1;
            for (i = nodes.length - 1; i >= 0; i--) {
                var node = nodes[i];
                if (pgcssh.isVariable(node) || node.type == 'decl') {
                    break;
                }
            }

            if (i < 0) i = -1;
            return i;
        },
        replaceNodeWith: function (node, newNode) {
            var index = node.parent.nodes.indexOf(node);
            newNode.parent = node.parent;
            node.parent.nodes[index] = newNode;
            newNode.pg_old_values = node.pg_old_values || {};
            return newNode;
        },
        replaceNodeWithAndUpdatePgId: function (node, newNode) {
            newNode = pinegrow.pgPostCSSHelper.replaceNodeWith(node, newNode);
            newNode['pg-node-id'] = node['pg-node-id'];
            newNode.pg_old_values = node.pg_old_values || {};
            // pinegrow.pgPostCSSHelper.addPgIdToSelector([newNode]);
            return newNode;
        },
        insertNodeAt: function (node, parent, index) {
            node.parent = parent;
            parent.nodes.splice(index, 0, node);
            return node;
        },
        insertNodesAt: function (nodes, parent, index) {
            for (var i = 0; i < nodes.length; i++) {
                pinegrow.pgPostCSSHelper.insertNodeAt(nodes[i], parent, index);
                index++;
            }
        },
        insertNodeBefore: function (node, beforeNode) {
            node.parent = beforeNode.parent;
            var nodeIndex = node.parent.nodes.indexOf(beforeNode);
            node.parent.nodes.splice(nodeIndex, 0, node);
            return node;
        },
        insertNodeAfter: function (node, beforeNode) {
            node.parent = beforeNode.parent;
            var nodeIndex = node.parent.nodes.indexOf(beforeNode);
            node.parent.nodes.splice(nodeIndex + 1, 0, node);
            return node;
        },
        insertNodeInside: function (node, parentNode) {
            node.parent = parentNode;
            if (!parentNode.nodes) parentNode.nodes = [];
            parentNode.nodes[parentNode.nodes.length] = node;
            return node;
        },
        insertNodesInside: function (nodes, parentNode) {
            for (var i = 0; i < nodes.length; i++) {
                pinegrow.pgPostCSSHelper.insertNodeInside(nodes[i], parentNode);
            }
        },
        isNodeEqualOrChildOf: function (node, newNode) {
            if(!node || !newNode) return false;
            var nodeStylesheet = pgStylesheetHelper.getStylesheetFromNode(node);
            var newNodeStylesheet = pgStylesheetHelper.getStylesheetFromNode(newNode);


            var isNodeChildren = false;
            if (nodeStylesheet && newNodeStylesheet && nodeStylesheet.getUniqueName() == newNodeStylesheet.getUniqueName()) {
                var newNodeObj = {
                    pgStylesheet: newNodeStylesheet,
                    node: newNode
                }

                newNodeObj['nodeIndices'] = pinegrow.pgPostCSSHelper.getNodeIndices(newNodeObj.node);
                newNodeObj['level']   = pinegrow.pgPostCSSHelper.getNodeLevelFromIndices(newNodeObj['nodeIndices']);

                newNodeObj['currLevelIndex'] = parseInt(newNodeObj.nodeIndices[newNodeObj.level]) || 0;
                newNodeObj['currParentLevelIndex'] = parseInt(newNodeObj.nodeIndices[newNodeObj.level - 1]) || 0;

                var selectedNode = {
                    pgStylesheet: nodeStylesheet,
                    node: node,
                    nodeIndices: pinegrow.pgPostCSSHelper.getNodeIndices(node)
                }

                selectedNode['level'] = pinegrow.pgPostCSSHelper.getNodeLevelFromIndices(selectedNode['nodeIndices']);

                selectedNode['currLevelIndex'] = parseInt(selectedNode.nodeIndices[newNodeObj.level]) || 0;
                selectedNode['currParentLevelIndex'] = parseInt(selectedNode.nodeIndices[newNodeObj.level - 1]) || 0;

                if (selectedNode.level <= newNodeObj.level) {

                    isNodeChildren = true;
                    for (var i = 0; i < selectedNode.level + 1; i++) {
                        if (selectedNode.nodeIndices[i] != newNodeObj.nodeIndices[i]) {
                            isNodeChildren = false;
                            break;
                        }
                    }
                }
            }

            return isNodeChildren;
        },
        getNodesWithRegex: function (type, prop, reg, node) {
            var nodesFromReg = [];
            var length = 0;
            var children = node.nodes;
            if (children) length = children.length;

            for (var i = 0; i < length; i++) {
                var currNode = children[i];
                if (currNode.type == type && currNode[prop]) {
                    var matched = currNode[prop].match(reg);
                    if (matched && matched.length > 1 && matched[1] == currNode[prop]) {
                        nodesFromReg.push(currNode);
                    }
                }
            }
            return nodesFromReg;
        },
        removeNodeFrom: function (node, parentNode) {
            if (!node) return;
            parentNode = parentNode || node.parent;

            var nodeIndex = parentNode.nodes.indexOf(node);
            parentNode.nodes.splice(nodeIndex, 1);
            node.is_removed = true;
            return node;
        },
        removeNodeWidthRegexFrom: function (type, prop, reg, node) {
            var children = node.nodes;
            for (var i = 0; i < children.length; i++) {
                var currNode = children[i];
                if (currNode.type == type && currNode[prop]) {
                    var matched = currNode[prop].match(reg);
                    if (matched && matched.length > 1 && matched[1] == currNode[prop]) {
                        pinegrow.pgPostCSSHelper.removeNodeFrom(currNode);
                        i--;
                    }
                }
            }
        },
        copyNode: function (node) {
            var stylesheet = pgcssh.getStylesheetForNode(node);

            if (node.type == "comment") {
                return PgPostCSSHelper().createNode(node.type, node.text, null, node, stylesheet);
            } else if (node.type == "decl") {
                return PgPostCSSHelper().createNode(node.type, node.prop, node.value, node, stylesheet);
            } else if (node.type == "rule") {
                return PgPostCSSHelper().createNode(node.type, node.selector, null, node, stylesheet);
            } else if (node.type == "atrule") {
                return PgPostCSSHelper().createNode(node.type, node.name, node.params, node, stylesheet);
            }
        },
        createNode: function (type, frVal, sdVal, helperNode, stylesheet) {
            var node = fillRawsInfos(helperNode, stylesheet);
            var postcssNode;

            if (type == "comment") {
                node.text = frVal;
                postcssNode = postcss.comment(node);
                postcssNode.raws = {
                    begin: "/*",
                    content: "/* " + frVal + " */"
                }
            }
            if (type == "decl") {
                node.prop = frVal;
                node.value = sdVal;
                node.raws.between = ": ";
                postcssNode = postcss.decl(node);
            } else if (type == "rule") {
                node.selector = frVal;
                node.raws.between = " ";
                node.raws.after = node.raws.before.replace(/\n\n/, '\n');
                node.nodes = [];
                postcssNode = postcss.rule(node);
            } else if (type == "atrule") {
                node.name = frVal;
                node.params = sdVal;
                node.raws.between = " ";
                node.raws.afterName = " ";
                node.raws.after = "\n";
                postcssNode = postcss.atRule(node);
            }

            return postcssNode;
        },
        renameRuleSelector: function (rule, selector) {
            if (rule.type == "rule" && rule.selector) {
                rule.selector = selector;
                return true;
            }
            return false;
        },
        closestNode: function (node, type) {
            var parent = node.parent;
            while (parent && parent.type != type) {
                parent = parent.parent;
            }

            return parent;
        },
        getNodeType: function (node) {
            if (node.type == 'atrule' && node.name == 'import') return 'import';

            return node.type;
        },
        getRootNodeFor: function (rule) {
            var currentNode = rule;
            while(currentNode.type != "root") {
                if (currentNode.parent) {
                    currentNode = currentNode.parent;
                }
                else {
                    break;
                }
            }
            return currentNode;
        },
        getTopLevelNodeFor: function (node) {
            var currentNode = node;
            while(currentNode.parent.type != "root") {
                if (currentNode.parent) {
                    currentNode = currentNode.parent;
                }
                else {
                    break;
                }
            }
            return currentNode;
        },
        getStylesheetForNode: function(node) {
            var root = this.getRootNodeFor(node);
            return root.pgStylesheet || null;
        },
        getChildrenDeclWithPropName: function (rule, declProp) {
            var props = [];

            for (var i = 0; i < rule.nodes.length; i++) {
                var currentNode = rule.nodes[i];
                if (currentNode && currentNode.type == "decl" && currentNode.prop == declProp) {
                    props.push(currentNode);
                    // return currentNode;
                }
            }
            return props;
            // return null;
        },
        addPgIdToSelector: function (nodes) {
            var addToAllNodes = function (nodes) {
                for (var i=0; i < nodes.length; i++) {
                    if (nodes[i].type == "rule") {
                        nodes[i]['selector'] = nodes[i]['selector'].replace(/\,\s\.pg-node-id-[0-9]*/g, '');
                        if (nodes[i]['pg-node-id']) {
                            nodes[i]['selector'] += ", .pg-node-id-" + nodes[i]['pg-node-id'];
                        }
                    }

                    if (nodes[i].nodes && nodes[i].nodes.length > 0) {
                        addToAllNodes(nodes[i].nodes);
                    }
                }
            }

            addToAllNodes(nodes);
        },
        shouldAddPlaceholder: function (rule) {
            var addPlaceholder = true;
            var nodes = rule.nodes;
            if (nodes) {
                for (var i = 0; i < nodes.length; i++) {
                    if (nodes[i].type == 'decl') {
                        addPlaceholder = false;
                        break;
                    }
                }
            }
            else {
                addPlaceholder = false;
            }
            return addPlaceholder;
        },
        addEmptyPlaceholderForEmptyRule: function(code, addPlaceholder) {
            if (addPlaceholder) {
                var bracketIndex = code.indexOf('{');
                if (bracketIndex > -1) {
                    code = code.substr(0, bracketIndex + 1) + '--pg-empty-placeholder:0px;' + code.substr(bracketIndex + 1, code.length);
                }
            }
            return code;
        },
        getMediaNode: function (node) {
            if (node.parent) {
                var parentNode = node.parent;
                while(parentNode) {
                    if (parentNode.type === "atrule" && parentNode.name === "media") {
                        return parentNode;
                    }
                    parentNode = parentNode.parent || null;
                }
                return null;

                if (parentNode.type == "atrule" && parentNode.name == "media") {
                    return parentNode;
                } else if(parentNode.type == "atrule" && parentNode.name == "supports" && parentNode.parent) {
                    if(parentNode.parent.type == 'atrule' && parentNode.parent.name == 'media') {
                        return parentNode.parent;
                    }
                }
            }
            return null;
        },
        getSupportsAtruleNode: function (node) {
            if (node.parent) {
                var parentNode = node.parent;
                while(parentNode) {
                    if (parentNode.type === "atrule" && parentNode.name === "supports") {
                        return parentNode;
                    }
                    parentNode = parentNode.parent || null;
                }
                return null;

                if (parentNode.type == "atrule" && parentNode.name == "supports") {
                    return parentNode;
                } else if(parentNode.type == "atrule" && parentNode.name == "media" && parentNode.parent) {
                    if(parentNode.parent.type == 'atrule' && parentNode.parent.name == 'supports') {
                        return parentNode.parent;
                    }
                }
            }
            return null;
        },
        getAtRuleValue: function (atRule) {
            if (atRule && atRule.type == "atrule") {
                return atRule.params;
            }
            return null;
        },
        getRuleForNode: function (node) {
            var currNode = node;
            if (currNode.type != "rule" && currNode.type != "root") {
                currNode = currNode.parent;
            }

                 if (currNode.type == "root") return null;
            else if (currNode.type == "rule") return currNode;
        },
        setNodeProperty: function(node, prop, value) {
            node.pg_old_values = node.pg_old_values || {}
            if(!(prop in node.pg_old_values)) {
                node.pg_old_values[prop] = node[prop] || null;
            }
            node[prop] = value;
        },
        getNodeOldValues: function(node) {
            return null; //disabled for now
            var r = node.pg_old_values || {};
            node.pg_old_values = {};
            return r;
        },
        getValuesFromNode: function (node) {
            if (node.type == 'decl') {
                return { prop: node.prop, value: node.value };
            }
            else if (node.type == 'rule') {
                return { selector: node.selector };
            }
            else if (node.type == 'atrule') {
                return { name: node.name, param: name.param };
            }
            else if (node.type == 'import') {
                return { name: node.name, param: name.param };
            }
            else if (node.type == 'comment') {
                return { text: node.text };
            }
        },
        copyOldValuesToNode: function(node, old_values) {
            if(old_values) {
                $.each(old_values, function(prop, value) {
                    node[prop] = value;
                })
            }
        },
        getSelectorWithoutPgId: function(selector) {
            return selector.replace(/,\s?\.pg\-node\-id\-[0-9]+/,'');
        },
        getNodeDisplayName: function(node) {
            if(!node) return 'node';
            var rule = this.getRuleForNode(node);
            var name = '';
            if (rule) {
                if(rule.selector) {
                    name = this.getSelectorWithoutPgId(rule.selector);
                } else if(rule.raw) {
                    name = rule.raw;
                }
            }
            else if (node.type == "comment") {
                name = node.text;
            }
            else if (node.type == "atrule") {
                name = node.name + ' ' + node.params;
            }
            name = shortenString(name, 10);
            if(rule != node && node.type == 'decl') {
                name += ' ' + node.prop;
            }
            return name;
        },
        getRuleDisplayName: function(r) {
            var p = r;
            var name = '';
            while(p.parent && p.type !== 'root') {
                if(p.type === 'rule') {
                    name = this.getSelectorWithoutPgId(p.selector) + (name.length ? ' ' : '') + name;
                } else if(p.type === 'atrule' && (p.name === 'media' || p.name === 'supports')) {
                    name = p.params + (name.length ? ' ' : '') + name;
                }
                p = p.parent;
            }
            return name;
        },
        filterStylesheetsList: function(list, options) {
            options = $.extend({
                loaded: true,
                ignored: true,
                imported: true,
                inline: true,
                locked: true,
                remote: true
            }, options || {} )

            var r = [];
            list.forEach(function(cs) {
                if(!cs.loaded && !options.loaded) return;
                if(cs.ignore && !options.ignored) return;
                if(cs.isImported() && !options.imported) return;
                if(cs.inline && !options.inline) return;
                if(cs.locked && !options.locked) return;
                if(!options.remote && cs.isRemote()) return;
                r.push(cs);
            })
            return r;
        },
        getClassesFromSelector: function(selector) {
            //support .\32 class\:aaa:hover ...
            var re = /\.(\\3[0-9]+\s[a-z0-9\\\:\-]+|[a-z0-9\\\:\-]+)/g;
            var pre = /([a-z0-9])(:[a-z:\-]+)/;
            var sre = /[\.\\]/g;

            var r = [];

            var m = selector.match(re);
            if(m) {
                for (var j = 0; j < m.length; j++) {
                    if(!m[j].startsWith('.pg-state-') && !m[j].startsWith('.pg-node-') && !m[j].startsWith('.pg-main-cs-')) {
                        var cls = m[j].replace(/\\3([0-9]+)\s?/, '$1');
                        cls = cls.replace(pre, '$1');
                        cls = cls.replace(sre, '');
                        r.push(cls);
                    }
                }
            }

            /*
            var m = this.getSelectorWithoutPgId(selector).match(/\.[^\(\)\#\.\:\s\"\,]+/g);
            if(!m) return [];
            var r = [];
            m.forEach(function(cls) {

                if(!cls.startsWith('.pg-state-') && !cls.startsWith('.pg-node-') && !cls.startsWith('.pg-main-cs-')) {
                    cls = cls.replace('.', '');
                    if(r.indexOf(cls) < 0) {
                        r.push(cls)
                    }
                }
            })
            */
            return r;
        },
        getUrlOfStylesheetForOverrides: function(page, suffix) {
            suffix = suffix || '';
            var pi = null;
            var parent_url = null;
            var key = 'urlOfStylesheetForOverrides' + suffix;

            var returnIfLinkedToPage = function(url) {
                var page = pinegrow.getSelectedPage();
                if(page) {
                    if(!page.hasAttachedStylesheetWithUrl(url)) {
                        url = '';
                    }
                }
                return url;
            }

            if(page) {
                if(page[key]) {
                    return returnIfLinkedToPage(page[key]);
                }
                parent_url = page.url;
                pi = page.getProjectInfo();
            } else {
                var p = pinegrow.getCurrentProject();
                if (p) {
                    if(p[key]) {
                        return returnIfLinkedToPage(p[key]);
                    }
                    pi = p.getProjectInfo();
                    parent_url = p.getUrl() + '/';
                }
            }
            if(pi) {
                var rel_v = pi.getSetting('cs-for-overrides' + suffix) || '';
                if(rel_v.length && parent_url) {
                    rel_v =  crsaGetAbsoluteUrlForRelativeUrlAndParentUrl(rel_v, parent_url);
                }
                return returnIfLinkedToPage(rel_v);
            }
            return returnIfLinkedToPage(this[key] || '');
        },
        setUrlOfStylesheetForOverrides: function(v, page, suffix) {
            suffix = suffix || '';
            var pi = null;
            var parent_url = null;
            var key = 'urlOfStylesheetForOverrides' + suffix;
            if(page) {
                pi = page.getProjectInfo();
                page[key] = v;
                parent_url = page.url;
            } else {
                var p = pinegrow.getCurrentProject();
                if (p) {
                    pi = p.getProjectInfo();
                    p[key] = v;
                    parent_url = p.getUrl();
                }
            }
            if(pi) {
                if(parent_url) {
                    var rel_v = crsaMakeLinkRelativeTo(v, parent_url);
                }
                pi.setSetting('cs-for-overrides' + suffix, rel_v);
                pi.save();
            }
            this[key] = v;
        },
        getGlobalNodeId: function(node) {
            var cs = this.getStylesheetForNode(node);
            return {
                csurl: cs ? cs.url : null,
                indices: this.getNodeIndices(node)
            }
        },
        getNodeFromGlobalId: function(nodeid) {
            if(nodeid && nodeid.csurl && nodeid.indices) {
                var cs = pinegrow.stylesheets.getByUrl(nodeid.csurl);
                if(cs && cs.getRootNode()) {
                    return this.getNodeByIndices(cs.getRootNode(), nodeid.indices);
                }
            }
        },
        findNodeInside: function (type, propName, value, parent, untilIndex) {
            var nodes = parent.nodes;
            untilIndex = untilIndex || nodes.length;

            for (var i = 0; i < untilIndex; i++) {
                var node = nodes[i];
                if (node.type == type) {
                    if (node[propName] == value) return node;
                }
            }

            return null;
        },
        searchInNode: function(parserNode, filter_re, pgStylesheet) {

            var containInFilter = function (val) {
                return val && val.match(filter_re);
            }

            if (parserNode.type == "comment") {
                if (!containInFilter(parserNode.text)) {
                    return false;
                }
            } else if (parserNode.type == "decl") {
                if (!containInFilter(parserNode.prop) && !containInFilter(parserNode.value)) {
                    return false;
                }
            } else if (parserNode.type == "atrule") {
                if (!containInFilter(parserNode.name) && !containInFilter(parserNode.params)) {
                    return false;
                }
            } else if (parserNode.type == "rule") {
                var selector = parserNode.selector;

                var ret = pgStylesheet.filter(parserNode);
                if (ret.haveTo == "update") {
                    selector = ret.newValue;
                }
                if ((ret.haveTo == "delete") || !containInFilter(selector)) {
                    return false;
                }
            }
            return true;
        },
        searchInNodeAndSubnodes: function(parserNode, filter_re, pgStylesheet) {
            if(this.searchInNode(parserNode, filter_re, pgStylesheet)) return true;
            //check subnodes
            if(parserNode.nodes) {
                for(var i = 0; i < parserNode.nodes.length; i++) {
                    if(this.searchInNode(parserNode.nodes[i], filter_re, pgStylesheet)) {
                        return true;
                    }
                }
            }
            return false;
        },
        isMedia: function (node) {
            return node.type == 'atrule' && node.name == 'media';
        },
        isVariable: function(node) {
            var isDecl = node.type == 'decl'   && node.prop && node.prop.startsWith('$');
            var isVar  = node.type == 'atrule' && node.variable;

            return isDecl || isVar;
            // return node.type == 'decl' && node.prop && (node.prop.startsWith('$') || node.prop.startsWith('@')) && (!node.nodes || node.nodes.length == 0);
        },
        getVariableProp: function (node) {
            if (node.type == 'decl') return node.prop;
            else if (node.type == 'atrule') return node.name;

            return null;
        }
    }
}
