var PgSelectDefaultStylesheet = function(page, current_cs) {

    this.show = function(done) {
        var html = '<p>Please select the stylesheet to which you wish to add new CSS rules:</p>';

        html += '<div class="cs-selector"></div>';

        html += '<p>Pinegrow will automatically add new CSS rules to this stylesheet when you create new CSS rules in the Active Style panel.</p>';

        html += '<p>A tip: it is not a good idea to directly modify CSS files of plugins and frameworks, for example bootstrap.css. You will lose changes if you update the framework to a newer version. Instead, create a separate stylesheet and define all the custom styling for your website there.</p>';


        var $d = pinegrow.showAlert(html, 'Choose default stylesheet for new CSS rules', 'Cancel', 'OK', function () {
            //cancel
            done(current_cs, true);
        }, function () {
            //ok
            if(in_input) {
                var url = in_input.getValue();
                current_cs = pinegrow.stylesheets.getByUrl(url);

                if (current_cs) {
                    pinegrow.setDefaultStylesheet(page, current_cs);
                }
                done(current_cs, current_cs ? false : true);
            }
        })

        var $s = $d.find('.cs-selector');

        var cslist = page.getAttachedStylesheets(
            {
                loaded: true,
                ignored: false,
                imported: true,
                inline: false,
                locked: false
            })

        if(cslist.length) {

            var list = cslist.map(function (cs) {
                var short = crsaGetNameFromUrl(cs.url)
                return {name: short, key: cs.url, html: short + ' <small>' + cs.url + '</small>'};
            })
            var in_input = new PgInputField('stylesheet', {
                type: 'select',
                show_empty: false,
                name: 'Stylesheet',
                without_text: false,
                default_value: list[0],
                options: list
            });

            in_input.appendTo($s);

            if (current_cs) {
                in_input.setValue(current_cs.url)
            }

            in_input.on('input', function (e) {

            })
        } else {
            $s.html('<p><b>No unlocked stylesheets were found.</b> Unlock a local stylesheet or add a new stylesheet in <b>File -&gt; Manage stylesheets</b> first. Note, inline stylesheets can\'t be used as default stylesheet.</p>');
        }
    }
}

var PgSelectorControl = function(pgel) {
    this.$element = $('<div class="pg-selector-control"></div>')
}

var PgSelectorMakerView = function(pgel, aselector, opts) {

    var _this = this;

    var defaults = {
        stylesheets: null,
        button_label: 'Create',
        add_classes: true,
        simple: false,
        labels: true,
        media_query: pgf.kids ? false : true,
        compact: false,
        save_inline: false
    }

    var save_inline_active = false;

    this.onCreate = function(maker) {};
    this.onCancel = function(maker) {};
    this.onCreateWithoutSelector = null;

    this.getSelector = function() {
        return selector;
    }

    this.setSelector = function(sel) {
        input.setValue(sel);
        input.$input.trigger('input');
    }

    this.getStylesheet = function() {
        return in_input ? in_input.getValue() : null;
    }

    this.getAddClasses = function() {
        if(addc_input && addc_input.getValue()) {
            return add_classes_list;
        }
        return [];
    }

    this.getMediaQuery = function() {
        return mq_input ? mq_input.getValue() : null;
    }

    this.getSaveInline = function() {
        var val = save_inline_active && save_inline_input && save_inline_input.getValue();
        if(save_inline_input) {
            pinegrow.setSetting('create-css-rule-save-inline', (save_inline_input.getValue()) ? '1' : '0');
        }
        return val;
    }

    this.setSaveInline = function(v) {
        save_inline_active = v;

        if(selector) {
            if (v) {
                $saveinline.show();
            } else {
                $saveinline.hide();
            }
        }
    }

    var options = $.extend(defaults, opts);

    this.getOptions = function() {
        return options;
    }

    var $element = $('<div class="pg-selector-maker"></div>');

    var $inputs = $('<div class="pg-selector-maker-inputs"></div>').appendTo($element);

    var view = new PgUIView($element);

    this.view = view;

    var selector = aselector || '';
    var add_classes_list = [];

    var input_changed = false;

    var input = new PgInputField('selector', {
        type: 'text',
        show_empty: false,
        name: 'Selector',
        without_text: !options.labels,
        placeholder: '.myclass, #id, h1...',
        slider_def_unit: '',
        with_clear_icon: true,
        validate: function(obj, fn, value) {
            if(!value) {
                //return 'Which elements will be styled by this rule?';
            } else {
                if(pgf.kids) {
                    var tags = ['html','body','div','section','a','p','iframe','main','h1','h2','h3','h4','img','header','footer','hr','button'];
                    //prevent myclass
                    var m = value.match(/^[a-z]+$/i);
                    if (m) {
                        var tag = m[0].toLowerCase();

                        if (tags.indexOf(tag) < 0 && pinegrow.getSelectedPage().sourceNode.findOne(tag) === null) {
                            return `Do you want to create a class? In that case use <code>.${tag}</code>.`;
                        }
                    }
                    //prevent .body
                    var classes = pgcssh.getClassesFromSelector(value);
                    for (var i = 0; i < classes.length; i++) {
                        if (tags.indexOf(classes[i]) >= 0) {
                            return `Are you sure you want to create a class <code>.${classes[i]}</code>? Remove the dot to target all <code>${classes[i]}</code> elements?`;
                        }
                    }
                }
                try {
                    $element.get(0).querySelector(value);
                } catch(err) {
                    return 'This is not a valid CSS selector.';
                }
            }
        }
    });

    input.appendTo($element);

    input.on('input', function(e) {
        selector = input.getValue();
        updateMatcher(true);
        input.$field.removeClass('error');
        updateAddClasses();
        if(updateRuleCreatedIn) {
            updateRuleCreatedIn();
        }
        if($saveinline && save_inline_active) {
            if(selector) {
                $saveinline.show();
            } else {
                $saveinline.hide();
            }
        }
        //input_changed = sel != getSelector();
        return;

        var tokens = sel.split(/s+/);

        $list.find('a.selected').removeClass('selected');
        var $lis = $list.children();

        var lis_i = $lis.length - 1;

        if(lis_i >= 0) {
            var next_must_match = false;
            var node = $lis.get(lis_i);

            i = tokens.length - 1;

            while(i >= 0 && lis_i >= 0) {
                var t = tokens[i];

                if(t == '>') {
                    next_must_match = true;
                    i--;
                } else {
                    if(isSelector(node, t, null, $li)) {
                        //hl selected parts
                        next_must_match = false;
                        i--;
                        lis_i--;
                    } else {
                        if(next_must_match) {
                            //invalid, we'lost
                        }
                        lis_i--;
                    }
                }
            }
        }

    })

    input.on('keydown', function(e) {
        if(e.which == 13) {
            onButtonPressed(e);
        } else if(e.which == 27 && !options.compact) {
            _this.onCancel();
        } else {
            return;
        }
        return false;
    })

    input.on('focus', function () {
        if(updateRuleCreatedIn) {
            updateRuleCreatedIn(true);
        }
    });

    var hideOnBlurTimer = null;

    input.on('blur', function () {
        hideOnBlurTimer = setTimeout(function() {
            if(updateRuleCreatedIn) {
                updateRuleCreatedIn();
            }
            hideOnBlurTimer = null;
        }, 500)
    });

    this.getInput = function() {
        return input;
    }

    var default_cs = null;

    if(options.compact) {
        var $ruleCreatedIn = $('<div class="pg-selector-dest"></div>').appendTo($element);

        $ruleCreatedIn.on('click', 'a', function(e) {
            e.preventDefault();
            var d = new PgSelectDefaultStylesheet(pinegrow.getSelectedPage(), default_cs);
            d.show(function(cs, cancelled) {

            })
            if(hideOnBlurTimer) {
                clearTimeout(hideOnBlurTimer);
                hideOnBlurTimer = null;
            }
        })

        var updateRuleCreatedIn = function(force) {
            if((selector.length || force) && default_cs) {
                if(default_cs) {
                    $ruleCreatedIn.html('<p>Rule will be created in <a href="#">' + crsaGetNameFromUrl(default_cs.url) + '</a></p>')
                } else {
                    $ruleCreatedIn.html('<p><a href="#">Choose stylesheet</a> for the rule</p>')
                }
                pg$Show($ruleCreatedIn);
            } else {
                pg$Hide($ruleCreatedIn);
            }
        }

        var onDefaultStylesheetSelected = function() {
            if(options.default_cs) {
                default_cs = options.default_cs;
                updateRuleCreatedIn();
            } else {
                pinegrow.getDefaultStylesheet(pinegrow.getSelectedPage(), function (cs) {
                    default_cs = cs;
                    updateRuleCreatedIn();
                })
            }
        }

        onDefaultStylesheetSelected();

        pinegrow.addEventHandler('on_default_stylesheet_selected', onDefaultStylesheetSelected);

        pg$Hide($ruleCreatedIn);

    }

    var states = ['', 'hover', 'active', 'focus', 'visited', 'first-child', 'first-of-type', 'last-child', 'first-letter', 'last-of-type', 'only-child', 'only-of-type', 'empty', 'after', 'before'];

    if(pgf.kids) {
        var states = ['', 'hover', 'active', 'focus'];
    }

    var $list = $('<ul class="pg-selector-maker-crumbs"></ul>');
    (!pgf.kids) && $list.appendTo($element);

    if(pgf.kids) {
        $(`<div class="pg-selector-maker-help">
<p>Which HTML elements do you wish to style?</p>
<ul>
<li><code>${pgel.tagName}</code> - All <b>&lt;${pgel.tagName}&gt;</b> elements</li>
<li><code>.myclass</code> - All elements with the <b>class</b> myclass.</li>
<li><code>#button1</code> - Element with the <b>id attribute</b> button1.</li>
</ul>
<p>Add <code>:hover</code>, <code>:active</code> or <code>:focus</code> to style states.</p>
</div>`).appendTo($element);
    }
    $list.on('click', 'a', function(e) {
        e.preventDefault();

        var $a = $(this);
        var $li = $a.closest('li');

        if(input_changed) {
            return;
        }

        var update = function() {
            selector = getSelector();
            input.setValue(selector);
            updateMatcher();
            updateAddClasses();
            if(updateRuleCreatedIn) {
                updateRuleCreatedIn();
            }
            input.$field.removeClass('error');
        }

        if($a.hasClass('pg-sm-untruncate')) {
            $li.removeClass('truncated');

        } else if($a.hasClass('pg-sm-state-opener')) {
            //$li.find('.pg-sm-state').toggle();

            var $state = $li.find('.pg-sm-state');
            var sel_state = $state.text();

            var menu = new CrsaContextMenu(null, $a);
            states.forEach(function(state) {
                menu.add(':' + state, null, function() {
                    $state.text(state)
                    update();
                }, null, {}, false);
            })
            menu.showForElement($a, null, 8);

        } else {
            var sel = $a.attr('data-pg-sel');

            $a.toggleClass('selected');

            //if > check if the next item is selected
            if (sel == '>' && $a.hasClass('selected')) {
                var $next_li = $li.next();
                if ($next_li.find('.selected').length === 0) {
                    $next_li.find('.pg-sm-tag').addClass('selected');
                }
            }

            if ($li.find('.selected').length === 0) {
                //is prev > selected? disable it
                $li.prev().find('.pg-sm-lg').removeClass('selected');
            }
            update();
        }

        return false;
    })

    if(options.simple) {
        pg$Hide($list);
    }

    var $matcher = $('<div class="pg-selector-maker-matcher"></div>').appendTo($element);

    pg$Hide($matcher);

    var matcherTimeout = null;
    var wrong_selector = false;

    var updateMatcher = function(delayed) {
        if(matcherTimeout) {
            clearTimeout(matcherTimeout);
            matcherTimeout = null;
        }

        var doTask = function() {
            try {
                if(selector.length) {
                    pg$Show($matcher);
                } else {
                    pg$Hide($matcher);
                    return;
                }

                var sel_without_states = selector.replace(/\:hover|\:active|\:focus|\:visited/g, '');

                var page = pgel.getPage();
                var m = $(page.getDocument()).find(sel_without_states);
                var num = m.length;

                pinegrow.highlightElement(m);

                var html = '';

                var sel_match = '';
                var sel_msg = '';
                var other_match = '';
                var other_msg = '';

                var $pgel = pgel.get$DOMElement();

                if($pgel.is(m)) {
                    sel_match = 'match';
                    sel_msg = 'This rule will style the selected element.';
                    num--;
                } else {
                    var cls = $pgel.attr('class');
                    _this.getAddClasses().forEach(function(c) {
                        $pgel.addClass(c);
                    })

                    if($pgel.is(sel_without_states)) {
                        sel_match = 'match-with-classes';
                        sel_msg = 'This rule will style the selected element if classes are added to the element.'
                    } else {
                        sel_match = 'no-match';
                        sel_msg = 'This rule will NOT style the selected element.';
                    }

                    if(cls === null) {
                        $pgel.removeAttr('class');
                    } else {
                        $pgel.attr('class', cls);
                    }
                }
                if(num > 0) {
                    other_match = 'match';
                    //other_msg = num + ' other elements on the page match the selector.';
                    other_msg = `This rule will style ${num} other elements on the page.`;
                } else {
                    other_match = 'no-match';
                    //other_msg = 'No other elements on the page match the selector.';
                    other_msg = `This rule will not style any other elements on the page.`;
                }

                html = `<span class="pg-sel-match ${sel_match}"><i class="icon icon-${pgf.kids ? 'check' : '267'}"></i></span>`;

                html += '<span class="pg-other-match ' + other_match + '"><i class="icon icon-no-affected"></i>' + num + '</span>';

                $matcher.html(html);

                addTooltip($matcher.find('.pg-sel-match'), sel_msg);
                addTooltip($matcher.find('.pg-other-match'), other_msg);
                //input.setError(null);
                //wrong_selector = false;
            } catch(err) {
                //input.setError('The selector is not valid.');
                $matcher.html('Could not count matches on the page. <b>Perhaps the selector is not valid?</b>');
                //wrong_selector = true;
            }
            matcherTimeout = null;
        }

        if(delayed) {
            setTimeout(doTask, 500);
        } else {
            doTask();
        }
    }

    updateMatcher();

    if(options.add_classes) {

        var $addclasses = $('<div class="pg-selector-maker-addc"></div>').appendTo($element).hide();
        var addc_input = new PgInputField('addclasses', {
            type: 'checkbox',
            show_empty: false,
            name: 'Add classes',
            without_text: false,
            default_value: 1,
            value: 1
        });

        addc_input.appendTo($addclasses);

        var $addc_label = addc_input.$field.find('label');
    }

    if(options.save_inline) {

        var $saveinline = $('<div class="pg-selector-maker-save-inline"></div>').appendTo($element).hide();
        var save_inline_input = new PgInputField('saveinline', {
            type: 'checkbox',
            show_empty: false,
            name: 'Save <span class="pg-text-orange">inline style</span> to this rule',
            without_text: false,
            default_value: pinegrow.getSetting('create-css-rule-save-inline', '1'),
            value: 1
        });

        save_inline_input.appendTo($saveinline);
    }

    if(options.media_query) {

        var mq_input = new PgInputField('mq', {
            type: 'text',
            show_empty: false,
            name: 'Media query',
            without_text: false,
            slider_def_unit: ''
        });

        mq_input.appendTo($element);

        if (options.media_value) {
            mq_input.setValue(options.media_value)
        }

        var mqHelper = new PgMediaQueryHelperView(mq_input.$input, pgel.getPage());
        mqHelper.view.get$Element().appendTo($element);
    }

    if(options.stylesheets) {
        var project = pinegrow.getCurrentProject();

        var list = options.stylesheets.map(function(s) {
            var short = crsaGetNameFromUrl(s);
            return {name: short, key: s, html: short + ' <small>' + (project ? project.getRelativeUrl(s) : s) + '</small>'};
        })
        var in_input = new PgInputField('stylesheet', {
            type: 'select',
            show_empty: false,
            name: 'Stylesheet',
            without_text: false,
            default_value: options.stylesheets[0],
            options: list
        });

        if(!pgf.kids) {
            in_input.appendTo($element);
        }

        if(options.stylesheet_value) {
            in_input.setValue(options.stylesheet_value)
        }

        in_input.on('input', function(e) {

        })
    }

    var $button = $('<button class="pg-button create btn-primary">' + options.button_label + '</button>');

    if(!options.compact) {
        $button.appendTo($element);
    } else {
        $button.appendTo(input.$field);

        $button.removeClass('btn-primary');

        $element.addClass('pg-selector-maker-compact');
    }

    var onButtonPressed = function(e) {
        e.preventDefault();
        if(!selector) {
            if(_this.onCreateWithoutSelector) {
                if(_this.onCreateWithoutSelector(_this, e)) return;
            }
            input.$field.addClass('error');
            if(_this.onEmptySelector) {
                _this.onEmptySelector();
            } else {
                pinegrow.showQuickError('Please enter the CSS rule selector');
            }
        } else if(pgf.kids && wrong_selector) {
            pinegrow.showQuickError('The value is not a correct CSS selector. Please enter a valid selector.');
        } else {
            if(_this.onCreate) _this.onCreate(_this);
        }
    }

    $button.on('click', onButtonPressed);



    var getSelector = function() {
        var sel = '';
        var add_space = false;
        $list.children().each(function(i, e) {
            var $li = $(e);
            $li.find('.selected[data-pg-sel]').each(function(i, a) {
                if(add_space) {
                    sel += ' ';
                    add_space = false;
                }
                var t = a.getAttribute('data-pg-sel');
                if(t == '&gt;') t = '>';
                sel += t;
            })
            var state = $li.find('.pg-sm-state').text();
            if(state) sel += ':' + state;
            add_space = true;
        })
        return $.trim(sel);
    }

    var updateCrumbs = function() {
        $list.empty();
        var $el = pgel.get$DOMElement();

        var page = pgel.getPage();

        while($el && $el.length && $el.get(0).nodeType === 1) {

            $list.prepend(createCrumb($el.get(0)));
            $el = $el.parent();

            if(pgf.kids) break;

            if($el.length && $el.get(0).nodeType === 1) {
                if(page && page.wrapper_node && $el.get(0).getAttribute('data-pg-wrapper')) {
                    break;
                }
                $list.prepend($('<li><a class="pg-sm-lg" href="#" data-pg-sel="&gt;">&gt;</a></li>'));
            }
        }
    }

    var createCrumb = function(node) {
        var r = [];

        var $li = $('<li></li>');

        r.push($('<a class="pg-sm-tag" href="#" data-pg-sel="' + node.tagName.toLowerCase() + '">' + node.tagName.toLowerCase() + '</a>'));

        var id = node.getAttribute('id');
        if(id && !id.startsWith('sizzle')) {
            r.push($('<a class="pg-sm-id" href="#" data-pg-sel="#' + id + '">#' + id + '</a>'));
        }

        var classList = pgMakeStringArrayUnique(pgGetElementClasses(node));
        for(var i = 0; i < classList.length; i++) {
            var cls = classList[i];
            var sel = '';
            if(options.select_class && cls == options.select_class) {
                options.select_class = null; //just the first one
                sel = ' selected'
            }
            r.push($('<a class="pg-sm-class' + sel + '" href="#" data-pg-sel=".' + cls + '">.' + cls + '</a>'));
        }

        if(classList.length > 5) {
            $li.addClass('truncated');
            r.push($('<a class="pg-sm-untruncate" href="#" title="Show all classes.">+</a>'));
        }

        r.push($('<a class="pg-sm-state-opener" href="#">:<span class="pg-sm-state"></span></a>'));

        /*
        states.forEach(function(state) {
            r.push($('<a class="pg-sm-state" href="#" data-pg-sel=":' + state + '">:' + state + '</a>').hide());
        })*/

        r.forEach(function($el) {
            $li.append($el);
        })
        return $li;
    }

    updateCrumbs();




    var updateAddClasses = function() {
        add_classes_list = [];
        if(!$addclasses) return;
        var a = $.trim(selector).split(/\s/);
        var list = [];
        if(a.length) {
            var last = a[a.length - 1];
            var m = last.match(/\.[a-z0-9\-\_]+/gi);
            if(m) {

                var $el = pgel.get$DOMElement();
                for(var i = 0; i < m.length; i++) {
                    var cls = m[i].replace('.', '');
                    if(!pgel.hasClass(cls) && !$el.hasClass(cls) && list.indexOf(cls) < 0) {
                        list.push(cls);
                    }
                }
            }
        }
        if(list.length) {
            $addc_label.html('Add ' + (list.length > 1 ? 'classes' : 'class') + ' <b>' + list.join('</b>, <b>') + '</b> to the selected element.')
            $addclasses.show();

            add_classes_list = list;
        } else {
            $addclasses.hide();
        }
    }



    updateAddClasses();


    var isSelector = function(node, sel, parsed_sel, $li) {
        if(!parsed_sel) {
            if(sel.indexOf(',') >= 0) {
                //handle multiple , separated selectors
                var list = crsaSplitAndTrimList(sel);
                for(var i = 0; i < list.length; i++) {
                    if(isSelector(node, list[i], null, $li)) return true;
                }
                return false;
            }
        }
        parsed_sel = parsed_sel || pgel.parseSelector(sel);

        var select = [];

        if(parsed_sel.id) {
            if(node.getAttribute('id') != parsed_sel.id[0].replace('#','')) return false;
            select.push('[data-pg-sel="' + parsed_sel.id[0] + '"]');
        }
        if(parsed_sel.tag) {
            if (parsed_sel.tag.length > 0 && parsed_sel.tag[0] != '*') {
                if(parsed_sel.tag.toLowerCase() != node.tagName.toLowerCase()) return false;
                select.push('[data-pg-sel="' + parsed_sel.tag[0] + '"]');
            }
        }
        if(parsed_sel.classes) {
            for(var i = 0; i < parsed_sel.classes.length; i++) {
                if(!node.classList.contains(parsed_sel.classes[i].replace('.', ''))) return false;
                select.push('[data-pg-sel="' + parsed_sel.classes[i] + '"]');
            }
        }

        if($li && select.length) {
            $li.find(select.join(',')).addClass('selected');
        }
        /*
         if(parsed_sel.attr) {
         if(this.hasAttr(parsed_sel.attr)) {
         if(parsed_sel.attrValue !== null && this.getAttr(parsed_sel.attr) != parsed_sel.attrValue) return false;
         } else {
         return false;
         }
         }*/
        return true;
    }


    if(options.selector_value) {
        input.setValue(options.selector_value)
        input.$input.trigger('input');
    }

    var onTargetSelected = function(page, apgel, $el, status) {
        if(in_target_mode) {
            pgel = apgel;
            pinegrow.highlightElement($el);
            updateCrumbs();
            pinegrow.showQuickMessage('New target element was selected.');
            status.stop = true;
        }
    }

    var in_target_mode = false;

    this.selectTarget = function() {
        in_target_mode = true;
        pinegrow.showQuickMessage('Click on the page to select a new target element.');
    }

    this.onShown = function() {
        input.$input.get(0).focus();
    }

    pinegrow.addEventHandler('on_element_clicked', onTargetSelected);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_element_clicked', onTargetSelected);

        if(options.compact) {
            pinegrow.removeEventHandler('on_default_stylesheet_selected', onDefaultStylesheetSelected);
        }

        if(matcherTimeout) {
            clearTimeout(matcherTimeout);
            matcherTimeout = null;
        }
    }
}

var PgQuickSelectorMaker = function(pgel, selector, options) {

    options = options || {};
    var _this = this;

    if(!options.stylesheet_value) {
        options.stylesheet_value = pgcssh.getUrlOfStylesheetForOverrides();
    }

    var sm = this.maker = new PgSelectorMakerView(pgel, selector, options);

    var win = new PgQuickWindow(options.title ? options.title : 'Create new CSS rule', sm, 'quickSelectorMaker', pgf.kids ? 500 : 570, pgf.kids ? 200 : 300);

    if(options.$position) {
        //var pos = options.$position.get(0).getBoundingClientRect();
        //win.showAtPosition(pos.left, pos.top, Math.max(pos.width, 600));
        win.showFor$El(options.$position, false, 0, 'down');
    } else {
        win.showForPgel(pgel);
    }

    sm.onShown();

    win.autoSize(120);

    if(!pgf.kids) {
        win.addIconToHeader('icon-267', 'Select another target element', function () {
            sm.selectTarget();
        })
    }

    win.onDestroy = function() {
    }

    this.onCreate = function(maker) {};

    sm.onCreate = function(maker) {
        if(_this.onCreate) _this.onCreate(maker);
        var s = maker.getStylesheet();
        if(s) {
            pgcssh.setUrlOfStylesheetForOverrides(s);
        }
        win.destroy();
    }

    sm.onCancel = function(maker) {
        win.destroy();
    }
}
