var PgVarsView = function () {
    var $element = $('<div class="crsa-stylesheet-vars"></div>');

    var view = new PgUIView($element);
    var pgStylesheetList = [];
    var stylesheetMap = {};
    var customStylesheet = null;

    var showActiveOnly = false;

    this.view = view;

    var filter = '';

    var $noteMessage = $('<div class="alert alert-info alert-panel-notice"></div>').appendTo($element);
    var $customStylesheetContainer = $('<div class="custom-stylesheet-container"></div>').appendTo($element);
    var $contentContainer = $('<div class="var-content-container"></div>').appendTo($element);

    var showHideNote = function () {
        var $sections = $element.find('.pg-ui-section');
        if ($sections.length > 0) {
            pg$Hide($noteMessage);
        }
        else {
            if (showActiveOnly) {
                var selectedElement = pinegrow.getSelectedElement();
                if (!selectedElement) {
                    $noteMessage.html('Element is not selected.');
                }
                else {
                    $noteMessage.html('There is no variables.');
                }
            }
            else if (filter.length > 0) {
                $noteMessage.html('There is no variables.');
            }

            pg$Show($noteMessage);
        }
    }

    var updateVarList = function () {
        var oldTop = $element.scrollTop();
        pgRemoveSpectrum($contentContainer);
        pgRemoveMultiselects($contentContainer);

        $contentContainer[0].innerHTML = '';

        var showVarsFor = function (cs) {
            var hasCustomStylesheet = customStylesheet && customStylesheet.getUniqueName() == cs.getUniqueName();
            if (hasCustomStylesheet) return;

            if (cs.varsPane) {
                cs.varsPane.remove();
                cs.varsPane = null
            }
            if (cs.varGroups.length > 0) {
                var $currVarContainer = $('<div/>').appendTo($contentContainer);
                cs.varsPane = new PgStylesheetVars(cs, $currVarContainer, {
                    customStylesheet: customStylesheet
                });
                cs.varsPane.filter = filter;
                cs.varsPane.showActiveOnly = showActiveOnly;
                cs.varsPane.show();
                // cs.varsPane.show(cs.varGroups[0]);
            }
        }

        for (var uniqueName in stylesheetMap) {
            showVarsFor(stylesheetMap[uniqueName]);
        }

        $element.scrollTop(oldTop);
        showHideNote();
    }

    var setAllStylesheets = function () {
        var addStylesheet = function (cs) {
            // pgStylesheetList.push(cs);
            if (cs.name == 'custom.scss') customStylesheet = cs;
            stylesheetMap[cs.getUniqueName()] = cs;
        }

        var searchForAllImportedFiles = function (cs) {
            var forAll = function (importedFiles) {
                for (var i = 0; i < importedFiles.length; i++) {
                    var importedFile = importedFiles[i];
                    addStylesheet(importedFile);

                    if (importedFile.importedFiles.length > 0) forAll(importedFile.importedFiles);
                }
            }

            forAll(cs.importedFiles);
        }

        if (pinegrow.stylesheets) {
            $.each(pinegrow.stylesheets.getAllCrsaStylesheets(), function(i, cs) {
                if (cs.inline || cs.ignore) return true;

                if (cs.compiledFile) cs = cs.sourceStylesheet;

                cs.compiledStylesheets.forEach(function (compiledStylesheet) {
                    addStylesheet(compiledStylesheet);
                    searchForAllImportedFiles(compiledStylesheet);
                });
            });
        }
    }

    var removeAllStylesheetList = function () {
        for (var uniqueName in stylesheetMap) {
            var cs = stylesheetMap[uniqueName];
            if (cs.varsPane) {
                cs.varsPane.remove();
                cs.varsPane = null
            }
        }
        stylesheetMap = {};
        customStylesheet = null;
    }

    var addCustomStylesheetSelection = function () {
        $customStylesheetContainer[0].innerHTML = '';

        var customStylesheet = [];
        for (var uniqueName in stylesheetMap) {
            var cs = stylesheetMap[uniqueName];
            customStylesheet.push({
                name: cs.name,
                key: uniqueName
            })
        }

        var stylesheetSelection = new PgInputField('stylesheet-selection', {
            type: 'select',
            name: 'Custom stylesheet',
            show_empty: true,
            icon: false,
            without_text: false,
            placeholder: 'Custom stylesheet',
            options: customStylesheet
        });

        stylesheetSelection.appendTo($customStylesheetContainer);
    }


    var updateStylesheetList = function () {
        removeAllStylesheetList();
        setAllStylesheets();
    }

    var showVars = function () {
        updateStylesheetList();
        updateVarList();
    }

    var onStylesheetsLoaded = function () {
        updateStylesheetList();

        view.whenVisible(function() {
            updateVarList();
        });
    }

    var onPageClosed = function () {
        showVars();
    }

    var onActiveRuleChanged = function (page, obj) {
        if (showActiveOnly) {
            if (obj && obj.stylesheet_changed) {
                updateStylesheetList();
            }

            updateVarList();
        }
    }

    var onRulesChanged = function (page) {
        view.whenVisible(function() {
            updateVarList();
        });
    }


    pinegrow.addEventHandler('on_page_stylesheets_loaded', onStylesheetsLoaded);
    pinegrow.addEventHandler('on_page_closed', onPageClosed);

    pinegrow.addEventHandler('on_css_active_rule_changed', onActiveRuleChanged);
    pinegrow.addEventHandler('on_css_active_var_changed', onActiveRuleChanged);

    pinegrow.addEventHandler('on_css_source_changed', onRulesChanged);

    view.whenVisible(function() {
        showVars();
    });

    view.on('crsa-var-search', function (event) {
        view.whenVisible(function() {
            filter = 'bg12';
            updateVarList();
        })
    });

    view.onResize = function() {
        for (var uniqueName in stylesheetMap) {
            var cs = stylesheetMap[uniqueName];
            if (cs.varsPane) {
                cs.varsPane.updateLayout();
            }
        }
    }

    view.onDestroy = function() {
        removeAllStylesheetList();

        pinegrow.removeEventHandler('on_page_stylesheets_loaded', onStylesheetsLoaded);
        pinegrow.removeEventHandler('on_page_closed', onPageClosed);

        pinegrow.removeEventHandler('on_css_active_rule_changed', onActiveRuleChanged);
        pinegrow.removeEventHandler('on_css_active_var_changed', onActiveRuleChanged);

        pinegrow.removeEventHandler('on_css_source_changed', onRulesChanged);
    }
}


var PgVarDialogView = function($html) {
    var _this = this;
    var stylesheetVar = null;

    var $element = $('<div class="pg-var-dialog"></div>');

    var view = this.view = new PgUIView($element);

    $element.append($html);

    this.setStylesheetVar = function (currStylesheetVar) {
        stylesheetVar = currStylesheetVar;
    }

    view.onDestroy = function() {
        if (stylesheetVar) stylesheetVar.remove();

        stylesheetVar = null;
        _this = null;
    }
}
