var PgBrowserSync = function() {
    var change_count = 0;

    var bs_url = pinegrow.getSetting('browsersync-url', 'http://localhost:3000');

    var reload = [];

    var tasks = new PgDelayedTasks();

    var enabled = pinegrow.getSetting('browsersync-enabled', '0') === '1';

    var last_error_msg = 0;

    var reloadDelayed = function() {
        tasks.run('reload', async function() {
            if(reload.length) {
                var url = bs_url + '/__browser_sync__?method=reload';
                reload.forEach(function(u) {
                    url += `&args=${u}`;
                })
                //console.log('BrowserSync: ' + url)
                reload = [];

                try {
                    await pgFetchRemoteContent(url);
                } catch(err) {
                    console.log(`BrowserSync is not running on ${bs_url}`);

                    var t = new Date().getTime();
                    if(t - last_error_msg > 60000) {
                        pinegrow.showQuickError(`BrowserSync is not running on ${bs_url}. Start it or switch off the sync in the top toolbar with <i class="icon icon-27"></i>.`);
                        last_error_msg = t;
                    }
                }
            }
        }, 250);
    }

    var changedUrl = function(url) {
        if(!enabled) return;
        var p = pinegrow.getCurrentProject();
        if (p) {
            //url = pinegrow.getProxyUrl(url);

            url = p.getRelativeUrl(url);

            if(reload.indexOf(url) < 0) {
                reload.push(url);
                reloadDelayed();
            }
        }

    }

    pinegrow.addEventHandler('on_page_changes_done', function (page) {
        changedUrl(page.url);
    })

    pinegrow.addEventHandler('on_css_stylesheet_changed', function (pagenull, cs) {
        cs.url && changedUrl(cs.url);
    })

    var $icon = $('.commands-toolbar');

    $icon.get(0).addEventListener('click', function(e) {
        if(e.target.matches('.icon-27') || e.target.querySelector('.icon-27')) {
            e.preventDefault();
            e.stopPropagation();

            var page = pinegrow.getSelectedPage();

            if(!page) {
                pinegrow.showQuickMessage('Nothing to preview in browser. Open a page first.');
                return;
            }

            var actions = [
                {
                    label: 'Open the page in browser',
                    //kbd: 'CMD O',
                    action: function() {
                        if(!enabled) {
                            page.openInBrowser();
                        } else {
                            var p = pinegrow.getCurrentProject();
                            if(!p) {
                                pinegrow.showQuickError('This feature requires projects. Open the page as Project first.');
                            } else {
                                if(!p.isUrlInProject(page.url)) {
                                    pinegrow.showQuickError('The page is not part of the current project. Opening in the browser instead...');
                                    page.openInBrowser();
                                } else {
                                    var url = p.getRelativeUrl(page.url);

                                    var gui = require('nw.gui');
                                    url = bs_url + '/' + url;

                                    gui.Shell.openExternal(url);
                                }
                            }
                        }
                    }
                },
                {
                    type: 'divider'
                },
                {
                    type: 'header',
                    label: 'BrowserSync'
                },
                {
                    label: 'Use BrowserSync',
                    check: enabled,
                    action: function() {
                        enabled = !enabled;
                        pinegrow.setSetting('browsersync-enabled', enabled ? '1' : '0');
                        pinegrow.showMode('BrowserSync', enabled)
                    }
                },
                {
                    label: 'How to start BrowserSync',
                    action: function() {
                        pinegrow.showAlert(`<p>BrowserSync must be running for this feature to work.</p><p>To start BrowserSync:</p><ul>
<li>Open the terminal / command prompt</li>
<li>Run <code>browser-sync start --proxy "${pinegrow.getApiUrl()}"</code></li>
</ul>
<p><a href="https://pinegrow.com/docs/master-pinegrow/using-external-code-editors/live-preview-on-any-device-or-browser-with-browsersync/" class="external">Learn more about using BrowserSync in Pinegrow.</a></p>`, 'How to start BrowserSync')
                    }
                },
                {
                    label: 'Settings...',
                    action: function() {
                        var fields = {
                            url: {
                                name: 'BrowserSync url',
                                type: 'text',
                                value: bs_url
                            }
                        };

                        pinegrow.showQuickDialog(`BrowserSync Settings`, 'Settings', `<p>BrowserSync is an external and free tool that let's you preview pages on multiple browsers / devices while you edit them in Pinegrow. All changes will be shown in the browser, without having to save them first.</p><p><a href="https://pinegrow.com/docs/master-pinegrow/using-external-code-editors/live-preview-on-any-device-or-browser-with-browsersync/" class="external">Learn more about using BrowserSync in Pinegrow.</a></p>`, fields, 'Cancel', 'Save', null, function(values) {
                            bs_url = values.url || 'http://localhost:3000';
                            pinegrow.setSetting('browsersync-url', bs_url);
                        })
                    }
                },
            ]

            var menu = new CrsaContextMenu(actions, $(e.target));
            menu.showForElement($(e.target));
        }
    }, true);

}

