$(function () {

    if(!pgf.components) return;

    $('body').one('pinegrow-ready', function (e, pinegrow) {//Wait for Pinegrow to start-up

        // Create new Pinegrow framework object
        var f = new PgFramework('pg.components', 'Components');
        pinegrow.addFramework(f);

        f.setScriptFileByScriptTagId('plugin-pg-components', 'lib/crsa/plugins/pg.components.js');

        var path = require('path');
        var fs = require('fs');

        var canUse = function() {
            return pinegrow.isPRO() || pinegrow.isProTrialActive()
        }

        f.detect = function(crsaPage) {
            return false;
        }

        f.default = true;
        f.has_actions = true;
        f.show_in_manager = false;
        //f.not_main_types = true;
        f.order = 1000; //show at the end of ACT panel

        var actions = [];
        var use_actions = [];

        var select_element_after_update = null;

        //Add insight handlers
        pinegrow.insight.addHandler(
            function(insight) { //begin
                insightext.master_pages = [];
                insightext.master_pages_changed = true;
                insightext.component_types_map = {};
                insightext.component_types_list = [];
            },
            function(insight, page) { //page
                //inspect nodes
                return function(insight, page, node) {
                    if(node.tagName == 'html') {
                        if(node.hasAttr('data-pgc-set-master')) {
                            insightAddMasterPage(page.name);
                        }
                    }
                }
            },
            function(insight) { //end
                if(insightext.master_pages_changed) {
                    insightext.master_pages.sort();
                    insightext.master_pages_changed = false;
                }
            }
        )
        var insightext = {
            getMasterPages : function() {
                var r = [];
                insightext.master_pages.forEach(function(name) {
                    r.push({key: name, name: name});
                })
                return r;
            },
            getComponents : function() {
                var r = [];
                var page = pinegrow.getSelectedPage();
                if(page) {
                    page.getFrameworks().forEach(function(f) {
                        if(f.components_loaded) {
                            $.each(f.getComponentTypes(), function(key, def) {
                                if(def.is_smart_component) {
                                    r.push({key: def.type, name: def.type, html: def.type + (def.type != def.name ? (' (' + def.name + ')') : '') + ' <small>' + f.name + '</small>'});
                                }
                            })
                        }
                    })
                }
                pinegrow.insight.sortOptionsList(r);
                return r;
            },
            getEditableFieldsForElement : function(pgel) {
                var r = [];
                var addFields = function(c, text) {
                    if(c.component_fields) {
                        $.each(c.component_fields, function(key, field) {
                            r.push({key: field.name, name: field.name, html: field.name + ' <small>' + (text || '') + c.name + '</small>'});
                        });
                    }
                }

                if (pgel) {
                    var page = pinegrow.getCrsaPageOfPgParserNode(pgel);
                    var c = getComponentInstance(pgel, page);
                    if (c) {
                        addFields(c);
                    } else {
                        //has master page?
                        var project = pinegrow.getCurrentProject();
                        if (project) {
                            var master_def = getMasterPage(page, project);
                            if (master_def) {
                                addFields(master_def, 'master page ');
                            } else {
                                r.push({key: '', name: '', html: '<small>Assign component or master page to see editable fields.</small>'})
                            }
                        }
                    }
                    pinegrow.insight.sortOptionsList(r);
                }
                return r;
            }
        }
        pinegrow.insight.ext.components = insightext;

        var insightAddMasterPage = function(name) {
            if(insightext.master_pages.indexOf(name) < 0) {
                insightext.master_pages.push(name);
                insightext.master_pages_changed = true;
            }
        }
        //end insight

        var pgc = new PgComponentType('pgc.define', 'Define component');
        pgc.short_name = 'Define comp.';
        pgc.selector = "[data-pgc-define]";
        pgc.attribute = 'data-pgc-define';
        pgc.action = true;
        pgc.not_main_type = true;
        pgc.sections = {
            'pgc.define.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-define' : {
                        type : 'text',
                        name : 'Id',
                        action : "element_attribute",
                        attribute : 'data-pgc-define',
                        attribute_keep_if_empty : true,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('my.poster'),
                        options: function() {
                            return pinegrow.insight.ext.components.getComponents();
                        }
                    },
                    'data-pgc-define-name' : {
                        type : 'text',
                        name : 'Display name',
                        action : "element_attribute",
                        attribute : 'data-pgc-define-name',
                        placeholder: f.textForExample('Poster')
                    },
                    'data-pgc-define-auto-update' : {
                        type : 'checkbox',
                        name : 'Update instances',
                        action : "element_attribute",
                        attribute: 'data-pgc-define-auto-update',
                        value: 'true',
                        negvalue: 'false',
                        default_value: 'true',
                        on_changed : function(pgel, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
                            if(value == 'false') {
                                pinegrow.showQuickMessage('Component instances <b>will not</b> be updated when you run Components -&gt; Update.', 3000);
                            } else {
                                pinegrow.showQuickMessage('Be careful, running Components -&gt; Update <b>will change</b> component instances.', 4000, false, true /* highlight */);
                            }
                        }
                    },
                    'data-pgc-define-map-urls' : {
                        type : 'checkbox',
                        name : 'Map urls',
                        action : "element_attribute",
                        attribute: 'data-pgc-define-map-urls',
                        value: 'true',
                        negvalue: 'false',
                        default_value: 'true',
                        helptext: `<p>Map urls if the component is used in another project location. For example link to index.html will become ../index.html if the component is used in a subfolder.</p>`
                    },
                    'data-pgc-define-description' : {
                        type : 'text',
                        name : 'Description',
                        action : "element_attribute",
                        attribute : 'data-pgc-define-description'
                    },
                    'data-pgc-define-photo-preview-only' : {
                        type : 'checkbox',
                        name : 'Use photo only for preview',
                        action : "element_attribute",
                        attribute: 'data-pgc-define-photo-preview-only',
                        empty_attribute: true,
                        value: 'true'
                    },
                    /*, DISABLED for now,
                    'data-pgc-define-selector' : {
                        type : 'text',
                        name : 'Selector',
                        action : "element_attribute",
                        attribute : 'data-pgc-define-selector',
                        placeholder: f.textForExample('.poster')
                    }, *//*
                    'data-pgc-define-advanced-label' : {
                        type : 'label',
                        name : 'Advanced',
                        action : "element_attribute",
                        attribute: 'dummy'
                    },
                    'data-pgc-define-parent' : {
                        type : 'text',
                        name : 'Parent selector',
                        action : "element_attribute",
                        attribute : 'data-pgc-define-parent',
                        placeholder: f.textForExample('.row')
                    }*/
                }
            }
        };

        if(pgIsDev() && pgf.smart_blocks) {
            pgc.sections['pgc.define.parameters'].fields['data-pgc-define-extract-css'] = {
                type : 'text',
                name : 'Extract CSS for selector',
                action : "element_attribute",
                attribute : 'data-pgc-define-extract-css'
            }
            pgc.sections['pgc.define.parameters'].fields['data-pgc-define-extract-css-button'] = {
                type : 'button',
                name : 'Extract CSS',
                func: function(pgel) {
                    var api = new PgApi();
                    api.extractComponent(pgel, function(d) {
                        console.log(d);
                    })
                }
            }
        }

        f.addComponentType(pgc);
        actions.push(pgc);

        pgc.meta = {
            helptext: 'Define Pinegrow component.<span class=\'more\'> Disable <b>Update instances</b> if you do not want component instances to be updated with Components -&gt; Update. In that case you can update them manually with Actions -&gt; Update component.</span>',
            helplink: 'https://pinegrow.com/docs/pinegrow-pro/components/'
        }

        var max_fields = 1;

        var getEditableActionTag = function(pgel, def) {
            var i = pgcVarParseValue(pgel.getAttr('data-pgc-field') || pgel.getAttr('data-pgc-edit'));
            return '<i class="icon icon-Editable"></i> ' + (i.name || '');
        }

        window.PgEditableAreaExtension = function(key, name, f) {
            this.key = key;
            this.section = {
                name : name, no_section_label: true, fields : {}
            };

            this.framework = f;

            var fields = {};

            var set_values = {};
            var get_values = {};
            var get_properties = {};
            var can_make_changes = {};
            var show_fields = {};

            this.getValue = function(field_key, node) {
                return get_values[field_key] ?  get_values[field_key](node) : null;
            }

            this.setValue = function(field_key, value, node) {
                return set_values[field_key] ? set_values[field_key](value, node) : null;
            }

            this.getProperty = function(field_key, value, pgel) {
                return get_properties[field_key] ? get_properties[field_key](value, pgel) : null;
            }

            this.canMakeChange = function(field_key, pgel, action, data) {
                return can_make_changes[field_key] ? can_make_changes[field_key](pgel, action, data) : null;
            }

            this.getSection = function(page, selectedElement) {
                var section = {
                    name : name, no_section_label: true, fields : {}
                };
                var has = false;
                $.each(fields, function(skey, fdef) {
                    if (!selectedElement || !show_fields[skey] || show_fields[skey](page, selectedElement)) {
                        section.fields[prefix + skey] = fdef;
                        if(!has) {
                            fdef.start_section = name;
                        }
                        has = true;
                    }
                })
                return has ? section : null;
            }

            this.framework = null;

            var prefix = 'data-pgc-edit-ext-';

            this.addField = function(field_key, opts) {
                var fdef = opts.fieldDefinition;
                if(!fdef.action) fdef.action = 'element_attribute';
                var fkey = prefix + key + (field_key ? ('-' + field_key) : '');
                if(!fdef.attribute) fdef.attribute = fkey;

                //this.section.fields[fkey] = fdef;

                var skey = fkey.replace(prefix, '');

                fields[skey] = fdef;

                if(opts.getValue) get_values[skey] = opts.getValue;
                if(opts.setValue) set_values[skey] = opts.setValue;
                if(opts.getProperty) get_properties[skey] = opts.getProperty;
                if(opts.canMakeChange) can_make_changes[skey] = opts.canMakeChange;
                if(opts.shouldShowField) show_fields[skey] = opts.shouldShowField;
            }

            this.addFieldFromProp = function(field_key, name, fdef, selector) {
                if(fdef.action != 'apply_class') {
                    console.log('PgEditableAreaExtension.addFromProp - Only apply_class is supported!', fdef);
                    return null;
                }

                var s = {};

                s.fieldDefinition = {
                    type: 'checkbox',
                    name: name,
                    value: '1',
                    empty_attribute: true
                }

                if (selector) {
                    s.shouldShowField = function(page, pgel) {
                        return pgel.isSelector(selector);
                    }
                }

                s.getValue = function(node) {
                    for(var i = 0; i < fdef.options.length; i++) {
                        if(node.hasClass(fdef.options[i].key)) {
                            return fdef.options[i].key;
                        }
                    }
                    return null;
                }

                s.setValue = function(value, node, $el) {
                    for(var i = 0; i < fdef.options.length; i++) {
                        node.removeClass(fdef.options[i].key);
                        if($el) $el.removeClass(fdef.options[i].key)
                    }
                    node.addClass(value);
                    if($el) $el.addClass(value);
                    return value;
                }

                s.canMakeChange = function (pgel, action, data) {
                    if(action == 'add_class' || action == 'remove_class') {
                        for(var i = 0; i < fdef.options.length; i++) {
                            if(data == fdef.options[i].key) {
                                return true;
                            }
                        }
                    }
                    return false;
                }

                s.getProperty = function (value, pgel) {
                    if (pgel) {
                        var fielddef = $.extend({}, fdef);
                        fielddef.action = 'custom';
                        fielddef.get_value = function(currPgel) {
                            return s.getValue(pgel);
                        }
                        fielddef.set_value = function(currPgel, value) {
                            return s.setValue(value, pgel, pgel.get$DOMElement());
                        }
                        return fielddef;
                    }
                }

                this.addField(field_key, s);

                return s;
            }
        }


        if(pgIsDev() && pgf.smart_blocks) {
            var pgcscript = new PgComponentType('pgc.define.script', 'Define component script');
            pgcscript.short_name = 'Define comp. script';
            pgcscript.selector = "[data-pgc-define-script]";
            pgcscript.attribute = 'data-pgc-define-script';
            pgcscript.action = true;
            pgcscript.not_main_type = true;
            pgcscript.sections = {
                'pgc.script.parameters' : {
                    name : 'Options',
                    fields : {
                        'data-pgc-define-script' : {
                            type : 'text',
                            name : 'Component id',
                            action : "element_attribute",
                            attribute : 'data-pgc-define-script',
                            attribute_keep_if_empty : true,
                            validate : f.fieldIsRequired,
                            placeholder: f.textForExample('my.poster'),
                            options: function() {
                                return pinegrow.insight.ext.components.getComponents();
                            }
                        }
                    }
                }
            };

            f.addComponentType(pgcscript);
            actions.push(pgcscript);
        }


        var pgcvar = new PgComponentType('pgc.edit', 'Define editable area');
        pgcvar.short_name = 'Define Editable';
        pgcvar.selector = "[data-pgc-edit]";
        pgcvar.attribute = 'data-pgc-edit';
        pgcvar.action = true;
        pgcvar.not_main_type = true;
        pgcvar.get_action_tag = getEditableActionTag;

        pgcvar.pgc_extensions = {};

        pgcvar.pgcAddExtension = function(e) {
            this.pgc_extensions[e.key] = e;
        }

        pgcvar.pgcGetExtensionForFieldKey = function(attr) {
            var key = attr.split('-');
            if(key.length) {
                key = key[0];
                return pgcvar.pgc_extensions[key] || null;
            } else {
                return null;
            }
        }

        pgcvar.on_before_show_fields = function(page, sections, selectedElement, def) {
            $.each(this.pgc_extensions, function(key, e) {
                if(!e.framework || page.hasFramework(e.framework)) {
                    if(!e.shouldShowFields || e.shouldShowFields(page, selectedElement)) {
                        var s = e.getSection(page, selectedElement);
                        if(s) sections.push(s);
                    }
                }
            })
        }

        var pgcVarParseValue = function(val) {
            var r = {name: null, attrs: [], content: true, repeat: false};
            if(!val) return r;
            var m = val.match(/^([^\[]*)(\[.*\]|)$/);
            if(m) {
                r.name = $.trim(m[1]);
                if(m[2].length > 0) {
                    var items = m[2].replace('[','').replace(']','').split(',');
                    for(var i = 0; i < items.length; i++) {
                        var item = $.trim(items[i]).toLowerCase();
                        if(item == 'no_content') {
                            r.content = false;
                        } else {
                            r.attrs.push(item);
                        }
                    }
                }
            }
            return r;
        }

        var pgcVarConstructValue = function(r) {
            var val = r.name;
            var params = r.attrs;
            if(!r.content) params.push('no_content');
            if(params.length) {
                val += '[' + params.join(', ') + ']';
            }
            return val;
        }

        var pgcVarSetAttr = function(pgel, name, content, attr_str, values) {
            if(name === -1) name = values['data-pgc-edit'];
            if(content === -1) content = values['data-pgc-edit-content'];
            if(attr_str === -1) attr_str = values['data-pgc-edit-attrs'];

            var r = {name: name, content: content == '1', attrs: null};
            r.attrs = attr_str ? attr_str.split(',') : [];
            for(var i = 0; i < r.attrs.length; i++) {
                r.attrs[i] = $.trim(r.attrs[i]);
            }
            //if(r.attrs == 0) r.content = true;
            var val = pgcVarConstructValue(r);
            pgel.attr(pgcvar.attribute, val);
        }

        f.getFieldDefinitionStringForEditableAttribute = function(current_val, attr, name) {
            var val = null;
            if(current_val) {
                var r = pgcVarParseValue(current_val);
                if(r.attrs.indexOf(attr) < 0) {
                    r.attrs.push(attr);
                }
                val = pgcVarConstructValue(r);
            } else {
                val = name + '[' + attr + ',no_content]';
            }
            return val;
        }

        pgcvar.sections = {
            'pgc.edit.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-edit' : {
                        type : 'text',
                        name : 'Field Id',
                        action : "custom",
                        options: function() {
                            var list = pinegrow.insight.getValuesForAttribute('data-pgc-edit');
                            list.forEach(function(item) {
                                var i = pgcVarParseValue(item.key);
                                if(i) {
                                    item.name = i.name;
                                    item.key = i.name;
                                }
                            })
                            return list;
                        },
                        get_value: function(pgel, fn, values, fv) {
                            var val = pgel.attr(pgcvar.attribute);
                            if (val) {
                                var r = pgcVarParseValue(val);
                                return r.name;
                            }
                            return null;
                        },
                        set_value: function(pgel, value, values, oldValue, eventType, fieldDef) {
                            if (!value) {
                                pgel.removeAttr(pgcvar.attribute);
                            } else {
                                pgcVarSetAttr(pgel, value, -1, -1, values);
                            }
                            return value;
                         },
                         validate : f.fieldIsRequired,
                         placeholder: f.textForExample('block_title')
                    },
                    'data-pgc-edit-content' : {
                        type : 'checkbox',
                        name : 'Inner content',
                        action : "custom",
                        value: '1',
                        get_value: function (pgel, fn, values, fv) {
                            var val = pgel.getAttr(pgcvar.attribute);
                            if (val) {
                                var r = pgcVarParseValue(val);
                                return r.content ? '1' : null;
                            }
                            return '1';
                        },
                        set_value: function (pgel, value, values, oldValue, eventType, fieldDef) {
                            pgcVarSetAttr(pgel, -1, value ? '1' : null, -1, values);
                            return value;
                        }
                    },
                    'data-pgc-edit-attrs' : {
                        type : 'select',
                        multiple: true,
                        can_add_items: true,
                        options: function(fdef, pgel) {
                            var r = [];
                            $.each( pgel.getAttrList(), function( index, attr ) {
                                if(!attr.name.startsWith('data-pg')) {
                                    r.push({key: attr.name, name: attr.name});
                                }
                            });
                            pinegrow.insight.sortOptionsList(r);
                            return r;
                        },
                        name : 'Attributes',
                        action : "custom",
                        get_value: function(pgel, fn, values, fv) {
                            var val = pgel.getAttr(pgcvar.attribute);
                            if (val) {
                                var r = pgcVarParseValue(val);
                                return r.attrs.length ? r.attrs.join(', ') : null;
                            }
                            return null;
                        },
                        set_value: function(pgel, value, values, oldValue, eventType, fieldDef) {
                            pgcVarSetAttr(pgel, -1, -1, value, values);
                            return value;
                        }
                    },
                    'data-pgc-edit-classes' : {
                        type : 'select',
                        multiple: true,
                        can_add_items: true,
                        name : 'Classes',
                        action : "element_attribute",
                        attribute : 'data-pgc-edit-classes',
                        placeholder: f.textForExample('active, inactive'),
                        options: function(fdef, pgel) {
                            var r = [];
                            $.each( (pgel.getAttr('class') || '').split(/\s+/), function( index, cls ) {
                                if (!cls.startsWith('crsa-')) {
                                    r.push({key: cls, name: cls});
                                }
                            } );
                            pinegrow.insight.sortOptionsList(r);
                            return r;
                        },
                    },
                    'data-pgc-edit-bckimage' : {
                        type : 'checkbox',
                        name : 'Bck. image',
                        action : "element_attribute",
                        attribute: 'data-pgc-edit-bckimage',
                        value: '1',
                        empty_attribute: true
                    },
                    'data-pgc-edit-types' : {
                        type : 'select',
                        multiple: true,
                        can_add_items: false,
                        options: function() {
                            return pinegrow.insight.ext.components.getComponents();
                        },
                        name : 'Components',
                        action : "element_attribute",
                        attribute: 'data-pgc-edit-types'
                    }
                    /*, DISABLED for now,
                    'data-pgc-edit-advanced-label' : {
                        type : 'label',
                        name : 'Advanced',
                        action : "element_attribute",
                        attribute: 'dummy'
                    },
                    'data-pgc-edit-selector' : {
                        type : 'text',
                        name : 'Selector (based on comp. root)',
                        action : "element_attribute",
                        attribute : 'data-pgc-edit-selector',
                        placeholder: f.textForExample('>p.desc or self')
                    } */
                }
            }
        };
        pgcvar.meta = {
            helptext: 'Make the element editable. Inner content of the element is editable by default.',
            helplink: 'https://pinegrow.com/docs/pinegrow-pro/editable-areas/'
        }
        f.addComponentType(pgcvar);
        use_actions.push(pgcvar);


        var pgcrepeat = new PgComponentType('pgc.repeat', 'Define repeatable area');
        pgcrepeat.short_name = 'Repeat';
        pgcrepeat.selector = "[data-pgc-repeat]";
        pgcrepeat.attribute = 'data-pgc-repeat';
        pgcrepeat.action = true;
        pgcrepeat.not_main_type = true;
        pgcrepeat.sections = {
            'pgc.repeat.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-repeat' : {
                        type : 'text',
                        name : 'Field Id',
                        action : "element_attribute",
                        attribute : 'data-pgc-repeat',
                        attribute_keep_if_empty : true,
                        //validate : f.fieldIsRequired,
                        placeholder: f.textDefaultValue('editable field id')
                    } /*, DISABLED for now,
                    'data-pgc-repeat-advanced-label' : {
                        type : 'label',
                        name : 'Advanced',
                        action : "element_attribute",
                        attribute: 'dummy'
                    },
                    'data-pgc-repeat-selector' : {
                        type : 'text',
                        name : 'Selector (based on comp. root)',
                        action : "element_attribute",
                        attribute : 'data-pgc-repeat-selector',
                        placeholder: f.textForExample('.item or self')
                    } */
                }
            }
        };
        pgcrepeat.meta = {
            helptext: 'The editable area can be repeated by duplicating it as many times as needed.',
            helplink: 'https://pinegrow.com/docs/pinegrow-pro/editable-areas/#repeatable-elements'
        }
        f.addComponentType(pgcrepeat);
        use_actions.push(pgcrepeat);



        var pgcoptional = new PgComponentType('pgc.optional', 'Define optional area');
        pgcoptional.short_name = 'Optional';
        pgcoptional.selector = "[data-pgc-optional]";
        pgcoptional.attribute = 'data-pgc-optional';
        pgcoptional.action = true;
        pgcoptional.not_main_type = true;
        pgcoptional.sections = {
            'pgc.optional.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-optional' : {
                        type : 'text',
                        name : 'Field Id',
                        action : "element_attribute",
                        attribute : 'data-pgc-optional',
                        attribute_keep_if_empty : true,
                        //validate : f.fieldIsRequired,
                        placeholder: f.textDefaultValue('editable field id')
                    },
                    'data-pgc-optional-restore' : {
                        type : 'checkbox',
                        name : 'Restore on update',
                        action : "element_attribute",
                        attribute: 'data-pgc-optional-restore',
                        empty_attribute : true,
                        value: 'true'
                    }
                }
            }
        };
        pgcoptional.meta = {
            helptext: 'The element can be deleted from component instances. <span class=\'more\'>To get back deleted optional areas use <b>Actions -&gt; Restore optional areas</b> on instances / child pages - or - check <b>Restore on update</b>, update and then uncheck it (that will restore that field in all updated instances).</span>',
            helplink: 'https://pinegrow.com/docs/pinegrow-pro/editable-areas/#optional-areas'
        }
        f.addComponentType(pgcoptional);
        use_actions.push(pgcoptional);




        var pgcparam = new PgComponentType('pgc.param', 'Parameter');
        pgcparam.short_name = 'Parameter';
        pgcparam.selector = 'pgcparam';
        f.addComponentType(pgcparam);

/*
        var pgcrequire = new PgComponentType('pgc.require', 'Require file or folder');
        pgcrequire.short_name = 'Require';
        pgcrequire.selector = "[data-pgc-require]";
        pgcrequire.attribute = 'data-pgc-require';
        pgcrequire.action = true;
        pgcrequire.not_main_type = true;
        pgcrequire.sections = {
            'pgc.require.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-require' : {
                        type : 'text',
                        name : 'Path relative to project',
                        action : "element_attribute",
                        attribute : 'data-pgc-require',
                        attribute_keep_if_empty : false,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('assets/css/style.css')
                    },
                    'data-pgc-require-global' : {
                        type : 'checkbox',
                        name : 'All components require this file',
                        action : "element_attribute",
                        attribute : 'data-pgc-require-global',
                        attribute_keep_if_empty : false,
                        value: "true"
                    }
                }
            }
        };
        f.addComponentType(pgcrequire);
        actions.push(pgcrequire);
*/

        var pgcinstance = new PgComponentType('pgc.use', 'Use component');
        pgcinstance.short_name = 'Use comp.';
        pgcinstance.selector = "[data-pgc]";
        pgcinstance.attribute = 'data-pgc';
        pgcinstance.action = true;
        pgcinstance.not_main_type = true;
        pgcinstance.sections = {
            'pgc.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc' : {
                        type : 'select',
                        name : 'Id',
                        action : "element_attribute",
                        attribute : 'data-pgc',
                        attribute_keep_if_empty : true,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('my.poster'),
                        show_empty: true,
                        can_add_items: true,
                        options: function() {
                            return pinegrow.insight.ext.components.getComponents();
                        }
                    },
                    'data-pgc-no-update' : {
                        type : 'checkbox',
                        name : 'Don\'t update',
                        action : "element_attribute",
                        attribute: 'data-pgc-no-update',
                        empty_attribute : true,
                        value: '1',
                        on_changed : function(pgel, prop, value, oldValue, fieldDef, $field, eventType, values, crsaPage) {
                            if(!value) {
                                if(wouldUpdateChangeComponentInstance(pgel.clone(true /* no ids */), crsaPage)) {
                                    pinegrow.showQuickMessage('Be careful, updating components <b>will change</b> this component instance.', 4000, false, true /* highlight */);
                                }
                            } else {
                                pinegrow.showQuickMessage('The component <b>will not</b> be updated when component definition changes.', 3000);
                            }
                        }
                    }
                }
            }
        };
        f.addComponentType(pgcinstance);
        actions.push(pgcinstance);

        pgcinstance.meta = {
            helptext: 'Set element to be an instance of a component.'
        }


        var pgcsection = new PgComponentType('pgc.section', 'Define section');
        pgcsection.short_name = 'Section';
        pgcsection.selector = "[data-pgc-section]";
        pgcsection.attribute = 'data-pgc-section';
        pgcsection.action = true;
        pgcsection.not_main_type = true;
        pgcsection.sections = {
            'pgc.section.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-section' : {
                        type : 'text',
                        name : 'Name',
                        action : "element_attribute",
                        attribute : 'data-pgc-section',
                        attribute_keep_if_empty : false,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('My section')
                    }
                }
            }
        }
        pgcsection.on_action_added = pgcsection.on_action_removed = function() {
            pinegrow.showQuickMessage('Update whole project to activate sections changes.', 3000);
        }
        f.addComponentType(pgcsection);
        actions.push(pgcsection);

        pgcsection.meta = {
            helptext: 'Components defined within the section element are listed under this section in LIB panel.'
        }


        var pgcfield = new PgComponentType('pgc.field', 'Use editable area');
        pgcfield.short_name = 'Use Editable';
        pgcfield.selector = "[data-pgc-field]";
        pgcfield.attribute = 'data-pgc-field';
        pgcfield.action = true;
        pgcfield.not_main_type = true;
        pgcfield.get_action_tag = getEditableActionTag;

        pgcfield.sections = {
            'pgc.field.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-field' : {
                        type : 'select',
                        name : 'Field Id',
                        action : "element_attribute",
                        attribute : 'data-pgc-field',
                        attribute_keep_if_empty : false,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('title'),
                        can_add_items: true,
                        options: function(fdef, pgel) {
                            return pinegrow.insight.ext.components.getEditableFieldsForElement(pgel);
                        }
                    }
                }
            }
        };
        f.addComponentType(pgcfield);
        use_actions.push(pgcfield);

        pgcfield.meta = {
            helptext: 'Use component field. This action is added automatically when components are used. <b>Removing this action will cause edits to be lost.</b>'
        }

        var pgclock = new PgComponentType('pgc.lock', 'Lock area');
        pgclock.short_name = 'Lock';
        pgclock.selector = "[data-pgc-lock]";
        pgclock.attribute = 'data-pgc-lock';
        pgclock.action = true;
        pgclock.not_main_type = true;
        // TODO: @add-icon lock icon (fa-lock)
        pgclock.get_action_tag = function() { return '<i class="icon icon-Lock_open"></i>'; };


        f.addComponentType(pgclock);
        use_actions.push(pgclock);

        pgclock.meta = {
            helptext: 'Lock all elements inside the element, except editable areas. This is similar to CMS mode.'
        }

        var master_actions = [];

        var pgcmaster = new PgComponentType('pgc.master', 'Set as master page');
        pgcmaster.short_name = 'Master';
        pgcmaster.selector = "[data-pgc-set-master]";
        pgcmaster.attribute = 'data-pgc-set-master';
        pgcmaster.action = true;
        pgcmaster.not_main_type = true;

        pgcmaster.on_action_added = function(pgel, cp, action_def, $el) {
            var fileNode = cp.getFileNode();
            var project = cp.getProject();
            if(project) {
                if(fileNode) {
                    project.updateTagsForFile( fileNode );
                }
            } else {
                masterPagesRequireProjectNotice();
            }
        }
        pgcmaster.on_action_removed = function(pgel, cp, action_def, $el) {
            var fileNode = cp.getFileNode();
            if(fileNode) {
                // setFileNodeTagMaster(false, fileNode);
                var project = cp.getProject();
                if(project) {
                    project.updateTagsForFile( fileNode );
                }
            }
        }

        pgcmaster.on_before_add_action = function(pgel, action_def) {
            if(pgel.tagName !== 'html') {
                var html = pgel.closest('html');
                if(html) {
                    pinegrow.selectElement(html);
                    pinegrow.showQuickMessage(`Adding the action to the top HTML element...`);
                }
            }
        }

        pgcmaster.on_can_add_action = function(pgel, cp, action_def, $el) {
            if(pgel.tagName != 'html') {
                pinegrow.showQuickMessage('<b>Wrong element!</b> Set master page actions on <b>the TOP node of the page</b> in the tree.', 5000, false, true /* error */);
                return false;
            }
            if((action_def == pgcmaster && pgel.hasAttr('data-pgc-master')) || (action_def == pgcusemaster && pgel.hasAttr('data-pgc-set-master'))) {
                pinegrow.showQuickMessage('The page can\'t BE A MASTER and USE A MASTER at the same time.', 5000, false, true /* error */);
                return false;
            }
            return true;
        }

        pgcmaster.meta = {
            helptext: 'Define the page as a master page that is used as a template for other pages.'
        }

        f.addComponentType(pgcmaster);
        master_actions.push(pgcmaster);

        var masterPagesRequireProjectNotice = function() {
            pinegrow.showAlert('<p>Master pages work with Projects. To use Master pages open the folder where your pages are located with <b>File -&gt; Open project</b>.</p>', 'Master pages and components require projects');
        }

        var pgcusemaster = new PgComponentType('pgc.master.use', 'Use master page');
        pgcusemaster.short_name = 'Use master';
        pgcusemaster.selector = "[data-pgc-master]";
        pgcusemaster.attribute = 'data-pgc-master';
        pgcusemaster.action = true;
        pgcusemaster.not_main_type = true;
        pgcusemaster.sections = {
            'pgc.master.use.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-master' : {
                        type : 'select',
                        name : 'Master page',
                        action : "element_attribute",
                        attribute : 'data-pgc-master',
                        attribute_keep_if_empty : false,
                        validate : f.fieldIsRequired,
                        placeholder: f.textForExample('master.html'),
                        can_add_items: true,
                        options: function() {
                            return pinegrow.insight.ext.components.getMasterPages();
                        }
                    }
                }
            }
        };
        pgcusemaster.on_action_added = function(pgel, cp, action_def, $el) {
            var project = cp.getProject();
            if(project) {
                var fileNode = cp.getFileNode();
                if(fileNode) {
                    // setFileNodeTagUseMaster(true, fileNode);
                    project.updateTagsForFile( fileNode );
                }
            } else {
                masterPagesRequireProjectNotice();
            }
            setHasUpdates(true);
        }
        pgcusemaster.on_action_removed = function(pgel, cp, action_def, $el) {
            var fileNode = cp.getFileNode();
            if(fileNode) {
                // setFileNodeTagUseMaster(false, fileNode);
                var project = cp.getProject();
                if(project) {
                    project.updateTagsForFile( fileNode );
                }
            }
            setHasUpdates(true);
        }
        pgcusemaster.on_can_add_action = pgcmaster.on_can_add_action;
        pgcusemaster.on_before_add_action = pgcmaster.on_before_add_action;

        pgcusemaster.meta = {
            helptext: 'Select which master page will be used as the template for this page.'
        }
        f.addComponentType(pgcusemaster);
        master_actions.push(pgcusemaster);

        var partials = [];

        var pgcsavepartial = new PgComponentType('pgc.partial.save', 'Partial');
        pgcsavepartial.short_name = 'Partial';
        pgcsavepartial.selector = "[data-pgc-save-partial]";
        pgcsavepartial.attribute = 'data-pgc-save-partial';
        pgcsavepartial.action = true;
        pgcsavepartial.not_main_type = true;
        pgcsavepartial.sections = {
            'pgc.save.partial.parameters' : {
                name : 'Options',
                fields : {
                    'data-pgc-save-partial' : {
                        type : 'text',
                        name : 'File',
                        action : "element_attribute",
                        attribute : 'data-pgc-save-partial',
                        attribute_keep_if_empty : false,
                        validate : f.fieldIsRequired,
                        file_picker : true,
                        file_picker_save : true,
                        file_picker_no_url : true,
                        file_picker_no_proxy : true,
                        file_picker_on_picked: function(url, pgel) {
                            var path = require('path');
                            var page = pgel.getPage();
                            if(page.localFile) {
                                var folder = path.dirname(page.localFile);
                                if(url.startsWith(folder)) {
                                    url = path.relative(path.dirname(page.localFile), url);
                                    //url = url.replace(folder + path.sep, '');
                                }
                            }
                            return url;
                        },
                        helptext: `<p>The file name where the partial is located. Relative paths are resolved from the page file location. For example, using "mypartial.html" will save the partial in the same folder where its parent page is located.</p>`
                    },
                    'data-pgc-save-partial-operation' : {
                        type : 'select',
                        name : 'Operation',
                        action : "element_attribute",
                        attribute: 'data-pgc-save-partial-operation',
                        default_value: 'save',
                        options: [
                            {key: 'save', name: 'Save'},
                            {key: 'load', name: 'Load'},
                            {key: 'save_load', name: 'Save & Load'},
                            {key: 'none', name: 'None'}
                        ],
                        helptext: `<p>What should happen with the partial:</p>
<ul>
<li><b>Save</b> - save the partial when the page is saved.</li>
<li><b>Load</b> - load the partial from its file when the page is opened.</li>
<li><b>Save &amp; Load</b> - save and load the partial on page save and load.</li>
<li><b>None</b> - don't do anything.</li>
</ul>`
                    },
                    'data-pgc-save-partial-content' : {
                        type : 'checkbox',
                        name : 'Use only the inner content',
                        action : "element_attribute",
                        attribute: 'data-pgc-save-partial-content',
                        empty_attribute : true,
                        value: '1'
                    },
                    'data-pgc-save-partial-save' : {
                        type : 'button',
                        name : 'Save',
                        func: function(pgel) {
                            savePartials(pgel.getPage(), [pgel], true);
                        }
                    },
                    'data-pgc-save-partial-reload' : {
                        type : 'button',
                        name : 'Reload',
                        func: function(pgel) {
                            pinegrow.makeChanges(pgel, 'Reload partial', function() {
                                if(loadPartials(pgel.getPage(), [pgel], true /* with events */, true /* force load */)) {
                                    pinegrow.showQuickMessage('The partial was reloaded.');
                                }
                            })
                        }
                    }
                }
            }
        }
        f.addComponentType(pgcsavepartial);
        partials.push(pgcsavepartial);

        pgcsavepartial.meta = {
            helptext: 'Save or load the element HTML code from a separate file.'
        }


        var getPartialOperation = function(pgel) {
            return pgel.getAttr('data-pgc-save-partial-operation') || 'save';
        }


        if(true || canUse()) {
            var section = new PgFrameworkLibSection('pg.components.master', 'Master pages');
            section.setComponentTypes( master_actions );
            //section.closed = false; //open by default
            f.addActionsSection(section);

            section = new PgFrameworkLibSection('pg.components', 'Components');
            section.setComponentTypes( actions );
            //section.closed = false; //open by default
            f.addActionsSection(section);

            section = new PgFrameworkLibSection('pg.components.editable', 'Editable areas');
            section.setComponentTypes( use_actions );
            //section.closed = false; //open by default
            f.addActionsSection(section);

            section = new PgFrameworkLibSection('pg.components.partials', 'Partials');
            section.setComponentTypes( partials );
            //section.closed = false; //open by default
            f.addActionsSection(section);
        }


        var resourceEditor;

        var showResourceEditorForProject = function(project) {
            if (!resourceEditor || (resourceEditor && !resourceEditor.getProject())) {
                resourceEditor = new PgResourceEditor(project);
            }
            resourceEditor.show();
        }

        f.on_project_menu = function(page, project) {
            var items = [];
/*
            items.push({
                label: 'Edit project library resources...',
                func: function() {
                    showResourceEditorForProject(project);
                }
            })

            items.push({
                label: 'Reload &amp; update components',
                key: 'SHIFT CMD U',
                func: function() {
                    f.reloadProjectAndUpdateComponents();
                }
            })

            items.push({
                label: 'Add or update resources...',
                kbd: null,
                func: function() {
                    pinegrow.showUpdateResourcesDialog(page, project);
                }
            })*/
            /*
            items.push({
                label: 'Reload components',
                func: function() {
                    refreshComponents(project);
                }
            })
            items.push({
                label: 'Update components',
                func: function() {
                    updateComponents(project);
                }
            })
            */
            return items;
        }

        var setFileNodeTagMaster = function(master, fileNode) {
            if(master) {
                fileNode.addTag(new CrsaFileTag('Master', 'primary', 'This is a master page.', '24'));
            } else {
                fileNode.removeTag('Master');
            }
        }

        var setFileNodeTagUseMaster = function(use_master, fileNode) {
            if(use_master) {
                fileNode.addTag(new CrsaFileTag('UseMaster', 'info', 'This page uses the master page ' + use_master + '.', '24'));
            } else {
                fileNode.removeTag('UseMaster');
            }
        }

        var setFileNodeTagDefinesComponents = function(defines, fileNode) {
            if(defines) {
                fileNode.addTag(new CrsaFileTag('DefC', 'primary', 'Components are defined on this page', 'actions'));
            } else {
                fileNode.removeTag('DefC');
            }
        }

        var setFileNodeTagUsesComponents = function(uses, fileNode) {
            if(uses) {
                fileNode.addTag(new CrsaFileTag('UseC', 'info', 'Components are used on this page', 'actions'));
            } else {
                fileNode.removeTag('UseC');
            }
        }

        var requireAtLeastOneOpenPage = function(project, func) {
            var found = null;
            project.forEachOpenPage(function(page) {
                found = page;
                if(func) func(page);
                func = null;
            })
            if(!found) {
                pinegrow.showAlert('<p>Updating the whole project needs at least one page from the project to be open in Pinegrow.</p><p><b>Open any editable page from the project</b> and run this action again.</p>', 'At least one page needs to be open.');
            }
        }

        var updateProjectResources = function (project) {
            if (!project || !project.framework) return;

            var info = project.getProjectInfo();
            var resources = info.getSetting("project_resources");
            var pf = project.framework;

            pf.resources.clear();
            if (resources) {
                for (var j=0; j < resources.length; j++) {
                    var url = resources[j].url;
                    var file = url;
                    if (!crsaIsAbsoluteUrl(url))
                        file = project.getAbsoluteUrl( url );
                    var footer = resources[j].footer || false;
                    var type = (resources[j].type == 'auto') ? '' : resources[j].type;

                    if(!type && pinegrow.isFileEditable(file)) {
                        type = 'pinegrow/pagelib';
                    }

                    if (file) {
                        if(type && type == 'pinegrow/pagelib') {
                            pf.addPageLib({
                                url: file,
                                name: crsaGetNameFromUrl(file),
                                selector: '*',
                                framework: pf
                            });
                        } else {
                            var res = new PgComponentTypeResource(file);
                            res.footer = footer;
                            res.relative_url = url;
                            res.source = file;
                            if (type) res.type = type;

                            pf.resources.add(res);
                        }
                    }
                }
            }
        }

        var refreshComponents = function(project, done, skip_frameworks_changed) {

            var all = [];
            var all_map = {};

            var global_resources = [];
            var master_pages = {};

            var sections = {'default': []};
            var sections_order = ['default'];

            project.forEachEditableFile( function( page, pageDone, status, fileNode ) {
                //on page
                var info = readComponentsInPage(page, global_resources, project);
                for(var i = 0; i < info.components.length; i++) {
                    if(info.components[i].type in all_map) {
                        status.warnings.push('Component <b>' + info.components[i].type + '</b> is <b>already defined</b> in ' + project.getRelativePath(all_map[info.components[i].type].sourceFile) + '.');
                    } else {
                        all.push(info.components[i]);
                        all_map[info.components[i].type] = info.components[i];

                        var sec = info.components[i].add_to_section || 'default';
                        if(!sections[sec]) {
                            sections[sec] = [];
                            sections_order.push(sec);
                        }
                        sections[sec].push(info.components[i]);
                    }
                }
                if(info.master) {
                    info.master.file = page.getFileNode();
                    info.master.image = crsaMakeUrlFromFile(getImagePathForPage(page, project));

                    if(master_pages[info.master.type]) {
                        status.warnings.push('Master page <b>' + info.master.type + '</b> is <b>already defined</b> in ' + project.getRelativePath(master_pages[info.master.type].file.path) + '.');
                    } else {
                        master_pages[info.master.type] = info.master;
                    }
                }

                setFileNodeTagUsesComponents(info.used_component_types.length, fileNode);
                setFileNodeTagDefinesComponents(info.components.length, fileNode);
                setFileNodeTagMaster(info.master, fileNode);
                setFileNodeTagUseMaster(info.use_master, fileNode);

                page.setSyntaxErrorIndicator(page.sourceNode && page.sourceNode.validateTree().length > 0, fileNode, project);
                //fileNode.addTag(new CrsaFileTag('Ok', 'success', 'Something dangerous'));
                //fileNode.addTag(new CrsaFileTag('Hmm', 'warning', 'Something dangerous'));

                project.updateTagsForFile( fileNode );

                pageDone(page, status);

            }, function() {
                //on done
                var f_key = project.getName();
                var pf = pinegrow.findFrameworkByKey(f_key);
                var new_f = false;
                if(!pf) {
                    pf = new PgFramework( f_key, f_key);
                    new_f = true;
                }
                pf.auto_updatable = true;
                pf.removeAllComponents();
                pf.master_pages = master_pages;

                pf.preview_images_base = 'previews';

                for(var i = 0; i < all.length; i++) {
                    pf.addComponentType(all[i]);
                }

                for(var i = 0; i < sections_order.length; i++) {
                    var section = new PgFrameworkLibSection(project.getName() + '.' + sections_order[i], sections_order[i] == 'default' ? project.getName() : sections_order[i]);
                    section.setComponentTypes( sections[sections_order[i]] );
                    pf.addLibSection(section);
                }

                if(new_f) {
                    pinegrow.addFramework(pf, -1000);
                }

                project.framework = pf;
                pf.project = project;

                updateProjectResources(project);

                if(!skip_frameworks_changed) {
                    pinegrow.frameworksChanged();
                }

                pf.components_loaded = true;

                pinegrow.showQuickMessage('Found ' + all.length + ' components in ' + project.getName() + '.');

                if(done) done(project, pf);

            }, 'Looking for components...', null, getIgnoreMap(project))
        }

        var areComponentsEqual = function(a, b) {
            return (a.definition_code == b.definition_code && a.code == b.code) && areComponentsEqualExceptCode(a, b);
        }

        var areComponentsEqualExceptCode = function(a, b) {
            return !(a.type != b.type || a.name != b.name || a.selector != b.selector || a.parent_selector != b.parent_selector || a.preview_image != b.preview_image || a.button_image != b.button_image || a.auto_update != b.auto_update || a.description != b.description || a.map_urls != b.map_urls);
        }



        var copyComponentToComponent = function(source, dest) {
            //type must be the same
            dest.name = source.name;
            dest.selector = source.selector;
            dest.parent_selector = source.parent_selector;
            dest.code = source.code;
            dest.definition_code = source.definition_code;
            dest.definition_attributes = source.definition_attributes;
            dest.preview_image = source.preview_image;
            dest.button_image = source.button_image;
            dest.description = source.description;
            dest.component_fields = source.component_fields;
            dest.component_repeat_fields = source.component_repeat_fields;
            dest.auto_update = source.auto_update;
            dest.map_urls = source.map_urls;
        }

        var quickRefreshComponentsInPage = function(page, project) {
            var f_key = project.getName();
            var pf = pinegrow.findFrameworkByKey(f_key);
            var new_f = false;
            if(!pf) {
                throw "No framework!";
            }

            var global_resources = [];
            var changed = false;

            var info = readComponentsInPage(page, global_resources, project);
            for(var i = 0; i < info.components.length; i++) {
                var c = info.components[i];
                var current = pf.getComponentType(c.type);
                if(!current) {
                    pf.addComponentType(c);
                    var section_key = project.getName() + '.' + (c.add_to_section || 'default');
                    var section = pf.getLibSection(section_key);
                    if(!section) {
                        section = new PgFrameworkLibSection(section_key, (c.add_to_section || project.getName()));
                        pf.addLibSection(section);
                    }
                    section.addComponentType(c);
                    changed = true;
                    pinegrow.showQuickMessage('Component ' + c.name + ' added.');
                } else {
                    //changed?
                    if(!areComponentsEqual(current, c)) {
                        copyComponentToComponent(c, current);
                        //pf.replaceComponent(current, c);
                        changed = changed || !areComponentsEqualExceptCode(current, c); //no need to refresh libs
                        pinegrow.showQuickMessage('Component ' + c.name + ' changed.');
                    }
                }
            }
            if(info.master) {
                pf.master_pages[info.master.type] = info.master;
            }
            if(changed) {
                pinegrow.frameworksChanged();
            }
        }

        var quickUpdateComponentsInPage = function(page, project, return_status, done) {
            return_status.changed_collection = new PgCollection();
            updateComponentsInPage(page, project, null, false, return_status, function(a, b, c, d) {
                if(return_status.changed_collection.getLength() > 0 && page.openedInEditor) {
                    pinegrow.changeManager.didChange(status.changed_collection, 'Quick components update', {}, page);
                }
                if(done) done(a, b, c, d);
            });
        }

        f.changeComponentDefinitionsToComponentInstances = function(pgel) {
            pgel.walkSelfAndChildren(function(node) {
                var comp_def = node.getAttr('data-pgc-define');
                if(comp_def) {
                    node.removeAttrIfStartsWith('data-pgc-define'); //remove def
                    node.setAttr('data-pgc', comp_def); //add use
                }
                var edit_area = node.getAttr('data-pgc-edit')
                var field_id = null;
                if(edit_area) {
                    var field_info = pgcVarParseValue(edit_area);
                    node.removeAttrIfStartsWith('data-pgc-edit');
                    node.setAttr('data-pgc-field', field_info.name);
                    field_id = field_info.name;
                }
                if(node.hasAttr('data-pgc-repeat')) {
                    var repeat = node.getAttr('data-pgc-repeat') || field_id;
                    node.removeAttrIfStartsWith('data-pgc-repeat');
                    if(!field_id) {
                        node.setAttr('data-pgc-field', repeat);
                    }
                }
                if(node.hasAttr('data-pgc-optional')) {
                    var optional = node.getAttr('data-pgc-optional') || field_id;
                    node.removeAttrIfStartsWith('data-pgc-optional');
                    if(!field_id) {
                        node.setAttr('data-pgc-field', optional);
                    }
                }
                if(node.hasAttr('data-pgc-section')) {
                    node.removeAttr('data-pgc-section');
                }

                return true;
            })
        }

        var getComponentDefinitionAttributes = function(cel) {
            var list = cel.getAttrList();
            var r = {};
            list.forEach(function(a) {
                if(a.name.startsWith('data-pgc-define')) {
                    r[a.name] = a.value;
                }
            })
            return r;
        }

        var readComponentsInPage = function(page, global_resources, project, only_def_type) {
            var list = [];

            var r = {components: [], master: null, use_master: null, used_component_types: []};

            if(!page.sourceNode) return r;

            var getSection = function(pgel) {
                var secel = pgel.closest('[data-pgc-section]');
                if(secel) {
                    return secel.getAttr('data-pgc-section');
                }
            }

            //var elements = page.sourceNode.find('[data-pgc-define]');

            var html = page.sourceNode.findOne('html');
            if(html) {
                r.master = html.hasAttr('data-pgc-set-master');
                r.use_master = html.getAttr('data-pgc-master');
            }

            var elements = [];

            if(r.master && !only_def_type) {
                elements.push(page.sourceNode); //master is a special whole-page component
            }

            page.sourceNode.walk(function(node) {
                var c = node.getAttr('data-pgc-define');
                if(c !== null && (!only_def_type || only_def_type == c)) {
                    elements.push(node);
                }
                var used_c = node.getAttr('data-pgc');
                if(used_c) {
                    if(r.used_component_types.indexOf(used_c) < 0) {
                        r.used_component_types.push(used_c);
                    }
                }
                return true;
            })

            for(var i = 0; i < elements.length; i++) {

                var is_master = elements[i] == page.sourceNode;

                var cel = elements[i].clone(true /* no pg ids */);
                var type = is_master ? ('master.' + page.name) : cel.getAttr('data-pgc-define');
                var name = is_master ? page.name : (cel.getAttr('data-pgc-define-name') || type);
                var custom_selector = cel.getAttr('data-pgc-define-selector');

                var definition_attributes = getComponentDefinitionAttributes(cel);

                if(is_master) {
                    var comp_defs = cel.find('[data-pgc-define]');
                    for(var j = 0; j < comp_defs.length; j++) {
                        f.changeComponentDefinitionsToComponentInstances(comp_defs[j]);
                    }
                    cel.removeAttrIfStartsWith('data-pgc-section', true /* in children also */);
                }

                cel.removeAttrIfStartsWith('data-pgc-define', false /* not in children also */);//mtch

                var params = cel.find('pgcparam');
                for(var j = 0; j < params.length; j++) {
                    params[j].remove();
                }

                if(type) {
                    var c = new PgComponentType(type, name);

                    c.definition_code = cel.toStringOriginal(true); //remember comp def so that we can see if def changed in quick update
                    c.definition_attributes = definition_attributes;

                    if(!custom_selector) {
                        cel.setAttr('data-pgc', type);
                    }

                    if(is_master) {
                        var cel_html = cel.findOne('html');
                        cel_html.removeAttr('data-pgc-set-master');
                        cel_html.setAttr('data-pgc-master', page.name);
                        c.master_page_source_node = cel;
                        c.master_page_url = page.url;
                    }

                    c.sourceFile = page.localFile || page.url;
                    c.sourceUrl = page.url;

                    c.selector = custom_selector || '[data-pgc="' + type + '"],[data-pgc-define="' + type + '"]';
                    c.tags = 'block';
                    c.priority = 10;
                    c.is_smart_component = true;

                    c.auto_update = elements[i].getAttr('data-pgc-define-auto-update') != 'false';
                    c.map_urls = elements[i].getAttr('data-pgc-define-map-urls') != 'false';

                    c.add_to_section = getSection(elements[i]);

                    if(isApp()) {
                        var preview_image = path.join(project.getDir(), 'previews', type + '.png');
                        if (crsaIsFileOrDir(preview_image, fs) == 'file') {
                            c.preview_image = type + '.png';
                            if (!elements[i].hasAttr('data-pgc-define-photo-preview-only')) {
                                c.button_image = c.preview_image;
                            } else {
                                c.button_image = null;
                            }
                        }
                    }

                    c.description = elements[i].getAttr('data-pgc-define-description');

                    c.component_fields = {};
                    c.component_repeat_fields = {};

                    var skipSubcomponents = function(node) {
                        return node.hasAttr('data-pgc-define');
                    }

                    //to make things less verbose replace the actuall field def in component code with field_id that will be used to lookup the field definition
                    var fields = cel.findIncludingSelf('[data-pgc-edit]', false, skipSubcomponents);
                    for(var m = 0; m < fields.length; m++) {
                        var field_def = fields[m].getAttr('data-pgc-edit');
                        var field_info = pgcVarParseValue(field_def);
                        field_info.selector = fields[m].getAttr('data-pgc-edit-selector');
                        var cls = fields[m].getAttr('data-pgc-edit-classes');
                        field_info.classes = crsaSplitAndTrimList(cls);
                        field_info.bckimage = fields[m].hasAttr('data-pgc-edit-bckimage');

                        var types = fields[m].getAttr('data-pgc-edit-types');
                        if(types) {
                            field_info.types = crsaSplitAndTrimList(types);
                            if(field_info.types.length) field_info.content = true;
                        }

                        field_info.extensions = {};
                        fields[m].findAttributesStartingWith('data-pgc-edit-ext-').forEach(function(attr) {
                            var key = attr.name.replace('data-pgc-edit-ext-', '');
                            field_info.extensions[key] = attr.value;
                        })

                        if(field_info.name) {
                            fields[m].setAttr('data-pgc-field', field_info.name);
                            //remember field def
                            c.component_fields[field_info.name] = field_info;
                        }
                        fields[m].removeAttrIfStartsWith('data-pgc-edit');
                    }

                    var repeatable = cel.findIncludingSelf('[data-pgc-repeat]', false, skipSubcomponents);
                    for(var m = 0; m < repeatable.length; m++) {
                        var field_id = repeatable[m].getAttr('data-pgc-field') || repeatable[m].getAttr('data-pgc-repeat');
                        if(field_id) {
                            repeatable[m].setAttr('data-pgc-field', field_id);
                            repeatable[m].removeAttrIfStartsWith('data-pgc-repeat');
                            c.component_repeat_fields[field_id] = true;
                            if(!(field_id in c.component_fields)) {
                                c.component_fields[field_id] = {name: field_id, content: false, attrs: [], classes: [], repeat: true};
                            } else {
                                c.component_fields[field_id].repeat = true;
                            }
                        }
                    }

                    var optional = cel.findIncludingSelf('[data-pgc-optional]', false, skipSubcomponents);
                    for(var m = 0; m < optional.length; m++) {
                        //if(optional[m].hasAttr('data-pgc-optional-restore')) continue;

                        var restore = optional[m].hasAttr('data-pgc-optional-restore');

                        var field_id = optional[m].getAttr('data-pgc-field') || optional[m].getAttr('data-pgc-optional');
                        if(field_id) {
                            optional[m].setAttr('data-pgc-field', field_id);
                            optional[m].removeAttrIfStartsWith('data-pgc-optional');

                            if(!(field_id in c.component_fields)) {
                                c.component_fields[field_id] = {name: field_id, content: false, attrs: [], classes: [], repeat: true, optional: !restore};
                            } else {
                                c.component_fields[field_id].optional = !restore;
                            }
                        }
                    }

                    if(!is_master) {
                        f.changeComponentDefinitionsToComponentInstances(cel);
                    }


                    //find and remember advanced editable areas
                    /*
                    c.editable_areas = [];
                    var editable = cel.findIncludingSelf('[data-pgc-edit-selector]');
                    for(var m = 0; m < editable.length; m++) {
                        var sel = editable[m].getAttr('data-pgc-edit-selector');
                        if(sel) {
                            var edit_def = pgcVarParseValue(editable[m].getAttr('data-pgc-edit'));
                            edit_def.selector = sel;
                            edit_def.classes = editable[m].getAttr('data-pgc-edit-classes');
                            c.editable_areas.push(edit_def);
                            editable[m].removeAttrIfStartsWith('data-pgc-edit');
                        }
                    }
                    */
                    //done advanced editable areas

                    //advanced repeats
                    /*
                    c.repeatable_areas = [];
                    var repeatable = cel.findIncludingSelf('[data-pgc-repeat-selector]');
                    for(var m = 0; m < repeatable.length; m++) {
                        var sel = repeatable[m].getAttr('data-pgc-repeat-selector');
                        if(sel) {
                            c.repeatable_areas.push({id: repeatable[m].getAttr('data-pgc-repeat'), selector: sel});
                            repeatable[m].removeAttrIfStartsWith('data-pgc-repeat');
                        }
                    }
*/

                    //data-pgc-define-auto-update

                    c.code = cel.toStringOriginal(true);

                    if(is_master) {
                        r.master = c;
                    } else {
                        r.components.push(c);
                    }

                    var info = project.getProjectInfo();
                    var resources = info.getSetting("project_resources");


                    /*
                    MT: Not sure if we use component specific resources anywhere. Let's comment this out and see if anything breaks...

                    if (resources) {
                        for (var j=0; j < resources.length; j++) {
                            var url = resources[j].url;
                            var file = url;
                            if (!crsaIsAbsoluteUrl(url))
                                file = project.getAbsoluteUrl( url );

                            var footer = resources[j].footer || false;
                            var type = (resources[j].type == 'auto') ? '' : resources[j].type;

                            if (file) {
                                if(type && type == 'pinegrow/pagelib') {

                                } else {
                                    var res = new PgComponentTypeResource(file);
                                    res.footer = footer;
                                    res.relative_url = url;
                                    res.source = file;
                                    if (type) res.type = type;

                                    c.addResource(res);
                                }
                            }
                        }
                    }*/

                    // var reqs = elements[i].find('[data-pgc-require]');
                    // for(var j = 0; j < reqs.length; j++) {
                    //     var relative_url = reqs[j].getAttr('data-pgc-require');
                    //     var file = project.getAbsoluteUrl( relative_url );
                    //     var global = reqs[j].getAttr('data-pgc-require-global') == 'true';
                    //     var footer = reqs[j].getAttr('data-pgc-require-footer') == 'true';

                    //     if(file) {
                    //         var res = new PgComponentTypeResource(file);
                    //         res.footer = footer;
                    //         res.relative_url = relative_url; //'components/' + project.getName() + '/' + relative_url;
                    //         res.source = project.getAbsolutePath(relative_url);

                    //         if(global) {
                    //             global_resources.push(res);
                    //         } else {
                    //             c.addResource(res);
                    //         }
                    //     }
                    // }
                }
            }
            return r;
        }

        var updateComponents = function(project, done, reference_page) {
            var return_status = new PgComponentsUpdateStatus();

            var component_types_info = reference_page.getComponentTypesInformation();

            project.forEachEditableFile( function( page, pageDone, status ) {
                //on page
                if(!page.openedInEditor) {
                    page.setComponentTypesInformation(component_types_info);
                }
                if(updateComponentsInPage(page, project, null, false, return_status)) {
                    status.changed = true;
                }
                pageDone(page, status);

            }, function() {
                //on done
                if(done) done();

                showSavedDataNotice(return_status);

                return_status.show();

            }, 'Updating components...', null, getIgnoreMap(project))
        }

        var getClassRegEx = function(cls) {
            if(cls.startsWith('/')) {
                try {
                    return new RegExp(cls.replace(/\//g, ''), '');
                }
                catch(err) {}
            }
            return null;
        }

        var getEditableData = function(node, field_info, fields) {
            var field_id = field_info.name;

            if(!fields[field_id]) fields[field_id] = {items: [], count: 0};
            var values = {restored: false};
            if(field_info.content) {
                values.html_content = node.html();
            }
            for(var ai = 0; ai < field_info.attrs.length; ai++) {
                values[field_info.attrs[ai]] = node.getAttr(field_info.attrs[ai]);
            }
            values.classes = {};
            var classes = /* node.getAttr('data-pgc-edit-classes') || */ field_info.classes;

            var cls_list = node.getClasses();

            for(var j = 0; j < field_info.classes.length; j++) {
                var clsre = getClassRegEx(field_info.classes[j]);
                if(clsre) {
                    for(var i = 0; i < cls_list.length; i++) {
                        if(cls_list[i].match(clsre)) {
                            values.classes[cls_list[i]] = true;
                        }
                    }
                } else {
                    values.classes[field_info.classes[j]] = node.hasClass(field_info.classes[j]);
                }
            }

            if(field_info.bckimage) {
                values.bckimage = node.getInlineStylePropertyValue('background-image', true);
            }

            $.each(field_info.extensions, function(k, v) {
                var ext = pgcvar.pgcGetExtensionForFieldKey(k);
                if(ext) {
                    var val = ext.getValue(k, node);
                    if(val !== null) {
                        values[k] = val;
                    }
                }
            })

            fields[field_id].items.push(values);
        }


        var fillSavedDataIntoComponent = function(node, field_info, fields) {
            var field_id = field_info.name;
            if(field_id && fields[field_id]) {
                var c = fields[field_id].count;
                if(c >= fields[field_id].items.length) c = 0;
                var values = fields[field_id].items[c];

                if(field_info.content && values.html_content) {
                    node.html(values.html_content);
                }
                for(var ai = 0; ai < field_info.attrs.length; ai++) {
                    if(values[field_info.attrs[ai]]) {
                        node.setAttr(field_info.attrs[ai], values[field_info.attrs[ai]]);
                    }
                }
                var classes = /* node.getAttr('data-pgc-edit-classes') || */ field_info.classes;

                var cls_list = node.getClasses();
                var clsres = [];

                for(var j = 0; j < field_info.classes.length; j++) {
                    var clsre = getClassRegEx(field_info.classes[j]);
                    if(clsre) {
                        for(var i = 0; i < cls_list.length; i++) {
                            if(cls_list[i].match(clsre)) {
                                node.removeClass(cls_list[i]);
                            }
                        }
                        clsres.push(clsre);
                    } else {
                        if(values.classes[field_info.classes[j]]) {
                            node.addClass(field_info.classes[j]);
                        } else {
                            node.removeClass(field_info.classes[j]);
                        }
                    }
                }
                if(clsres.length) {
                    for(var i = 0; i < clsres.length; i++) {
                        $.each(values.classes, function(cls, has) {
                            if(cls.match(clsres[i])) {
                                if(has) {
                                    node.addClass(cls);
                                } else {
                                    //already removed above
                                }
                            }
                        })
                    }
                }

                if(field_info.bckimage) {
                    if(values.hasOwnProperty('bckimage')) {
                        node.setInlineStylePropertyValue('background-image', values.bckimage, true);
                    }
                }

                $.each(field_info.extensions, function(k, v) {
                    var ext = pgcvar.pgcGetExtensionForFieldKey(k);
                    if(ext) {
                        ext.setValue(k, values[k] || null, node);
                    }
                })

                values.restored = true;
                c++;
                fields[field_id].count = c;
            }
        }

        var getUnrestoredFields = function(fields) {
            var list = [];
            $.each(fields, function(id, data) {
                for(var i = 0; i < data.items.length; i++) {
                    if(!data.items[i].restored) {
                        data.items[i].field_id = id;
                        list.push(data.items[i]);
                    }
                }
            })
            return list;
        }

        var getSavedDataForElement = function(pgel, remove) {
            var data = null;
            for(var i = 0; i < pgel.children.length; i++) {
                if(pgel.children[i].comment) {
                    if(pgel.children[i].attributes.indexOf('[Pinegrow components saved data') >= 0) {
                        data = $.trim(pgel.children[i].attributes.replace(/\[Pinegrow components saved data[^\]]*\]/, '').replace(/\-\-$/,''));
                        data = data.replace(/\[\[\[\-\-/g, '<!--').replace(/\-\-\]\]\]/g, '-->');
                        if(remove) {
                            pgel.children[i].remove();
                        }
                    }
                }
            }
            return data;
        }

        var removeSavedDataFromElement = function(node) {
            var r = [];
            node.walkSelfAndChildren(function(pgel) {
                if(pgel.comment) {
                    if(pgel.attributes.indexOf('[Pinegrow components saved data') >= 0) {
                        r.push(pgel);
                    }
                }
                return true; //continue
            })
            for(var i = 0; i < r.length; i++) {
                r[i].remove();
            }
            return r.length;
        }

        var clearSavedDataFromPage = function(page) {
            var count;
            pinegrow.makeChanges(page, 'Clear saved unused edits', function() {
                page.sourceNode.withoutEvents(function() {
                    count = removeSavedDataFromElement(page.sourceNode);
                })
            })
            return count;
        }

        //Get field_info for field_id from component def c
        var getFieldInfo = function(field_id, c) {
            if(c.component_fields && (field_id in c.component_fields)) return c.component_fields[field_id];
            return null;
        }

        var walkSelfAndChildrenButDontGoInComponents = function(start_node, func, master) {
            start_node.walkSelfAndChildren(function(node) {
                //why !master ???
                //if(!master && node != start_node && (node.hasAttr('data-pgc') || node.hasAttr('data-pgc-define'))) return 'skip_children';
                //try without
                if(node != start_node && (node.hasAttr('data-pgc') || node.hasAttr('data-pgc-define'))) return 'skip_children';
                return func(node);
            });
        }

        /* node - old element, pgel - new element, c - component definition */
        var copyEditableDataFromOldElementToNew = function(node, pgel, c, restore_optional, status, dont_save_unused) {
            var fields = {};
            var repeated = {};
            var optional_present = {};
            var detached_editable_areas = 0;
            var detached_editable_areas_list = [];
            var detached_editable_areas_html = '';

            status = status || {saved: 0, deleted: []};


            var saved_data_html = getSavedDataForElement(node, true);
            var saved_nodes = [];
            var append_nodes = [];

            if(saved_data_html) {
                var p = new pgParser();
                p.assignIds = false;
                p.parse(saved_data_html);
                for(var i = 0; i < p.rootNode.children.length; i++) {
                    var n = p.rootNode.children[i];
                    var field_id = n.getAttr('data-pgc-field');
                    if(field_id) {
                        var field_info = getFieldInfo(field_id, c);
                        if(field_info && field_info.optional) {
                            continue; //skip optional
                        }
                    }
                    saved_nodes.push(n);
                }
            }

            for(var i = 0; i < saved_nodes.length; i++) {
                node.append(saved_nodes[i]);
            }

            //get data out of standard editable area
            walkSelfAndChildrenButDontGoInComponents(node, function(node) {

                for(var i = 0; i < max_fields; i++) {
                    var key = 'data-pgc-field' + (i == 0 ? '' : '-' + i);
                    var field_id = node.getAttr(key);
                    if(field_id) {
                        var field_info = getFieldInfo(field_id, c);
                        if(field_info) {
                            getEditableData(node, field_info, fields);

                            if(field_info.optional) {
                                optional_present[field_id] = true;
                            }
                        } else {
                            detached_editable_areas_html += node.toStringOriginal(true);
                            detached_editable_areas++;
                            detached_editable_areas_list.push({field_id: field_id, node: node});
                        }

                        //standard repeats
                        var repeat_id = (c.component_repeat_fields && c.component_repeat_fields[field_id]) ? field_id : null;
                        if(repeat_id) {
                            if(!repeated[repeat_id]) repeated[repeat_id] = 0;
                            repeated[repeat_id]++;
                        }
                    }
                }
                return true;
            }, c.master_page_url || false)

            for(var i = 0; i < saved_nodes.length; i++) {
                saved_nodes[i].remove();
            }


            //get data from advanced editable areas
            /*
            for(var i = 0; i < c.editable_areas.length; i++) {
                var elements = node.find(c.editable_areas[i].selector);

                var field_info = c.editable_areas[i];
                var field_id = field_info.name;

                for(var j = 0; j < elements.length; j++) {
                    getEditableData(elements[j], field_info, fields);
                }
            }
            */
            //ok, got data from editable areas

            //advanced repeats
            /*
            for(var i = 0; i < c.repeatable_areas.length; i++) {
                var elements = node.find(c.repeatable_areas[i].selector);
                var repeated_id = c.repeatable_areas[i].id;
                repeated[repeat_id] = elements.length;
            }

            */
            //node is the existing component

            //repeat advanced repeats
            /*
            for(var i = 0; i < c.repeatable_areas.length; i++) {
                var elements = pgel.find(c.repeatable_areas[i].selector);
                var repeated_id = c.repeatable_areas[i].id;
                if(repeated[repeat_id]) {
                    for(var i = elements.length; i < repeated[repeat_id]; i++) {
                        var new_el = elements[elements.length-1].clone();
                        new_el.insertAfter(elements[elements.length-1]);
                    }
                }
            }
            */
            //advanced repeats done

            var remove_optional_elements = [];

            walkSelfAndChildrenButDontGoInComponents(pgel, function(node) {

                var repeat_id = node.getAttr('data-pgc-field');
                if(repeat_id && repeated[repeat_id]) {
                    for(var i = 1; i < repeated[repeat_id]; i++) {
                        var new_el = node.clone();
                        new_el.insertAfter(node);
                    }
                    repeated[repeat_id] = 0;
                }

                //fill standard data
                for(var i = 0; i < max_fields; i++) {
                    var key = 'data-pgc-field' + (i == 0 ? '' : '-' + i);
                    var field_id = node.getAttr(key);
                    if(field_id) {
                        var field_info = getFieldInfo(field_id, c);
                        if(field_info) {
                            fillSavedDataIntoComponent(node, field_info, fields);

                            if(field_info.optional) {
                                if(!optional_present[field_id] && !restore_optional) {
                                    remove_optional_elements.push(node);
                                }
                            }
                        }
                    }
                }
                return true;
            }, c.master_page_url || false)

            for(var i = 0; i < remove_optional_elements.length; i++) {
                remove_optional_elements[i].remove();
            }

            //var unrestored = getUnrestoredFields(fields);

            if(detached_editable_areas && !dont_save_unused) {
                //console.log(unrestored);
                //pinegrow.showQuickMessage(unrestored.length + ' fields were lost');

                var key = '[Pinegrow components saved data - this data is no longer used in the component. We\'re saving it just in case. Clear this with Components -> Clear saved unused edits]';
                //var json = JSON.stringify(unrestored);

                var c = '<!--' + key + detached_editable_areas_html.replace(/\<\!\-\-/g,'[[[--').replace(/\-\-\>/g, '--]]]') + '-->';
                pgel.append(pgCreateNodeFromHtml(c));

                status.saved = detached_editable_areas;
                status.deleted_areas = detached_editable_areas_list;
            }


            //fill advanced data
            /*
            for(var i = 0; i < c.editable_areas.length; i++) {
                var elements = pgel.find(c.editable_areas[i].selector);

                var field_info = c.editable_areas[i];

                for(var j = 0; j < elements.length; j++) {
                    fillSavedDataIntoComponent(elements[j], field_info, fields);
                }
            }
            */
            //end adv data

            return status;
        }

        //returns null or "changed" or "refresh"
        var updateComponent = function(node, c, page, update_tree_list, restore_optional_areas, test, ignore_auto_update, dont_save_unused) {

            var status = {refresh_page: false, changed: false, saved:0, new_pgel: null, deleted_areas: null};

            if(!c.auto_update && !ignore_auto_update) return status;

            var project = pinegrow.getCurrentProject();

            //create new component instance
            var code = pinegrow.getCodeFromDefinition(c);
            var pgel = pgCreateNodeFromHtml(code);
            var old_code = node.toStringOriginal(true);

            node = test ? node.clone(true /* no ids */) : node; /* do test on a clone to prevent losing saved info on the original */

            node.withoutEvents(function() {

                //console.log('Updating COMP ' + c.name + ' - ' + node.getOpeningTag() + '...' + node.getClosingTag());
                if(c.map_urls && page && c.sourceUrl && c.framework && c.framework.project && c.framework.project === project) {
                    //belongs to the same project. Lets update links if diff location
                    walkSelfAndChildrenButDontGoInComponents(pgel, function (node) {
                        node.mapLinksToNewLocation(c.sourceUrl, page.url);
                        return true;
                    })
                }

                copyEditableDataFromOldElementToNew(node, pgel, c, restore_optional_areas, status, dont_save_unused);

                if(page) {
                    page.callFrameworkHandler('on_component_updated', pgel, c);
                }

                //if changed
                var new_code = pgel.toStringOriginal(true);
                if(old_code != new_code) {
                    status.changed = true;

                    if(!test) {
                        node.replaceWith( pgel, true /* detach */ );
                        //update $element
                        var $el = null;
                        var new_pgel = null;

                        if(page && page.openedInEditor) {
                            if(pinegrow.isElementSelected(node)) {
                                select_element_after_update = pgel;
                            }
                            var select_new = pinegrow.isElementOrDescendantSelected(node);

                            status.new_pgel = pgel;

                            if(select_new) {
                                select_element_after_update = pgel;
                            }

                            //status.refresh_page = true;  ????
                        }

                        node.remove();
                    } else {
                        pgel.remove();
                    }
                } else {
                    pgel.remove(); //throw it away
                }
            })
            if(!test && status.changed && page && page.openedInEditor) {
                pgel.update(old_code, node.getId())
            }
            return status;
        }

        var wouldUpdateChangeComponentInstance = function(node, page) {
            var comp_type = node.getAttr('data-pgc');
            if(comp_type) {
                var c = page.getTypeDefinition(comp_type);
                if(c) {
                    return updateComponent(node, c, null, [], false, true /* test */).changed;
                }
            }
            return false;
        }

        var updateComponentsInPage = function(page, project, subtree, restore_optional_areas, return_status, done) {
            var changed = false;
            var refresh_page = false;
            var saved = 0;
            var components_on_the_page = 0;

            select_element_after_update = null;

            if(!restore_optional_areas && pinegrow.getSetting('pgc-ignore-optional', '0') == '1') {
                restore_optional_areas = true;
                pinegrow.showQuickMessage('Optional areas are ignored.');
            }

            var update_tree_list = [];

            return_status = return_status || new PgComponentsUpdateStatus();

            var interactive = return_status.on_before_update || return_status.on_before_update_master || false;


            completion_status = 'done';

            var onDone = function () {
                if (changed) {
                    if (page.openedInEditor) {

                        if (refresh_page) {
                            page.refresh();
                        } else {
                            if (select_element_after_update) {
                                pinegrow.selectElement(select_element_after_update);
                            }
                        }
                    }
                }
                if (done) done(changed, completion_status);
            }

            var doUpdateComponents = function() {

                if (!subtree) subtree = page.sourceNode;

                var node_list = [];

                subtree.walkSelfAndChildren(function (node) {
                    if (node.isElement) {
                        var c = null;
                        var comp_type = node.getAttr('data-pgc');
                        if (!node.hasAttr('data-pgc-define')) {
                            if (comp_type) {
                                if (node.hasAttr('data-pgc-no-update')) return true; //do not update
                                c = page.getTypeDefinition(comp_type);
                            } else {
                                //c = page.getMainType(null, node, true, true);
                            }
                        }
                        if (c) {
                            node_list.push({c: c, node: node, comp_type: comp_type});
                        }
                    }
                    return true;
                });



                var doNode = function (idx) {

                    if (idx < node_list.length) {
                        var node = node_list[idx].node;
                        var c = node_list[idx].c;
                        var comp_type = node_list[idx].comp_type;

                        components_on_the_page++;

                        var do_update = true;

                        var doUpdate = function (dont_save_unused) {
                            var status = updateComponent(node, c, page, update_tree_list, restore_optional_areas, null, null, dont_save_unused);
                            changed = changed || status.changed;
                            saved += status.saved;

                            return_status.saved += saved;

                            if(status.changed) {
                                if(return_status.changed_collection) {
                                    return_status.changed_collection.add(node);
                                }
                            }

                            if (status.refresh_page) {
                                refresh_page = true;
                            }
                        }

                        var test_status = updateComponent(node, c, page, update_tree_list, restore_optional_areas, true);
                        if (!test_status.changed) {
                            do_update = false;
                            if(interactive) {
                                doNode(idx + 1);
                            }
                        } else {
                            if (interactive) {
                                var callback = false;
                                if (return_status.on_before_update) {
                                    callback = true;
                                    return_status.on_before_update(node, c, test_status, function (action) {
                                        switch (action) {
                                            case 'update':
                                            case 'update-delete':
                                                doUpdate(true /* dont save unused */);
                                                break;
                                            case 'update-save':
                                                doUpdate();
                                                break;
                                            case 'skip':

                                                break;
                                            case 'stop':
                                                completion_status = 'stop';
                                                onDone();
                                                return;
                                                break;
                                        }
                                        doNode(idx + 1);
                                    });
                                }
                                if (!callback) {
                                    doUpdate();
                                    doNode(idx + 1);
                                }
                            }
                            else {
                                doUpdate();
                            }
                        }
                    } else {
                        onDone();
                    }
                }

                if (interactive) {
                    doNode(0);
                } else {
                    for (var i = 0; i < node_list.length; i++) {
                        doNode(i);
                    }
                    onDone();
                }
            }

            var doApplyMasterPage = function() {
                if (applyMasterPage(page, project, return_status)) {
                    changed = true;
                    refresh_page = true;
                }
            }

            //update master page
            if(!subtree) {
                if(!interactive) {
                    doApplyMasterPage();
                    doUpdateComponents();
                } else {
                    if(return_status.on_before_update_master) {
                        var test_status = applyMasterPage(page, project, return_status, false, true /* test */);
                        if(test_status) {
                            return_status.on_before_update_master(page, test_status.master_def, test_status, function (action) {
                                switch (action) {
                                    case 'update':
                                        //Update master
                                        doApplyMasterPage();
                                        doUpdateComponents();
                                        break;

                                    case 'skip':
                                        doUpdateComponents();
                                        break;
                                    case 'stop':
                                        completion_status = 'stop';
                                        onDone();
                                        return
                                        break;
                                }
                            });
                        } else {
                            //nothing to update, no mp or master page would not change
                            doUpdateComponents();
                        }
                    } else {
                        doApplyMasterPage();
                        doUpdateComponents();
                    }
                }
            } else {
                doUpdateComponents();
            }

            return changed;
        }

        var updatePageScreenshot = function (page, project, force) {
            if(project) {
                if (!page) {
                    pinegrow.showQuickMessage("You have to open the page in order to update the screenshot");
                    return false;
                }

                var file_path = getImagePathForPage(page, project);
                if (force || !crsaIsFileExist(file_path)) {
                    pinegrow.takePhotoOfPage(page, file_path, function () {
                        pinegrow.showQuickMessage("Screenshot updated successfully");
                    });
                    return true;
                }
                else {
                    pinegrow.showQuickMessage("Screenshot not updated");
                    return false;
                }
            }
        }

        var isMasterPage = function (page) {
            if (!page) return;

            var html = page.sourceNode.findOne('html');
            if(html) {
                if (html.hasAttr('data-pgc-set-master'))
                    return true;
                else
                    return false;
            }
            return false;
        }

        /*

        var getMasterPage = function(url) {
            var page = pinegrow.getCrsaPageForUrl( url );
            return page;
        }
        */

        var getMasterPage = function(page, project, status) {
            var html = page.sourceNode.findOne('html');
            if(html) {
                if(!project.framework) return null;

                var master_pages = project.framework.master_pages;

                var use_master = html.getAttr('data-pgc-master');
                if(use_master) {
                    var key = 'master.' + use_master;
                    if(key in master_pages) {
                        return master_pages[key];
                    } else {
                        if(status) {
                            status.addError('Page <b>' + page.name + '</b>: master page <b>' + use_master + '</b> not found.');
                        }
                    }
                }
            }
            return null;
        }


        var applyMasterPageWithDidChange = function(page, project, return_status, restore_optional_areas, test, skip_refresh) {
            if(applyMasterPage(page, project, return_status, restore_optional_areas, test)) {
                pinegrow.changeManager.didChangePage(page, 'Apply master page');
                if(!skip_refresh) page.refresh();
            }
        }

        var applyMasterPage = function(page, project, return_status, restore_optional_areas, test) {
            var master_def = getMasterPage(page, project, return_status);
            if(master_def) {

                var new_page_source_node = master_def.master_page_source_node.clone();
                var old_source_node = page.sourceNode;

                //f.changeComponentDefinitionsToComponentInstances(new_page_source_node);

                new_page_source_node.setDocument(new_page_source_node);

                var status;

                console.log(`Apply master page ${master_def.master_page_url} to page ${page.url}...`)

                page.sourceNode.withoutEvents(function() {
                    //map links
                    new_page_source_node.walk(function(node) {
                        node.mapLinksToNewLocation(master_def.master_page_url, page.url);
                        return true;
                    })

                    //update editable areas
                    status = copyEditableDataFromOldElementToNew(page.sourceNode, new_page_source_node, master_def, restore_optional_areas);
                })

                if(status) {
                    status.master_def = master_def;
                }

                var old_code = page.sourceNode.toStringOriginal(true);

                if(old_code != new_page_source_node.toStringOriginal(true)) {

                    if(test) {
                        new_page_source_node.withoutEvents(function() {
                            new_page_source_node.remove(); //throw it away
                        })
                        return status;
                    }

                    page.setCode(new_page_source_node, {}, true /* skip didChange */)

                    /*
                    pinegrow.fixLinks(project, page, master_def.master_page_url, page.url, function (badLinksFound) {
                        if (badLinksFound) {
                            pinegrow.showQuickMessage("This page has broken links");
                        }
                    }, false, true);
                    */

                    return_status.changed_collection.add(new_page_source_node);

                    return status; //changed
                }

            }
            return false;
        }


        f.on_project_loaded = function(selectedPage, project) {

            if(pinegrow.getSetting('pgc-auto-reload', '1') == '1') {
                refreshComponents(project);
            } else {
                showReloadComponentsManuallyNotice();
            }

            var pi = pinegrow.getCurrentProjectInfo();
            var settings = pi.getSetting('project-ignore-files') || {};

            project.walkThroughFiles(function (file) {
                updateIgnoreTagOnFile(file, project, settings)
            })

        }

        f.on_project_refreshed = function(selectedPage, project) {

            if(pinegrow.getSetting('pgc-auto-reload', '1') == '1') {
                refreshComponents(project);
            } else {
                showReloadComponentsManuallyNotice();
            }

            var pi = pinegrow.getCurrentProjectInfo();
            var settings = pi.getSetting('project-ignore-files') || {};

            project.walkThroughFiles(function (file) {
                updateIgnoreTagOnFile(file, project, settings)
            })
        }

        f.on_project_closed = function() {
            setHasUpdates(false);
        }

        f.reloadProjectAndUpdateComponents = function() {

            pinegrow.stats.using('components.updateproject');

            setHasUpdates(false);
            requireProject(function(project) {
                requireAtLeastOneOpenPage(project, function(reference_page) {
                    project.reload(function() {
                        project.showProjectExplorer($('#crsa-project'));

                        // TODO: @add-icon file changed icon (fa-asterisk)
                        pinegrow.showNotice('<p><b>Update whole project</b> will go through <b>ALL</b> editable files (HTML, HTM, or file types listed in Editable file types in Support -&gt; Settings) in your project and apply changes to these files.</p><p><b>Changed files will not be saved.</b> Changed files will be marked in the <b>Project view</b> with <b><i class="" style="color:#444;">&#x2731;</i></b> icon.<ul><li><b>Click on file</b> in the Project panel to open the changed file</li><li><b>Right-click</b> on the file and choose <b>Save</b></li><li>Use <b>File -&gt; Save all</b> to save all changed files.</li></ul></p><p><b>Be careful:</b> this action can change lots of your files. It is strongly <b>recommended</b> that you use a <b>source versioning system</b> (like Git) for this project.</p>', 'About updating project', 'comps-update-project', function() {
                            refreshComponents(project, function() {
                                updateComponents(project, null, reference_page);
                            });

                            var pi = pinegrow.getCurrentProjectInfo();
                            var settings = pi.getSetting('project-ignore-files') || {};

                            project.walkThroughFiles(function (file) {
                                updateIgnoreTagOnFile(file, project, settings)
                            })
                        });
                    })
                })
            })
        }



        if(true || canUse()) {
            f.on_key_pressed = function(page, event, status) {

                if(event.which == 85 && event.pgCtrlKey) {
                    if(event.shiftKey) {
                        f.reloadProjectAndUpdateComponents();
                    } else {
                        quickRefreshAndUpdate();
                    }
                    status.done = true;
                }
            }
        }

        var getImagePathForPage = function (page, project) {
            var path = require('path');
            var name = crsaRemoveExtFromUrl(page.name) + ".png";

            var file_path = path.join(project.getDir(), "screenshots");
                file_path = path.join(file_path, name);

            return file_path;
        }

        var PgComponentsUpdateStatus = function() {
            this.saved = 0;
            this.errors = [];
            this.changed_collection = new PgCollection();

            this.addError = function(msg) {
                this.errors.push(msg);
            }

            this.getErrors = function() {
                if(this.errors.length) {
                    return '<ul><li>' + this.errors.join('</li><li>') + '</li></ul>';
                }
                return null;
            }

            this.hasErrors = function() {
                return this.errors.length > 0;
            }

            this.show = function(completion_status) {
                if(!this.hasErrors()) {
                    if(completion_status == 'stop') {
                        pinegrow.showQuickMessage('Components update was stopped.');
                    } else {
                        pinegrow.showQuickMessage('Components refreshed and updated.');
                    }
                } else {
                    pinegrow.showAlert('<p>The following problems happened during the ' + (completion_status == 'stop' ? 'stopped ' : '') + 'update:</p>' + this.getErrors(), 'Oops, the update didn\'t go smoothly');
                }
            }
        }

        var showSavedDataNotice = function(status) {
            if(status.saved) {
                pinegrow.showNotice('<p>Some editable areas were removed from component definitions (or their field ids were changed). To prevent the corresponding edits from being lost, the edits were saved in component instances in the form of special invisible HTML comments.</p><p><b>To show this data again</b> restore editable areas in component definition.</p><p><b>To remove saved data</b> on the selected page use <b>Components -&gt; Clear saved data</b>.</p>', 'Unused editable areas', 'pgc-saved-data', function(showed) {
                    if(!showed) {
                        pinegrow.showQuickMessage(status.saved + ' unused edits were saved in components.', 3000);
                    }
                })
            }
        }

        var quickRefreshAndUpdate = function(done) {

            pinegrow.stats.using('components.quickupdate');

            requireProject(function(project) {
                var pages = pinegrow.getAllPages();
                for(var i = 0; i < pages.length; i++) {
                    if(project.isPageInProject(pages[i])) {
                        quickRefreshComponentsInPage(pages[i], project);
                    }
                }
                var status = new PgComponentsUpdateStatus();

                status.on_before_update_master = status.on_before_update = function(pgel_or_page, comp_def, status, response_func) {
                    var is_master = pgel_or_page instanceof CrsaPage;
                    var pgel = is_master ? pgel_or_page.sourceNode.findOne('html') : pgel_or_page;

                    if(status.deleted_areas) {
                        var str = '<ul>';
                        status.deleted_areas.forEach(function (a) {
                            var field_el = findEditableElementInComponent(pgel, a.field_id);
                            str += field_el ? '<li><b><a href="#" data-pg-select-element="' + field_el.getId() + '">' + a.field_id + '</a></b>' : '<li><b>' + a.field_id + '</b>';
                            var t = a.node.html();
                            if (t.length) {
                                t = crsaGetSummaryStr(escapeHtmlCode(t), 60, true);
                                str += ' | ' + t;
                            }
                            str += '</li>';
                        })
                        str += '</ul>';

                        var onSelect = function (action) {
                            response_func(action);
                        }

                        var title = is_master ? 'Master page and Child page have different editable areas' : 'Some information will be lost after update';

                        var msg = is_master ? '<p>Applying master page <b>' + comp_def.name + '</b> to the child page <b>' + pgel_or_page.name + '</b> will delete the information in some of its editable areas because these editable areas don\'t exist in the master page anymore (did you rename field ids or select a different master page?):</p>' + str + '<p>How would you like to proceed?</p>' : '<p>Updating this <b>' + comp_def.name + '</b> component instance (see selected element) will delete the information in some of its editable areas because these editable areas don\'t exist in the new component definition anymore (did you rename field ids or change component types?):</p>' + str + '<p>How would you like to proceed?</p>';

                        var buttons = null;

                        if(is_master) {
                            buttons = [
                                {
                                    label: 'Do not update',
                                    action: 'skip',
                                    func: onSelect,
                                    tooltip: 'Don\'t update this child page.',
                                    primary: true
                                },
                                {
                                    label: 'Update and discard',
                                    action: 'update',
                                    func: onSelect,
                                    tooltip: 'Update the page and discard the unused information'
                                },
                                {
                                    label: 'Stop update',
                                    action: 'stop',
                                    func: onSelect,
                                    tooltip: 'Stop the process of updating components.'
                                }

                            ]
                        } else {
                            buttons = [
                                {
                                    label: 'Do not update',
                                    action: 'skip',
                                    func: onSelect,
                                    tooltip: 'Don\'t update this component instance.',
                                    primary: true
                                },
                                {
                                    label: 'Update and save',
                                    action: 'update-save',
                                    func: onSelect,
                                    tooltip: 'Deleted areas will be saved as HTML comment inside the instance. The information will be restored if the editable area comes back to the component. Use <b>Components -&gt; Clear saved data</b> to remove saved information.',
                                    primary: true
                                },
                                {
                                    label: 'Update and discard',
                                    action: 'update-delete',
                                    func: onSelect,
                                    tooltip: 'Update the component and discard the unused information'
                                },
                                {
                                    label: 'Stop update',
                                    action: 'stop',
                                    func: onSelect,
                                    tooltip: 'Stop the process of updating components.'
                                }

                            ]
                        }

                        var ask = new PgAskAboutElements([pgel], msg, title, buttons);
                        ask.show();
                    } else {
                        response_func('update');
                    }

                }

                pages = pinegrow.getAllPages().concat([]);

                var completion_status = 'done';

                var doPage = function() {
                    if(pages.length) {
                        quickUpdateComponentsInPage(pages.shift(), project, status, function(changed, status) {
                            if(status == 'stop') {
                                completion_status = 'stop';
                                pages = [];
                            }
                            doPage();
                        });
                    } else {

                        status.show(completion_status);

                        showSavedDataNotice(status);

                        if(done) done(completion_status);
                    }
                }

                doPage();
            })
            setHasUpdates(false);
        }

        if(canUse()) {
            f.on_page_changed = function(page, pgel) {
                if(pgel) {
                    var c = pgel.closest('[data-pgc-define]');

                    if (c || pgel.closest('[data-pgc-set-master]')) {
                        setHasUpdates(true);
                    }
                }
            }

            f.on_duplicate_element_async = function(page, pgel, done) {
                var comp_type;
                if(comp_type = isComponentDefinition(pgel, page)) {
                    var name = getComponentTypeName(comp_type, page);
                    pinegrow.showAlert('<p>This element is the definition of the component <b>' + name + '</b>.</p><p>Would you like to duplicate the component definition - or - make the duplicated element an instance of this component?</p>', 'Duplicating component definition', 'Duplicate definition', 'Create instance', function() {
                        //nothing to do
                        done(pgel);
                        pinegrow.showQuickMessage('Definition duplicated. Change component id if you want to keep both.');
                    }, function() {
                        //change to instance
                        f.changeComponentDefinitionsToComponentInstances(pgel);
                        done(pgel);
                    })
                } else {
                    done(pgel);
                }
            }

            f.on_before_delete_element_async = function(page, pgel, done) {
                var comp_type;
                if(comp_type = isComponentDefinition(pgel, page)) {
                    var name = getComponentTypeName(comp_type, page);
                    pinegrow.showAlert('<p>This element is the definition of the component <b>' + name + '</b>.</p><p>Deleting the component definition will break all instances of this component.</p><p>Would you like to delete it?</p>', 'Deleting component definition', 'Do not delete', 'Delete definition', function() {
                        //nothing to do
                        done(null); //cancel delete
                    }, function() {
                        //delete
                        done(pgel);
                        pinegrow.showQuickMessage('Component definition was deleted.');
                    })
                } else {
                    done(pgel);
                }
            }
        }

        var getComponentTypeName = function(comp_type, page) {
            var c = page.getTypeDefinition(comp_type);
            return c ? c.name : comp_type;
        }

        var isComponentDefinition = function(pgel, page) {
            //just quick check
            return pgel.getAttr('data-pgc-define');
            if(comp_type) {
                if(!page) page = pinegrow.getCrsaPageOfPgParserNode(pgel);
                return page.getTypeDefinition(comp_type);
            }
            return null;
        }

        var restoreOptionalAreasOnMasterPage = function(page, project) {
            var return_status = new PgComponentsUpdateStatus();
            if(applyMasterPage(page, project, return_status, true /* restore optional areas */)) {
                pinegrow.changeManager.didChangePage(page, 'Restore optional areas');
                page.refresh();
            } else {
                //page.undoStack.remove();
            }
            pinegrow.showQuickMessage('Optional areas restored');
        }


        f.on_page_menu = function(page, items) {
            var project = pinegrow.getCurrentProject();
            if(project && project.isPageInProject(page) && getMasterPage(page, project)) {
                items.push({
                    type: 'divider'
                    //header: 'Components'
                })

                items.push({
                    label: 'Restore optional areas',
                    action: function() {
                        restoreOptionalAreasOnMasterPage(page, project);
                    }
                })
            }
        }

        var goToDefinition = function(c) {
            pinegrow.stats.using('components.gotodefinition');

            if(c.sourceUrl) {
                pinegrow.openOrShowPage(c.sourceUrl, function(def_page) {
                    pinegrow.setSelectedPage(def_page);
                    var defel = def_page.sourceNode.findOneWithAttrValue('data-pgc-define', c.type);
                    if(defel) {
                        pinegrow.selectElement(defel, true);
                        defel.scrollTo();
                    }
                }, false, true);
            } else {
                pinegrow.showQuickMessage('Sorry, don\'t know where the component is defined.', 4000, false, 'error');
            }
        }

        var duplicateDefinition = function(c, insert_after) {
            f.requireSelectedElement(function(page, pgel) {

                var duplicateIt = function(new_id) {
                    var code = c.definition_code;
                    var new_el = pgCreate(code);
                    new_el.attr('data-pgc-define', new_id);

                    if(canMakeChange(insert_after ? pgel.parent : pgel, 'insert_element')) {
                        pinegrow.makeChanges(pgel, 'Duplicate definition', function() {
                            if(insert_after) {
                                new_el.insertAfter(pgel);
                            } else {
                                pgel.append(new_el);
                            }
                        })
                    }
                    pinegrow.selectElement(new_el, true);
                    pinegrow.showQuickMessage('Component definition duplicated.');
                    setHasUpdates(true);
                }

                pinegrow.showPrompt('<p>Enter the new component\'s id:</p>', 'Choose the unique id for the component', c.type, '', null, function(val) {
                    var existing = page.getTypeDefinition(val);
                    if(existing) {
                        pinegrow.showAlert('Component definition with the id <b>' + val + '</b> already exists. Are you sure you want to use this id?', 'The id is not unique', 'Cancel', 'Use it!', function() {
                            //cancel
                        }, function() {
                            //use it
                            duplicateIt(val);
                        })
                    } else {
                        duplicateIt(val);
                    }
                })
            })
        }

        var makeInstanceOf = function(c) {
            f.requireSelectedElement(function(page, pgel) {

                //pgel is pgQuery

                var select = new PgCollection();
                var col = pinegrow.getCollectionForElement(pgel);

                pinegrow.makeChanges(col, "Make instance of", function() {

                    col.forEachElement(function(pgel) {
                        if(pgel) {
                            var cel = pgCreateNodeFromHtml(c.code);

                            var fields = cel.findIncludingSelf('[data-pgc-field]');
                            for(var i = 0; i < fields.length; i++) {

                                var field_id = fields[i].getAttr('data-pgc-field');
                                var sel = pgFindUniqueSelectorForElement(fields[i], cel, {tag: true, path: true});

                                if(sel) {
                                    var els = pgel.find(sel);
                                    //console.log('Selector = ' + sel + ' - found ' + els.length);
                                    for(var j = 0; j < els.length; j++ ) {
                                        els[j].setAttr('data-pgc-field', field_id);
                                        //console.log(els[j].getOpeningTag());
                                    }
                                }
                            }

                            pgel.setAttr('data-pgc', c.type);

                            pgel.setAttr('data-pgc-dummy', 'hello'); //will force update because new code will differ

                            var update_tree_list = [];

                            var status = updateComponent(pgel, c, page, update_tree_list, false);

                            if(status.new_pgel) {
                                select.add(status.new_pgel);
                                //pinegrow.selectElement(status.new_pgel);
                            } else {
                                select.add(pgel)
                            }
                        }
                    })
                })

                pinegrow.selectedElements.selectMany(select, true)
            })
        }

        var isProjectComponent = function(def) {
            return def.framework.project ? true : false;
        }

        if(canUse()) {
            f.on_build_lib_actions_menu = function(page, actions, def) {

                if(isProjectComponent(def) && !isContributorModeOrTest()) {
                   actions.push(
                        {label: "Make selected an instance of <b>" + def.name + "</b>", class: 'action-make-instance-of', kbd: null, manage_change: true, action: function() {
                            makeInstanceOf(def);
                        }});
                    actions.push(
                        {label: "Go to definition", class: 'action-go-to-definition', kbd: null, manage_change: true, action: function() {
                            goToDefinition(def);
                        }});

                    actions.push(
                        {label: "Duplicate definition...", class: 'duplicate-definition', kbd: null, manage_change: true, action: function() {
                            duplicateDefinition(def, true);
                        }});
                }
            }

            f.on_project_item_get_context_menu = function(page, file, project) {
                var items = [];

                if (!resourceEditor || (resourceEditor && !resourceEditor.getProject())) {
                    var project = pinegrow.getCurrentProject();
                    resourceEditor = new PgResourceEditor(project);
                }

                items.push({
                    divider: true,
                    header: 'Components'
                })


                var addRequireFile = function(global) {
                    f.requireSelectedElement(function(cp, pgel) { //todo30

                        if(pgel.hasAttr('data-pgc-define')) {
                            var url = cp.makeRelativeUrl(file.url);

                            var code = '<pgcparam data-pgc-require="' + url + '"';
                            if(global) {
                                code += ' data-pgc-require-global="true"';
                            }
                            code += '></pgcparam>';
                            var param = (new pgQuery()).create(code);

                            pinegrow.makeChanges(pgel, 'Insert component parameter', function() {
                                pgel.prepend(param);
                            })
                            pinegrow.showQuickMessage('Parameter added.');
                        } else {
                            pinegrow.showAlert('<p><b>A component definition must be selected.</b></p><p>Select a component definition and then set which files and folders are required by that component or by the whole library.</p>', 'This only works on component definitions');
                        }
                    });
                }

                items.push({
                    label: 'Add to resources...',
                    func: function() {
                        //addRequireFile(false);
                        // pinegrow.showQuickMessage('Not yet supported. Use <b>Required by all</b>.', 5000);
                        resourceEditor.requireFileFromUrl(file.url);
                    }
                })

                /*
                items.push({
                    label: 'Required ' + (file.isFile ? 'file' : 'folder') + ' by all: ' + file.name,
                    func: function() {
                        addRequireFile(true);
                    }
                })
    */
                if (isMasterPage(page)) {
                    items.push({
                        label: 'Update screenshot',
                        func: function () {
                            var project = pinegrow.getCurrentProject();
                            updatePageScreenshot(page, project, true);
                            pinegrow.cache_breaker++;
                        }
                    })
                }

                return items;
            }
        }

        if(canUse()) {
            f.on_build_actions_menu = function(page, menus, pgel, for_all_selected) {

                var target_pgel = for_all_selected ? null : pgel;

                var getCompDef = function(pgel) {
                    var comp_type = pgel.getAttr('data-pgc') || pgel.getAttr('data-pgc-define');
                    if (comp_type) {
                        return pgel.getPage().getTypeDefinition(comp_type);
                    }
                    return null;
                }

                if(!isContributorModeOrTest()) {
                    var project = pinegrow.getCurrentProject();
                    if (pgel.tagName == 'html' && project && project.isPageInProject(page) && getMasterPage(page, project)) {
                        menus.push({
                            type: 'divider'
                        })
                        menus.push({
                            type: 'header',
                            label: 'Master page'
                        })

                        menus.push({
                            label: 'Restore optional areas',
                            manage_change: true,
                            action: function () {
                                restoreOptionalAreasOnMasterPage(page, project);
                            }
                        })
                    }

                    var comp_type = pgel.getAttr('data-pgc') || pgel.getAttr('data-pgc-define');
                    var c = getCompDef(pgel);

                    if (c) {
                        if (c.framework.project) {

                            menus.push({
                                type: 'divider'
                            })
                            menus.push({
                                type: 'header',
                                label: 'Components'
                            })

                            if (pgel.getAttr('data-pgc') && (!c.auto_update || pgel.hasAttr('data-pgc-no-update'))) {
                                menus.push(
                                    {
                                        label: "Update component instance",
                                        class: 'action-update-component',
                                        kbd: null,
                                        manage_change: true,
                                        action: function () {
                                            var update_tree_list = [];
                                            var col = pinegrow.getCollection(target_pgel);
                                            var changed_col = new PgCollection();

                                            col.forEachElement(function(pgel) {
                                                var c = getCompDef(pgel);
                                                if(c && c.framework.project && updateComponent(pgel, c, pgel.getPage(), update_tree_list, false, false, true /* force update */)) {
                                                    changed_col.add(pgel);
                                                }
                                            })
                                            if(changed_col.getLength()) {
                                                pinegrow.changeManager.didChange(changed_col, 'Component update');
                                                pinegrow.showQuickMessage('Component instance updated!');
                                            } else {
                                                pinegrow.showQuickMessage('Component was not changed.');
                                            }
                                        }
                                    });
                            }

                            menus.push(
                                {
                                    label: "Go to definition",
                                    class: 'action-go-to-definition',
                                    kbd: null,
                                    manage_change: true,
                                    action: function () {
                                        goToDefinition(c);
                                    }
                                });

                            menus.push(
                                {
                                    label: "Duplicate definition...",
                                    class: 'duplicate-definition',
                                    kbd: null,
                                    manage_change: true,
                                    action: function () {
                                        duplicateDefinition(c, true);
                                    }
                                });

                            menus.push(
                                {
                                    label: "Restore optional areas",
                                    class: 'action-restore-optional',
                                    kbd: null,
                                    manage_change: true,
                                    action: function () {
                                        var update_tree_list = [];
                                        var col = pinegrow.getCollection(target_pgel);
                                        var changed_col = new PgCollection();

                                        col.forEachElement(function(pgel) {
                                            var c = getCompDef(pgel);
                                            if(c && c.framework.project && updateComponentsInPage(pgel.getPage(), c.framework.project, pgel, true)) {
                                                changed_col.add(pgel);
                                            }
                                        })
                                        if(changed_col.getLength()) {
                                            pinegrow.changeManager.didChange(changed_col, 'Restore optional area');
                                            pinegrow.showQuickMessage('Optional areas were restored!');
                                        } else {
                                            pinegrow.showQuickMessage('Component was not changed.');
                                        }
                                    }
                                });

                            if(isApp()) {
                                menus.push(
                                    {
                                        label: "Take component photo",
                                        class: 'action-take-photo',
                                        kbd: null,
                                        manage_change: true,
                                        action: function () {
                                            var preview_dir = path.join(c.framework.project.getDir(), 'previews');
                                            var filename = path.join(preview_dir, comp_type + '.png');
                                            crsaCreateDirs(preview_dir, fs);

                                            var $el = pgel.get$DOMElement();

                                            //some links have no dimension
                                            if ($el.outerWidth() == 0 && $el.children().length) {
                                                $el = $($el.children().get(0));
                                            }

                                            pinegrow.takePhotoOfElement($el, 400.0, filename, function () {
                                                pinegrow.showQuickMessage('&lt;project&gt;/previews/<b>' + comp_type + '.png</b> saved.');
                                                pinegrow.cache_breaker++;

                                                var gui = require('nw.gui');
                                                gui.App.clearCache();

                                                c.preview_image = comp_type + '.png';
                                                c.button_image = c.preview_image;

                                                if (c.definition_attributes && 'data-pgc-define-photo-preview-only' in c.definition_attributes) {
                                                    c.button_image = null;
                                                }

                                                $('body').trigger('crsa-update-lib-display');
                                                //pinegrow.frameworksChanged();
                                            });
                                        }
                                    });
                            }
                        }
                    }
                }
            }
        }

        pinegrow.loadProjectAsLibrary = function(folder, done, done_activated) {
            //Create empty framework
            var f_key = isApp() ? path.basename(folder) : crsaGetNameFromUrl(folder);

            var pf = pinegrow.findFrameworkByKey(f_key);

            var folder_url = crsaMakeUrlFromFile(folder);

            if(pf) {
                if(pf.pluginUrl == folder_url) {
                    //same project already loaded, use it
                } else {
                    pinegrow.showAlert('<p>Library <b>' + f_key + '</b> is already loaded.</p><p>Note that the library folder name is used as the name of the library. This name must be unique - only one library with a certain name can be loaded.</p>', f_key + ' is already loaded');
                    if(done) done();
                    return;
                }
            }

            if(!pf) pf = new PgFramework( f_key, f_key);

            pf.pluginUrl = folder_url;

            pinegrow.addFramework(pf, -1000);
            pinegrow.frameworksChanged();

            pf.components_loaded = pf.components_loaded || false;

            pf.on_init_plugin = function(init_done_func) {
                //Load it when first activated
                if(!pf.components_loaded) {
                    var project;

                    pinegrow.stats.using('components.loadprojectaslibrary');

                    var onProjectLoaded = function(project) {

                        refreshComponents( project, function(project, framework) {
                            if(done_activated) done_activated(project, framework);
                            if(init_done_func) init_done_func(framework);
                        })
                    }

                    var current_project = pinegrow.getCurrentProject();

                    if(current_project && current_project.getDir() == folder) {
                        project = current_project;
                        project.framework = pf;
                        onProjectLoaded(project);
                    } else {
                        project = new CrsaProject();
                        project.framework = pf;

                        project.fromFolder(folder, function(project) {
                            onProjectLoaded(project);
                        }, true);
                    }

                    pf.components_loaded = true;
                } else {
                    if(init_done_func) init_done_func(pf);
                }
            }

            pf.on_plugin_activated = function(page, init_done_func) {
                pf.on_init_plugin(init_done_func);
            }

            var flist = pinegrow.getFrameworksListFromStorage();
            if(flist.indexOf(folder_url) < 0) {
                flist.push(folder_url);
                crsaStorage.setValue('frameworks', flist);
            }

            if(done) done(pf);
            return pf;
        }

        pinegrow.selectAndLoadLibrary = function(done) {
            crsaChooseFile(function(url, file) {
                pinegrow.loadProjectAsLibrary(file, function(f) {

                    pinegrow.activateFrameworkInCurrentContext(f);
                    if(done) done(f);
                })
            }, false, false, null, true);
        }

        var findEditableElementInComponent = function(pgel, field_id, index) {
            index = index || 0;
            var r = findAllEditableElementsInComponent(pgel, field_id);
            return index < r.length ? r[index] : null;
        }

        var findAllEditableElementsInComponent = function(pgel, field_id) {
            var r = [];
            walkSelfAndChildrenButDontGoInComponents(pgel, function(node) {
                if(node.isElement) {
                    var editable = node.getAttr('data-pgc-edit');
                    if(editable) {
                        var info = pgcVarParseValue(editable);
                        if(info && info.name == field_id) {
                            r.push(node);
                        }
                    } else if(node.getAttr('data-pgc-field') == field_id) {
                        r.push(node);
                    }
                }
                return true;
            })
            return r;
        }

        var test_contributor_mode = false;

        var setTestContributorMode = function(val) {
            test_contributor_mode = val;
            pinegrow.reselectElement();
            pinegrow.repaintTree();
        }

        var isContributorModeOrTest = function() {
            if(test_contributor_mode) return true;
            return pinegrow.isContributorMode();
        }

        f.on_show_properties = function(page, sections, pgel, defs, showPropertiesView) {
            var locked_desc = null;
            var locked_reason = null;
            if(locked_reason = isElementInLockedArea(pgel)) {
                if(f.on_can_make_change(page, pgel, 'edit') === null) {
                    //is in editable field with editable content
                    locked_desc = 'The element is fully editable';
                } else {
                    sections.splice(0, sections.length);
                    locked_desc = locked_reason == 'cc_mode' ? 'The element is not editable.' : 'The element is not editable because it is in the locked area.';
                }
            }
            var new_sections = [];

            if(defs) {
                for(var i = 0; i < defs.length; i++) {
                    if(defs[i].component_fields) {

                        var c = defs[i];
                        var s = {
                            framework : c.framework,
                            name : c.name,
                            fields : {},
                            component_fields: c.component_fields,
                            default_closed: false
                        }
                        new_sections.push(s);

                        if(locked_desc) {
                            locked_desc = 'Only certain properties of this element are editable.';
                        }
                    }
                }
            }
            if(!new_sections.length) {
                //look for plain editable area without component
                var field_info = getEditableAreaInfoOfElement(pgel, isContributorModeOrTest() || true /* include definition */, page /* include master page fields */);

                if(!field_info) {
                    var ret_obj = {
                        pgel: null,
                        info: null,
                        closest: false
                    }

                    pinegrow.dispatchEvent('on_cms_get_dynamic_field', page, pgel, ret_obj);

                    if(ret_obj.pgel && ret_obj.info) {
                        field_info = ret_obj.info;
                    }
                }

                //how about master page fields?

                if(field_info) {
                    var cf = {};
                    cf[field_info.name] = field_info;

                    var s = {
                        framework : f,
                        name : 'Editable properties',
                        fields : {},
                        component_fields: cf,
                        default_closed: false
                    }
                    new_sections.push(s);

                    if(locked_reason) {
                        locked_desc = locked_reason == 'cc_mode' ? 'Only certain properties of this element are editable.' : 'Only certain properties of this element are editable because it is in the locked area.';
                    }
                }
            }
            if(new_sections.length) {
                for(var i = 0; i < new_sections.length; i++) {
                    if(new_sections[i].component_fields) {

                        var has_props = false;

                        var s = new_sections[i];
                        $.each(s.component_fields, function(key, field) {

                            var all = findAllEditableElementsInComponent(pgel, field.name);

                            $.each(all, function(index, el) {
                                var index_str = all.length == 1 ? '' : (' ' + (index+1));
                                var index_key = '.' + (index+1);

                                if(field.content) {
                                    if(field.types && field.types.length) {
                                        var fdef = {};
                                        fdef.name = field.name + index_str + ' content';
                                        fdef.type = 'custom';
                                        fdef.control = function() {
                                            //Create control, needs a unique id
                                            var c = new PgCustomPropertyControl('pg.components.field.control.' + field.name + index_key + '.content');

                                            //Register subfields.
                                            c.onDefine = function() {
                                                //no subfields
                                            }

                                            //Show control. Return the control $el.
                                            c.onShow = function() {
                                                var $d = $('<div class="crsa-field"><label class="">Add <nocap>to</nocap> ' + field.name + ':</label></div>');
                                                field.types.forEach(function(t) {
                                                    var tname = getComponentTypeName(t, page);
                                                    var $b = $('<button class="btn btn-primary btn-sm"><i class="icon icon-plus"></i> ' + tname + '</button>').on('click', function(e) {
                                                        e.preventDefault();
                                                        var cdef = page.getTypeDefinition(t);
                                                        var pgel = c.obj;
                                                        var fel = findEditableElementInComponent(pgel, field.name, index);
                                                        if(fel) {
                                                            var code = getCodeFromDefinition(cdef);
                                                            var api = new PgApi();
                                                            var inserted_ids = api.insertElements(code, fel, 'append', {
                                                                scroll: true,
                                                                highlight: true
                                                            });
                                                            if (inserted_ids.getLength()) {
                                                                pinegrow.showQuickMessage(tname + ' was appended to ' + field.name);
                                                            }
                                                        }
                                                    })
                                                    $d.append($b);
                                                })
                                                return $d;
                                            }

                                            //Called when control is recycled and new values are set
                                            c.onSetValues = function() {
                                            }

                                            c.onHasData = function() {
                                                return false;
                                            }

                                            //Return the control
                                            return c;
                                        };
                                        /*
                                        fdef.show = function($c, obj, fn, fdef, values, $scrollParent) {
                                            var $d = $('<div class="crsa-field"><label class="full-width">Add <nocap>to</nocap> ' + field.name + ':</label></div>');
                                            field.types.forEach(function(t) {
                                                var tname = getComponentTypeName(t, page);
                                                var $b = $('<button class="btn btn-primary btn-sm"><i class="fa fa-plus"></i> ' + tname + '</button>').on('click', function(e) {
                                                    e.preventDefault();
                                                    var cdef = page.getTypeDefinition(t);
                                                    var pgel = getElementPgNode(obj.data);
                                                    var fel = findEditableElementInComponent(pgel, field.name, index);
                                                    if(fel) {
                                                        var code = getCodeFromDefinition(cdef);
                                                        var api = new PgApi();
                                                        var inserted_ids = api.insert(code, fel.getId(), 'append', {
                                                            scroll: true,
                                                            highlight: true
                                                        });
                                                        if (inserted_ids.length) {
                                                            pinegrow.showQuickMessage(tname + ' was appended to ' + field.name);
                                                        }
                                                    }
                                                })
                                                $d.append($b);
                                            })
                                            $d.appendTo($c);
                                        }*/
                                    } else {
                                        var fdef = {};
                                        fdef.name = field.name + index_str + ' content';
                                        fdef.type = 'text';
                                        fdef.action = 'custom';
                                        fdef.get_value = function (pgel) {
                                            var fel = findEditableElementInComponent(pgel, field.name, index);
                                            return fel ? fel.html() : null;
                                        }
                                        fdef.set_value = function (pgel, value, values, oldValue, eventType) {
                                            pinegrow.willChangeDom();
                                            var fel = findEditableElementInComponent(pgel, field.name, index);
                                            if (fel) {
                                                fel.html(value || '');
                                            }
                                            return value;
                                        }
                                    }
                                    fdef.on_can_make_change = function() { return true; };

                                    s.fields['pg.components.field.' + field.name + index_key + '.content'] = fdef;
                                    has_props = true;
                                }
                                for(var ai = 0; ai < field.attrs.length; ai++) {
                                    var attr = field.attrs[ai];

                                    if(attr == 'wp-include-template-part-set') continue;

                                    (function(attr) {

                                        var fdef = {};
                                        fdef.name = field.name + index_str + ' ' + attr;
                                        fdef.type = 'text';
                                        fdef.action = 'custom';
                                        fdef.options = function(fdef, pgel) {
                                            var fel = findEditableElementInComponent(pgel, field.name, index);
                                            return pinegrow.insight.getValuesForAttribute(attr, ['img','script','link','iframe'].indexOf(fel.tagName) >= 0 ? fel.tagName : null);
                                        },
                                        fdef.get_value = function(pgel) {
                                            var fel = findEditableElementInComponent(pgel, field.name, index);
                                            return fel ? fel.getAttr(attr) : null;
                                        }
                                        fdef.set_value = function(pgel, value, values, oldValue, eventType) {
                                            var fel = findEditableElementInComponent(pgel, field.name, index);
                                            if (fel) {
                                                fel.setAttr(attr, value);
                                            }
                                            return value;
                                        }
                                        fdef.on_can_make_change = function() { return true; };

                                        if(attr == 'src' || attr == 'href') {
                                            fdef.file_picker = true;
                                        }

                                        s.fields['pg.components.field.' + field.name + index_key + '.' + attr] = fdef;
                                    })(attr);
                                    has_props = true;
                                }
                                var classes = [];
                                for(var ci = 0; ci < field.classes.length; ci++) {
                                    var clsre = getClassRegEx(field.classes[ci]);
                                    if(!clsre) {
                                        classes.push({key: field.classes[ci], name: field.classes[ci]});
                                    }
                                }

                                if(classes.length) {
                                    var fdef = {};
                                    fdef.name = field.name + index_str + ' classes';
                                    fdef.type = 'select';
                                    fdef.options = classes;
                                    fdef.show_empty = true;
                                    fdef.action = 'custom';
                                    fdef.multiple = classes.length > 1;
                                    fdef.get_value = function(pgel) {
                                        //debugger;
                                        var fel = findEditableElementInComponent(pgel, field.name, index);
                                        var r = [];
                                        if(fel) {
                                            for(var ii = 0; ii < classes.length; ii++) {
                                                if(fel.hasClass(classes[ii].key)) {
                                                    r.push(classes[ii].key);
                                                }
                                            }
                                            if(r.length) return r.join(',');
                                        }
                                        return null;
                                    }
                                    fdef.set_value = function(pgel, value, values, oldValue, eventType) {
                                        var fel = findEditableElementInComponent(pgel, field.name, index);
                                        if(fel) {
                                            for(var ii = 0; ii < classes.length; ii++) {
                                                if(fel.hasClass(classes[ii].key)) {
                                                    fel.removeClass(classes[ii].key);
                                                }
                                            }
                                            if(value) {
                                                var list = value.split(',');
                                                list.forEach(function(cls) {
                                                    fel.addClass(cls);
                                                })
                                            }
                                        }
                                        return value;
                                    }
                                    fdef.on_can_make_change = function() { return true; };

                                    s.fields['pg.components.field.' + field.name + index_key + '.classes'] = fdef;
                                    has_props = true;
                                }

                                //bckimage
                                if(field.bckimage) {
                                    var fdef = {};
                                    fdef.name = field.name + index_str + ' Bck. Image';
                                    fdef.type = 'text';
                                    fdef.action = 'custom';

                                    fdef.get_value = function(pgel) {
                                        var fel = findEditableElementInComponent(pgel, field.name, index);
                                        return fel ? fel.getInlineStylePropertyValue('background-image', true) : null;
                                    }
                                    fdef.set_value = function(pgel, value, values, oldValue, eventType) {
                                        var fel = findEditableElementInComponent(pgel, field.name, index);
                                        if (fel) {
                                            fel.setInlineStylePropertyValue('background-image', value, true);
                                        }
                                        return value;
                                    }
                                    fdef.on_can_make_change = function() { return true; };
                                    fdef.file_picker = true;
                                    fdef.file_picker_no_proxy = true;

                                    s.fields['pg.components.field.' + field.name + index_key + '.bckimage'] = fdef;
                                    has_props = true;
                                }

                                //extensions
                                $.each(field.extensions, function(k, v) {
                                    var ext = pgcvar.pgcGetExtensionForFieldKey(k);
                                    if(ext && ext.getProperty) {
                                        var fdef = ext.getProperty(k, v, findEditableElementInComponent(pgel, field.name, index));
                                        fdef.name = field.name + index_str + ' / ' + fdef.name;
                                        fdef.on_can_make_change = function() { return true; };

                                        s.fields['pg.components.field.' + field.name + index_key + '.' + k] = fdef;
                                        has_props = true;
                                    }
                                })
                            })
                        })
                        if(has_props) sections.unshift(s);
                    }
                }
            }
            if(locked_desc /*&& isContributorModeOrTest()*/) {
                showPropertiesView.showMessage((test_contributor_mode ? 'CMS mode is ON.<br><br>' : '') + locked_desc);
            }
        }

        f.on_set_inline_style = function(page, o) {
            o.css += '\
            html.crsa-pgc-mark-areas [data-pgc-define], html.crsa-pgc-mark-areas [data-pgc-save-partial] {\
            outline: 1px solid rgba(0, 236, 255, 1);\
            outline-offset:0;\
            }\
            html.crsa-pgc-mark-areas [data-pgc] {\
            outline: 1px solid rgba(255, 214, 0, 1);\
            outline-offset:0;\
            }\
            html.crsa-pgc-mark-areas [data-pgc-field], html.crsa-pgc-mark-areas [data-pgc-edit] {\
            outline: 1px solid rgba(255, 214, 0, 0.3);\
            outline-offset:0;\
            }\
            html.crsa-pgc-mark-areas[data-pgc-set-master] [data-pgc-edit] {\
            outline: 1px solid rgba(255, 214, 0, 1);\
            outline-offset:0;\
            }\
            ';
        }

        var $pgcmenu = $('<li class="dropdown pgc-menu"><a href="#" class="dropdown-toggle" data-toggle="dropdown">Components</a>\
                        <ul class="dropdown-menu with-checkboxes">\
                            <li><a href="#" class="pgc-menu-quick-update">Quick update (only open pages)</a></li>\
                            <li><a href="#" class="pgc-menu-update-project">Update the whole project</a></li>\
                            <li><a href="#" class="pgc-menu-reload">Reload components from project</a></li>\
                            <li class="divider"></li>\
                            <li class="dropdown-header">Libraries</li>\
                            <li><a href="#" class="pgc-menu-edit-resources">Edit resources...</a></li>\
                            <li><a href="#" class="pgc-menu-refresh-libs">Refresh loaded libraries</a></li>\
                            <li><a href="#" class="pgc-menu-resources">Add/update resources from libraries...</a></li>\
                            <li class="divider"></li>\
                            <li class="dropdown-header">Tools</li>\
                            <li><a href="#" class="pgc-menu-auto-reload"><i class="icon icon-check check-icon"></i>Auto reload on project refresh</a></li>\
                            <li><a href="#" class="pgc-menu-mark"><i class="icon icon-check check-icon"></i>Mark components</a></li>\
                            <li><a href="#" class="pgc-menu-test-cm"><i class="icon icon-check check-icon"></i>Test CMS mode</a></li>\
                        <!--<li><a href="#" class="pgc-menu-ignore-optional"><i class="icon icon-check check-icon"></i>Ignore optional areas on update</a></li>-->\
                            <li><a href="#" class="pgc-menu-clear-saved-edits">Clear unused edits...</a></li>\
                            <li><a href="#" class="pgc-menu-remove">Remove component information...</a></li>\
                            <li><a href="#" class="pgc-menu-help">Documentation</a></li>\
            </ul></li>');

        crsaAddKbd($pgcmenu.find('.pgc-menu-quick-update'), 'CMD U');
        crsaAddKbd($pgcmenu.find('.pgc-menu-update-project'), 'SHIFT CMD U');

        $pgcmenu.find('.pgc-menu-quick-update')
            .on('click', function(e) {
                e.preventDefault();
                quickRefreshAndUpdate();
            });

        $pgcmenu.find('.pgc-menu-update-project')
            .on('click', function(e) {
                e.preventDefault();
                f.reloadProjectAndUpdateComponents();
            });

        $pgcmenu.find('.pgc-menu-reload')
            .on('click', function(e) {
                e.preventDefault();
                requireProject(function(project) {
                    //setHasUpdates(false);
                    refreshComponents(project, function() {
                        //updateComponents(project);
                    });
                })
            });


        $pgcmenu.find('.pgc-menu-resources')
            .on('click', function(e) {
                e.preventDefault();
                requireProject(function(project) {
                    pinegrow.showUpdateResourcesDialog(pinegrow.getSelectedPage(), project);
                })
            });

        $pgcmenu.find('.pgc-menu-edit-resources')
            .on('click', function(e) {
                e.preventDefault();
                requireProject(function(project) {
                    showResourceEditorForProject(project);
                })
            });



        $pgcmenu.find('.pgc-menu-refresh-libs')
            .on('click', function(e) {
                e.preventDefault();
                var list = [];

                var current_project = pinegrow.getCurrentProject();

                $.each(pinegrow.getFrameworks(), function(key, f) {
                    if(f.project && f.components_loaded) {
                        if(f.project != current_project) {
                            list.push(f);
                        }
                    }
                });

                var doOne = function() {
                    if(list.length) {
                        var f = list.shift();
                        f.project.reload(function(project) {
                            refreshComponents(project, function() {
                                doOne();
                            }, true /* skip frameworks changed */);
                        })
                    } else {
                        pinegrow.frameworksChanged();
                    }
                }
                doOne();
            });

        $pgcmenu.find('.pgc-menu-help')
            .on('click', function(e) {
                e.preventDefault();
                pinegrow.openExternalUrl(pinegrow.pro_info_url);
            });

        $pgcmenu.find('.pgc-menu-apply-master')
            .on('click', function(e) {
                e.preventDefault();
                var project = pinegrow.getCurrentProject();
                var page = pinegrow.getSelectedPage();
                if(project && page) {
                    applyMasterPageWithDidChange(page, project);
                }
            });

        $pgcmenu.find('.pgc-menu-clear-saved-edits')
            .on('click', function(e) {
                e.preventDefault();
                requireProject(function(project) {

                    var $d = pinegrow.showAlert('Would you like to clear saved unused edits in <b>the selected page</b> or from <b>the whole project</b>?', 'Where to clear?', 'Cancel', 'Selected page only', null, function() {
                        var page = pinegrow.getSelectedPage();
                        if(page) {
                            clearSavedDataFromPage(page);
                            pinegrow.showQuickMessage('Saved data cleared!');
                        }
                    })
                    var $b = $('<button class="btn btn-primary btn-sm">Whole project</button>');
                    $d.find('.modal-footer').append($b);
                    $b.on('click', function(e) {
                        e.preventDefault();
                        $d.modal('hide');

                        project.forEachEditableFile( function( page, pageDone, status ) {
                            //on page
                            if(clearSavedDataFromPage(page)) {
                                status.changed = true;
                            }
                            pageDone(page, status);

                        }, function() {
                            pinegrow.showQuickMessage('Saved data cleared!');
                        }, 'Clearing saved unused edits...', null, getIgnoreMap(project))

                    });
                });
            });

        $pgcmenu.find('.pgc-menu-remove')
            .on('click', function(e) {
                e.preventDefault();

                var where = pinegrow.selectedElements.hasSelection() ? 'selected elements' : 'page';
                var $d = pinegrow.showAlert('Would you like to remove all information about components and editable areas from the <b>' + where + '</b>? All <code>data-pgc-*</code> attributes will be removed.', 'Are you sure', 'Cancel', 'Remove', null, function() {
                    var page = pinegrow.getSelectedPage();
                    if(page) {
                        pinegrow.makeChanges(page, 'Remove component info', function() {
                            if(pinegrow.selectedElements.hasSelection()) {
                                pinegrow.selectedElements.forEachElement(function(pgel) {
                                    pgel.removeAttrIfStartsWith('data-pgc', true);
                                })
                            } else {
                                page.sourceNode.findOne('html').removeAttrIfStartsWith('data-pgc', true);
                            }
                        })
                        pinegrow.showQuickMessage('Component information removed!');
                        pinegrow.refreshCurrentProject(null, false, true);
                        page.refresh();
                    }
                })
            });

        var $mark_areas = $pgcmenu.find('.pgc-menu-mark');

        var markAreasHelper = new PgPageViewHelper('Mark components', 'pgc-mark-areas', 'crsa-pgc-mark-areas', $mark_areas);

        //auto reload
        var $auto_reload = $pgcmenu.find('.pgc-menu-auto-reload');
        var $auto_reload_checkbox = $auto_reload.find('i');

        var set_auto_reload = function(value) {
            if(value) {
                $auto_reload_checkbox.show();
            } else {
                $auto_reload_checkbox.hide();
            }
        }

        var showReloadComponentsManuallyNotice = function() {
            pinegrow.showQuickMessage('Use <b>Components -&gt; Reload components from project</b> to reload components.', 4000);
        }

        set_auto_reload(pinegrow.getSetting('pgc-auto-reload', '1') == '1');

        $auto_reload
            .on('click', function(e) {
                e.preventDefault();
                var value = pinegrow.getSetting('pgc-auto-reload', '1') == '1';
                value = !value;
                set_auto_reload(value);
                pinegrow.setSetting('pgc-auto-reload', value ? '1' : '0');

                if(!value) {
                    showReloadComponentsManuallyNotice();
                }
            });
        //end reload



        //test contributor mode
        var $test_cm = $pgcmenu.find('.pgc-menu-test-cm');
        var $test_cm_checkbox = $test_cm.find('i');

        var set_test_cm = function(value, show_msg) {
            if(value) {
                $test_cm_checkbox.show();
                if(show_msg) {
                    pinegrow.showNotice('<p><b>Test CMS mode is now active.</b> To continue editing the page, switch it off in the Components menu.</p><p>In <b>CMS mode</b> only elements marked with Editable area actions can be edited, the rest of the page is locked.</p><p>Learn more about <a href="https://pinegrow.com/docs/pinegrow-pro/cms/" class="external">using Pinegrow as CMS</a>.</p>', 'About the CMS mode', 'pgc-content-contributor', function(shown) {
                        if(!shown) {
                            pinegrow.showQuickMessage('CMS test mode is ON. Only editable areas can be edited.');
                        }
                    })
                }
            } else {
                if(show_msg) {
                    pinegrow.showQuickMessage('CMS test mode is OFF.');
                }
                $test_cm_checkbox.hide();
            }
        }

        set_test_cm(test_contributor_mode);

        $test_cm
            .on('click', function(e) {
                e.preventDefault();
                var value = test_contributor_mode;
                value = !value;
                set_test_cm(value, true);
                setTestContributorMode(value);
            });
        //end reload


/*
        //ignore optional on update
        var $ignore_optional = $pgcmenu.find('.pgc-menu-ignore-optional');
        var $ignore_optional_checkbox = $ignore_optional.find('i');

        var set_ignore_optional = function(value) {
            if(value) {
                $ignore_optional_checkbox.show();
            } else {
                $ignore_optional_checkbox.hide();
            }
        }

        set_ignore_optional(pinegrow.getSetting('pgc-ignore-optional', '0') == '1');

        $ignore_optional
            .on('click', function(e) {
                e.preventDefault();
                var value = pinegrow.getSetting('pgc-ignore-optional', '0') == '1';
                value = !value;
                set_ignore_optional(value);
                pinegrow.setSetting('pgc-ignore-optional', value ? '1' : '0');

                if(value) {
                    pinegrow.showQuickMessage('Optional areas will be ignored on update.', 4000);
                }
            });
        //end ignore optional on update
        */

        if((true || pinegrow.isPRO() || pinegrow.isProTrialActive()) && !pinegrow.isContributorMode()) {
            pinegrow.addPluginControlToTopbar(f, $pgcmenu, true);
        }

        var $updateButton = $('<span class="update-components"><i class="icon icon-REFRESH"></i></span>')
            .on('click', function(e) {
                e.preventDefault();
                e.stopPropagation();
                quickRefreshAndUpdate();
            })

        addTooltip($updateButton, 'Quick update components ' + pgShowKbdCombo('CMD+U'));

        //pinegrow.addPluginControlToTopbar(f, $updateButton);

        $updateButton.appendTo($pgcmenu.find('>a'));

        var setHasUpdates = function(has) {
            if(isContributorModeOrTest()) has = false;
            if(has) {
                $updateButton.removeClass('update-components-hide');
            } else {
                $updateButton.addClass('update-components-hide');
            }
        }

        f.setHasUpdates = setHasUpdates;

        setHasUpdates(false);


        f.useMasterPage = function (page, project, master_page_file, skip_refresh) {
            var html = page.sourceNode.findOne('html');
            html.withoutEvents(function() {
                html.removeAttr('data-pgc-set-master');
                html.setAttr('data-pgc-master', master_page_file.name);
            })
            var return_status = new PgComponentsUpdateStatus();
            applyMasterPageWithDidChange(page, project, return_status, null, null, skip_refresh);
        }

        f.updateEditableAreas = function (page) {
            f.changeComponentDefinitionsToComponentInstances(page.sourceNode);
        }

        f.on_page_loaded = function(crsaPage) {
            //set_mark_areas(pinegrow.getSetting('pgc-mark-areas', '1') == '1');
        }

        f.on_page_refreshed = function(crsaPage) {
            //set_mark_areas(pinegrow.getSetting('pgc-mark-areas', '1') == '1');
        }

        f.on_page_saved = function(crsaPage, saved_csss, first_save) {
            savePartials(crsaPage);

            //perhaps this is a open partial?
            updateSavedPartialsOnOpenPages(crsaPage);
        }

        var savePartials = function(crsaPage, list, force) {
            var path = require('path');
            var partials = list || crsaPage.sourceNode.find('[data-pgc-save-partial]');

            if(partials.length) {
                var pi = pinegrow.getCurrentProjectInfo();
                var partials_map = pi.getSetting('project-partials') || {};

                var page_rel_url = pi.makeUrlRelative(crsaPage.url);

                for (var i = 0; i < partials.length; i++) {
                    var filename = partials[i].getAttr('data-pgc-save-partial');

                    var op = getPartialOperation(partials[i]);
                    if(op.indexOf('save') < 0 && !force) continue;

                    try {
                        var c = partials[i].clone(true /* no ids */);
                        c.removeAttrIfStartsWith('data-pgc-save-partial', true);
                        var html;
                        if (partials[i].hasAttr('data-pgc-save-partial-content')) {
                            html = c.toStringContent(true, pinegrow.getFormatHtmlOptions());
                        } else {
                            html = c.toStringOriginal(true, pinegrow.getFormatHtmlOptions());
                        }
                        filename = path.resolve(path.dirname(crsaPage.localFile || ''), filename);

                        if(filename === crsaPage.localFile) {
                            throw `Partial can't overwrite its parent file.`;
                        }

                        var is_new = !crsaIsFileExist(filename);

                        crsaWriteFileWithBackup(null, filename, html, "utf8");
                        pinegrow.showQuickMessage('Partial <b>' + crsaGetSummaryStr(filename, 50) + '</b> saved!');
                        is_new && pinegrow.refreshCurrentProjectDelayed(true);

                        var file_url = crsaMakeUrlFromFile(filename);
                        partials_map[pi.makeUrlRelative(file_url)] = page_rel_url;

                    } catch (err) {
                        pinegrow.showAlert('<p>Partial file ' + filename + ' could not be saved: ' + err + '</p>', 'Error saving partial file');
                    }
                }

                pi.setSetting('project-partials', partials_map);
                pi.save();
            }
        }


        f.pgcGetPartialOpeningOptions = function(page_url, pi, name) {
            pi = pi || pinegrow.getCurrentProjectInfo();
            var partials_map = pi.getSetting('project-partials') || {};

            var file_url = pi.makeUrlRelative(page_url);
            if(partials_map[file_url]) {

                return [
                    {
                        label: 'As partial',
                        action: function () {
                            var path = require('path');
                            var found = false;
                            var url = pi.makeUrlAbsolute(partials_map[file_url]);
                            pinegrow.openOrShowPage(url, function (page) {
                                var partials = page.sourceNode.find('[data-pgc-save-partial]');
                                for (var i = 0; i < partials.length; i++) {
                                    var filename = partials[i].getAttr('data-pgc-save-partial');

                                    if(!filename) continue;

                                    filename = path.resolve(path.dirname(crsaMakeFileFromUrl(url)), filename);
                                    var p_url = crsaMakeUrlFromFile(filename);
                                    if (p_url === page_url) {
                                        pinegrow.selectElement(partials[i]);
                                        partials[i].scrollTo();
                                        pinegrow.showQuickMessage(`Partial ${name} is selected.`);
                                        found = true;
                                        break;
                                    }
                                }
                                if (!found) {
                                    pinegrow.showQuickError(`Partial ${name} was not found in page ${page.name}`);
                                }
                            });
                        }
                    },
                    {
                        label: 'Directly',
                        action: function () {
                            pinegrow.openOrShowPage(page_url);
                        }
                    },
                    {
                        label: 'Edit code',
                        action: function () {
                            pinegrow.openFileInCodeEditor(crsaMakeFileFromUrl(page_url));
                        }
                    }
                ]
            }
            return null;
        }

        var loadPartials = function(page, list, with_events, force, filter_func, stat_obj) {
            stat_obj = stat_obj || {};

            stat_obj.changed = false;
            stat_obj.col = new PgCollection();

            var path = require('path');
            var fs = require('fs');

            var partials = list || page.sourceNode.find('[data-pgc-save-partial]');

            var errors = [];

            if(!crsaIsFileUrl(page.url)) {
                return;
            }
            var page_file = crsaMakeFileFromUrl(page.url);

            var doTask = function(p) {
                for(var i = 0; i < partials.length; i++) {

                    var op = getPartialOperation(partials[i]);

                    if(op.indexOf('load') < 0 && !force) continue;

                    var filename = partials[i].getAttr('data-pgc-save-partial');

                    if(!filename) continue;

                    filename = path.resolve(path.dirname(page_file), filename);

                    if(filter_func && !filter_func(partials[i], filename)) continue;

                    if(filename === page.localFile) {
                        errors.push(`Partial <code>${filename}</code> can't be loaded from its parent file.`)
                        continue;
                    }

                    try {
                        var html = fs.readFileSync(filename, {encoding: 'utf8'});
                    } catch (err) {
                        errors.push(`Can't read the file <code>${filename}</code>: ${err}.`);
                        continue;
                    }
                    if (partials[i].hasAttr('data-pgc-save-partial-content')) {
                        partials[i].html(html);
                    } else {
                        try {
                            var pgel = pgCreate(html.trim(), true);

                            if(pgel.script_info) {
                                errors.push(`The top element should not be a ${pgel.script_info.name} tag. If you need to do this, define a Partial that uses only the inner content.`);
                                continue;
                            }

                            pgel.removeAttrIfStartsWith('data-pgc-save-partial', false);

                            partials[i].getAttrList().forEach(function (a) {
                                pgel.setAttr(a.name, a.value);
                            })

                            partials[i].replaceWith(pgel);

                        } catch (err) {
                            errors.push(`Invalid content in file <code>${filename}</code>. Only one top HTML element is allowed.`)
                        }
                    }
                    stat_obj.changed = true;
                    stat_obj.col.add(partials[i]);
                }
            }

            if(with_events) {
                doTask();
            } else {
                page.sourceNode.withoutEvents(function () {
                    doTask();
                })
            }

            if(errors.length) {
                pinegrow.showAlert(`<p>Could not load all partials in <b>${page.name}</b>:</p><ul><li>${errors.join('</li><li>')}</li></ul>`, `Error(s) while loading partials`);
                return false;
            } else {
                return true;
            }
        }

        pinegrow.addEventHandler('on_page_parsed_in_proxy', function(page, sourceNode, url, is_browser_preview) {
            if(page.openedInEditor && !is_browser_preview) {
                //doing loadPartials with browser preview changes the document. Should not happen.
                loadPartials(page);
            }
        })

        var updateSavedPartialsOnOpenPages = function(codePage) {
            var file = codePage.localFile;

            if(file) {
                var has = false;

                pinegrow.getAllPages().forEach(function (page) {
                    var stat = {};

                    loadPartials(page, null, true, false, function(pgel, filename) {
                        return file === filename;
                    }, stat);

                    if(stat.changed) {
                        pinegrow.changeManager.didChange(stat.col, 'Load partials on ' + page.name);
                        has = true;
                    }
                })
                if(has) {
                    pinegrow.showQuickMessage('Saved partial was updated on open pages.');
                }
            }
        }

        pinegrow.addEventHandler('on_code_page_saved', function(codePage) {
            updateSavedPartialsOnOpenPages(codePage);
        })

        /*
        pinegrow.addEventHandler('on_got_source_node_of_url_from_file', function(pagenull, sourceNode, url) {

        })
        */

        //Ignore folder and files

        var shouldIgnoreFile = function(file, project, settings) {
            var url = project.getRelativeUrl(file.url);
            return settings[url] || false;
        }

        var getIgnoreMap = function(project) {
            var pi = pinegrow.getCurrentProjectInfo();
            return pi.getSetting('project-ignore-files') || {};
        }

        var updateIgnoreTagOnFile = function(file, project, settings) {
            if(!settings) {
                project = pinegrow.getCurrentProject();
                var pi = pinegrow.getCurrentProjectInfo();
                settings = pi.getSetting('project-ignore-files') || {};
            }
            var tag = file.getTagByType('ignorescan');

            var ignore = shouldIgnoreFile(file, project, settings);

            var icon;
            var text;
            var tip;
            var cls;

            if(ignore) {
                icon = 'hide';
                text = tip = `Don't scan for components &amp; master pages.`;
                //tip = 'Will not be scanned.';
                cls = 'pg-file-tag-white';
            } else {
                icon = 'check';
                text = tip = `Scan for components &amp; master pages.`;
                //tip = 'Will be scanned.'
                cls = 'pg-file-tag-gray'
            }

            if(ignore) {
                if (!tag) {
                    tag = new CrsaFileTag(text, 'ignorescan', tip, icon, cls);
                    file.addTag(tag);
                    file.updateTags();
                } else {
                    /*
                    tag.text = text;

                    tag.icon = icon;
                    tag.desc = tip;
                    tag.class = cls;
                    */
                }

            } else {
                if(tag) {
                    file.removeTagByType('ignorescan');
                    file.updateTags();
                }
            }
        }

        pinegrow.addEventHandler('on_project_item_get_context_menu_regular', function(page, file, project, menu_list) {

            var pi = pinegrow.getCurrentProjectInfo();

            var settings = pi.getSetting('project-ignore-files') || {};

            var items = [];

            items.push({
                divider: true,
                header: 'Components'
            })

            var rel_url = project.getRelativeUrl(file.url);
            var ignore = settings[rel_url] || false;

            var setCopy = function(val) {
                if(!val) {
                    delete settings[rel_url];
                    pinegrow.showQuickMessage(`<b>${file.name}</b> will be scanned for components and master pages. Reload project to apply the change.`)
                } else {
                    settings[rel_url] = true;
                    pinegrow.showQuickMessage(`<b>${file.name}</b> will <b>not</b> be scanned for components and master pages. Reload project to apply the change.`)
                }

                pi.setSetting('project-ignore-files', settings)
                pi.save();

                updateIgnoreTagOnFile(file, project, settings)
            }

            items.push({
                label: 'Scan for components &amp; master pages',
                check: !shouldIgnoreFile(file, project, settings),
                submenu: [
                    {
                        label: 'Yes',
                        check: !ignore,
                        action: function() {
                            setCopy(false)
                        }
                    },
                    {
                        label: 'No',
                        check: ignore,
                        action: function() {
                            setCopy(true)
                        }
                    }
                ]
            })
            menu_list.push(items);

            //return items;
        })

        //End ignore folder

        //Only editable areas
        var getComponentInstance = function(pgel, page, extras) {
            var cel = pgel.closest('[data-pgc]');
            if(cel) {
                if(extras) extras.element = cel;
                return page.getTypeDefinition(cel.getAttr('data-pgc'));
            }
        }

        var getComponentInstanceOrDefinition = function(pgel, page, extras) {
            var cel = pgel.closest('[data-pgc],[data-pgc-define]');
            if(cel) {
                if(extras) extras.element = cel;
                return page.getTypeDefinition(cel.getAttr('data-pgc') || cel.getAttr('data-pgc-define'));
            }
        }



        var describeEditableArea = function(i, page) {
            if(!i) return '';
            var s = '<p>The following parts of <b>' + i.name + '</b> are editable:</p><ul>';
            if(i.content) {
                if(i.types && i.types.length) {
                    var names = i.types.map(function(t) {
                        var c = page.getTypeDefinition(t);
                        return '<b>' + (c ? c.name : t) + '</b>';
                    })
                    s += '<li>Component instances of ' + names.join(', ') + ' can be added and deleted from the content</li>';
                } else {
                    s += '<li>The content</li>';
                }
            }
            if(i.attrs.length) {
                s += '<li>Attributes: <b>' + i.attrs.join('</b>, <b>') + '</b></li>';
            }
            if(i.classes.length) {
                s += '<li>Classes: <b>' + i.classes.join('</b>, <b>') + '</b></li>';
            }
            if(i.bckimage) {
                s += '<li>Background image</li>';
            }
            s += '</ul>';
            return s;
        }

        var isElementInLockedArea = function(pgel) {
            if(isContributorModeOrTest()) return "cc_mode";
            return pgel.closest('[data-pgc-lock]') ? "locked" : false;
            return false;
        }

        f.on_custom_tree_filter = function(page, obj) {

            if(!isContributorModeOrTest()) return;

            var list = ['[data-pgc],[data-pgc-define],[data-pgc-edit],[data-pgc-field]'];

            pinegrow.dispatchEvent('on_cms_get_selectors_for_dynamic_fields', page, list);

            try {
                page.get$Html().find( list.join(',') ).each(function(i, e) {
                    var treeid = e.getAttribute('data-pg-id');
                    if(treeid) {
                        obj.selected_ids[treeid] = true;
                        if(!obj.changed) obj.changed = true;

                        //check if editable content area

                        var field_el = getDomElementPgNode(e);
                        var info = getEditableAreaInfoOfElement(field_el, true);
                        if(info && info.content) {
                            var sel = '*';
                            if(info.types && info.types.length) {
                                sel = info.types.map(function(t) {
                                    return '[data-pgc="' + t + '"],[data-pgc-define="' + t + '"]';
                                }).join(',');
                            }
                            $(e).find(sel).each(function(idx, el) {
                                var treeid = el.getAttribute('data-pg-id');
                                if(treeid) {
                                    obj.selected_ids[treeid] = true;
                                }
                            })
                        }
                    }
                })
            } catch(err) {
                debugger;
            }
            /////
        }

        f.on_can_make_change = function(page, pgel, action, data) {
            //another framework already prevented it
            if(page.getCurrentFrameworkHandlerReturnValue('on_can_make_change')) return page.getCurrentFrameworkHandlerReturnValue('on_can_make_change');

            var lock_non_editable = isElementInLockedArea(pgel);

            var can_edit = false;
            var error_msg = '';

            var project = pinegrow.getCurrentProject();
            var extras = {};
            var field_info = null;
            var c = lock_non_editable === 'cc_mode'  ? getComponentInstanceOrDefinition(pgel, page, extras) : getComponentInstance(pgel, page, extras); //mtch

            //prevent comp def deletes in cc mode
            if(isContributorModeOrTest() && isComponentDefinition(pgel) && action == 'delete_element' && c) {
                return new PgEditException('<p>The element can\'t be deleted because it is the definition of the <b>' + c.name + '</b> component.</p><p>Deleting the component definition would break all instances of this component.</p>', 'This element can\'t be deleted');
            }

            var parent_actions = ['delete_element', 'duplicate_element'];

            if(c && extras.element == pgel && parent_actions.indexOf(action) >= 0) {
                //this concerns the parent
                c = lock_non_editable ? getComponentInstanceOrDefinition(pgel.parent, page, extras) : getComponentInstance(pgel.parent, page, extras); //mtch
            }

            if (project && action == "add_action" && data.attribute == "data-pgc-set-master")
                updatePageScreenshot(page, project);

            var canEditField = function(field_info, field_el) {
                var is_field = pgel == field_el;

                //classes
                if(is_field && (action == 'add_class' || action == 'remove_class')) {
                    //check if class attr is editable, first
                    if(field_info.attrs.indexOf('class') >= 0) {
                        can_edit = true;
                    } else {
                        for(var i = 0; i < field_info.classes.length; i++) {
                            var clsre = getClassRegEx(field_info.classes[i]);
                            if(clsre) {
                                if(data.match(clsre)) {
                                    can_edit = true;
                                    break;
                                }
                            } else {
                                if(data == field_info.classes[i]) {
                                    can_edit = true;
                                    break;
                                }
                            }
                        }
                    }
                }

                //attributes
                if(is_field && action == 'attr') {
                    if(field_info.attrs.indexOf(data) >= 0) {
                        can_edit = true;
                    } else {
                        error_msg = field_info.attrs.length == 1 ? 'Only the attribute <b>' + field_info.attrs[0] + '</b> is editable.' : 'Only attributes <b>' + field_info.attrs.join(', ') + '</b> are editable.';
                    }
                }

                var isAllowedType = function(pgel) {
                    var types = page.getAllTypes(null, pgel, true /* skip actions */);
                    for(var i = 0; i < types.length; i++) {
                        if(field_info.types.indexOf(types[i].type) >= 0) {
                            return true;
                        }
                    }
                    return false;
                }

                //content
                if(field_info.content) {
                    if(is_field) {
                        if(field_info.types && field_info.types.length) {
                            //if only specific types are allowed
                            if(action == 'insert_element' && data && data.inserted && isAllowedType(data.inserted)) {
                                can_edit = true;
                            }
                        } else {
                            var allowed = ['insert_element', 'edit_content' /*, 'edit_code' */];
                            if (allowed.indexOf(action) < 0) {
                                error_msg = 'Content of the element can\'t be changed.';
                            } else {
                                can_edit = true;
                            }
                        }
                    } else {
                        //if types specified, only that types can be deleted from top content
                        if(field_info.types && pgel.parent == field_el) {
                            var allowed = ['delete_element', 'duplicate_element'];
                            if(allowed.indexOf(action) >= 0 && isAllowedType(pgel)) {
                                can_edit = true;
                            }
                        } else {
                            can_edit = true; //everything should be ok inside the field
                        }
                    }
                }

                //bckimage
                if(field_info.bckimage && is_field && action == 'element_bckimage') {
                    can_edit = true;
                }

                if(field_info.repeat && is_field) {
                    if(action == 'duplicate_element' || action == 'delete_element' || action == 'move_element') {
                        can_edit = true;
                    }
                }

                if(field_info.optional && is_field) {
                    if(action == 'delete_element') {
                        can_edit = true;
                    }
                }

                //is this reordening of repeatable element?
                if(action == 'insert_element' && data && data.moved) {
                    var movedel = data.moved;
                    var moved_info = getEditableAreaInfoOfElement(movedel);
                    if(moved_info && moved_info.repeat) {
                        //are next or prev same field type?
                        var isType = function(type, el) {
                            if(el) {
                                var i = getEditableAreaInfoOfElement(el);
                                return i && i.name == type;
                            }
                            return false;
                        }

                        if(isType(moved_info.name, data.prev) || isType(moved_info.name, data.next)) {
                            can_edit = true;
                        }
                    }
                }

                //extensions
                $.each(field_info.extensions, function(k, v) {
                    var ext = pgcvar.pgcGetExtensionForFieldKey(k);
                    if(ext) {
                        if(ext.canMakeChange(k, pgel, action, data)) {
                            can_edit = true;
                        }
                    }
                })
            }

            if(!c) {
                //maybe a master page?
                if(project) {
                    c = getMasterPage(page, project);
                }
            }
            if(c) {

                var comp_el = extras.element || null;
                var field_el = lock_non_editable ? pgel.closest('[data-pgc-field],[data-pgc-edit]', comp_el) : pgel.closest('[data-pgc-field]', comp_el);
                var is_comp = pgel == comp_el;

                if(!c.auto_update) {
                    can_edit = true;
                } else if(comp_el && comp_el.hasAttr('data-pgc-no-update')) {
                    can_edit = true;
                } else if(field_el) {
                    var field_id = field_el.getAttr('data-pgc-field');
                    if(!field_id && lock_non_editable) {
                        var info = pgcVarParseValue(field_el.getAttr('data-pgc-edit'));
                        if(info) field_id = info.name;
                    }
                    field_info = (field_id && c.component_fields) ? (c.component_fields[field_id] || null) : null;
/*
                    edit_code
                    insert_element
                    move_element
                    delete_element
                    duplicate_element
                    remove_tags
                    edit_content
                    remove_class
                    add_class
                    add_action
                    remove_action
                    edit
                    attr
   */
                    if(field_info) {
                        canEditField(field_info, field_el);
                    } else {
                        can_edit = true;
                        //pinegrow.showQuickMessage('Warning: Field <b>' + field_id + '</b> is not defined in component <b>' + c.name + '</b>', 5000);
                    }
                }
                if(!can_edit && is_comp) {
                    can_edit = ['delete_element', 'duplicate_element', 'move_element'].indexOf(action) >= 0;
                }


                if(!can_edit && action == 'edit_code') {
                    can_edit = true;
                    //pinegrow.showQuickMessage('Be carefull - you\'re editing a component instance!', 3000);
                }
                if(!can_edit && pgel.tagName == 'html') {
                    if(action == 'add_action' || action == 'remove_action') {
                        if(data == pgcmaster || data == pgcusemaster) {
                            can_edit = true;
                        }
                    } else if(action == 'attr' && data.startsWith('data-pgc-master')) {
                        can_edit = true;
                    }
                }
                if(!can_edit) {
                    if(is_comp && (action == 'add_action' || action == 'remove_action')) {
                        if(data == pgc || data == pgcinstance) {
                            can_edit = true;
                        }
                    }
                }
                if(!can_edit && is_comp && action == 'attr' && (data == 'data-pgc-no-update' || data == 'data-pgc')) {
                    can_edit = true;
                }


                if(!can_edit) {
                    if(!isContributorModeOrTest()) {
                        return new PgEditException('<p>The element can\'t be changed because it is defined in component <b>' + c.name + '</b>.</p>' + describeEditableArea(field_info, page) + '<p>Choose "Go to component definition" to edit the component definition.</p>', 'This element is locked', function ($dialog) {
                            var $b = $('<button class="btn btn-default btn-sm" type="button">Go to component definition</button>').on('click', function (e) {
                                e.preventDefault();
                                goToDefinition(c);
                                $dialog.modal('hide');
                            }).insertBefore($dialog.find('.btn.ok'));
                        });
                    } else {
                        return new PgEditException('<p>This element is not editable.</p>' + describeEditableArea(field_info, page), 'This element is locked');
                    }
                }
            } else {
                //not component, not master page
                if(lock_non_editable && !(action == 'remove_action' && data == pgclock)) {
                    var field_el = pgel.closest('[data-pgc-edit]');

                    var info = null;

                    if(field_el) {
                        info = getEditableAreaInfoOfElement(field_el, true);
                    }

                    if(!field_el) {
                        var ret_obj = {
                            pgel: null,
                            info: null,
                            closest: true
                        }

                        pinegrow.dispatchEvent('on_cms_get_dynamic_field', page, pgel, ret_obj);

                        if(ret_obj.pgel && ret_obj.info) {
                            field_el = ret_obj.pgel;
                            info = ret_obj.info;
                        }
                    }

                    if(info) {

                        canEditField(info, field_el);

                        if(!can_edit) {
                            return new PgEditException('<p>This element is not editable.</p>' + describeEditableArea(info, page), 'This element is locked');
                        }
                        return null;
                    }
                }
            }
            //
            if(lock_non_editable && !field_info && !can_edit) {
                if(action == 'remove_action' && data == pgclock) {
                    //ok
                } else {
                    return new PgEditException((test_contributor_mode ? '<p><b>Test CMS mode is ON</b> (you can switch it off in the Components menu).</p>' : '') + '<p>The element is not editable.</p><p>Only elements marked with <i class="icon icon-Editable" style="color:#6cadd6;"></i> in the tree are editable.</p>', 'This element is locked');
                }
            }

            //

            return null;
        }

        var getEditableAreaInfoOfElement = function(field_el, include_definition, child_page) {
            var info = null;
            var field_id = field_el.getAttr('data-pgc-field');
            if(field_id) {
                //look for component
                var extras = {};
                var c = getComponentInstance(field_el, field_el.getPage(), extras);
                if(c) {
                    var info = getFieldInfo(field_id, c);
                    if(info) return info;
                }
            }
            if(include_definition) {
                info = pgcVarParseValue(field_el.getAttr('data-pgc-edit'));
                if(!info.name) info = null;
                if(info) {
                    var cls = field_el.getAttr('data-pgc-edit-classes');
                    info.classes = crsaSplitAndTrimList(cls);
                    info.bckimage = field_el.hasAttr('data-pgc-edit-bckimage');

                    var types = field_el.getAttr('data-pgc-edit-types');
                    if(types) {
                        info.types = crsaSplitAndTrimList(types);
                        if(info.types.length) info.content = true;
                    }
                    info.extensions = {};
                    field_el.findAttributesStartingWith('data-pgc-edit-ext-').forEach(function(attr) {
                        var key = attr.name.replace('data-pgc-edit-ext-', '');
                        info.extensions[key] = attr.value;
                    })

                }
            }
            if(field_id && child_page) {
                var project = pinegrow.getCurrentProject();
                if(project && project.isPageInProject(child_page)) {
                    var master = getMasterPage(child_page, project);
                    if(master) {
                        return getFieldInfo(field_id, master);
                    }
                }
            }


            return info;
        }

        var requireProject = function(func) {
            if(!canUse()) {
                pinegrow.showProductUpgradeMessage('PRO', 'Pinegrow PRO', 'Your edition of Pinegrow does not include Pinegrow PRO features. Please upgrade to the PRO edition to continue using projects, master pages and components.')
                return;
            }
            var p = pinegrow.getCurrentProject();
            if(p) {
                func(p);
            } else {
                pinegrow.showAlert('<p>Components work with <b>Projects</b>. To use components open the folder where your pages are located with <b>File -&gt; Open project</b>.</p>', 'Components work with Projects');
            }
        }

        /*
        f.on_is_element_locked = function(page, pgel) {
            return true;
        }
*/
        //test
        //pinegrow.loadProjectAsLibrary('/Users/Matjaz/Dropbox/Development/DVWeb/DV/Pinegrow/lab/ComponentsDemo');

    });


});
