var PgSpecialChars = function() {

    var chars = [
        {
            name: 'Space',
            chars: [' ','&nbsp;']
        },
        {
            name: 'Quotes',
            chars: ['&quot;|"','&apos;|\'','&ldquo;|“','&rdquo;|”','&bdquo;|„','&lsquo;|‘','&rsquo;|’','&sbquo;|‚','&laquo;|«','&raquo;|»', '&lsaquo;|‹', '&rsaquo;|›', '&lang;|⟨', '&rang;|⟩']
        },
        {
            name: 'Arrows',
            chars: ['&uarr;|↑','&darr;|↓','&larr;|←','&rarr;|→','&harr;|↔','&larr;|⇐','&rarr;|⇒','&harr;|⇔']
        },
        {
            name: 'Symbols',
            chars: ['&ndash;|–', '&mdash;|—', '&hellip;|…', '&amp;', '&copy;|©','&reg;|®','&bigstar;|★','&euro;|€','&cent;|¢','&pound;|£','&yen;|¥','&para;|¶','&permil;|‰']
        },
        {
            name: 'Bullets',
            chars: ['&bull;|•','&middot;|·','&sdot;|⋅']
        },
        {
            name: 'Emojis',
            chars: ['&#128522;|😊', '&#128512;|😀', '&#128513;|😁', '&#128514;|😂', '&#128521;|😉', '&#128528;|😐', '&#128533;|😕', '&#127794;|🌲']
        }
    ]


    var processed = false;

    var replace_map = {}
    var replace_re = null;

    this.getChars = function() {
        if(!processed) {
            processed = true;

            var e = [];

            chars.forEach(function(d) {
                d.chars = d.chars.map(function(i) {
                    if(i.indexOf('|') >= 0) {
                        var a = i.split('|');
                        replace_map[a[1]] = a[0];
                        e.push(escapeRegExp(a[1]));
                        return a[0];
                    } else {
                        return i;
                    }
                })
            })

            var ret = `(${e.join('|')})`;
            //console.log(ret);
            replace_re = new RegExp(ret, 'g');
        }
        return chars;
    }

    this.replaceCharsWithEntities = function(str) {
        if(!processed) {
            this.getChars();
        }
        return str.replace(replace_re, function(a, b) {
            return replace_map[a];
        })
    }

}

var pgSpecialChars = new PgSpecialChars();


var PgQuickTextEdit = function(pgel) {

    var ac = new PgTextEditView();

    var win = new PgQuickWindow('Edit text for ' + pgel.getName({short: true}), ac, 'quickTextEdit', 640, 320);

    win.showForPgel(pgel);

    ac.show(pgel);

    win.onDestroy = function() {
    }

}



var PgTextEditView = function() {

    var $element = $('<div class="pg-text-edit"></div>');

    var $editor = $('<div class="pg-text-edit-editor"></div>').appendTo($element);

    var view = new PgUIView($element);

    this.view = view;

    var _this = this;

    var current_pgel;
    var page;

    this.show = function(pgel) {
        page = pgel.getPage();
        current_pgel = pgel;
        $editor.html(page.getViewInnerHTMLForElement(pgel, true /* disable js */));

        $editor.editor({
            iframe: true,
            toolbarButtons: ['undo', 'redo' , '|', 'bold', 'italic', 'underline', 'strikeThrough', 'subscript', 'superscript', 'outdent', 'indent', 'clearFormatting', 'insertTable', 'html', 'insertImage', 'insertLink']
        })
    }



    var onElementSelected = function(page, pgel, info) {
    }

    var onPageChanged = function(page, pgel, info) {
        if(pgel == current_pgel && (!info.source || info.source != _this)) {
        }
    }

    view.onResize = function() {
    }

    pinegrow.addEventHandler('on_element_selected', onElementSelected);
    pinegrow.addEventHandler('on_page_changed', onPageChanged);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_element_selected', onElementSelected);
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        //$editor.editor('destroy');
    }
}




var PgInlineEditorManager = function() {

    var _this = this;

    var currentEditor = null;
    var currentPageView = null;
    var currentPage = null;
    var currentPgel = null;
    var editor_id = 0;

    var textMode = false;

    this.domHtmlToSource = function(html, ret_nodes, assign_ids) {
        if(typeof assign_ids == 'undefined') assign_ids = false;
        var p = new pgParser();
        p.assignIds = assign_ids;
        p.replaceExistingIds = true;
        p.parse(html);
        p.rootNode.find('.pinegrow-ui-helper').forEach(function(pgel) {
            pgel.remove();
        })
        p.rootNode.walkSelfAndChildren(function(node) {
            if(node.isElement) {
                node.removeAttribute('data-pg-id');
                node.removeAttribute('data-pg-tree-id');
                node.removeClass('crsa-selected');
                node.removeAttrIfStartsWith('data-pg-temp-id-');

                if(node.getAttribute('target') == '_pg_blank') {
                    node.removeAttribute('target');
                }

                if(node.getAttribute('type') == 'pinegrow/disabled') {
                    node.removeAttribute('type');
                }

                if(node.hasAttribute('data-pg-inline-edit-remove')) {
                    node.remove();
                    return true;
                }

                var list = node.getAttrList();
                for(var i = 0; i < list.length; i++) {
                    if(list[i].name.startsWith('pg-disabled-')) {
                        list[i].name = list[i].name.replace(/^pg\-disabled\-/, '');
                        node.attrListChanged = true;
                    }
                }

                var href = null;
                var attr = null;
                var attrs = ['src', 'href', 'action'];
                for(var i = 0; i < attrs.length; i++) {
                    var val = node.getAttribute(attrs[i]);
                    if(val !== null && val.length) {
                        href = val;
                        attr = attrs[i];
                        break;
                    }
                }
                if(attr) {
                    node.setAttribute(attr, pinegrow.getOriginalUrl(href));
                }
            } else if(node.textNode) {
                if(node.content) {
                    node.content = pgSpecialChars.replaceCharsWithEntities(node.content);
                }
            }
            return true;
        })
        if(ret_nodes) {
            ret_nodes.rootNode = p.rootNode;
        }
        return assign_ids ? p.rootNode.toStringWithIds(true, pinegrow.getFormatHtmlOptions())
            : p.rootNode.toStringOriginal(true, pinegrow.getFormatHtmlOptions());
    }

    this.isTextMode = function() {
        return textMode;
    }

    this.setTextMode = function(val, sourceButton, skip_edit) {
        textMode = val;
        if(!skip_edit) {
            if (!val) {
                this.stop();
            } else {
                var pgel = pinegrow.selectedElements.getLastSelected();
                if (pgel) {
                    this.edit(pgel.get$DOMElement());
                }
            }
        }
        if (sourceButton != this.textButton) {
            this.textButton.setActive(textMode);
        }
        pinegrow.selectedElements.setShowMenus(!textMode);
    }

    pinegrow.addEventHandler('on_element_selected', function(page, pgel, info) {

        if(textMode && pgf.auto_text_edit_done && currentPgel !== pgel) {
            _this.stop();

        } else if(textMode) {
            if(pgel) {
                if(info && info.sourceType && info.sourceType == 'undo' && pgel == currentPgel) return;

                //check if el is editable
                if(!canMakeChange(pgel, 'edit_content', null, true /* silent */)) {
                    pinegrow.showQuickError('This element is not editable in Text editing mode.');
                    _this.stop();
                    return;
                }

                var $el = (info && info.clicked_element) ? info.clicked_element : pgel.get$DOMElement();

                if ($el) {
                    if(_this.is$ElementInsideEdited($el) && false) {
                        //let it be selected
                    } else {
                        _this.edit($el, false, info && info.source /* skip focus */, (info && info.event) ? info.event : null);
                    }
                }
            } else {
                _this.stop();
            }
        }
    })

    this.is$ElementEdited = function($el) {
        return $el.is('[contentEditable="true"]');
    }

    this.is$ElementInsideEdited = function($el) {
        return ($el.closest('[contentEditable="true"]').length > 0) && !$el.is('[contentEditable="true"]');
    }

    this.getEditorId = function() {
        return 'pgte_' + editor_id++;
    }

    var show_instructions = true;

    this.edit = function($el, double_click, skip_focus, event, ignore_errors) {

        var pgel = getElementPgNode($el);
        //pgel = pgel || pinegrow.selectedElements.getLastSelected();
        //var qte = new PgQuickTextEdit(pgel);
        //return;

        //is element already in edit and we just have a double click to select text?
        if(double_click && $el.closest('[contentEditable="true"]').length) {
            //Ignore
            return true;
        }

        this.stop();

        this.setTextMode(true, null, true);

        try {
            //var pgel = getElementPgNode($el);
            //Handle situation where .stop above invalidates $el
            var node = $el.get(0);
            if(!node.parentNode) {
                node = node.ownerDocument.documentElement.querySelector('[data-pg-id="' + pgel.getId() + '"]');
                $el = $(node);
            }

            currentPgel = pgel;
            var page = currentPage = pgel.getPage();
            var pageView = currentPageView = page.getPageViewOf$DOMElement($el);

            var status = {
                allowed: true,
                msg: ''
            }
            page.callFrameworkHandler('on_before_inline_edit', $el, status);

            if(!status.allowed) {
                pinegrow.showQuickError(status.msg);
                this.stop();
                this.setTextMode(false, null, true);
                return null;
            }

            var e = new PgInlineEditor(pgel, pageView, $el, show_instructions);

            var r = e.startEdit(skip_focus, event, ignore_errors);
            if (r === true) {
                currentEditor = e;

                show_instructions = false;

            } else if(typeof r == 'object') {
                //$el can't be edited, r is its parent
                return this.edit(r, double_click, skip_focus);
            } else if(r === 'code') {
                e.destroy();
                this.stop();
                this.setTextMode(false, null, true);
                pinegrow.editElementCode(pgel);
                return r;
            } else if(r === 'exit') {
                this.stop();
                this.setTextMode(false, null, true);
                return null;
            }

            //pinegrow.selectedElements.temporaryHideMenus();
            return r;

        } catch(err) {

        }
    }

    this.stop = function(pv) {
        var match = pv === undefined;
        if(pv) {
            if(pv instanceof PgPageView) {
                match = pv === currentPageView;
            } else if(pv instanceof CrsaPage) {
                match = pv === currentPage;
            }
        }

        if(currentEditor && match) {
            currentEditor.destroy();

            pinegrow.dispatchEvent('on_inline_editing_done', currentPage, currentPgel);

            if(pinegrow.selectedElements.isSelectedElement(currentPgel)) {
                pinegrow.selectedElements.updateMenuForPgel(currentPgel);
            }

            currentEditor = null;
            currentPgel = null;

            pinegrow.inlineEditManager.textButton && pinegrow.inlineEditManager.textButton.setActive(false);
        }

        //pinegrow.selectedElements.restoreShowMenus();
    }

    this.isInEdit = function() {
        return currentEditor != null;
    }

    this.getPage = function() {
        return currentPgel ? currentPgel.getPage() : null;
    }

    this.getEditedPgel = function() {
        return currentPgel;
    }


}


var PgInlineEditor = function(pgel, pageView, $el, show_instructions) {

    var _this = this;
    var changeTimer = null;
    var $toolbar = null;
    var changed = false;
    var $parent = null;

    var previous_html = pgel.html(null, true);

    var $edit_el = null;
    var $style;
    var editor_id = null;

    var mediumEditor;

    var page = pgel.getPage();
    var $html = $el.closest('html');

    var toolbarPositioner = null;

    var split_str = '##PGSPLITHERE_DO_UNDO_ONE_MORE_TIME##';

    $toolbar = $('<div class="pg-inline-edit-toolbar"></div>').appendTo(pageView.getPageParent$Body());

    /*
    $('body').on('crsaWillMakeChange', function(e) {
        _this.endEdit();
    });


    var onPageChanged = function() {
        if(changeTimer) {
            clearTimeout(changeTimer);
            changeTimer = null;
        }
        if(keyupTimer) {
            clearTimeout(keyupTimer);
            keyupTimer = null;
        }
        changeTimer = setTimeout(needsLayout, 750);

        didMakeChange(_this.selectedPage, $parent);
    }
*/

    var copySourceFromDOMToHtml = function(html) {
        /*
        html = removeCrsaClassesFromHtml(html);
        html = html.replace(/\sdata\-pg\-tree\-id="[0-9]+"/g, '');
        html = html.replace(/\starget="_pg_blank"/g, '');

        var p = new pgParser();
        p.replaceExistingIds = false;
        p.assignIds = true;
        p.parse(html);
        */

        var p = {};
        html = pinegrow.inlineEditManager.domHtmlToSource(html, p, true);

        pinegrow.undoManager.getHistoryForPage(page).ignore(function() {
            pgel.withEmitEvent(function () {

                while (pgel.children.length) {
                    pgel.children[0].remove();
                }
                for (var i = 0; i < p.rootNode.children.length; i++) {
                    pgel.addChild(p.rootNode.children[i]);
                }
                var edit_el = $edit_el.get(0);
                pgel.mapIdsToDomElement(edit_el);

                //check if href or target changed, changed by edit link function
                var attrs = ['href', 'target'];
                attrs.forEach(function(a) {
                    var val = edit_el.getAttribute(a);
                    if(a === 'target' && val === '_pg_blank') {
                        val = null;
                    }
                    if(pgel.getAttribute(a) !== val) {
                       if(!edit_el.hasAttribute(a)) {
                           pgel.removeAttribute(a);
                       } else {
                           pgel.setAttribute(a, val);
                       }
                    }
                })

            }, 'html', {html: html, sourcePageView: pageView});
        })

        // if(!splitOnNextUpdate) {
        //create custom undo
        var pgEvent = new PgNodeEvent(pgel, 'updateInner', null, {
            oldValue: previous_html.replace(split_str, '')
        }, page);

        var undoEvent = new PgUndoEvent(pgEvent);
        pinegrow.undoManager.getHistoryForPage(page).addEvent(undoEvent);
        /////
        //  }

        if(pgel.hasClass('pg-empty-placeholder')) {
            pgel.removeClass('pg-empty-placeholder');
        }
        if(pgf.empty_ph_attribute && pgel.hasAttribute(pgf.empty_ph_attribute)) {
            pgel.removeAttribute(pgf.empty_ph_attribute);
        }

        pinegrow.changeManager.didChange(pgel, 'Inline text edit', {source: _this, undoCombineKey: 'inline_edit'});

        previous_html = pgel.html(null, true);
    }

    var updateTimer = null;

    var doUpdate = function(skip_br_check) {
        updateTimer = null;

        //try to catch <br> that is sometimes inserted at the end
        if(!skip_br_check) {
            $edit_el.find('br').each(function (i, br) {
                var $br = $(br);
                if (!br.hasAttribute('data-pg-id')) {
                    //the br was inserted
                    if ($br.next().length === 0) {
                        //last one
                        //$br.remove();
                        $br.attr('data-pg-inline-edit-remove', 'true');
                    }
                }
            })
        }
        var html = mediumEditor.getContent();
        copySourceFromDOMToHtml(html);
        pinegrow.highlightElement(null);
    }

    var splitOnNextUpdate = false;

    this.startEdit = function(skip_focus, event, ignore_errors) {

        pinegrow.stats.using('edit.inlinetext');

        if(!canMakeChange(pgel, 'edit_content')) return false;

        if(pgel.containsServerSideScripts(true)) {
            pinegrow.showQuickMessage('The element contains Server side code. Using <b>Edit code</b>...', 3000, false);
            //return false;
            return 'code';
        }

        $edit_el = $el.closest('div,body,label,section,table,[data-pgc-field],[data-pgc-edit]');

        $edit_el = $el;

        if($edit_el.length == 0) $edit_el = $el;

        var simple = false;

        var $p = $edit_el.parents('p');
        if($p.length) {
            //$edit_el = $p;

            simple = true;
        }

        var $span = $edit_el.closest('span,b,strong,em,i');
        if($span.length) {
            //$edit_el = $span.parent();
            simple = true;
        }

        var headings = true;

        if($edit_el.closest('h1,h2,h3,h4,h5,h6').length) {
            headings = false;
        }

        //make edit more selective
        if($edit_el.is('img') && !$edit_el.is('[data-pgc-field],[data-pgc-edit]')) {
            $edit_el = $edit_el.parent();
        }

        if($edit_el.find('php').length) {
            var $withoutphp = $el;
            while($withoutphp.get(0) != $edit_el.get(0) && $withoutphp.parent().find('php').length == 0) {
                $withoutphp = $withoutphp.parent();
            }
            $edit_el = $withoutphp;
        }

        var pgbody = getElementPgNode($edit_el);

        if(['html', 'head', 'aaabody'].indexOf(pgbody.tagName) >= 0) {
            pinegrow.showQuickError('Text editing mode is on. Can\'t edit text of ' + pgbody.tagName.toUpperCase() + ' element.' + (pgbody.tagName == 'body' ? ' Select a sub-element.' : ''));
            return;
        }

        if((pgNodeContainsDynamic($edit_el) || !pgbody) && !ignore_errors) {
            $edit_el = $el;
            pgbody = getElementPgNode($edit_el)
            if(pgNodeContainsDynamic($edit_el) || !pgbody) {
                var $d = showAlert('<p>Editing this element is not recommended because:</p><ul><li>The element contains dynamic elements that were created by JavaScript code.</li><li>The HTML code is not valid (for example, a tag is not closed or a H1 is nested inside a P element).</li></ul><p>You can use Page -&gt; Edit code to edit the code of the whole page.</p><p>Or, chose "Edit anyway" option, but note that it can lead to unintended results. Use Undo if things go wrong.</p>', 'Can not edit this element', null, "OK, Do not edit");

                var $b = $('<button type="button" class="btn btn-default btn-sm">Edit anyway</button>').insertBefore($d.find('button.ok'));

                $b.on('click', function(e) {
                    e.preventDefault();
                    $d.modal('hide');
                    pinegrow.inlineEditManager.edit($edit_el, false, false, event, true);
                })
                return 'exit';
            }
        }

        if($edit_el.get(0) != $el.get(0)) {
            return $edit_el;
        }

        var $new_selected_element = null;

        if(page.javascriptEnabled || $edit_el.find('.pinegrow-ui-helper').length) {
            //in case JS code changed the DOM html
            /*if(pinegrow.isDescendantSelected($edit_el)) {
                $new_selected_element = $edit_el;
            }*/
            var nojs_code = page.getViewInnerHTMLForElement(pgbody, true /* disable js */);

            var code = function(c) {
                c = c.replace(/[\s\n\r]/g,'');
                c = c.replace(/data\-pg\-[a-z\-]*\="[^"]*"/g, '');
                return c;
            }

            if(code(nojs_code) != code($edit_el.html())) {
                //pinegrow.showQuickMessage('MISKOLIN');
                //console.log(code(nojs_code));
                //console.log(code($edit_el.html()));
                $edit_el.html(nojs_code); //only switch if js stuff is present
            }
        }


        if($edit_el.is('button')) {
            $edit_el.on('keydown.inlineedit', function(e) {
                if(e.key == ' ') {
                    //var caretPositions = MediumEditor.selection.getCaretOffsets(e.target);
                    var ch = ' ';
                    /*
                    var r = document.getSelection().getRangeAt(0);
                    console.log(r);
                    var test = '###TOKEN###';
                    var inner = e.target.innerHTML;
                    MediumEditor.util.insertHTMLCommand(e.target.ownerDocument, test);

                    if(e.target.innerHTML.startsWith(test) || e.target.innerHTML.endsWith(test)) {
                        ch = '&nbsp';
                    }
                    e.target.innerHTML = e.target.innerHTML.replace(test, '');
                    */
                    MediumEditor.util.insertHTMLCommand(e.target.ownerDocument, ch);
                    e.preventDefault();
                }
            })
        }


        toolbarPositioner = new PgPositionNextToPage$El(pageView, $toolbar, $edit_el, null, 12, true /* top */, true /* use fastdom */);

        var $page = pageView.get$Page();
        editor_id = pinegrow.inlineEditManager.getEditorId();

        $page.attr('id', editor_id);

        $style = $('<style></style>').appendTo($page);

        pageView.addViewChangeHandler(onPageViewChanged, true);

        var buttons = ['phraseBold', 'phraseItalic', 'phraseUnderline', 'phraseSpan', 'pgAnchor', /*'h1', 'h2', 'h3', 'h4', 'orderedlist', 'unorderedlist', 'quote',*/ 'pgInsertSpecialChar','pgInsert', 'pgExitEdit'];

        buttons = [];
        if($edit_el.closest('b,strong').length == 0) buttons.push('phraseBold');
        if($edit_el.closest('em,i').length == 0) buttons.push('phraseItalic');
        buttons.push('phraseUnderline');

        buttons.push('phraseSpan');
        buttons.push('phraseCode');
        if($edit_el.closest('a').length == 0 || $edit_el.is('a')) buttons.push('pgAnchor');
        buttons.push('pgInsertSpecialChar');
        //buttons.push('pgInsert');
        buttons.push('pgExitEdit');

        if(simple) {
            //buttons = ['pgExitEdit'];
        } else if(!headings) {
            //buttons = ['phraseBold', 'phraseItalic', 'phraseUnderline', 'phraseSpan', 'pgAnchor', 'pgInsert', 'pgExitEdit'];
        }

        if(show_instructions) {
            buttons.push('pgInstructionsEdit');
        }

        var opts = {
            elementsContainer: $page.get(0),
            contentWindow: pageView.getWindow(),
            ownerDocument: pageView.getDocument(),
            spellcheck: false,
            toolbar: {
                buttons: buttons,
                'static': true,
                relativeContainer: $toolbar.get(0),
                sticky: true,
                updateOnEmptySelection: true
            },
            paste: {
                cleanPastedHTML: true,
                forcePlainText: true
            },
            anchorPreview: false,
            extensions: {
                'pgExitEdit': new PgInlineEditExitButton(),
                'pgInsert': new PgInlineEditInsertButton(),
                'pgInsertSpecialChar': new PgInlineEditInsertSpecialCharButton(),
                'pgAnchor': new PgInlineEditAnchor(),
                pgInstructionsEdit: new PgInlineEditInstructionsButton(),
                phraseSpan: new MediumEditorPhrase({
                    name: 'phraseSpan',
                    phraseTagName: 'span',
                    aria: 'span',
                    contentDefault: 'span',
                    phraseClassList: []
                }),
                phraseCode: new MediumEditorPhrase({
                    name: 'phraseCode',
                    phraseTagName: 'code',
                    aria: 'code',
                    contentDefault: 'code',
                    phraseClassList: []
                }),
                phraseBold: new MediumEditorPhrase({
                    name: 'phraseBold',
                    phraseTagName: 'b',
                    aria: 'b',
                    contentDefault: '<b>B</b>',
                    phraseClassList: []
                }),
                phraseItalic: new MediumEditorPhrase({
                    name: 'phraseItalic',
                    phraseTagName: 'i',
                    aria: 'i',
                    contentDefault: '<b><i>I</i></b>',
                    phraseClassList: []
                }),
                phraseUnderline: new MediumEditorPhrase({
                    name: 'phraseUnderline',
                    phraseTagName: 'u',
                    aria: 'u',
                    contentDefault: '<b><u>U</u></b>',
                    phraseClassList: []
                })
            }
        }

        var in_p = false;

        var is_inline = ($edit_el.css('display') || '').indexOf('inline') >= 0;
        var no_p_sel = 'p,a,span,button,h1,h2,h3,h4,h5,h6';

        opts.paste.unwrapTags = ['div', 'code'];

        if(simple || $edit_el.is(no_p_sel) || $edit_el.closest(no_p_sel).length || is_inline) {
            in_p = true;
            //opts.disableReturn = true;
            opts.paste.unwrapTags.push($edit_el.get(0).tagName);
        }


        //delete MediumEditor.extensions.button.prototype.defaults.bold.style;

        mediumEditor = new MediumEditor($edit_el.get(0), opts);

        $toolbar.find('.medium-editor-instructions').parent().addClass('medium-editor-instructions-parent')

        //mediumEditor = new MediumEditor($('body'), {});

        //do call stuff here
        pinegrow.dispatchEvent('on_inline_edit_started', page, pgel, $edit_el);


        splitOnNextUpdate = false;

        var splitElement = function(split_str) {
            pinegrow.inlineEditManager.stop();

            //lets split pgbody
            var html = pgbody.html(null, true /* with ids */);
            var a = html.split(split_str);

            if(a.length > 1) {
                pgbody.withoutEvents(function() {
                    pgbody.html(html.replace(split_str, ''));
                })
                pgbody.html(a[0]);
            }

            var api = new PgApi();
            var pgel_next = null;

            api.duplicateElements(pgbody, function(new_col) {
                new_col.forEachElement(function(pgel, i) {
                    pgel.html(a[1]);
                    if(!pgel_next) {
                        pgel_next = pgel;
                    }
                })
                if(pgel_next) {
                    var $pgel = pgel_next.get$DOMElement($html);
                    if($pgel) {
                        pinegrow.selectElement(pgel_next, null, false, $pgel);

                        if(pgf.auto_text_edit_done) {
                            pinegrow.inlineEditManager.edit($pgel);
                        }
                    }
                }
            })
        }

        var shift_enter_pressed = false;

        if(in_p || simple) {
            mediumEditor.subscribe("editableKeydownEnter", function (event, element) {
                if(!event.shiftKey) {
                    if(!simple) {
                        if(pgf.auto_text_edit_done && false) {
                            mediumEditor.options.ownerDocument.execCommand.apply(mediumEditor.options.ownerDocument, ['insertHTML', false, '<br>']);
                            //MediumEditor.util.insertHTMLCommand(mediumEditor.options.ownerDocument, "<br>")
                        } else {
                            var node = MediumEditor.selection.getSelectionStart(mediumEditor.options.ownerDocument)
                            //MediumEditor.util.insertHTMLCommand(mediumEditor.options.ownerDocument, "<br>&nbsp;")
                            splitOnNextUpdate = true;
                            MediumEditor.util.insertHTMLCommand(mediumEditor.options.ownerDocument, split_str)
                        }
                    }

                    //mediumEditor.options.ownerDocument.execCommand('DefaultParagraphSeparator', false, 'br')
                    //mediumEditor.options.ownerDocument.execCommand('InsertParagraph')
                    //mediumEditor.options.ownerDocument.execCommand("insertHtml", false, "world" + "<br> ");
                    event.preventDefault();

                    //node = MediumEditor.selection.getSelectionStart(this.options.ownerDocument)
                    //console.log(MediumEditor.selection.getCaretOffsets(node));
                    //var p = mediumEditor.options.ownerDocument.createElement('br');
                    //node.previousElementSibling.parentNode.insertBefore(p, node);
                } else {
                    shift_enter_pressed = true;
                }
            }.bind(mediumEditor));
        }

        if(!skip_focus) {
            var orig_pg_app_win_ignore_focus_blur = pg_app_win_ignore_focus_blur;
            pg_app_win_ignore_focus_blur = true;
            mediumEditor.events.focusElement($edit_el.get(0));
            pg_app_win_ignore_focus_blur = orig_pg_app_win_ignore_focus_blur;
        }

        function unwrap(wrapper) {
            // place childNodes in document fragment
            var docFrag = document.createDocumentFragment();
            while (wrapper.firstChild) {
                var child = wrapper.removeChild(wrapper.firstChild);
                docFrag.appendChild(child);
            }

            // replace wrapper with document fragment
            wrapper.parentNode.replaceChild(docFrag, wrapper);
        }

        mediumEditor.subscribe("editableInput", function (event, element) {
            var node = MediumEditor.selection.getSelectionStart(mediumEditor.options.ownerDocument);

            //Not sure why this was here. It messes up SPAN command
            /*
            if(node && node.tagName == 'SPAN' && !node.getAttribute('data-pg-id')) {
                unwrap(node);
            }
            */

            if(updateTimer) {
                clearTimeout(updateTimer);
                updateTimer = null;
            }
            var skip_br_check = shift_enter_pressed;
            shift_enter_pressed = false;

            if(splitOnNextUpdate) {
                doUpdate(skip_br_check);
                splitElement(split_str);
            } else {
                updateTimer = setTimeout(function () {
                    doUpdate(skip_br_check);
                }, 50);
            }

        }.bind(mediumEditor));

        if(event && false) { //Disabled because it prevented the element to be focused. Not sure what this does.
            var doc = pageView.getDocument();
            //var page_pos = pageView.globalPointToPagePoint({left: event.pageX, top: event.pageY});
            var page_pos = {left: event.clientX, top: event.clientY};
            var pel = doc.elementFromPoint(page_pos.left, page_pos.top);
            if(pel) {
                var range = doc.caretRangeFromPoint(page_pos.left, page_pos.top);
                if(range) {
                    sel = pageView.getWindow().getSelection();

                    sel.removeAllRanges();
                    sel.addRange(range);
                }
            }
        }

        return true;
    }

    var onPageViewChanged = function(pv, what, info) {
        if(toolbarPositioner) {
            if(pv instanceof PgPageView) {
                if (pv.isVisible()) {
                    pg$Show($toolbar);
                } else {
                    pg$Hide($toolbar);
                }
            }
            toolbarPositioner.reposition();
        }
    }

    pinegrow.addEventHandler('on_page_changes_done', onPageViewChanged);

    this.destroy = function() {
        if(updateTimer) {
            clearTimeout(updateTimer);
            updateTimer = null;
            doUpdate();
        }
        if(mediumEditor) {
            mediumEditor.destroy();
            $edit_el.removeAttr('contenteditable');
            //$edit_el.editor('destroy');
            toolbarPositioner.destroy();
            $style.remove();
            $toolbar.remove();

            $edit_el.off('keydown.inlineedit');
        }
        pageView.removeViewChangeHandler(onPageViewChanged);
        pinegrow.removeEventHandler('on_page_changes_done', onPageViewChanged);
    }

}

var PgInlineEditExitButton = MediumEditor.Extension.extend({
    name: 'pgExitEdit',

    init: function () {
        this.button = this.document.createElement('button');
        this.button.classList.add('medium-editor-action');
        this.button.classList.add('pg-text-edit-menu-done');

        this.button.innerHTML = `<b>Done <input type="checkbox" ${pgf.auto_text_edit_done ? 'checked' : ''} style="position: relative;top: 2px;margin-left: 4px;margin-top:0;"></b>`;

        this.checkbox = this.button.querySelector('input');

        this.on(this.checkbox, 'click', function(event) {
            event.stopPropagation();
        });

        var _this = this;

        this.on(this.checkbox, 'change', function(event) {
            event.stopPropagation();
            pgf.auto_text_edit_done = _this.checkbox.checked;
            pinegrow.setSetting('auto-text-edit-done', pgf.auto_text_edit_done ? '1' : '0');
            pinegrow.showMode('Exit text editing when selecting another element', pgf.auto_text_edit_done);
        });

        this.on(this.button, 'click', this.handleClick.bind(this));

        addTooltip($(this.button), 'Exit the text editing mode.<br>Check the checkbox to auto-exit when selecting another element. Uncheck it to keep the text editing mode active when moving between elements.');
    },

    getButton: function () {
        return this.button;
    },

    handleClick: function (event) {
        pinegrow.inlineEditManager.setTextMode(false);
        //this.base.checkContentChanged();
    }
});

var PgInlineEditInsertButton = MediumEditor.Extension.extend({
    name: 'pgInsert',

    init: function () {
        this.button = this.document.createElement('button');
        this.button.classList.add('medium-editor-action');
        this.button.innerHTML = '<b>Insert</b>';

        this.on(this.button, 'click', this.handleClick.bind(this));
    },

    getButton: function () {
        return this.button;
    },

    handleClick: function (event) {
        pinegrow.showQuickInsert(pinegrow.inlineEditManager.getEditedPgel(), $(this.button));
    }
});


var PgInlineEditInsertSpecialCharButton = MediumEditor.Extension.extend({
    name: 'pgInsertSpecialChar',

    init: function () {
        this.button = this.document.createElement('button');
        this.button.classList.add('medium-editor-action');
        this.button.classList.add('pg-insert-char-menu-link');
        this.button.innerHTML = '<b>&bigstar;</b>';

        this.on(this.button, 'click', this.handleClick.bind(this));
    },

    getButton: function () {
        return this.button;
    },

    handleClick: function (event) {

        var editor = this;

        var $b = $(this.button);

        var insert = function() {
            MediumEditor.util.insertHTMLCommand(editor.document, this.char)
        }

        var chars = pgSpecialChars.getChars();

        var a = []

        var map = {
            ' ' : 'Regular',
            '&nbsp;': 'Non-breaking'
        }

        chars.forEach(function(sec) {
            var sub = sec.chars.map(function(ch) {
                return {
                    label: map[ch] ? map[ch] : `<span style="font-size:16px;">${ch}</span>`,
                    char: ch,
                    action: insert
                }
            })

            a.push({
                label: sec.name,
                submenu: [
                    {
                        buttons: sub
                    }
                ]
            })
        })

        var menu = new CrsaContextMenu(a, $b);
        menu.showForElement($b, null, 12);

        event.stopImmediatePropagation();
        event.stopPropagation();
    }


});


var PgInlineEditInstructionsButton = MediumEditor.Extension.extend({
    name: 'pgInstructionsEdit',

    init: function () {
        this.button = this.document.createElement('div');
        this.button.classList.add('medium-editor-action');
        this.button.classList.add('medium-editor-instructions');
        this.button.innerHTML = pgShowKbdCombo('SHIFT + ENTER') + ' to insert &lt;br&gt; line break. ' + pgShowKbdCombo('ESC') + ' to exit text mode.';

        this.on(this.button, 'click', function(e) {
            e.preventDefault();
        });

        var _this = this;
        setTimeout(function() {
            if(_this.button && _this.button.parentNode) {
                _this.button.parentNode.classList.add('animated');
            }
        }, 5000);
    },

    getButton: function () {
        return this.button;
    }
});

//Define a function that will create the control
var pgCreateAnchorControl = function() {
    //will keep helper functions and vars private

    //Create control, needs a unique id
    var c = new PgCustomPropertyControl('html-anchor-control');

    var defHref = {
        type: 'text',
        name: 'Href',
        placeholder: 'http://example.com',
        action: 'element_attribute',
        attribute: 'href',

    }

    var defLabel = {
        type: 'text',
        name: 'Label',
        placeholder: 'Link name',
        action: 'element_html'
    }

    var defTarget = {
        type: 'select',
        name: 'Open in',
        action: 'element_attribute',
        attribute: 'target',
        show_empty: true,
        options: [
            {key: '_blank', name: 'New window'}
        ]
    }

    c.keyHref = 'html-anchor-control-href';
    c.keyLabel = 'html-anchor-control-label';
    c.keyTarget = 'html-anchor-control-target';

    //Register subfields.
    c.onDefine = function() {
        this.registerInputField(c.keyHref, defHref);
        this.registerInputField(c.keyLabel, defLabel);
        this.registerInputField(c.keyTarget, defTarget);
    }

    //Show control. Return the control $el.
    c.onShow = function() {
        var $c = $('<div></div>');
        this.showInputField($c, c.keyHref, defHref);
        this.showInputField($c, c.keyLabel, defLabel);
        this.showInputField($c, c.keyTarget, defTarget);
        return $c;
    }

    //Called when control is recycled and new values are set
    c.onSetValues = function() {
        //Nothing to do in this case. PG will take care of updating sub fields with new values
    }

    //Return the control
    return c;
};

var PgAnchorView = function(editor, win) {

    var $element = $('<div class="pg-anchor"></div>');

    var view = new PgUIView($element);

    this.view = view;

    var _this = this;

    this.win = null;

    var current_pgel = null;

    var onElementSelected = function(page, pgel) {
        view.whenVisible(function() {
        })
        current_pgel = pgel;
    }

    var control = null;
    var $control;

    var existing_link_node = null;

    var page = pinegrow.inlineEditManager.getPage();

    var href_input = new PgInputField('href', {
        type: 'text',
        name: 'Href',
        file_picker: true,
        file_picker_parent_url: page ? page.url : null,
        placeholder: 'http://example.com'
    });

    href_input.appendTo($element);

    var label_input = new PgInputField('linklabel', {
        type: 'text',
        name: 'Label'
    });

    label_input.appendTo($element);


    var target_input = new PgInputField('target', {
        type: 'select',
        show_empty: true,
        name: 'Open in',
        options: [
            {key: '_blank', name: 'New window'}
        ]
    });

    target_input.appendTo($element);

    this.show = function(pgel) {
        var values = {};

        var label = '';
        var target = '';
        var href = '';

        existing_link_node = null;

        if(editor) {
            var range = MediumEditor.selection.getSelectionRange(editor.document);

            if(range.startContainer.nodeName.toLowerCase() === 'a') {
                existing_link_node = range.startContainer;
            } else if(range.endContainer.nodeName.toLowerCase() === 'a') {
                existing_link_node = range.endContainer;
            } else {
                var n = MediumEditor.util.getClosestTag(MediumEditor.selection.getSelectedParentElement(range), 'a');
                if(n) {
                    existing_link_node = n;
                }
            }
            if(existing_link_node) {
                var ex_pgel = getElementPgNode($(existing_link_node));
                label = ex_pgel.html();
                href = ex_pgel.getAttribute('href');
                target = ex_pgel.getAttribute('target');

                $remove.show();
                $button.text('Save changes');

            } else if(!range.collapsed) {
                label = MediumEditor.selection.getSelectionHtml(editor.document);
                label = pinegrow.inlineEditManager.domHtmlToSource(label);

                $remove.hide();
            }
        }
        href_input.setValue(href);
        label_input.setValue(label);
        target_input.setValue(target);

        href_input.focus();
    }


    var onCreate = function(e) {
        if(e) e.preventDefault();
        if(editor) {
            var href = href_input.getValue(href);
            var label = label_input.getValue(label);
            var target = target_input.getValue(target);

            if(href.length) {
                var random_url ='#pgtempurl_' + Math.random();

                if (existing_link_node) {
                    existing_link_node.setAttribute('href', random_url);
                } else {
                    var opts = {
                        value: random_url
                    }
                    var selection = MediumEditor.selection.getSelectionStart(editor.document);

                    if(!selection || (selection !== editor.base.elements[0] && !editor.base.elements[0].contains(selection))) {
                        MediumEditor.selection.selectNode(editor.base.elements[0], editor.document);
                        //label = '';
                    }
                    //var html = MediumEditor.selection.getSelectionHtml(editor.document);
                    //editor.base.elements
                    editor.execAction('createLink', opts);
                    existing_link_node = MediumEditor.selection.getSelectionStart(editor.document);
                }

                var node = existing_link_node;

                if(editor.base.elements[0].getAttribute('href') === random_url) {
                    node = editor.base.elements[0];
                } else {
                    node = editor.base.elements[0].querySelector(`a[href="${random_url}"]`);
                }

                if(node) {
                    node.setAttribute('href', href);
                    if (target) {
                        node.setAttribute('target', target);
                    } else {
                        node.removeAttribute('target');
                    }
                    if (label.length) node.innerHTML = label;
                }

                editor.base.events.triggerCustomEvent('editableInput', {}, MediumEditor.selection.getSelectionElement(editor.window));
            }
        }
        if(_this.win) {
            _this.win.destroy();
        }
    }

    var onRemove = function(e) {
        if(e) e.preventDefault();
        if(editor) {
            var range = MediumEditor.selection.getSelectionRange(editor.document);

            if(range.collapsed) {
                if(editor.base.elements[0] === existing_link_node) {
                    //Remove link on the selected element
                    var pgel = getElementPgNode($(existing_link_node));
                    pinegrow.inlineEditManager.stop();
                    crsaFuncs.detagElement(pgel);

                } else {
                    var $e = $(existing_link_node);
                    $e.contents().insertBefore($e);
                    $e.remove();
                    editor.base.checkContentChanged();
                }
            } else {
                editor.execAction('unlink');
            }
        }
        if(_this.win) {
            _this.win.destroy();
        }
    }
    var $buttons = $('<div class="pg-field-buttons"></div>').appendTo($element);
    var $button;
    $button = $('<button class="pg-button pg-button-primary">Create link</button>').appendTo($buttons);
    $button.on('click', onCreate);

    this.onEnter = function() {
        onCreate();
        return false;
    }

    var $remove = $('<button class="pg-button">Remove the link</button>').appendTo($buttons);
    $remove.on('click', onRemove);

    view.onResize = function() {
    }

    pinegrow.addEventHandler('on_element_selected', onElementSelected);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_element_selected', onElementSelected);
        _this.win = null;
    }
}

var PgQuickAnchor = function(pgel, editor, $target) {

    var ac = new PgAnchorView(editor);

    var win = new PgQuickWindow('Edit link', ac, 'quickAnchor', 400, 190);

    this.win = win;
    ac.win = win;

    win.onEnter = function() {
        return ac.onEnter();
    }

    win.showFor$El($target);

    ac.show(pgel);

}

/*
var PgInlineEditExitButton = MediumEditor.Extension.extend({
    name: 'pgExitEdit',

    init: function () {
        this.button = this.document.createElement('button');
        this.button.classList.add('medium-editor-action');
        this.button.innerHTML = '<b>Done</b>';

        this.on(this.button, 'click', this.handleClick.bind(this));
    },

    getButton: function () {
        return this.button;
    },

    handleClick: function (event) {
        pinegrow.inlineEditManager.setTextMode(false);
        //this.base.checkContentChanged();
    }
});*/


var PgInlineEditAnchor = MediumEditor.Extension.extend({
    name: 'pgAnchor',
    init: function () {
        this.button = this.document.createElement('button');
        this.button.classList.add('medium-editor-action');
        this.button.classList.add('pg-text-edit-menu-link');
        this.button.innerHTML = '<b>Link</b>';

        this.on(this.button, 'click', this.handleClick.bind(this));
        this.subscribe('editableKeydown', this.handleKeydown.bind(this));
    },

    getButton: function () {
        return this.button;
    },

    handleClick: function (event) {
        event.preventDefault();
        event.stopPropagation();

        var range = MediumEditor.selection.getSelectionRange(this.document);
        /*
                if (range.startContainer.nodeName.toLowerCase() === 'a' ||
                    range.endContainer.nodeName.toLowerCase() === 'a' ||
                    MediumEditor.util.getClosestTag(MediumEditor.selection.getSelectedParentElement(range), 'a')) {
                    return this.execAction('unlink');
                }*/
        this.aqw = new PgQuickAnchor(null, this, $(this.button));

        this.aqw.win.onDestroy = (function() {
            this.aqw = null;
        }).bind(this);

        return false;
    },

    // Called when user hits the defined shortcut (CTRL / COMMAND + K)
    handleKeydown: function (event) {
        if (MediumEditor.util.isKey(event, MediumEditor.util.keyCode.K) && MediumEditor.util.isMetaCtrlKey(event) && !event.shiftKey) {
            this.handleClick(event);
        }
    },
    destroy: function () {
        if(this.aqw) {
            this.aqw.win.destroy();
        }
    },

    isAlreadyApplied: function (node) {
        return node.nodeName.toLowerCase() === 'a';
    },

    isActive: function () {
        return this.button.classList.contains('medium-editor-button-active');
    },

    setInactive: function () {
        this.button.classList.remove('medium-editor-button-active');
    },

    setActive: function () {
        this.button.classList.add('medium-editor-button-active');
    }
});


