/**
 * Created by Matjaz on 06/11/15.
 */

class PgLinkChecker {

    constructor() {
        var _this = this;
        this.errors = [];

        pinegrow.addEventHandler('on_page_menu', function(page, items) {
            items.push({
                label: 'Check for broken links...',
                kbd: null,
                action: function() {
                    _this.checkLinksAndDisplayErrors(page.sourceNode);
                }
            })

            return items;
        });
    }

    addError(pgel, msg, attr, url) {
        this.errors.push({pgel: pgel, attr: attr, url: url, error: msg});
    }

    elementSelector(pgel) {
        return '<b data-pg-id="' + pgel.getId() + '">' + pgel.getName({short: true}) + '</b>';
    }

    validate(root, on_update, on_done) {
        var _this = this;
        this.errors = [];
        var page = root.getPage();

        var check_list = [];
        var p = pinegrow.getCurrentProject();

        root.goThroughAllUrls(function(url, attr, pgel) {
            try {
                if(!url.startsWith('javascript:') && !url.startsWith('#') && !url.startsWith('mailto:') && !url.startsWith('tel:')) {
                    var abs_url = null;
                    if(url.startsWith('/')) {
                        if(p) {
                            abs_url = p.makeUrlAbsolute(url);
                        }
                    } else {
                        abs_url = new URL(url, page.url).toString();
                    }
                    abs_url && check_list.push({
                        pgel: pgel,
                        attr: attr,
                        url: abs_url,
                        rel_url: url
                    })
                }
            } catch(err) {}
            return url;
        })

        var total = check_list.length;

        on_update && on_update(0, total);

        var nextOrDone = function(delay) {
            setTimeout(function() {
                on_update && on_update(total - check_list.length, total);
                if(check_list.length) {
                    checkLink();
                } else {
                    on_done && on_done();
                }
            }, delay)
        }

        var checkLink = function() {
            var item = check_list.shift();
            if(crsaIsFileUrl(item.url)) {
                var path = crsaMakeFileFromUrl(item.url);
                if(!crsaIsFileExist(path)) {
                    _this.addError(item.pgel, `File <code>${item.rel_url}</code> not found.`, item.attr, item.rel_url);
                }
                nextOrDone(50);
            } else {
                fetch(item.url, { /* method: 'HEAD' */})
                    .then(res => {
                        if (res.ok) {
                            nextOrDone(0);
                        } else {
                            //console.log('Error getting remote file ' + item.url, error);
                            _this.addError(item.pgel, `URL <code>${item.rel_url}</code> not found.`, item.attr, item.rel_url);
                            nextOrDone(0);
                        }
                    }).catch(error => {
                        console.log('Error getting remote file ' + item.url, error);
                        _this.addError(item.pgel, `URL <code>${item.rel_url}</code> error.`, item.attr, item.rel_url);
                        nextOrDone(0);
                });
                /*
                pgFetchRemoteContent(item.url, {method: 'HEAD'}).catch(function (error) {
                    console.log('Error getting remote file ' + item.url, error);
                    _this.addError(item.pgel, `URL ${item.rel_url} not found.`, item.attr, item.rel_url);
                    on_update && on_update(total - check_list.length, total);
                    nextOrDone(0);
                }).then(function (data) {
                    nextOrDone(0);
                });*/
            }
        }

        nextOrDone(0);

    }

    hasErrors() {
        return this.errors.length > 0;
    }

    describeErrors() {
        var html = '';
        if(this.errors.length === 0) {
            html += '<li>No errors found.</li>';
        } else {
            this.errors.forEach(function (e) {
                var text = crsaGetSummaryStr(e.pgel.text(null, true) || '', 10, true);
                html += '<li data-pg-id="' + e.pgel.getId() + '"><span>' + e.pgel.getName({short: true}) + (text.length ? ` | <span style="opacity:0.65;">${text}</span>` : '') + '</span> - ' + e.error + '</li>';
            })
        }
        return html;
    }

    checkLinksAndDisplayErrors(node) {

        var validator = this;

        if(!validator.hasErrors()) {
        }

        var msg = '<p class="status"></p><p>Broken links:</p><ul style="user-select: text;">';
        msg += '</ul>';

        var page = node.getPage();

        var q = PgQuickDialog('Broken links in ' + page.getName(), function($element) {
            msg += '<p class="help">Click on the element on the above list to select it. If you can\'t see the element in the tree, switch to Source code <i class="icon icon-KODA"></i> tree mode.';
            msg += '</p>';

            var update = function(i, total) {
                var $ul = $element.find('ul');
                $ul.html( validator.describeErrors());
                $element.find('p.status').html(`Checking... ${i} of ${total} done.`)
            }

            $element.addClass('pg-code-validator-results');

            $element.on('click', '[data-pg-id]', function(e) {
                e.preventDefault();
                var pgid = $(this).attr('data-pg-id');
                var pgel = getPgNodeByPgId(pgid);
                if(pgel) {
                    pgel.scrollTo();
                    setTimeout(function() {
                        pinegrow.selectElement(pgel);
                        pinegrow.showQuickMessage(pgel.getName({short:true}) + ' was selected.')
                    }, 100)
                }
            })

            msg += '<button class="pg-button pg-button-primary refresh">Refresh</button>';

            var validate = function() {
                validator.validate(node, update, function() {
                    pinegrow.showQuickMessage('Done.');
                    $element.find('ul').html(validator.hasErrors() ? validator.describeErrors() : '<li>No errors found.</li>');
                    $element.find('p.status').html(`Done.`);
                });
            }

            $element.on('click', 'button.refresh', function(e) {
                e.preventDefault();
                validate();
            })

            validate();

            return msg;
        }, {width:480})

    }

}



$(function () {

    if(!pgf.edit_html) return;

    var pgLinkChecker = null;

    $('body').one('pinegrow-ready', function (e, pinegrow) {//Wait for Pinegrow to start-up
        pgLinkChecker = new PgLinkChecker();
    })
});