var PgCollection = function(pgel_or_$el) {

    var list = [];

    this.getList = function () {
        return list;
    }

    this.setList = function (l) {
        list = l;
    }

    var _this = this;

    if (pgel_or_$el) {
        if (pgel_or_$el instanceof pgParserNode) {
            this.add(pgel_or_$el);
        } else {
            pgel_or_$el.each(function (i, e) {
                var pgel = getElementPgNode($(e));
                if (pgel) {
                    _this.add(pgel);
                }
            })
        }
    }
}

PgCollection.prototype.getLength = function() {
    return this.getList().length;
}

PgCollection.prototype.select = function(pgel) {
    this.clear();
    this.add(pgel);
}

PgCollection.prototype.clear = function() {
    this.setList([]);
}

PgCollection.prototype.clone = function() {
    var n = new PgCollection();
    n.setList(this.getList().slice());
    return n;
}

PgCollection.prototype.add = function(pgel) {
    if(pgel instanceof PgCollection) {
        var _this = this;
        pgel.forEachElement(function(cpgel) {
            _this.add(cpgel);
        })
    } else {
        var list = this.getList();
        var idx = list.indexOf(pgel);
        if (idx < 0) {
            list.push(pgel);
        }
    }
}

PgCollection.prototype.remove = function(pgel) {
    var list = this.getList();
    var idx = list.indexOf(pgel);
    if(idx >= 0) {
        list.splice(idx, 1);
    }
    return idx;
}

PgCollection.prototype.cleanDeleted = function() {
    var r = [];
    var changed = false;
    this.forEachElement(function(pgel) {
        if(pgel.isDeleted) {
            changed = true;
        } else {
            r.push(pgel);
        }
    })
    this.setList(r);
    return changed;
}

PgCollection.prototype.forEachElement = function(func) {
    var list = this.getList();
    var current = list.slice(0);
    for(var i = 0; i < current.length; i++) {
        if (func(current[i], i, current.length) === false) break;
    }
}

PgCollection.prototype.contains = function(pgel) {
    var list = this.getList();
    return list.indexOf(pgel) >= 0;
}

PgCollection.prototype.getIndex = function(pgel) {
    var list = this.getList();
    return list.indexOf(pgel);
}

PgCollection.prototype.getLast = function() {
    var list = this.getList();
    return list.length > 0 ? list[list.length - 1] : null;
}

PgCollection.prototype.getFirst = function() {
    var list = this.getList();
    return list.length > 0 ? list[0] : null;
}

PgCollection.prototype.scrollTo = function() {
    var pgel = this.getFirst();
    pgel && pgel.scrollTo();
}

PgCollection.prototype.normalize = function() {
    var list = this.getList();
    var r = [];

    for(var i = 0; i < list.length; i++) {
        var contained = false;
        for(var j = 0; j < list.length; j++) {
            if(i != j && list[i].isDescendantOf(list[j])) {
                contained = true;
                break;
            }
        }
        if(!contained) {
            r.push(list[i]);
        }
    }
    this.setList(r);
}

PgCollection.prototype.reverse = function() {
    var list = this.getList();
    this.setList(list.reverse());
}

PgCollection.prototype.orderByDocumentPosition = function(fast) {
    var list = this.getList();
    var r = [];

    for(var i = 0; i < list.length; i++) {
        r.push({
            pgel: list[i],
            pos: list[i].getPositionInDocument(fast)
        })
    }

    r.sort(function(a, b) {
        return a.pos - b.pos;
    })

    list = [];
    r.forEach(function(d) {
        list.push(d.pgel);
    })
    this.setList(list);
}

PgCollection.prototype.to$List = function() {
    var r = [];
    this.forEachElement(function(pgel) {
        var $el = pgel.get$DOMElement();
        if($el) {
            r.push($el.get(0));
        }
    })
    return $(r);
}

PgCollection.prototype.getName = function(opts, skip_def) {
    var list = this.getList();
    if(list.length === 0) return '&lt;No element selected&gt';
    var options = $.extend({short: true}, opts || {})
    if(!options.def && !skip_def) {
        var page = list[0].getPage();
        if(page) {
            options.def = page.getMainType(null, list[0])
        }
    }
    return list[0].getName(options) + (list.length > 1 ? (' + ' + (list.length - 1)) : '');
}

PgCollection.prototype.getPages = function() {
    var r = [];
    this.forEachElement(function(pgel) {
        var cp = pgel.getPage();
        if(cp && r.indexOf(cp) < 0) {
            r.push(cp);
        }
    })
    return r;
}

PgCollection.prototype.forEachPage = function(func) {
    var r = [];
    var cols = [];
    this.forEachElement(function(pgel) {
        var cp = pgel.getPage();
        if(!cp && pgel.replacedWith) cp = pgel.replacedWith.getPage();

        if(cp) {
            var idx = r.indexOf(cp);
            if(idx < 0) {
                r.push(cp);
                cols.push(new PgCollection());
                idx = r.length - 1;
            }
            cols[idx].add(pgel);
        }
    })
    for(var i = 0; i < r.length; i++) {
        func(r[i], cols[i]);
    }
}

PgCollection.prototype.forEachElementAsync = function(func, done) {
    var results = [];
    var count = this.getLength();

    if(count === 0) {
        if(done) done(results);
        return;
    }

    this.forEachElement(function(pgel) {
        func(pgel, function(data) {
            results.push({
                pgel: pgel,
                data: data
            })
            count--;
            if(count === 0) {
                if(done) done(results);
            }
        })
    })
}

////end collection


var PgSelectedElements = function() {

    var _this = this;

    var collection = new PgCollection();

    var menus = {};
    var show_menus = true;

    var changedTimer = null;

    var delayedTasks = new PgDelayedTasks();

    var changedInfo = {};
    var setChangedInfo = function(info) {
        if(info) {
            changedInfo = info || {};
        }
    }

    var triggerChanged = function() {
        if(changedTimer) clearTimeout(changedTimer);
        changedTimer = setTimeout(function() {
            pinegrow.dispatchEvent('on_selected_elements_changed', null, _this.getPages(), changedInfo);
            changedInfo = {};
            changedTimer = null;
        }, 10)
    }

    var getMenuId = function(pgel) {
        return pgel.getId();
    }

    this.getCollection = function() {
        return collection;
    }

    this.setCollection = function(c) {
        collection = c;
    }

    this.select = function(pgel) {
        this.clear();
        return this.add(pgel);
    }

    this.clear = function() {
        var c = collection.clone();

        collection.clear();

        $.each(menus, function(key, m) {
            m.destroy();
        })

        menus = {};

        c.forEachElement(function(pgel) {
            pgel.deselected();
        })

        triggerChanged();

        shift_direction = 'down';
    }

    this.cleanDeleted = function() {
        if(collection.cleanDeleted()) {
            triggerChanged();
        }
    }

    this.add = function(pgel, info) {
        setChangedInfo(info);
        collection.add(pgel);

        var m = this.updateMenuForPgel(pgel);

        pgel.selected();

        triggerChanged();

        return m;
    }

    this.updateMenuForPgel = function(pgel) {
        var mid = getMenuId(pgel);

        if(menus[mid]) menus[mid].destroy();

        menus[mid] = new PgElementMenus(pgel, collection.getIndex(pgel));

        if(!show_menus) {
            menus[mid].hide();
        }

        hideAllButLastMenu();

        return menus[mid];
    }

    this.selectParent = function() {
        var col = new PgCollection();
        this.forEachElement(function(pgel) {
            if(pgel.parent) {
                col.add(pgel.parent);
            }
        })
        if(col.getLength() === 0) {
            pinegrow.showQuickError('The element doesn\'t have a parent.');
            return;
        }
        this.selectCollection(col);
    }

    this.selectFirstChild = function() {
        var col = new PgCollection();
        this.forEachElement(function(pgel) {
            var child = pgel.getChildAtPos(0, true);
            if(child) {
                col.add(child);
            }
        })
        if(col.getLength() === 0) {
            this.selectNext();
            //pinegrow.showQuickError('The element doesn\'t have children.');
            return;
        }
        this.selectCollection(col);
    }

    this.selectNext = function() {
        var col = new PgCollection();
        this.forEachElement(function(pgel) {
            var next = pgel.next();
            if(!next) {
                next = pgel.parent.first();
            }
            if(next) {
                col.add(next);
            }
        })
        if(col.getLength() === 0) {
            //pinegrow.showQuickError('The element doesn\'t have children.');
            return;
        }
        this.selectCollection(col);
    }

    this.selectPrevious = function() {
        var col = new PgCollection();
        this.forEachElement(function(pgel) {
            var next = pgel.prev();
            if(!next) {
                next = pgel.parent.last();
            }
            if(next) {
                col.add(next);
            }
        })
        if(col.getLength() === 0) {
            //pinegrow.showQuickError('The element doesn\'t have children.');
            return;
        }
        this.selectCollection(col);
    }

    var shift_direction = 'down';

    this.selectNextWithShift = function() {
        var col = this.getCollection().clone();
        col.orderByDocumentPosition();
        var last = shift_direction == 'down' ? col.getLast() : col.getFirst();
        //last = this.getLastSelected();
        if(last) {
            if(shift_direction == 'down') {
                var pgel = last.next();
                if (pgel) {
                    this.add(pgel);
                }
            } else {
                if (col.getLength() > 1) {
                    this.remove(last);
                } else {
                    shift_direction = 'down';
                    this.selectNextWithShift();
                }
            }
        }
    }

    this.selectPreviousWithShift = function() {
        var col = this.getCollection().clone();
        col.orderByDocumentPosition();
        var last = shift_direction == 'down' ? col.getLast() : col.getFirst();
        //last = this.getLastSelected();
        if(last) {
            if(shift_direction == 'down') {
                if (col.getLength() > 1) {
                    this.remove(last);
                } else {
                    shift_direction = 'up';
                    this.selectPreviousWithShift();
                }
            } else {
                var pgel = last.prev();
                if (pgel) {
                    this.add(pgel);
                }
            }
            /*
            var pgel = last.prev();
            if(pgel) {
                this.addOrRemove(pgel);
            }*/
        }
    }

    this.triggerElementSelectedEvent = function(pgel, info) {
        if(pgel) {
            $('body').trigger('crsa-element-selected', pgel, info);
            var page = pgel.getPage();
            page.callFrameworkHandler('on_element_selected', pgel, info);
            setChangedInfo(info);
            triggerChanged();
        }
    }

    this.setShowMenus = function(val) {
        show_menus = val;
        $.each(menus, function(key, m) {
            if(show_menus) {
                m.show();
            } else {
                m.hide();
            }
        })
    }

    var original_show_menus;
    var original_show_borders;

    this.temporaryHideMenus = function(borders) {
        var $body = $('body');
        original_show_menus = $body.hasClass('pg-show-sel-menu');
        $body.removeClass('pg-show-sel-menu');

        if(borders === true) {
            original_show_borders = !$body.hasClass('pg-show-sel-no-border');
            $body.addClass('pg-show-sel-no-border')
        }
        //this.setShowMenus(false);
    }

    this.restoreShowMenus = function(borders) {
        var $body = $('body');
        if(original_show_menus) {
            $body.addClass('pg-show-sel-menu')
        }
        if(borders === true) {
            if(original_show_borders) {
                $body.removeClass('pg-show-sel-no-border')
            }
        }
        //this.setShowMenus(original_show_menus);
    }

    this.selectIds = function(ids, info) {
        setChangedInfo(info);
        this.clear();
        if(ids) {
            ids.forEach(function(pgid) {
                var pgel = getPgNodeByPgId(pgid);
                if(pgel) {
                    _this.add(pgel);
                }
            })
            var last = this.getLastSelected();
            if(last) {
                this.triggerElementSelectedEvent(last, info);
            }
        }
    }

    this.selectCollection = function(col) {
        var ids = [];
        col.forEachElement(function(pgel) {
            ids.push(pgel.getId())
        })
        this.selectIds(ids);
    }

    this.addOrRemove = function(pgel) {
        if(collection.contains(pgel)) {
            this.remove(pgel);
        } else {
            return this.add(pgel);
        }
    }

    this.addRange = function(pgel) {
        var last = this.getLastSelected();
        if(!last) {
            this.add(pgel);
            return true;
        } else {
            //get el with same parent
            while(pgel.parent && pgel.parent !== last.parent) {
                pgel = pgel.parent;
            }
            if(last.parent !== pgel.parent || last === pgel) {
                return false;
            }
            var start = last.parent.getChildPos(last);
            var end = pgel.parent.getChildPos(pgel);
            var f = start > end ? -1 : 1;
            for(var i = start + f; i != end + f; i = i + f) {
                if(last.parent.children[i].isElementOrScript) {
                    this.add(last.parent.children[i]);
                }
            }
            return true;
        }
    }

    this.getLastSelected = function() {
        return collection.getLast();
    }

    this.getLastSelectedOnPage = function(page) {
        var last = null;
        collection.forEachElement(function(pgel) {
            if(pgel.getPage() === page) last = pgel;
        })
        return last;
    }

    this.getSelectedOnPage = function(page) {
        var r = [];
        collection.forEachElement(function(pgel) {
            if(pgel.getPage() === page) r.push(pgel);
        })
        return r;

    }

    this.remove = function(pgel, info) {
        if(collection.remove(pgel) >= 0) {
            setChangedInfo(info);

            var mid = getMenuId(pgel);

            menus[mid].destroy();

            delete menus[mid];

            hideAllButLastMenu();

            pgel.deselected();

            triggerChanged();
        }
    }

    this.removeForPage = function(page) {
        var c = collection.clone();
        c.forEachElement(function(pgel) {
            if(pgel.getPage() === page) {
                _this.remove(pgel);
            }
        })
    }

    this.areAllFromTheSamePage = function() {
        var p = null;
        var all = true;
        collection.forEachElement(function(pgel) {
            var page = pgel.getPage();
            if(p === null) {
                p = page;
            } else if(p !== page) {
                all = false;
            }
        })
        return all;
    }

    this.getElementRect = function(pgel) {
        var mid = getMenuId(pgel);
        if(menus[mid]) return menus[mid].getElementRect();
        return null;
    }

    this.replace = function(pgel, withel) {
        if(this.isSelectedElement(pgel)) {
            this.remove(pgel);
            this.add(withel);
        }
    }

    this.destroyMenusForPageView = function(pv) {
        $.each(menus, function(mid, m) {
            m.destroyForPageView(pv);
        })
    }

    this.getMenuPoints = function(pv) {
        var list = [];
        $.each(menus, function(mid, m) {
            list = list.concat(m.getMenuPoints(pv));
        })
        return list;
    }


    this.reselect = function() {
        var _this = this;
        //todo30 - could be faster maybe
        var list = collection.clone();
        this.clear();

        var dolist = [];

        list.forEachElement(function(pgel, i, total) {
            pgel = pgel.replacedWith || getPgNodeByPgId(pgel.getId());//maybe el was replaced by undo/redo
            if(pgel && !pgel.isDeleted) {
                dolist.push(pgel)
            }
        })

        if(dolist.length) {
            for(var i = 0; i < dolist.length; i++) {
                var pgel = dolist[i];
                if (i == dolist.length - 1) {
                    pinegrow.selectElement(pgel, null, true);
                } else {
                    _this.add(pgel);
                }
            }
        }
    }

    this.reselectDelayed = function(delay) {
        delayedTasks.run('reselect', function() {
            _this.reselect();
        }, delay || 1000);
    }

    this.getIds = function() {
        var r = [];
        collection.forEachElement(function(pgel) {
            var pgid = pgel.getId();
            if(pgid) {
                r.push(pgid);
            }
        })
        return r;
    }

    this.forEachElement = function(func) {
        collection.forEachElement(func);
    }

    this.forEachSubsequentElement = function(skip_pgel, func) {
        collection.forEachElement(function(a, b, c, d) {
            if(a != skip_pgel) {
                func(a, b, c, d);
            }
        })
    }

    this.forEachElementOrderedByDocumentPosition = function(func) {
        var c = collection.clone();
        c.orderByDocumentPosition();
        c.forEachElement(func);
    }

    this.forEachSubsequentElementOrderedByDocumentPosition = function(skip_pgel, func) {
        this.forEachElementOrderedByDocumentPosition(function(a, b, c, d) {
            if(a != skip_pgel) {
                func(a, b, c, d);
            }
        })
    }

    this.getPages = function() {
        return collection.getPages();
    }

    this.forEachPage = function(func) {
        this.getPages().forEach(func);
    }

    this.getMenu = function(pgel) {
        return menus[getMenuId(pgel)] || null;
    }

    this.repositionMenus = function(on_scroll) {
        $.each(menus, function(key, m) {
            m.reposition(on_scroll);
        })

        if(this.onRepositionMenus) this.onRepositionMenus();
    }

    this.repositionMenusDelayed = function() {
        delayedTasks.run('repositionMenus', function() {
            _this.repositionMenus();
        }, 200)
    }

    this.hasSelection = function() {
        return collection.getLength() > 0;
    }

    this.getLength = function() {
        return collection.getLength();
    }

    this.isInSelectedElement = function(pgel) {
        var list = collection.getList();
        for(var i = 0; i < list.length; i++) {
            if(pgel.isDescendantOf(list[i])) return list[i];
        }
        return null;
    }

    this.isDescendantSelected = function(pgel) {
        var list = collection.getList();
        for(var i = 0; i < list.length; i++) {
            if(list[i].isDescendantOf(pgel)) return list[i];
        }
        return null;
    }

    this.isSelectedElement = function(pgel) {
        return collection.contains(pgel);
    }

    this.isElementOrDescendantSelected = function(pgel) {
        return this.isSelectedElement(pgel) || this.isDescendantSelected(pgel);
    }

    this.isSelectedElementOrInSelectedElement = function(pgel) {
        return this.isSelectedElement(pgel) || this.isInSelectedElement(pgel);
    }

    this.isSelected$Element = function($el) {
        return this.isSelectedElement(getElementPgNode($el));
    }

    this.getRange = function() {
        var r = {first: null, count: 0}
        var last_idx = -1;
        var first_idx = -1;
        collection.forEachElement(function(pgel, i, total) {
            if(r.first && r.first.parent != pgel.parent) return null;
            var idx = pgel.parent.getChildPos(pgel);
            if(r.first == null || first_idx > idx) {
                r.first = pgel;
                first_idx = idx;
            }
            if(last_idx < 0 || last_idx < idx) {
                last_idx = idx;
            }
        })
        if(!r.first) return null;
        r.count = last_idx - first_idx + 1;
        r.all = [];
        for(var i = first_idx; i <= last_idx;i++) {
            r.all.push(r.first.parent.children[i]);
        }
        return r;
    }

    this.getName = function(opts) {
        return collection.getName(opts);
    }

    this.selectMany = function(c, clear) {
        if(clear) {
            this.clear();
        }
        c.forEachElement(function(pgel, i, total) {
            if(i == total - 1) {
                pinegrow.selectElement(pgel, null, true);
            } else {
                _this.add(pgel);
            }
        })
    }

    this.deselectMany = function(c) {
        c.forEachElement(function(pgel, i, total) {
            _this.remove(pgel);
        })
    }


    var hideAllButLastMenu = function() {
        collection.forEachElement(function(pgel, i, total) {
            var menus = _this.getMenu(pgel);
            if(menus) {
                if(i == total -1) {
                    menus.show();
                } else {
                    menus.hideMenu();
                }
            }
        })
    }

    var reselect_timer = null;

    pinegrow.addEventHandler('on_page_changes_done', function(page, info) {
        if(info && info.sourceType && info.sourceType == 'undo') {
            //undo will do its own reselect
            if(reselect_timer) {
                clearTimeout(reselect_timer);
                reselect_timer = null;
            }
            return;
        }

        setChangedInfo(info);

        var remove = [];
        collection.forEachElement(function(pgel) {
            if(pgel.isDeleted) {
                remove.push(pgel);
            }
        })
        var main = remove.indexOf(_this.getLastSelected()) >= 0;

        remove.forEach(function(pgel) {
            if(pgel.replacedWith) {
                _this.replace(pgel, pgel.replacedWith);
            } else {
                _this.remove(pgel);
            }
        })
        if(main) {
            if(reselect_timer) {
                clearTimeout(reselect_timer);
            }
            reselect_timer = setTimeout(function() {
                reselect_timer = null;
                var pgel = _this.getLastSelected();
                var page = pgel ? pgel.getPage() : null;
                $('body').trigger('crsa-element-selected', pgel, info);
                pinegrow.dispatchEvent('on_element_selected', page, pgel, info);
            }, 10)
        }
    });

}

var PgChangeManager = function() {

    var changed = new PgCollection();
    var will_change = new PgCollection();

    var _this = this;

    var get$Element = function(pgel, $el) {
        return $el || pgel.get$DOMElement();
    }

    var getPage = function(pgel, cp) {
        return cp || pgel.getPage();
    }

    this.willChange = function(pgel, action) {
        will_change.add(pgel);
        var cp = pinegrow.getCrsaPageOfPgParserNode(pgel);
        willMakeChange(cp.$iframe, action + " / " + getElementName(pgel.get$DOMElement(cp.get$Html())));
    }

    this.didChange = function(pgel_or_collection, action, info, page_arg, skip_defs_on_change) {
        info = info || {};
        info.replacedIds = {};
        info.action = action;

        if(info.skip_change_event) return;

        if(pgel_or_collection) {
            var col = pinegrow.getCollection(pgel_or_collection);
            if(in_batch) {
                batch_count++;
                batch_name = action;
                batch_collection.add(col);
                return;
            }

            pinegrow.dispatchEvent('on_before_changes_done', null, col, info);

            var emulated_pages = [];

            col.forEachElement(function(pgel) {
                if(pgel.replacedWith) {
                    var rdone = false;
                    var pgid = pgel.getId();
                    $.each(info.replacedIds, function(rid, wid) {
                        if(wid == pgid) {
                            info.replacedIds[rid] = pgel.replacedWith.getId();
                            rdone = true;
                            return false;
                        }
                    })
                    if(!rdone) {
                        info.replacedIds[pgel.getId()] = pgel.replacedWith.getId();
                    }
                    pgel = pgel.replacedWith;
                }
                var page = getPage(pgel, page_arg) || pgel.removedFromPage;

                if(page) {
                    if (page instanceof PgEmulatePageForUndo) {
                        if(emulated_pages.indexOf(page) < 0) {
                            emulated_pages.push(page);
                        }
                    } else if (page.openedInEditor) {
                        if (!skip_defs_on_change) {
                            callOnDefs(pgel, 'on_changed', pgel, page, null, null, page);
                        }
                        page.callFrameworkHandler('on_page_changed', pgel, info);
                    }
                }
            })

            emulated_pages.forEach(function(page) {
                pinegrow.dispatchEvent('on_emulated_page_changed', page, info);
            })

            var page_done = false;

            col.forEachPage(function(page, page_col) {
                if(page.openedInEditor) {
                    page.callFrameworkHandler('on_page_changes_done', info, page_col);//all changes to the page are done, in case of multiple selected elements.
                    page.setPageChanged(true);

                    if (page === page_arg) page_done = true;
                }
            })

            if(page_arg && !page_done && page_arg.openedInEditor) {
                page_arg.callFrameworkHandler('on_page_changes_done', info, col);
                page_arg.setPageChanged(true);
            }

        } else {
            if(!in_batch) {
                if(page_arg instanceof PgEmulatePageForUndo) {
                    pinegrow.dispatchEvent('on_emulated_page_changed', page_arg, info);
                } else if(page_arg.openedInEditor) {
                    page_arg.callFrameworkHandler('on_page_changed', null, info);
                    page_arg.callFrameworkHandler('on_page_changes_done', info);
                    page_arg.setPageChanged(true);
                }
            } else {
                batch_count++;
                batch_name = action;
            }
        }
    }

    this.didChangePage = function(page, action, info) {
        this.didChange(null, action, info, page);
    }

    this.didDelete = function(pgel_or_collection, page_arg) {
        if(pgel_or_collection) {
            pinegrow.getCollection(pgel_or_collection).forEachElement(function(pgel) {
                var page = page_arg || pgel.removedFromPage;
                if(page) {
                    _this.elementWasDeleted(pgel, page);
                }
            })
        }
        if(in_batch) {
            batch_collection.add(pgel_or_collection);
            batch_count++;
            batch_name = "Delete elements";
        } else {
            this.didChange(pgel_or_collection, "Delete elements", null, page_arg, true);
        }
    }

    this.didMove = function(pgel_or_collection, from_page, to_page) {
        if(pgel_or_collection) {
            pinegrow.getCollection(pgel_or_collection).forEachElement(function(pgel) {
                from_page = getPage(pgel, from_page);
                _this.elementWasMoved(pgel, from_page, to_page);
            })
        }
        if(in_batch) {
            batch_collection.add(pgel_or_collection);
            batch_count++;
            batch_name = "Move elements";
        } else {
            this.didChange(pgel_or_collection, "Move elements", null, from_page, true);
        }
    }

    this.didInsert = function(pgel_or_collection, orig_def, format_html) {
        if(pgel_or_collection) {
            pinegrow.getCollection(pgel_or_collection).forEachElement(function(pgel) {
                _this.elementWasInserted(pgel, orig_def, true);
                pgel.formatHtml = format_html; //MTCH TEST CODE FORMAT
            })
        }
        if(in_batch) {
            batch_collection.add(pgel_or_collection);
            batch_count++;
            batch_name = "Insert elements";
        } else {
            this.didChange(pgel_or_collection, "Insert elements", null, null, true);
        }
    }

    var in_batch = false;
    var batch_collection = null;
    var batch_count = 0;
    var batch_name;

    this.batchChanges = function(func, name) {
        //func();
        //return;
        in_batch = true;
        batch_count = 0;
        batch_collection = new PgCollection();
        func();
        in_batch = false;
        this.didChange(batch_collection, batch_count == 1 ? batch_name : name);
    }

    var callOnDefs = function(pgel, handler, a, b, c, d, cp) {
        return pgel.callDefHandlers(handler, a, b, c, d, cp)
    }

    var callOnParents = function(pgel, handler, a, b, c, d, cp) {
        pgel.callDefHandlersOnParents(handler, a, b, c, d, cp)
    }

    this.elementWasMoved = function(pgel, from_cp, to_cp) {
        removeEmptyPlaceholderFromParents(pgel);
        callOnDefs(pgel, 'on_moved', pgel, from_cp, to_cp); //todo30 - pgel
    }

    this.elementWasDeleted = function(pgel, cp) {
        cp = cp || pgel.getPage();
        if(!cp) return;
        callOnDefs(pgel, 'on_deleted', pgel, cp, null, null, cp);//todo30 - pgel
        cp.callFrameworkHandler('on_element_deleted', pgel);

        //Update CS list, not the best place to put it here btw
        if(pgel.tagName === 'link' && pgel.getAttribute('rel') === 'stylesheet') {
            cp.updateStylesheetsList();
        }
    }

    this.elementWasInserted = function(pgel, orig_def, skip_highlight) {
        var cp = pgel.getPage();
        if(!cp) return;

        var defs = callOnDefs(pgel, 'on_inserted', pgel, cp);

        /*
        if(orig_def && defs.indexOf(orig_def) < 0 && orig_def.on_inserted) {
            orig_def.on_inserted(pgel, cp);
        }
*/

        var def = defs.length ? defs[0] : null;

        if(orig_def && orig_def != def && defs.indexOf(orig_def) < 0) {
            if(orig_def.on_inserted) {
                orig_def.on_inserted(pgel, cp)
            }
        }

        cp.callFrameworkHandler('on_element_inserted', pgel, /* $el, */ def, defs);//todo30 - remove $el in handlers

        removeEmptyPlaceholderFromParents(pgel);

        callOnParents(pgel, 'on_child_inserted', -1, cp, pgel);

        if(orig_def && orig_def.empty_placeholder && !pgf.empty_ph_attribute) {
            if(pinegrow.getSetting('show-placeholders', '1') == '1') {
                showNotice('<p><b>Empty ' + pgel.tagName + ' </b>was just added to the page. Pinegrow added <b>pg-empty-placeholder</b> class to it, with <code>min-height:100px</code> so that you can see the element on the page. Once you add content to the element the class is removed. This does not affect how page looks like outside of Pinegrow - it\'s just a help to make editing easier.</p><p>You can enable or disable this behaviour in Support -&gt; Settings.</p>', 'A note about empty elements', 'empty-placeholder-2');
            } else {

            }
        }
        //scroll and highlight element
        if(!skip_highlight) {
            pgel.scrollTo();
            setTimeout(function () {
                pinegrow.highlightElement(pgel);//todo30
            }, 200);
        }

        //Update CS list, not the best place to put it here btw
        if(pgel.tagName === 'link' && pgel.getAttribute('rel') === 'stylesheet') {
            cp.updateStylesheetsList();
        }

        if(orig_def) {
            if(orig_def.resources && orig_def.resources.list.length) {
                var project = pinegrow.getCurrentProject();

                pinegrow.addFrameworkResourcesToPageOrProject(orig_def.framework, project || cp, {
                    resources: orig_def.resources,
                    only_add_to_page: cp,
                    overwrite: true
                })
                //set locked for css files
                var pi = pinegrow.getCurrentProjectInfo(cp);
                orig_def.resources.list.forEach(function(res) {
                    if(res.isCss() && res.relative_url) {
                        pi.setSettingForUrl(res.relative_url, 'locked', true);
                        pi.setSettingForUrl(res.relative_url, 'locked_reason', 'This stylesheet is a library resource and will be overwritten on update. Over-ride CSS rules in the your project\'s stylesheet.');
                    }
                })
                pi.save();
            }
            if(orig_def.required_plugins) {
                orig_def.required_plugins.forEach(function(pkey) {
                    var plugin = pinegrow.findFrameworkByKey(pkey);
                    if(plugin) {
                        pinegrow.activateFrameworkInCurrentContext(plugin, function() {
                            pinegrow.addFrameworkResourcesToPageOrProject(plugin, cp, {
                                done: function () {
                                    cp.refresh();
                                },
                                overwrite: false
                            })
                        }, cp);
                    }
                })
            }
        }
        if(pgf.smart_blocks) {
            var project = pinegrow.getCurrentProject();

            if(project) {
                if (project.smartStylesheet) {
                    if (orig_def && orig_def.add_style) {
                        project.smartStylesheet.addPart(orig_def.type, orig_def.add_style, pgel);
                        project.smartStylesheet.update();
                    }
                }

                if (project.smartScript) {
                    if (orig_def && orig_def.add_script) {
                        project.smartScript.addPart(orig_def.type, orig_def.add_script, pgel);
                        project.smartScript.update();
                    }
                }
            }
        }
    }

    /*
    this.done = function() {
        changed.forEachElement(function(pgel, $el, cp) {
            didMakeChange(cp.$iframe, $el, $el);
        })
        changed.normalize();
        pinegrow.updateTreeMultiple(changed.to$List());
    }
    */
}

var PgElementMenus = function(pgCurrent, element_idx) {

    var menus = [];

    var shown = true;

    var cp = pinegrow.getCrsaPageOfPgParserNode(pgCurrent);

    cp.getPageViews().forEach(function(pv) {
        menus.push(new PgElementPageViewMenu(pgCurrent, cp, pv, element_idx));
    })

    this.destroyForPageView = function(pv) {
        for(var i = 0; i < menus.length; i++) {
            if(menus[i].pageView === pv) {
                menus[i].destroy();
                menus.splice(i, 1);
                break;
            }
        }
    }

    this.getMenuPoints = function(pv) {
        var list = [];
        for(var i = 0; i < menus.length; i++) {
            if(menus[i].pageView === pv) {
                list = list.concat(menus[i].getPoints())
            }
        }
        return list;
    }

    this.getMenus = function() {
        return menus;
    }

    this.show = function() {
        if(!shown) {
            menus.forEach(function (m) {
                m.show();
            })
            shown = true;
        }
    }

    this.hide = function() {
        if(shown) {
            menus.forEach(function (m) {
                m.hide();
            })
            shown = false;
        }
    }

    this.hideMenu = function() {
        menus.forEach(function (m) {
            m.hideMenu();
        })
    }

    this.destroy = function() {
        menus.forEach(function(m) {
            m.destroy();
        })
        menus = null;
    }

    this.reposition = function(on_scroll) {
        menus.forEach(function(m) {
            m.reposition(on_scroll);
        })
    }

    this.getElementRect = function() {
        for(var i = 0; i < menus.length; i++) {
            var r = menus[i].getElementRect();
            if(r) return r;
        }
        return null;
    }
}


var PgElementPageViewMenuPoint = function(placement, menu, no_drag, text) {
    var _this = this;

    this.$element = $(`<div class="pg-sel-menu-point"><span>${text || ''}</span><div class="pg-sel-menu-point-label"></div></div>`);

    this.$label = this.$element.find('.pg-sel-menu-point-label');
    pg$Hide(this.$label);

    this.width = 20;
    this.height = 20;
    this.placement = placement;
    this.snapX = 1;
    this.snapY = 1;

    this.axis = null;

    switch(this.placement) {
        case 'left':
        case 'right':
            this.axis = 'x';
            break;
        case 'top':
        case 'bottom':
            this.axis = 'y';
            break;
    }
    this.in_drag = false;
    this.tooltip = null;

    if(this.axis == 'y') {
        this.$element.addClass('pg-sel-menu-point-y');
    }

    if(!no_drag) {

        var dragHelper = this.dragHelper = new PgDragHelper(this.$element);

        var ox, oy;

        dragHelper.on('dragStart', function (x, y, event) {

            this.in_drag = true;

            $.fn.crsapages('showOverlays');
            ox = parseInt(this.$element.css('left'));
            oy = parseInt(this.$element.css('top'));

            pinegrow.highlightElement(null);

            pinegrow.shouldHighlightElements = false;

            this.onDragStart.call(this, x, y);

        }, this);

        dragHelper.on('drag', function (x, y, event) {
            if (this.axis == 'x') y = 0;
            if (this.axis == 'y') x = 0;

            if (this.snapX > 1) {
                x = Math.floor(x / this.snapX) * this.snapX;
            }
            if (this.snapY > 1) {
                y = Math.floor(y / this.snapY) * this.snapY;
            }

            var o = this.onDrag.call(this, x, y);

            if (o) {
                x = o.x;
                y = o.y;
            }

            this.$element.css({
                left: (ox + x) + 'px',
                top: (oy + y) + 'px'
            })


        }, this);

        dragHelper.on('dragStop', function (x, y) {
            this.in_drag = false;
            pinegrow.shouldHighlightElements = true;
            $.fn.crsapages('showOverlays', true);
            this.onDragStop.call(this, x, y);
        }, this);
    }

    this.$element.on('click', function(e) {
        e.preventDefault();
        if(_this.onClick) _this.onClick.call(_this);
        return false;
    })

}

PgElementPageViewMenuPoint.prototype.add = function(menu, $container) {
    this.$element.appendTo($container);
    this.menu = menu;

    if(this.tooltip) {
        var _this = this;

        addTooltip(this.$element, null, {
            title: function() {
                return _this.tooltip;
            }
        });
    }
}

PgElementPageViewMenuPoint.prototype.setTooltipText = function(txt) {
    this.tooltip = txt;
}

PgElementPageViewMenuPoint.prototype.setText = function(txt) {
    this.$element.find('>span').html(txt);
    this.width = this.$element.outerWidth();
}

PgElementPageViewMenuPoint.prototype.setLabel = function(txt) {
    if(typeof txt == "number") txt = txt + '';
    if(txt !== null && txt !== undefined && txt.length) {
        this.$label.html(txt);
        pg$Show(this.$label);
        var w = this.$label.outerWidth();
        this.$label.css('left', (this.width / 2 - w/2) + 'px')
    } else {
        pg$Hide(this.$label);
    }
}

PgElementPageViewMenuPoint.prototype.hide = function(txt) {
    pg$Hide(this.$element);
}

PgElementPageViewMenuPoint.prototype.show = function(txt) {
    pg$Show(this.$element);
}

PgElementPageViewMenuPoint.prototype.setIsNotSet = function(val) {
    if(val) {
        this.$element.addClass('pg-sel-menu-point-not-set');
    } else {
        this.$element.removeClass('pg-sel-menu-point-not-set');
    }
}

PgElementPageViewMenuPoint.prototype.destroy = function() {
    pg$Hide(this.$element);
    this.$element.remove();
    if(this.dragHelper) {
        this.dragHelper.destroy();
    }
}

PgElementPageViewMenuPoint.prototype.onDragStart = function() {};
PgElementPageViewMenuPoint.prototype.onDrag = function() {};
PgElementPageViewMenuPoint.prototype.onDragStop = function() {};
PgElementPageViewMenuPoint.prototype.onReposition = function() {};
PgElementPageViewMenuPoint.prototype.onClick = function() {};


var PgElementPageViewMenuPointButton = function(menu, label, tooltip) {

    PgElementPageViewMenuPoint.call(this, 'menu' /* 'inside-top-left' */, menu);

    this.$element.addClass('pg-sel-menu-point-button');

    if(label.endsWith('</i>')) this.$element.addClass('just-icon');

    this.label = label;
    this.menu = menu;
    this.tooltip = tooltip;
    this.update();
}

extendObject(PgElementPageViewMenuPointButton, PgElementPageViewMenuPoint)

PgElementPageViewMenuPointButton.prototype.update = function(x, y) {
    this.setText(this.label);
}

PgElementPageViewMenuPointButton.prototype.onReposition = function() {
    var rect = this.menu.getElementRect();
    if(rect) {
        this.$element.css('max-width', rect.width - 16);
    }
};




var PgElementPageViewMenuPointForStyleProp = function(prop) {

    PgElementPageViewMenuPoint.call(this);

    this.value = null;
    this.prop = prop;
}

extendObject(PgElementPageViewMenuPointForStyleProp, PgElementPageViewMenuPoint)


PgElementPageViewMenuPointForStyleProp.prototype.onDragStart = function(x, y) {
    switch(this.prop) {
        case 'width':
            this.value = this.menu.$el.outerWidth();
            break;
    }
}

PgElementPageViewMenuPointForStyleProp.prototype.onDrag = function(x, y) {
    var _this = this;
    pinegrow.getCollectionForElement(this.menu.pgel).forEachElement(function(pgel) {
        pgel.setInlineStylePropertyValue(_this.prop, Math.round(_this.value + x * _this.menu.pageView.zoom) + 'px');
    })
    this.menu.reposition();
}

PgElementPageViewMenuPointForStyleProp.prototype.onDragStop = function(x, y) {

    pinegrow.changeManager.didChange(pinegrow.getCollectionForElement(this.menu.pgel), 'Resize with drag');
    pinegrow.selectedElements.reselect();
}







var PgElementPageViewMenu = function(pgCurrent, cp, pv, idx) {

    this.element_idx = idx || 0;

    var points = [];
    var show_points_in_menu = [];

    var selectedCrsaPage = cp;

    var positionTaskKey = pgFastdomTasks.getUniqueKey();

    //var current_left;
    //var current_top;

    this.pageView = pv;
    this.pgel = pgCurrent;

    var current = this.$el = pgCurrent.get$DOMElement(pv.get$Html());
    var currentNode = current ? this.$el.get(0) : null;
    var currentDocument = currentNode ? currentNode.ownerDocument : null;

    var getCurrent = function() {
        if(!current) return null;
        try {
            if (!current.get(0).ownerDocument.defaultView) {
                current = pgCurrent.get$DOMElement(pv.get$Html());
                if (!current || !current.get(0).ownerDocument.defaultView) return null;
            }
            return current;
        }
        catch(err) {
            return null;
        }
    }

    var $this = pv.get$Iframe();

    var _this = this;

    var positionObserver = new PgElementPositionObserver(function(force) {
        var c = getCurrent(force);
        return c ? c.get(0) : null;
    }, function(rect) {
        _this.reposition();
    })

    this.get$Element = function() {
        return $im;
    }

    this.getPoints = function() {
        return points;
    }

    var show = true;



    this.show = function() {
        $im.removeClass('hidden').show();
        showBorder();
        points.forEach(function(p) {
            p.show();
        })
        show = true;
    }

    this.hide = function() {
        $im.hide();
        hideBorder();
        points.forEach(function(p) {
            p.hide();
        })
        show = false;
    }

    this.hideMenu = function() {
        $im.addClass('hidden').hide();
    }

    var hideBorder = function() {
        if($border) {
            $border.each(function (i, e) {
                pgHide(e);
            })
        }
    }

    var showBorder = function() {
        if($border) {
            $border.each(function (i, e) {
                pgShow(e);
            })
        } else {
            _this.reposition();
        }
    }

    this.destroy = function() {
        positionObserver.destroy();

        pinegrow.removeEventHandler('on_page_views_updated', onPageViewChanged);
        pinegrow.removeEventHandler('on_page_changes_done', onPageViewChanged);
        pinegrow.removeEventHandler('on_page_shown_in_live_view', onPageViewChanged);

        pinegrow.removeEventHandler('on_css_node_deleted', onPageViewChanged);
        pinegrow.removeEventHandler('on_css_node_added', onPageViewChanged);
        pinegrow.removeEventHandler('on_css_node_updated', onPageViewChanged);

        onLinkUpdated && pinegrow.removeEventHandler('on_element_link_changed', onLinkUpdated);

        if(currentDocument) {
            currentDocument.removeEventListener("transitionend", onAnimationEnd);
            currentDocument.removeEventListener("animationend", onAnimationEnd);
            currentDocument.removeEventListener("load", onAnimationEnd, false);
        }

        if(currentNode) {
            currentNode.removeEventListener("load", onAnimationEnd, false);
        }


        document.addEventListener('mousemove', onMouseMove);

        if($border) {
            pg$Hide($border);
            //$border.remove();
            $border = null;
        }
        if(current) current.removeClass('crsa-selected');
        if($im) $im.stop().remove();
        $im = null;
        //if(timer) clearTimeout(timer);

        if(dragHelper) {
            dragHelper.destroy();
        }

        points.forEach(function(p) {
            p.destroy();
        })

        show_points_in_menu.forEach(function(p) {
            p.destroy();
        })

        points = [];

        if(highlightTask) {
            highlightTask.stop();
        }

        pgFastdomTasks.stopAll(positionTaskKey);

        this.destroyed = true;
    }

    var current_position = null;

    var onPageViewChanged = function() {
        if(pgCurrent.isDeleted) {
        } else {
            _this.reposition();
        }
    }

    pinegrow.addEventHandler('on_page_views_updated', onPageViewChanged);
    pinegrow.addEventHandler('on_page_changes_done', onPageViewChanged);
    pinegrow.addEventHandler('on_page_shown_in_live_view', onPageViewChanged);
    pinegrow.addEventHandler('on_css_node_deleted', onPageViewChanged);
    pinegrow.addEventHandler('on_css_node_added', onPageViewChanged);
    pinegrow.addEventHandler('on_css_node_updated', onPageViewChanged);

    var onAnimationEnd = function(e) {
        if(currentNode) {
            if (e.target == currentNode || e.target.contains(currentNode) || currentNode.contains(e.target)) {
                _this.reposition();
            }
        }
    }

    if(currentDocument) {
        currentDocument.addEventListener("transitionend", onAnimationEnd, false);
        currentDocument.addEventListener("animationend", onAnimationEnd, false);
    }

    if(currentNode) {
        currentNode.addEventListener("load", onAnimationEnd, false);
    }

    var border_rect = null;

    var onMouseMove = function(e) {
        if(border_rect) {
            crsaFuncs.showInsertionLine(current, 'inside-top');
        }
    }

    //document.addEventListener('mousemove', onMouseMove, true, false);

    var repositionElement = function(e, ox, oy) {
        var cur_pos = e.pgCurrentPosition || {x: 0, y: 0};
        pgPositionAt(e, cur_pos.x + ox, cur_pos.y + oy);
        /*
        var style = e.getAttribute('style');
        if (style) {
            var m = style.match(/left:\s*([0-9\.\-]*)/);
            var left = parseFloat(m ? m[1] : 0);
            m = style.match(/top:\s*([0-9\.\-]*)/);
            var top = parseFloat(m ? m[1] : 0);

            style = style.replace(/left:\s*[0-9\.\-]*dummy/, 'left:' + (left + ox));
            style = style.replace(/top:\s*[0-9\.\-]*dummy/, 'top:' + (top + oy));

            e.setAttribute('style', style);
        }*/
    }

    var repositionBorder = function(ox, oy /* offsets */) {
        if($border) {
            $border.each(function (i, e) {
                repositionElement(e, ox, oy)
            })
        }
        points.forEach(function(p) {
            if(!p.in_drag) {
                repositionElement(p.$element.get(0), ox, oy)
            }
        })
    }

    this.getElementRect = function() {
        if($border) {
            var $size = $border.filter('.crsa-hl-size-sel');
            if($size.length) {
                return $size.get(0).getBoundingClientRect();
            }
        }
        return null;
    }

    this.reposition = function(on_scroll) {
        var current = getCurrent();

        if(!current) return;

        crsaFuncs.positionElementMenu($im, current, $this, positionTaskKey, function(new_position) {

            if(on_scroll && current_position) {

                repositionBorder(new_position.left - current_position.left, new_position.top - current_position.top);


                 if($im) {
                 var im = $im.get(0);
                 if(timer) clearTimeout(timer);

                 if(!im.classList.contains('crsa-disable-hover')) {
                 im.classList.add('crsa-disable-hover')
                 }
                 timer = setTimeout(function(){
                 im.classList.remove('crsa-disable-hover')
                 },250);
                 }

            } else {
                if($border) {

                    $prev_border = $border;
                    $border = null;
                }
                if(show) {
                    highlightTask = highlightElementsInPageView(pv, current, '', null, 'pg-hl-select', '-sel', taskSuffix, onBorderCreated, null, null, true /* no name */, _this.element_idx);

                } else if($prev_border) {
                    $prev_border.remove();
                    $prev_border = null;
                }

                points.forEach(function(p) {
                    p.onReposition();
                })
            }
            current_position = new_position;

        });


    }

    if(current) {
        current.addClass('crsa-selected');
        var $border;
        var $prev_border;

        var onBorderCreated = function (el_list, show_point_func) {
            $border = $(el_list);
            if (!show) {
                hideBorder();
            }

            points.forEach(function(p) {
                show_point_func(p.$element, p.width, p.height, typeof p.placement === 'function' ? p.placement.call(p) : p.placement, p.in_drag)
            });

            /*
            if($prev_border) {
                var remove = [];
                $prev_border.each(function(i, el) {
                    if(el_list.indexOf(el) < 0) {
                        //not used anymore
                        remove.push(el);
                    }
                })
                $prev_border = null;

                if(remove.length) {
                    fastdom.mutate(function() {
                        remove.forEach(function(n) {
                            n.parentNode && n.parentNode.removeChild(n);
                        });
                    })
                }
            }

             */

            //border_rect = $border.offset();
        }

        var taskSuffix = 'Menu' + pgCurrent.getId();

        var b = pg$Closest($this, '.content');

        //points
        points = [];
        pinegrow.dispatchEvent('on_get_menu_points', pv.getPage(), pgCurrent, this, points);

        pinegrow.dispatchEvent('on_get_menu_points_with_callback', pv.getPage(), pgCurrent, this, function(point_list) {
            if(point_list.length && !_this.destroyed) {
                point_list.forEach(function (p) {
                    p.add(_this, b);
                    points.push(p);
                })
                _this.reposition();
            }
        });
          ///

        show_points_in_menu = [];

        points.forEach(function(p) {
            if(p.placement === 'menu') {
                show_points_in_menu.push(p);
            } else {
                p.add(_this, b);
            }
        })

        var highlightTask = highlightElementsInPageView(pv, current, '', null, 'pg-hl-select', '-sel', taskSuffix, onBorderCreated, null, null, /* no name */ true, _this.element_idx);

        var dyn = pgCurrent.dynamic;


        var simple = pgf.simple_element_menu || false;

        var menu_cmds = [];
        var menu_options = {
            simple: simple
        }

        pv.getPage().callFrameworkHandler('on_get_menu_commands', pgCurrent, menu_cmds, this.$el, menu_options);

        pgCurrent.callDefHandlers('on_get_menu_commands', pgCurrent, menu_cmds, this.$el);

        simple = menu_options.simple;


        var $im = $('<div/>', {'class': `crsa-inline-menu ${dyn ? 'dyn' : ''} ${simple ? ' crsa-inline-menu-simple' : ''}`}).appendTo(b);

        if(pgf.edit_html) {

            if(dyn) {
                var $dyn_info = $(`<span class="dyn-info">DYN</span>`).appendTo($im);
                addTooltip($dyn_info, 'This is a dynamic element created by Javascript code. You can style it with the Style panel. Note that inline styles can\'t be saved. Click on DYN for more information.');

                $dyn_info.on('click', function(e) {
                    e.preventDefault();
                    pinegrow.openExternalUrl(pgf.docs.dyn_elements_help);
                })
            } else {
                var $b_move = $('<i class="icon icon-Pin6 crsa-inline-menu-move"></i>').appendTo($im);

                var $updown = $(`<div class="crsa-inline-menu-updown"></div>`).appendTo($im);

                addTooltip($('<i class="icon icon-up crsa-inline-menu-move-up"></i>').appendTo($updown)
                    .on('click', function (e) {
                        e.preventDefault();
                        var api = new PgApi();
                        api.moveUp();
                    }), 'Move the element up (before the preceding element).');

                addTooltip($('<i class="icon icon-down crsa-inline-menu-move-down"></i>').appendTo($updown)
                    .on('click', function (e) {
                        e.preventDefault();
                        var api = new PgApi();
                        api.moveDown();
                    }), 'Move the element down (after the next element)');

                if(!pgCurrent.canMove()) {
                    $updown.css('visibility', 'hidden');
                }
            }

            //$('<i/>', {'class' : 'fa fa-trash-o'}).appendTo($im);

            if(!dyn) {

                var addProps = function() {
                    var $b_props = $('<i/>', {'class': 'icon icon-settings crsa-inline-menu-prop'}).appendTo($im);
                    $b_props.on('click', function (e) {
                        e.preventDefault();
                        var q = PgQuickProperties(null, pgCurrent, current);
                    });
                    addTooltip($b_props, 'Show element properties ' + pgShowKbdCombo('P'));
                }

                if(simple) {
                    var $b_name = $(`<span class="crsa-inline-menu-name">${pinegrow.selectedElements.getName()}</span>`).appendTo($im);


                }

                var $b_text = $('<i/>', {'class': 'icon icon-text_edit'}).appendTo($im);
                $b_text.on('click', function (e) {
                    e.preventDefault();
                    pinegrow.inlineEditManager.edit(current, true, false, e);
                });

                addTooltip($b_text, 'Edit text');

                if(!isApp()) {

                    var $b_click = $(`<a href="#" target="_blank" style="text-decoration:none;"><i class="icon icon-touch"></i></a>`).appendTo($im);
                    addTooltip($b_click, 'Click the link.');

                    $b_click.on('click', function (e) {
                        var href = $b_click.attr('href');

                        if (!crsaIsAbsoluteUrl(href) && !href.match(/^[a-z]+\:/)) {
                            e.preventDefault();
                            var p = pinegrow.getCurrentProject();
                            var url = p.getAbsoluteUrl(href);
                            var page = pinegrow.getCrsaPageByUrl(url);

                            var file = p.getFileForUrl(url);

                            if (page || file) {
                                pinegrow.openOrShowPage(url);
                                //pinegrow.setSelectedPage(page);
                            } else {
                                pinegrow.showQuickError(`Page ${href} not found.`);
                            }
                        }
                    });

                    var updateLink = function(pgCurrent) {
                        var link_pgel = pgCurrent.getLinkElement() || pgCurrent.closest('a');

                        if(link_pgel) {
                            var href = link_pgel.getAttr('href');
                            $b_click.attr('href', href);
                            if(href) {
                                pg$Show($b_click, 'inline');
                                return;
                            }
                        }
                        pg$Hide($b_click);
                    }

                    updateLink(pgCurrent);

                    var onLinkUpdated = function() {
                        updateLink(pgCurrent);
                    }

                    pinegrow.addEventHandler('on_element_link_changed', onLinkUpdated);
                }

                if(!simple) {
                    var $b_props = $('<i/>', {'class': 'icon icon-settings crsa-inline-menu-prop'}).appendTo($im);
                    $b_props.on('click', function (e) {
                        e.preventDefault();
                        var q = PgQuickProperties(null, pgCurrent, current);
                    });
                    addTooltip($b_props, 'Show element properties ' + pgShowKbdCombo('P'));
                }
            }

            if(!simple) {
                var $b_style = $('<i/>', {'class': 'icon icon-Pin15 crsa-inline-menu-style'}).appendTo($im);
                $b_style.on('click', function (e) {
                    e.preventDefault();
                    pinegrow.showCreateCssRule();
                });
                addTooltip($b_style, 'Style with CSS ' + pgShowKbdCombo('R'));
            }

            if (!dyn && pgCurrent.canDuplicate()) {
                var $b_duplicate = $('<i/>', {'class': 'icon icon-duplicate crsa-inline-menu-duplicate'}).appendTo($im);
                $b_duplicate.on('click', function (e) {
                    e.preventDefault();
                    crsaFuncs.duplicateCurrnetElement(/*current - do all selected*/null, e.altKey);
                });

                if (pinegrow.selectedElements.getLength() > 1) {
                    addTooltip($b_duplicate, 'Duplicate ' + pgShowKbdCombo('CMD+D') + '<br>Use ' + pgShowKbdCombo('ALT') + ' to duplicate as group');
                } else {
                    addTooltip($b_duplicate, 'Duplicate ' + pgShowKbdCombo('CMD+D'));
                }
            }

            if(!dyn && !simple) {
                var $b_add = $('<i/>', {'class': 'icon icon-import crsa-inline-menu-add'}).appendTo($im);
                $b_add.on('click', function (e) {
                    e.preventDefault();
                    pinegrow.showQuickInsert(pgCurrent, current);
                });

                addTooltip($b_add, 'Insert element ' + pgShowKbdCombo('PLUS'));
            }
        } else {
            $im.css('padding-left', '10px');
        }



        menu_cmds.forEach(function(cmd) {
            if(cmd.icon || cmd.html) {
                var $b_cmd = $(cmd.html || `<i class="icon ${cmd.icon}"></i>`).appendTo($im);
                $b_cmd.on('click', function (e) {
                    e.preventDefault();
                    cmd.action(pgCurrent, $(this));
                    return false;
                });

                addTooltip($b_cmd, cmd.tooltip);
            }
        })

        if(pgf.edit_html && !dyn && ['html','head','body'].indexOf(pgCurrent.tagName) < 0) {
            var $b_delete = $('<i/>', {'class': 'icon icon-bin crsa-inline-menu-delete'}).appendTo($im);
            $b_delete.on('click', function (e) {
                e.preventDefault();
                crsaFuncs.deleteCurrentElement(/* current - all selected */);
            });

            addTooltip($b_delete, 'Delete ' + pgShowKbdCombo('DELETE'));
        }

        if(pgf.empty_ph_attribute && pgCurrent.canHaveChildren()) {
            var has_empty_attr = pgCurrent.hasAttribute(pgf.empty_ph_attribute) || pgCurrent.hasClass('pg-empty-placeholder');
            if(has_empty_attr || true /* pgCurrent.isEmpty() */) {
                var $b_empty = $('<i/>', {'class': 'pg-el-menu-placeholder icon icon-Pin21' + (has_empty_attr ? ' active' : '')}).appendTo($im);
                $b_empty.on('click', function (e) {
                    e.preventDefault();

                    var api = new PgApi();

                    pinegrow.makeChanges(pgCurrent, 'Toggle empty placeholder', function() {
                        pinegrow.getCollection().forEachElement(function(pgCurrent) {
                            if(pgCurrent.hasAttribute(pgf.empty_ph_attribute) || pgCurrent.hasClass('pg-empty-placeholder')) {
                                pgCurrent.removeAttribute(pgf.empty_ph_attribute);
                                pgCurrent.removeClass('pg-empty-placeholder');
                                pgCurrent.placeholderWasAutoRemoved = false;
                            } else {
                                pgCurrent.setAttribute(pgf.empty_ph_attribute /*, 'always' */);
                            }
                        })
                    })

                    pinegrow.selectedElements.reselect();
                });

                addTooltip($b_empty, has_empty_attr ? 'Switch off the empty placeholder' : 'Turn on the empty placeholder');
            }
        }

        show_points_in_menu.forEach(function(p) {
            $im.append(p.$element);
            if(p.tooltip) {
                addTooltip(p.$element, p.tooltip);
            }
        })

        var buildActionsMenu = function () {
            var action_menu = new PgDropdownButton(null, function (e) {
                e.preventDefault();

                var def_actions = $.fn.crsa('getActionsMenuFor', pgCurrent, true /* for all selected elements */);

                if(pgf.kids) {
                    def_actions = def_actions.filter(function(a) {
                        if(a.label == 'Transform' && pgCurrent.canDelete()) {
                            a.label = 'Change to';

                            a.submenu = a.submenu.filter(function(m) {
                                return (m.group && m.group == 'change_to') ? true : false;
                            })

                            return true;
                        }
                        return false;
                    })
                }
                if(menu_cmds.length) {
                    var idx = 0;
                    var header = null;
                    menu_cmds.forEach(function(cmd) {
                        if(cmd.header && cmd.header != header) {
                            def_actions.splice(idx++, 0, {
                                type: 'header',
                                label: cmd.header
                            })
                            header = cmd.header;
                        }
                        if(cmd.label) {
                            def_actions.splice(idx++, 0, cmd);
                        }
                    })
                }

                !pgf.kids && selectedCrsaPage.callFrameworkHandler('on_build_actions_menu', def_actions, pgCurrent, true);

                var menu = new CrsaContextMenu(def_actions, action_menu.$element);
                menu.targetPgel = pgCurrent;

                menu.showForElement(action_menu.$element, action_menu.$element.closest('.page'));

            });

            if(!pgf.kids || pgCurrent.canDelete()) {
                action_menu.$element.appendTo($im);
                action_menu.setTooltip('More actions');
            }
        }

        //if(!pgf.kids) {
            buildActionsMenu();
        //}

        var $start_iframe = null;


        if (current && current.is('body,html,head,script') && $b_move) $b_move.hide();

        var move_copy = false;
        var current_style = null;

        if($b_move) {

            if (!pgCurrent) {
                $b_move
                    .on('mousedown', function (e) {

                        var problems = new pgParserSourceProblem(pgCurrent, current);

                        if (!pgCurrent) {
                            problems.add('element', getElementName(current), 'change');
                        }
                        if (!problems.ok()) {
                            showAlert(problems.toString(), "Can't move this element");
                            return;
                        }
                        e.preventDefault();
                        e.stopImmediatePropagation();
                    })
            }
        }

        if (crsaIsInEdit()) {
            if($b_move) {
                $b_move
                    .on('mousedown', function (e) {
                        if (crsaIsInEdit()) {
                            showAlert("Go out of content editing mode before moving elements.", "Can't move during content editing")
                            e.preventDefault();
                            e.stopImmediatePropagation();
                        }
                        if (!canMakeChange(pgCurrent, 'move_element')) {
                            e.preventDefault();
                            e.stopImmediatePropagation();
                        }

                    })
            }
        } else {
            var draggedMenu;
            var draggableDone = false;

            var dragHelper = new PgDragHelper($im, '.crsa-inline-menu-move');

            dragHelper.onBefore = function($el) {
                if(!pgf.edit_html) return false;
                return true;
            }

            dragHelper.on( 'dragStart', function(x, y, event) {
                $.fn.crsapages('showOverlays');

                var cp = getCrsaPageOfElement(current);
                draggedMenu = new PgDragMenu(cp.getMainType(null, pgCurrent), true /* use clone */, true /* show menu */, null, pgCurrent);

                draggedMenu.pgel = pgCurrent;
                draggedMenu.movedPgel = pgCurrent;

                crsaFuncs.setDraggedMenu(draggedMenu);
                draggedMenu.handleMoveEvent(event, draggedMenu);

                $im.hide();

            }, this );

            dragHelper.on( 'drag', function(x, y, event) {

                draggedMenu.handleMoveEvent(event, draggedMenu);
            }, this );

            dragHelper.on( 'dragStop', function(x, y) {
                $.fn.crsapages('showOverlays', true);

                if($im) $im.show();

                if (!draggedMenu) return; //was already handled by the tree
                if (draggedMenu.insert()) {
                } else {
                }

                draggedMenu.destroy();
                draggedMenu = null;

                crsaFuncs.setDraggedMenu(null);

                crsaFuncs.showInsertionLine(null);
                pinegrow.highlightElement(null);

            }, this );

            crsaFuncs.positionElementMenu($im, current, $this, positionTaskKey, function(new_position) {
                current_position = new_position;
            });
            var inlineMenu = $im;


            var timer;
            inlineMenu
                .on('mousewheel', function (e) {

                    var $ul = $(e.target.closest('ul'));
                    if (!$ul.hasClass('dropdown-menu')) {

                        var im = inlineMenu.get(0);

                        if (!im.classList.contains('crsa-disable-hover')) {
                            fastdom.mutate(function() {
                                im && im.classList.add('crsa-disable-hover')
                            })
                        }
                        timer && clearTimeout(timer);

                        timer = setTimeout(function () {
                            fastdom.mutate(function() {
                                im && im.classList.remove('crsa-disable-hover')
                            })
                        }, 250);
                    }

                })

            if (current.is('html')) $im.hide();

        }

        positionObserver.check();
    }

}

var PgQuickSelectElement = function(options) {
    var v = new PgSelectElementView(options)
    var win = new PgQuickWindow(options.title, v, 'quickSelectElement', 400, 250);
    v.quickWin = win;
    win.showFor$El(options.target);
    win.autoSize(80);

    win.$element.addClass('important');
}

var PgSelectElementView = function(options) {

    var defs = {
        label: 'Save changes',
        current_col: null,
        remove_label: 'remove',
        select: true,
        on_changed: null,
        remove: true
    }

    options = $.extend(defs, options || {})

    var _this = this;

    var $element = $('<div class="pg-select-element"><div class="pg-select-element-description"></div><div class="pg-select-element-help">Click on an element on the page or in the tree. ' + pgShowKbd('CMD') + ' + Click to select multiple. Click on the listed element to reveal it on the page.</div><ul class="pg-select-element-list"></ul><button class="pg-button pg-button-primary done">' + options.label + '</button> <button class="pg-button cancel ">Cancel</button></div>');

    var $desc = $element.find('.pg-select-element-description');
    var $list = $element.find('.pg-select-element-list');
    var $button = $element.find('button.done');

    //pg$Hide($button)

    $button.on('click', function(e) {
        e.preventDefault();
        col.forEachElement(function(pgel) {
            removed_col.remove(pgel)
        })

        if(options.func) {
            options.func(col, removed_col);
        }

        if(_this.quickWin) {
            _this.quickWin.destroy();
        }
        if(options.reselect) {
            pinegrow.selectedElements.reselect();
        }
    })

    $element.find('button.cancel').on('click', function(e) {
        e.preventDefault();

        if(_this.quickWin) {
            _this.quickWin.destroy();
        }
        if(options.reselect) {
            pinegrow.selectedElements.reselect();
        }
    })

    $list.on('click', '[data-pg-id]', function(e) {
        e.preventDefault();
        var pgid = $(this).attr('data-pg-id');
        var pgel = getPgNodeByPgId(pgid);
        if(pgel) {
            pgel.scrollTo();
            setTimeout(function() {
                pinegrow.highlightElement(pgel.get$DOMElement())
            }, 100)
        }
    })

    $list.on('click', '.remove', function(e) {
        e.preventDefault();
        var pgid = $(this.closest('li')).attr('data-pg-id');
        var pgel = getPgNodeByPgId(pgid);
        if(pgel) {
            removed_col.add(pgel);
            existing_col.remove(pgel);
            col.remove(pgel);
            update(col);
            onChanged();
        }
    })

    $desc.html(options.description)

    var view = this.view = new PgUIView($element);

    var col = new PgCollection();
    var existing_col = new PgCollection();
    var removed_col = new PgCollection();

    var did_select = false;

    var onSelected = function() {
        col = existing_col.clone();
        col.add(pinegrow.selectedElements.getCollection());

        update(col);
        onChanged();
        did_select = true;
    }

    var update = function(col) {
        if(col && col.getLength() > 0) {
            var html = '';
            col.forEachElement(function(pgel) {
                html += '<li data-pg-id="' + pgel.getId() + '"><span class="name">' + pgel.getName({short: false}) + '</span>';
                if(options.remove === 'all' || (existing_col.contains(pgel) && options.remove)) {
                    html += '<span class="remove"><i class="remove icon icon-close"></i>' + options.remove_label + '</span>';
                }
                html += '</li>';
            })
            $list.html(html);
            //pg$Show($button, 'inline-block');
        } else {
            $list.html('<li>No element is selected.</li>');
            //pg$Hide($button);
        }
        if(_this.quickWin) {
            _this.quickWin.autoSize(80);
        }
    }

    var onChanged = function() {
        if(options.on_changed) {
            options.on_changed(col);
        }
    }

    if(options.current_col) {
        existing_col = col = typeof options.current_col == 'function' ? options.current_col() : options.current_col;
        if(!existing_col) existing_col = new PgCollection();
        update(existing_col);
    }

    var onElementClicked = function(crsaPage, pgel, $el, status) {
        status.stop = true;

        if(status.multi) {
            if(col.contains(pgel)) {
                col.remove(pgel);
            } else {
                col.add(pgel);
            }
        } else {
            col = new PgCollection();
            col.add(pgel);
        }
        update(col);
        onChanged();

        if(options.updated_message) {
            pinegrow.showQuickMessage(options.updated_message);
        }
    }

    if(!options.select) {
        pinegrow.addEventHandler('on_element_clicked', onElementClicked);
    }

    pinegrow.addEventHandler('on_selected_elements_changed', onSelected);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_selected_elements_changed', onSelected);

        if(!options.select) {
            pinegrow.removeEventHandler('on_element_clicked', onElementClicked);
        }
    }
}

//disabled for now
class PgElementPositionObserver {
    constructor(dom_node, on_change) {
        this.element = dom_node;
        this.on_change = on_change;
        this.rect = {x: 0, y: 0, width: 0, height: 0};
        //this.timer = setInterval(this.check.bind(this), 20);
        this.force_resolve_element = false;
    }

    check() {
        return;
        const _this = this;
        fastdom.measure(function() {
            const el = typeof _this.element === 'function' ? _this.element(_this.force_resolve_element) : _this.element;
            _this.force_resolve_element = false;
            if(el) {
                const rect = el.getBoundingClientRect();
                if (_this.rect) {
                    if (_this.rect.x !== rect.x || _this.rect.y !== rect.y || _this.rect.width !== rect.width || _this.rect.height !== rect.height) {
                        _this.rect.x = rect.x;
                        _this.rect.y = rect.y;
                        _this.rect.width = rect.width;
                        _this.rect.height = rect.height;

                        _this.on_change(rect);
                    } else if(rect.width === 0 && rect.height === 0) {
                        _this.force_resolve_element = true;
                    }
                } else {
                    _this.rect = rect;
                }
            }
        })
    }

    destroy() {
        clearInterval(this.timer);
    }
}