var PgCSSComposerView = function() {

    var active = true;

    var $element = $('<div/>', {id: 'pg-tree', class: 'pg-tree pg-css-composer'});

    var api = new PgApi();

    var search = new PgSearchBox({
        toolbar_on: true
    });

    var toolbar = this.toolbar = search.getToolbar();

    var stylesButton = new PgButtonToolbarItem('icon-filter', 'Select what kind of elements to show in the tree.');

    toolbar.addItem('settings', stylesButton);

    var filters = {
        position: {
            label: 'Position',
            desc: 'Show only the elements with inline styles',
            re: /_(top|right|bottom|left|position|z\-index)_/i
        },
        text: {
            label: 'Text',
            desc: 'Show only images',
            re: /font|line-height|_color_|letter-spacing|text-decoration|text-align|text-transform|white-space|word-break|_direction_|text-indent/i
        }
    }

    var default_filter_options = {};

    $.each(filters, function(key, name) {
        default_filter_options[key] = false;
    })

    var filter_options = default_filter_options;

    stylesButton.onClick = function () {

        var menu = new CrsaContextMenu(null, this.$element);

        menu.add({
            type: 'header',
            label: 'Show only:'
        });

        $.each(filters, function(key, info) {
            menu.add({
                label: info.label,
                action: function() {
                    var on = !(filter_options[key] === true);
                    _this.setFilterOption(on ? key : null);
                },
                check: filter_options[key] === true
            })
        })

        menu.add({
            type: 'divider'
        });

        menu.add({
            label: 'Clear filter',
            action: function() {
                _this.setFilterOption(null);
            }
        })

        menu.showForElement(this.$element, null, 8);
    }

    this.setFilterOption = function(key, skip_update) {
        filter_options = $.extend({}, default_filter_options);
        if(key) {
            var info = filters[key];
            filter_options[key] = true;
            stylesButton.setActive(true);
            pinegrow.showMode(info.desc, true);
            setFilterMessage(info.label);
        } else {
            stylesButton.setActive(false);
            setFilterMessage();
        }
        cs_tree.setFilterOptions(filter_options);
        if(!skip_update) {
            cs_tree.research(true);
        }
    }

    search.$element.appendTo($element);

    var $selects = search.$element.find('.pg-button-toolbar-section.icons');

    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    search.onSearch = function(filter) {
        var p = new CrsaProfile();
        //if(cs_tree) cs_tree.search(filter);

        if(filter.length) {
            pg$Show($selects, 'inline-block');
        } else {
            pg$Hide($selects);
        }

        p.show('SEARCH');
    }

    pg$Hide($selects);

    var $filterMsg = $('<div class="pg-tree-filter-msg"><div>Show only: <span></span><i class="close icon icon-close"></i></div></div>').appendTo($element);

    pg$Hide($filterMsg);

    $filterMsg.on('click', function(e) {
        e.preventDefault();
        _this.setFilterOption(null);
    })

    var setFilterMessage = function(msg) {
        if(msg) {
            $filterMsg.find('span').html(msg);
            pg$Show($filterMsg);
            $element.addClass('with-filter-msg');
        } else {
            pg$Hide($filterMsg);
            $element.removeClass('with-filter-msg');
        }
    }

    var $panel = $('<div/>', {class: 'pg-css-panel'}).appendTo($element);

    var composerPanel = pgComposerManager.getInstance('PgComposerCSSPanel');
    $panel.append(composerPanel.element);

    var view = this.view = new PgUIView($element, '', 'CSS Composer');

    var view_prefix = `pg_ct${view.getId()}_`;

    var flag_overriden = view_prefix + 'ov';
    var flag_inherited = view_prefix + 'inh';
    var flag_not_inherited = view_prefix + 'ninh';

    view.classForContainer = 'pg-tree-panel-container';

    var _this = this;

    view.tabIcon = 'icon-Pin11';

    var currentPage;

    var activeControls = new PgCSSTreeActiveCommandsView();
    $element.prepend(activeControls.view.get$Element());

    activeControls.onStateChanged = function() {
        onElementSelected();
    }

    view.onShow = function() {
        //scroll to selected element if not visible
        return;
    }


    var dimension;


    var scrubValue = null;
    var colorValue = null;
    var editValue = null;
    var countHandler = null;

    var get$ValueElementValue = function($num) {
        var $value = $num.closest('value,name');
        var text = '';

        var node = $value.get(0);

        for (var i = 0; i < node.childNodes.length; i++) {
            var cn = node.childNodes[i];
            if(cn.nodeType === 3) {
                text += cn.textContent;
            } else if(cn.nodeType === 1) {
                if(cn.tagName === 'COLOR') {
                    text += $(cn).find('> span').text();
                } else {
                    text += cn.textContent;
                }
            }
        }
        return text;
    }


    var changeInlineStyleProperty = function(rule, event) {
        var info = {}
        if(!event) info.skip_change_event = true;
        api.setInlineStyle(pinegrow.getCollection(), rule, info);
    }

    view.on('crsa-rule-selected', function(event, obj) {
        view.whenVisible(function() {
            if(obj) {

                //showCssGui(obj.rule, obj.options);
                //selectedRule = obj.rule;
            }
        })
    });

    view.on('crsa-rules-changed', function(event, info) {
        onElementSelected();
        onPageStylesheetsLoaded(currentPage);
    });

    var onNodeSelectedDeselected = function(pagenull, nodes, info) {

    }

    pinegrow.addEventHandler('on_css_node_selected_deselected', onNodeSelectedDeselected);

    var onNodeChanged = function(pagenull, obj) {
        if(obj.edited_by !== _this) {
            var node = obj.node;

        }
    }

    var onNodeDeleted = function(pagenull, obj) {
        if(obj.edited_by !== _this) {
            onElementSelected();
            onPageStylesheetsLoaded(currentPage);
        }
    }

    var onNodeAdded = function(pagenull, obj) {
        if(obj.edited_by !== _this) {
            onElementSelected();
            onPageStylesheetsLoaded(currentPage);
        }
    }

    pinegrow.addEventHandler('on_css_node_deleted', onNodeDeleted);
    pinegrow.addEventHandler('on_css_node_updated', onNodeChanged);
    pinegrow.addEventHandler('on_css_node_added', onNodeAdded);

    var wireTreeElements = function() {

        var onChangeValue = function($num, value, is_done) {
            if(value === undefined) value = get$ValueElementValue($num);
            var event = is_done || false;

            var node = getCSSNodeForTree$Element($num);
            if(node) {
                var rule = node.parent;

                if(rule.inline) {
                    if(rule.inline === 'source') {
                        node.value = value;
                        changeInlineStyleProperty(rule, event)
                    }
                } else {
                    var cs = pgcssh.getStylesheetForNode(node);
                    cs.ruleValueChanged(rule, node.prop, value, function () {
                        //on done
                    }, false, _this /* edited by */, event, null, null)
                }
            }
        }

        var onChangeProp = function($name) {

        }

        //var trigger

        countHandler = new PgCSSCountHandler(tableGrid.$container);

        countHandler.onGetSelector = function($count) {
            var rule = getCSSNodeForTree$Element($count);
            return rule ? rule.selector : null;
        }

        countHandler.onGetElements = function(selector) {
            try {
                return currentPage.getDocument().querySelectorAll(selector);
            } catch(err) {
                return []
            }
        }

        scrubValue = new PgCSSEditableValue(tableGrid.$container, 'num');

        scrubValue.onChange = function($num, val, is_done) {
            onChangeValue($num, undefined, is_done);
        }

        colorValue = new PgCSSEditableColor(tableGrid.$container, 'colorbox');
        colorValue.onChange = function($color) {
            onChangeValue($color);
        }

        editValue = new PgCSSEditableText(tableGrid.$container, 'name,value');
        editValue.onGetValue = function($el) {
            return $el.text();
        }
        editValue.onChange = function($name, is_done, value) {
            var data = getDataForTree$Element($name)
            var node = data ? data.node : null;
            if(node) {
                var cs = pgcssh.getStylesheetForNode(node);

                if($name.get(0).tagName === 'NAME') {

                    //var pos = pgcssh.getNodePositionInSource(node);
                    //console.log('position in source', pos);

                    switch (node.type) {
                        case 'rule':
                            node.selector = value;
                            cs.updateSingleRule(node);
                            break;
                        case 'decl':
                            node.prop = value;
                            cs.ruleValueChanged(node.parent, node.prop, node.value, function () {
                                //on done
                            }, false, 'csstree', false, null, null)
                            break;
                    }
                } else {
                    //Value
                    switch (node.type) {
                        case 'rule':
                            break;
                        case 'decl':
                            onChangeValue($name, value);
                            break;
                    }
                }
            }
            if(is_done) {
                refreshDataItem(data)
            }
        }
        editValue.onGetOptions = function($field) {
            if($field.get(0).tagName === 'NAME') {
                return function() {
                    return pinegrow.devTools.getCSSPropertiesOptions();
                }
            }
            return null;
        }
        //return;

        var getSubmenu = function(list, add_list) {
            var show_desc = pinegrow.getSetting('css-menu-show-desc', '1') === '1';

            return function() {
                var r = add_list ? add_list.slice(0) : [];
                list.forEach(function(item) {
                    if(item.name === '-') {
                        var d = {
                            type: 'divider'
                        }
                    } else if(item.name === 'pgmore') {
                        var d = {
                            label: 'More...',
                            action: function() {}
                        }
                        if (item.sub && item.sub.length) {
                            d.submenu = getSubmenu(item.sub);
                        }
                    } else {
                        var d = {
                            class: 'pg-css-menu-prop',
                            label: '<i class="icon icon-minus"></i>' + '<code>' + item.name + '</code>',
                            action: function () {
                            }
                        }
                        var info = pinegrow.devTools.getCSSPropertyInfo(item.name);
                        var add = []

                        if(info) {
                            if(show_desc) {
                                if (info.desc) {
                                    add.push({
                                        type: 'html',
                                        html: info.desc_html,
                                        on_click: function ($li, event) {

                                        }
                                    })
                                    add.push({
                                        type: 'divider'
                                    })
                                }
                            }
                            if(info.examples && info.examples.length) {
                                add.push({
                                    label: 'Examples',
                                    submenu: info.examples.map(function(e) {
                                        return {
                                            label: '<code>' + e + '</code>'
                                        }
                                    })
                                })
                            }
                            if(info.values) {
                                var buttons = []

                                /*
                                add.push({
                                    type: 'header',
                                    label: 'Values'
                                })*/
                                add.push({
                                    buttons: buttons
                                })
                                info.values.forEach(function(val) {
                                    buttons.push({
                                        label: val,
                                        class: 'pg-css-menu-value',
                                        action: function() {}
                                    })
                                })
                            }
                            if(add.length) {
                                add.push({
                                    type: 'divider'
                                })
                            }
                        }
                        if((item.sub && item.sub.length) || add.length) {
                            console.log(item.name, item.sub, add)
                            d.submenu = getSubmenu(item.sub || [], add);
                        }
                    }
                    r.push(d)
                })
                return r;
            }
        }

        var getInsertCSSPropMenu = function() {
            var a = [];

            var info = pinegrow.devTools.getCSSPropsInfo();
            $.each(info, function(section, def) {

                var d = {
                    label: `<i class="icon ${def.icon}" style="margin-right:8px;"></i>${def.name}`,
                    submenu: getSubmenu(def.sub),
                    class: 'pg-css-menu-cat'
                }
                a.push(d);
            })

            a.push({
                type: 'divider'
            })
            a.push({
                label: 'Show descriptions',
                check: function() {
                    return pinegrow.getSetting('css-menu-show-desc', '1') === '1';
                },
                action: function() {
                    var val = !(pinegrow.getSetting('css-menu-show-desc', '1') === '1');
                    pinegrow.setSetting('css-menu-show-desc', val ? '1' : '0')
                }
            })
            return a;
        }

        tableGrid.$container
            .on('contextmenu', '.pg-table-grid-cell', function (event) {
                event.preventDefault();
                event.stopPropagation();

                var $el = $(this)
                var data = getDataForTree$Element($el);

                var actions = []

                var cmds = {
                    insert_prop: {
                        label: 'Insert property',
                        submenu: function() {
                            return getInsertCSSPropMenu()
                        }
                    },
                    insert_rule: {
                        label: 'Insert rule',
                        action: function() {}
                    },
                    important: {
                        label: 'Important',
                        action: function() {},
                        check: data.node.important || false
                    },
                    override: {
                        label: 'Override',
                        action: function() {}
                    },
                    override_in: {
                        label: 'Override in',
                        action: function() {}
                    },
                    duplicate: {
                        label: 'Duplicate',
                        action: function() {}
                    },
                    duplicate_in: {
                        label: 'Duplicate in',
                        action: function() {}
                    },
                    media_query: {
                        label: 'Media query',
                        action: function() {}
                    },
                    supports: {
                        label: 'Supports',
                        action: function() {}
                    },
                    div: {
                        type: 'divider'
                    }
                }

                if(data.node) {
                    switch(data.node.type) {
                        case 'decl':
                            actions.push(cmds.insert_prop)
                            actions.push(cmds.div)
                            actions.push(cmds.important)
                            break;
                        case 'rule':
                            actions.push(cmds.insert_prop)
                            actions.push(cmds.div)
                            actions.push(cmds.insert_rule)
                            actions.push(cmds.div)
                            actions.push(cmds.media_query)
                            actions.push(cmds.supports)
                            actions.push(cmds.div)
                            actions.push(cmds.override)
                            actions.push(cmds.override_in)
                            actions.push(cmds.duplicate)
                            actions.push(cmds.duplicate_in)
                            break;
                    }
                }

                if(actions.length) {
                    var menu = new CrsaContextMenu(actions, $el);
                    menu.showForElement($el)
                }
                return false;
            })
            .on('mouseover', '.pg-table-grid-cell', function (event) {
                if(!tableGrid.inScroll) {
                    var data = getDataForEvent(event);

                    if (!pinegrow.getDragMenu()) {

                    }

                }
                event.stopImmediatePropagation();
            })
            .on('mouseout', '.pg-table-grid-cell', function (event) {
                if(!tableGrid.inScroll) {
                    if (!pinegrow.getDragMenu()) {

                    }

                }
                event.stopImmediatePropagation();
            })
            .on('click', 'i.collapse-icon', function (event) {

                var data = getDataForEvent(event);

                data.collapsed = !(data.collapsed || false);

                if(event.altKey) {
                    tableGrid.forEachVisibleSibling(data, function(ch) {
                        ch.collapsed = data.collapsed;

                    })
                }

                tableGrid.refresh();
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', 'i.icon-hide', function (event) {

                var data = getDataForEvent(event);
                showItem(data);
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', '.pg-table-grid-cell', function (event) {

                //event.stopImmediatePropagation();
                //event.preventDefault();

                var $node = $(this);
                var data = getDataForTree$Element($node);

                if(quickInsertPosition) {
                    addGridLine(data, quickInsertPosition);
                    return;
                }

                // addGridLine(data);
                // return;

                if(data && data.node) {

                    var rule = pgcssh.getRuleForNode(data.node);

                    if(rule) {
                        if(!pinegrow.isSelectedRule(rule)) {
                            pinegrow.selectRule(rule);
                        }
                    }

                    var selected_done = false;

                    if(event.shiftKey) {

                        var selected = [];
                        var selected_datas = [];
                        var last_node = pinegrow.selectedCSSNodes.getLastSelected();

                        if (last_node) {
                            var last_data = getDataForCSSNode(last_node);
                            if(last_data) {
                                var inc = last_data.visible_idx > data.visible_idx ? -1 : 1;

                                for (var i = last_data.visible_idx + inc; i !== data.visible_idx + inc; i = i + inc) {
                                    if (tableGrid.visibleData[i].level == last_data.level) {
                                        var node = tableGrid.visibleData[i].node;
                                        if(!pinegrow.selectedCSSNodes.isSelected(node)) {
                                            pinegrow.selectedCSSNodes.select(node);
                                            selected.push(node);
                                            selected_datas.push(tableGrid.visibleData[i]);
                                        }
                                    }
                                }
                            }

                            if(selected.length) {
                                selected_done = true;
                                pinegrow.selectedCSSNodes.dispatchEvent(selected, {
                                    datas: selected_datas,
                                    source: _this
                                })
                            } else {
                                selected_done = true;
                                pinegrow.showQuickError('SHIFT + Click can only select elements on the same level.');
                            }
                        }

                    } else if(crsaIsCmdPressed(event)) {
                        pinegrow.selectedCSSNodes.toggle(data.node);

                        pinegrow.selectedCSSNodes.dispatchEvent([data.node], {
                            source: _this,
                            datas: [data]
                        })

                        selected_done = true;

                    }
                    if(!selected_done) {
                        pinegrow.selectedCSSNodes.clear({
                            source: _this
                        })
                        pinegrow.selectedCSSNodes.select(data.node);
                        pinegrow.selectedCSSNodes.dispatchEvent([data.node], {
                            source: _this,
                            datas: [data]
                        })
                    }

                }

                return;
                //selectNode(node);

                var rule = pgcssh.getRuleForNode(node)

                if(rule) {
                    pinegrow.selectRule(rule);
                }
                return;

                if(pgel) {
                    if(quickInsertPosition) {
                        showQuickInsertIntoTree(data, quickInsertPosition);
                    } else {
                        var sel_done = false;

                        if (event.shiftKey) {
                            var selected = false;
                            var last_pgel = pinegrow.selectedElements.getLastSelectedOnPage(currentPage);
                            if (last_pgel) {
                                var last_data = currentPage.tree.getTreeDataItemForPgId(last_pgel.getId());
                                var inc = last_data.visible_idx > data.visible_idx ? -1 : 1;
                                var col = new PgCollection();
                                for (var i = last_data.visible_idx + inc; i !== data.visible_idx + inc; i = i + inc) {
                                    if(tableGrid.visibleData[i].level == last_data.level) {
                                        var cpgel = _this.getPgelOfDataItem(tableGrid.visibleData[i]);
                                        if (cpgel) {
                                            //pinegrow.selectedElements.add(cpgel);
                                            col.add(cpgel);
                                            selected = true;
                                        }
                                    }
                                }
                                pgel.scrollTo();
                                if(selected) {
                                    pinegrow.selectedElements.selectMany(col);
                                }
                                sel_done = true;

                                if(!selected) {
                                    pinegrow.showQuickMessage('SHIFT + Click can only select elements on the same level.');
                                }
                            }
                        }
                        if (!sel_done) {
                            pinegrow.selectElement(pgel, 'tree', crsaIsCmdPressed(event));
                            pgel.scrollTo();

                            if($(event.target).is('.inline-styles')) {
                                pinegrow.showUIPanel('style', false);
                                //console.log(pinegrow.pgActiveRules.getActiveNodes());
                                pinegrow.activeCSSView.updateRulesDisplayOnElementSelect(true);
                            }
                        }
                    }
                } else {
                    //todo30 - handle dyn el select
                    pgDynElementMessage(quickInsertPosition ? 'Can\'t edit dynamic elements.' : 'Can\'t select a dynamic element.');
                }

                $('body').trigger('click');
                // return false;
            });

    }


    view.onDestroy = function() {
        resizeObserver.disconnect();

        pinegrow.removeEventHandler('on_css_loaded', onCSSLoaded);
        pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected);
        pinegrow.removeEventHandler('on_page_stylesheets_loaded', onPageStylesheetsLoaded);

        pinegrow.removeEventHandler('on_css_node_selected_deselected', onNodeSelectedDeselected);
        pinegrow.removeEventHandler('on_css_node_deleted', onNodeDeleted);
        pinegrow.removeEventHandler('on_css_node_updated', onNodeChanged);
        pinegrow.removeEventHandler('on_css_node_added', onNodeAdded);

        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        pinegrow.removeEventHandler('on_page_loaded', onPageLoaded);
        pinegrow.removeEventHandler('on_page_refreshed', onPageLoaded);
        pinegrow.removeEventHandler('on_page_selected', onPageSelected);
        pinegrow.removeEventHandler('on_drag_drop_operation_done', onDragDropDone);
        pinegrow.removeEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);
        pinegrow.removeEventHandler('on_element_isolated', onElementIsolated);

        $element.off('mousemove', onMouseMoved);
        $element.off('mouseleave', onMouseOut);
        $element.off('mouseenter', onMouseIn);
        //$element.off('mouseenter mouseleave', noop);
    }

    var onPageChanged = function(page, pgel, info) {
    }



    var onCSSLoaded = function(pagenull, cs) {

    }

    var onPageStylesheetsLoaded = function(page) {

    }

    var inherited_props = [
        "border-spacing", "border-collapse" , "color-adjust" , "empty-cells" , "caret-color" , "font-variant-ligatures" , "font" , "caption-side" , "font-style" , "list-style-image" , "color" , "object-position" , "font-family" , "letter-spacing" , "line-height" , "font-stretch" , "font-variant-numeric" , "rotate" , "list-style-position" , "font-size" , "font-optical-sizing" , "font-weight" , "pointer-events" , "perspective" , "text-align" , "overflow-wrap" , "text-align-last" , "quotes" , "list-style-type" , "image-rendering" , "word-wrap" , "text-indent" , "visibility" , "font-variant" , "mix-blend-mode" , "transform" , "word-spacing" , "scale" , "text-shadow" , "transform-style" , "translate" , "word-break" , "font-variant-caps" , "white-space" , "text-transform" , "cursor" , "hyphens" , "position" , "z-index"]

    var inherited_props_map = {}
    inherited_props.forEach(function(p) {
        inherited_props_map[p] = true;
    })


    var clearFlags = function(rules) {
        rules.forEach(function(rule) {
            if(rule.node.nodes) {
                rule.node.nodes.forEach(function(decl) {
                    decl[flag_not_inherited] = false;
                    decl[flag_overriden] = false;
                })
            }
        })
    }

    var markAndFilterInherited = function(rules) {
        var r = [];
        rules.forEach(function(rule) {
            var has = false;
            if(rule.node.nodes) {
                rule.node.nodes.forEach(function(decl) {
                    if(decl.type === 'decl') {
                        if(decl.prop in inherited_props_map) {
                            decl[flag_not_inherited] = false;
                            has = true;
                        } else {
                            decl[flag_not_inherited] = true;
                        }
                    }
                })
            }
            if(has) {
                r.push(rule)
            }
        })
        return r;
    }

    var markOverriden = function(rules) {
        //mark overriden
        var defined = {}

        rules.forEach(function(rule) {
            if(rule.node.nodes) {
                rule.node.nodes.forEach(function(decl) {
                    if(decl.type === 'decl') {
                        if(decl.prop in defined) {
                            if(decl.important && !defined[decl.prop].important) {
                                defined[decl.prop][flag_overriden] = true;
                                defined[decl.prop] = decl;
                                decl[flag_overriden] = false;
                            } else {
                                decl[flag_overriden] = true;
                            }
                        } else {
                            defined[decl.prop] = decl;
                            decl[flag_overriden] = false;
                        }
                    }
                })
            }
        })
    }



    var getMatchedForPgelStack = function(pgel, done) {
        var p = new CrsaProfile(true, 'getMatchedForStack');

        pinegrow.pgActiveRules.clear();
        pinegrow.pgActiveRules.setPgel(pgel);

        pinegrow.pgActiveRules.updateMatchedRules(true, function (rules) {
            var an = pinegrow.pgActiveRules.getActiveNodes();
            console.log('BUBI', an)

            var list = []
            if(an) {
                var d = {
                    pgel: pgel,
                    rules: an.inline_rules.map(function(r) { return {node: r} }).concat(an.nodes_to_show)
                }

                list.push(d);

                clearFlags(d.rules);
                markOverriden(d.rules)

                an.inherited.forEach(function(info) {

                    markOverriden(info.data.nodes_to_show)
                    var rules = markAndFilterInherited(info.data.nodes_to_show)

                    if(rules.length) {
                        list.push({
                            pgel: info.pgel,
                            rules: rules
                        })
                    }

                })

                var computed = {}
                for(var i = list.length - 1; i >= 0; i--) {
                    list[i].rules.forEach(function(rule) {
                        if(rule.node.nodes) {
                            rule.node.nodes.forEach(function(node) {
                                if(node.prop) {
                                    if(computed[node.prop]) {
                                        if(computed[node.prop].important && !node.important) {
                                            //
                                        } else {
                                            computed[node.prop] = node;
                                        }
                                    } else {
                                        computed[node.prop] = node;
                                    }
                                }
                            })
                        }
                    })
                }
            }
            done(list);
            console.log('Computed', computed)
        })
    }

    var onElementSelected = function() {

        view.whenVisible(function() {
            var pgel = pinegrow.selectedElements.getLastSelected();

            if (pgel) {
                var p = new CrsaProfile(true, 'Show active rules');

                getMatchedForPgelStack(pgel, function (list) {

                    p.show('Got from devtools');
                    console.log(list);

                    if(list.length) {
                        composerPanel.setData(list);
                    }
                    p.show('Updated composer');
                    return;

                    activeControls.update(pgel);

                    var top = {
                        rootNode: {
                            type: 'root',
                            pgStylesheet: {
                                name: 'Active'
                            },
                            nodes: []
                        }
                    }

                    list.forEach(function (data) {

                        var cs = {
                            rootNode: {
                                type: 'root',
                                pgStylesheet: {
                                    name: data.pgel.getName()
                                },
                                nodes: []
                            }
                        }

                        var current_cs = null;
                        var nodes = [];

                        data.rules.forEach(function (rule_info) {

                            if(rule_info.node.inline) {
                                cs.rootNode.nodes.push(rule_info.node);

                            } else {
                                if (current_cs !== rule_info.pgStylesheet) {
                                    if (nodes.length) {
                                        cs.rootNode.nodes.push({
                                            type: 'root',
                                            pgStylesheet: current_cs,
                                            nodes: nodes
                                        });
                                    }
                                    nodes = [];
                                    current_cs = rule_info.pgStylesheet;
                                }
                                nodes.push(rule_info.node);
                            }
                        })

                        if (nodes.length) {
                            cs.rootNode.nodes.push({
                                type: 'root',
                                pgStylesheet: current_cs,
                                nodes: nodes
                            });
                        }

                        top.rootNode.nodes.push(cs.rootNode);
                    });

                    p.show('Tree shown')
                })

            }
        })
    }

    var onPageLoaded = function(page) {
    }

    var onPageFrameworksChanged = function(page) {
    }

    var onElementIsolated = function(page, pgel) {
    }

    var onPageSelected = function(page) {
        currentPage = page;
    }

    var onDragDropDone = function(page, drag_menu) {
    }


    pinegrow.addEventHandler('on_css_loaded', onCSSLoaded);
    pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
    pinegrow.addEventHandler('on_page_stylesheets_loaded', onPageStylesheetsLoaded);

    pinegrow.addEventHandler('on_page_changed', onPageChanged);
    pinegrow.addEventHandler('on_page_loaded', onPageLoaded);
    pinegrow.addEventHandler('on_page_refreshed', onPageLoaded);
    pinegrow.addEventHandler('on_page_selected', onPageSelected);
    pinegrow.addEventHandler('on_drag_drop_operation_done', onDragDropDone);
    pinegrow.addEventHandler('on_element_isolated', onElementIsolated);

    pinegrow.addEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);


    this.onSelected = function(data) {

    }

    this.onDeselected = function(data) {

    }
}



