
var PgComboDividerBox = function($divider, view, without_title) {
    var $box = null;

    var fastdom_key = view.getFastdomTasks().getUniqueKey();

    var divider = $divider.get(0);
    var $container;

    var addedToDom = function() {
        if($box) $box.remove();
        $box = $('<div class="pg-combined-panel-box"></div>');
        if(without_title) {
            $box.addClass('without-title');
        }
        $container = $divider.closest('.lm_item_container,.pg-quick-win');
        $container.append($box);
        updatePosition();
    }

    var updatePosition = function(fastdom) {

        if($container) {
            if(!$container.get(0).parentNode) {
                addedToDom();
            }
            var task = fastdom ? view.getFastdomTasks().create(fastdom_key) : view.getFastdomTasks().createDummy();

            task.measure(function() {
                var rect = divider.getBoundingClientRect();
                var rect_c = $container.get(0).getBoundingClientRect();

                task.mutate(function() {
                    $box.css({
                        top: (rect.top - rect_c.top) + 'px'
                    })
                })
            })
        }
    }

    this.onAddedToLayout = function() {
        addedToDom();

        view.whenVisible(function() {
            addedToDom();
        }, 'onComboDividerFirstShow')
    }

    this.update = function(fastdom) {
        view.whenVisible(function() {
            updatePosition(fastdom);
        }, 'updateComboDivider');
    }

    this.destroy = function() {
        if($box) $box.remove();
        view.getFastdomTasks().stopAll(fastdom_key);
    }

    this.hide = function() {
        if($box) pg$Hide($box);
    }

    this.show = function() {
        if($box) {
            pg$Show($box);
            this.update();
        }
    }

    view.whenVisible(function() {
        addedToDom();
    }, 'onComboDividerFirstShow')

}

var PgCombinedPanelView = function(key, master_view, helper_view, title, default_h) {

    var _this = this;

    this.minHelperSize = 50;
    
    var key_open = key + '-open';
    var key_size = key + '-size';

    var $element = $('<div class="pg-combined-panel"></div>');

    var view = this.view = new PgUIView($element);

    var $main = $element;

    var $master = $('<div class="pg-combined-panel-master"></div>').appendTo($main);

    var $helper = $('<div class="pg-combined-panel-master-helper"><h1><span class="text">' + (title || '') + '</span><span class="handle"><i class="icon icon225 icon-drag-vert-x225"></i></span><!-- <i class="icon icon-ddown hide-gui"></i><i class="icon icon-dup show-gui"></i>--></h1></div>').appendTo($main);

    if(!title) $helper.addClass('without-title');

    master_view.get$Element().appendTo($master);
    helper_view.get$Element().appendTo($helper);

    var $resizer = $helper.find('> h1');

    $resizer.find('.hide-gui').on('click', function(e) {
        e.preventDefault();
        hideHelper();
        return false;
    })

    var box = new PgComboDividerBox($resizer, view, !title);

    view.onResize = function() {
        var w = $element.width();

        if(!is_static) {
            parent_h = 0;
            setHelperSize(percent);
        } else {
            _this.enterStaticMode(static_h);
        }
        box.update();
    }

    var helper_open = pinegrow.getSetting(key_open, "1") === "1";

    var hideHelper = function() {
        return;
        $main.addClass('gui-hidden');
        helper_view.hide();
        pinegrow.setSetting(key_open, "0");
        helper_open = false;
    }

    var showHelper = function() {
        return;
        $main.removeClass('gui-hidden');
        helper_view.show();
        view.onResize();
        pinegrow.setSetting(key_open, "1");
        helper_open = true;
    }

    var percent = parseFloat(pinegrow.getSetting(key_size, (default_h || 50).toString()));

    var setHelperSize = function(p) {
        if(!parent_h) parent_h = $element.height();
        if(p > 90) p = 90;
        if(parent_h && p * parent_h / 100.0 < _this.minHelperSize) {
            p = _this.minHelperSize / parent_h * 100.0;
        }
        $helper.css({
            'min-height': p + '%',
            'max-height': p + '%'
        })

        box.update();
    }

    setHelperSize(percent);

    var is_static = false;
    var static_h = 0;

    this.enterStaticMode = function(h) {
        static_h = h;
        var p = 100 - ((h * 100) / $element.height());
        setHelperSize(p);
        $resizer.addClass('static')
        is_static = true;
    }

    this.exitStaticMode = function() {
        is_static = false;
        $resizer.removeClass('static')
        percent = parseFloat(pinegrow.getSetting(key_size, (default_h || 50).toString()));
        setHelperSize(percent);
    }

    this.isStatic = function() {
        return is_static;
    }

    if(!helper_open) {
        hideHelper();
    }

    $resizer.on('click', function(e) {
        e.preventDefault();
        if(in_resize) return;
        helper_open = !helper_open;
        if(helper_open) {
            showHelper();
        } else {
            hideHelper();
        }
    })

    var isSecondHidden = false;

    this.hideSecondPanel = function() {
        if(isSecondHidden) return;

        pg$Hide($helper);
        isSecondHidden = true;
        box.hide();
        pgUIViewManager.updateVisibility();
    }

    this.showSecondPanel = function() {
        if(!isSecondHidden) return;

        pg$Show($helper);
        isSecondHidden = false;
        box.show();
        pgUIViewManager.updateVisibility();
    }

    this.onAddedToLayout = function() {
        box.onAddedToLayout();
    }

    //var drag_resize = new GoldenLayout.__lm.utils.DragListener($resizer.find('.handle'), null, false);
    var drag_resize = new PgDragHelper($resizer, '.handle', null, false);

    var resize_original_y = 0;
    var in_resize = false;
    var parent_h = 0;

    drag_resize.onBefore = function() {
        return !is_static;
    }

    drag_resize.on( 'dragStart', function(x, y) {
        resize_original_y = $helper.outerHeight();
        parent_h = $element.height();
        in_resize = true;

        $.fn.crsapages('showOverlays');
    }, this );

    drag_resize.on( 'drag', function(x, y, event) {

        var h = Math.max(0, resize_original_y - y);
        percent = (h / parent_h) * 100;
        setHelperSize(percent);

    }, this );

    drag_resize.on( 'dragStop', function(x, y) {
        $.fn.crsapages('showOverlays', true);
        setTimeout(function() {
            in_resize = false;
        }, 100)

        pinegrow.setSetting(key_size, percent);
    }, this );

}