$('body').one('pinegrow-ready', function(e, pinegrow) {

    class PgLiveBuild {
        constructor() {
            this.last_live_version = parseFloat(pinegrow.getSetting('last-live-version', 0));
            this.descriptions = [];

            this.log();

            if(this.isLiveUpdated()) {
                this.showLog();
                pinegrow.setSetting('last-live-version', this.getLiveVersion());
            }
        }

        getVersionFloat(major, live) {
            return parseFloat(major) * 100 + (live / 1000)
        }

        //version format 6.21, live 10 => 621.010
        getLiveVersion() {
            const minor = 2;
            return this.getVersionFloat(crsaGetVersion(), minor);
        }

        isLiveUpdated() {
            return this.getLiveVersion() > this.last_live_version;
        }

        describeVersion(major_version, live_version, date, html) {
            this.descriptions.push({
                major_version: major_version,
                live_version: live_version,
                date: date,
                html: html
            })
        }

        showLog() {
            const _this = this;
            var dm = pinegrow.findFrameworkByType('pg.docs.maker');
            if (dm) {
                var html = '<h1>What\'s New in Pinegrow Live<small>Update log</small></h1><p class="lead">Pinegrow Live has been updated. <a href="https://www.getdrip.com/forms/362224443/submissions/new" class="external">Receive updates by email.</a></p><div class="pg-live-log">';
                this.descriptions.forEach(function(d) {
                    var v = _this.getVersionFloat(d.major_version, d.live_version);
                    if(v > _this.last_live_version || true) {
                        html += `<div data-pgdm-item-section data-pgdm-item-section-open><h2>${d.date} - ${d.major_version}.${d.live_version } <span style="color: #333;
    text-transform: none;"><img src="https://pinegrow.com/pg4assets/img/matjaz_1.jpg" style="width: 24px;
    height: 24px;
    border-radius: 50%;
    margin-left: 12px;
    position: relative;
    top: -2px;"> Matjaz</span> </h2>${d.html}</div>`;
                    }
                })
                html += '</div>';
                dm.docsAddTopic('live.log', 'What\'s New in Pinegrow Live', html);
                pinegrow.showDocsTopic('live.log');
            }
        }

        log() {
            this.describeVersion(6.21, 2, 'Dec 9, 2021', `
<h3>Fixed input / button type in Properties panel</h3>
<p>Button types were shown in the input type select if a button element was selected before the input element. Reported on <a href="https://forum.pinegrow.com/t/add-more-type-values-for-input-html-properties/5986" class="external">this forum thread</a>.</p>   

<h3>Fix previewing files with partials</h3>
<p>Previewing a document with partials in the browser made the element dynamic.</p>  

<h3>Fix PUG link in the Library panel</h3>
<p>Jade was renamed to Pug a while ago. The link was updated to <a href="https://pugjs.org/language/tags.html" class="external">the tags Pug page</a>. Reported on <a href="https://forum.pinegrow.com/t/insert-after-panel-link-pug-doesn-t-redirect-to-the-correct-website/5978" class="external">this forum thread</a>.</p>         
            `)


            this.describeVersion(6.21, 1, 'Dec 1, 2021', `
<p>Welcome to the first Pinegrow Live update!</p><ul>

    <h3>Performance optimizations</h3>
    <p>Release 6.2 was noticebly slower than previous versions. The cause was updating to a newer NWJS build. It appears that newer versions of Chromium are more sensitive to DOM updates done in non-optimal way. This is forcing us to dig under the hood and fix code that caused unneccessary layout and style recalculation in the browser. So, it's not all bad :) Our tests show that Pinegrow is now noticably faster (at least compared to 6.2). Let us know how it goes for you. More to come.</p>
  
    <h3>Fixed WordPress post field as background image</h3>
    <p><a href="https://forum.pinegrow.com/t/post-field-background-image-bug/5944" class="external">The bug was reported on the forum</a> and is now fixed.</p>

   <h3>Broken links checker</h3>
   <p>We forgot to tell you about this new handy tool in 6.21 release notes. Select <b>Page -&gt; Check for broken links</b> to open the link checker. The tool will check both local and remote links on the selected page and notify you if any links are broken.</p>

    <h3>Runaway process</h3>
    <p>Sometimes a process called <code>nwjs helper (Renderer)</code> is left after Pinegrow quits and it consumes CPU and memory. This appears to be a NWJS issue and we can not fix it directly. In this update we added a code that explicitly terminates the process on Pinegrow exit. Please let us know if this has any positive impact.</p>

    <h3>WordPress adds rel="noopener" when target in blocks is set to _blank</h3>
    <p>In reponse to <a href="https://forum.pinegrow.com/t/gutenberg-blocks-with-links-that-have-target-blank-need-rel-noopener/5972/2" class="external">this forum post</a>. I'm not that keen on adding a magic behavior (something that happens without direct instruction by the user) to the code generated by Pinegrow. In this case, this made sense. So, Pinegrow will automatically add rel="noopener" to block links if the rel is not set already and if target is set to _blank.</p>
           
            `);


        }
    }

    pinegrow.liveBuild = new PgLiveBuild();

});