class PgImageBrowserProvider {

    constructor(name) {
        this.name = name;
        this.custom_dim = null;
        this.default_size = 'regular';
        this.supports_custom_dimensions = true;
        this.supports_add_image = true;
        this.supports_delete = true;
        this.parameter = null;
    }

    onError(error) {}

    setParameter(val) {
        this.parameter = val;
    }

    registerCustomDimensions = function(dim) {
        this.custom_dim = dim;
    }

    getDefaultSize = function() {
        return this.custom_dim ? 'custom' : this.default_size;
    }

    getImageUrlForInsertingToPage(d, size) {
        var url = this.getImageUrl(d, size);
        var p = pinegrow.getCurrentProject();
        if(p) {
            url = p.mapDOMUrlToSourceUrl(url);
        }
        return url;
    }

    getImageUrl(d, size, dim) {
    }

    getImageSrcset(d) {
    }

    getImageTag(pic, size, responsive) {
        var srcset = this.getImageSrcset(pic) || '';
        if(srcset && responsive) {
            srcset = ` srcset="${srcset}"`;
        } else {
            srcset = '';
        }
        return `<img src="${this.getImageUrl(pic, size)}"${srcset}>`;
    }

    getImageTagForInsertingToPage(pic, size, responsive) {
        var srcset = this.getImageSrcset(pic) || '';
        if(srcset && responsive) {
            srcset = ` srcset="${srcset}"`;
        } else {
            srcset = '';
        }
        return `<img src="${this.getImageUrlForInsertingToPage(pic, size)}"${srcset}>`;
    }

    getLink(url) {
        return url;
    }

    triggerDownload(d) {
    }

    getImageDimensions(d, size) {
        return {w:0, h:0}
    }

    getImageSizes(pic) {
    }

    getRandomPhoto(params, done, error) {
    }

    search(term, params, done, error) {
    }

    withCurrentResults(done) {
        return false;
    }
}

var PgSafeUnsplashImages = function() {

    var cats = {}

    var cats_list = [];

    var simple_sizes = {
        'small': {w: 400, h: 300},
        'square': {w: 800, h: 800},
        'regular': {w: 1024, h: 0},
        'large': {w: 1600, h: 0}
    }

    this.load = function(url, done) {
        fetch(url)
            .then(function(response) { return response.json(); })
            .then(function(data) {

                cats = data;

                var doCat = function(name, data) {
                    //can be main cat or with images
                    var cat = {
                        key: name,
                        sizes: simple_sizes
                    }

                    var images = {}

                    if(data.has_images) {
                        $.each(data, function(id, url) {
                            if(id !== 'has_images') {
                                images[id] = url;
                            }
                        })
                        cat.images = images;

                    } else {
                        cat.subcats = {};

                        $.each(data, function(subcat, subdata) {
                            subdata.has_images = true;
                            cat.subcats[subcat] = doCat(subcat, subdata);
                        })
                    }
                    return cat;
                }

                $.each(data, function(key, data) {

                    cats[key] = doCat(key, data);

                    cats_list.push(cats[key])
                })

                var custom_cats = [];
                pinegrow.dispatchEvent('on_image_categories_loaded', null, custom_cats);

                custom_cats.forEach(function(cat) {
                    cats_list.unshift(cat);
                    cats[cat.key] = cat;
                })

                if(done) done();
            })
            .catch(function(error) {
                console.log('GetSafeImages error: ', error.message);
            });
    }

    var cap = function(str) {
        return str.substring(0, 1).toUpperCase() + str.substring(1);
    }

    this.getCategoriesMenu = function(action) {
        var m = [];

        var photos = [];

        //$.each(cats, function(cat_name, images) {
        cats_list.forEach(function(cat) {
            var cat_id = cat.key;

            if(cat.top) {
                //display its cats on the top
                m.push({
                    type: 'header',
                    label: cat.header
                })
                $.each(cat.subcats, function(subcat, subdata) {
                    m.push({
                        label: cap(subcat),
                        key: cat_id + '.' + subcat,
                        action: action
                    })
                })
                m.push({
                    type: 'divider'
                })
                m.push({
                    label: 'Photos',
                    submenu: photos
                })
            } else if(cat.images) {
                photos.push({
                    label: cap(cat.key),
                    key: cat_id,
                    action: action
                })
            } else {
                var sub = [];
                $.each(cat.subcats, function(subcat, subdata) {
                    sub.push({
                        label: cap(subcat),
                        key: cat_id + '.' + subcat,
                        action: action
                    })
                })
                photos.push({
                    label: cap(cat.key),
                    submenu: sub
                })
            }
        })
        return m;
    }

    this.getDefaultCategory = function() {
        if(cats_list.length) {
            var c = cats_list[0];
            if(c.subcats) {
                var first = null;
                $.each(c.subcats, function(catid, d) {
                    first = first || catid;
                })
                return c.key + '.' + first;
            } else {
                return c.key;
            }
        }
        return null;
    }

    this.getImagesForCategory = function(cat_id) {
        var a = cat_id.split('.');
        var d = cats;
        for(var i = 0; i < a.length; i++) {
            if(i == 0) {
                d = d[a[i]];
            } else {
                d = d.subcats[a[i]];
            }
        }
        var r = [];
        $.each(d.images, function(image_id, url) {
            var sizes = d.sizes;

            var i = {
                id: image_id,
                urls: {
                    raw: url
                },
                links: null
            }

            if(sizes) {
                $.each(sizes, function (size, dim) {
                    var surl = url + 'q=85&fm=jpg&crop=entropy&cs=srgb';
                    if (dim.w) {
                        surl += '&w=' + dim.w;
                    }
                    if (dim.h) {
                        surl += '&h=' + dim.h;
                    }
                    i.urls[size] = surl;
                })
            } else {
                i.urls.regular = url;
                i.name = crsaGetNameFromUrl(url);
            }
            r.push(i);
        })
        return r;
    }
}

class PgUnsplashImageProvider extends PgImageBrowserProvider {

    constructor() {
        super('Unsplash');

        this.appId = pinegrow.defaults.unsplashAppId;
        this.api = 'https://api.unsplash.com/';

        this.random_list = [];

        this.supports_add_image = false;
        this.supports_delete = false;

        this.current_params_hash = null;
        this.current_params_max_pages = 1;
        this.current_params_max_pages_hash = null;

        this.featured = null;
    }

    createUrl(endpoint, params) {
        var url = this.api + endpoint + '?client_id=' + this.appId;
        if(params) {
            $.each(params, function (key, val) {
                url += '&' + key + '=' + encodeURIComponent(val);
            })
        }
        return url;
    }

    call(url, done, error) {
        var _this = this;
        //console.log('fetch ' + url);
        fetch(url)
            .then(function(response) { return response.json(); })
            .then(function(data) {
                //console.log(data);
                if(done) done(data);
            })
            .catch(function(err) {
                console.log('Unsplash API error: ', err.message);
                _this.onError(err.message);
                error && error();
        });
    }

    getImageUrl(d, size, dim) {
        size = size || this.getDefaultSize();
        dim = dim || this.custom_dim;

        if(size == 'custom' && dim) {
            var url = d.urls.raw + 'q=85&fm=jpg&crop=faces&cs=srgb';
            if(dim.w) {
                url += '&w=' + dim.w;
            }
            if(dim.h) {
                url += '&h=' + dim.h;
            }
            url += '&fit=' + ((dim.w * dim.h > 0) ? 'crop' : 'max');
            return url;
        }
        var url = d.urls[size] || d.urls.regular;
        return url;
    }

    getLink(url) {
        return url + '?utm_source=Pinegrow&utm_medium=referral';
    }

    triggerDownload(d) {
        if(!d.links || !d.links.download_location) return;

        var dl_url = d.links.download_location + '?client_id=' + this.appId;
        fetch(dl_url);
        //call( createUrl('photos/' + d.id + '/download'));
    }

    getImageDimensions(d, size) {
        var url = this.getImageUrl(d, size);
        var dim = size == 'full' ? {w: d.width, h: d.height} : {w: 0, h: 0};
        var m = url.match(/&w=([0-9]+)/);
        if(m) dim.w = parseInt(m[1]);
        m = url.match(/&h=([0-9]+)/);
        if(m) dim.h = parseInt(m[1]);
        return dim;
    }

    getImageSizes(pic) {

        const check_sizes = ['full', 'large', 'square', 'regular', 'small', 'thumb'];

        var sizes = {};

        for(var i = 0; i < check_sizes.length; i++) {
            if(pic.urls[check_sizes[i]]) {
                sizes[check_sizes[i]] = this.getImageDimensions(pic, check_sizes[i])
            }
        }

        if(this.custom_dim) {
            sizes.custom = this.getImageDimensions(pic, 'custom');
        }
        return sizes;
    }



    getRandomPhoto(params, done, error) {

        const _this = this;

        params = $.extend({
            count: 30
        }, params)

        if(!params.query) {
            //params.content_filter = 'high';
        }

        var getRandomItem = function(list) {
            var idx = Math.floor(Math.random() * (list.length - 1));
            var i = list[idx];
            list.splice(idx, 1);
            return i;
        }

        var hash = JSON.stringify(params);

        if(this.random_list.length && (!this.current_params_hash || this.current_params_hash === hash)) {
            if(done) done(getRandomItem(this.random_list))
        } else {
            this.current_params_hash = hash;

            //random api returns bad photos
            if(false) {
                this.call(this.createUrl('photos/random', params), function (data) {
                    _this.random_list = data;
                    if (done) done(getRandomItem(_this.random_list));
                })
            } else {
                var page = 1;
                var per_page = 1;

                if(this.current_params_max_pages_hash === hash) {
                    var max_page = this.current_params_max_pages;
                    if (this.current_params_max_pages > 40) {
                        max_page = Math.min(Math.round(this.current_params_max_pages * 0.25), 20);
                    }
                    if (max_page > 1) {
                        page = Math.floor(Math.random() * max_page) + 1;
                    }
                    per_page = 20;
                }

                this.search(params.query || '', {
                    per_page: per_page,
                    page: page,
                    featured: true
                }, function (data) {
                    _this.random_list = data.results;
                    _this.current_params_max_pages = data.total_pages;
                    _this.current_params_max_pages_hash = hash;
                    if (done) done(getRandomItem(_this.random_list));
                }, error)
            }
        }
    }

    search(term, params, done, error) {
        params = $.extend({
            page: 1,
            per_page: 40
        }, params)

        params.query = term;

        this.call(this.createUrl('search/photos', params), function (data) {
            if (done) done(data);
        }, error)
    }


    searchFeatured(done) {
        const _this = this;
        if(this.featured) {
            done(this.featured);
        } else {
            var params = {
                count: 30,
                featured: true
            }
            this.call(this.createUrl('photos/random', params), function (data) {
                _this.featured = {
                    results: data,
                    total_pages: 1
                }
                done(_this.featured);
            })
        }
    }
}

class PgBackendImageProvider extends PgImageBrowserProvider {

    constructor() {
        super('Backend');

        this.supports_custom_dimensions = false;
        this.default_size = 'full';

        this.size_map = {
            small: 'medium'
        }

        this.random_list = [];

        this.current_params_hash = null;
        this.current_params_max_pages = 1;
        this.current_params_max_pages_hash = null;

        this.featured = null;

        this.results_cache = {};
    }

    getImageUrl(d, size, dim, for_browser) {
        size = size || this.getDefaultSize();
        if(this.size_map[size]) {
            size = this.size_map[size];
        }

        if(!d.sizes[size]) size = this.getDefaultSize();
        var url = d.sizes[size] ? (d.url_base + d.sizes[size].file) : d.url;

        if(!for_browser) {
            var project = pinegrow.getCurrentProject();
            url = project.mapDOMUrlToSourceUrl(url);
        }
        return url;
    }

    getImageSrcset(d) {
        return d.srcset || null;
    }

    /*
    getImageTagForInsertingToPage(pic, size) {
        var project_url = pinegrow.browserPreview.getPublicUrlForProject();

        var srcset = this.getImageSrcset(pic) || '';
        if(srcset) {
            //make links to images from the same project relative
            srcset = srcset.split(' ').map(function(i) {
                return i.replace( project_url, '');
            }).join(' ');
            srcset = ` srcset="${srcset}"`;
        }
        var src = this.getImageUrlForInsertingToPage(pic, size);

        //make links to images from the same project relative
        src = src.replace( project_url, '');

        return `<img src="${src}"${srcset}>`;
    }*/

    triggerDownload(d) {
    }

    getImageSizes(pic) {

        var sizes = {};

        $.each(pic.sizes, function(size, def) {
            sizes[size] = {w: def.width, h: def.height};
        })
        return sizes;
    }



    getRandomPhoto(params, done, error) {

        const _this = this;

        params = $.extend({
            count: 30
        }, params)

        if(!params.query) {
            //params.content_filter = 'high';
        }

        var getRandomItem = function(list) {
            var idx = Math.floor(Math.random() * (list.length - 1));
            var i = list[idx];
            list.splice(idx, 1);
            return i;
        }

        var hash = JSON.stringify(params);

        if(this.random_list.length && (!this.current_params_hash || this.current_params_hash === hash)) {
            if(done) done(getRandomItem(this.random_list))
        } else {
            this.current_params_hash = hash;

            var page = 1;
            var per_page = 1;

            if(this.current_params_max_pages_hash === hash) {
                var max_page = this.current_params_max_pages;
                if (this.current_params_max_pages > 40) {
                    max_page = Math.min(Math.round(this.current_params_max_pages * 0.25), 20);
                }
                if (max_page > 1) {
                    page = Math.floor(Math.random() * max_page) + 1;
                }
                per_page = 20;
            }

            this.search(params.query || '', {
                per_page: per_page,
                page: page,
                featured: true
            }, function (data) {
                _this.random_list = data.results;
                _this.current_params_max_pages = data.total_pages;
                _this.current_params_max_pages_hash = hash;
                if (done) done(getRandomItem(_this.random_list));
            }, error)
        }
    }

    getResultsKey() {
        return 'results_' + (this.term || '') + '_' + (this.parameter + '');
    }

    search(term, params, done, error) {
        this.term = term;

        params = $.extend({
            term: term,
            page: 1,
            per_page: 40,
            project_id: this.parameter || null
        }, params)

        params.query = term;

        const _this = this;

        pinegrow.backend.getImages(params, function(data) {
            const key = _this.getResultsKey();
            if(params.page === 1 || !_this.results_cache[key]) {
                _this.results_cache[key] = [];
            }
            _this.results_cache[key] = _this.results_cache[key].concat(data.results);
            done && done(data);
        }, error);
    }


    searchFeatured(done) {
        this.search(null, {}, done);
    }

    deleteImage(pic, done) {
        const _this = this;
        pinegrow.backend.deleteImage(pic.id, done, function(e) {
            _this.onError && _this.onError(e);
        })
    }

    withCurrentResults(done) {
        const key = this.getResultsKey();
        if(this.results_cache[key]) {
            done({results: this.results_cache[key]});
            return true;
        }
        return false;
    }
}

class PgImageBrowserProvidersManager {
    constructor() {
        this.list = [];

        const _this = this;

        $('body').one('pinegrow-ready', function() {
            if(!isApp()) {
                let backend = new PgBackendImageProvider();
                _this.list.push(backend);
            }

            let unsplash = new PgUnsplashImageProvider();
            _this.list.push(unsplash);
        })
    }

    getProviders() {
        return this.list;
    }
}

//const pgImageBrowserProviderManager = new PgImageBrowserProvidersManager();

var pgQuickImageBrowserSearchTerm = null;

var PgImageBrowserView = function() {

    var mark_images = false;

    var marked = {}

    let provider_backend = isApp() ? null : new PgBackendImageProvider();
    let provider_unsplash = new PgUnsplashImageProvider();

    if(mark_images) {
        var mi = localStorage.getItem('markedImages');
        if(mi) {
            marked = JSON.parse(mi);
        }
    }

    var markToggle = function(pid, url) {
        var cat = search_term;
        if(!marked[cat]) {
            marked[cat] = {};
        }
        if(marked[cat][pid]) {
            delete marked[cat][pid];
        } else {
            marked[cat][pid] = url;
        }
        localStorage.setItem('markedImages', JSON.stringify(marked));
        console.log(marked)
    }

    var isMarked = function(pid) {
        var cat = search_term;
        if(marked[cat] && marked[cat][pid]) return true;
        return false;
    }

    var safe = pgf.kids || false;

    if(safe) {
        var safe_images = window.PgSafeUnsplashImages;
        //safe_images.load('images/kids/kids-images.json');
    }

    var custom_sizes = !pgf.kids;

    var picapi = provider_backend || provider_unsplash;

    this.setImageProvider = function(provider, param) {
        picapi = provider;
        provider.setParameter(param);

        if(provider.supports_custom_dimensions) {
            input_h.show();
            input_w.show();
        } else {
            input_h.hide();
            input_w.hide();
        }
        doSearch(true);
    }

    var $element = $('<div class="pg-image-browser"></div>');

    var view = new PgUIView($element);

    this.view = view;

    this.images_shown = false;

    var _this = this;

    this.drag_enabled = true;
    this.show_sizes = true;
    this.enable_unsplash = true;


    var per_page = 40;
    var page = 1;
    var total_pages = 0;
    var current_data = {results: [], total_pages: 1};
    var search_options = {
        placeholder: 'Search',
        placeholder_active: null,
        toolbar_on: false,
        delay: false,
        delay_time: 800,
        search_on_input: false
    }

    if(safe_images) {
        search_options.input_type = 'menu';
        search_options.input_get_menu_options = safe_images.getCategoriesMenu
        search_options.search_on_input = true;
        search_options.no_search_icon = true;
        search_options.input_menu_no_not_set_option = true;
    }

    var provider_selector = new PgInputField('provider_selector', {
        type: 'menu',
        without_text: true,
        name: 'Select',
        menu_no_not_set_option: true,
        get_menu_options: function() {
            var actions = [];

            if(pinegrow.currentUser && pinegrow.currentUser.canUploadImages()) {
                actions.push({
                    key: 'backend:' + pinegrow.backend.getProjectId(pinegrow.getCurrentProject()),
                    label: 'Current project'
                })

                actions.push({
                    key: 'backend:',
                    label: 'All my images'
                })

                var projects = []

                if (pinegrow.backend.list_of_loaded_projects) {
                    pinegrow.backend.list_of_loaded_projects.forEach(function (pi) {
                        projects.push({
                            label: pi.name,
                            key: 'backend:' + pi.project_id
                        })
                    })
                }

                if (projects.length) {
                    actions.push({
                        key: '',
                        label: 'Projects',
                        submenu: projects
                    })
                }
            }

            if(_this.enable_unsplash) {
                actions.push({
                    type: 'divider'
                })

                actions.push({
                    key: 'unsplash',
                    label: 'Unsplash'
                })
            }

            return actions;
        }
    });

    provider_selector.$field.css({
        width: 'calc(100% - 10px)',
        margin: '0px 0 5px'
    });

    if(!isApp()) {
        provider_selector.appendTo($element);
    } 

    provider_selector.on('change', function() {
        var v = provider_selector.getValue();
        if(v.startsWith('backend:')) {
            v = v.replace('backend:', '');
            _this.setImageProvider(provider_backend, v);
        } else {
            _this.setImageProvider(provider_unsplash);
        }
    })

    var search = new PgSearchBox(search_options);

    var toolbar = search.getToolbar();

    search.$element.appendTo($element);

    var search_term_re = null;
    var search_term = null;

    search.onSearch = function(f) {
        //_this.search(f);
        try {
            search_term = f;
            pgQuickImageBrowserSearchTerm = search_term;
            doSearch();
        }
        catch(err) {}
    }

    this.onShown = function() {
        search.focus();
        //doSearch();
        if(!isApp() && pinegrow.currentUser.canUploadImages()) {
            provider_selector.setValue('backend:' + pinegrow.backend.getProjectId(pinegrow.getCurrentProject()));
            this.setImageProvider(provider_backend, pinegrow.backend.getProjectId(pinegrow.getCurrentProject()));
        } else {
            provider_selector.setValue('unsplash');
            this.setImageProvider(provider_unsplash);
        }
    }

    this.getSearchTerm = function() {
        return search_term;
    }

    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    var dim_key = 'unsplash-dim-';

    var updateCustomDim = function(skip_update) {
        var dim = {w: input_w.getValue() || null, h: input_h.getValue() || null};
        if(dim.w !== null) dim.w = parseInt(dim.w);
        if(dim.h !== null) dim.h = parseInt(dim.h);
        dim = (dim.w + dim.h > 0) ? dim : null;
        picapi.registerCustomDimensions(dim);
        if(!skip_update) {
            updateResults();
            pinegrow.setSetting(dim_key + 'w', dim ? dim.w : '');
            pinegrow.setSetting(dim_key + 'h', dim ? dim.h : '');
        }
    }



    var input_w = new PgInputField('custom_w', {
        type: 'slider',
        name: 'W'
    });

    var w_val = pinegrow.getSetting(dim_key + 'w', '');
    if(w_val == 0 || !custom_sizes) w_val = '';
    input_w.setValue(w_val);

    input_w.on('change', function() {
        pinegrow.showQuickMessage('Dimensions updated.')
        updateCustomDim();
    })

    custom_sizes && input_w.appendTo(search.$element);

    var input_h = new PgInputField('custom_h', {
        type: 'slider',
        name: 'H'
    });

    var h_val = pinegrow.getSetting(dim_key + 'h', '');
    if(h_val == 0 || !custom_sizes) h_val = '';
    input_h.setValue(h_val);

    input_h.on('change', function() {
        pinegrow.showQuickMessage('Dimensions updated.')
        updateCustomDim();
    })

    custom_sizes && input_h.appendTo(search.$element);

    updateCustomDim(true);

    var $searchButton = $('<button class="pg-button pg-button-primary">Search</button>');
    !safe_images && $searchButton.appendTo(search.$element);

    var doSearch = function(use_current_results) {
        _this.images_shown = true;
        search_term = search.getValue();
        $items.empty();
        current_data.results = [];
        page = 1;

        if(safe_images) {
            if(search_term) {
                updateResults({
                    results: safe_images.getImagesForCategory(search_term),
                    total_pages: 1
                });
            }
        } else {
            if(use_current_results && picapi.withCurrentResults(updateResults)) {
                //current results were shown
            } else {
                if (search_term) {
                    picapi.search(search_term, {}, updateResults);
                } else {
                    //show featured
                    picapi.searchFeatured(updateResults);
                }
            }
        }
    }

    $searchButton.on('click', function(e) {
        e.preventDefault();
        doSearch();
    })

    picapi.onError = function(msg) {
        updateMoreButton();
        pinegrow.showQuickError('Oops... API call failed.');
    }

    var nextPage = function() {
        page++;
        picapi.search( search_term, {page: page}, updateResults);
    }

    var $container = $('<div class="pg-image-browser-c"></div>').appendTo($element);

    var $items = $('<div class="pg-image-browser-items"></div>').appendTo($container);

    var $more = $('<button class="pg-button more-button"></button>').appendTo($container);
    $more_pages = $more.find('span');

    pg$Hide($more);

    $more.on('click', function(e) {
        e.preventDefault();
        nextPage();
        //pg$Hide($more);
        updateMoreButton(true);
    })

    var used = function(pic) {
        picapi.triggerDownload(pic);
    }

    crsaHandleExternalLinks($container);

    var $popel = null;

    var destroyPopup = function() {
        if($popel) {
            $popel.popover('destroy');
            $popel = null;
        }
    }

    $container.on('click', '.pg-image-browser-upload', function(e) {
        e.preventDefault();
        e.stopPropagation();

        pinegrow.backend.uploadFilesWithDialog({}, function(data) {
            current_data.results.unshift(data.image);
            get$ItemForImage(data.image).insertAfter($items.find('.pg-image-browser-upload'));
        })

        return;

        var $d = pinegrow.showAlert(`<p>Choose one or more JPG, GIF or PNG files and click Upload. Images will be uploaded into the WordPress media library.</p><form enctype="multipart/form-data" style="margin-bottom: 30px;"><input type="file" class="file" name="file" multiple /></form>`, 'Upload images', 'Close', 'Upload', null, function() {
            var file = $d.find('input.file').get(0);
            $d.find('.error').remove();

            if (file.files.length < 1) {
                pinegrow.showQuickError('Please select one or more image files.');
                return false;
            }

            var showError = function(msg, sel) {
                if(!sel) {
                    $d.find('.modal-body').prepend($(`<div class="error alert alert-danger">${msg}</div>`))
                } else {
                    var $msg = $d.find(`[data-file="${fileitem.name}"]`);
                    $msg.append(' ' + msg).removeClass('alert-info').addClass('alert-error');
                }
            }

            var allowed_types = ['image/jpeg', 'image/png', 'image/jpg', 'image/gif'];
            if(pinegrow.currentUser.isAdmin()) {
                allowed_types.push('image/svg');
                allowed_types.push('image/svg+xml');
            }

            var c = 0;
            var has_errors = false;

            for(var i = 0; i < file.files.length; i++) {
                var fileitem = file.files[i];
                var file_type = fileitem.type;

                if (allowed_types.indexOf(file_type) < 0) {
                    showError('Only JPG, PNG and GIF images can be uploaded.');
                    return false;
                }
            }

            $d.find('button.ok').attr('disabled', 'disabled').html('Uploading...');

            for(var i = 0; i < file.files.length; i++) {
                (function(fileitem) {
                    var $pitem = $(`<div data-file="${fileitem.name}" class="pg-upload-file-item error alert alert-info" style="margin-top:10px;margin-bottom: 10px;"><div class="upload-progress"></div><p>Uploading ${fileitem.name}....<span></span></p></div>`);

                    $d.find('.modal-body').append($pitem)

                    var $bar = $pitem.find('.upload-progress');

                    var updateProgress = function(p, msg, done) {
                        $bar.css('width', p + '%');
                        if(msg) {
                            $pitem.find('span').html(`<b> ${msg}</b>`);
                        }
                        if(done) {
                            $pitem.removeClass('alert-info').addClass('alert-success');
                        }
                    }

                    updateProgress(0);

                    pinegrow.backend.uploadFile(fileitem, function(data) {
                        c++;
                        if(data.data.error) {
                           showError(`The file could not be uploaded. ${data.data.error}`, fileitem.name);
                            has_errors = true;
                        } else {
                            current_data.results.unshift(data.data.image);
                            get$ItemForImage(data.data.image).insertAfter($items.find('.pg-image-browser-upload'));

                            $d.find(`[data-file="${fileitem.name}"]`);
                            updateProgress(100, 'Done.', true);

                            if(c >= file.files.length && !has_errors) {
                                $d.modal('hide');
                                /*
                                $d.find('button.ok').html('Upload').get(0).removeAttribute('disabled');
                                $d.find('form').get(0).reset();

                                 */
                            }

                            var p = pinegrow.getCurrentProject();
                            if(data.data.file) {
                                p.setContentOfUrl(data.data.file.source_url, data.data.file.content);
                                p.addSourceUrlToDOMUrlMap(data.data.file.source_url, data.data.file.mapped_url);
                                pinegrow.refreshCurrentProjectDelayed(true);
                            }
                        }
                    }, function(error) {
                        //error
                        c++;
                        showError(`The file could not be uploaded. ${error}`, fileitem.name);
                        has_errors = true;
                    }, function(progress) {
                        updateProgress(progress, progress === 100 ? 'Processing...' : '');
                    });
                })(file.files[i]);
            }

            return false;
        });
    })

    $container.on('click', '.pg-image-browser-item, .pg-image-browser-item-size', function(e) {
        if($(e.target).is('a.external')) return;

        e.preventDefault();
        e.stopPropagation();

        pgCloseContextMenus();

        var $pic = $(this);

        var pic = $pic.data('picdata');
        var pic_size = $pic.data('picsize');
        var picurl = picapi.getImageUrlForInsertingToPage(pic, pic_size);
        var srcset = picapi.getImageSrcset(pic);

        var replaced = false;
        var col = new PgCollection();

        pinegrow.selectedElements.forEachElement(function(pgel) {
            if(pgel.tagName === 'img') {
                col.add(pgel);
            }
        })

        destroyPopup();

        if(mark_images) {
            markToggle(pic.id, pic.urls.raw);
            var $item = $pic.closest('.pg-image-browser-item');

            if(isMarked(pic.id)) {

                $item.css({
                    outline: '4px solid yellow',
                    'outline-offset': '-4px'
                })
            } else {
                $item.css({
                    outline: 'none',
                })
            }
            return;
        }

        if(_this.onImageClicked) {
            _this.onImageClicked(picurl, pic, srcset);
            return;
        }

        if(col.getLength() === 0) {
            $popel = $pic;//.closest('.pg-image-browser-c');

            pinegrow.showQuickError('Drag the image to the page. Click to replace the selected image. No image is selected.');
            return;
        } else {
            pinegrow.makeChanges(col, 'Set image', function() {
                col.forEachElement(function(pgel) {
                    pgel.setAttribute('src', picurl);
                    if(srcset) {
                        pgel.setAttribute('srcset', srcset);
                    } else if(pgel.hasAttribute('srcset')) {
                        pgel.removeAttribute('srcset');
                    }
                })
            })
            pinegrow.selectedElements.reselect();

            pinegrow.showQuickMessage('Replacing image...', null, true);
            used(pic);
        }

        pinegrow.dispatchEvent('on_image_browser_image_selected', null, picurl /* picapi.getImageUrl(pic) */, pic)

        return false;
    })

    $container.on('contextmenu', '.pg-image-browser-item,.pg-image-browser-item-size', function(e) {
        e.preventDefault();
        var $pic = $(this);

        var pic = $pic.data('picdata');
        var pic_size = $pic.data('picsize');
        var pic_responsive = false;

        if(pic_size.endsWith('_r')) {
            pic_responsive = true;
            pic_size = pic_size.replace('_r', '');
        }

        var picurl = picapi.getImageUrlForInsertingToPage(pic, pic_size);

        if(pic_responsive) {
            var srcset = picapi.getImageSrcset(pic);
        }

        var menu = new CrsaContextMenu(null, $pic);

        var pgel = pinegrow.selectedElements.getLastSelected();

        destroyPopup();

        if(pgel) {
            var name = pinegrow.selectedElements.getName({short: true, num_classes: 0});

            if(pgel.tagName == 'img') {
                menu.add({
                    type: 'divider'
                })
                menu.add({
                    type: 'header',
                    label: 'Replace'
                })

                menu.add("Replace image " + name, null, function() {
                    pinegrow.makeChanges(pinegrow.selectedElements.getCollection(), 'Set image', function() {
                        pinegrow.selectedElements.forEachElement(function(pgel) {
                            if(pgel.tagName == 'img') {
                                pgel.setAttribute('src', picurl)
                            }
                            if(srcset) {
                                pgel.setAttribute('srcset', srcset);
                            } else if(pgel.hasAttribute('srcset')) {
                                pgel.removeAttribute('srcset');
                            }
                        })
                    })
                    used(pic);
                });
            } else {
                menu.add({
                    type: 'divider'
                })
                menu.add({
                    type: 'header',
                    label: 'Background'
                })

                menu.add("Set background for " + name, null, function() {
                    pinegrow.makeChanges(pinegrow.selectedElements.getCollection(), 'Set background image', function() {
                        pinegrow.selectedElements.forEachElement(function(pgel) {
                            if(pgel.tagName !== 'img') {
                                pgel.setInlineStylePropertyValue('background-image', picurl, true);
                                pinegrow.dispatchEvent('on_css_active_rule_changed', null);
                            }
                        })
                    })
                    used(pic);
                });
            }

            menu.add({
                type: 'divider'
            })
            menu.add({
                type: 'header',
                label: 'Insert'
            })

            var insert = function(placement) {
                var api = new PgApi();
                if(api.insertElements(picapi.getImageTagForInsertingToPage(pic, pic_size, pic_responsive), null /* all selected */, placement, { repeater: true}).getLength()) {
                    pinegrow.showQuickMessage('Image inserted.');
                }
            }

            if(pgel.canHaveChildren()) {
                menu.add('Prepend to ' + name, null, function() {
                    insert('prepend');
                })
                menu.add('Append to ' + name, null, function() {
                    insert('append');
                })
            }
            menu.add('Insert before ' + name, null, function() {
                insert('insertBefore');
            })
            menu.add('Insert after ' + name, null, function() {
                insert('insertAfter');
            })

            menu.add({
                type: 'divider'
            })
            menu.add({
                type: 'header',
                label: 'Clipboard'
            })
            menu.add('Copy image URL', null, function () {
                copyCodeToClipboard(picurl, true);
            })

        } else {
            menu.add('Select one or more elements on the page first.', null, function() {})
        }

        if(picapi.supports_delete) {
            menu.add({
                type: 'divider'
            })
            menu.add({
                label: 'Delete...',
                action: function() {
                    pinegrow.showAlert(`<p>Do you want to delete this image? Deleting the image will remove the image from all your projects, and it can not be undone.</p>`, 'Are you sure about deleting this image?', 'Cancel', 'Yes, delete it', null, function() {
                        pinegrow.showQuickMessage('Deleting the image...');
                        pinegrow.backend.deleteFile(pic.id, function() {
                            pinegrow.showQuickMessage('Deleted.');
                            $pic.remove();
                            var idx = current_data.results.indexOf(pic);
                            if(idx >= 0) current_data.results.splice(idx, 1);
                        }, function(error) {
                            pinegrow.showQuickError('Oops... could not delete it.')
                        })
                    })
                }
            })
        }

        if(pgf.use_project_featured_images && pic.id && picapi.supports_delete) {
            menu.add({
                type: 'divider'
            })

            menu.add({
                label: 'Set as project featured image',
                action: function() {
                    pinegrow.backend.setProjectFeaturedImage(pinegrow.getCurrentProject(), pic.id, function() {
                        pinegrow.showQuickMessage('Featured image set.');
                    })
                }
            })
        }

        menu.showAt(e.pageX, e.pageY);
        return false;
    })

    var dragMenu = null;
    var dragged_pic = null;

    var dragHelper = new PgDragHelper($container, '.pg-image-browser-item,.pg-image-browser-item-size', null, true);

    dragHelper.onBefore = function() {
        return _this.drag_enabled;
    }

    dragHelper.on( 'dragStart', function(x, y, event) {
        $.fn.crsapages('showOverlays');

        var $pic = dragHelper.dragTarget;
        var pic = $pic.data('picdata');
        var pic_size = $pic.data('picsize');
        var pic_responsive = false;

        if(pic_size.endsWith('_r')) {
            pic_responsive = true;
            pic_size = pic_size.replace('_r', '');
        }

        dragged_pic = pic;

        destroyPopup();

        //console.log(`image ${pic_size}`)

        var pgel = pgCreate(picapi.getImageTagForInsertingToPage(pic, pic_size, pic_responsive));

        var def = null;
        var page = pinegrow.getSelectedPage();
        if(page) {
            def = page.getMainType(null, pgel);
        }

        dragMenu = new PgDragMenu(def, false, true);

        dragMenu.pgel = pgel;
        dragMenu.selectNew = true;

        dragMenu.handleMoveEvent(event, dragMenu);

        $element.trigger('pg-drag-start');

    }, this );

    dragHelper.on( 'drag', function(x, y, event) {

        dragMenu.handleMoveEvent(event, dragMenu);
    }, this );

    dragHelper.on( 'dragStop', function(x, y) {
        $.fn.crsapages('showOverlays', true);

        if(dragMenu.insert()) {
            used(dragged_pic);
        }
        dragMenu.destroy();
        dragMenu = null;
        dragged_pic = null;

        crsaFuncs.showInsertionLine(null);
        pinegrow.highlightElement(null);

        $element.trigger('pg-drag-stop');

    }, this );

    var get$ItemForImage = function(r) {
        var def_size = picapi.getDefaultSize();

        var info = '';

        if(r.user) {
            info = `<p><a href="${picapi.getLink(r.user.links.html)}" class="external" target="_blank">${r.user.name}</a> on <a href="${picapi.getLink('https://unsplash.com/')}" class="external" target="_blank">Unsplash</a>.</p>`;
        } else if(r.name) {
            info = `<p style="text-align: left;padding-left: 8px;background: rgba(0,0,0,0.6);">${r.name}</p>`;
        }

        var $item = $('<div draggable class="pg-image-browser-item"><img src="' + picapi.getImageUrl(r, 'small', null, true)+ '">' + info + '</div>');
        $item.data('picdata', r);
        $item.data('picsize', def_size);

        var $sizes = $('<div class="pg-image-browser-item-sizes"></div>').appendTo($item);

        var sizes = picapi.getImageSizes(r);
        var num_sizes = 0;

        if(_this.show_sizes) {
            $.each(sizes, function (size, dim) {
                var $s = $(`<div draggable class="pg-image-browser-item-size">${size} ${dim.w > 0 ? dim.w : '0'}&times;${dim.h > 0 ? dim.h : '0'}</div>`).appendTo($sizes);
                $s.data('picdata', r);
                $s.data('picsize', size);
                num_sizes++;
            })

            if (picapi.getImageSrcset(r)) {
                var $s = $(`<div draggable class="pg-image-browser-item-size">Responsive</div>`).appendTo($sizes);
                $s.data('picdata', r);
                $s.data('picsize', picapi.getDefaultSize() + '_r');
                num_sizes++;
            }
        }
        if(num_sizes <= 1) {
            pg$Hide($sizes);
        }

        if(mark_images) {
            //console.log(r);
            if(isMarked(r.id)) {
                $item.css({
                    outline: '4px solid yellow',
                    'outline-offset': '-4px'
                })
            } else {
                $item.css({
                    outline: 'none',
                })
            }
        }
        return $item;
    }

    var updateResults = function(data) {

        var show_upload = false;

        if(!data) {
            if(current_data) {
                data = current_data;
                $items.empty();
            } else {
                return;
            }
        } else {
            current_data.results = current_data.results.concat(data.results);
            current_data.total_pages = data.total_pages;
        }

        show_upload = picapi.supports_add_image && $items.get(0).children.length === 0;

        if(show_upload) {
            $(`<div class="pg-image-browser-upload"><i class="icon icon-plus"></i>Upload images</div>`).appendTo($items);
        }

        data.results.forEach(function(r) {
            get$ItemForImage(r).appendTo($items);
        })

        total_pages = data.total_pages;

        updateMoreButton();
    }

    var updateMoreButton = function(loading) {
        if(loading) {
            $more.attr('disabled', 'disabled');
            $more.html('Loading...');
        } else {
            $more.removeAttr('disabled');
            if(total_pages > 0) {
                if (total_pages <= page) {
                    pg$Hide($more);
                } else {
                    pg$Show($more);
                    $more.html(`Show more - <span>${total_pages - page}</span> pages to go`);
                }
            } else {
                pg$Show($more);
                $more.html(`Show more`);
            }
        }
    }

    var onElementSelected = function(page, pages, info) {
    }

    var onPageChanged = function(page, pgel, info) {
    }

    view.onResize = function() {
    }

    //pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
    //pinegrow.addEventHandler('on_page_changed', onPageChanged);

    var onImageCatsLoaded = function() {
        if(safe_images && !search_term) {
            setTimeout(function() {
                var defcat = safe_images.getDefaultCategory();
                defcat && search.setValue(defcat);
            }, 20)
        }
    }

    pinegrow.addEventHandler('on_image_categories_loaded', onImageCatsLoaded);

    view.onDestroy = function() {
        pinegrow.removeEventHandler('on_image_categories_loaded', onImageCatsLoaded);
       // pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected);
        //pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        destroyPopup();
    }

    onImageCatsLoaded();
}

var PgQuickImageBrowser = function($target, title) {
    var _this = this;

    var v = this.imageBrowser = new PgImageBrowserView();

    var win = new PgQuickWindow(title || (pgf.kids ? 'Image browser' : 'Search photos on Unsplash'), v, 'quickImageBrowser', 440, 600, null, true);

    win.noRightEdge();
    win.hideOnClose = true;

    $target = $target || $('body');

    if($target) {
        win.showFor$El($target);
    }

    v.onShown();

    //v.show();

    win.onHide = function() {
        if(_this.with_focus_rect) {
            pinegrow.focusRect.hide();
        }
        if(_this.onHide) _this.onHide();
    }

    var onFocusRectClosed = function() {
        if(_this.with_focus_rect) {
            win.hide();
        }
    }

    pinegrow.addEventHandler('on_focus_rect_closed', onFocusRectClosed);

    win.onDestroy = function() {
        pinegrow.removeEventHandler('on_focus_rect_closed', onFocusRectClosed);
    }

    this.setOnSelected = function(func) {
        v.onImageClicked = func;
    }

    this.show = function($target) {
        win.show(true);
        if($target) {
            win.showFor$El($target);
        }
        pgQuickWindowManager.focus(win);
    }

    this.getSearchTerm = function() {
        return v.getSearchTerm();
    }

    this.openCategoryDropdown = function() {
        setTimeout(function() {
            v.view.get$Element().find('.pg-search-box .value-holder').trigger('click');
        }, 200)
    }
}

var pgQuickImageBrowser = null;

var pgShowQuickImageBrowser = function($target, with_focus_rect, on_selected) {

    if(!pgQuickImageBrowser) {
        pgQuickImageBrowser = new PgQuickImageBrowser($target);
    } else {
        pgQuickImageBrowser.show($target);
    }
    pgQuickImageBrowser.imageBrowser.onImageClicked = null;
    pgQuickImageBrowser.setOnSelected(on_selected);
    pgQuickImageBrowser.with_focus_rect = with_focus_rect;
    return pgQuickImageBrowser;
}

/*
var PgResourceLocalizer = function(url) {

    var _this = this;

    this.url = url;

    var project = pinegrow.getCurrentProject();
    var outside_of_project = project && !project.isUrlInProject(url);

    this.isLocal = function() {
        return !outside_of_project;
    }

    var showError = function(msg) {
        pinegrow.showAlert(`<p>Could not copy ${_this.url}:</p><div>${msg}</div>`, 'Could not copy the resource');
    }

    this.localize = function(done) {
        pinegrow.httpServer.readRemoteUrl(url, function(error, buffer, mime) {
            if(error) {
                showError(error);
            } else {

            }
        })
    }
}
*/

class PgRandomLocalImage {

    constructor() {
        this.cache = {}
    }

    getRandomImageFromFolder(folder, done) {
        if(!this.cache[folder]) {
            const fs = require('fs');
            try {
                var list = fs.readdirSync(folder, {
                    withFileTypes: true
                });
            } catch(err) {
                pinegrow.showQuickError(`Could not read content of the folder ${folder}.`);
            }
            var images = list.filter(function(de) {
                return (de.isFile() && pinegrow.getSimpleMimeType(de.name) === 'image');
            }).map(function(de) {
                return de.name;
            });

            this.cache[folder] = {
                all: images,
                current: []
            }
        }
        if(this.cache[folder].current.length === 0) {
            this.cache[folder].current = this.cache[folder].all.slice(0);
        }
        if(this.cache[folder].current.length === 0) {
            pinegrow.showQuickError(`No images were found in ${folder}.`);
            done && done(null);
            return;
        }

        var idx = Math.floor(Math.random() * (this.cache[folder].current.length - 1));
        var i = this.cache[folder].current[idx];
        this.cache[folder].current.splice(idx, 1);

        i = crsaMakeUrlFromFile( require('path').join(folder, i));
        done && done(i);
    }

    refresh() {
        this.cache = {}
    }
}