class PgWindowMenu {
    constructor() {
        var menuView = this.menuView = new PgDropdownMenu($('.menu-window'));
        menuView.onGetActions = function(menu) {



            var panels_menu = []

            var cat = null;
            var first = true;

            var isVisible = function(d) {
                if(d.on_is_visible) {
                    return d.on_is_visible();
                }
                return !pinegrow.isHiddenUIPanel(d.key);
            }

            var showPanel = function(d, reveal) {
                if(d.on_show) {
                    d.on_show();
                } else {
                    pinegrow.showUIPanel(d.key, false, false, reveal);
                }
            }

            var hidePanel = function(d) {
                if(d.on_hide) {
                    d.on_hide();
                } else {
                    pinegrow.hideUIPanel(d.key);
                }
            }

            pinegrow.panelManager.forEachPanel(function(d) {
                if(cat !== d.category && d.category) {
                    if(!first) {
                        panels_menu.push({
                            type: 'divider'
                        })
                    }
                    panels_menu.push({
                        type: 'header',
                        label: d.category
                    })
                    first = false;
                    cat = d.category;
                }
                var name = d.name;
                if(d.icon) {
                    name = `<i class="icon ${d.icon}" style="color:rgba(255,255,255,0.55);margin-right:2px;position:relative;top:2px;"></i> ` + name;
                }
                panels_menu.push({
                    label: name,
                    check: function() {
                        return isVisible(d)
                    },
                    action: function() {
                        var visible = isVisible(d);

                        if(!visible) {
                            showPanel(d, true)
                        } else {
                           hidePanel(d)
                        }
                    },
                    helptext: d.description,
                    kbd: d.kbd || null
                })
            })

            panels_menu.push({
                type: 'divider'
            })

            panels_menu.push({
                label: 'Show all',
                action: function() {
                    pinegrow.panelManager.forEachPanel(function(d) {
                        if(!isVisible(d)) {
                            showPanel(d, false)
                        }
                    })
                }
            })

            panels_menu.push({
                label: 'Hide all',
                action: function() {
                    pinegrow.panelManager.forEachPanel(function(d) {
                        if(isVisible(d)) {
                            hidePanel(d)
                        }
                    })
                }
            })

            menu.add({
                label: 'Show / hide panels',
                submenu: panels_menu
            })

            var workspace_menu = []

            workspace_menu.push({
                type: 'header',
                label: 'Layout'
            })

            pgGetWorkspaces().forEach(function (s) {
                workspace_menu.push({
                    label: s.name,
                    action: function (none, e, $li) {
                        pinegrow.showWorkspace(s.key);
                    }
                })
            })

            workspace_menu.push({
                type: 'divider'
            })

            workspace_menu.push({
                type: 'header',
                label: 'My layouts'
            })

            pgCustomWorkspaces.getList().forEach(function(item) {
                var a = {
                    label: item.name,
                    action: function() {
                        pinegrow.showWorkspace(item);
                    },
                    check: '<i class="icon icon-close"></i>',
                    check_action: function(e) {
                        pgCustomWorkspaces.remove(item.name);
                        e.preventDefault();
                        e.stopPropagation();
                        menu.close();
                        pinegrow.showQuickMessage('<b>' + item.name + '</b> was removed.');
                    },
                    check_tooltip: 'Remove'
                }

                workspace_menu.push(a);
            })

            workspace_menu.push({
                label: 'Save the current layout...',
                action: function (none, e, $li) {
                    pinegrow.showPrompt('<p>Please enter the name for the saved workspace layout. Use the name of an existing layout to update that layout.</p>', 'Save workspace layout', '', '', null, function(value) {
                        value = $.trim(value);
                        if(value) {
                            pgCustomWorkspaces.save(value);
                            pinegrow.showQuickMessage('The workspace layout <b>' + value + '</b> was saved.');
                        }
                    });
                }
            })

            workspace_menu.push({
                type: 'divider'
            })

            workspace_menu.push({
                label:'Refresh the current layout',
                action:function (none, e, $li) {
                    var conf = $.extend(true, myLayout.toConfig(), {});
                    pinegrow.showWorkspace({
                        config: JSON.stringify(conf),
                        ccmode: pinegrow.isContributorMode()
                    });
                }
            })

            menu.add({
                label: 'Workspace',
                submenu: workspace_menu
            })

            if(isApp() || true) {
                var theme_menu = []

                var themes = [
                    {
                        name: 'Dark gray (default)',
                        key: 'gray'
                    },
                    {
                        name: 'Light',
                        key: 'light'
                    }
                ]

                var current_theme = pinegrow.getWorkspaceTheme();

                themes.forEach(function (t) {
                    theme_menu.push({
                        label: t.name,
                        action: function (none, e, $li) {
                            pinegrow.setWorkspaceTheme(t.key);
                        },
                        check: function () {
                            return current_theme == t.key;
                        }
                    })
                })

                menu.add({
                    label: 'Theme',
                    submenu: theme_menu
                })
            }

            if(isApp()) {

                var zoom_menu = [
                    {
                        type: 'header',
                        label: 'Make UI smaller or larger'
                    }
                ]

                var zoomUISizes = [
                    {key: -2, name: '50%'},
                    {key: -1, name: '75%'},
                    {key: 0, name: '100%'},
                    {key: 1, name: '125%'},
                    {key: 2, name: '150%'},
                    {key: 4, name: '200%'},
                ].forEach(function(s) {
                    zoom_menu.push({
                        label: s.name,
                        action: function(none, e, $li) {
                            setUIZoom(s.key);
                            pinegrow.setSetting('ui-zoom', s.key);
                        },
                        check: function() {
                            return getUIZoom() == s.key;
                        }
                    })
                })

                menu.add({
                    label: 'UI Zoom',
                    submenu: zoom_menu
                })
            }

            menu.add({
                type: 'divider'
            })

            var open_pages = []

            var project = pinegrow.getCurrentProject();

            pinegrow.getAllPages().forEach(function(page) {
                var url = project ? project.getRelativeUrl(page.url) : page.url;
                url = pgUrlToDisplayString(url);

                var l = shortenString(url, 40, '&hellip;', true);
                if(page.hasChanges()) {
                    l += ' *';
                }
                open_pages.push({
                    label: l,
                    action: function() {
                        pinegrow.openOrShowPage(page.url);
                    }
                })
            })

            if(open_pages.length === 0) {
                open_pages.push({
                    type: 'header',
                    label: 'No documents are open.'
                })
            }

            menu.add({
                label: 'Open documents',
                submenu: open_pages
            })

            if(pgIsDev() && false) {

                menu.add({
                    type: 'divider'
                })

                menu.add({
                    type: 'header',
                    label: 'Dev mode'
                })


                menu.add({
                    label: 'Enable product',
                    submenu: [
                        {
                            label: 'Web Editor',
                            check: pgGetApp() === 'PG',
                            action: function() {
                                pinegrow.setSetting('app', 'PG');
                                pinegrow.restartPinegrow();
                            }
                        },
                        {
                            label: 'Magic',
                            check: pgGetApp() === 'MAGIC',
                            action: function() {
                                pinegrow.setSetting('app', 'MAGIC');
                                pinegrow.restartPinegrow();
                            }
                        },
                        {
                            label: 'WP Converter',
                            check: pgGetApp() === 'PGTC',
                            action: function() {
                                pinegrow.setSetting('app', 'PGTC');
                                pinegrow.restartPinegrow();
                            }
                        },
                        {
                            label: 'Snap!',
                            check: pgGetApp() === 'SNAP',
                            action: function() {
                                pinegrow.setSetting('app', 'SNAP');
                                pinegrow.restartPinegrow();
                            }
                        }
                    ]
                })
            }

            if(pgf.hide_panels) {
                menu.add({
                    type: 'divider'
                })

                menu.add({
                    label: 'Quickly hide panels',
                    kbd: pgShowKbd('TAB'),
                    check: pgArePanelsQuicklyHidden(),
                    action: function() {
                        pgToggleFloatingPanels();
                    },

                })
            }
        };

    }
}