var PgAttributesEditorView = function(standalone) {

    var api = new PgApi();

    var $container = $('<div class="attributes-panel"></div>');

    $('<div class="alert alert-info">This element has no attributes</div>').prependTo($container);


    var view = new PgUIView($container);

    this.view = view;
    var _this = this;

    var attributeOptions = {
        'add-new-placeholder-text': 'Add attribute',
        'item-length-to-add-placeholder': 100
    };
    var crsaSelectBoxOption = {
        itemClass: 'crsa-node-item'
    };

    var pgFastEditor = new PgFastEditor($container, attributeOptions);
    var crsaSelectBox = new CrsaSelectBox($container, crsaSelectBoxOption);

    attributeOptions.isAllowedToShow = function (name, value) {
        if (name.trim() == 'data-pg-collapsed') return false;
        /*
        if(value && value !== escapeHtmlCode(value)) {
            return false;
        }*/
        return true;
    }

    attributeOptions.onItemChanged = function (name, value, changeType, oldValue) {
        if (changeType == "prop") {
            api.removeAttribute(null, oldValue, {source: _this});
        }

        if(value === '') value = null;

        api.setAttribute(null, name, value, {source: _this});

        pinegrow.dispatchEventDelayedForObject(current_pgel, 'on_prop_field_changed', null, current_pgel, {source: _this})
        return true;
    }

    attributeOptions.onItemAdded = function (name, value) {
        var pgel = pinegrow.getSelectedElement();
        if(pgel) {
            if(value === '') value = null;
            api.setAttribute(null, name, value, {source: _this});
            pinegrow.dispatchEventDelayedForObject(current_pgel, 'on_prop_field_changed', null, current_pgel, {source: _this})
        }
        return true;
    }

    attributeOptions.onItemRemoved = function (name, value) {
        var pgel = pinegrow.getSelectedElement();
        if(pgel) {
            api.removeAttribute(null, name, {source: _this});
            pinegrow.dispatchEventDelayedForObject(current_pgel, 'on_prop_field_changed', null, current_pgel, {source: _this})
        }
        return true;
    }

    attributeOptions.parseCodeAndGetArray = function (code) {
        var pgel = pinegrow.getSelectedElement();
        var attrs = [];

        if (pgel) {
            var arr = [];
            pgel.getAttrList().forEach(function(a) {
                var attr = {
                    prop: a.name,
                    value: a.value || ''
                }
                //pgel.setAttr(attr.prop, attr.value); ???
                attrs.push(attr);
            })
        }

        return attrs;
    }

    var current_pgel = null;

    var onElementSelected = function(event, pgel) { //todo30 - el selected
        current_pgel = pgel;
        view.whenVisible(function() {
            if(pgel) {
                var cant = [];
                var attributes = pgel.getAttrList();

                var list = [];
                for (var i = 0; i < attributes.length; i++) {
                    var attr = attributes[i];

                    var a = {
                        prop: attr.name,
                        value: attr.value || ''
                    }

                    if(attr.value && (attr.value.indexOf('>') >= 0 || attr.value.indexOf('<') >= 0)) {
                        cant.push(attr.name);
                    } else {
                        list.push(a);
                    }
                }

                pgFastEditor.setList(list);
                pgFastEditor.show();
                $container.show();

                if(cant.length) {
                    var $cantedit = $('<div class="alert alert-info">Can\'t edit attributes with special characters: <span>' + cant.join(', ') + '</span>. Please use element code edit.</div>').appendTo($container);
                    $cantedit.prependTo($container);
                }
            } else {
                $container.hide();
            }
        })
    }

    this.show = function(pgel) {
        onElementSelected(null, pgel);
    }

    var onPageChanged = function(page, pgel, info) {
        if(pgel && pgel == current_pgel && (!info || !info.source || info.source != _this)) {
            onElementSelected(null, current_pgel);
        }
    }

    pinegrow.addEventHandler('on_page_changed', onPageChanged);

    if(standalone) {
        view.on('crsa-element-selected', onElementSelected);
    }

    view.onDestroy = function () {
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
    }

}


//PROP control version of the attr editor

var pgCreateAttributesEditorControl = function() {
    //will keep helper functions and vars private

    var editor = new PgAttributesEditorView(false);
    editor.view.alwaysVisible = true;
    editor.view.updateVisibility();

    //Create control, needs a unique id
    var c = new PgCustomPropertyControl('pg-attribute-editor');

    //Register subfields.
    c.onDefine = function() {
        //no subfields
    }

    //Show control. Return the control $el.
    c.onShow = function() {
        return editor.view.get$Element();
    }

    //Called when control is recycled and new values are set
    c.onSetValues = function() {
        editor.show(c.obj);
    }

    c.onHasData = function() {
        return c.obj.getAttrList().length > 0;
    }

    //Return the control
    return c;
};

