var PgQuickCSSRuleCodeEdit = function(rule, $target) {

    var editor = new PgCSSRuleCodeEditView(rule);

    var cs = pgStylesheetHelper.getStylesheetFromNode(rule);

    if(!cs.canEdit(true)) {
        editor.setReadOnly(true)
    }

    var win = new PgQuickWindow(cs.filterSource(rule.selector) || '' + ' | Edit CSS code', editor, 'quickCSSRuleCodeEdit', 570, 260);

    editor.setWindow(win);
    win.$element.addClass('pg-css-rule-code-editor');

    win.showFor$El($target);

    editor.show();

    win.onDestroy = function() {
    }

}

var PgCSSRuleCodeEditView = function(rule) {

    var $element = $('<div class=""><div class="pg-element-code-editor"></div><div class="pg-element-code-editor-controls"></div></div>');

    var $content = $element.find('.pg-element-code-editor');

    var notice = new PgCodePanelNotice();
    notice.hideOnEmpty = true;
    notice.get$Element().appendTo($element);

    var view = this.view = new PgUIView($element);

    var _this = this;

    var settingsPrefix = 'edit-css-rule-html';
    var editorData = null;

    var cs = pgStylesheetHelper.getStylesheetFromNode(rule);

    var oldRule = rule;

    var ruleObj = {
        pgStylesheet: cs,
        node: rule,
        nodeIndices: pinegrow.pgPostCSSHelper.getNodeIndices(rule)
    }

    var ignore_change = false;
    var ignore_select_element = false;
    var has_changes = false;
    var has_content = false;
    var has_errors = false;
    var skipUpdatingIndices = false;

    var wrap = true;

    var code_ignore_change_timeout = null;

    var settingRuleCode = false;
    var code_changed = false;

    this.win = null;

    var getNodeUsingRuleObj = function () {
        return pinegrow.pgPostCSSHelper.getNodeByIndices(ruleObj.pgStylesheet.getRootNode(), ruleObj.nodeIndices);
    }

    var updateRuleSource = function (obj, deleted) {
        var rule = getNodeUsingRuleObj();

        if (pinegrow.pgPostCSSHelper.isNodeEqualOrChildOf(rule, obj.node)) {
            if (deleted) {
                var index = rule.nodes.indexOf(obj.node);
                var clonedRule = rule.clone();
                clonedRule.nodes.splice(index, 1);
                rule = clonedRule;
            }

            setRuleSource(rule);
        }
    }

    var onNodeAdded = function(page, obj) {
        if (skipUpdatingIndices) return;

        updateRuleIndices(ruleObj, obj, 1, 'added');
        updateRuleSource(obj);
    }

    var onNodeUpdated = function(page, obj) {
        updateRuleSource(obj);
    }

    var onNodeBeforeDeleted = function(page, obj) {
        if (skipUpdatingIndices) return;

        updateRuleIndices(ruleObj, obj, -1, 'deleted');
        updateRuleSource(obj, true);
    }

    pinegrow.addEventHandler('on_css_node_added',   onNodeAdded);
    pinegrow.addEventHandler('on_css_node_updated', onNodeUpdated);
    pinegrow.addEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);


    view.onResize = function() {
        if(editorData) editorData.mirror.layout();
    }

    var restore_on_cancel = true;
    var skipUpdate = false;

    var onRulesChangedInExternalEditor = function(p, data) {
        if (data && data.list && data.eventType && data.eventType == 'editor' && data.editor == 'external_editor') {
            if (data.list.indexOf(cs) >= 0) {
                pinegrow.showQuickMessage('Stylesheet was modified in external editor.');
                restore_on_cancel = false;
            }
        }
        if (data.list.indexOf(cs) >= 0) {
            skipUpdate = true;
            // _this.win.destroy();
        }
    }

    pinegrow.addEventHandler('on_css_source_changed', onRulesChangedInExternalEditor);

    var onCSSLockedChanged = function(pagenull, cs_param, locked) {
        if(cs === cs_param) {
            _this.setReadOnly(locked);
        }
    }

    pinegrow.addEventHandler('on_css_stylesheet_locked_changed', onCSSLockedChanged);

    view.onDestroy = function() {
        if (rule_change_timeout) {
            clearTimeout(rule_change_timeout);
            if (!skipUpdate) doUpdate();
        }
        pinegrow.code_editors.unregister(CodeEditorVsc);
        pinegrow.removeEventHandler('on_css_node_added',   onNodeAdded);
        pinegrow.removeEventHandler('on_css_node_updated', onNodeUpdated);
        pinegrow.removeEventHandler('on_css_before_node_deleted', onNodeBeforeDeleted);
        pinegrow.removeEventHandler('on_css_stylesheet_locked_changed', onCSSLockedChanged);

        if(noticeTimer) {
            clearTimeout(noticeTimer);
            noticeTimer = false;
        }
        setNotice('');
    }

    this.setWindow = function (win) {
        this.win = win;
    }

    this.show = function() {
        view.whenVisible(function () {

            if (!editorData) {
                editorData = makeEditor();
            }
            editRuleSource();
        })
    }

    // pinegrow.addEventHandler('on_element_selected', onElementSelected);
    // pinegrow.addEventHandler('on_page_changed', onPageChanged);

    var noticeTimer = null;

    var setNotice = function(msg) {
        if(noticeTimer) {
            clearTimeout(noticeTimer);
            noticeTimer = false;
        }
        if(!msg) {
            notice.setMessage('');
        } else {
            noticeTimer = setTimeout(function() {
                var cp = CodeEditorVsc.getPosition();
                var si = { left: CodeEditorVsc.getScrollLeft(), top: CodeEditorVsc.getScrollTop() };
                // cp.top -= si.top;
                var top = CodeEditorVsc.getContentHeight() - si.top * 2 < 60;

                notice.setMessage(msg, top);
                noticeTimer = null;
            }, 750)
        }
    }

    var onChange = function() {
        applySource(editorData.mirror.getValue());
    }

    var onBlur = function () {
        var pgStylesheet = ruleObj.pgStylesheet;
        var node = getNodeUsingRuleObj();

        // skipUpdatingIndices = true;
        // pinegrow.dispatchEvent('on_css_before_node_deleted', null, {node: node, stylesheet: pgStylesheet});
        // pinegrow.dispatchEvent('on_css_node_added', null, {node: node, stylesheet: pgStylesheet});
        // skipUpdatingIndices = false;
        if (code_changed) {
            pinegrow.dispatchEvent('on_css_node_replaced', null, {oldNode: oldRule, node: node, stylesheet: pgStylesheet});
            pinegrow.selectRule(node);
        }
        code_changed = false;

        // $('body').trigger('crsa-stylesheets-changed', {list: [cs]});
        // pinegrow.reselectElement();
    }

    var rule_change_timeout = null

    var doUpdate = function() {
        if(read_only) return;

        if (settingRuleCode) {
            settingRuleCode = false;
            return;
        }

        var s = CodeEditorVsc.getValue();
        s = $.trim(s);
        var rule = pinegrow.pgPostCSSHelper.getNodeByIndices(cs.getRootNode(), ruleObj.nodeIndices);

        cs.updateNodeStylesheetWithSource(rule, s, function (newRule) {
            code_changed = true;
            if (newRule) rule = newRule;

            var error = cs.getError();
            rule_change_timeout = null;

            if (error) {
                notice.setMessage(error);
            } else {
                notice.setMessage('');
            }

            $('body').trigger('crsa-rule-changed-in-editor', {list: [cs]});
            pinegrow.dispatchEvent('on_css_rule_changed_in_code_editor', null, {
                list: [cs],
                node: rule
            })
        });
    }

    var applySource = function() {
        if(ignore_change) return false;

        if (rule_change_timeout) {
            clearTimeout(rule_change_timeout);
        }

        rule_change_timeout = setTimeout(doUpdate, 300);
    }

    this.setWrap = function(w) {
        wrap = w;
        if(CodeEditorVsc) {
            CodeEditorVsc.updateOptions({wordWrap: w ? "on": "off"});
        }
    }

    var read_only = false;

    this.setReadOnly = function(val) {
        read_only = val;
        if(CodeEditorVsc) {
            CodeEditorVsc.updateOptions({readOnly:read_only});
        }
    }

    var CodeEditorVsc;

    var lint = pinegrow.getSetting('edit-element-css-lint', '1') === '1';

    //todo PEP enable lint into monaco

    // this.setLint = function(v) {
    //     pinegrow.setSetting('edit-element-css-lint', v ? '1' : '0');
    //     if(mirror) {
    //         if (v) {
    //             mirror.setOption('lint', true);
    //         } else {
    //             mirror.setOption('lint', false);
    //         }
    //     }
    //     lint = v;
    // }

    var makeEditor = function() {

        var $editor_el = null;

        let win = $content.get(0).ownerDocument.defaultView;

        CodeEditorVsc = monacoInstance.openEditor(null, null, win, $content, true, "css")

        let model = CodeEditorVsc.getModel();

        CodeEditorVsc.updateOptions({
            contextmenu:false
        });

        pinegrow.code_editors.register(CodeEditorVsc);

        // $editor_el.attr('data-has-codemirror','').data('codeMirrorInstance', mirror);

        CodeEditorVsc.onContextMenu(function(e) {
            e.event.preventDefault();
            return false;
        });

        _this.setWrap(wrap);

        _this.setReadOnly(read_only);

        setTimeout(function() {
            CodeEditorVsc.focus();
        }, 100);

        CodeEditorVsc.onDidChangeModelContent( function() {
            onChange();
        });

        CodeEditorVsc.onDidBlurEditorWidget(function () {
            onBlur();
        });

        return {mirror: CodeEditorVsc, editor_el: $editor_el,
            close: function() {
            }};
    }

    this.invalidate = function() {
    }

    var setCode = function(source) {
        ignore_change = true;

        if (code_ignore_change_timeout) {
          clearTimeout(code_ignore_change_timeout);
          code_ignore_change_timeout = null;
        }

        // mirror.operation(function() {
        //
        // });

        let cp = CodeEditorVsc.getPosition();
        var si = { left: CodeEditorVsc.getScrollLeft(), top: CodeEditorVsc.getScrollTop() };
        CodeEditorVsc.setValue(source);

        // mirror.getDoc().clearHistory();
        // mirror.scrollTo(si.left, si.top);

        // CodeEditorVsc.setPosition(cp);
        // CodeEditorVsc.revealPositionInCenter(cp, 0);

        code_ignore_change_timeout = setTimeout(function () {
          ignore_change = false;
          code_ignore_change_timeout = null;
        }, 100);

        has_errors = false;
        setNotice("");
    }

    var setRuleSource = function (rule) {
        var before = '';
        var shouldFormat = pinegrow.autoFormatCss != 'off';
        if (!shouldFormat) before = rule.raws.before;

        var original_source = before + cs.getSourceForRules(rule, false, shouldFormat);
        original_source = cs.filterSource(original_source, true);

        settingRuleCode = true;
        CodeEditorVsc.setValue(original_source);
        // mirror.getDoc().clearHistory();

        // Source needs to get reformatted after formatting single rule
        cs.setNeedFormatting(true);
    }

    var editRuleSource = function() {
        setRuleSource(oldRule)
    }

    this.flash = function() {
        pg$Flash($content);
    }
}
