var PgQuickCodeEdit = function (pgel, options) {

    var defs = {
        winId: 'quickCodeEdit',
        pin: pgf.edit_html,
        focus: true,
        title: false,
        copy: true,
        $target: null
    }

    options = $.extend(defs, options || {})

    var editors = this.editors = new PgCodeEditorsView(options);
    editors.handleSelected = false;

    if (!options.title && pgel) {
        options.title = pgel.getName({short: true}) + ` | ${pgf.edit_html ? 'Edit code' : 'View code'}`;
    }


    editors.quickWin = true;

    var win = this.win = new PgQuickWindow(options.title, editors, options.winId, 570, 260, null, true);

    editors.quickWin = win;

    if (options.pin) {
        win.addIconToHeader('icon-Pin', 'Pin to CODE panel', function () {
            var e = pinegrow.codeEditorsView.getEditorForPgel(editors.getEditor(0).getAllNodes()[0]);
            if (e) {
                pinegrow.showQuickMessage('Code of this element is already edited in the Code panel.', 4000);
                pinegrow.showTab('element_code_panel');
                e.view.setPinned(true);
                e.view.flash();
                win.destroy();
            } else {
                pinegrow.showTab('element_code_panel');
                pinegrow.codeEditorsView.editCode(editors.getEditor(0).getAllNodes());
                win.destroy();
                pinegrow.showQuickMessage('Pinned to CODE panel.');
            }
        })
    }

    if (options.$target) {
        win.showFor$El(options.$target)
    } else if (pgel) {
        win.showForPgel(pgel);
    } else {
        win.showAtPosition(300, 300);
    }

    editors.editCode(pgel);

    if (options.focus) {
        editors.focus();
    }

    this.setCustomCode = function (code) {
        editors.setCustomCode(code || '');
    }

    win.onDestroy = function () {
    }

}

//todo PEP find a way to reduce monaco model in the select element editor
var PgCodeEditorsView = function (options) {

    var defs = {
        key: 'element-code',
        jade: isApp(),//true,
        lint: true,
        editable: pgf.edit_html,
        copy: true,
        handleSelect: true,
        custom: false,
        mode: "application/x-httpd-php",
        copy: false,
        view_name: 'element_code'
    }

    options = $.extend(defs, options || {})

    //View implementation is responsible for creating DOM and adding it to document.

    var $element = $('<div class="pg-element-code-editors"></div>');

    var view = new PgUIView($element, defs.view_name, 'Element code');

    this.view = view;
    this.quickWin = false;

    view.tabIcon = 'icon-KODA';

    this.handleSelected = options.handleSelect;

    var _this = this;
    var editors = [];
    var jade = options.jade && (pinegrow.getSetting(options.key + '-jade', '0') === '1');
    var wrap = pinegrow.getSetting(options.key + '-html', '1') === '1';
    var lint = options.lint && (pinegrow.getSetting(options.key + '-html-lint', '1') === '1');

    var setUseJade = function (j) {
        jade = j;
        pinegrow.setSetting(options.key + '-jade', jade ? '1' : '0');
        editors.forEach(function (e) {
            e.view.setUseJade(jade)
        })
    }

    var setWrap = function (w) {
        wrap = w;
        pinegrow.setSetting(options.key + '-html', wrap ? '1' : '0');
        editors.forEach(function (e) {
            e.view.setWrap(wrap)
        })
    }

    var setLint = function (w) {
        lint = w;
        pinegrow.setSetting(options.key + '-html-lint', lint ? '1' : '0');
        editors.forEach(function (e) {
            e.view.setLint(lint)
        })
        pinegrow.showQuickMessage('Display HTML errors and warnings: ' + (lint ? 'ON' : 'OFF'));
    }

    var search = new PgSearchBox({
        placeholder: 'search',
        placeholder_active: null,
        toolbar_on: true
    });

    var toolbar = this.toolbar = search.getToolbar();

    var $toolbar_section = toolbar.addSection('settings').$element;

    if (options.jade) {
        var pugButton = new PgButtonToolbarItem('icon-270', 'Use simplified (Pug, ex-Jade) syntax.');

        toolbar.addItem('settings', pugButton);

        pugButton.onClick = function () {
            jade = !this.getActive();
            this.setActive(jade);

            setUseJade(jade);
            pinegrow.showMode('Use simplified Pug syntax', jade)
        }

        pugButton.setActive(jade);
    }

    var wrapButton = new PgButtonToolbarItem('icon-266', 'Wrap lines.');

    toolbar.addItem('settings', wrapButton);

    wrapButton.onClick = function () {
        var active = !this.getActive();
        this.setActive(active);

        setWrap(active);
        pinegrow.showMode('Wrap lines', active)
    }

    wrapButton.setActive(wrap);

    if (options.lint) {
        var lintButton = new PgButtonToolbarItem('icon-269', 'Display HTML errors and warnings (in HTML mode only).');

        toolbar.addItem('settings', lintButton);

        lintButton.onClick = function () {
            lint = !this.getActive();
            this.setActive(lint);
            setLint(lint);
        }

        lintButton.setActive(lint);
    }

    if (options.copy) {
        toolbar.addSection('tools');

        var copyButton = new PgButtonToolbarItem('icon-273', 'Copy code to clipboard.');

        toolbar.addItem('tools', copyButton);

        copyButton.onClick = function () {
            if (editors.length) {
                var code = editors[0].view.getCode();
                copyCodeToClipboard(code, true);
            }
        }

    }

    search.$element.appendTo($element);

    search.onSearch = function (f) {
        _this.search(f);
    }

    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    this.search = function (f) {
        editors.forEach(function (e) {
            e.view.search(f);
        })
    }

    var notice = new PgPanelNotice();
    notice.$element.appendTo($element);

    var grid = new PgUISectionGrid();
    grid.fit = true;

    grid.onGetNumberOfColumns = function (w, h, num_sections, num_c) {
        //try with 1
        var good_w = 500;
        if (w / good_w > 2) {
            num_c = Math.round(w / good_w);
        } else if (h / num_sections > 140) {
            num_c = 1;
        }
        if (num_sections < num_c && num_sections > 0) {
            num_c = num_sections;
        }
        return num_c;
    }

    $element.append(grid.$element);

    this.focus = function () {
        if (editors.length) {
            editors[0].view.focus();
        }
    }

    this.editCode = function (pgel) {
        view.whenVisible(function () {
            var e = getEditorForPgel(pgel);
            if (e) {
                if (e.view.pinned) {
                    e.view.flash();
                    return;
                }
            }
            createEditor(pgel);
        })
    }

    this.setCustomCode = function (code) {
        view.whenVisible(function () {
            editors[0].view.setCustomCode(code);
        }, 'setCustomCode', 1)
    }

    var createEditor = function (pgel) {
        var d = {};
        d.pinnable = true;
        if (pgel) {
            //d.pinned = true;
        }
        d.section = new PgUISection('Editor', null, d);
        d.view = new PgCodeEditView(null, options, grid.$element /* this will be the parent element */);
        d.view.section = d.section;
        d.section.appendTo(d.view.view.get$Element());
        d.view.view.alwaysVisible = true;
        d.view.view.updateVisibility();
        d.view.setWrap(wrap);
        d.view.setUseJade(jade);
        d.view.setLint(lint);
        d.view.show(pgel);
        d.view.view.onResizeOriginal = d.view.view.onResize;
        d.view.view.onResize = null;

        addTooltip(d.section.$title.find('.icon-Pin').css('top', 0), 'Show this editor even if element is not selected.');

        if (pgel) {
            //d.view.setPinned(true);
        }
        d.section.onLayoutChanged = function () {
            d.view.view.onResizeOriginal();
        }

        d.section.onPin = function (s) {
            d.view.setPinned(true);
        }

        d.section.onClose = function (s) {
            _this.closeEditor(d);
        }
        if (d.view.pinned) {
            var idx = -1;
            for (var i = 0; i < editors.length; i++) {
                if (!editors[i].view.pinned) {
                    idx = i;
                    break;
                }
            }
            if (idx >= 0) {
                editors.splice(idx, 0, d);
            } else {
                editors.push(d);
            }
        } else {
            editors.push(d);
        }
        showEditors();
    }

    this.closeEditor = function (d) {
        var idx = editors.indexOf(d);
        if (idx >= 0) {
            editors.splice(idx, 1);
            d.view.view.destroy();
            showEditors();
        }
        if (editors.length == 0 && this.quickWin) {
            this.quickWin.destroy();
        }
    }

    this.getEditor = function (num) {
        return editors[num].view;
    }

    var showEditors = function () {
        grid.clear();

        if (editors.length == 0) {
            notice.show('Element is not selected.');
            grid.$element.hide();
        } else {
            notice.show();
            grid.$element.show();

            editors.forEach(function (e) {
                grid.addSection(e.section)
            })
            grid.updateLayout();
        }
    }

    showEditors();

    var getEditorForPgel = this.getEditorForPgel = function (pgel) {
        if (!pgel) return null;
        for (var i = 0; i < editors.length; i++) {
            if (editors[i].view.isPgElBeingEdited(pgel)) return editors[i];
        }
        return null;
    }

    var getEditorForSelected = function () {
        for (var i = 0; i < editors.length; i++) {
            if (!editors[i].view.pinned) return editors[i];
        }
        return null;
    }

    var onElementSelected = function (page, pages, info) {
        var pgel = pinegrow.selectedElements.getLastSelected();
        if (info && info.source && info.source instanceof PgCodeEditView) {
            return;
        }
        if (!pgel) {
            var e = getEditorForSelected();
            if (e) {
                _this.closeEditor(e);
            }
            return;
        }

        //undo?
        if (info && info.undoSourceId) {
            for (var i = 0; i < editors.length; i++) {
                if (editors[i].view.view.getId() == info.undoSourceId) {
                    //This was undo operation originating in this editor
                    if (editors[i].view.handleUndoSelect(pgel, info)) {
                        //editor knew how to handle it
                        return;
                    } else {
                        break;
                    }
                }
            }
        }
        if (!_this.handleSelected) return;

        view.whenVisible(function () {

            var e = getEditorForPgel(pgel);
            if (e) {
                if (e.view.pinned) {
                    //this el is already in pinned editor. Just flash that.
                    e.view.flash();
                    return;
                }
            }
            e = getEditorForSelected();
            if (!e) {
                createEditor(pgel);
            } else {
                e.view.show(pgel);
            }
        })
    }

    var onPageClosed = function (page) {
        var remove = [];
        for (var i = 0; i < editors.length; i++) {
            if (page === editors[i].view.getPage()) {
                remove.push(editors[i]);
            }
        }
        remove.forEach(function (d) {
            _this.closeEditor(d);
        })
    }

    var onPageChanged = function (page, pgel, info) {
        if (!pgel) return;
        for (var i = 0; i < editors.length; i++) {
            if (editors[i].view.invalid) continue;
            if (info && info.source && info.source == editors[i].view) continue;
            if (info && info.undoSourceId && info.undoSourceId == editors[i].view.view.getId()) continue;

            if (info && info.source && info.source instanceof PgCodeEditView && info.removedIds) {
                //change in another editor
                //1. top el is removed (changed in another editor) - invalidate
                if (editors[i].view.isTopId(info.removedIds)) {
                    editors[i].view.invalidate();
                } else if (editors[i].view.isIdInList(info.removedIds)) {
                    //2. child el is changed - update
                    editors[i].view.updateEditor();
                }
            } else {
                if (editors[i].view.checkIfEditedElementWasDeleted()) {
                    _this.closeEditor(editors[i]);
                } else {
                    //normal change
                    var next = pgel.next();
                    var prev = pgel.prev();
                    if (editors[i].view.isPgElBeingEdited(pgel) ||
                        (next && editors[i].view.isPgElBeingEdited(next)) ||
                        (prev && editors[i].view.isPgElBeingEdited(prev))) {
                        editors[i].view.updateEditor();
                    }
                }
            }
        }
    }

    view.onResize = function () {
        grid.updateLayout();
    }

    view.onShow = function () {
        editors.forEach(function (e) {
            e.view.view.onShow();
        })
    }

    if (!options.custom) {
        pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);
        pinegrow.addEventHandler('on_page_changed', onPageChanged);
        pinegrow.addEventHandler('on_page_closed', onPageClosed);
    }

    view.onDestroy = function () {
        pinegrow.removeEventHandler('on_selected_elements_changed', onElementSelected);
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        pinegrow.removeEventHandler('on_page_closed', onPageClosed);
    }
}


var PgCodeEditView = function (aPgEl, options, $future_parent) {

    // var $element = $('<div class=""><div class="pg-element-code-editor"></div><div class="pg-element-code-editor-controls"</div>');
    var $element = $('<div class=""><div class="pg-element-code-editor"></div>');

    var $content = $element.find('.pg-element-code-editor');
    //var $controls = $element.find('.pg-element-code-editor-controls');

    this.setPinned = function (pin) {
        this.pinned = pin;
        if (pin) {
            //$pin.html('UNPIN');
            this.section.$title.addClass('pinned');
        } else {
            //$pin.html('PIN');
            this.section.$title.removeClass('pinned');
        }
        setTitle();
    }

    var notice = new PgCodePanelNotice();
    notice.hideOnEmpty = true;
    notice.get$Element().appendTo($element);

    var view = this.view = new PgUIView($element);

    var _this = this;

    this.pinned = false;
    this.isSelected = false;

    var pug = require('pug');

    var settingsPrefix = options.key + '-html';
    var editorData = null;

    var ignore_change = false;
    var ignore_select_element = false;
    var has_changes = false;
    var has_content = false;
    var has_errors = false;

    var id_to_markers = {};

    var use_jade = true;
    var do_jade;
    var wrap = true;

    var code_ignore_change_timeout = null;

    var pgEl = aPgEl;
    var nodeCount = 1; //always starts with editing just one el
    var selectAll = false;
    var allNodes = [];
    var page = null;
    var editUpToPgel = null;

    var allIds = [];

    var path = null;
    var path_last = null;

    this.getAllNodes = function () {
        return allNodes;
    }

    this.checkIfEditedElementWasDeleted = function () {
        if (allNodes.length && allNodes[0].isDeleted) {
            //our element is gone
            return true;
        }
        return false;
    }

    var updateAllIds = function () {
        allIds = [];
        allNodes.forEach(function (pgel) {
            pgel.walkSelfAndChildren(function (node) {
                var id = node.getId();
                if (id) {
                    allIds.push(id);
                }
                return true;
            })
        })
        allIds.sort();

        path = allNodes[0].getPath();
        path_last = allNodes.length > 1 ? allNodes[allNodes.length - 1].getPath() : null;

        page = allNodes[0].getPage();
        editUpToPgel = allNodes[allNodes.length - 1].next();
    }

    this.isTopId = function (ids) {
        for (var i = 0; i < allNodes.length; i++) {
            var id = allNodes[i].getId();
            if (id && ids.indexOf(id) >= 0) return true;
        }
        return false;
    }

    this.isIdInList = function (ids) {
        var ids_i = 0;
        var all_i = 0;
        while (ids_i < ids.length && all_i < allIds.length) {
            if (ids[ids_i] === allIds[all_i]) return true;
            if (ids[ids_i] < allIds[all_i]) {
                ids_i++;
            } else {
                all_i++;
            }
        }
        return false;
    }

    view.onResize = function () {
        if (editorData) {
            if ($mirror_el?.get(0).ownerDocument) {
                // CodeEditorVsc.pgSetWindow($mirror_el.get(0).ownerDocument.defaultView);
                CodeEditorVsc.layout()
            }
            editorData.CodeEditorVsc.layout();
        }
    }

    view.onShow = function () {
        if (editorData) editorData.CodeEditorVsc.layout();
    }

    view.onDestroy = function () {
        //dispose the editor instance
        if (editorData) {
            editorData.CodeEditorVsc.dispose();
            editorData = null;
        }

        pinegrow.code_editors.unregister(CodeEditorVsc);
        //pinegrow.removeEventHandler('on_element_selected', onElementSelected);
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        if (reselectElementsTimer) {
            clearTimeout(reselectElementsTimer);
        }
        if (updateTimer) {
            clearTimeout(updateTimer);
        }
        setNotice('');
    }

    view.onMovedToAnotherWindow = function (new_win) {
        //Called when the view is moved from one app window to another. In this case we need to recreate the editor
        if (editorData) {
            editorData.CodeEditorVsc.dispose();
            editorData = null;
        }
        _this.show(pgEl);
    }

    var checkIfElementCanBeEdited = function (pgel, info) {
        info = info || {};
        info.msg = 'This element can not be edited here.'
        var problems = new pgParserSourceProblem(pgEl);

        if (!pgel) {
            problems.add('element', pgel.getName(), 'find');
        }
        if (!problems.ok()) {
            return false;
        }
        if (!canMakeChange(pgel, 'edit_code', null, true /* silent */)) {
            info.msg = 'Can not edit this element because it is locked.';
            return false;
        }

        if (pgel.tagName == 'body' || pgel.tagName == 'head' || pgel.find('body').length) {
            info.msg = 'Can not edit HEAD and BODY elements. Please use Page -> Edit code to edit the whole page.';
            return false;
        }
        return true;
    }

    this.show = function (pgel) {
        if (options.custom) {
            view.whenVisible(function () {
                if (!editorData) {
                    editorData = makeEditor();
                }
            })

        } else {
            var info = {pgel: pgel}

            pgel.getPage().callFrameworkHandler('on_before_element_code', info)

            pgel = info.pgel;

            pgEl = pgel;

            this.isSelected = pinegrow.selectedElements.isElementOrDescendantSelected(pgel);

            view.whenVisible(function () {
                if (!editorData) {
                    editorData = makeEditor();
                }
                editElementSource(pgel);
            })
        }
    }

    this.setCustomCode = function (code) {
        if (code.match(/<\?php/igm)) {
            CodeEditorVsc.setModel(monaco.editor.createModel(CodeEditorVsc.getValue(), "php"));
            //await pgEditorAttributeIntellisense.setHtmlCompletion(pageUrl, null, "php");
        }
        editorData.CodeEditorVsc.setValue(code);
    }

    this.getCode = function () {
        return editorData.CodeEditorVsc.getValue();
    }

    this.isPgElBeingEdited = function (pgel) {
        for (var i = 0; i < allNodes.length; i++) {
            if (pgel == allNodes[i] /*|| pgel.isDescendantOf(allNodes[i]) */) {
                return true;
            }
        }
        /*if(allNodes.length > 1 && (allNodes[0].parent == pgel || pgel.isDescendantOf(allNodes[0].parent))) return true;*/

        return false;
    }

    var updateTimer = null;

    var onPageChanged = function (page, pgel, info) {
        if (options.custom) return;
        if (info && info.source && info.source == _this) return;
        if (_this.invalid) return;

        if (!pgel || _this.isPgElBeingEdited(pgel)) {
            if (updateTimer) {
                clearTimeout(updateTimer);
            }
            updateTimer = setTimeout(function () {
                updateRange();
                original_source = getCode();
                setCode(original_source, info);
                updateTimer = null;
            }, 50)
        }
    }

    this.updateEditor = function (info) {
        onPageChanged(null, null, info);
    }

    this.handleUndoSelect = function (pgel, info) {
        //reselect after undo editor op happened
        this.invalidate(info);
        return true;
    }

    // pinegrow.addEventHandler('on_element_selected', onElementSelected);
    pinegrow.addEventHandler('on_page_changed', onPageChanged);

    var noticeTimer = null;

    var setNotice = function (msg) {
        if (noticeTimer) {
            clearTimeout(noticeTimer);
            noticeTimer = false;
        }
        if (!msg) {
            notice.setMessage('');
        } else {
            noticeTimer = setTimeout(function () {
                updateNoticePosition(msg);
                noticeTimer = null;
            }, 750)
        }
    }

    var updateNoticePosition = function (msg) {
        if (!notice) return;
        msg = msg === undefined ? notice.getMessage() : msg;
        if (msg) {
            var cp = CodeEditorVsc.getPosition();
            var si = {left: CodeEditorVsc.getScrollLeft(), top: CodeEditorVsc.getScrollTop(), clientHeight: CodeEditorVsc.$el.clientHeight};
            var topy = CodeEditorVsc.getTopForPosition(cp.lineNumber, cp.column);
            //console.log(topy, si.top, si.clientHeight);
            topy -= si.top;
            var top = si.clientHeight - topy < 60;
            notice.setMessage(msg, top);
            //CodeEditorVsc.setPosition(cp);
            //CodeEditorVsc.revealPositionInCenter(cp, 0);
        }
    }

    var onChange = function () {
        if (options.custom) return;

        if (applySource(editorData.CodeEditorVsc.getValue())) {

        }
        return has_errors;
    }

    var cursor_position_before_change = null;
    var has_error = false;

    var applySource = function (s, force) {
        if (ignore_change) return false;
        s = $.trim(s);

        try {
            if (do_jade) {
                s = pug.render(s, {});
            }

            var pgNewElList = pgCreateNodeFromHtml(s, false /* dont assign id & add to catalogue */, true /* multi */);
            pgCurrentRequestContext.setCurrentRequestContext(page.url, page.sourceNode);
            var pgNewEl = null;
            var firstPgNewEl = null;
            var only_elements = true;
            var new_node_count = 0;
            var new_nodes = [];

            var jade = [];

            for (var i = 0; i < pgNewElList.length; i++) {
                pgNewEl = pgNewElList[i];

                if (!firstPgNewEl) firstPgNewEl = pgNewEl;

                if (!pgNewEl.isElementOrScript) only_elements = false;

                var html = pgNewEl.toStringOriginal(true, pinegrow.getFormatHtmlOptions(), function (linkNode, str, type) {
                    return pgCurrentRequestContext.createProxyUrlNodeOutputFilter(linkNode, str, type);
                });

                if (html.length === 0) continue;

                new_node_count++;
                new_nodes.push(pgNewEl);

                var html_err = 'HTML syntax error (open tags, invalid nesting of elements...)';

                if (pgNewEl.isElement && !pgNewEl.closed) {
                    throw 'Tag ' + pgNewEl.tagName + ' is not closed.';
                }

                if (pgNewEl.validateTree().length && !force) {
                    throw html_err;
                }

                var $newEl = $(getIframeDocument(pgEl.getPage().$iframe.get(0)).createElement('div')).append(html).contents();

                if ($newEl.length !== 1 && !force) {
                    throw html_err;
                }

                if ($newEl.get(0).nodeType !== 1) {
                    //throw "Must be HTML element, not just text.";
                }

                //all ok, assign ids
                pgNewEl.assignIdAndAddToCatalog(true);
                jade.push(pgNewEl.toJade());

            }


            var multi = (new_node_count > 1) || (nodeCount > 1);

            var updatePgEl = multi ? pgEl.parent : firstPgNewEl;
            var updatePgId = multi ? updatePgEl.getId() : pgEl.getId();

            //var oldHtml = multi ? updatePgEl.toStringWithIds() : pgEl.toStringWithIds();

            var undoPgel = pgEl.parent;
            var undoOldHtml = undoPgel.html(null, true);

            var removedIds = [];
            var deselect = [];

            var selected_before_change = getElementsToSelect();

            updatePgEl.withoutEvents(function () {
                pgEl.replaceWith(firstPgNewEl, true /* detach */);

                //pinegrow.selectedElements.remove(pgEl);
                deselect.push(pgEl);
                pgEl.remove(false, removedIds);

                var next = firstPgNewEl.next(true);
                for (var i = 1; i < nodeCount; i++) {
                    if (next) {
                        if (next === editUpToPgel) break;

                        var n = next.next(true);
                        if (next.isElementOrScript) {
                            //pinegrow.selectedElements.remove(next);
                            deselect.push(next);
                        }
                        next.remove(false, removedIds);
                        next = n;
                    }
                }

                var last = firstPgNewEl;
                for (var i = 1; i < new_nodes.length; i++) {
                    new_nodes[i].insertAfter(last);
                    last = new_nodes[i];
                }
            });
            pinegrow.undoManager.getHistoryForPage(page).ignore(function () {
                updatePgEl.update(/*oldHtml*/null, updatePgId /* perhaps id changed, tell which dom el to update */);
            })

            //create custom undo for parent
            var undoData = {
                cursorPosition: cursor_position_before_change
            }

            var pgEvent = new PgNodeEvent(undoPgel, 'updateInner', null, {
                oldValue: undoOldHtml
            }, page);
            var undoEvent = new PgUndoEvent(pgEvent);
            pinegrow.undoManager.getHistoryForPage(page).addEvent(undoEvent);
            /////

            if (!_this.pinned) {
                pinegrow.undoManager.getHistoryForPage(page).setSelectedElementsAtStart(selected_before_change.map(function (pgel) {
                    return pgel.getId();
                }));
            }

            allNodes = new_nodes;
            updateAllIds();

            pgEl = firstPgNewEl;
            nodeCount = pgNewElList.length;
            setNotice('');

            if (!_this.pinned) {
                pinegrow.undoManager.getHistoryForPage(page).setSelectedElementsAtEnd(getElementsToSelect().map(function (pgel) {
                    return pgel.getId();
                }));
            }

            deselect.forEach(function (pgel) {
                pinegrow.selectedElements.remove(pgel, {source: _this});
            })

            reselectElements();

            removedIds.sort();

            cursor_position_before_change = CodeEditorVsc.getPosition();

            pinegrow.changeManager.didChange(pgEl, 'Edit code', {
                source: _this,
                removedIds: removedIds,
                undoCombineKey: 'edit_code',
                undoSourceId: view.getId(),
                undoData: undoData
            });

            last_good_source = s;
            has_error = false;
            return true;
        } catch (err) {
            if (err instanceof pgParserException) {
                setNotice('Only one element can be edited.', 'text-danger');
            } else {
                setNotice((err.msg || err.toString()), 'text-danger');
            }
            has_error = true;
        }

        return false;
    }

    var reselectElementsTimer = null;

    var getElementsToSelect = function () {
        var r = [];
        if (allNodes[0].isElementOrScript) {
            r.push(allNodes[0]);
        }
        if (selectAll) {
            for (var i = 1; i < allNodes.length; i++) {
                if (allNodes[i].isElementOrScript) {
                    r.push(allNodes[i]);
                }
            }
        }
        return r;
    }

    var reselectElements = function () {
        if (reselectElementsTimer) {
            clearTimeout(reselectElementsTimer);
            reselectElementsTimer = null;
        }
        if (!_this.pinned || _this.isSelected) {
            reselectElementsTimer = setTimeout(function () {
                var list = getElementsToSelect();
                list.forEach(function (pgel) {
                    pinegrow.selectedElements.add(pgel, {source: _this});
                })
                pinegrow.selectedElements.triggerElementSelectedEvent(list[0], {source: _this});
                reselectElementsTimer = null;
            }, 750);
        } else {

        }
    }

    this.setWrap = function (w) {
        wrap = w;
        if (CodeEditorVsc) {
            CodeEditorVsc.updateOptions({wordWrap: w ? "on" : "off"});
        }
    }

    this.setUseJade = function (j, dont_set_code) {
        use_jade = j;
        if (CodeEditorVsc) {
            if (j) {
                CodeEditorVsc.setModel(monaco.editor.createModel(CodeEditorVsc.getValue(), "pug"))
            } else {
                CodeEditorVsc.setModel(monaco.editor.createModel(CodeEditorVsc.getValue(), "html"));
            }
            if (!dont_set_code) {
                if (updateTimer) {
                    clearTimeout(updateTimer);
                    updateTimer = null;
                }
                original_source = getCode();
                setCode(original_source);
            }
        }
    }

    var CodeEditorVsc;
    var $mirror_el = null;

    var lint = pinegrow.getSetting(options.key + '-html-lint', '1') === '1';

    this.setLint = function (v, pgel_list) { // todo PEP work on lint here
        pinegrow.setSetting(options.key + '-html-lint', v ? '1' : '0');
        if (CodeEditorVsc) {
            var ev = v;
            var has_script = false;
            if (pgel_list) {
                for (var i = 0; i < pgel_list.length; i++) {
                    if (pgel_list[i].containsServerSideScripts()) {
                        has_script = true;
                        break;
                    }
                }
            }
            if (v && has_script) {
                ev = false;
                //pinegrow.showQuick
            }
            if (ev) {
                // CodeEditorVsc.setOption('lint', pinegrow.getCodeMirrorLintOptions());
            } else {
                // CodeEditorVsc.setOption('lint', false);
            }
            CodeEditorVsc.pgCheckForErrors = ev;
        }
        lint = v;
    }

    this.focus = function () {
        if (CodeEditorVsc) {
            CodeEditorVsc.focus();
        }
    }

    let pageUrl;

    var makeEditor = function () {


        var $editor_el = null;

        //At this point the editor view is not yet appended to the DOM, so defaultView of the $content will always match the main app window. That's why we pass in the future parent reference and get its window.
        let win = ($future_parent || $content).get(0).ownerDocument.defaultView;
        CodeEditorVsc = monacoInstance.openEditor(null, null, win, $content, false, "html");

        let model = CodeEditorVsc.getModel();
        pinegrow.code_editors.register(CodeEditorVsc);

        pageUrl = pinegrow.getSelectedPage().url
        //pgEditorAttributeIntellisense.setHtmlCompletion(pageUrl, null, model._languageIdentifier.language );
        pgEditorAttributeIntellisense.setHtmlCompletion(pageUrl, null, model.getLanguageId() );

        // $editor_el.attr('data-has-codemirror','').data('codeMirrorInstance', CodeEditorVsc);

        _this.setWrap(wrap);
        // _this.setUseJade(use_jade, true /* dont set code */);

        if (!options.editable) {
            CodeEditorVsc.updateOptions({readOnly: true});
        }

        CodeEditorVsc.onDidChangeModelContent(function () {
            if (_this.invalid) return;

            onChange();
            if (did_highlight) {
                pinegrow.highlightElement(null);
            }
        });

        // CodeEditorVsc.onDidChangeModelLanguage(async e => {
        //     await pgEditorAttributeIntellisense.setHtmlCompletion(url, null, e.newLanguage);
        // })

        var did_highlight = false;

        CodeEditorVsc.onDidFocusEditorText(function () {
            pgEditorAttributeIntellisense.focusedUrl = pinegrow.getSelectedPage().url;
            if (_this.invalid) return;
            if (options.custom) return;

            if (_this.pinned) {
                var $els = [];
                allNodes.forEach(function (pgel) {
                    var $e = pgel.get$DOMElement();
                    if ($e) {
                        $els.push($e.get(0));
                    }
                })

                if ($els.length) {
                    $els = $($els);
                    pinegrow.highlightElement($els);
                    did_highlight = true;
                }
            }
        });

        CodeEditorVsc.onDidChangeCursorPosition(function () {
            updateNoticePosition();
            cursor_position_before_change = CodeEditorVsc.getPosition();
        })

        CodeEditorVsc.onContextMenu(function (e) {
            e.event.preventDefault();
            return false;
        });

        return {
            CodeEditorVsc: CodeEditorVsc, editor_el: $editor_el,
            close: function () {
                has_changes = false;
                close();
            }
        };
    }

    this.invalidate = function (info) {
        //setCode('invalid');
        var pgel = page.sourceNode.getNodeFromPath(path);
        if (pgel) {
            allNodes = [pgel];
            pgEl = pgel;
            if (path_last) {
                var last_pgel = page.sourceNode.getNodeFromPath(path_last);
                if (last_pgel != pgel && last_pgel.parent == pgel.parent) {
                    allNodes.push(last_pgel);
                }
            }
            this.updateEditor(info);
        }
    }

    let isPhp = false

    var setCode = async function (source, info) {
        var source = removeCrsaClassesFromHtml(source);

        if (source.match(/<\?php/igm) && !isPhp) {
            CodeEditorVsc.setModel(monaco.editor.createModel(CodeEditorVsc.getValue(), "php"));
            await pgEditorAttributeIntellisense.setHtmlCompletion(pageUrl, null, "php");
            isPhp = true;
        }

        ignore_change = true;

        if (code_ignore_change_timeout) {
            clearTimeout(code_ignore_change_timeout);
            code_ignore_change_timeout = null;
        }

// removed the code scroll
        // var cp = (info && info.undoData && info.undoData.cursorPosition) ? info.undoData.cursorPosition : CodeEditorVsc.getPosition();
        //
        // var si = {left: CodeEditorVsc.getScrollLeft(), top: CodeEditorVsc.getScrollTop()};
        //
        // CodeEditorVsc.setPosition(cp);
        // CodeEditorVsc.revealPositionInCenter(cp, 0);

        has_errors = false;
        setNotice('');

        code_ignore_change_timeout = setTimeout(function () {
            CodeEditorVsc.setValue(source);
            ignore_change = false;
            code_ignore_change_timeout = null;
        }, 100);


    }

    this.search = function (f) {
        let editor = CodeEditorVsc.getModel();
        let find = editor.findMatches(f, true, false, true, null, true);
        let selections = []

        if (f.length !== 0) {
            Array.from(find).forEach(v => {
                let range = v.range;
                selections.push({
                    positionColumn: range.endColumn,
                    positionLineNumber: range.endLineNumber,
                    selectionStartColumn: range.startColumn,
                    selectionStartLineNumber: range.startLineNumber
                })
            });
            if (selections.length > 0)
                CodeEditorVsc.setSelections(selections);
        } else {
            CodeEditorVsc.setSelection({
                startColumn: 0,
                startLineNumber: 0,
                endColumn: 0,
                endLineNumber: 0
            });
        }
        // CodeEditorVsc.execCommand('doSearchMT');
    }

    var updateRange = function () {
        if (allNodes.length > 1) {
            var first = null;
            var min = null;
            var max = null;

            for (var i = 0; i < allNodes.length; i++) {
                var idx = allNodes[i].parent.getChildPos(allNodes[i]);
                if (first === null) {
                    first = allNodes[i];
                    min = idx;
                    max = idx;
                } else if (first.parent == allNodes[i].parent) {
                    if (idx < min) {
                        min = idx;
                    }
                    if (idx > max) {
                        max = idx;
                    }
                }
            }
            var r = [];
            for (var i = min; i <= max; i++) {
                r.push(first.parent.children[i]);
            }
            allNodes = r;
            nodeCount = allNodes.length
            pgEl = allNodes[0];
            updateAllIds();
        } else {
            updateAllIds();
        }
    }

    var setTitle = function () {
        if (options.custom) return;
        if (!pgEl) {
            _this.section.setTitle('Can\'t edit this element');
        } else {
            if (_this.pinned) {
                _this.section.setTitle(pgEl.getName({short: true}) + ' | Pinned');
            } else {
                _this.section.setTitle(pgEl.getName({short: true}) + ' | Selected element');
            }
        }
    }

    var editElementSource = function (pgel) {

        pinegrow.stats.using('edit.elementcode');

        var all = null;
        selectAll = false;

        if (!pgel) {
            var range = pinegrow.selectedElements.getRange();
            if (range) {
                all = range.all;
                pgel = range.first;
                nodeCount = range.count;
                selectAll = nodeCount > 1;
            } else {
                nodeCount = 1;
                pgel = pinegrow.selectedElements.getLastSelected();
            }
        }

        if (pgel && Array.isArray(pgel)) {
            all = pgel;
            pgel = pgel[0];
        }

        var CodeEditorVsc = editorData.CodeEditorVsc;

        var info = {};
        if (!pgel || !checkIfElementCanBeEdited(pgel, info)) {
            _this.invalid = true;
            CodeEditorVsc.updateOptions({readOnly: true});
            //
            // CodeEditorVsc.setOption('readOnly', true);
            // CodeEditorVsc.setOption('lint', false);
            if (info) {
                CodeEditorVsc.setValue(info.msg || '');
            }
            setTitle();
            return false;
        }
        _this.invalid = false;
        CodeEditorVsc.updateOptions({
            readOnly: !options.editable
        });
        // CodeEditorVsc.setOption('lint', lint ? pinegrow.getCodeMirrorLintOptions() : false);

        if (!all) all = [pgel];

        allNodes = all;
        updateAllIds();

        pgEl = pgel;

        setTitle();

        var crsaPage = pinegrow.getCrsaPageOfPgParserNode(pgEl);

        var dynmirror = null;

        _this.setLint(lint, allNodes);

        var original_source = getCode();

        //original_source = pgEl.toStringWithIds(true, pinegrow.getFormatHtmlOptions());

        //original_source = pgEl.toStringOriginal(true, pinegrow.getFormatHtmlOptions());


        var last_good_source = original_source;
        var has_error = false;

        var codeEditor = pinegrow.codeEditor;

        setCode(original_source);

    }

    var canUseJade = function () {
        for (var i = 0; i < allNodes.length; i++) {
            if (!allNodes[i].canConvertToJade()) return false;
        }
        return true;
    }

    var getCode = function () {
        var original_source = '';// = element.outerHTML;

        do_jade = use_jade;

        if (use_jade && !canUseJade()) {
            do_jade = false;
        }

        allNodes.forEach(function (pgel) {
            if (do_jade) {
                var jade = pgel.toJade();

                //console.log(pgel.toDebug());
                if (jade !== null) {
                    if (original_source.length) original_source += "\n";
                    original_source += jade;
                }
            } else {
                if (original_source.length) original_source += "\n";
                var options = pinegrow.getFormatHtmlOptions();
                options.remove_collapsed = true;
                options.remove_hidden = true;
                original_source += pgel.toStringOriginal(true, options);
            }
        })

        return original_source;
    }

    this.flash = function () {
        pg$Flash($content);
    }

    this.getPage = function () {
        return page;
    }
}
