var editor_y = 100;

var PgInternalCodeEditor = function(
  id, url, existing_editor, not_html, win, let_monaco_handle_undo) {
  this.id = id;
  this.url = url;
  var CodeEditorVsc = null;
  var code = null;

  var _this = this;

  this.page = null;

  var $mirror_el = null;

  var wrap = pinegrow.getSetting('edit-page-html-wrap', '1') === '1';
  var lint = pinegrow.getSetting('edit-page-html-lint', '1') === '1';

  var $textEdit = $('<div class="pg-page-code-editor"></div>');

  this.isFocued = function() {
    return $(document.activeElement).closest('.pg-page-code-editor').is(
      $textEdit);
  };

  this.appendTo = function($c) {
    // $c.append($textEdit);
    $textEdit = $c;
    $mirror_el?.css('height', '');
    CodeEditorVsc.render(true);
    CodeEditorVsc.layout();
  };

  this.resize = function() {
    if($mirror_el?.get(0)?.ownerDocument) {
      CodeEditorVsc.layout();
    }
    // this.refresh();
  };

  this.setWrap = function(w) {
    wrap = w;
    pinegrow.setSetting('edit-page-html-wrap', wrap ? '1' : '0');
    if(CodeEditorVsc) {
      CodeEditorVsc.updateOptions({wordWrap: w ? 'on' : 'off'});
      CodeEditorVsc.layout();
    }
  };

  var read_only = false;

  this.setReadOnly = function(val) {
    read_only = val;
    if(CodeEditorVsc) {
      CodeEditorVsc.updateOptions({readOnly: read_only});
    }
  };
//Todo PEP move this to editor action
  this.search = function(f) {
    CodeEditorVsc._pgSearchQuery = f;
    // CodeEditorVsc.execCommand('doSearchMT');
    let model = CodeEditorVsc.getModel();
    let EOL = model.getEOL();
    let spliter = f.split(EOL);
    let findStart = model.findMatches(spliter[0]);
    let findEnd = model.findMatches(spliter[spliter.length - 1]);

    let selections = [];

    if(f.length !== 0) {
      Array.from(findEnd).find((v, i) => {
        if(v.range.endLineNumber ==
          ( findStart[0].range.startLineNumber + spliter.length - 1 )) {
          CodeEditorVsc.setSelection({
            startColumn: findStart[0].range.startColumn,
            startLineNumber: findStart[0].range.startLineNumber,
            endColumn: v.range.endColumn,
            endLineNumber: v.range.endLineNumber,
          });

          CodeEditorVsc.revealLineInCenter(findStart[0].range.startLineNumber);
        }
      });
    }
  };

  this.focus = function() {
    CodeEditorVsc.focus();
  };

  this.setLint = function(v) {
      lint = v;
      CodeEditorVsc.pgCheckForErrors = lint;
  }

  var undo_depth = 0;

  CodeEditorVsc = monacoInstance.openEditor(url, id, win, null,
    let_monaco_handle_undo);
  this.setWrap(wrap);

  CodeEditorVsc.pgCheckForErrors = lint;

  if(read_only) {
    CodeEditorVsc.updateOptions({readOnly: read_only});
  }

  pinegrow.code_editors.register(CodeEditorVsc);

  var last_selected_path = null;

  let mousePos;
  let onMousePos;

  function DOMSelector() {
    let model = CodeEditorVsc.getModel();
    var profile = new CrsaProfile();
    var pos = mousePos ?? onMousePos;
    if(pos) {
      var idx = model.getOffsetAt(pos);
      doWithCurrentSourceNode(function(sourceNode) {
        var node = sourceNode.findNodeAtSourceIndex(idx);
        if(node) {
          last_selected_path = node.getPath();

          _this.onElementSelectedInEditor(_this, last_selected_path);
        }
      });
      onMousePos = mousePos = null;
    }
    return false;
  }

    //switch(CodeEditorVsc.getModel()._languageIdentifier.language) {
  switch(CodeEditorVsc.getModel().getLanguageId()) {
    case 'html':
    case 'php':

      monacoInstance.actions['domSelector'] = 'pinegrow.editor.selectDom';
      CodeEditorVsc.addAction({
        id: monacoInstance.actions.domSelector,
        label: 'Select on the page',
        keybindings: [monaco.KeyMod.Alt | monaco.KeyCode.KEY_S],
        precondition: null,
        keybindingContext: null,
        contextMenuGroupId: 'navi',
        contextMenuOrder: 100,
        run: (editor, ...fn) => {
          // fn()
          DOMSelector();
        },
      });

      monacoInstance.actions['browserPreview'] = 'pinegrow.editor.action.browserPreviewer';
      CodeEditorVsc.addAction({
        id: monacoInstance.actions.browserPreview,
        label: 'Preview In Browser',
        keybindings: [monaco.KeyMod.CtrlCmd | monaco.KeyCode.KEY_B],
        precondition: null,
        keybindingContext: null,
        contextMenuGroupId: 'navi',
        contextMenuOrder: 98,
        run: (editor, ...page) => {
          _this.page.openInBrowser();
        },
      });
      break;

    default:
      break;
  }

  CodeEditorVsc.onMouseDown((e) => {

    mousePos = e.target.position;

    if(e.event.altKey && e.event.leftButton) {
      DOMSelector();
    }
  });

  CodeEditorVsc.onDidChangeCursorPosition(e => {
    updateNoticePosition();
    if(e.reason === 3) { //Explicit user gesture only
        cursor_position_before_change = CodeEditorVsc.getPosition();
    }
  });

  var ignore_change = 0;

  var change_timer = null;
  var change_delay = 500;
  var syntax_error_timer = null;

  var doChange = function() {
    var st = new Date().getTime();
    var page = pinegrow.getCrsaPageByUrl(url);
    let model = CodeEditorVsc.getModel();

    code = model.getValue();
    currentSourceNode = null;
    change_timer = null;

    var cp = CodeEditorVsc.getPosition();
    if(_this.onChange) {

      var undoData = {
          cursorPosition: cursor_position_before_change,
          cursorPositionAfter: cp
      };
      _this.onChange(_this, undoData);
    }

    cursor_position_before_change = cp;

    if(syntax_error_timer) {
      clearTimeout(syntax_error_timer);
      syntax_error_timer = null;
    }

    var status = page?.applyChangesToSource(code, true, false, false, 'editor');

    if(status?.errors?.length) {

      syntax_error_timer = setTimeout(function() {
        pinegrow.showQuickMessage('The document has syntax errors.', 3000,
          true);
        syntax_error_timer = null;
      }, 3000);

      _this.setErrors(status.errors);

    } else {
        _this.setErrors(null);
    }

    var took = new Date().getTime() - st;
    if(took < change_delay) {
      change_delay = Math.max(took, 300);
    } else {
      change_delay = Math.min(took, 1000);
    }
    //console.log('change delay ' + change_delay)
  };

  CodeEditorVsc.onDidChangeModelContent((e) => {
    if(ignore_change > 0) {
      ignore_change--;
      return;
    }
    if(change_timer) {
      clearTimeout(change_timer);
      change_timer = null;
    }
    change_timer = setTimeout(function() {
      doChange();// update the view
    }, change_delay);
  });

  var cursor_position_before_change = null;


  var setPageStackCodeEditorActive = function(v) {
    var pages_stack = myLayout.root.getItemsByFilter(function(item) {
      if(item.config.pages) return true;
    })[0];
    v
      ? pages_stack.header.element.addClass('code-editor-active')
      : pages_stack.header.element.removeClass('code-editor-active');
  };

  this.tabWasActivated = function() {
    // if(_this.page.onlyCode) {
    pinegrow.lastActivePage.setPage(_this.page, function() {
      if(_this.page) { //check if editor not destroyed yet
        _this.view.tab.tab.element.removeClass('active-editor');
      }
      setPageStackCodeEditorActive(false);
    });
    //  } else {
    //      pinegrow.lastActivePage.setPage(null)
    //  }
    _this.view.tab.tab.element.parent().find('.active-editor').removeClass(
      'active-editor');
    _this.view.tab.tab.element.addClass('active-editor');
    setPageStackCodeEditorActive(true);
  };

  CodeEditorVsc.onDidFocusEditorText(() => {
    _this.tabWasActivated();
    pgEditorAttributeIntellisense.focusedUrl = url;
  });


  CodeEditorVsc.onKeyDown(e => {
    if(read_only) {
      pinegrow.showQuickError('The stylesheet is locked.', true);
    }
    // this.forceApplyChanges();
  });

  this.dispose = function() {
    // model.dispose();
    CodeEditorVsc.dispose();
  };

  this.destroy = function() {
    if(change_timer) {
      clearTimeout(change_timer);
      change_timer = null;
      doChange();
    }

    if(noticeTimer) {
      clearTimeout(noticeTimer);
      noticeTimer = false;
    }

    this.page = null;
    pinegrow.code_editors.unregister(CodeEditorVsc);
    $textEdit.remove();
    this.dispose();
  };

  let isPhp = false;
  /**
   * @param {string} c
   * **/
  this.setCode = function(c, info, skip_do_change) {
    this.view.view.whenVisible(async function() {
      let model = CodeEditorVsc.getModel();

      let has_focus = CodeEditorVsc.hasWidgetFocus();

      code = c;
      if(code !== null && code.length) ignore_change++;
      currentSourceNode = null;
      var cp;
      if(info && info.undoData && info.undoOrRedo) {
          if(info.undoOrRedo === 'undo') {
              if(info.undoData.cursorPosition) cp = info.undoData.cursorPosition;
          } else if(info.undoData.cursorPositionAfter) { //redo
              cp = info.undoData.cursorPositionAfter;
          }
          _this.setErrors(null);
      }

      !cp && (cp = CodeEditorVsc.getPosition());

      // var si = {left: CodeEditorVsc.getScrollLeft(), top: CodeEditorVsc.getScrollTop()};
      CodeEditorVsc.setValue(code);
      has_focus && CodeEditorVsc.focus(); //only set focus if it had the focus before
      cp && CodeEditorVsc.setPosition(cp);

      cursor_position_before_change = cp;
      // CodeEditorVsc.revealPositionInCenter(cp, 0);

      if(c.match(/<\?php/igm) && !isPhp) {
        CodeEditorVsc.setModel(
          monaco.editor.createModel(CodeEditorVsc.getValue(), 'php'));
        await pgEditorAttributeIntellisense.setHtmlCompletion(url, null, 'php');

        isPhp = true;
      }
    }, 'setCode');
  };

  this.getCode = function() {
    return code;
  };

  this.getDoc = function() {
    return CodeEditorVsc.getValue();
  };

  this.elementWasSelectedInPreview = function(path) {
    doWithCurrentSourceNode(function(sourceNode) {
      var nodeAtPath = sourceNode.getNodeFromPath(path);
      if(nodeAtPath) {
        selectNode(nodeAtPath, last_selected_path != path /* scroll if not selected in editor */);
      }
    });
  };

  this.onChange = null;
  this.onElementSelectedInEditor = null;

  var currentSourceNode = null;

  var doWithCurrentSourceNode = function(func) {
    if(currentSourceNode) {
      func(currentSourceNode);
    } else {
      var p = new pgParser();
      p.assignIds = false;
      p.parse(code, function() {
        currentSourceNode = p.rootNode;
        func(currentSourceNode);
      });
    }
  };

  var selectNode = function(node, scroll) {
    var sourcePos = node.getPositionInSource();
    let model = CodeEditorVsc.getModel();

    var posStart = model.getPositionAt(sourcePos.start);
    var posEnd = model.getPositionAt(sourcePos.end);

    var range = {
      startColumn: posStart.column,
      startLineNumber: posStart.lineNumber,
      endColumn: posEnd.column,
      endLineNumber: posEnd.lineNumber,
    };

    // if (scroll){
    CodeEditorVsc.revealRangeInCenter(range);
    // }

    CodeEditorVsc.setSelection(range);
  };

  this.highlightRange = function(range) {
      if(!(range instanceof monaco.Range)) {
          range = new monaco.Range(range.start.line, range.start.ch, range.end.line, range.end.ch);
      }
      CodeEditorVsc.revealRangeInCenter(range);
      CodeEditorVsc.setSelection(range);
  }

  var notice = null;

  var noticeTimer = null;

  var errors = null;

  this.forceApplyChanges = function() {
    if( !errors) {
      //nothing to apply
    } else {
      pinegrow.showAlert(
        '<p>Are you sure you want to apply code with syntax error to the page?</p><p>This can cause serious problems. Only proceed if you are sure you know what you are doing.</p>',
        'Are you sure?', 'Cancel', 'Proceed', null, function() {
          var code = CodeEditorVsc.getValue();
          _this.page.setCode(code, {action: 'Apply page code'});
          _this.page.refresh();
          _this.setErrors(null);

        });
    }
  };

  this.setErrors = async function(list) {
    errors = list;
    if(!list && !notice) return;

    if( !notice) {
      notice = new PgCodePanelNotice();
      notice.hideOnEmpty = true;
      notice.get$Element().appendTo($textEdit.parent());

      notice.get$Element().on('click', 'button.apply-changes', function(e) {
        e.preventDefault();
        _this.forceApplyChanges();
      });
    }
    if(noticeTimer) {
      clearTimeout(noticeTimer);
      noticeTimer = false;
    }

    if(list && list.length) {
      noticeTimer = setTimeout(function() {
        updateNoticePosition(
          'The code has syntax errors. Changes have not been applied to the page. <button class="pg-button apply-changes">Apply changes</button>');
        noticeTimer = null;
      }, 750);
    } else {
      notice.setMessage('');
    }

  };

  var updateNoticePosition = function(msg) {
    if( !notice) return;
    msg = msg === undefined ? notice.getMessage() : msg;
    if(msg) {
      // var cp = CodeEditorVsc.getPosition();
      var si = {
        left: CodeEditorVsc.getScrollLeft(),
        top: CodeEditorVsc.getScrollTop(),
      };
      // cp.top -= si.top;
      var top = CodeEditorVsc.getContentHeight() - si.top * 2 < 60;
      notice.setMessage(msg, top);
    }
  };

  //pgEditorAttributeIntellisense.setHtmlCompletion(url, null,
    //CodeEditorVsc.getModel()._languageIdentifier.language);

pgEditorAttributeIntellisense.setHtmlCompletion(url, null,
        CodeEditorVsc.getModel().getLanguageId());


};

var PgInternalCodeEditorView = function(
  current_page, read_only, $future_parent) {
  var _this = this;

  var internalEditor = null;

  var $element = $('<div class="pg-page-code"></div>');

  var view = new PgUIView($element, null, 'Edit code');

  this.view = view;

  var wrap = pinegrow.getSetting('edit-page-html-wrap', '1') === '1';

  var search = new PgSearchBox({
    placeholder: 'search',
    placeholder_active: null,
    toolbar_on: true,
  });

  var toolbar = search.getToolbar();

  var $toolbar_section = toolbar.addSection('settings').$element;

  var wrap_control = new PgCheckbox($toolbar_section, 'Wrap', '1');
  wrap_control.show();
  wrap_control.get$input().on('change', function(e) {
    internalEditor.setWrap(wrap_control.getValue());
  });
  wrap_control.setValue(wrap ? '1' : '0');

  //search.$element.appendTo($element);

  search.onSearch = function(f) {
    internalEditor.search(f);
  };

  var spacer = new PgSearchBoxSpacer();
  //spacer.$element.appendTo($element);

  var $content = $('<div class="pg-page-code-content"></div>').appendTo(
    $element);

  this.search = function(f) {
    internalEditor.search(f);
  };

  this.getEditor = function() {
    return internalEditor;
  };

  this.setWrap = function(w) {
    internalEditor.setWrap(w);
  };

  this.setLint = function(w) {
    internalEditor.setLint(w);
  };

  this.setReadOnly = function(val) {
    internalEditor.setReadOnly(val);
  };

  this.show = function(page) {
    var not_html = !pinegrow.isFileEditable(page.url) &&
      !crsaIsRemoteUrl(page.url) && !page.url.endsWith('.php');
    var $textEdit = $('<div class="pg-page-code-editor"></div>').appendTo(
      $content);

    var let_monaco_handle_undo = false;

    if( !page.openedInEditor &&
      !pinegrow.stylesheets.getByUrl(page.url, true)) {
      //let Monaco handle undo/redo
      let_monaco_handle_undo = true;
    }

    internalEditor = pinegrow.codeEditors.open(page.url, page.getSource(), null,
      not_html, $content.get(0).ownerDocument.defaultView,
      let_monaco_handle_undo, this);
    internalEditor.page = page;
    // internalEditor.setLint(internalEditor.getLint()); //prevent lint on pages with scripts
    internalEditor.appendTo($textEdit);
    //internalEditor.view = this;

    internalEditor.setReadOnly(read_only);
    internalEditor.focus();
  };

  var onPageSelected = function(page) {
    // _this.show(page);
  };

  // pinegrow.addEventHandler('on_page_selected', onPageSelected);

  view.onResize = function() {
    if(internalEditor) internalEditor.resize();
  };

  view.onDestroy = function() {
    pinegrow.codeEditors.closeEditor(internalEditor);
    internalEditor.destroy();
    // pinegrow.removeEventHandler('on_page_selected', onPageSelected);
    _this.tab.pgEditorView = null;
    _this.tab = null;
    current_page.codeEditor = null;

    //if page was only opened in editor close it
    if(current_page.onlyCode) {
      current_page.close();
    }

    pgPageCodeEditorsStack.unregisterEditor(_this);
    pgPageCodeEditorsStack.hideIfNoEditors();
  };

  view.onBeforeMovedToAnotherWindow = function(win, el) {
    if(internalEditor) {
      monacoInstance.editorsArr.push(current_page.url);
      pgEditorAttributeIntellisense.Context[current_page.url]?.htmlIntellisense.dispose();
      internalEditor.destroy();
      _this.closeTab();
      internalEditor = null;
    }
  };

  this.getPage = function() {
    return current_page;
  };

  this.getUrl = function() {
    return internalEditor.url;
  };

  this.setModified = function(v) {
    if(v) {
      this.tab.tab.element.addClass('pg-modified');
    } else {
      this.tab.tab.element.removeClass('pg-modified');
    }
  };

  pgPageCodeEditorsStack.addTab(this);

  this.show(current_page);

  pgPageCodeEditorsStack.registerEditor(this);

  this.closeTab = function() {
    pgPageCodeEditorsStack.closeTab(this);
  };

  this.updateTitle = function() {
    this.tab.tab.element.find('.lm_title').text(current_page.getName());
  };

  this.pageWasRenamed = function() {
    internalEditor.url = current_page.url;
    this.updateTitle();
  };
};

var PgPageCodeEditorsStack = function() {

  var code_stack = null;

  var _this = this;
  this.auto = true;

  var getStack = function() {
    try {
      if( !code_stack) {
        code_stack = myLayout.root.getItemsByFilter(function(item) {
          if(item.config.codeEditors) return true;
        })[0];
      }
      return code_stack;
    } catch(err) {

    }
    return null;
  };

  this.getStack = function() {
    return getStack();
  };

  this.setStack = function(val) {
    code_stack = val;
    if(this.isShown) {
      this.show();
    } else {
      this.hide();
    }
    getStack();
    if(toolbar) {
      code_stack.header.element.append(toolbar.$element);
    }
  };

  this.beforeDestroy = function() {
    if(toolbar) {
      //will attach it again. this prevents from being destoryed
      toolbar.$element.detach();
    }
  };

  this.isShown = false;

  this.showEditor = function(editor) {
    this.show();
    editor.nextContentItemChangedIsNotInteractive = true;
    getStack().setActiveContentItem(editor.tab);
  };

  this.isEditorShown = function(editor) {
    return this.isShown &&
      ( !editor || getStack().getActiveContentItem().pgEditorView == editor );
  };

  this.show = function() {
    var stack = getStack();
    stack.config.isHidden = true;
    stack.showStack();
    this.isShown = true;
    pinegrow.editCodeButton.setActive(true);

    var win = pgGetAppWindowForDOMElement(stack.element.get(0));

    if(win.window !== window) {
      win.show();
      pgUpdateAppWindowTitle(win);
    }
  };

  this.hide = function() {
    var stack = getStack();
    stack.hideStack();
    this.isShown = false;
    if(pinegrow.editCodeButton) {
      pinegrow.editCodeButton.setActive(false);
    }

    var win = pgGetAppWindowForDOMElement(stack.element.get(0));

    if(win.window !== window) {
      //it is in external window and only one stack lets hide the window
      if(stack.element.closest('body').find('.lm_stack').length === 1) {
        win.hide();
      }
    }

    var pages_stack = myLayout.root.getItemsByFilter(function(item) {
      if(item.config.pages) return true;
    })[0];
    pages_stack.header.element.removeClass('code-editor-active');
  };

  this.addTab = function(editor) {
    var new_tab = myLayout._$normalizeContentItem({
      title: editor.getPage().name,
      type: 'component',
      componentName: 'code_editor',
      componentState: {},
      /*validDropLocations: ".pg-fullsize-item",
       checkDropArea: function(area) {
       return area.contentItem.config.pages || false;
       }*/
    });
    new_tab.pgEditorView = editor;

    var stack = getStack();
    stack.addChild(new_tab);

    new_tab.tab.element.addClass('file-name-tab');

    $('<modified>*</modified>').insertAfter(
      new_tab.tab.element.find('.lm_title'));

    $('<div class="pg-selected-page-line"></div>').insertAfter(
      new_tab.tab.element.find('.lm_title'));

    new_tab.container.getElement().append(editor.view.get$Element());

    editor.view.addedToLayout(new_tab.container);

    this.show();

    editor.tab = new_tab;

    return new_tab;
  };

  this.closeTab = function(editor) {
    var stack = getStack();
    editor.tab.remove(); //will call editor destroy
    if(stack.contentItems.length == 0) {
      this.hide();
    }
  };

  var editors = [];

  this.registerEditor = function(e) {
    editors.push(e);
  };

  this.unregisterEditor = function(e) {
    var idx = editors.indexOf(e);
    if(idx >= 0) {
      editors.splice(idx, 1);
    }
  };

  var allEditors = function() {
    return editors;
  };

  this.hideIfNoEditors = function() {
    if(allEditors().length == 0) {
      this.hide();
    }
  };

  var toolbar = null;

  var currentEditor = null;

  this.getCurrentEditor = function() {
    return currentEditor;
  };

  $('body').one('pinegrow-ready', function() {
    pinegrow.addEventHandler('on_panel_open_in_floating_window',
      function(pagenull, stack, column, win_options) {
        if(column && column.contentItems.length === 1 &&
          column.contentItems[0].config.codeEditors) {
          win_options.show = _this.isShown;
        }
      });
  });

  $('body').one('pinegrow-layout-ready', function() {

    if( !pgf.edit_page_code) return;

    pinegrow.addEventHandler('on_page_selected', function(page) {
      if(_this.auto && _this.isShown && page) {
        if(page.codeEditor) {
          if( !_this.isEditorShown(page.codeEditor)) {
            _this.showEditor(page.codeEditor);
          }
        } else {
          page.codeEditor = new PgInternalCodeEditorView(page);
        }
      }
    });

    _this.auto = pinegrow.getSetting('edit-page-code-auto', '1') === '1';
    _this.wrap = pinegrow.getSetting('edit-page-html-wrap', '1') === '1';
    _this.lint = pinegrow.getSetting('edit-page-html-lint', '1') === '1';
    _this.highlight_css_rule = pinegrow.getSetting('edit-page-css-rule',
      '1') === '1';

    var stack = getStack();

    if( !stack) {
      pinegrow.showAlert(
        '<p>Please restart Pinegrow. The workspace will be restored to the default layout.</p>',
        'Ups... There is a problem with the workspace');
      delete localStorage['pg_layout_state' + pgf.layout_prefix];
    if(pgf.settings_on_backend) {
        pinegrow.setSetting('pg_layout_state' + pgf.layout_prefix, '');
    }

      return;
    }

    stack.on('activeContentItemChanged', function(contentItem) {
      if(contentItem) {
        currentEditor = contentItem.pgEditorView;
        if( !currentEditor.nextContentItemChangedIsNotInteractive) {
          var e = currentEditor.getEditor();
          e && e.tabWasActivated();
        } else {
          currentEditor.nextContentItemChangedIsNotInteractive = false;
        }
      } else {
        currentEditor = null;
      }
    });

    stack.on('activeContentItemClicked', function(contentItem) {
      if(contentItem) {
        currentEditor = contentItem.pgEditorView;
        var e = currentEditor.getEditor();
        e && e.tabWasActivated();
      }
    });

    toolbar = new PgButtonToolbar();
    toolbar.addSection('icons');

    var autoButton = new PgButtonToolbarItem('icon-down',
      'Code editor options');

    toolbar.addItem('icons', autoButton);

    autoButton.onClick = function() {
      var actions = [
        {
          type: 'header',
          label: 'General options',
        },
        {
          label: 'Wrap lines',
          check: function() {
            return _this.wrap;
          },
          action: function() {
            _this.wrap = !_this.wrap;
            allEditors().forEach(function(e) {
              e.setWrap(_this.wrap);
            });
            // pinegrow.setSetting('edit-page-html-wrap', _this.wrap ? "1" : "0");
            pinegrow.showMode('Wrap lines', _this.wrap);
          },
        },
        {
          type: 'divider',
        },
        {
          type: 'header',
          label: 'HTML pages',
        },
        {
          label: 'Show code for the selected page',
          check: function() {
            return _this.auto;
          },
          action: function() {
            _this.auto = !_this.auto;
            pinegrow.setSetting('edit-page-code-auto', _this.auto ? '1' : '0');
            pinegrow.showMode('Show code for selected page', _this.auto);
          },
        },
        {
          label: 'Display HTML errors and warnings',
          check: function() {
            return _this.lint;
          },
          action: function() {
            _this.lint = !_this.lint;
            allEditors().forEach(function(e) {
              e.setLint(_this.lint);
            });
            pinegrow.setSetting('edit-page-html-lint', _this.lint ? '1' : '0');
            pinegrow.showMode('Display HTML errors and warnings', _this.lint);
          },
        },
        {
          type: 'divider',
        },
        {
          type: 'header',
          label: 'CSS stylesheets',
        },
        {
          label: 'Search for the selected CSS rule<br><span style="opacity:0.7;font-size:11px;">(if the stylesheet is open in the code editor)</span><br><span style="opacity:0.7;font-size:11px;">Use ' +
            pgShowKbdCombo('CMD+G') + ' to move between search hits.</span>',
          check: function() {
            return _this.highlight_css_rule;
          },
          action: function() {
            _this.highlight_css_rule = !_this.highlight_css_rule;
            pinegrow.setSetting('edit-page-css-rule',
              _this.highlight_css_rule ? '1' : '0');
            pinegrow.showMode('Search for the selected CSS rule',
              _this.highlight_css_rule);
          },
        },
      ];

      var menu = new CrsaContextMenu(actions, this.$element);
      menu.showForElement(this.$element);
      return false;
    };

    /*
     var winButton = new PgButtonToolbarItem('icon-window', 'Open in separate window.');

     toolbar.addItem('icons', winButton);

     autoButton.onClick = function() {
     var active = !this.getActive();
     this.setActive(active);
     }
     */
    stack.header.element.append(toolbar.$element);
    //stack.header.element.find('.lm_tabs').css('padding-right', '90px');

    _this.hide();
  });

};

var pgPageCodeEditorsStack = new PgPageCodeEditorsStack();

