var PgProjectDBTreeCollapseMap = function() {

    var map = {}

    var getKey = function(node) {

        switch(node.tagName) {
            case 'document':
                return node.getAttr('url');
            case 'element':
                var parent = node.parent.parent;
                return getKey(parent) + '-' + node.getAttr('selector');
            case 'action':
                return getKey(node.parent) + '-' + node.getAttr('id');
        }
        return '';
    }

    this.setCollapsed = function(node, val) {
        var key = getKey(node);
        if(key) {
            map[key] = val;
        }
    }

    this.isCollapsed = function(node) {
        var key = getKey(node);
        if(key) {
            return map[key] || false;
        }
    }
}

var PgProjectDBTreeView = function() {

    pinegrow.registerTreeView(this);

    pinegrow.projectDBTreeView = this;

    var $element = $('<div/>', {id: 'pg-tree', class: 'pg-tree pg-db-tree'});

    var api = new PgApi();

    var sel_helper = new PgElementSelectorHelper();

    var collapsed_map = new PgProjectDBTreeCollapseMap();

    var delayed_tasks = new PgDelayedTasks();

    var $tooltip_element = null;

    var getStatusDescription = function(s, node, mapped_to_pgel) {
        switch(s) {
            case 'ok':
                return `<b>The mapping is ok.</b> This action set is correctly mapped to the element ${node.getAttr('desc')}.`;
                break;

            case 'unused':
                return `<b>Not used.</b> The action set is currently not mapped to any element on the page.`;
            case 'none':
                return `<p><b>The target element is not found.</b> The action set's selector doesn't map to any element on the page. Previously it mapped to ${node.getAttr('desc')}.</p><p>Did you change the page structure, element ids or classes that are used in this selector?</p><p><b>To fix the issue</b>, right-click on the Action set and remap it to the selected element.</p>`;

            case 'taken':
                return `<p><b>Target element is already mapped to another action set.</b> The action set's selector points to the element on the page, but that element is already mapped to another action set.</p><p><b>To fix the issue</b>, right-click on the Action set and remap it to the original element or map it to another element on the page.</p>`

            case 'multiple':
                return `<p><b>Multiple elements on the page match the selector.</b> The selector for this action set is not unique.</p><p><b>To fix the issue</b>, right-click on the Action set and remap it to the original element or map it to another element on the page.</p>`

        }
        return null;
    }

    var search = new PgSearchBox({
        toolbar_on: true,
        placeholder: 'search'
    });

    search.addIcon({
        icon: 'icon-130',
        label: 'Select search results',
        func: function() {
            searchSelect('select');
        }
    })

    search.addIcon({
        icon: 'icon-137',
        label: 'Add search results to selection',
        func: function() {
            searchSelect('add');
        }
    })

    search.addIcon({
        icon: 'icon-138',
        label: 'Deselect search results',
        func: function() {
            searchSelect('deselect');
        }
    })

    var toolbar = this.toolbar = search.getToolbar();

    var $toolbar_section = toolbar.addSection('settings').$element;

    var stylesButton = new PgButtonToolbarItem('icon-filter', 'Select what kind of elements to show in the tree.');

    toolbar.addItem('settings', stylesButton);

    var filters = {
        issues: {
            label: 'Issues',
            desc: 'Show only Action sets with mapping issues'
        },
        unused: {
            label: 'Unused',
            desc: 'Show only unused Action sets'
        }
    }

    var default_filter_options = {};

    $.each(filters, function(key, name) {
        default_filter_options[key] = false;
    })

    var filter_options = default_filter_options;

    stylesButton.onClick = function () {

        var menu = new CrsaContextMenu(null, this.$element);

        menu.add({
            type: 'header',
            label: 'Show only:'
        });

        $.each(filters, function(key, info) {
            menu.add({
                label: info.label,
                action: function() {
                    var on = !(filter_options[key] === true);
                    _this.setFilterOption(on ? key : null);
                },
                check: filter_options[key] === true
            })
        })

        menu.add({
            type: 'divider'
        });

        menu.add({
            label: 'Clear filter',
            action: function() {
                _this.setFilterOption(null);
            }
        })

        menu.showForElement(this.$element, null, 8);
    }

    this.setFilterOption = function(key, skip_update) {
        filter_options = $.extend({}, default_filter_options);
        if(key) {
            var info = filters[key];
            filter_options[key] = true;
            stylesButton.setActive(true);
            pinegrow.showMode(info.desc, true);
            setFilterMessage(info.label);
        } else {
            stylesButton.setActive(false);
            setFilterMessage();
        }
        cs_tree.setFilterOptions(filter_options);
        if(!skip_update) {
            cs_tree.research(true);
        }
    }

    search.$element.appendTo($element);

    var $selects = search.$element.find('.pg-button-toolbar-section.icons');

    var spacer = new PgSearchBoxSpacer();
    spacer.$element.appendTo($element);

    search.onSearch = function(filter) {
        var p = new CrsaProfile();
        if(cs_tree) cs_tree.search(filter);

        if(filter.length) {
            pg$Show($selects, 'inline-block');
        } else {
            pg$Hide($selects);
        }

        p.show('SEARCH');
    }

    pg$Hide($selects);

    var $filterMsg = $('<div class="pg-tree-filter-msg"><div>Show only: <span></span><i class="close icon icon-close"></i></div></div>').appendTo($element);

    pg$Hide($filterMsg);

    $filterMsg.on('click', function(e) {
        e.preventDefault();
        _this.setFilterOption(null);
    })

    var setFilterMessage = function(msg) {
        if(msg) {
            $filterMsg.find('span').html(msg);
            pg$Show($filterMsg);
            $element.addClass('with-filter-msg');
        } else {
            pg$Hide($filterMsg);
            $element.removeClass('with-filter-msg');
        }
    }

    var $tree = $('<div/>', {class: 'pg-tree-container'}).appendTo($element);

    var $content = $element.find('.editor');

    var view = this.view = new PgUIView($element, 'db_tree', 'Mappings');

    view.classForContainer = 'pg-tree-panel-container';

    var _this = this;

    view.tabIcon = 'icon-mapping';

    var currentPage;

    view.onShow = function() {
        //scroll to selected element if not visible
        return;

        if(currentPage && currentPage.tree.scrollToSelected) {
            tableGrid.scrollToDelayed(currentPage.tree.scrollToSelected);
            currentPage.tree.scrollToSelected = null;
        }
    }


    var menuItems = [
        {
            icon: 'icon-bin',
            name: 'Delete',
            func: function(data) {

            }
        },
        {
            icon: 'icon-duplicate',
            name: 'Duplicate',
            func: function(data) {

            }
        }
    ]


    var menu = new PgTreeNodeMenu(menuItems);

    menu.hide();

    var dimension;

    var hasHScrollBar = true;
    var hasVScrollBar = true;

    var tableGrid = new PgTableGrid();

    tableGrid.draggable = false;

    var quickInsertConnector = null;

    tableGrid.onScroll = function() {
        updateMenuPosition();
    }

    $tree.append(tableGrid.$element);

    menu.$element.appendTo($element);

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell-document"><i data-element="icon" class="collapse-icon icon-down"></i><name data-element="name"></name></div>')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-project"><i data-element="icon" class="collapse-icon icon-down"></i><name data-element="name"></name></div>', 'project')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-document"><i data-element="icon" class="collapse-icon icon-down"></i><i class="icon icon-Page decoration"></i><pagename data-element="name"></pagename></div>', 'document')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-element"><i data-element="icon" class="collapse-icon icon-down"></i></i><name data-element="name"></name><problem data-element="problem"></problem></div>', 'element')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-selector"><i class="icon icon-267 decoration"></i><name data-element="name"></name></div>', 'selector')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-mapped"><span data-element="deco"></span><name data-element="name"></name></div>', 'mapped')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-action"><i data-element="icon" class="collapse-icon icon-down"></i><i class="icon icon-WP decoration"></i><action data-element="name"></action></div>', 'action')

    tableGrid.setTemplate('<div class="pg-pdb-tree-cell pg-pdb-tree-cell-action-param"><name data-element="name"></name><value data-element="value"></value></div>', 'action_param')

    tableGrid.onSetDataToCell = function(data, cell) {

        cs_tree.populateDataItem(data, currentPage);

        var pgel = data.pgel;

        switch(pgel.tagName) {
            case 'document':
                cell.setHtml('name', pgel.getAttr('url'));
                break;
            case 'action':
                cell.setHtml('name', data.ai.def.name);
                break;
            case 'action_param':
                cell.setHtml('name', pgel.name + (pgel.value !== null ? ': ' : ''));
                cell.setHtml('value', pgel.value);
                break;
            case 'element':
                cell.setHtml('name', pgel.getAttr('name') || '<span style="opacity:0.5;">No name</span>');
                var status = data.pgel.getAttr('status');
                if(status === 'ok' || status === 'unused') {
                    cell.$element.removeClass('problem');
                    cell.addClass('problem', 'hide')

               /* } else if(status === 'unused') {
                    cell.$element.removeClass('problem');
                    cell.removeClass('problem', 'hide')
                    cell.setHtml('problem', `<span><i class="icon icon-269"></i> Not used.</span>`)*/
                } else {
                    cell.$element.addClass('problem');
                    cell.removeClass('problem', 'hide')

                    var err = '';

                    switch(status) {
                        case 'none':
                            err = 'No element matches the selector.';
                            break;
                        case 'multiple':
                            err = 'Multiple elements match the selector.';
                            break;
                        case 'taken':
                            err = 'The matched element is already linked.'
                            break;
                        case 'unused':
                            err = 'Not used.'
                            break;
                    }
                    //cell.setHtml('problem', `<i class="icon-no-affected"></i> ${status === 'none' ? `0 <span><i class="icon icon-269"></i> No element matches the selector</span>` : 5}`)

                    cell.setHtml('problem', `<span><i class="icon icon-269"></i> ${err}</span>`)
                }
                //var status = pgel.getAttr('status');
                break;
            case 'selector':
                cell.setHtml('name', data.element_pgel.getAttr('selector') || 'Not used.');

                break;
            case 'mapped':
                cell.setHtml('name', data.element_pgel.getAttr('desc'));
                var status = data.element_pgel.getAttr('status');
                if(status === 'ok' || status === 'taken' || status === 'multiple') {
                    cell.$element.removeClass('problem');
                    cell.setHtml('deco', `<i class="icon icon-right-arrow decoration"></i>`)
                } else {
                    cell.$element.addClass('problem');
                    cell.setHtml('deco', `<i class="icon icon-close decoration"></i>`)
                }
                break;
            default:
                cell.setHtml('name', pgel.tagName);
        }


        if((pgel.pg_selected || false) !== (cell.isSelected || false)) {
            if (pgel.pg_selected) {
                cell.$element.addClass('selected');
            } else {
                cell.$element.removeClass('selected');
            }
            cell.isSelected = pgel.pg_selected || false;
        }

        if(data.border !== (cell.hasBorder || null)) {
            if(data.border) {
                cell.$element.removeClass('no-border');
            } else {
                cell.$element.addClass('no-border');
            }
            cell.hasBorder = data.border;
        }
        if(tableGrid.hasChildren(data)) {
            //cell.removeClass('icon', 'noicon');
            if (data.collapsed) {
                cell.addClass('icon', 'icon-right');
                cell.removeClass('icon', 'icon-minus', 'icon-down');
                cell.$element.addClass('collapsed');
            } else {
                cell.removeClass('icon', 'icon-minus', 'icon-right');
                cell.addClass('icon', 'icon-down');
                cell.$element.removeClass('collapsed');
            }
        } else {
            cell.removeClass('icon', 'icon-down', 'icon-right');
            cell.addClass('icon', 'icon-minus');
            cell.$element.removeClass('collapsed');
        }

        //var search = data.search || 0;
        //cell.$element.removeClass('search-1 search-2 search-0').addClass('search-' + search);
    }

    tableGrid.onRepaint = function() {
        menu.setBottom(tableGrid.getHScrollBarSize());
        menu.setRight(tableGrid.getVScrollBarSize());
        updateMenuPosition();
    }

    tableGrid.onSetData = function() {
        //updated borders of visible data
        var list = tableGrid.visibleData;
        for(var i = 0; i < list.length; i++) {
            list[i].border = true;
            if(list[i].template_type === 'element') {
                if(i > 0) {
                    list[i - 1].border = false;
                }
            }
        }
    }

    var refreshDataItem = function(data) {
        if(data) {
            data.populated = false;
            if(tableGrid) {
                tableGrid.updateItem(data);
            }
        }
    }

    var refreshDataItems = function(datas) {
        datas.forEach(function(data) {
            data.populated = false;
        })
        if(tableGrid) {
            tableGrid.updateItems(datas);
        }
    }

    var refreshGrid = function(done) {
        nextFrame.do('refreshGrid', function() {
            //_this.research();
            if(tableGrid) {
                tableGrid.refresh();

                if(!view.isVisible()) {
                    view.whenVisible(function() {
                        tableGrid.show();
                        if(done) done();
                    })
                } else {
                    if(done) done();
                }
            }
        })
    }


    var wireTreeElements = function() {



        tableGrid.$container
            .on('contextmenu', '.pg-table-grid-cell', function (event) {
                event.preventDefault();
                event.stopPropagation();

                var project = pinegrow.getCurrentProject();

                var $el = $(this)
                var data = getDataForTree$Element($el);

                var selected_pgel = pinegrow.selectedElements.getLastSelected();

                var actions = []
                
                var closest_element = data.element_pgel || (data.pgel ? data.pgel.closest('element') : null);
                
                if(closest_element) {

                    var page = sel_helper.getOpenPageForElement(closest_element);

                    if(!sel_helper.getOpenPageForElement(closest_element)) {
                        actions.push({
                            type: 'html',
                            html: '<p class="pg-menu-html-desc">Please open the page first. You can click on the Action set to open its page.</p>'
                        })

                    } else {
                        var status = closest_element.getAttr('status');
                        var selector = closest_element.getAttr('selector');
                        var namedesc = closest_element.getAttr('desc');

                        if(status === 'taken') {
                            var should_be_mapped_to_pgel = sel_helper.getPgelTargetedBySelector(closest_element);
                            if(should_be_mapped_to_pgel) {
                                var taken_by_element = project.db.getElementNodeForSelector(should_be_mapped_to_pgel.getAttr('data-pg-sel'), page);
                                if(taken_by_element) {
                                    actions.push({
                                        type: 'html',
                                        html: `<p class="pg-menu-html-desc">This action set should be mapped to element ${should_be_mapped_to_pgel.getName({short:true})}. But that element is already mapped to action set ${sel_helper.getActionSetName(taken_by_element)}.</p>`
                                    })

                                    actions.push({
                                        label: `Jump to ${sel_helper.getActionSetName(taken_by_element)}`,
                                        action: function() {
                                            selectActionSet(taken_by_element);
                                        }
                                    })

                                    actions.push({
                                        type: 'divider'
                                    })
                                }
                            }
                        }

                        var mapTo = function() {
                            if(pinegrow.selectedElements.getLength() > 1) {
                                actions.push({
                                    type: 'html',
                                    html: '<p class="pg-menu-html-desc">Multiple elements are selected on the page. Please select just one element.</p>'
                                })
                                return;
                            }
                            var sel = selected_pgel.getAttr('data-pg-sel');
                            if (sel || sel_helper.isAutoSelectorOn()) {
                                actions.push({
                                    label: `Map to ${selected_pgel.getName({short: true})}`,
                                    action: function () {
                                        sel_helper.mapDBElementToPgel(closest_element, selected_pgel, sel, status === 'multiple' /* force new selector */);
                                    }
                                })

                            } else {

                                var submenu = sel_helper.getActionsForMenu(selected_pgel, function (sel) {
                                    sel_helper.mapDBElementToPgel(closest_element, selected_pgel, sel);
                                })

                                submenu.unshift({
                                    type: 'header',
                                    label: 'Map using the selector:'
                                })
                                actions.push({
                                    label: `Map to ${selected_pgel.getName({short: true})}`,
                                    submenu: submenu
                                })
                            }
                        }

                        actions.push({
                            type: 'header',
                            label: 'Map to:'
                        })
                        if (selected_pgel) {
                            var selected_pgel_element = project.db.getElementNodeForPgel(selected_pgel);

                            if (selected_pgel_element && selected_pgel_element.findOne('action') === null) {
                                //has no actions
                                selected_pgel_element = null;
                            }

                            if (selected_pgel_element) {
                                if (selected_pgel_element === closest_element) {
                                    //How about if it is mapped to multiple?
                                    if(status === 'multiple') {
                                        mapTo();

                                    } else {
                                        //already mapped to itself
                                        actions.push({
                                            type: 'html',
                                            html: '<p class="pg-menu-html-desc">The action set is already mapped to the selected element.</p>'
                                        })
                                        /*
                                         actions.push({
                                         type: 'divider'
                                         })

                                         actions.push({
                                         label: 'Remove from the selected element (and keep it in the project)',
                                         class: 'wrap',
                                         action: function () {
                                         sel_helper.unmapDBElementFromPgel(closest_element, selected_pgel);
                                         }
                                         })*/
                                    }
                                } else {
                                    actions.push({
                                        type: 'html',
                                        html: `<p class="pg-menu-html-desc">${selected_pgel.getName({short: true})} is already mapped to another Action set ${selected_pgel_element.getAttr('name')} | ${selected_pgel_element.getAttr('desc')}</p>`,
                                    })

                                    actions.push({
                                        label: 'Replace it',
                                        action: function () {
                                            sel_helper.unmapOneDBElementFromPgelAndMapAnother(selected_pgel_element, closest_element, selected_pgel);
                                        }
                                    })
                                }
                            } else {
                                mapTo();
                            }

                        } else {
                            actions.push({
                                type: 'html',
                                html: `<p class="pg-menu-html-desc">Select an element on the page first.</p>`,
                            })
                        }
                        if (actions.length) {
                            actions.push({
                                type: 'divider'
                            })
                            actions.push({
                                label: 'Edit name...',
                                action: function() {
                                    pinegrow.showPrompt('<p>Edit the name of the Action set:</p>', 'Choose a descriptive name', closest_element.getAttr('name'), '', null, function(val) {
                                        sel_helper.changeActionSetName(closest_element, val);
                                    })
                                }
                            })
                        }
                        if (actions.length) {
                            actions.push({
                                type: 'divider'
                            })

                            actions.push({
                                label: 'Remove from the page (and keep it in the project as Unused)',
                                class: 'wrap',
                                action: function () {
                                    sel_helper.makeElementUnused(closest_element);
                                }
                            })

                            actions.push({
                                type: 'divider'
                            })

                            actions.push({
                                label: 'Delete...',
                                action: function () {
                                    pinegrow.showAlert(`<p>Do you want to delete this Action set? Deleted actions will be removed from the mapped element and the action set will be removed from the project.</p><p>Deleting the Action set can't be undone.</p><p>If you want to keep the Action set in the project, do the "Remove from the page" instead.</p>`, 'Are you sure?', 'Cancel', 'Delete it', null, function () {
                                        sel_helper.deleteElement(closest_element);
                                    }, 'Just remove it from the page', function() {
                                        sel_helper.makeElementUnused(closest_element);
                                    })
                                }
                            })
                        }
                    }

                    if(actions.length) {
                        var menu = new CrsaContextMenu(actions, $el);
                        menu.showForElement($el)
                    }
                }
                
                return false;

            })
            .on('mouseenter', '.pg-table-grid-cell', function (event) {
                if(!tableGrid.inScroll) {

                    var $e = $(this)
                    var data = getDataForTree$Element($e);

                    //var data = getDataForEvent(event);

                    if(data.template_type === 'element') {
                        var s = data.pgel.getAttr('status');
                        if(s !== 'ok') {
                            var $e = $(this);
                           delayed_tasks.run('elementTooltip', function() {
                                addTooltip($e, getStatusDescription(s, data.pgel), {trigger: 'manual', close_delay: 0})
                                $tooltip_element = $e;
                           }, 300)
                        }
                    }

                }
                event.stopImmediatePropagation();
            })
            .on('mouseleave', '.pg-table-grid-cell', function (event) {
                var $e = $(this);
                if($tooltip_element) {
                    $tooltip_element.tooltip('destroy');
                    $tooltip_element = null;
                }
                delayed_tasks.cancel('elementTooltip')

                if(!tableGrid.inScroll) {

                }
                event.stopImmediatePropagation();
            })
            .on('click', 'i.collapse-icon', function (event) {

                var data = getDataForEvent(event);

                data.collapsed = !(data.collapsed || false);
                if(data.pgel) {
                    collapsed_map.setCollapsed(data.pgel, data.collapsed);
                }

                if(event.altKey) {
                    tableGrid.forEachVisibleSibling(data, function(ch) {
                        ch.collapsed = data.collapsed;
                        if(data.pgel) {
                            collapsed_map.setCollapsed(ch.pgel, data.collapsed);
                        }
                    })
                }

                tableGrid.refresh();
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', 'i.icon-hide', function (event) {

                var data = getDataForEvent(event);
                showItem(data);
                event.preventDefault();
                event.stopPropagation();
            })
            .on('click', '.pg-table-grid-cell', function (event) {

                //event.stopImmediatePropagation();
                //event.preventDefault();

                pgCloseContextMenus();

                var $el = $(this)
                var data = getDataForTree$Element($el);

                var selected_pgel = pinegrow.selectedElements.getLastSelected();

                switch(data.pgel.tagName) {
                    case 'element':
                        sel_helper.selectMappedPgel(data.pgel)
                        break;
                    case 'mapped':
                        sel_helper.selectMatchedPgels(data.element_pgel)
                        break;
                    case 'selector':
                        sel_helper.selectMatchedPgels(data.element_pgel)
                        //sel_helper.selectMappedPgel(data.element_pgel)
                        break;
                    case 'action':
                        sel_helper.selectAction(data.pgel)
                        break;
                    case 'action_param':
                        sel_helper.selectAction(data.pgel.action_pgel)
                        break;
                    case 'document':
                        sel_helper.openDocument(data.pgel);
                        break;
                }
                return false;
                // return false;
            });

    }


    var getDataForEvent = function(event) {
        var div = $(event.currentTarget).closest('div');
        return tableGrid.visibleData[parseInt(div.attr('data-visible-idx'))];
    }

    var getDataForTree$Element = function($el) {
        var div = $el.closest('div');
        return tableGrid.visibleData[parseInt(div.attr('data-visible-idx'))];
    }

    var getCSSNodeForTree$Element = function($el) {
        var data = getDataForTree$Element($el);
        return data ? data.node : null;
    }

    var getDataForCSSNode = function(node) {
        var list = tableGrid.data;
        for(var i = 0; i < list.length; i++) {
            if(list[i].node === node) return list[i];
        }
        return null;
    }

    var getDatasForCSSNodes = function(nodes) {
        var list = tableGrid.data;
        var r = []
        for(var i = 0; i < list.length; i++) {
            if(nodes.indexOf(list[i].node) >= 0) {
                r.push(list[i])
            }
        }
        return r;
    }

    var getDataForNode = function(node) {
        var list = tableGrid.data;
        for(var i = 0; i < list.length; i++) {
            if(list[i].pgel === node) return list[i];
        }
        return null;
    }


    var getPgelOfTreeNode = function ($li) {
        return getPgelOfTreeNodeFast($li.get(0));
    }

    var getPgelOfTreeNodeFast = function (li) {
        var pgid = li.getAttribute('data-pg-id');
        if (pgid) {
            return getPgNodeByPgId(pgid);
        } else {
            return null;
        }
    }

    var getTreeNodeForPgel = function(pgel) {
        var pgid = pgel instanceof pgParserNode ? pgel.getId() : pgel;
        var node = treeTop.find('[data-pg-id="' + pgid + '"]');
        if(node.length) return node;
        return null;
    }


    this.isOverTree = function(event, drag_menu) {

        return null;
    }



    wireTreeElements();



    view.onResize = function() {

    }

    var resizeObserver = new ResizeObserver(function() {
        dimension = tableGrid.$element.get(0).getBoundingClientRect();
        tableGrid.onRepaint();
        tableGrid.show();
        menu.updateLayout();
    });
    resizeObserver.observe(tableGrid.$element.get(0));

    view.onDestroy = function() {
        resizeObserver.disconnect();
        resizeObserver = null;

        delayed_tasks.destroy();

        pinegrow.removeEventHandler('on_projectdb_changed', onProjectDBChanged);
        pinegrow.removeEventHandler('on_project_loaded', onProjectLoaded);
        pinegrow.removeEventHandler('on_project_closed', onProjectClosed);
        pinegrow.removeEventHandler('on_page_changed', onPageChanged);
        pinegrow.removeEventHandler('on_page_loaded', onPageLoaded);
        pinegrow.removeEventHandler('on_page_refreshed', onPageLoaded);
        pinegrow.removeEventHandler('on_page_selected', onPageSelected);
        pinegrow.removeEventHandler('on_drag_drop_operation_done', onDragDropDone);
        pinegrow.removeEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);
        pinegrow.removeEventHandler('on_element_isolated', onElementIsolated);

        $element.off('mousemove', onMouseMoved);
        $element.off('mouseleave', onMouseOut);
        $element.off('mouseenter', onMouseIn);
        //$element.off('mouseenter mouseleave', noop);
    }

    var onPageChanged = function(page, pgel, info) {
    }

    var showTree = function(cs_tree) {
        tableGrid.showInsertionLineAt(-1);
        tableGrid.setEmptyLineAt(-1);
        tableGrid.borderItem(null);

        if(cs_tree) {
            var has_isolated = false;

            cs_tree.setFilterOptions(filter_options);
            cs_tree.setTreeView(_this);
            cs_tree.setTableGrid(tableGrid);

            /*
             var pgel = pinegrow.selectedElements.getLastSelectedOnPage(page);
             if (!pgel) {
             currentMenuData = null;
             menu.hide();
             } else {
             currentMenuData = page.tree.getTreeDataItemForPgId(pgel.getId());
             menu.show();
             updateMenuPosition();
             }*/
        } else {
            currentMenuData = null;
            tableGrid.setData([]);
            menu.hide();
        }
        menu.updateLayout();
    }

    var cs_tree = null;

    var onProjectLoaded = function(page, project) {
        cs_tree = new PgProjectDBTree(project.db, collapsed_map);
        cs_tree.create();
        showTree(cs_tree);
    }

    var onProjectClosed = function(page, project) {
        _this.view.setNotificationBadge(null)
    }

    var onProjectDBChanged = function(pagenull, col, info) {
        var project = pinegrow.getCurrentProject();

        if(true /* !col || col.getLength() === 0 */) {

            if(project.db.selectedElement) {
                project.db.selectedElement.pg_selected = false;

                var pgel = pinegrow.selectedElements.getLastSelected();

                if (pgel) {
                    var element_of_pgel = project.db.getElementNodeForPgel(pgel);
                    project.db.selectedElement = element_of_pgel;
                    if(element_of_pgel) {
                        project.db.selectedElement.pg_selected = true;
                    }
                }
            }

            cs_tree = new PgProjectDBTree(project.db, collapsed_map);
            cs_tree.create();
            showTree(cs_tree);
        }

        updateNotificationBadge();
        if($tooltip_element) {
            $tooltip_element.tooltip('destroy');
            $tooltip_element = null;
        }
    }

    var updateNotificationBadge = function() {
        var project = pinegrow.getCurrentProject();
        var num = 0;
        if(project && project.db) {
            var list = project.db.getElementsWithIssues();
            num = list.length;

            if(doWeHaveAnyNewIssues(list)) {
                sel_helper.displayIssuesNotice(list.length);
            }
        }
        _this.view.setNotificationBadge(num > 0 ? `${num}` : null, `Please resolve ${num} mapping issue${num > 1 ? 's' : ''}.`);
    }

    var currentIssues = [];

    var doWeHaveAnyNewIssues = function(list) {
        var r = list.length > currentIssues.length;
        currentIssues = list;
        return r;
    }

    var selectActionSet = function(element) {
        var datas = [];
        var data = null;
        var project = pinegrow.getCurrentProject();
        if(!project) return;
        if(project.db.selectedElement) {
            project.db.selectedElement.pg_selected = false;
            data = getDataForNode(project.db.selectedElement);
            if(data) {
                datas.push(data)
            }
            project.db.selectedElement = null;
        }

        if(element) {
            element.pg_selected = true;
            project.db.selectedElement = element;

            data = getDataForNode(element);
            if(data) {
                datas.push(data)
                tableGrid.scrollToDelayed(data)
            }
        }
        if(datas.length) {
            refreshDataItems(datas);
        }
    }

    var onElementSelected = function() {
        var project = pinegrow.getCurrentProject();
        if(!project) return;

        var pgel = pinegrow.selectedElements.getLastSelected();
        if(pgel) {
            var element = project.db.getElementNodeForPgel(pgel);
        }
        selectActionSet(element);
    }


    var onPageLoaded = function(page) {
        return;

        page.tree.create();
        if(page === currentPage) {
            showTreeForPage(page);
        }
    }

    var onPageFrameworksChanged = function(page) {
        return;

        if((page && page === currentPage) || (page === null && currentPage)) {
            if(currentPage && currentPage.tree) {
                currentPage.tree.create();
                showTreeForPage(currentPage);
            }
        }
    }

    var onElementIsolated = function(page, pgel) {
        return;

        page = page || currentPage;
        if(page && page.tree) {
            page.tree.create();
            showTreeForPage(page);
        }
    }

    var onPageSelected = function(page) {
        currentPage = page;
        return;

        view.whenVisible(function () {
            dimension = $element.get(0).getBoundingClientRect();
            showTreeForPage(page);
        })
    }


    var noop = function() {}

    var currentMenuData = null;

    var hoveredCell = null;

    var mouseSpeed = new PgMouseSpeed();

    var quickInsertPosition = null;
    var quickInsertPositionTimer = null;


    var onMouseOut = function(event) {
        tableGrid.highlightLevel(-1);
        tableGrid.quickInsertLine.hide();
        if(quickInsertPositionTimer) {
            clearTimeout(quickInsertPositionTimer);
            quickInsertPositionTimer = null;
        }
        menu.hide();
        if (hoveredCell) {
            hoveredCell.$element.removeClass('hovered');
        }
        hoveredCell = null;
    }

    var onMouseIn = function(event) {
        if(currentMenuData) menu.show();
    }

    $element.on('mouseleave', onMouseOut);
    $element.on('mouseenter', onMouseIn);

    pinegrow.addEventHandler('on_selected_elements_changed', onElementSelected);

    pinegrow.addEventHandler('on_project_loaded', onProjectLoaded);
    pinegrow.addEventHandler('on_project_closed', onProjectClosed);
    pinegrow.addEventHandler('on_projectdb_changed', onProjectDBChanged);

    pinegrow.addEventHandler('on_page_changed', onPageChanged);
    pinegrow.addEventHandler('on_page_loaded', onPageLoaded);
    pinegrow.addEventHandler('on_page_refreshed', onPageLoaded);
    pinegrow.addEventHandler('on_page_selected', onPageSelected);
    //pinegrow.addEventHandler('on_drag_drop_operation_done', onDragDropDone);
    //pinegrow.addEventHandler('on_element_isolated', onElementIsolated);

    pinegrow.addEventHandler('on_page_frameworks_changed', onPageFrameworksChanged);

    var updateMenuPosition = function() {
        if(currentMenuData) {
            var so = tableGrid.getScrollOffset();
            var top = tableGrid.getYForVisibleDataIndex(currentMenuData.visible_idx) - so.top;

            menu.showFor(top, top + tableGrid.cellHeight, currentMenuData.selected || false, currentMenuData);
        }
    }

    this.onSelected = function(data) {

        currentMenuData = data;

        var pgel = this.getPgel(data);

        if(pgel) {
            var $item = menu.find('hide');
            $item.find('i').removeClass('icon-hide icon-see').addClass(pgel.isHidden() ? 'icon-see' : 'icon-hide');
        }
        updateMenuPosition();
    }

    this.onDeselected = function(data) {
        //if(currentMenuData === data) setClassesForMenu(data);
        if(!pinegrow.selectedElements.hasSelection()) {
            currentMenuData = null;
            menu.hide();
        }
    }
}







var PgProjectDBTree = function(db, collapsed_map) {

    var _this = this;

    var max_tree_level = 0;
    var li_id_count = 0;
    var currentTreeRoot;

    var filter_options = {
        styles: false,
        images: false
    }

    var has_filter_options = false;

    var treeData = [];
    var treeData_PgIdMap = {};
    var helper;

    var wpf = pinegrow.findFrameworkByKey('wordpress.pinegrow');

    this.setFilterOptions = function(opt) {
        var o = has_filter_options;
        has_filter_options = false;
        $.each(opt, function(key, val) {
            if(val) {
                has_filter_options = true;
                return true;
            }
        })
        filter_options = opt;
        research_on_show = o !== has_filter_options;
    }

    var treeView;

    this.setTreeView = function(tv) {
        treeView = tv;
    }

    var tableGrid = null;

    this.setTableGrid = function(tg) {
        tableGrid = tg;
        this.$element = tableGrid.$element;
        tg.setData(treeData);

        if(research_on_show) {
            this.research();
        }
    }

    var treeTop = $('<div/>').addClass('crsa-tree-branch pg-tree-ul');

    this.$element = null;

    this.highlightOnMouseOver = true;

    this.destroy = function() {
        nextFrame.cancelAll();
        tableGrid = null;
        treeView = null;
        if(helper) helper.destroy();
        helper = null;
        treeData = null;
        treeData_PgIdMap = null;
        _this = null;
    }

    this.getDataFor$Element = function($el) {
        return tableGrid.getDataFor$Element($el);
    }

    this.updatePgIdMap = function() {
        var p = new CrsaProfile();
        treeData_PgIdMap = {};
        treeData.forEach(function(data) {
            if(data.pgid) {
                treeData_PgIdMap[data.pgid] = data;
            }
        })
        p.show('updatePgIdMap');
    }

    var removeItemPgidMap = function(data) {
        if(data && data.pgid) delete treeData_PgIdMap[data.pgid];
    }

    var removeItemAndChildrenFromPgidMap = function(data) {
        helper.withDataAndChildren(data, removeItemPgidMap);
    }

    var getTreeDataItemForPgId = function(pgid) {
        return treeData_PgIdMap[pgid] || null;
    }

    this.getTreeDataItemForPgId = function(pgid) {
        return getTreeDataItemForPgId(pgid);
    }

    var getDataItemsForPgIdAndDescendants = function(pgid) {
        var data = getTreeDataItemForPgId(pgid);
        if(data) {
            var list = [data];
            var num = data.idx + helper.getNumberOfDescendants(data.idx, treeData);
            for(var i = data.idx + 1; i <= num; i++) {
                list.push(treeData[i]);
            }
            return list;
        }
        return null;
    }

    var createDataItemsForPgEl = function(pgel, level) {
        var list = [];
        createDataItemForPgEl(list, pgel, false, level);
        return list;
    }

    var refreshGrid = function() {
        nextFrame.do('refreshGrid', function() {
            _this.research();
            if(tableGrid) {
                tableGrid.refresh();

                if(treeView && !treeView.view.isVisible()) {
                    treeView.view.whenVisible(function() {
                        tableGrid.show();
                    })
                }
            }
        })
    }

    var previous_selector_was_invalid = false;

    const SEARCH_IN = 1;
    const SEARCH_OUT = 3;
    const SEARCH_HAS_CHILD = 2;

    var currentFilter = null;

    this.research = function(force) {
        var obj = {
            selected_ids: {},
            changed: false
        }

        if(currentFilter || obj.changed || has_filter_options || force) this.search(currentFilter, true, obj);
    }

    var research_on_show = false;

    this.search = function(filter, research, custom_obj) {
        if(!treeView) {
            research_on_show = true;
            return;
        }

        research_on_show = false;

        if(filter === null) filter = '';
        var re = (filter.length > 0) ? new RegExp(escapeRegExp(filter), 'i') : null;

        var force_sel = false;
        var new_filter = filter;

        //try searching for selector
        var selected_ids = {};

        currentFilter = filter;

        var obj;

        if(custom_obj) {
            obj = custom_obj;
        } else {
            var obj = {
                selected_ids: {},
                changed: false
            }
            //page.callFrameworkHandler('on_custom_tree_filter', obj);
        }

        var custom_ids = obj.selected_ids;
        var custom_filter = obj.changed;

        if(!filter && !custom_filter && !has_filter_options) {
            treeData.forEach(function(data) {
                data.search = 0;
                data.force_search = false;
                if(data.peek_in) {
                    data.collapsed = true;
                    data.peek_in = false;
                }
            })
            tableGrid.refresh();
            return;
        }

        if(!helper) return;

        treeData.forEach(function(data) {
            data.search = 0;
            if(data.peek_in) {
                data.collapsed = true;
                data.peek_in = false;
            }
        })

        helper.forEachWithParents(function(data, parents) {

            var inc = false;

            if(!re || data.search_string.match(re)) {
                inc = true;
                if(data.template_type === 'document') {
                    inc = false; //don't matc documents
                }
            }

            if(has_filter_options) {

                var is_el = data.template_type === 'element';

                var status = is_el ? data.pgel.getAttr('status') : null;

                if(filter_options.issues) {
                    if(is_el && status !== 'ok' && status !== 'unused') {
                        inc = inc && true;
                    } else {
                        inc = false;
                    }
                }
                if(filter_options.unused) {
                    if(is_el && status === 'unused') {
                        inc = inc && true;
                    } else {
                        inc = false;
                    }
                }
            }


            if(inc) {
                //selected_ids[data.pgid] = true;
                data.search = SEARCH_IN;

                for(var i = parents.length - 1; i >= 0; i--) {
                    if(parents[i].search === SEARCH_OUT) {
                        parents[i].search = SEARCH_HAS_CHILD;
                        if(parents[i].collapsed) {
                            parents[i].peek_in = true;
                            //data.peeked = true;
                        }
                        parents[i].collapsed = false;
                    } else {
                        break;
                    }
                    //parents[i].collapsed = false;
                }
            } else {
                data.search = SEARCH_OUT;

                for(var i = parents.length - 1; i >= 0; i--) {
                    if(parents[i].template_type === 'element' && parents[i].search === SEARCH_IN) {
                        data.search = SEARCH_IN;
                        break;
                    }
                }
                /*
                if(false && data.force_search && research) {
                    for(var i = parents.length - 1; i >= 0; i--) {
                        if(parents[i].search === SEARCH_OUT) {
                            parents[i].search = SEARCH_HAS_CHILD;
                        } else {
                            break;
                        }
                        //parents[i].collapsed = false;
                    }
                } else {
                    data.search = SEARCH_OUT;
                }
                */
            }
        })

        tableGrid.refresh();
    }

    var unpopulateDataItem = function(data) {
        if(data) {
            data.populated = false;
            data.def = null;
        }
    }

    var updateBranchForPgel = function(pgel, data) {
        if(data === undefined) {
            data = getTreeDataItemForPgId(pgel.getId())
        }
        var list = createDataItemsForPgEl(pgel);
        helper.replaceWith(data, list);
    }

    var updateSelectedItems = function() {
        pinegrow.selectedElements.forEachElement(function(pgel) {
            var data = getTreeDataItemForPgId(pgel.getId());
            if(data) data.selected = true;
        })
    }

    var forceDisplayInSearch = function(list_or_data, val) {
        if(Array.isArray(list_or_data)) {
            list_or_data.forEach(function(data) {
                data.force_search = val;
            })
        } else {
            list_or_data.force_search = val;
        }
    }

    this.scrollToSelected = null;

    this.handleEvent = function(event) {

    }

    var ignore_scripts = false;
    var force_parent = null;

    this.create = function() {

        treeData = [];
        treeData_PgIdMap = {};

        var $ul = treeTop;
        var only_kids = false;

        var current_max_tree_level = max_tree_level;

        max_tree_level = 0;

        var test = false;

        var treeRoot = db.getProjectNode();

        force_parent = null;

        var profile = new CrsaProfile(true);

        var start_level = 0;

        var html = createDataItemForNode(treeData, treeRoot, only_kids);

        profile.show('create ProjectDB ' + treeData.length + ' tree data items');

        helper = new PgTreeDataHelper(treeData);

        helper.reindexData();

        this.research();
    }

    //helper methods
    var getCSSValueHTML = new PgGetCSSValueHTML();

    this.populateDataItem = function(data, page) {
        if(data.populated) return;

        switch(data.pgel.tagName) {
            case 'action_param':
                //if()
                break;
        }

        data.populated = true;
    }

    var actions_map = new Map();

    var getActionInfo = function(id) {
        var a = actions_map.get(id);
        if(a) return a;

        var def = wpf.getComponentType(id);

        if(def) {
            var params = def.getActionParameters();
            var params_map = {}

            params.forEach(function(p) {
                params_map[p.name] = p;
            })

            a = {
                def: def,
                params_map: params_map,
                params: params
            }

            actions_map.set(id, a)
            return a;
        }
        return null;
    }

    var createDataItemForNode = this.createDataItemForNode = function (treeData, node, only_kids, level, element_fata) {
        if (level === undefined) level = -1;

        var cls = 'sort';
        var no_sort = false;
        var type = null;

        var getFata = function() {
            return {pgel: node, shown: true, level: level, tags: '', search_string: ''};
        }
        var fata = getFata();

        if(node.tagName === 'element') {
            fata.template_type = 'element';
            treeData.push(fata);

            element_fata = fata;

            fata.search_string += `_${node.getAttr('selector') || ''}_${node.getAttr('name') || ''}_`

            var fata_sel = getFata();
            fata_sel.level++;
            fata_sel.pgel = {
                tagName: 'selector'
            }
            fata_sel.element_pgel = node;
            fata_sel.template_type = 'selector';
            treeData.push(fata_sel);

            var fata_map = getFata();
            fata_map.level++;
            fata_map.pgel = {
                tagName: 'mapped'
            }
            fata_map.element_pgel = node;
            fata_map.template_type = 'mapped';

            fata.collapsed = isNodeCollapsed(node)

            treeData.push(fata_map);

        } else if(node.tagName !== 'project') {

            switch (node.tagName) {
                case 'project':
                case 'document':
                    cls = 'no-sort';
                    no_sort = true;
                    type = 'label';
                    fata.tags = 'tag-block';
                    fata.search_string +=  `_${node.getAttr('url') || ''}_`;
                    break;
                case 'action_param':

                    element_fata.search_string += `_${node.value || ''}_`
                    break;
                case 'action':
                    element_fata.search_string += `_${node.getAttr('name') || ''}_`;
                    break;
            }

            type = node.tagName;

            if(type === 'document') {
                var document_elements = node.findOne('>elements').children;
                if(document_elements.length === 0) {
                    return;
                }
            }

            var force_collapsed = false;
            var focused = false;

            fata.classes = cls;
            fata.collapsed = isNodeCollapsed(node) || force_collapsed;
            fata.no_sort = no_sort;
            fata.focused = focused;
            fata.template_type = type;

            treeData.push(fata);
        }

        if(node.tagName === 'action') {
            var ai = getActionInfo(node.getAttr('id'))
            if(ai) {
                fata.ai = ai;

                var list = []
                node.getAttrList().forEach(function(attr) {
                    if(attr.name !== 'id') {
                        var param = ai.params_map[attr.name] || null;
                        if(param) {
                            var not_set = '______not_set___';
                            var value = node.getAttribute(attr.name, not_set)
                            if(value !== not_set) {
                                list.push({
                                    tagName: 'action_param',
                                    name: param.def.name || attr.name,
                                    value: value,
                                    action_pgel: node
                                })

                            }
                        }
                    }
                })
                list.forEach(function (ch) {
                    createDataItemForNode(treeData, ch, false, level + 1, element_fata);
                })
            }
        } else if(node.children) {
            var list = node.children;

            if (node.tagName === 'project') {
                list = node.findOne('>documents').children;
            } else if (node.tagName === 'document') {
                list = document_elements;
            }

            list.forEach(function (ch) {
                if (ch.isElement) {
                    createDataItemForNode(treeData, ch, false, level + 1, element_fata);
                }
            })
        }
    }


    var isDescendantOf = function(node, parent) {
        while(node) {
            if(node.parent === parent) return true;
            node = node.parent || null;
        }
        return false;
    }

    var isNodeCollapsed = function(node) {
        return collapsed_map.isCollapsed(node);
    }

}