
var PgSnapView = function(type, icon) {

    var $element = $('<div class="pg-snap"></div>');
    var view = new PgUIView($element, 'snap', 'Snap');

    this.view = view;

    view.tabIcon = null;

    var snaps = []
    var recent_urls = []
    var options = {}

    var $panel = $(`<div class="pg-snap-controls"></div>`).appendTo($element);

    //<span>${pgShowKbd('CMD')} ${pgShowKbd('S')}</span>

    var $snap_button = $(`<button class="pg-button pg-button-primary pg-take-snap">Take a snapshot</button>`).appendTo($panel);

    var $open_button = $(`<button class="pg-button">Open url...</button>`).appendTo($panel);

    var $options_button = $(`<button class="pg-button pg-options-button">Options...</button>`).appendTo($panel);

    var $title = $(`<h3>Snapshots</h3>`).appendTo($element);

    var $snaps = $(`<div class="pg-snap-snaps"></div>`).appendTo($element);

    var project = null;

    var path = require('path');
    var fs = require('fs');

    $options_button.on('click', function(e) {
        e.preventDefault();

        var d = new PgDialog('Snapshot options');
        d.values = $.extend({}, options);
        d.okButtonLabel = 'OK';

        d.before = `<p>How do you like your snapshots?</p>`;

        d.sections = {
            snapoptions: {
                name: 'Options',
                is_closable: false,
                fields: {
                    formathtml: {
                        type: 'checkbox',
                        name: 'Format HTML',
                        value: 'true',
                        helptext: `<p>Format the HTML code to make it more readable. In some cases this can introduce layout errors because of extra spacings.</p>`
                    },
                    removejs: {
                        type: 'checkbox',
                        name: 'Remove all Javascript code',
                        value: 'true',
                        helptext: `<p>Removing scripts ensures that the HTML snapshot is trully a static representation of the current page state. But any interactive elements that are powered by Javascript code will not work.</p>`
                    },
                    addcustomcss: {
                        type: 'checkbox',
                        name: 'Add custom.css',
                        value: 'true',
                        helptext: `<p>Link custom.css stylesheet to the snapshot. Useful for overriding stylesheet styles.</p>`
                    }
                }
            },
            appendcode: {
                name: 'Add code',
                fields: {
                    addtohead: {
                        type: 'text',
                        textarea: true,
                        name: 'Append this HTML code to the &lt;head&gt; element:',
                        validate: function (pgel, prop, value, fieldDef, $field, values) {
                            if(value) {
                                var p = new pgParser();
                                p.assignIds = null;
                                p.parse(value);
                                if(p.rootNode.validateTree().length) {
                                    return 'Should be a valid HTML. Are tags closed?'
                                }
                            }
                        },
                        helptext: `<p>Add this code to the HEAD element. Useful for linking development stylesheets. Should be a valid HTML code.</p>`
                    }
                }
            },
            snapmaps: {
                name: 'URL Mapping',
                is_closable: true,
                fields: {

                }
            }
        }

        var max_maps = 5;
        var has_maps = false;

        for(var i = 1; i <= max_maps; i++) {
            (function(i) {
                d.sections.snapmaps.fields[`map${i}_label`] = {
                    type : 'label',
                    name : 'Map ' + i,
                    action : "element_attribute",
                    class : 'pg-action-info-label'
                }
                d.sections.snapmaps.fields[`map${i}_search`] = {
                    type: 'text',
                    name: `If URL contains`,
                    helptext: `<p>Enter the string to search for or regular expression in the form /.../.</p>`,
                    validate: function (pgel, prop, value, fieldDef, $field, values) {
                        if(values[`map${i}_re`] && value) {
                            if(!value.endsWith('/') || !value.startsWith('/') || value.length < 3) {
                                return 'Regular expression should start and end with /, without flags (ig is assumed).';
                            }
                        }
                    }
                }
                d.sections.snapmaps.fields[`map${i}_replace`] = {
                    type: 'text',
                    name: `Replace the URL with`,
                    helptext: `<p>Enter the new URL or the regular expression replace pattern.</p>`
                }
                d.sections.snapmaps.fields[`map${i}_re`] = {
                    type: 'checkbox',
                    name: 'Use regular expression',
                    value: 'true',
                    helptext: '<p>If checked, Search should be a regular expression in the form /.../ and Replace should be a regular expression replace pattern.</p>'
                }
                has_maps = has_maps || (d.values[`map${i}_search`] || d.values[`map${i}_replace`]);
            })(i);
        }

        if(!has_maps) {
            d.sections.snapmaps.closed = true;
            d.sections.appendcode.closed = true;
        }

        d.onOK = function () {
            var pi = pinegrow.getCurrentProjectInfo();
            pi.setSetting('snap-options', d.values);
            pi.save();
            options = d.values;
        }

        var $d = d.show();
        $d.find('.pg-ui-section:first-child h2').hide();
        $d.find('.pg-ui-section:first-child').css({
            paddingTop: 0,
            marginTop: 0
        })

    })

    $open_button.on('click', function(e) {
        e.preventDefault();

        pinegrow.showEnterUrlPrompt(function(url) {
            if(url.startsWith('file://')) {
                pinegrow.showQuickError('file:// urls are not supported. Please open the page from the server.')
            } else {
                addRecent(url);
                pinegrow.openPage(url, function (page) {

                }, true)
                //recentFiles.add(url);
            }
        }, null, recent_urls)
    })

    var takeSnap = function(name, url, replace_idx) {
        var page = pinegrow.getSelectedPage();
        page.activeView.getHTMLSnapshot(function(html, dom_html) {
            var file = path.join(project.getDir(), name + '.html');
            fs.writeFileSync(file, html, {encoding: 'utf8'});

            var originals_dir = path.join(project.getDir(), 'originals');
            crsaCreateDirs(originals_dir, fs);

            var file_original = path.join(originals_dir, name + '.html');
            fs.writeFileSync(file_original, html, {encoding: 'utf8'});

            //file = path.join(project.getDir(), name + '_dom.html');
            //fs.writeFileSync(file, dom_html, {encoding: 'utf8'});

            var thumb_dir = path.join(project.getDir(), 'thumbnails');
            crsaCreateDirs(thumb_dir, fs);
            var thumb = path.join(thumb_dir, name + '.png');

            if(options.addcustomcss) {
                var custom_css = path.join(project.getDir(), 'custom.css');
                if(!crsaIsFileExist(custom_css)) {
                    fs.writeFileSync(custom_css, ``, {encoding: 'utf8'});
                }
            }

            pinegrow.takePhotoOfPage(page, thumb, function() {
                pinegrow.showQuickMessage('Snapshot was saved!');

                if(replace_idx === undefined) {
                    snaps.push({
                        name: name,
                        url: url,
                        version: 0
                    })
                } else {
                    snaps[replace_idx].url = url;
                    snaps[replace_idx].version++;
                }
                saveSnaps();
                updateSnaps();

                pinegrow.thumbnailMaker.setProjectThumbnail(project, thumb);
                pinegrow.dispatchEvent('on_thumbnail_generated', null, project.getDir());
            }, 400)
        }, options)
    }

    $snap_button.on('click', function(e) {
        e.preventDefault();

        var page = pinegrow.getSelectedPage();
        if(!page) {
            pinegrow.showQuickError('Please open a url first.');
        } else {
            pinegrow.showPrompt(`<p>The snapshot will be saved as {name}.html.</p>`, 'Give the new snapshot a name', '', 'dashboard', null, function (name) {
                if (!name) {
                    pinegrow.showQuickError('The name is required.');
                } else {
                    if(name.endsWith('.html')) {
                        name = name.replace(/\.html$/, '');
                    }

                    var name_lower = name.toLowerCase();
                    var idx = null;

                    for(var i = 0; i < snaps.length; i++) {
                        if(snaps[i].name.toLowerCase() === name_lower) {
                            idx = i;
                            break;
                        }
                    }

                    var file_exists = false;
                    var file = path.join(project.getDir(), name + '.html');
                    file_exists = crsaIsFileExist(file);

                    if(idx === null && !file_exists) {
                        takeSnap(name, page.activeView.getViewUrl());
                        addRecent(page.activeView.getViewUrl());
                    } else {
                        var msg = (idx === null && file_exists) ? `File <b>${name}.html</b> already exists in the project folder. Do you want to overwrite it? ` : `<p>Snapshot <b>${name}.html</b> already exists. Do you want to overwrite it?</p>`;

                        pinegrow.showAlert(msg, 'Overwrite ' + name + '?', 'Cancel', 'Yes, overwrite it', null, function() {
                            if(idx !== null) {
                                takeSnap(name, page.activeView.getViewUrl(), idx);
                            } else {
                                takeSnap(name, page.activeView.getViewUrl());
                            }
                            addRecent(page.activeView.getViewUrl());
                        })
                    }
                }
            })
        }
    })

    var updateSnaps = function() {
        $snaps.empty();

        var html = '';

        var i = 0;
        snaps.forEach(function(snap) {
            var img_src = crsaMakeUrlFromFile(path.join(project.getDir(), 'thumbnails', snap.name + '.png?v=' + snap.version));
            html += `<div class="pg-snap-snaps-snap" data-idx="${i}"><img src="${img_src}" /><p>${snap.name}.html</p></div>`;
            i++;
        })
        $snaps.html(html);
    }

    var getSnapFromEvent = function(e) {
        return snaps[parseInt(e.currentTarget.getAttribute('data-idx'))];
    }

    var getSnapUrl = function(snap) {
        return project.getUrl() + '/' + snap.name + '.html';
    }

    $snaps.on('aaaclick', '.pg-snap-snaps-snap', function(e) {
        e.preventDefault();
        var snap = getSnapFromEvent(e);

        pinegrow.openExternalUrl(getSnapUrl(snap));
    })

    $snaps.on('click contextmenu', '.pg-snap-snaps-snap', function(e) {
        e.preventDefault();
        var snap = getSnapFromEvent(e);

        var idx = parseInt(e.currentTarget.getAttribute('data-idx'));

        var page = pinegrow.getSelectedPage();
        var page_url = page ? page.activeView.getViewUrl() : null;


        var actions = [];

        if(page) {
            actions.push({
                label: 'Update the snapshot',
                action: function() {
                    takeSnap(snap.name, page_url, idx);
                }
            })
            actions.push({
                type: 'divider'
            })
        }



        actions = actions.concat([
            {
                label: 'Edit the snapshot in Pinegrow Web Editor',
                action: function() {
                    pinegrow.showAlert(`<p>Sorry, the feature of automatically opening the HTML snapshot in Pinegrow Web Editor is not implemented yet.</p><p>Instead, <b>run Pinegrow Web Editor and open your whole project  or just this snapshot there</b>.</p><p>You can edit both HTML and CSS in Pinegrow Web Editor.</p>`, 'Open the snapshot in Pinegrow Web Editor')
                    /*
                    var file = path.join(project.getDir(), snap.name + '.html');
                    var cp = require("child_process");
                    cp.spawn(`/Applications/Pinegrow.app/Contents/MacOS/nwjs`, [file]);*/
                }
            },
            {
                type: 'divider'
            },
            {
                label: 'Open URL ' + crsaGetSummaryStr(snap.url || '', 45),
                action: function() {
                    pinegrow.openOrShowPage(snap.url, null, true);
                    addRecent(snap.url);
                }
            },
            {
                label: 'Open the snapshot in external browser',
                action: function() {
                    pinegrow.openExternalUrl(getSnapUrl(snap));
                }
            },
            {
                type: 'divider'
            },
            {
                label: 'Delete...',
                action: function() {

                    pinegrow.showAlert(`<p>Are you sure you want to delete this amazing snapshot?</p>`, 'Are you sure?', 'No, I want to keep it', 'Yes, throw it away', null, function() {

                        try {
                            var file = path.join(project.getDir(), snap.name + '.html');
                            fs.unlinkSync(file);

                            var originals_dir = path.join(project.getDir(), 'originals');
                            var file_original = path.join(originals_dir, snap.name + '.html');
                            fs.unlinkSync(file_original);

                            var thumb_dir = path.join(project.getDir(), 'thumbnails');
                            var thumb = path.join(thumb_dir, snap.name + '.png');
                            fs.unlinkSync(thumb);
                        }
                        catch(err) {
                            pinegrow.showAlert(`<p>Could not delete snapshot or its thumbnail.</p><p>${err}</p>`, 'Files are not deleted');
                        }
                        snaps.splice(idx, 1);
                        saveSnaps();
                        updateSnaps();

                        pinegrow.showQuickMessage(`${snap.name} deleted.`);
                    })
                }
            },
            {
                type: 'divider'
            },
            {
                label: 'Compare...',
                action: function() {
                    var file = path.join(project.getDir(), snap.name + '.html');
                    var originals_dir = path.join(project.getDir(), 'originals');
                    var file_original = path.join(originals_dir, snap.name + '.html');

                    var html = fs.readFileSync(file, {encoding: 'utf8'});
                    var html_original = fs.readFileSync(file_original, {encoding: 'utf8'});

                    console.log(pgFindChangedNodesInPage(html_original, html));
                }
            }
        ])

        var menu = new CrsaContextMenu(actions, $(e.target));
        menu.showAt(e.pageX, e.pageY);
        return false;
    })

    var saveSnaps = function() {
        var pi = pinegrow.getCurrentProjectInfo();
        pi.setSetting('snap-snaps', snaps);
        pi.save();
    }

    var addRecent = function(url) {
        var idx = recent_urls.indexOf(url);
        if(idx >= 0) {
            recent_urls.splice(idx, 1);
        }
        recent_urls.unshift(url);
        if(recent_urls.length > 10) {
            recent_urls.splice(10, recent_urls.length - 10);
        }
        saveRecent();
    }

    var saveRecent = function() {
        var pi = pinegrow.getCurrentProjectInfo();
        pi.setSetting('snap-recent', recent_urls);
        pi.save();
    }

    pinegrow.addEventHandler('on_project_loaded', function(pagenull, p) {
        project = p;

        var pi = pinegrow.getCurrentProjectInfo();
        var values = pi.getSetting('snap-options');

        if(!values) {
            options = {
                removejs: 'true',
                addcustomcss: 'true',
                formathtml: 'true'
            }
        } else {
            options = $.extend({}, values);
        }

        snaps = pi.getSetting('snap-snaps') || [];

        recent_urls = pi.getSetting('snap-recent') || [];
        updateSnaps();
    });

    pinegrow.addEventHandler('on_key_pressed', function(page, e, status) {
        if(status.key == 's' && status.ctrl) {
            $snap_button.trigger('click');
            status.done = true;
        }
    })



    view.onAddedToLayout = function(c) {
    }

    view.onResize = function() {

    }

    view.onDestroy = function() {
    }
}
