
var PgTableGrid = function(options) {

    var _this = this;

    options = $.extend({}, {
        highlightLevels: true
    }, options);

    var cells = [];
    var cells_to_data_map = {};

    var default_type = 'default';

    var zoom = 1;

    this.data = [];
    this.visibleData = [];

    this.delayedTasks = new PgDelayedTasks();

    this.draggable = true;

    var width = 0;
    var height = 0;
    var top, bottom;
    var contentWidth = 0;
    var maxVisibleLevel = 0;

    this.cellHeight = 26;
    this.variableHeight = false;
    this.levelPadding = 18;
    this.inlineRowGap = 4;

    this.$element = $('<div class="pg-table-grid"></div>');

    this.$container = $('<div class="pg-table-grid-container"></div>').appendTo(this.$element);
    //this.template = '<div data-element="name"></div>';

    var template_types = {
    }

    this.setTemplate = function(temp, type) {
        template_types[type || default_type] = temp;
    }

    var getTemplateForData = function(data) {
        return template_types[data.template_type || default_type];
    }


    var $border = $('<div class="pg-table-grid-item-border"></div>');
    var $levelBorder = null;
    if(options.highlightLevels) {
        $levelBorder = $('<div class="pg-table-grid-level-border"></div>').appendTo(this.$container);
    }

    var emptyLineAt = -1;

    var helper;

    var borderedItem = null;
    var borderedItemRows = 0;

    this.onRepaint = null;

    this.getContentWidth = function() {
        return contentWidth;
    }

    this.getVScrollBarSize = function() {
        var div = this.$element.get(0);
        return div.scrollHeight > div.clientHeight ? 10 : 0;
    }

    this.getHScrollBarSize = function() {
        var div = this.$element.get(0);
        return div.scrollWidth > div.clientWidth ? 10 : 0;
    }

    this.highlightChildren = function(data) {
        if(!$levelBorder) return;
        if(!data) {
            $levelBorder.css({
                left: '-8px',
                top: '0px',
                height: '0px'
            });
            return;
        }

        var level = data.level + 1;

        $levelBorder.css({
            left: ((level) * this.levelPadding) + 'px',
            top: this.getYForVisibleDataIndex(data.visible_idx) + this.cellHeight + 'px',
            height: (this.getVisualHeightOfData(data) -1 ) * this.cellHeight + 'px'
        });
    }

    this.highlightLevel = function(idx, level) {
        if(!$levelBorder) return;
        if(idx < 0 || idx >= this.visibleData.length) {
            $levelBorder.css({
                left: '-8px',
                top: '0px',
                height: '0px'
            });
            return;
        }

        var data = this.visibleData[idx];

        if(level >= data.level) level = data.level - 1;

        var parent = this.getParentWithLevel(idx, Math.max(0,  level));

        if(!parent) {
            return this.highlightLevel(-1);
        }

        $levelBorder.css({
            left: ((parent.level + 1) * this.levelPadding) + 'px',
            top: this.getYForVisibleDataIndex(parent.visible_idx) + this.cellHeight + 'px',
            height: (this.getVisualHeightOfData(parent) -1 ) * this.cellHeight + 'px'
        });
    }

    this.borderItem = function(data, add_empty, force_rows) {
        add_empty = add_empty || false;
        if(borderedItem && !data) {
            $border.detach();
        } else if(data && !borderedItem) {
            $border.appendTo(this.$container);
        }

        borderedItem = data;
        if(data && 'visible_idx' in data) {
            if(force_rows !== undefined) {
                borderedItemRows = force_rows;
            } else {
                borderedItemRows = 1 + helper.getNumberOfDescendants(data.visible_idx, this.visibleData);

                if (add_empty) {
                    if (emptyLineAt >= data.visible_idx && emptyLineAt <= data.visible_idx + borderedItemRows) {
                        borderedItemRows++;
                    }
                } else {
                    if (emptyLineAt > data.visible_idx && emptyLineAt < data.visible_idx + borderedItemRows) {
                        borderedItemRows++;
                    }
                }
            }
            this.updateBorder();
        }
        return $border;
    }

    this.updateBorder = function() {
        if(!borderedItem) return;

        var b_top = this.getYForVisibleDataIndex(borderedItem.visible_idx);
        var b_bottom = b_top + borderedItemRows * this.cellHeight;
        var b_left = borderedItem.level * this.levelPadding;
        var b_right = contentWidth;

        if(b_top < top) b_top = top - 10;
        if(b_bottom > bottom) b_bottom = bottom + 10;

        $border.css({
            left: b_left + 'px',
            top: b_top + 'px',
            width: (b_right - b_left) + 'px',
            height: (b_bottom - b_top + 1) + 'px'
        })
    }


    //quick insert line
    var QuickInsertLine = function(grid) {

        var $element = $('<div class="pg-tree-quick-insert-line"></div>');
        $element.appendTo(grid.$container);
        pg$Hide($element);

        var h = 1;
        var data, pos, level;

        this.show = function(adata, alevel, apos) {
            if(data != adata) {
                data = adata;
                level = alevel;
                pos = apos;

                var _this = this;

                fastdom.mutate(function() {
                    if (!data) {
                        pg$Hide($element);
                    } else {
                        pg$Show($element);
                    }

                    _this.update();
                })
            }
        }

        this.hide = function() {
            this.show(null);
        }

        this.update = function() {
            if(data && data.visible_idx < grid.visibleData.length) {
                var x = level * grid.levelPadding;
                var y = grid.getYForVisibleDataIndex(data.visible_idx);
                var h = 1;
                var b_right = grid.getContentWidth();
                var w = (b_right - x);

                if (pos === 'bottom') {
                    y += grid.cellHeight;
                } else if(pos === 'right') {
                    //y = y;
                    w = 2;
                    h = grid.getCellHeightForData(data);
                    x = grid.getXForVisibleData(data);
                    x = x + grid.getCellWidthForData(data);
                }

                $element.css({
                    left: x + 'px',
                    top: y + 'px',
                    width: w + 'px',
                    height: h + 'px'
                })
            }
        }

    }

    this.quickInsertLine = new QuickInsertLine(this);

    var currentScrollOffset = null;

    this.getScrollOffset = function() {
        if(currentScrollOffset === null) {
            currentScrollOffset = {
                top: this.$element.scrollTop(),
                left: this.$element.scrollLeft()
            }
        }
        return currentScrollOffset;
    }

    this.updateDataIndexes = function(from) {

        var hide_level = 0;
        this.visibleData = [];
        var visible_idx = 0;
        var max_level = 0;

        if(from === undefined) from = 0;
        for(var i = from; i < this.data.length; i++) {
            this.data[i].idx = i;

            var hidden = false;

            if(hide_level > 0) {
                if(this.data[i].level >= hide_level) {
                    hidden = true;
                } else {
                    hide_level = 0;
                }
            }
            if(hide_level === 0) {
                if (this.data[i].collapsed) {
                    hide_level = this.data[i].level + 1;
                }
            }

            var search = this.data[i].search || 1;

            if(search === 3) {
                hidden = true;
            }

            if(!hidden) {
                this.visibleData.push(this.data[i]);
                this.data[i].visible_idx = visible_idx++;
                if(this.data[i].level > max_level) max_level = this.data[i].level;
            }
            this.data[i].hidden = hidden;
        }
        maxVisibleLevel = max_level;
    }

    this.onSetDataToCell = function(data, cell) {}

    var scrollTimer = null;

    this.$element.on('scroll', function(e) {
        top = _this.$element.scrollTop();
        currentScrollOffset = null;
        _this.showCells();

        if(scrollTimer) {
            clearTimeout(scrollTimer);
        } else {
            _this.inScroll = true;
        }
        scrollTimer = setTimeout(function() {
            _this.inScroll = false;
            scrollTimer = null;
        }, 500);

        if(_this.onScroll) _this.onScroll();
    })

    this.getHelper = function() {
        return helper;
    }

    this.setData = function(data) {
        cells.forEach(function(cell) {
            cell.setData(null);
        });
        this.data = data;

        helper = new PgTreeDataHelper(this.data);

        this.updateDataIndexes();

        if(this.variableHeight) {
            this.calculatePositionsForVisibleCells();
        }

        if(this.onSetData) {
            this.onSetData(this)
        }

        this.borderItem(null);
        this.show();

        this.highlightLevel(-1);

        if(this.onRepaint) this.onRepaint();
    }

    this.getContainerHeight = function() {
        if(this.variableHeight) {
            let h = 0;
            if(this.visibleData.length > 0) {
                h = this.getYForVisibleDataIndex(this.visibleData.length - 1) + this.getCellHeightForData(this.visibleData[this.visibleData.length - 1]);
            }
            /*
            for(let i = 0; i < this.visibleData.length; i++) {
                h += this.getCellHeightForData(this.visibleData[i]);
            }*/
            if(emptyLineAt >= 0) {
                h += this.cellHeight;
            }
            return h;
        } else {
            return (this.visibleData.length + (emptyLineAt >= 0 ? 1 : 0)) * this.cellHeight;
        }
    }
    this.updateContainerSize = function() {
        this.$container.css({
            height: this.getContainerHeight() + 'px',
            width: contentWidth + 'px'
        })
    }

    this.refresh = function() {
        var so = this.getScrollOffset();
        var origBorderedItem = borderedItem;
        this.setData(this.data);
        this.borderItem(origBorderedItem);

        this.$element.scrollTop(so.top);
        this.$element.scrollLeft(so.left);

        if(so.top > 0) {
            this.showCells();
        }
    }

    this.show = function() {
        this.updateLayout();
        this.updateContainerSize();
        this.showCells();
    }

    this.updateLayout = function() {
        width = this.$element.innerWidth();
        height = this.$element.innerHeight();

        const contentHeight = this.getContainerHeight();

        contentWidth = Math.max(width - (height < contentHeight ? 10 : 0 ), maxVisibleLevel * this.levelPadding + width);

        //contentWidth = Math.max(width - ((this.visibleData.length > height / this.cellHeight) ? 10 : 0 ), maxVisibleLevel * this.levelPadding + width);

        top = this.$element.scrollTop();

        /*
        Will create cells on show

        var num_cells = Math.ceil(height / this.cellHeight / zoom) + 2;

        while(num_cells < cells.length) {
            var cell = cells.pop();
            cell.destroy();
        }

        while(cells.length < num_cells) {
            var cell = new PgTableGridCell(this.template);
            cells.push(cell);
        }
        */
    }

    this.getCellForData = function(data) {
        for(var i = 0; i < cells.length; i++) {
            if(cells[i].data === data) return cells[i];
        }
        return null;
    }

    this.createCellForData = function(data) {
        var cell;
        var type = data.template_type || default_type;
        if(typeof template_types[type] === 'function') {
            cell = new (template_types[type])(type, !this.draggable);
        } else {
            cell = new PgTableGridCell(template_types[type], type, !this.draggable);
        }
        cell.tableGrid = this;
        return cell;
    }

    this.forEachCellOfDataAndDescendants = function(data, func) {
        var num = helper.getNumberOfDescendants(data.visible_idx, this.visibleData) + data.visible_idx + 1;
        for(var i = data.visible_idx; i < num; i++) {
            var cell = this.getCellForData(this.visibleData[i]);
            if(cell) {
                func(cell);
            }
        }
    }

    this.forEachDataAndDescendants = function(data, func) {
        var num = helper.getNumberOfDescendants(data.idx, this.data) + data.idx + 1;
        for(var i = data.idx; i < num; i++) {
            func(this.data[i])
        }
    }

    this.forEachVisibleChild = function(data, func) {
        for(var i = data.visible_idx + 1; i < this.visibleData.length; i++) {
            if(this.visibleData[i].level <= data.level) break;
            if(this.visibleData[i].level === data.level + 1 ) {
                func(this.visibleData[i]);
            }
        }
    }

    this.hasChildren = function(data) {
        return data.idx + 1 < this.data.length && this.data[data.idx + 1].level === data.level + 1;
    }

    this.forEachVisibleSibling = function(data, func) {
        var parent = this.getVisibleParent(data);
        if(parent) {
            this.forEachVisibleChild(parent, func);
        } else {
            for(var i = 0; i < this.visibleData.length; i++) {
                if(this.visibleData[i].level === 0 && this.visibleData[i] !== data) {
                    func(this.visibleData[i]);
                }
            }
        }
    }

    var getXForData = this.getXForVisibleData = function(data) {
        if(_this.variableHeight) {
            return data.pos_x;
        } else {
            return data.level * _this.levelPadding;
        }
    }

    this.showCells = function() {
        //top = this.$element.scrollTop();
        bottom = top + height / zoom;

        var y = top;
        var cell_idx = 0;

        //reposition existing cells
        var unused_cells = {};
        var data_done = {};

        //console.log('showCells', cells.length);

        cells.forEach(function(cell) {
            if(cell.data && !cell.data.hidden && _this.isDataIndexVisible(cell.visible_data_idx)) {
                cell.showAt(getXForData(cell.data), _this.getYForVisibleDataIndex(cell.visible_data_idx), _this.$container, cell.data.cellWidth || 0, cell.data.cellHeight || 0);
                data_done[cell.visible_data_idx] = true;
            } else {
                if(!unused_cells[cell.type]) {
                    unused_cells[cell.type] = [];
                }
                //console.log('add unused for ' + cell.type);
                unused_cells[cell.type].push(cell);
            }
        })

        const idx_from = this.getVisibleDataIndexForY(Math.max(top - this.cellHeight, 0));
        let idx_to = Math.min(this.getVisibleDataIndexForY(bottom), this.visibleData.length - 1);

        //console.log('showCells', cells.length, idx_from, idx_to);

        if((idx_to >= 0 || idx_from >= 0) && (idx_to < idx_from || idx_to < 0)) idx_to = this.visibleData.length - 1;

        for(let visible_data_idx = idx_from; visible_data_idx <= idx_to; visible_data_idx++) {
        //while(y <= bottom) {

            //var visible_data_idx = this.getVisibleDataIndexForY(y);

            if(visible_data_idx >= 0) {

                if (visible_data_idx in data_done) {
                    //nothing to do, already repositioned above
                } else {
                    var data = this.visibleData[visible_data_idx];
                    var type = data.template_type || default_type;

                    var cell;

                    if(unused_cells[type] && unused_cells[type].length) {
                        cell = unused_cells[type].pop();
                        //console.log('using unused ', cell.type)
                    } else {
                        cell = this.createCellForData(data);
                        //cell = new PgTableGridCell(template_types[type], type, !this.draggable);
                        cells.push(cell);

                        //console.log('making new ', type);
                    }

                    var data_y = this.getYForVisibleDataIndex(visible_data_idx);

                    (function(data, cell) {
                        cell.getFastDomTask().mutate(function() {
                            _this.onSetDataToCell(data, cell);
                        })
                    })(data, cell);

                    cell.$element.attr('data-visible-idx', visible_data_idx);
                    cell.setData(data);
                    cell.visible_data_idx = visible_data_idx;
                    cell.showAt(getXForData(data), data_y, this.$container, cell.data.cellWidth || 0, cell.data.cellHeight || 0);
                }
                y = data_y + this.getCellHeightForData(data || this.visibleData[visible_data_idx]);
            } else {
                y += this.cellHeight;
            }

        }
        $.each(unused_cells, function(type, list) {
            list.forEach(function(cell) {
                cell.hide();
                cell.onHidden && cell.onHidden(cell);
            });
        })

        this.updateBorder();
        this.quickInsertLine.update();
    }

    this.getCellWidthForData = function() { return 0 };

    this.getCellHeightForData = function() { return this.cellHeight;}

    this.cellSizeForDataChanged = function(data) {
        var idx = this.visibleData.indexOf(data);
        if(idx >= 0) {
            for(var i = idx; i < this.visibleData.length; i++) {
                this.visibleData[i].has_pos = false;
            }
            this.calculatePositionsForVisibleCells(idx);
        }
    }

    this.calculatePositionsForVisibleCells = function(from_idx, to_idx) {
        var p = new CrsaProfile('calculatePositionsForVisibleCells');
        if(from_idx === undefined) from_idx = 0;
        if(to_idx === undefined) to_idx = this.visibleData.length - 1;

        let cy = 0;
        let cx = 0;
        let last_h = 0;
        let prev_inline = false;
        const avail_w = width - 20;

        if(from_idx > 0) {
            while(from_idx > 0 && !this.visibleData[from_idx - 1].has_pos) {
                from_idx--;
            }
            const prev_data = this.visibleData[from_idx - 1];
            cy = prev_data.pos_y + this.getCellHeightForData(prev_data);
            cx = prev_data.pos_x + this.getCellWidthForData(prev_data)
        }

        for(let idx = from_idx; idx <= to_idx; idx++) {
            const data = this.visibleData[idx];
            //if(cx === 0) {
            //    cx = data.level * this.levelPadding;
            //}
            const w = this.getCellWidthForData(data);
            const h = this.getCellHeightForData(data);
            if(w === 0 || cx + w > avail_w) {
                //needs to go into a new row
                cy += last_h;
                cx = data.level * this.levelPadding;
                last_h = h;
                data.pos_x = cx;
                data.pos_y = cy;
                prev_inline = w !== 0;
                cx += w + this.inlineRowGap;
            } else {
                if(!prev_inline) {
                    cy += last_h;
                    cx = data.level * this.levelPadding;
                }
                data.pos_x = cx;
                data.pos_y = cy;
                cx += w + this.inlineRowGap;
                last_h = h;
                prev_inline = true;
            }
            last_w = w;
            data.has_pos = true;
            /*
            if(idx < 20) {
                console.log(`cell pos ${idx} at ${data.pos_x}, ${data.pos_y}`);
            }

             */
        }
        p.show('done');
    }

    this.getVisibleDataIndexForY = function(y, first) {
        if(this.variableHeight) {
            let last_good_idx = -1;

            for(let idx = 0; idx < this.visibleData.length; idx++) {
                const data = this.visibleData[idx];
                if(data.pos_y <= y && data.pos_y + this.getCellHeightForData(data) >= y) {
                    last_good_idx = idx;
                    if(first) return idx;
                } else if(data.pos_y > y && last_good_idx >= 0) {
                    return last_good_idx;
                }
            }
            return last_good_idx;
        } else {
            var idx = Math.floor(y / this.cellHeight);
            if (emptyLineAt >= 0) {
                if (emptyLineAt === idx) return -2;
                if (emptyLineAt < idx) idx--;
            }
            return idx < this.visibleData.length ? idx : -1;
        }
    }

    this.getYForVisibleDataIndex = function(idx) {
        if(this.variableHeight) {
            return this.visibleData[idx].pos_y;
        } else {
            if (emptyLineAt >= 0 && emptyLineAt <= idx) idx++;
            return idx * this.cellHeight;
        }
    }

    this.getDataFor$Element = function($el) {
        var idx = $el.attr('data-visible-idx');
        return (idx !== null && idx < this.visibleData.length) ? this.visibleData[idx] : null;
    }

    this.getVisibleParent = function(data) {
        var idx = data.visible_idx - 1;
        while(idx >= 0) {
            if(this.visibleData[idx].level < data.level) return this.visibleData[idx];
            idx--;
        }
        return null;
    }

    this.isDataIndexVisible = function(idx, for_scroll) {
        var f = for_scroll ? -1 : 1;
        var y = this.getYForVisibleDataIndex(idx);
        if(this.variableHeight) {
            const top_idx = this.getVisibleDataIndexForY(top - this.cellHeight * f);
            const bottom_idx = this.getVisibleDataIndexForY(bottom + f * this.cellHeight );
            if(top_idx <= idx && (bottom_idx === -1 || bottom_idx >= idx)) {
                return true;
            } else {
                return false;
            }
        } else {
            return y > top - f * this.cellHeight && y < bottom + f * this.cellHeight;
        }
    }

    this.scrollTo = function(data) {
        this.delayedTasks.cancel('scrollToDelayed');

        if(!this.isDataIndexVisible(data.visible_idx, true)) {
            var y = this.getYForVisibleDataIndex(data.visible_idx);
            this.$element.scrollTop(Math.round(y - height*0.2));
        }
        var w = this.$element.width();
        var sl = this.$element.scrollLeft();
        var x = data.level * this.levelPadding;

        var d = Math.floor(w/2);
        var left_space = 3 * this.levelPadding;

        if(sl + w - d < x) {
            this.$element.scrollLeft(x - left_space);
        } else if(x < sl) {
            this.$element.scrollLeft(x - left_space);
        }
    }

    var invalidateCellForData = function(data) {
        for(var i = 0; i < cells.length; i++) {
            if(cells[i].data === data) {
                cells[i].data = null;
                break;
            }
        }
    }

    this.updateItem = function(data) {
        invalidateCellForData(data);
        if(this.isDataIndexVisible(data.visible_idx)) {
            this.showCells();
        }
    }

    this.updateItems = function(datas) {
        var has_visible = false;
        datas.forEach(function(data) {
            invalidateCellForData(data);
            has_visible = has_visible || _this.isDataIndexVisible(data.visible_idx);
        })

        if(has_visible) {
            this.showCells();
        }
    }

    this.refreshDelayed = function() {
        nextFrame.do('tree-refresh', this.refresh);
    }

    this.scrollToDelayed = function(data) {
        stopAutoScroll();
        this.delayedTasks.run('scrollToDelayed', function() {
            _this.scrollTo(data);
        }, 1)
        //nextFrame.do('tree-scrollto', function() {
        //_this.scrollTo(data);
        //});
    }

    this.getNumberOfDescendants = function(idx, list) {
        var c = 0;
        var level = list[idx].level;
        idx++;
        while(idx < list.length && list[idx].level > level) {
            c++;
            idx++;
        }
        return c;
    }

    //get number of data and all its descendants
    this.getVisualHeightOfData = function(data) {
        var c = helper.getNumberOfDescendants(data.visible_idx, this.visibleData);
        var h = c + 1;
        if(emptyLineAt > data.visible_idx && emptyLineAt <= data.visible_idx + c) {
            h++;
        }
        return h;
    }

    this.getParentWithMinLevel = function(idx, min_level) {
        var my_level = this.visibleData[idx].level;
        var max = null;
        while(idx >= 1) {
            idx--;
            if(this.visibleData[idx].level < min_level) break;
            if(!max || max.level > this.visibleData[idx].level) max = this.visibleData[idx];
        }
        return max;
    }

    this.getParentWithLevel = function(idx, level) {
        while(idx >= 1) {
            idx--;
            if(this.visibleData[idx].level === level) return this.visibleData[idx];
        }
        return null;
    }

    var draggedOverData = null;
    var draggedOverDataAtX = 0;
    var draggedData = null;



    this.dragStarted = function(data) {
        draggedData = data;
        this.forEachCellOfDataAndDescendants(data, function(cell) {
            cell.setOpacity(0.5);
        })
    }

    this.dragStopped = function() {
        if(draggedData) {
            this.forEachCellOfDataAndDescendants(draggedData, function (cell) {
                cell.setOpacity(1);
            })
        }
        draggedData = null;
        dragPos = null;
        this.setEmptyLineAt(-1);
        this.showInsertionLineAt(-1);
        this.borderItem(null);
        stopAutoScroll();
    }

    var lastDragInfo = null;
    var lastDragIdx = -1;
    var dragPos = null;

    this.getDragPos = function() {
        return dragPos;
    }

    this.getDragInfo = function(pos) {
        var dd_options = {};//pinegrow.drag_drop_options;

        dd_options.force_parent = null;

        if(false && pinegrow.keyState.shift) {
            debugger;
        }
        var idx = this.getVisibleDataIndexForY(pos.top);

        var data, position;
        var on_empty = false;

        var moveLevel = function() {
            if(!dragPos) return null;

            var level_offset = Math.round((pos.left - draggedOverDataAtX) / _this.levelPadding);

            var more = {level_offset: level_offset};

            if(level_offset < 0) {
                //debugger;
            }

            if(!dd_options.force_parent && canInsertAtIdxAndLevel(draggedData, dragPos.idx, dragPos.orig_level + level_offset, more)) {
                dragPos.level = dragPos.orig_level + level_offset;
            } else {
                if('ok_level' in more) {
                    dragPos.level = more.ok_level;
                }
            }
            return dragPosToInfo(dragPos);
        }

        if(idx == -2 && lastDragInfo) {
            data = lastDragInfo.data;
            position = lastDragInfo.position;
            on_empty = true;
            //idx = lastDragIdx;
        }

        if(idx >= 0) {

            data = this.visibleData[idx];

            if(data.no_sort) return null;

            if(dd_options.force_parent) {
                if(!data.pgel) return null;
                if(!data.pgel.isDescendantOf(dd_options.force_parent)) return null;
                if(!data.pgel.isChildOf(dd_options.force_parent)) {
                    return null;
                    var pdata = data;
                    data = null;
                    while(pdata) {
                        pdata = this.getVisibleParent(pdata);
                        if (pdata.pgel && pdata.pgel.parent === dd_options.force_parent) {
                            data = pdata;
                            idx = data.visible_idx;
                            break;
                        }
                    }
                    if(!data) return null;
                }
            }

            var top = this.getYForVisibleDataIndex(idx);
            var bottom = top + this.lineHeight;

            if(data !== draggedData && top + this.getCellHeightForData(data) / 2 < pos.top) {

                if(this.visibleData.length === idx + 1) {
                    //data is the last item in tree
                    if(!dragPos || dragPos.idx !== idx + 1) {
                        //data = this.visibleData[idx];
                        //data = this.getVisibleParent(data);
                        dragPos = {
                            data: data,
                            idx: idx + 1,
                            level: data.level,
                            orig_level: data.level,
                            on_self: false,
                            position: 'bottom'
                        };
                        if(data.pgel) {
                            if(data.pgel.tagName === 'body') {
                                dragPos.position = 'inside-bottom';
                                dragPos.level++;
                                dragPos.orig_level++;
                            }
                        }
                        return dragPos;
                    } else {
                        return moveLevel();
                    }
                } else {
                    idx++;
                    data = this.visibleData[idx];
                }
            }
        } else if(idx === -1) {
            //at the end?
            idx = this.visibleData.length - 1;
            if(!dragPos || dragPos.idx !== idx) {
                data = this.visibleData[idx];
                if(data.pgel && data.pgel.tagName !== 'body') {
                    data = this.getVisibleParent(data);
                }
                dragPos = {
                    data: data,
                    idx: idx + 1,
                    level: data.level + 1,
                    orig_level: data.level + 1,
                    on_self: false,
                    position: 'inside-bottom'
                };
                return dragPos;
            } else {
                return moveLevel();
            }
        }
        if(data) {
            var info;

            if(idx >= 0) {

                if(!dragPos || dragPos.idx !== idx) {
                    var next = (this.visibleData.length > idx) ? this.visibleData[data.visible_idx] : null;
                    var prev = (idx - 1 >= 0) ? this.visibleData[idx - 1] : null;

                    draggedOverDataAtX = pos.left;

                    if(data == draggedData) {
                        //on self, just move levels
                        dragPos = {idx: idx, level: data.level, orig_level: data.level, on_self: true};
                        return dragPosToInfo(dragPos);
                    }  else if(next == draggedData || prev == draggedData) {
                        //dragPos = null;
                        return lastDragInfo;
                    }  else {
                        dragPos = {idx: idx, level: data.level, orig_level: data.level};
                        if (prev && (!next || next.level <= prev.level)) {
                            info = {
                                data: prev,
                                position: 'bottom'
                            }
                            dragPos.level = prev.level;
                            dragPos.orig_level = prev.level;
                        } else {
                            info = {
                                data: data,
                                position: 'top'
                            }
                        }
                        return info;
                    }
                } else {
                    return moveLevel();
                    //return lastDragInfo;
                }
            } else if(idx == -2) {
                return moveLevel();
                //return lastDragInfo;
            } else {
                return null;
            }
        }
    }
    /*

     xxxx
     pppp
     aaaa
     zzzzz
     bbb
     cccc
     oooo
     ddd

     */

    this.canChangeLevel = function(data, level) {
        var od = draggedData;
        draggedData = data;
        var r = canInsertAtIdxAndLevel(data, data.visible_idx, level, {});
        draggedData = od;
        return r;
    }

    var dragPosToInfo = function(dragPos) {
        if(!dragPos) return null;

        var next = getNext(dragPos.idx);
        var prev = (dragPos.idx - 1 >= 0) ? _this.visibleData[dragPos.idx - 1] : null;

        var info = {};

        if(prev && prev.level == dragPos.level) {
            info.data = prev;
            info.position = 'bottom';
        } else if(next && next.level == dragPos.level) {
            info.data = next;
            info.position = 'top'
        } else if(prev && prev.level + 1 == dragPos.level) {
            info.data = prev;
            info.position = 'inside-top';
        } else if(!next || (next && next.level < dragPos.level)) {
            info.data = _this.getParentWithLevel(dragPos.idx, dragPos.level);
            info.position = 'bottom';
        }
        return info.data ? info : null;
    }

    var canInsertAtIdxAndLevel = function(draggedData, before_idx, level, more) {
        var count = 0;//helper.getNumberOfDescendants(draggedData.visible_idx, _this.visibleData);

        var next = getNext(before_idx);
        var prev = (before_idx - 1 >= 0) ? _this.visibleData[before_idx - 1] : null;

        //console.log(level, next, prev);

        if(prev && level == prev.level + 1) {
            return true;
        } else if(prev && level == prev.level) {
            if(next && next.level > level) {
                return false;
            }
            return true;
        } else if(prev && level > prev.level + 1) {
            more.ok_level = prev.level + 1;
            return false;
        } else if(!next) {
            return true;
        } else if(next.level <= level) {
            return true;
        } else {
            more.ok_level = next.level;
            return false;
        }
        //console.log(info);
        return info;
    }

    var getNext = function(before_idx, movedData) {
        if(!movedData) movedData = draggedData;

        var next = (_this.visibleData.length > before_idx) ? _this.visibleData[before_idx] : null;

        if(next && movedData == next) {
            //moving self
            var count = helper.getNumberOfDescendants(movedData.visible_idx, _this.visibleData);

            next = (_this.visibleData.length > before_idx + 1 + count) ? _this.visibleData[before_idx + 1 + count] : null;
        }
        while(next && movedData && movedData.selected && next.level == movedData.level && next.selected) {
            next = getNext(next.visible_idx, next);
        }
        return next;
    }

    this.setEmptyLineAt = function(at) {
        if(emptyLineAt != at) {
            emptyLineAt = at;
            this.showCells();
        }
    }


    var $insertionLine = this.$insertionLine = $('<div class="pg-table-grid-insertion-line"></div>');
    var insertionLineShown = false;
    var insertionLineAt = -1;
    var insertionLineLevel = 0;
    var insertionLineRows = 0;

    this.showInsertionLineAt = function(idx, level, span_rows) {
        if(span_rows === undefined) span_rows = 0;
        if(insertionLineAt !== idx || level !== insertionLineLevel || insertionLineRows !== span_rows) {
            insertionLineAt = idx;
            insertionLineLevel = level;
            insertionLineRows = 1;

            this.updateInsertionLine();
        }
        return $insertionLine;
    }

    this.getYForInsertionLine = function() {
        if(insertionLineAt === 0) {
            return 0;
        } else if(insertionLineAt > 0) {
            var y = this.getYForVisibleDataIndex(insertionLineAt - 1) + this.getCellHeightForData(this.visibleData[insertionLineAt - 1]);
            return y;
        }
        return 0;
    }

    this.updateInsertionLine = function() {
        if(insertionLineAt < 0) {
            if(insertionLineShown) {
                $insertionLine.detach();
                insertionLineShown = false;
            }
        } else {
            if(!insertionLineShown) {
                insertionLineShown = true;
                $insertionLine.appendTo(this.$container);
            }
            $insertionLine.css({
                //top: insertionLineAt * this.cellHeight + 'px',
                top: this.getYForInsertionLine() + 'px',
                left: insertionLineLevel * this.levelPadding + 'px',
                width: (contentWidth - insertionLineLevel * this.levelPadding) + 'px',
                height: (insertionLineRows ? (insertionLineRows * this.cellHeight) : 1) + 'px'
            })
        }
    }

    this.showInsertionPoint = function(info) {
        if(info) {

            if('level_change' in info) {
                //just changing the level of dragged data
                this.showInsertionLineAt(info.dragged_data.visible_idx, info.dragged_data.level + info.level_change, 1);

            } else {
                //find the new parent
                var parent;
                var level;
                var before_idx = -1;
                var line_at = -1;

                if (info.position == 'inside-top' || info.position == 'inside-bottom') {
                    parent = info.data;
                } else {
                    parent = this.getVisibleParent(info.data) || info.data;
                }
                level = parent.level + 1;

                before_idx = getInsertedIdxForInfo(info);

                //this.borderItem(parent);
                if(dragPos.on_self) {
                    this.setEmptyLineAt(-1);
                    this.showInsertionLineAt(dragPos.idx, dragPos.level);
                    this.highlightChildren(parent);
                } else {
                    this.setEmptyLineAt(dragPos.idx);
                    this.showInsertionLineAt(dragPos.idx, dragPos.level);
                    this.highlightChildren(parent);
                }
            }
        } else {
            this.showInsertionLineAt(-1, 0);
            this.borderItem(null);
        }
    }

    var getInsertedIdxForInfo = function(info) {
        var before_idx = -1;
        switch (info.position) {
            case 'inside-top':
                before_idx = info.data.visible_idx + 1;
                //line_
                break;
            case 'inside-bottom':
                before_idx = info.data.visible_idx + 1 + helper.getNumberOfDescendants(info.data.visible_idx, _this.visibleData);
                break;
            case 'left':
            case 'top':
                before_idx = info.data.visible_idx;
                break;
            case 'bottom':
            case 'right':
                before_idx = info.data.visible_idx + 1 + helper.getNumberOfDescendants(info.data.visible_idx, _this.visibleData);
                break;
        }
        return before_idx;
    }

    var autoScrollTimer = null;
    var autoScrollHotZone = 50;
    var autoScrollStartSpeed = 100; //100 px / s
    var autoScrollEndSpeed = 500; //100 px / s
    var autoScrollSpeed = 100; //100 px / s
    var autoScrollAniFrame = null;
    var autoScrollCurrentDirection = null;

    var autoScroll = function(up) {

        var f = up ? -1 : 1;
        if(autoScrollCurrentDirection === f) return;

        var start = null;
        var prev = null;

        autoScrollSpeed = autoScrollStartSpeed;


        autoScrollCurrentDirection = f;

        var step = function(ts) {
            if(!start) {
                start = ts;
                prev = ts;
            } else {
                var so = _this.getScrollOffset();
                autoScrollSpeed = Math.min(autoScrollSpeed * (1.0 + 0.75 * ((ts - prev) / 1000.0)), autoScrollEndSpeed);
                var dy = autoScrollSpeed * (ts - prev) / 1000.0;
                _this.$element.scrollTop(so.top + dy * f);
                prev = ts;
            }
            autoScrollAniFrame = requestAnimationFrame(step);
        }
        autoScrollAniFrame = requestAnimationFrame(step);
    }

    var stopAutoScroll = this.stopAutoScroll = function() {
        if(autoScrollAniFrame) {
            cancelAnimationFrame(autoScrollAniFrame);
            autoScrollAniFrame = null;
            autoScrollCurrentDirection = null;
        }
        if(autoScrollTimer) {
            clearTimeout(autoScrollTimer);
            autoScrollTimer = null;
        }
    }

    this.dragOver = function(pos) {
        var so = this.getScrollOffset();

        //auto scroll
        var delay = 1000;
        if(autoScrollTimer) {
            clearTimeout(autoScrollTimer);
            autoScrollTimer = null;
            stopAutoScroll();
        }

        if(so.top > 0 && pos.top < autoScrollHotZone) {
            autoScrollTimer = setTimeout(function() {
                autoScroll(true);
            }, delay);
        } else if(pos.top > height - autoScrollHotZone) {
            autoScrollTimer = setTimeout(function() {
                autoScroll();
            }, delay);
        } else {
            // stopAutoScroll();
        }


        pos.top += so.top;
        pos.left += so.left;

        var info = this.getDragInfo(pos);

//        console.log(info)

        lastDragInfo = info;

        if(info) {
            return info;
        } else {
            return null;
        }
    }

    this.destroy = function() {
        cells.forEach(function(cell) {
            cell.destory()
        })
        cells = [];
    }
}


var pgTableGridCellCount = 0;

var PgTableGridCell = function(template, type, no_drag) {
    this.type = type;
    this.data = null;
    this.data_idx = 0;
    this.visible_data_idx = 0;
    this.opacity = 1.0;
    this.tableGrid = null;

    this.fastDomTaskId = 'gridCell' + pgTableGridCellCount++;

    this.$element = $(template).addClass('pg-table-grid-cell pg-table-grid-cell-' + type);
    if(!no_drag) {
        this.$element.attr('draggable', 'true');
    }

    this.dataElements = {}

    this.shown = false;

    var _this = this;

    this.$element.find('[data-element]').add(this.$element).each(function(i, e) {
        _this.dataElements[e.getAttribute('data-element')] = e;
    })

    this.eventHandlers = []
    this.tooltips = []

    this.cache = {}
}

PgTableGridCell.prototype.update = function() {
    this.tableGrid.updateItem(this.data);
}

PgTableGridCell.prototype.on = function(key, event, handler) {
    this.dataElements[key].addEventListener(event, handler);
    this.eventHandlers.push({
        key: key,
        event: event,
        handler: handler
    })
}

PgTableGridCell.prototype.tooltip = function(key, tooltip) {
    if(this.tooltips.indexOf(key) < 0) {
        addTooltip($(this.dataElements[key]), tooltip, {
            use_title_attr: 'data-pg-title'
        });
        this.tooltips.push(key);
    }
    this.setAttribute(key, 'data-pg-title', tooltip);
}

PgTableGridCell.prototype.getFastDomTask = function() {
    return pgFastdomTasks.create(this.fastDomTaskId);
}

PgTableGridCell.prototype.setData = function(data) {
    this.data = data;
    this.setOpacity(1.0);
}

PgTableGridCell.prototype.onPopulate = function() {
}

PgTableGridCell.prototype.showAt = function(x, y, $container, w, h) {
    if(!this.shown) {
        this.$element.appendTo($container);
        this.shown = true;
    }
    var el = this.$element.get(0);
    pgPositionAt(el, x, y);

    w && (el.style.width = w + 'px');
    h && (el.style.height = h + 'px');
}

PgTableGridCell.prototype.hide = function() {
    this.setData(null);
    if(this.shown) {
        this.$element.detach();
        this.shown = false;
    }
}

PgTableGridCell.prototype.setText = function(key, text) {
    const ck = key + '__text';
    if(!(ck in this.cache) || this.cache[ck] !== text) {
        this.dataElements[key].textContent = text;
        this.cache[ck] = text;
    }
}

PgTableGridCell.prototype.setHtml = function(key, html) {
    const ck = key + '__html';
    if(!(ck in this.cache) || this.cache[ck] !== html) {
        this.dataElements[key].innerHTML = html;
        this.cache[ck] = html;
    }
}

PgTableGridCell.prototype.addClass = function(key, cls, a, b, c) {
    if(this.dataElements[key]) {
        this.dataElements[key].classList.add(cls, a, b, c);
    }
}

PgTableGridCell.prototype.removeClass = function(key, cls, a, b, c) {
    if(this.dataElements[key]) this.dataElements[key].classList.remove(cls, a, b, c);
}

PgTableGridCell.prototype.setAttribute = function(key, attr, val) {
    const ck = key + '__attr_' + attr;
    if(!(ck in this.cache) || this.cache[ck] !== val) {
        this.dataElements[key].setAttribute(attr, val);
        this.cache[ck] = val;
    }
}

PgTableGridCell.prototype.getElement = function(key) {
    return this.dataElements[key];
}

PgTableGridCell.prototype.showElement = function(key, d) {
    const ck = key + '__vis';
    if(!(ck in this.cache) || this.cache[ck] !== true) {
        pgShow(this.dataElements[key], d || 'block');
        this.cache[ck] = true;
    }
}

PgTableGridCell.prototype.hideElement = function(key) {
    const ck = key + '__vis';
    if(!(ck in this.cache) || this.cache[ck] !== false) {
        pgHide(this.dataElements[key]);
        this.cache[ck] = false;
    }
}

PgTableGridCell.prototype.setOpacity = function(opa) {
    if(this.opacity !== opa) {
        this.$element.css('opacity', opa);
    }
    this.opacity = opa;
}


PgTableGridCell.prototype.destroy = function() {
    const _this = this;
    this.eventHandlers.forEach(function(i) {
        _this.dataElements[i.key].removeEventListener(i.event, i.handler);
    })
    this.eventHandlers = null;

    this.tooltips.forEach(function(key) {
        $(_this.dataElements[key]).tooltip('destroy');
    })
    this.tooltips = null;

    this.data = null;
    this.$element.remove();
    this.$element = null;
    this.dataElements = null;
}


var PgMouseSpeed = function() {

    var lastX = null;
    var lastY = null;
    var lastTs = null;

    this.speedX = 0;
    this.speedY = 0;

    var _this = this;

    var timer;

    this.track = function(event) {
        if(timer) {
            clearTimeout(timer);
        }
        var ts = new Date().getTime();
        if(lastTs !== null) {
            var d = (ts - lastTs) / 1000.0;
            if(d > 0) {
                this.speedX = (event.pageX - lastX) / d;
                this.speedY = (event.pageY - lastY) / d;

                if(d > 1) {
                    lastTs = ts;
                    lastX = event.pageX;
                    lastY = event.pageY;
                }
            }
        } else {
            lastTs = ts;
            lastX = event.pageX;
            lastY = event.pageY;
        }

        timer = setTimeout(function() {
            timer = null;
            lastTs = null;
            _this.speedX = 0;
            _this.speedY = 0;
        }, 200);
    }

    this.getSpeedX = function() {
        var dir = '';
        if(this.speedX > 0) {
            dir = 'right';
        } else if(this.speedX < 0) {
            dir = 'left';
        }
        if(Math.abs(this.speedX) > 20) dir += '-fast';

        console.log(this.speedX, dir);
        return dir;
    }
}

var PgNextFrame = function() {

    var refs = {};

    this.do = function(key, func) {
        if(refs[key]) {
            cancelAnimationFrame(refs[key]);
        }
        refs[key] = requestAnimationFrame(function() {
            delete refs[key];
            func();
        })
    }

    this.cancel = function(key) {
        if(refs[key]) {
            cancelAnimationFrame(refs[key]);
            delete refs[key];
        }
    }

    this.cancelAll = this.destory = function() {
        $.each(refs, function(key, val) {
            cancelAnimationFrame(val);
        })
        refs = {};
    }
}

window.nextFrame = new PgNextFrame();


var PgTreeDataHelper = function(data) {
    this.data = data;
}

PgTreeDataHelper.prototype.removeItem = function(data, func) {
    var num = this.getNumberOfDescendants(data.idx, this.data) + 1;
    if(func) {
        var to = Math.min(this.data.length, data.idx + num);
        for(var i = data.idx; i < to; i++) {
            func(this.data[i]);
        }
    }
    this.data.splice(data.idx, num);
}

PgTreeDataHelper.prototype.getNumberOfDescendants = function(idx, list) {
    if(!list) list = this.data;
    var c = 0;
    var level = list[idx].level;
    idx++;
    while(idx < list.length && list[idx].level > level) {
        c++;
        idx++;
    }
    return c;
}

PgTreeDataHelper.prototype.removeIfAlreadyInData = function(list, insert_idx) {
    var idx = this.data.indexOf(list[0]);
    if(idx >= 0) {
        this.data.splice(idx, list.length);
        if(idx < insert_idx) insert_idx -= list.length;
    }
    return insert_idx;
}

PgTreeDataHelper.prototype.updateLevels = function(list, level) {
    var top = list[0].level;
    list.forEach(function(d) {
        d.level = d.level - top + level;
    })
}

PgTreeDataHelper.prototype.insertAfter = function(dest, list) {
    var insert_idx = this.removeIfAlreadyInData(list, dest.idx);
    this.updateLevels(list, dest.level);
    this.data.splice.apply(this.data, [insert_idx + this.getNumberOfDescendants(insert_idx, this.data) + 1, 0].concat(list));
}

PgTreeDataHelper.prototype.insertBefore = function(dest, list) {
    var insert_idx = this.removeIfAlreadyInData(list, dest.idx);
    this.updateLevels(list, dest.level);
    this.data.splice.apply(this.data, [insert_idx, 0].concat(list));
}

PgTreeDataHelper.prototype.findIndexOfChild = function(parent, idx, list /* el that will be inserted */) {
    var ci = -1;
    for(var i = parent.idx + 1; i < this.data.length; i++) {
        if(list && list[0] === this.data[i]) {
            //this el is already a child, ignore it
            i += list.length - 1;
            continue;
        }
        if(this.data[i].level === parent.level + 1) {
            ci++;
            if(ci === idx) return i;
        } else if(this.data[i].level <= parent.level) return i;
    }
    return this.data.length; /* - 1 */;
}

PgTreeDataHelper.prototype.insertAtIndex = function(dest, list, idx) {
    var insert_idx = this.findIndexOfChild(dest, idx, list);
    insert_idx = this.removeIfAlreadyInData(list, insert_idx);
    this.updateLevels(list, dest.level + 1);
    this.data.splice.apply(this.data, [insert_idx, 0].concat(list));
}

PgTreeDataHelper.prototype.append = function(dest, list) {
    var insert_idx = this.removeIfAlreadyInData(list, dest.idx);
    this.updateLevels(list, dest.level + 1);
    this.data.splice.apply(this.data, [insert_idx + this.getNumberOfDescendants(insert_idx, this.data) + 1, 0].concat(list));
}

PgTreeDataHelper.prototype.prepend = function(dest, list) {
    var insert_idx = this.removeIfAlreadyInData(list, dest.idx);
    this.updateLevels(list, dest.level + 1);
    this.data.splice.apply(this.data, [insert_idx + 1, 0].concat(list));
}

PgTreeDataHelper.prototype.removeDescendants = function(data) {
    this.data.splice(data.idx + 1, this.getNumberOfDescendants(data.idx, this.data));
}

PgTreeDataHelper.prototype.delevel = function(data) {
    var c = this.getNumberOfDescendants(data.idx, this.data);
    for(var i = data.idx + 1; i <= data.idx + c; i++) {
        this.data[i].level--;
    }
    this.data.splice(data.idx, 1);
}

PgTreeDataHelper.prototype.peek = function(data) {
    if(data.hidden === undefined) {
        data.hidden = false;
        this.withAllParents(data, function(p) {
            if (p.collapsed || p.hidden) {
                data.hidden = true;
                return false;
            }
        })
    }
    if(data.hidden) {
        data.peeked = true;

        this.withAllParents(data, function(p) {
            if (p.collapsed) {
                p.collapsed = false;
                p.peek_in = true;
            }
            if (!p.hidden) return false;
        })
        return true;
    } else {
        return false;
    }
}

PgTreeDataHelper.prototype.unpeek = function(data) {
    if(data.peeked) {
        data.peeked = false;

        this.withAllParents(data, function(p) {
            if (p.peek_in) {
                p.peek_in = false;
                p.collapsed = true;
            }
        })
        return true;
    } else {
        return false;
    }
}

PgTreeDataHelper.prototype.withAllParents = function(data, func) {
    var level = data.level - 1;

    for(var i = data.idx - 1; i >= 0; i--) {
        if(this.data[i].level === level) {
            if(func(this.data[i]) === false) break;
            level--;
        }
    }
}

PgTreeDataHelper.prototype.withChildren = function(data, func) {
    for(var i = data.idx + 1; i < this.data.length; i++) {
        if(this.data[i].level <= data.level) break;
        if(func(this.data[i]) === false) break;
    }
}

PgTreeDataHelper.prototype.withDataAndChildren = function(data, func) {
    if(func(data) === false) return;
    this.withChildren(data, func);
}

PgTreeDataHelper.prototype.forEachWithParents = function(func) {
    var parents = [];
    var prev = null;
    for(var i = 0; i < this.data.length; i++) {
        var data = this.data[i];

        if(prev) {
            if(prev.level < data.level) {
                parents.push(prev);
            } else if(prev.level > data.level) {
                parents.splice(parents.length - (prev.level - data.level), prev.level - data.level);
            }
        }
        if(func(data, parents) === false) break;

        prev = data;
    }
}

PgTreeDataHelper.prototype.replaceContentWith = function(data, list) {
    var count = this.getNumberOfDescendants(data.idx);
    var idx = this.removeIfAlreadyInData(list, data.idx + 1);
    this.updateLevels(list, data.level + 1);
    this.data.splice.apply(this.data, [idx, count].concat(list));
}

PgTreeDataHelper.prototype.replaceWith = function(data, list) {
    var count = this.getNumberOfDescendants(data.idx);
    var idx = this.removeIfAlreadyInData(list, data.idx);
    this.updateLevels(list, data.level);
    this.data.splice.apply(this.data, [idx, 1 + count].concat(list));
}

PgTreeDataHelper.prototype.reindexData = function(from, update_hidden) {
    if(from === undefined) from = 0;
    if(!update_hidden) {
        for (var i = from; i < this.data.length; i++) {
            this.data[i].idx = i;
        }
    } else {
        var hide_level = 0;

        for (var i = from; i < this.data.length; i++) {
            this.data[i].idx = i;
            var hidden = false;

            if (hide_level > 0) {
                if (this.data[i].level >= hide_level) {
                    hidden = true;
                } else {
                    hide_level = 0;
                }
            }
            if (hide_level === 0) {
                if (this.data[i].collapsed) {
                    hide_level = this.data[i].level + 1;
                }
            }

            var search = this.data[i].search || 1;

            if (search === 3) {
                hidden = true;
            }
            this.data[i].hidden = hidden;
        }
    }
}

PgTreeDataHelper.prototype.destroy = function() {
    this.data = null;
}




var pgDragHelperObjectId = 0;

//Based on Goldenlayout DragListener
var PgDragHelper = function($container, selector, mydoc, use_dd)
{
    var lm = GoldenLayout.__lm;

    lm.utils.EventEmitter.call(this);

    mydoc = mydoc || document;

    //use_dd = false;

    if(!isApp()) use_dd = false;

    this.use_dd = use_dd;

    this.object_id = pgDragHelperObjectId++;

    this.$container = $container;
    this.selector = selector;
    this._oDocument = $(mydoc);
    this._eBody = $(mydoc.body);
    this._oAnotherDocument = null;

    this.onGetDragImage = null;
    this.onSetDragData = null;

    this.onBefore = null;

    /**
     * The delay after which to start the drag in milliseconds
     */
    this._nDelay = 200;

    /**
     * The distance the mouse needs to be moved to qualify as a drag
     */
    this._nDistance = 10;//TODO - works better with delay only

    this._nX = 0;
    this._nY = 0;

    this._nOriginalX = 0;
    this._nOriginalY = 0;

    this._prevX = -999999;
    this._prevY = -999999;

    this._bDragging = false;

    this._fMove = lm.utils.fnBind( this.onMouseMove, this );
    this._fUp = lm.utils.fnBind( this.onMouseUp, this );
    this._fDown = lm.utils.fnBind( this.onMouseDown, this );
    this._fDragStart = lm.utils.fnBind( this.onDragStart, this );
    this._fDragOver = lm.utils.fnBind( this.onDragOver, this );

    this.movedElement = null;
    this.currentWindow = null;

    if(use_dd) {
        this.$container.on('dragstart', this.selector, this._fDragStart);
    } else {
        if (this.selector) {
            this.$container.on('mousedown touchstart', this.selector, this._fDown);
        } else {
            this.$container.on('mousedown touchstart', this._fDown);
        }
    }
};

//PgDragHelper.timeout = null;

GoldenLayout.__lm.utils.copy( PgDragHelper.prototype, {
    setDocument: function() {
        var el = this.$container.get(0);
        var doc = el.ownerDocument || document;
        this._oDocument = $(doc);
        this._eBody = $(doc.body);
    },
    destroy: function() {
        if(this.use_dd) {
            this.$container.off('dragstart', this.selector, this._fDragStart);
        } else {
            if (this.selector) {
                this.$container.off('mousedown touchstart', this.selector, this._fDown);
            } else {
                this.$container.off('mousedown touchstart', this._fDown);
            }
        }
    },
    onDragStart: function(oEvent)
    {
        this.setDocument();
        this.movedElement = $(oEvent.currentTarget);

        if(this.onBefore && !this.onBefore(this.movedElement)) return false;

        oEvent.stopPropagation();

        var _this = this;

        var coordinates = this._getCoordinates( oEvent );

        this._nOriginalX = coordinates.x;
        this._nOriginalY = coordinates.y;

        //this._oDocument.on('drag', this._fMove);
        //this._oDocument.on('mousemove touchmove', this._fMove);
        this._oDocument.one('dragend', this._fUp);

        this.currentWindow = oEvent.view;

        var frame_id = 0;
        this.currentFrameId = frame_id;
        this.processedFrameId = -1;

        var my_win = this.currentWindow;

        var onFrame = function() {
            _this.currentFrameId = frame_id++;
            my_win.requestAnimationFrame(onFrame);
        }

        my_win.requestAnimationFrame(onFrame);

        pgGetAppWindows().forEach(function(win) {
            if(win.window !== oEvent.view || true) {
                $(win.window).on('dragover', _this._fDragOver);
            }
        })

        this.lastEvent = oEvent;
        this.dragTarget = $(oEvent.currentTarget);

        if(this.onGetDragImage) {
            var img = this.onGetDragImage(this.dragTarget);
            if(img == 'none') {
                img = document.createElement("img");
                //img = _this.dragTarget.get(0).cloneNode(true);
                //img.src = 'https://d33wubrfki0l68.cloudfront.net/ce237d2a79fb3c3781e5833ad45e2367d781d3c0/c5b6c/pg4assets/img/matjaz_1.jpg';
                img.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
                _this._oDocument.get(0).body.appendChild(img);
                _this.dragGhost = img;
            }
            oEvent.originalEvent.dataTransfer.setDragImage(img, 0, 0);
        }

        oEvent.originalEvent.dataTransfer.effectAllowed = "copyMove";
        if(this.onGetDropEffect) {
            oEvent.originalEvent.dataTransfer.dropEffect = this.onGetDropEffect(this.dragTarget);
        }

        setTimeout(function() {
            _this._startDrag(oEvent);
            if(_this.onSetDragData) {
                _this.onSetDragData(oEvent.originalEvent.dataTransfer);
            }
        }, 1);

    },
    onDragOver: function(oEvent)
    {
        pinegrow.keyState.keyChanged(oEvent);
        if(this.processedFrameId === this.currentFrameId) {
            //already handled event in this frame
            return;
        }
        this.processedFrameId = this.currentFrameId;

        this._fMove(oEvent);
    },
    onMouseDown: function(oEvent)
    {
        this.setDocument();
        this.movedElement = $(oEvent.currentTarget);

        if(this.onBefore && !this.onBefore(this.movedElement)) return;

        oEvent.preventDefault();
        oEvent.stopPropagation();

        var coordinates = this._getCoordinates( oEvent );

        this._nOriginalX = coordinates.x;
        this._nOriginalY = coordinates.y;

        this._oDocument.on('mousemove touchmove', this._fMove);
        this._oDocument.one('mouseup touchend', this._fUp);

        if(this._oAnotherDocument) {
            this._oAnotherDocument.one('mouseup touchend', this._fUp);
            this._oAnotherDocument.on('mousemove touchmove', this._fMove);
        }


        this.lastEvent = oEvent;
        this.dragTarget = $(oEvent.currentTarget);

        clearTimeout( this._timeout );
        this._timeout = setTimeout( GoldenLayout.__lm.utils.fnBind( this._startDrag, this ), this._nDelay );
    },

    onMouseMove: function(oEvent)
    {
        oEvent.preventDefault();

        this.lastEvent = oEvent;

        var coordinates = this._getCoordinates( oEvent );
        //console.log(this.object_id, coordinates, this.currentFrameId)

        this._nX = coordinates.x - this._nOriginalX;
        this._nY = coordinates.y - this._nOriginalY;

        if( this._bDragging === false ) {
            if(
                Math.abs( this._nX ) > this._nDistance ||
                Math.abs( this._nY ) > this._nDistance
            ){
                clearTimeout( this._timeout );
                this._startDrag(oEvent);
            }
        }

        this.currentWindow = oEvent.view;

        if( this._bDragging )
        {
            if(this._nX !== this._prevX || this._nY !== this._prevY) {
                this._prevX = this._nX;
                this._prevY = this._nY;
                this.emit('drag', this._nX, this._nY, oEvent);
            }
        }
    },

    onMouseUp: function(oEvent)
    {
        clearTimeout( this._timeout );
        this._eBody.removeClass( 'lm_dragging' );
        this._oDocument.unbind( 'mousemove touchmove', this._fMove);
        this._oDocument.unbind( 'drag', this._fMove);

        if(this._oAnotherDocument) {
            this._oAnotherDocument.unbind( 'mousemove touchmove', this._fMove);
        }

        var _this = this;

        pgGetAppWindows().forEach(function(win) {
            if(win.window !== oEvent.view || true) {
                $(win.window).off('dragover', _this._fDragOver);
            }
        })

        if( this._bDragging === true )
        {
            this._bDragging = false;
            this.emit('dragStop', oEvent, this._nOriginalX + this._nX);
        }
        if(this.dragGhost) {
            $(this.dragGhost).remove();
            this.dragGhost = null;
        }
    },

    _startDrag: function(oEvent)
    {
        this._bDragging = true;
        this._eBody.addClass( 'lm_dragging' );
        this.emit('dragStart', this._nOriginalX, this._nOriginalY, this.lastEvent);
    },

    _getCoordinates: function( event ) {
        var coordinates = {};

        if( event.type.substr( 0, 5 ) === 'touch' ) {
            coordinates.x = event.originalEvent.targetTouches[ 0 ].pageX;
            coordinates.y = event.originalEvent.targetTouches[ 0 ].pageY;
        } else {
            coordinates.x = event.pageX;
            coordinates.y = event.pageY;
        }

        return coordinates;
    }
});





var PgTreeNodeMenu = function(menuItems) {

    var $menuContainer = $('<div class="pg-tree-menu-container"></div>');
    var $menu = $('<div class="pg-tree-menu"></div>');//.appendTo($menuContainer);

    var $door = $('<div class="pg-tree-menu-door"></div>').appendTo($menuContainer);

    this.$element = $menu;//$menuContainer;
    this.$menu = $menu;

    var currentData;

    var hidden = false;

    this.width = 32;

    var menuTop = 0;

    this.isYAlignedWithMenu = function(y, tolerance) {
        if(tolerance === undefined) tolerance = 0;
        return (y >= menuTop - tolerance && y <= menuTop + height + tolerance);
    }


    for(var i = 0; i < menuItems.length; i++) {
        var code = menuItems[i].icon.startsWith('<') ? menuItems[i].icon : '<i class="' + menuItems[i].icon + '"></i>';
        var $item = $('<div data-menu-id="' + i + '">' + code + '</div>').data('menu-item', menuItems[i]);
        $menu.append($item);
        if(menuItems[i].name) {
            addTooltip($item, menuItems[i].name, {placement: 'left'});
        }
        menuItems[i].$item = $item;
    }

    $menu.on('click', '>div', function(e) {
        var m = menuItems[parseInt($(e.currentTarget).attr('data-menu-id'))];
        m.func(currentData, $(this));
    })

    var height = null;
    var width = null;
    var container_height = null;
    var left = null;

    this.find = function(key) {
        for(var i = 0; i < menuItems.length; i++) {
            if(menuItems[i].key && menuItems[i].key == key) {
                return $menu.find('> [data-menu-id="' + i + '"]');
            }
        }
        return null;
    }

    this.hide = function() {
        if(!hidden) {
            $menu.hide();
            //$menuContainer.hide();
            hidden = true;
        }
    }

    this.show = function() {
        if(hidden /*&& !pgf.kids*/) {
            $menu.show();
            //$menuContainer.show();
            hidden = false;
        }
    }

    this.isHidden = function() {
        return hidden;
    }

    this.updateLayout = function() {
        //container_height = $menuContainer.height();
        container_height = $menu.parent().height();
        width = $menu.width();
        left = $menu.parent().width() - width;
    }

    this.showFor = function(top, bottom, selected, data) {

        currentData = data;

        //update visibility
        var ctx = {} //let show function share info for example data -> pgel
        for(var i = 0; i < menuItems.length; i++) {
            var show = true;
            if(menuItems[i].show) {
                show = menuItems[i].show(data, ctx);
            }
            show ? pg$Show(menuItems[i].$item) : pg$Hide(menuItems[i].$item);
        }

        //if(!height) height = $menu.height();

        height = $menu.height();

        if(hidden) {
            this.show();
        }

        var min = 38;

        if(pgf.kids) min = -8;

        var mid = (top + bottom) / 2 + min;

        var sy = Math.round(mid - height / 2);

        if(sy < min) sy = min;

        if(sy + height > (container_height - min)) {
            sy = (container_height - min) - height;
        }
        $menu.css({
            top: sy + 'px'
        })
        /* $door.css({
         top: top + 'px',
         height: (bottom - top) + 'px'
         })*/
        menuTop = sy;

        this.setSelected(selected);
    }

    this.showForVert = function(top, data, so_top) {

        currentData = data;

        if(!height) height = $menu.height();

        if(hidden) {
            this.show();
        }

        var sy = top - height + 4;

        if(sy < so_top) sy = so_top;

        if(sy + height > container_height) {
            //     sy = container_height - height;
        }
        $menu.css({
            top: sy + 'px'
        })
        /*  $door.css({
         top: top + 'px',
         height: (bottom - top) + 'px'
         })*/
        menuTop = sy;

        //this.setSelected(selected);
    }


    this.showInCell = function(cell) {

        currentData = cell.data;

        if(!height) height = $menu.height();

        cell.$element.append($menu);

        if(hidden) {
            this.show();
        }

        var sy = - height + 4;

        $menu.css({
            top: sy + 'px'
        })
        /*  $door.css({
         top: top + 'px',
         height: (bottom - top) + 'px'
         })*/
        menuTop = sy;

        //this.setSelected(selected);
    }

    this.isOverMenu = function(pos) {
        return !hidden && pos.left >= left && pos.top >= menuTop && pos.top <= menuTop   + height;
    }

    this.setSelected = function(selected) {
        if(selected) {
            $menu.addClass('for-selected');
            $menuContainer.addClass('for-selected');
        } else {
            $menu.removeClass('for-selected');
            $menuContainer.removeClass('for-selected');
        }
    }

    this.setBottom = function(b) {
        //$menu.css('bottom', b + 'px');
    }

    this.setRight = function(b) {
        $menu.css('right', b + 'px');
    }
}